#include <stdarg.h>

#include "libcad3d.h"


AXES::AXES (double Size, int FontBase /* = -1 */) :
  CAD_FILE (),
  UNIT     (),
  Size     (Size),
  FontBase (FontBase)
{
  Id = CS_ID;

  if (FontBase == -1)
    this->FontBase = OpenGL->GetFontBase ();

  Create ();
}


AXES::AXES (double Size, Display *Displ,
	    const char *FontSpec /* = "fixed" */, ...) :
  CAD_FILE (),
  UNIT     (),
  Size     (Size)
{
  char    FontName[256];
  va_list Arguments;

  va_start (Arguments, FontSpec);
  vsprintf (FontName, FontSpec, Arguments);
  va_end   (Arguments);

  FontBase = OpenGL->MakeFont (FontName);
  Id = CS_ID;

  Create ();
}


AXES::~AXES ()
{
  PRO_METHOD ("AXES~", "")

  PRO_END
}


bool AXES::Read (DATAFILE &File, double Version)
{
  PRO_METHOD ("COLOR", "\n")

  if (File.Parse("<\\[AXES\\]>\n") < 0)
    return FALSE;

  if (! CAD_FILE::Read(File, Version))
    return FALSE;

  if (File.Parse("  <Size:>    (%f)\n", &Size) < 1)
    return FALSE;

  Create ();

  PRO_END
  return TRUE;
}


bool AXES::Write (FILE *fp /* = NULL */, int Indent /* = 0 */) const
{
  PRO_METHOD ("AXES", "\n")

  if (fp == NULL)
    fp = stdout;

  Space   (fp, Indent);
  fprintf (fp, "[AXES]\n");
  Indent += INDENT;

  if (! CAD_FILE::Write(fp, Indent))
    PRO_RETURN ("Writing failed", FALSE)

  Space   (fp, Indent);
  fprintf (fp, "Size:   %.3f\n", Size);

  PRO_END
  return TRUE;
}


void AXES::Draw () const
{
  PRO_METHOD ("AXES", "")

  if (Id >= 0)
  {
    glMatrixMode (GL_MODELVIEW);
    glCallList (Id);
    glFlush ();
  }

  PRO_END
}


void AXES::Draw (H_MATRIX Matrix) const
{
  PRO_METHOD ("AXES", "")

  if (Id >= 0)
  {
    glPushMatrix ();
    glMatrixMode (GL_MODELVIEW);

    glLoadMatrixd ((GLdouble *)Matrix.m);
    glCallList (Id);
    glPopMatrix ();
    glFlush ();
  }
  
  PRO_END
}


void AXES::SetSize (double Size)
{
  this->Size = Size;
  Create ();
}


bool AXES::Create ()
{
  COLOR Color;

  PRO_METHOD ("AXES", "")

  Id = OpenGL->GetID (Id, "AXES");
  glNewList (Id, GL_COMPILE);
  glIndexi  (Id);

  glDisable (GL_DEPTH_TEST);
  glEnable  (GL_COLOR_MATERIAL);
  glDisable (GL_LIGHTING);

  Color.SetID (COLOR::MEDRED);
  Color.Set ();
  glBegin (GL_LINES);
    glVertex3f ( 0.0, 0.0, 0.0);
    glVertex3f (Size, 0.0, 0.0);
  glEnd ();
  if (FontBase >= 0)
    Print (Size, 0.0, 0.0, "x");

  Color.SetID (COLOR::MEDGREEN);
  Color.Set ();
  glBegin (GL_LINES);
    glVertex3f (0.0,  0.0, 0.0);
    glVertex3f (0.0, Size, 0.0);
  glEnd ();
  if (FontBase >= 0)
    Print (0.0, Size, 0.0, "y");

  Color.SetID (COLOR::MEDBLUE);
  Color.Set ();
  glBegin (GL_LINES);
    glVertex3f (0.0, 0.0, 0.0);
    glVertex3f (0.0, 0.0, Size);
  glEnd ();
  if (FontBase >= 0)
    Print (0.0, 0.0, Size, "z");

  glEnable  (GL_LIGHTING);
  glDisable (GL_COLOR_MATERIAL);
  glEnable  (GL_DEPTH_TEST);

  glEndList ();

  PRO_END
  return TRUE;
}


void AXES::Print (double x, double y, double z, const char *Msg)
{
  PRO_METHOD ("AXES", "")

  glRasterPos3d (x, y, z);
  glPushAttrib  (GL_LIST_BIT);
  glListBase    (FontBase);
  glCallLists   (strlen(Msg), GL_UNSIGNED_BYTE, (GLubyte *)Msg);
  glPopAttrib   ();

  PRO_END  
}
