#include <assert.h>
#include <math.h>

#include "libcad3d.h"


static POINT3D Corner[8] =
{
  {-0.5, -0.5, -0.5},
  { 0.5, -0.5, -0.5},
  {-0.5,  0.5, -0.5},
  { 0.5,  0.5, -0.5},
  {-0.5, -0.5,  0.5},
  { 0.5, -0.5,  0.5},
  {-0.5,  0.5,  0.5},
  { 0.5,  0.5,  0.5}
};

static double Dir[3][3] =
{
  {1.0, 0.0, 0.0},
  {0.0, 1.0, 0.0},
  {0.0, 0.0, 1.0}
};



BOX::BOX (double SizeX, double SizeY, double SizeZ,
	  H_MATRIX Center/* = Matrix_1 */) :
  CAD_FILE (),
  UNIT     (),
  Len      (1.0),
  PTB      (Center),
  Color    (COLOR::GOLD)
{
  PRO_METHOD ("BOX", "")

  SetSize (SizeX, SizeY, SizeZ);

  PRO_END
}


BOX::BOX (double MinX, double MaxX,
	  double MinY, double MaxY, 
	  double MinZ, double MaxZ) :
  CAD_FILE (),
  UNIT     (),
  Len      (1.0),
  Color    (COLOR::GOLD)
{
  PRO_METHOD ("BOX", "")

  SetRange (MinX, MaxX, MinY, MaxY, MinZ, MaxZ);

  PRO_END
}


BOX::~BOX ()
{
  PRO_METHOD ("~BOX", "")

  PRO_END
}


BOX &BOX::operator = (const BOX &Box)
{
  if (this != &Box)
  {
    Len   = Box.Len;
    Color = Box.Color;
    PTB   = Box.PTB;
    Size  = Box.Size;

    Create ();
  }
  return *this;
}


bool BOX::Read (DATAFILE &File, double Version)
{
  PRO_METHOD ("BOX", "\n")

  if (File.Parse("<\\[BOX\\]>\n") < 0)
    PRO_RETURN ("Key '[BOX]' not found", FALSE)

  if (! CAD_FILE::Read(File, Version))
    PRO_RETURN ("Reading FILE failed", FALSE)

  if (File.Parse("  <Size:>    (%f) x (%f) x (%f)\n"
		 "  <Length:>  (%f)\n",
		 &Size.x, &Size.y, &Size.z,
		 &Len) < 4)
    PRO_RETURN ("Reading data 1 failed", FALSE)

  // TODO: PTB

  if (! Color.Read(File, Version))
    PRO_RETURN ("Reading data 2 failed", FALSE)

  Create ();

  PRO_END
  return TRUE;
}


bool BOX::Write (FILE *fp /* = NULL */, int Indent /* = 0 */) const
{
  PRO_METHOD ("BOX", "\n")

  if (fp == NULL)
    fp = stdout;

  Space   (fp, Indent);
  fprintf (fp, "[BOX]\n");
  Indent += INDENT;

  if (! CAD_FILE::Write(fp, Indent))
    PRO_RETURN ("Writing FILE failed", FALSE)

  Space   (fp, Indent);
  fprintf (fp, "Size:    %.2f x %.2f x %.2f\n", Size.x, Size.y, Size.z);

  Space   (fp, Indent);
  fprintf (fp, "Length:  %.3f\n", Len);

  // TODO: PTB

  if (! Color.Write(fp))
    PRO_RETURN ("Writing data failed", FALSE)

  PRO_END
  return TRUE;
}


bool BOX::Create ()
{
  int     count1, count2;
  double  Length;
  POINT3D Point;

  PRO_METHOD ("BOX", "\n")

  Id = OpenGL->GetID (Id, "BOX");
  glNewList (Id, GL_COMPILE);
  glIndexi  (Id);

  glEnable  (GL_COLOR_MATERIAL);
  glDisable (GL_LIGHTING);
  Color.Set ();

  glPushMatrix ();

  // TODO: wird noch nicht benutzt: (alle Achsen gleiche Lnge)
  Length = 0.05 * MIN (Size.x, MIN(Size.y, Size.z));

  glMultMatrixd ((GLdouble *)PTB.m);
  glBegin (GL_LINES);
    for (count1 = 0; count1 < 8; ++count1)
    {
      Point.x = Corner[count1].x * Size.x;
      Point.y = Corner[count1].y * Size.y;
      Point.z = Corner[count1].z * Size.z;

      for (count2 = 0; count2 < 3; ++count2)
      {
	glVertex3f (Point.x, Point.y, Point.z);
	glVertex3f ((1.0 - Dir[count2][X]*Len) * Point.x,
		    (1.0 - Dir[count2][Y]*Len) * Point.y,
		    (1.0 - Dir[count2][Z]*Len) * Point.z);
      }
    }
  glEnd ();

  glPopMatrix ();

  glEnable    (GL_LIGHTING);
  glDisable   (GL_COLOR_MATERIAL);

  glEndList ();

  PRO_END
  return TRUE;
}


void BOX::Place (H_MATRIX Center)
{
  PTB = Center;
  Create ();
}


void BOX::SetRange (double MinX, double MaxX, 
		    double MinY, double MaxY, 
		    double MinZ, double MaxZ)
{
  PRO_METHOD ("BOX", "\n")

  Size.x = fabs (MaxX - MinX);
  Size.y = fabs (MaxY - MinY);
  Size.z = fabs (MaxZ - MinZ);

  PTB = GEOMakeTrans (0.5*(MinX+MaxX), 0.5*(MinY+MaxY), 0.5*(MinZ+MaxZ));

  Create ();

  PRO_END
}


void BOX::SetSize (double SizeX, double SizeY, double SizeZ)
{
  Size.x = fabs (SizeX);
  Size.y = fabs (SizeY);
  Size.z = fabs (SizeZ);

  Create ();
}


void BOX::SetLen (double Length /* = 0.2 */)
{
  if (Length != Len)
  {
    Len = Length;
    Create ();
  }
}


void BOX::SetColor (COLOR Color /* = COLOR(COLOR::GOLD) */)
{
  this->Color = Color;
  Create ();
}


POINT3D BOX::GetSize () const
{
  return Size;
}


// Values are relative to part due to PTB lies in center of part
void BOX::GetRange (double *MinX, double *MaxX,
		    double *MinY, double *MaxY,
		    double *MinZ, double *MaxZ)
{
  int     count;
  POINT3D Point, Min, Max;

  Point.x = Corner[0].x * Size.x;
  Point.y = Corner[0].y * Size.y;
  Point.z = Corner[0].z * Size.z;
  Min = Max = GEOMatrixMult (PTB, Point);

  for (count = 1; count < 8; ++count)
  {
    Point.x = Corner[count].x * Size.x;
    Point.y = Corner[count].y * Size.y;
    Point.z = Corner[count].z * Size.z;
    Point = GEOMatrixMult (PTB, Point);
    
    Min.x = MIN (Min.x, Point.x);
    Max.x = MAX (Max.x, Point.x);
    Min.y = MIN (Min.y, Point.y);
    Max.y = MAX (Max.y, Point.y);
    Min.z = MIN (Min.z, Point.z);
    Max.z = MAX (Max.z, Point.z);
  }

  if (MinX != NULL)
    *MinX = Min.x;
  if (MaxX != NULL)
    *MaxX = Max.x;
  if (MinY != NULL)
    *MinY = Min.y;
  if (MaxY != NULL)
    *MaxY = Max.y;
  if (MinZ != NULL)
    *MinZ = Min.z;
  if (MaxZ != NULL)
    *MaxZ = Max.z;
}


double BOX::GetLen () const
{
  return Len;
}


COLOR &BOX::GetColor ()
{
  return Color;
}


