#include "libcad3d.h"


static PART3D *DEVICE_GetMount (DEVICE3D *DevList, int &No);


DEVICE3D::DEVICE3D (const char   *DevName /* = NULL */,
		          PART3D *Part    /* = NULL */,
 		          WORLD  *World   /* = NULL */) :	
  CAD_FILE (),
  UNIT     (),
  PrgName  (NULL),
  ConfigNo (0),
  KinType  (PART3D::TRANS | PART3D::ROT),
  MountNo  (0),
  Offset   (Matrix_1),
  HomeNo   (0),
  Home     (NULL),
  Alpha    (1.0),
  Robot    (NULL),
  WTD      (Matrix_1),
  DOF      (6),
  State    (STATE_NORMAL),
  PartList (Part),
  Fix      (NULL),
  Next     (NULL),
  World    (World)
{
  PRO_METHOD ("DEVICE3D", "\n")

  No = 0;
  if (DevName != NULL)
  {
    Name = new char[strlen(DevName)+1];
    strcpy (Name, DevName);
  }
  else
    Name = NULL;

  if (Part != NULL)
  {
    PartNo   = 1;
    Part->SetParent (this);
  }
  else
    PartNo = 0;

  PRO_END
}


DEVICE3D::DEVICE3D (DEVICE3D *Dev, WORLD *World /* = NULL */) :
  CAD_FILE (),
  UNIT     (),
  Fix      (NULL),
  Next     (NULL),
  World    (World)
{
  PART3D *Ptr;

  PRO_METHOD ("DEVICE3D", "\n")

  if (Dev == NULL)
    throw ERROR (ERROR::NO_DATA);

  if (Dev != this)
  {
    ConfigNo = Dev->ConfigNo;
    KinType  = Dev->KinType;
    MountNo  = Dev->MountNo;
    Offset   = Dev->Offset;
    HomeNo   = Dev->HomeNo;
    if (HomeNo > 0 && Dev->Home != NULL)
    {
      Home = new HOME_DATA[HomeNo];
      memcpy (Home, Dev->Home, sizeof(HOME_DATA) * HomeNo);
    }
    else
      Home = NULL;

    Alpha = Dev->Alpha;
    Robot = Dev->Robot; // kopieren ?
    WTD   = Dev->WTD;
    DOF   = Dev->DOF;
    State = Dev->State;
    No    = Dev->No;
    if (Dev->Name != NULL)
    {
      Name = new char[strlen(Dev->GetName())+1];
      strcpy (Name, Dev->GetName());
    }
    else
      Name = NULL;

    PartNo = 0;
    PartList = NULL;

    Ptr = Dev->GetPartList ();
    while (Ptr != NULL)
    {
      AddPart (Ptr->Copy(this));
      Ptr = Ptr->GetNext ();
    }
  }

  PRO_END
}


DEVICE3D::~DEVICE3D ()
{
  PART3D *Part, *Chain;
  PART3D *Ptr, *Next;

  PRO_METHOD ("DEVICE3D~", "\n")

  if (Robot != NULL)
    delete Robot;

  Part = PartList;
  while (Part != NULL)
  {
    Chain = Part->GetChain ();

    Ptr = Part->GetNext ();
    while (Ptr != NULL)
    {
      Next = Ptr->GetNext ();
      delete (Ptr);
      Ptr = Next;
    }

    if (Part->GetAttach() != NULL)
      delete (Part->GetAttach());

    delete (Part);
    Part = Chain;
  }

  if (Name != NULL)
    delete Name;

  PRO_END
}


DEVICE3D *DEVICE3D::Copy ()
{
  PART3D   *Ptr, *Copy;
  DEVICE3D *Result = new DEVICE3D (Name);

  Ptr = PartList;
  while (Ptr != NULL)
  {
    Copy = Ptr->Copy (this); 
    Result->AddPart (Copy);
    Ptr = Ptr->GetNext ();
  }

  return Result;
}


static bool DEVICE_Create (PART3D *PartList, COLOR Color)
{  
  PART3D   *Ptr, *Chain;
  DEVICE3D *Attach;

  Ptr = PartList;
  while (Ptr != NULL)
  {
    Ptr->Create (Color);

    if ((Chain = Ptr->GetChain()) != NULL)
      if (! DEVICE_Create(Chain, Color))
	return FALSE;

    if ((Attach = Ptr->GetAttach()) != NULL)
      if (! Attach->Create(Color))
	return FALSE;

    Ptr = Ptr->GetNext ();
  }
  return TRUE;
}


bool DEVICE3D::Create ()
{
  return Create (COLOR(COLOR::OBJECT));
}


bool DEVICE3D::Create (COLOR Color)
{
  return DEVICE_Create (this->GetPartList(), Color);
}


bool DEVICE3D::Write (FILE *fp /* = NULL */, int Indent /* = 0 */)
{
  PART3D *Ptr;

  PRO_METHOD ("DEVICE3D", "\n")

  if (fp == NULL)
    fp = stdout;

  Space   (fp, Indent);
  fprintf (fp, "[DEVICE]\n");
  Indent += INDENT;

  if (! CAD_FILE::Write(fp, Indent))
    PRO_RETURN ("Writing failed", FALSE)
  
  Space   (fp, Indent);
  fprintf (fp, "Number:      %d\n", No);

  Space   (fp, Indent);
  fprintf (fp, "Name:        %s\n", Name);

  // TODO: ProgramName 

  Space   (fp, Indent);
  fprintf (fp, "Mount:       %d\n", MountNo);

  Space   (fp, Indent);
  fprintf (fp, "Offset:      \n");

  Space   (fp, Indent);
  fprintf (fp, "Homes:       %d\n", HomeNo);

  Space   (fp, Indent);
  fprintf (fp, "Kiniematic:  \n");

  Space   (fp, Indent);
  fprintf (fp, "WTD:         \n");

  Space   (fp, Indent);
  fprintf (fp, "DOF:         %d\n", DOF);

  Space   (fp, Indent);
  fprintf (fp, "Parts:       %d\n", GetPartNumber());

  Ptr = PartList;
  while (Ptr != NULL)
  {
    Ptr->Write (fp, Indent+2);
    fprintf (fp, "\n");
    Ptr = Ptr->GetNext ();
  }

  Space   (fp, Indent);
  fprintf (fp, "[Affix]\n");

  return TRUE;
}


bool DEVICE3D::Save (const char *FileName)
{
  bool  Result;
  FILE *fp;

  if ((fp = fopen(FileName, "w+")) == NULL)
    return FALSE;

  Result = Write (fp);
  fclose (fp);

  return Result;
}


int DEVICE3D::GetPartNumber (int Mask /* = 0 */)
{
  int     Result = 0;
  PART3D *Ptr;

  Ptr = PartList;
  while (Ptr != NULL)
  {
    Result += Ptr->GetPartNumber (Mask);
    Ptr = Ptr->GetNext ();
  }
  
  if ((Mask & NEXT_MASK) && Next != NULL)
    Result += Next->GetPartNumber (Mask);

  return Result;
}


int DEVICE3D::GetDevNumber (int Mask /* = 0 */)
{
  int Result = 1;

  Result += PartList->GetDevNumber (Mask);

  if ((Mask & NEXT_MASK) && Next != NULL)
    Result += Next->GetDevNumber (Mask);

  return Result;
}


void DEVICE3D::SetName (char *DevName)
{
  if (Name != NULL)
    delete Name;

  if (DevName != NULL)
  {
    Name = new char[strlen(DevName)+1];    
    strcpy (Name, DevName);
  }
}


void DEVICE3D::SetPos (H_MATRIX M)
{
  GetPartList()->GetPart()->WTP = M;
}


PART3D *DEVICE3D::GetPart (const char *PartName)
{
  PART3D *Ptr;
  PART3D *Result;

  Ptr = PartList;
  while (Ptr != NULL)
  {
    if ((Result = Ptr->GetPart(PartName)) != NULL)
      return Result;
    Ptr = Ptr->GetNext();
  }

  return NULL;
}


#if 1
// TODO: Gehrt in WORLD: Hier nur fr ein einziges Device mit seinen Parts
void DEVICE3D::AddDevice (DEVICE3D *Dev)
{
  DEVICE3D *Ptr;

  PRO_METHOD ("DEVICE3D", "")

  if (Dev == NULL)
    PRO_ABORT ("No device")

  if ((Ptr = Next) == NULL)             // No other device present
    Next = Dev;
  else                                  // Add at end of list
  {
    while (Ptr->Next != NULL)
      Ptr = Ptr->Next;
    Ptr->Next = Dev;
  }

  // Update references of the device and compute new positions
  Dev->Fix = Fix;
  Dev->UpdatePos ();

  PRO_END
}
#endif


void DEVICE3D::AddPart (PART3D *Part, PART3D *Prev /* = NULL */)
{
  PART3D *Ptr;

  PRO_METHOD ("DEVICE3D", "")

  if (Part == NULL)
    PRO_ABORT ("No Part")

  Part->SetParent (this);

  if (Prev == NULL)                     // Add part to the device's partlist
    if ((Ptr = PartList) == NULL)       // list is empty
      PartList = Part;
    else                                // append at end of list
    {
      while (Ptr->GetNext() != NULL)
	Ptr = Ptr->GetNext ();
      Ptr->SetNext (Part);
    }
  else                                  // Add part to chain-list
    Prev->AddChain (Part);

  Part->UpdatePos ();

  while (Part != NULL)                  // Update part counter
  {
    PartNo++;
    if (World != NULL)
      World->IncPartNum ();
    Part = Part->GetNext ();
  }

  PRO_END
}    


bool DEVICE3D::DetachPart (PART3D *Part)
{
  PART3D *Ptr;

  PRO_METHOD ("DEVICE3D", "")

  if (Part == NULL)
    PRO_RETURN ("No part", TRUE)

  if ((Ptr = PartList) == Part)         // Detach first part
  {
    PartList = Ptr->GetNext();
    PartNo--;
  }
  else
  {
    while (Ptr->GetNext() != NULL && Ptr->GetNext() != Part)
      Ptr = Ptr->GetNext();
    if (Ptr->GetNext() == Part)
    {
      Ptr->SetNext(Ptr->GetNext()->GetNext());
      PartNo--;
    }
  }

  Part->SetParent (NULL);
  // TODO: Mte fr das Part ein Device erzeugt werden ?

  PRO_END
  return (Ptr != NULL);
}


H_MATRIX DEVICE3D::GetTCP ()
{
  PART3D   *Mount;
  H_MATRIX  Result = Matrix_1;

  PRO_METHOD ("DEVICE3D", "\n")

  if ((Mount = GetMount()) != NULL)
    Result = GEOMatrixProd (Mount->GetWTP(), Offset);

  PRO_END
  return Result;
}


H_MATRIX DEVICE3D::GetWTD ()
{
  return WTD;
}


static PART3D *DEVICE_GetMount (PART3D *PartList, int &No)
{
  PART3D   *Result = NULL;
  PART3D   *Chain, *Ptr;
  DEVICE3D *Attach;
  
  Ptr = PartList;
  while (Ptr != NULL && No >= 0)
  {
    if (No == 0)
      return Ptr;

    No--;

    if ((Chain = Ptr->GetChain()) != NULL)
      Result = DEVICE_GetMount (Chain, No);

    if ((Attach = Ptr->GetAttach()) != NULL)
      Result = DEVICE_GetMount (Attach, No);

    Ptr = Ptr->GetNext ();
  }

  return Result;
}

static PART3D *DEVICE_GetMount (DEVICE3D *DevList, int &No)
{
  PART3D   *Result = NULL;
  DEVICE3D *Dev;

  Dev = DevList;
  while (Dev != NULL && Result == NULL)
  {
    Result = DEVICE_GetMount (Dev->GetPartList(), No);
    Dev = Dev->GetNext ();
  }
  return Result;
}

PART3D *DEVICE3D::GetMount ()
{
  int     No = 0;
  PART3D *Mount;

  PRO_METHOD ("DEVICE3D", "")

  PRO1 ("\\MountNo: %d...", MountNo)

  // Search the MountNo-th part in the device
  No = MountNo;         // 2nd parameter will be changes, do not use MountNo
  Mount = DEVICE_GetMount (PartList, No);
  if (Mount != NULL)
    PRO1 ("Mount-Part: %s\n", Mount->GetName());

  PRO_END
  return Mount;
}


void DEVICE3D::GetRange (double *MinX, double *MaxX, 
			 double *MinY, double *MaxY,
			 double *MinZ, double *MaxZ)
{
  POINT3D  Min = Vector_0, Max = Vector_0;
  PART3D  *Ptr;

  Ptr = PartList;
  while (Ptr != NULL)
  {
    Ptr->GetRange (&Min.x, &Max.x, &Min.y, &Max.y, &Min.z, &Max.z,
		   CHAIN_MASK | ATTACH_MASK);

    Ptr = Ptr->GetNext();
  }
}

void DEVICE3D::Scale (double ScaleX, double ScaleY, double ScaleZ,
		      int Mask /* = 0 */)
{
  PART3D *Part;

  PRO_METHOD ("DEVICE3D", "\n")

  if ((Part = GetPartList()) != NULL)
    Part->Scale (ScaleX, ScaleY, ScaleZ, ALL_MASK);

  if (Mask & NEXT_MASK && Next != NULL)
    Next->Scale (ScaleX, ScaleY, ScaleZ, Mask);

  PRO_END
}


void DEVICE3D::Place (H_MATRIX WTD)
{
  PART3D *Part;

  // TODO: berlegen: korrekt, wenn alle Parts auf WTD zu liegen kommen ?

  Part = PartList;
  while (Part != NULL)
  {
    Part->Place (WTD);
    Part = Part->GetNext ();
  }
}


void DEVICE3D::UpdatePos ()
{
  if (Fix != NULL)
    UpdatePos (Fix->GetWTP());
  else
    UpdatePos (WTD);
}


void DEVICE3D::UpdatePos (H_MATRIX WTD)
{
  PART3D *Ptr;

  Ptr = PartList;
  while (Ptr != NULL)
  {
    Ptr->UpdatePos (WTD);
    Ptr = Ptr->GetNext ();
  }
}


void DEVICE3D::SetColorID (COLOR::ID ColorID /* = OBJECT */)
{
  PART3D *Next;

  Next = PartList;
  while (Next != NULL)
  {
    Next->SetColorID (ColorID);
    Next = Next->GetNext ();
  }
}


void DEVICE3D::SetColor (COLOR Color /* = COLOR(COLOR::OBJECT_COLOR)*/)
{
  PART3D *Next;

  Next = PartList;
  while (Next != NULL)
  {
    Next->SetColor (Color);
    Next = Next->GetNext ();
  }
}


int DEVICE3D::GetRange (int No, double &Min, double &Val, double &Max)
{
  int     count = 0;
  PART3D *Part;

  Part = PartList;
  while (count < No && Part != NULL)
    Part = Part->GetChain (); 

  if (Part == NULL)
    return -1;

  JOINT Joint = Part->GetJoint ();

  count = POS_X;
  while (count <= ROT_Z && ! Joint.IsFree (count))
    count++;
  if (count <= ROT_Z)
  {
    Min = Joint.GetMin (count);
    Max = Joint.GetMax (count);
  }
  Val = Part->GetVal (0);

  return 1;
}


static int DEVICE_SetJoint (PART3D *PartList, double *Joint, int No, 
			                      bool first)
{
  int       count = 0;
  PART3D   *Next, *Ptr;
  DEVICE3D *Attach;

  Next = PartList;
  while (Next != NULL && count < No)
  {
    if (! first && count < No)
      count += Next->SetJoint (Joint+count);
    else
      first = FALSE;  // skip base-part

    if ((Ptr = Next->GetChain()) != NULL)
      count += DEVICE_SetJoint (Ptr, Joint+count, No-count, FALSE);

    if ((Attach = Next->GetAttach()) != NULL)
      count += Attach->SetJoint (Joint+count, No-count, TRUE);

    Next = Next->GetNext ();
  }
  return count;
}

int DEVICE3D::SetJoint (double *Joint, int No, bool attached /* = FALSE */)
{
  int       count = 0;
  DEVICE3D *Dev;

  Dev = this;
  while (Dev != NULL && count < No)
  {
    count += DEVICE_SetJoint (Dev->GetPartList(), 
			      Joint+count, No-count, TRUE); 
 
    if (! attached)
      return count;

    Dev = Dev->GetNext();
  }
  return count;
}

static void DEVICE_SetAllJoints (PART3D *PartList, double Value)
{
  int       count;
  double    Min, Max;
  double    Joints[6];
  PART3D   *Next, *Ptr;
  DEVICE3D *Attach;
  JOINT     Joint (0);

  Next = PartList;
  while (Next != NULL)
  {
    Joint = Next->GetJoint ();
    if (Joint.GetAxes() > 0)
    {
      memset (Joints, 0, sizeof(double)*6);
      for (count = POS_X; count <= ROT_Z; ++count)
        if (Joint.IsFree(count))
	{
	  Min = Joint.GetMin (count);
	  Max = Joint.GetMax (count);
	  Joints[count] = Min + (0.5 * (Value+1.0) * (Max - Min));
	}
      Next->SetJoint (Joints);
    }

    if ((Ptr = Next->GetChain()) != NULL)
      DEVICE_SetAllJoints (Ptr, Value);

    if ((Attach = Next->GetAttach()) != NULL)
      Attach->SetAllJoints (Value);

    Next = Next->GetNext ();
  }
}


/* Set all joints of the device to a value, which is relative to the 
   Joints range (-1.0 correspondents with the minimum joint-value, 
   1.0 correspondents with the maximum value) */
void DEVICE3D::SetAllJoints (double Value /* = 0.0 */)
{
  DEVICE3D *Dev;

  if (Value < -1.0)
    Value = -1.0;

  if (Value > 1.0)
    Value = 1.0;

  Dev = this;
  while (Dev != NULL)
  {
    DEVICE_SetAllJoints (Dev->GetPartList(), Value);

    Dev = Dev->GetNext ();
  }
}


static int DEVICE_GetJoints (PART3D *PartList, double **Joints, int No)
{
  int     Result;
  PART3D *Ptr;

  Result = No;
  if (PartList != NULL)
  {
    if ((Ptr = PartList->GetChain()) != NULL)
      Result = DEVICE_GetJoints (Ptr, Joints, No+1);
    elsif (Joints != NULL)
      *Joints = (double *)malloc (sizeof(double) * No);

    if (*Joints != NULL && No > 0)
      (*Joints)[No-1] = PartList->GetVal (5);
  }

  return Result;
}


int DEVICE3D::GetJoints (double **Joints)
{
  return DEVICE_GetJoints (this->GetPartList(), Joints, 0);
}


bool DEVICE3D::MoveTCP (H_MATRIX Rel)
{
  int       count;
  double    J[18];  // TODO: dynamisch
  PART3D   *Mount;
  JOINTS    Joint;
  H_MATRIX  RTT, Next;

  if (Robot == NULL)
    return FALSE;

  if ((Mount = GetMount()) == NULL)
    return FALSE;

  RTT = GEOMatrixProd (GEOMatrixInv(WTD), GetTCP());
  Next = GEOMatrixProd (RTT, Rel);

  try
  {
    Joint = Robot->InvKin (TO_FRAME(Next));
    for (count = 0; count < Joint.GetNumber(); ++count)
      J[count] = Joint[count];
    SetJoint (J, Joint.GetNumber());

    RTT = Next;
    RPY rpy = GEOGetRPY (RTT);
  }
  catch (KINEMATIC::OutOfWorkspace OOW)
  {
    return FALSE;
  }
  return TRUE;
}


char *DEVICE3D::GetName ()
{
  return Name; 
}


int DEVICE3D::GetConfigNo ()
{
  return ConfigNo; 
}


int DEVICE3D::GetMountNo ()
{
  return MountNo; 
}


H_MATRIX DEVICE3D::GetOffset ()
{ 
  return Offset;
}


double DEVICE3D::GetAlpha ()
{ 
  return Alpha; 
}


int DEVICE3D::GetHomeNo ()
{ 
  return HomeNo; 
}


int DEVICE3D::GetDOF ()                   // Calculate Degree Of Freedom
{
  int     No = 0;
  int     Axes;
  PART3D *Ptr;

  Ptr = PartList;
  while (Ptr != NULL)
  {
    Axes = Ptr->GetJoint().GetAxes ();
    if (Axes == 0 && Ptr->GetChain() == NULL)
      No += 6;
    else
      while (Axes != 0)
      {
	if (Axes & 1)
	  No++;
	Axes = Axes >> 1;
      }
    Ptr = Ptr->GetChain();
  }

  printf ("DOF %s: %d (%d)\n", Name, No, DOF);

  return DOF;
}


int DEVICE3D::GetState ()
{ 
  return State; 
}


int DEVICE3D::GetPartNo ()
{ 
  return PartNo; 
}


ROBOT *DEVICE3D::GetRobot () 
{ 
  return Robot;
}


PART3D *DEVICE3D::GetPartList ()
{ 
  return PartList; 
}


PART3D *DEVICE3D::GetFix ()
{
  return Fix; 
}


DEVICE3D *DEVICE3D::GetNext ()
{ 
  return Next; 
}


void DEVICE3D::SetConfigNo (int ConfigNo)
{
  this->ConfigNo = ConfigNo;
}


void DEVICE3D::SetMountNo (int MountNo)
{ 
  this->MountNo = MountNo;
}


void DEVICE3D::SetOffset (H_MATRIX Offset)
{ 
  this->Offset = Offset;
}


void DEVICE3D::SetHomeNo (int HomeNo)
{ 
  this->HomeNo = HomeNo;
}


void DEVICE3D::SetWTD (H_MATRIX WTD)
{
  this->WTD = WTD;
  UpdatePos ();
}


void DEVICE3D::SetDOF (int DOF)
{
  this->DOF = DOF;
}


void DEVICE3D::SetState (int State)
{ 
  this->State = State;
}


void DEVICE3D::SetPartNo (int PartNo)
{
  this->PartNo = PartNo;
}


void DEVICE3D::SetPart (PART3D *Part)
{
  this->PartList = Part;
}


void DEVICE3D::SetFix (PART3D *Fix)
{ 
  this->Fix = Fix;
}


void DEVICE3D::SetNext (DEVICE3D *Dev)
{
  this->Next = Dev;
}


void DEVICE3D::SetAlpha (double Alpha)
{
  this->Alpha = Alpha;
}


void DEVICE3D::SetRobot (ROBOT *Robot)
{
  this->Robot = Robot;
}


void DEVICE3D::IncPartNo (int Value)
{ 
  PartNo += Value;   
  if (Fix != NULL && Fix->GetParent() != NULL)
    Fix->GetParent()->IncPartNo (Value);
}


// This method does NOT change the parts pointer (Parent, Prev, Chain, 
// Attach, Next). All chained parts or attached devices will be removed also.
// The part is only removed from the list, in which it is stored. So, take
// care not to use the prev, parent or next pointer. They are not actual, but
// may be used to reattach the chain-parts or attached devices accordingly.
// It does not process the following devices
bool DEVICE3D::Remove (PART3D *Part)
{
  bool    Result;
  PART3D *Ptr; 

  if (Part == NULL)
    return TRUE;

  Ptr = PartList;
  if ((Result = (Ptr == Part)))               // Remove 1st part of list
    PartList = Part->GetNext ();
  else
    Result = Ptr->Remove (Part);

  while (! Result && Ptr->GetNext() != NULL)  // Search in part list
  {
    if ((Result = (Ptr->GetNext() == Part)))  // Part found...
      Ptr->SetNext (Part->GetNext());
    else
      Result = Ptr->Remove (Part);

    Ptr = Ptr->GetNext ();
  }

  if (Result)
  {
    PartNo -= Part->GetPartNumber (CHAIN_MASK | ATTACH_MASK);
    // TODO: Falls device attached wurde, PartNum in Fix aktualisieren
  }

  return Result;
}


bool DEVICE3D::Remove (DEVICE3D *Dev)
{
  bool    Result = FALSE;
  PART3D *Ptr;

  if (Dev == NULL)
    return TRUE;

  Ptr = PartList;
  while (! Result && Ptr != NULL)
  {
    Result = Ptr->Remove(Dev);
    Ptr = Ptr->GetNext ();
  }

  if (Result)
  {
    PartNo -= Dev->GetPartNumber (CHAIN_MASK | ATTACH_MASK);
    // TODO: Falls Device attached is, auch fix aktualisieren
  }

  return Result;
}


static void DEVICE_CalcDHP (PART3D *PartList, H_MATRIX Matr)
{
  PART3D *Ptr, *Chain;

  Ptr = PartList;
  while (Ptr != NULL)
  {
    double   Theta, d, a, Alpha;
    H_MATRIX M = Ptr->GetMatrix ();

    FRAME An (M.m[N][X], M.m[N][Y], M.m[N][Z], 
	      M.m[O][X], M.m[O][Y], M.m[O][Z], 
	      M.m[A][X], M.m[A][Y], M.m[A][Z], 
	      M.m[P][X], M.m[P][Y], M.m[P][Z]);
    An.GetDHP (&Theta, &d, &a, &Alpha);
    printf ("DHP %s: THETA=%7.2f D=%6.2f A=%6.2f ALPHA=%7.2f\n", 
	    Ptr->GetName(), Theta, d, a, Alpha);

    if ((Chain = Ptr->GetChain()) != NULL)
      DEVICE_CalcDHP (Chain, Matr);

    Ptr = Ptr->GetNext ();
  }
}

void DEVICE3D::CalcDHP ()
{
  DEVICE_CalcDHP (PartList, Matrix_1);
}
