#include <assert.h>

#include "libcad3d.h"


FLOOR::FLOOR (MODE Mode /* RASTER */, int LineWidth /* = 1 */) :
  CAD_FILE  (),
  UNIT      (),
  Mode      (Mode),
  SizeX     (10),
  SizeY     (10),
  SpaceX    (10.0),
  SpaceY    (10.0),
  Place     (RPY_0),
  Material  (COLOR::SILVER),
  LineWidth (LineWidth),
  Grid      (COLOR::BLACK),
  ID        (-1)
{
  PRO_METHOD ("FLOOR", "\n")

  Create ();

  PRO_END
}
  

FLOOR::~FLOOR ()
{
  PRO_METHOD ("FLOOR~", "")

  OpenGL->FreeID (ID);

  PRO_END
}


double &FLOOR::operator [] (const int &i)
{
  assert (i >= PX &&  i <= RZ);

  switch (i)
  {
    case PX : return Place.px;
    case PY : return Place.py;
    case PZ : return Place.pz;
    case RX : return Place.y;
    case RY : return Place.p;
    case RZ : return Place.r;
    default : printf ("invalid operator");
  }

  return Place.px;
}


double FLOOR::operator [] (const int &i) const
{
  assert (i >= PX &&  i <= RZ);

  switch (i)
  {
    case PX : return Place.px;
    case PY : return Place.py;
    case PZ : return Place.pz;
    case RX : return Place.y;
    case RY : return Place.p;
    case RZ : return Place.r;
    default : printf ("invalid operator");
  }
  return 0.0;
}    



bool FLOOR::Read (DATAFILE &File, double Version)
{
  PRO_METHOD ("FLOOR", "\n")

  if (File.Parse("<\\[FLOOR\\]>\n") < 0)
    PRO_RETURN ("Key [FLOOR] not found", FALSE)

  if (! CAD_FILE::Read(File, Version))
    PRO_RETURN ("Reading FILE failed", FALSE)

  if (File.Parse("  <Mode:>      (OFF|ON)\n"
		 "  <Size:>      (%d) x (%d)\n"
		 "  <Space:>     (%f) x (%f)\n"
		 "  <Position:>  (%f) / (%f) / (%f)\n"
		 "  <Rotation:>  (%f) / (%f) / (%f)\n",
		 &Mode, 
		 &SizeX,    &SizeY, 
		 &SpaceX,   &SpaceY,
		 &Place.px, &Place.py, &Place.pz,
		 &Place.y,  &Place.p,  &Place.r) < 11)
    PRO_RETURN ("Reading DATA failed", FALSE)

  if (! Material.Read(File, Version))
    PRO_RETURN ("Reading MATERIAL failed", FALSE)

  if (! Grid.Read(File, Version))
    PRO_RETURN ("Reading GRID failed", FALSE)

  Create ();

  PRO_END
  return TRUE;
}


bool FLOOR::Write (FILE *fp /* = NULL */, int Indent /* = 0 */) const
{
  PRO_METHOD ("FLOOR", "\n")

  if (fp == NULL)
    fp = stdout;

  Space   (fp, Indent);
  fprintf (fp, "[FLOOR]\n");
  Indent += INDENT;

  if (! CAD_FILE::Write(fp, Indent))
    PRO_RETURN ("Writing failed", FALSE)

  Space   (fp, Indent);
  fprintf (fp, "Mode:      %s\n",          (Mode==ON)? "ON":"OFF");

  Space   (fp, Indent);
  fprintf (fp, "Size:      %d x %d\n",     SizeX, SizeY);

  Space   (fp, Indent);
  fprintf (fp, "Space:     %.2f x %.2f\n", SpaceX, SpaceY);

  Space   (fp, Indent);
  fprintf (fp, "Position:  %.3f / %.3f / %.3f\n", 
	       Place.px, Place.py, Place.pz);

  Space   (fp, Indent);
  fprintf (fp, "Rotation:  %.3f / %.3f / %.3f\n", 
	       Place.y, Place.p, Place.r);

  if (! Material.Write(fp, Indent))
    PRO_RETURN ("Writing MATERIAL failed", FALSE)

  if (! Grid.Write(fp, Indent))
    PRO_RETURN ("Writing GRID failed", FALSE)

  PRO_END
  return TRUE;
}


void FLOOR::Draw ()
{
  PRO_METHOD ("FLOOR", "")

  if (Mode == ON && ID >= 0)
    glCallList (ID);

  PRO_END
}


void FLOOR::SetMode (MODE NewMode /* = RASTER */)
{
  PRO_METHOD ("FLOOR", "\n")

  Mode = NewMode;
  Create ();

  PRO_END
}


void FLOOR::SetSize (int Width, int  Height)
{
  PRO_METHOD ("FLOOR", "\n")

  assert (Width >= 0 && Height >= 0);

  SizeX = Width;
  SizeY = Height;

  Create ();
  PRO_END
}


void FLOOR::SetPos (double px, double py, double pz)
{
  PRO_METHOD ("FLOOR", "\n")

  Place.px = px;
  Place.py = py;
  Place.pz = pz;

  Create ();
  PRO_END
}


void FLOOR::SetRot (double rx, double ry, double rz)
{
  PRO_METHOD ("FLOOR", "\n")

  Place.y = rx;
  Place.p = ry;
  Place.r = rz;

  Create ();
  PRO_END
}


void FLOOR::SetSpacing (double DistX, double DistY)
{
  PRO_METHOD ("FLOOR", "\n")

  assert (DistX >= 0.0 && DistY >= 0.0);

  SpaceX = DistX;
  SpaceY = DistY;

  Create ();

  PRO_END
}


void FLOOR::SetMaterial (const MATERIAL &NewMaterial)
{
  PRO_METHOD ("FLOOR", "\n")

  Material = NewMaterial;
  Create ();

  PRO_END
}


void FLOOR::SetLineWidth (int Width /* = 1 */)
{
  LineWidth = Width;
  Create ();
}


void FLOOR::SetGridColor (const COLOR &Color)
{
  PRO_METHOD ("FLOOR", "\n")
 
  Grid = Color;
  Create ();

  PRO_END
}


int FLOOR::GetMode () const
{
  return Mode;
}


int FLOOR::GetSizeX  () const
{
  return SizeX;
}


int FLOOR::GetSizeY () const
{
  return SizeY;
}


double FLOOR::GetSpaceX () const
{
  return SpaceX;
}


double FLOOR::GetSpaceY () const
{
  return SpaceY;
}


MATERIAL *FLOOR::GetMaterial ()
{
  return &Material;
}


int FLOOR::GetLineWidth () const
{
  return LineWidth;
}


COLOR &FLOOR::GetGridColor ()
{
  return Grid;
}


bool FLOOR::Create ()
{
  int      count;
  double   MinX, MaxX;
  double   MinY, MaxY;
  double   MinZ, MaxZ, StepZ;
  double   dx, dy, dz;
  double   Stretch[2] = {1.0, 1.0};
  H_MATRIX WTF, FTG, WTG;
  TEXTURE *Texture;

  PRO_METHOD ("FLOOR", "\n")

  MaxX = Place.px + 0.5 * (SizeX * SpaceX);
  MaxY = Place.py + 0.5 * (SizeY * SpaceY);
  MinX = Place.px - MaxX;
  MinY = Place.py - MaxY;
  MinZ = MaxZ = Place.pz;                   // Use same z-coordinates yet

  if ((Texture = Material.GetTexture()) != NULL)
  {
    Stretch[X] = Texture->GetStretchS ();
    Stretch[Y] = Texture->GetStretchT ();
  }

  ID = OpenGL->GetID (ID, "FLOOR");
  glNewList (ID, GL_COMPILE);

  Material.Set ();

  WTF = GEORPY2Frame (Place.px, Place.py, Place.pz, Place.y, Place.p, Place.r);
  FTG = GEOMakeTrans (SpaceX, SpaceY, 0.0);
  WTG = GEOMatrixProd (WTF, FTG);
  dx = WTG.m[P][X] - WTF.m[P][X];
  dy = WTG.m[P][Y] - WTF.m[P][Y];
  dz = WTG.m[P][Z] - WTF.m[P][Z];

  glPushMatrix ();
  glBegin (GL_QUADS);
    glNormal3d (0.0, 0.0, 1.0);

    // TODO: Eckpunkte mit Tranformationsberechnung bestimmen

    if (Material.HasTexture())
      glTexCoord2f (0.0, 0.0);
#if 1
    glVertex3d (Place.px-0.5*SizeX*dx, Place.py-0.5*SizeX*dy, Place.pz-0.5*SizeX*dz);
#else
    glVertex3d (Place.px+MinX, Place.py+MinY, MinZ);
#endif

    if (Material.HasTexture())
      glTexCoord2f (SizeX/10.0 * Stretch[X], 0.0);
#if 1
    glVertex3d (Place.px+0.5*SizeX*dx, Place.py-0.5*SizeX*dy, Place.pz);
#else
    glVertex3d (Place.px+MaxX, Place.py+MinY, 0.5*(MinZ+MaxZ));
#endif

    if (Material.HasTexture())
      glTexCoord2f (SizeX/10.0 * Stretch[X], SizeY/10.0 * Stretch[Y]);
#if 1
    glVertex3d (Place.px+0.5*SizeX*dx, Place.py+0.5*SizeX*dy, Place.pz+0.5*SizeX*dz);
#else
    glVertex3d (Place.px+MaxX, Place.py+MaxY, MaxZ);
#endif

    if (Material.HasTexture())
      glTexCoord2f (0.0, SizeY/10.0*Stretch[Y]);
#if 1
    glVertex3d (Place.px-0.5*SizeX*dx, Place.py+0.5*SizeX*dy, Place.pz);
#else
    glVertex3d (Place.px+MinX, Place.py+MaxY, 0.5*(MinZ+MaxZ));
#endif
  glEnd ();

  if (Material.HasTexture())
    glDisable (GL_TEXTURE_2D);

  if (LineWidth > 0)
  {
    glDisable   (GL_DEPTH_TEST);
    glEnable    (GL_COLOR_MATERIAL);
    glLineWidth (LineWidth);
    Grid.Set ();

    StepZ = (MaxZ - MinZ) / double(SizeX+1);
    for (count = 0; count <= SizeX; ++count)
    {
      glBegin (GL_LINES);
        glVertex3d (MinX+count*dx, MinY, MinZ+count*StepZ);
        glVertex3d (MinX+count*dx, MaxY, MinZ+count*StepZ);
      glEnd ();
    }

    StepZ = (MaxZ - MinZ) / double(SizeY+1);
    for (count = -SizeX/2; count <= SizeY/2; ++count)
    {
      glBegin (GL_LINES);
        glVertex3d (Place.px-0.5*SizeX*dx, Place.py+count*dy, MinZ+count*StepZ);
        glVertex3d (Place.px+0.5*SizeX*dx, Place.py+count*dy, MinZ+count*StepZ);
      glEnd ();
    }

    glDisable (GL_COLOR_MATERIAL);
    glEnable  (GL_DEPTH_TEST);
  }
  glPopMatrix ();

  glEndList ();

  PRO_END
  return TRUE;
}
