#include <assert.h>

#include "libcad3d.h"


FOG::FOG (MODE Mode /* = OFF */, double Density /* = 0.5 */) :
  CAD_FILE (),
  UNIT     (),
  Mode     (Mode),
  Hint     (FASTEST),
  Color    (COLOR::WHITE),
  Density  (Density),
  Start    (0.0),
  End      (1.0)
{
  PRO_METHOD ("FOG", "")

  PRO_END
}


bool FOG::Read (DATAFILE &File, double Version)
{
  PRO_METHOD ("FOG", "\n")

  if (File.Parse("<\\[FOG\\]>\n") < 0)
    return FALSE;

  if (! CAD_FILE::Read(File, Version))
    return FALSE;

  if (File.Parse("  <Mode:>     (OFF|LINEAR|EXPONENT|EXP2)\n"
		 "  <Hint:>     (DONT_CARE|FASTEST|NICEST)\n"
		 "  <Density:>  (%f)\n"
		 "  <Start:>    (%f)\n"
		 "  <End:>      (%f)\n",
		 &Mode, &Hint, &Density, &Start, &End) < 5)
    return FALSE;

  if (! Color.Read(File, Version))
    return FALSE;

  PRO_END
  return TRUE;
}


bool FOG::Write (FILE *fp /* = NULL */, int Indent /* = 0 */) const
{
  PRO_METHOD ("FOG", "\n")

  if (fp == NULL)
    fp = stdout;

  Space   (fp, Indent);
  fprintf (fp, "[FOG]\n");
  Indent += INDENT;

  if (! CAD_FILE::Write(fp, Indent))
    PRO_RETURN ("Writing failed", FALSE)

  Space   (fp, Indent);
  fprintf (fp, "Mode:     ");
  switch (Mode)
  {
    case    OFF : fprintf (fp, "OFF\n");
                  break;
    case LINEAR : fprintf (fp, "LINEAR\n");
                  break;
    case    EXP : fprintf (fp, "EXPONENT\n");
                  break;
    case   EXP2 : fprintf (fp, "EXP2\n");
                  break;
        default : fprintf (fp, "INVALID\n");
  }

  Space   (fp, Indent);
  fprintf (fp, "Hint:     %s\n", 
	       (Hint==FASTEST)?"FASTEST":(Hint==NICEST)?"NICEST":"DONT_CARE");

  Space   (fp, Indent);
  fprintf (fp, "Density:  %f\n", Density);

  Space   (fp, Indent);
  fprintf (fp, "Start:    %f\n", Start);

  Space   (fp, Indent);
  fprintf (fp, "End:      %f\n", End);

  Color.Write (fp, Indent);

  PRO_END
  return TRUE;
}


void FOG::Set () const
{
  GLfloat FogColor[4] = {GLfloat(Color[RED]),  GLfloat(Color[GREEN]), 
			 GLfloat(Color[BLUE]), GLfloat(Color[ALPHA])};

  PRO_METHOD ("FOG", "\n")

  if (Mode != OFF)
  {
    glEnable (GL_FOG);
    glFogfv  (GL_FOG_COLOR, FogColor);
    switch (Hint)
    {
      case FASTEST :
	glHint (GL_FOG_HINT, GL_FASTEST);
	break;

      case NICEST :
	glHint (GL_FOG_HINT, GL_NICEST);
	break;

      default :
    	glHint (GL_FOG_HINT, GL_DONT_CARE);
    }
  }

  switch (Mode)
  {
    case LINEAR : 
      glFogi (GL_FOG_MODE, GL_LINEAR);  
      glFogf (GL_FOG_START, 5.0 * Density * Start);
      glFogf (GL_FOG_END,   5.0 * Density * End);
      break;

    case EXP :
      glFogi (GL_FOG_MODE,    GL_EXP);
      glFogf (GL_FOG_DENSITY, Density * 0.005);
      break;

    case EXP2 :
      glFogi (GL_FOG_MODE,    GL_EXP2);
      glFogf (GL_FOG_DENSITY, Density * 0.005);
      break;

    default :
      glDisable (GL_FOG);
  }

  PRO_END
}


void FOG::SetMode (MODE Mode /* = OFF */)
{
  this->Mode = Mode;

  Set ();
}


void FOG::SetHint (HINT Hint /* = FASTEST */)
{
  this->Hint = Hint;

  Set ();
}


void FOG::SetColor (COLOR Color /* = COLOR(COLOR::WHITE) */)
{
  this->Color = Color;

  Set ();
}


void FOG::SetDensity (double Density)
{
  this->Density = Density;
  
  Set ();
}


void FOG::SetRange (double Start /* = 0.0 */, double End /* = 1.0 */)
{
  this->Start = Start;
  this->End = End;

  Set ();
}


FOG::MODE FOG::GetMode () const
{
  return Mode;
}


FOG::HINT FOG::GetHint () const
{
  return Hint;
}


COLOR &FOG::GetColor ()
{
  return Color;
}


double FOG::GetDensity () const
{
  return Density;
}


double FOG::GetStart () const
{
  return Start;
}


double FOG::GetEnd () const
{
  return End;
}
