#include <stdarg.h>
#include <math.h>

#include <gltt/FTEngine.h>
#include <gltt/FTFace.h>
#include <gltt/FTInstance.h>
#include <gltt/FTGlyph.h>
#include <gltt/FTFont.h>
#include <gltt/GLTTOutlineFont.h>
#include <gltt/GLTTFont.h>
#include <gltt/GLTTGlyphPolygonizer.h>
#include <gltt/GLTTGlyphTriangulator.h>

#include <libtools.h>

#include "libcad3d.h"

/* 
   The gltt-library seems to have problems creating some letters if the 
   font size is to small (e.g. 'S' will not be created in times with <= 7 pt).
   Possibly there are rounding problems when using int-values.
   So let's create the letters with a larger size (Factor: ENLARGE) and 
   shrink them back to get the desired size.
*/
#define ENLARGE 10
#define SHRINK  (1.0/ENLARGE)


class GLTTGlyphTriangles : public GLTTGlyphTriangulator
{
  public:
    struct Triangle
    {
      FTGlyphVectorizer::POINT *p1;
      FTGlyphVectorizer::POINT *p2;
      FTGlyphVectorizer::POINT *p3;
    };

    int          nTriangles;
    Triangle    *Triangles;
    GLTTboolean  count_them;
 
    GLTTGlyphTriangles (FTGlyphVectorizer *Vectorizer) :
      GLTTGlyphTriangulator (Vectorizer)
    {
      nTriangles = 0;
      Triangles  = NULL;
      count_them = GLTT_TRUE;
    }

    virtual ~GLTTGlyphTriangles ()
    {
      delete[] Triangles;
      Triangles = 0;
    }

    void alloc()
    {
      delete[] Triangles;
      Triangles = new Triangle [nTriangles + 1];
    }

    virtual void triangle (FTGlyphVectorizer::POINT *p1,
			   FTGlyphVectorizer::POINT *p2,
			   FTGlyphVectorizer::POINT *p3)
    {
      if (! count_them)
      {
	Triangles[nTriangles].p1 = p1;
	Triangles[nTriangles].p2 = p2;
	Triangles[nTriangles].p3 = p3;
      }
      ++nTriangles;
    }
};


OBJECT *MakeLetter (FTGlyphVectorizer *Vec, GLTTGlyphTriangles *Tri, 
		    double Depth, H_MATRIX M, const MATERIAL &Material)
{
  int                         count, j;
  int                         Index = 0;
  double                     *p1, *p2, *p3;
  double                     *n1, *n2, *n3;
  POLYGON                    *Polygon;
  POLYEDER                   *Polyeder;
  OBJECT                     *Obj;
  FTGlyphVectorizer::Contour *Contour;

  if (Tri->nTriangles == 0)
    return NULL;

  Obj = OBJNewObject ();
  Obj->ColorIndex = Material.GetColorID();
  Obj->Type       = PART_PLANE;
  Obj->VertexNo   = Tri->nTriangles * 3 * 2;     // each side one triangle
  if (Depth != 0.0)
    for (count = 0; count < Vec->getNContours(); ++count)
      if ((Contour = Vec->getContour(count)) != NULL)
	(Obj->VertexNo) += (Contour->nPoints+1) * 2;

  Polyeder = GEONewPolyeder ();
  Polyeder->Number = 0;

  Obj->VertexList = (POINT3D *)malloc (sizeof(POINT3D) * Obj->VertexNo);

  // draw faces of letters...
  for (j = 0; j < Tri->nTriangles; ++j)
  {
    GLTTGlyphTriangles::Triangle &t = Tri->Triangles[j];

    p1 = ((double *)t.p1->data);
    p2 = ((double *)t.p2->data);
    p3 = ((double *)t.p3->data);
    n1 = p1 + 3;
    n2 = p2 + 3;
    n3 = p3 + 3;

    // Front-side of letter
    Obj->VertexList[Index  ] = 
      GEOMakeVector (SHRINK*p3[0], SHRINK*p3[1], SHRINK*p3[2]);
    Obj->VertexList[Index+1] = 
      GEOMakeVector (SHRINK*p2[0], SHRINK*p2[1], SHRINK*p2[2]);
    Obj->VertexList[Index+2] = 
      GEOMakeVector (SHRINK*p1[0], SHRINK*p1[1], SHRINK*p1[2]);

    Polygon = GEONewPolygon ();
    Polygon->Points = Obj->VertexList;
    Polygon->Number = 3;
    Polygon->Vertex = (int *)malloc (sizeof(int) * Polygon->Number);
    Polygon->Vertex[0] = Index;
    Polygon->Vertex[1] = Index+1;
    Polygon->Vertex[2] = Index+2;
    Polygon->Normals = (VECTOR3D *)malloc (sizeof(VECTOR3D) * 3);
    Polygon->Normals[0] = GEOMakeVector (n3[0], n3[1], n3[2]);
    Polygon->Normals[1] = GEOMakeVector (n2[0], n2[1], n2[2]);
    Polygon->Normals[2] = GEOMakeVector (n1[0], n1[1], n1[2]);

    Polygon->Info.ID = j;
    Polygon->Info.Center = GEOCalcPolygonCenter (Polygon);
    GEOCalcHesseForm (Polygon);
    Polygon->Triangle = (TRIANGLE *)malloc (sizeof(TRIANGLE));
    Polygon->Triangle[0][0] = Polygon->Vertex[0];
    Polygon->Triangle[0][1] = Polygon->Vertex[1];
    Polygon->Triangle[0][2] = Polygon->Vertex[2];
    if (Polyeder->List == NULL)
      Polyeder->List = Polygon;
    else
      Polyeder->Last->Next = Polygon;
    Polyeder->Last = Polygon;
    (Polyeder->Number)++;
      
    Index += 3;

    // Back-side of letter
    Obj->VertexList[Index  ] = 
      GEOMakeVector (SHRINK*p1[0], SHRINK*p1[1], SHRINK*p1[2]+Depth);
    Obj->VertexList[Index+1] = 
      GEOMakeVector (SHRINK*p2[0], SHRINK*p2[1], SHRINK*p2[2]+Depth);
    Obj->VertexList[Index+2] = 
      GEOMakeVector (SHRINK*p3[0], SHRINK*p3[1], SHRINK*p3[2]+Depth);

    Polygon = GEONewPolygon ();
    Polygon->Points = Obj->VertexList;
    Polygon->Number = 3;
    Polygon->Vertex = (int *)malloc (sizeof(int) * Polygon->Number);
    Polygon->Vertex[0] = Index;
    Polygon->Vertex[1] = Index+1;
    Polygon->Vertex[2] = Index+2;
    Polygon->Normals = (VECTOR3D *)malloc (sizeof(VECTOR3D) * 3);
    Polygon->Normals[0] = GEOMakeVector (-n1[0], 0.0, -n1[2]);
    Polygon->Normals[1] = GEOMakeVector (-n2[0], 0.0, -n2[2]);
    Polygon->Normals[2] = GEOMakeVector (-n3[0], 0.0, -n3[2]);

    Polygon->Info.ID = j;
    Polygon->Info.Center = GEOCalcPolygonCenter (Polygon);
    GEOCalcHesseForm (Polygon);
    Polygon->Triangle = (TRIANGLE *)malloc (sizeof(TRIANGLE));
    Polygon->Triangle[0][0] = Polygon->Vertex[0];
    Polygon->Triangle[0][1] = Polygon->Vertex[1];
    Polygon->Triangle[0][2] = Polygon->Vertex[2];
    Polyeder->Last->Next = Polygon;
    Polyeder->Last = Polygon;
    (Polyeder->Number)++;

    Index += 3;
  }

  if (Depth != 0.0)    // Draw sides between both faces...
  {
    int       j, j0, j1;
    double   *p0, *p1, *e;
    double    vx, vy, vz;
    double    nx, ny, nz;
    VECTOR3D  Normals[4] = {{0.0, 0.0, 0.0}, {0.0, 0.0, 0.0},
                            {0.0, 0.0, 0.0}, {0.0, 0.0, 0.0}};
    FTGlyphVectorizer::POINT *point0, *point1;

    for (count = 0; count < Vec->getNContours(); ++count)
      if ((Contour = Vec->getContour(count)) != NULL)
	for (j = 0; j <= Contour->nPoints; ++j)
	{
	  j1 = (j < Contour->nPoints) ? j : (j-Contour->nPoints);
	  j0 = (j1 == 0) ? (Contour->nPoints-1) : (j1-1);

	  point0 = Contour->points + j0;
	  point1 = Contour->points + j1;
	  p0  = (double *)point0->data;
	  p1  = (double *)point1->data;
	  e   = p0 + 3;
	  vx  = p1[0] - p0[0];
	  vy  = p1[1] - p0[1];
	  vz  = p1[2] - p0[2];
	  nx  =           - vy * e[2];
	  ny  = e[2] * vx - vz * e[0];
	  nz  = e[0] * vy;
  
	  Obj->VertexList[Index  ] = 
	    GEOMakeVector (SHRINK*p0[0], SHRINK*p0[1], SHRINK*p0[2]);
	  Obj->VertexList[Index+1] = 
	    GEOMakeVector (SHRINK*p0[0]-e[0]*Depth, 
			   SHRINK*p0[1], SHRINK*p0[2]-e[2]*Depth);

	  Normals[0] = Normals[2];
	  Normals[1] = Normals[3];
	  Normals[2] = GEOMakeVector (nx, ny, nz);
	  Normals[3] = GEOMakeVector (nx, ny, nz);

	  Index += 2;

	  if (j > 0)
	  {
	    Polygon = GEONewPolygon ();
	    Polygon->Points = Obj->VertexList;
	    Polygon->Number = 4;
	    Polygon->Vertex = (int *)malloc (sizeof(int) * Polygon->Number);
	    Polygon->Vertex[0] = Index-4;
	    Polygon->Vertex[1] = Index-3;
	    Polygon->Vertex[2] = Index-1;
	    Polygon->Vertex[3] = Index-2;
	    Polygon->Normals = (VECTOR3D *)malloc (sizeof(VECTOR3D) * 4);
	    Polygon->Normals[0] = Normals[0];
	    Polygon->Normals[1] = Normals[1];
	    Polygon->Normals[2] = Normals[3];
	    Polygon->Normals[3] = Normals[2];
	      
	    Polygon->Info.ID = Polyeder->Number+1;
	    Polygon->Info.Center = GEOCalcPolygonCenter (Polygon);
	    GEOCalcHesseForm (Polygon);

	    // GEOTriangulatePolygon is restricted to planar polygons only.
	    // Because the letters might be blended, we triangulate the 
	    // polygons manually.
	    Polygon->Triangle = (TRIANGLE *)malloc (sizeof(TRIANGLE) * 2);
	    Polygon->Triangle[0][0] = Polygon->Vertex[0];
	    Polygon->Triangle[0][1] = Polygon->Vertex[1];
	    Polygon->Triangle[0][2] = Polygon->Vertex[3];
	    Polygon->Triangle[1][0] = Polygon->Vertex[0];
	    Polygon->Triangle[1][1] = Polygon->Vertex[3];
	    Polygon->Triangle[1][2] = Polygon->Vertex[2];
	    Polyeder->Last->Next = Polygon;
	    Polyeder->Last       = Polygon;
	    (Polyeder->Number)++;
	  }
	}
  }

  Obj->Data.Polyeder = Polyeder;
  Obj->Info = OBJCalcInfo (Obj);

  return Obj;
}


/* TODO:
   Fertige Transform-Funktionen fr:
   - Condensed
   - Italic (X-/Z-Achse)
   - ...
   - Rckgabe der Textbreite ? (auch im device ermittelbar)
   - Eigene Tranformationsfunktion fr jeden einzelnen Buchstaben
*/


DEVICE3D *WORLD::CreateText (const char *FontName, int FontSize, double Depth,
			     double Precision, const MATERIAL &Material,
			     TRANSFORM_FUNC Transform, bool relative,
			     const char *Text, ...)
{
  bool                        first = TRUE;
  int                         count, TextLen;
  int                         c, j;
  double                      MinY = 0.0, MaxY = 0.0;
  double                      OffsX = 0.0;
  double                     *Pos, *Norm;
  char                        Message[1024];
  char                        FontFile[PATH_MAX+1];
  va_list                     Arguments;
  FTGlyph                    *Glyph;
  VECTOR3D                    Vector, Normal;
  FTFace                     *TTFace; 
  H_MATRIX                    WTL = Matrix_1;
  OBJECT                     *Obj = NULL;
  PART                       *Part;
  PART3D                     *Letter, *Last = NULL;
  DEVICE3D                   *Dev;
  FTGlyphVectorizer          *Vec = NULL;
  GLTTGlyphTriangles         *Tri = NULL;
  FTGlyphVectorizer::Contour *Contour;
  FTGlyphVectorizer::POINT   *Point;

  if (Text == NULL)
    return NULL;

  TOOLParseFileName (FontName, FontFile, PATH_MAX);

  TTFace = new FTFace;
  if (! TTFace->open(FontFile))
  {
    printf ("error loading font '%s'", FontFile);
    return NULL;
  }

  GLTTFont TTFont (TTFace);
  if (! TTFont.create(ENLARGE * FontSize))
  {
    printf ("Create failed");
    return NULL;
  }

  va_start (Arguments, Text);
  vsprintf (Message, Text, Arguments);
  va_end   (Arguments);

  TextLen = strlen (Message);

  Vec = new FTGlyphVectorizer;
  Tri = new GLTTGlyphTriangles (Vec);
  
  Dev = new DEVICE3D ("text");

  for (count = 0; count < TextLen; ++count)
  {
    // get glyph for character
    Glyph = TTFont.getFont()->getGlyph ((unsigned char)Message[count]);
    if (Glyph == NULL)
      continue;

    Precision = MAX (0.1, MIN(1.0, Precision));
    Vec->setPrecision (1.0 / Precision);
    if (Vec->init(Glyph) == 0)
      continue;

    if (! Vec->vectorize())
      continue;

    // lookup extreme values and allocate point's memory...
    for (c = 0; c < Vec->getNContours(); ++c)
      if ((Contour = Vec->getContour(c)) != NULL && Contour->nPoints > 0)
	for (j= 0; j < Contour->nPoints; ++j)
	{
	  Point = Contour->points + j;
	  if (first)
	  {
	    MinY = MaxY = Point->y;
	    first = FALSE;
	  }
	  else
	  {
	    if (Point->y < MinY)
	      MinY = Point->y;
	    if (Point->y > MaxY) 
	      MaxY = Point->y;
	  }
	  Point->data = (void *)new double[6];  // 0-2: points, 3-5: normals
	}

    if (Tri->init(Glyph) != 0)
    {
      Tri->count_them = GLTT_TRUE;
      Tri->nTriangles = 0;
      Tri->triangulate ();

      Tri->count_them = GLTT_FALSE;
      Tri->alloc ();   // Speicherleck ???
      Tri->nTriangles = 0;
      Tri->triangulate ();
    }

    // Compute Positions of contour/triangle-points 
    for (c = 0; c < Vec->getNContours(); ++c)
      if ((Contour = Vec->getContour(c)) != NULL)
	for (j = 0; j < Contour->nPoints; ++j)
	{
	  Point = Contour->points + j;

	  Pos  = (double *)Point->data;
          Norm = Pos + 3;

	  // copy coordinates into vector for calling Transform 
	  if (relative)
	    Vector.x = Point->x;
	  else
	    Vector.x = WTL.m[P][X] + Point->x;  // use absolute coordinates
	  Vector.y = Point->y;
	  Vector.z = 0.0;

	  Normal = GEOMakeVector (0.0, 0.0, -1.0);

	  if (Transform != NULL)  // User transformation
	  {
	    Transform (&Vector, &Normal);
	    if (! relative)
	      Vector = GEOMatrixMult (GEOMatrixInv(WTL), Vector);
	  }

	  Pos[X] = Vector.x;
	  Pos[Y] = Vector.y;
	  Pos[Z] = Vector.z;
	  
	  Norm[X] = Normal.x;
	  Norm[Y] = Normal.y;
	  Norm[Z] = Normal.z;
	}

    Part = PARTNewPart ();
    PARTSetName (Part, "text_%c", Message[count]);
    //    Part->ObjectNo = Vec->getNContours ();
    Part->ObjectNo = 1;

    if (Tri->nTriangles > 0)
    {
      Obj = MakeLetter (Vec, Tri, Depth, WTL, Material);

      Part->ObjectList = Obj;       // There's only one object in each part
      PARTCalcInfo (Part);

      Letter = new PART3D (Part, this, Material);
      Letter->GetJoint().SetAxes (0x3F);   // (1<<POS_X) | ... | (1<<TRANS_Z)

      if (Last == NULL)
	Dev->AddPart (Letter);
      else
      {
	Letter->SetMatrix (GEOMakeTrans(OffsX, 0.0, 0.0));
	Last->AddChain (Letter);
      }
      
      Last = Letter;
      OffsX = SHRINK * Vec->getAdvance ();      // Offset for next character
    
      // advance next part's position
      WTL.m[P][X] += OffsX;
    }
  }

  // Free allocated memory...
  for (c= 0; c < Vec->getNContours(); ++c)
    if ((Contour = Vec->getContour(c)) != NULL)
      for (j= 0; j < Contour->nPoints; ++j)
      {
	Point = Contour->points + j;
	delete [] (double *)Point->data;
       	Point->data= 0;
      }

  delete Tri;
  delete Vec;

  return Dev;
}


double WORLD::GetTextWidth (char *FontName, int FontSize, char *Text, ...)
{
  int                count, TextLen;
  double             Width = 0.0;
  char               Message[1024];
  va_list            Arguments;
  FTFace            *TTFace; 
  FTGlyphVectorizer *Vec = NULL;

  if (Text == NULL)
    return 0.0;

  TTFace = new FTFace;
  if (! TTFace->open(FontName))
  {
    printf ("error loading fon");
    return 0.0;
  }

  GLTTFont TTFont (TTFace);
  if (! TTFont.create(ENLARGE * FontSize))
  {
    printf ("Create failed");
    return 0.0;
  }

  va_start (Arguments, Text);
  vsprintf (Message, Text, Arguments);
  va_end   (Arguments);

  TextLen = strlen (Message);

  Vec = new FTGlyphVectorizer[TextLen];
  
  for (count = 0; count < TextLen; ++count)
  {
    FTGlyphVectorizer &v = Vec[count];
    v.setPrecision (20);

    FTGlyph *g = TTFont.getFont()->getGlyph ((unsigned char)Message[count]);
    if (g != NULL && v.init(g) != 0)
      Width += SHRINK * v.getAdvance ();
  }

  delete [] Vec;

  return Width;
}
