#include <assert.h>

#include "libcad3d.h"


JOINT::JOINT (int Axes) :
  CAD_FILE (),
  Axes     (Axes),
  Speed    (0.0, 0.0, 0.0),
  Accel    (0.0, 0.0, 0.0),
  Time     (0.0, 0.0, 0.0)
{
  int count;

  PRO_METHOD ("JOINT", "")

  for (count = POS_X; count <= ROT_Z; ++count)
    if (! (Axes & (1 << count)))
      Range[count].SetRange (0.0, 0.0);

  PRO_END
}


bool JOINT::Read (DATAFILE &File, double Version)
{
  PRO_METHOD ("JOINT", "\n")

  if (File.Parse("<\\[JOINT\\]>\n") < 0)
    return FALSE;

  if (! CAD_FILE::Read(File, Version))
    return FALSE;

  if (File.Parse("  <Axes:>  {POS_X|POS_Y|POS_Z|ROT_X|ROT_Y|ROT_Z}\n"
		 "  <Speed:> (%f) (%f) (%f)\n"
		 "  <Accel:> (%f) (%f) (%f)\n"
		 "  <Time:>  (%f) (%f) (%f)\n",
		 &Axes, 
		 &Speed[RANGE::MINIMUM], &Speed[RANGE::VALUE],
		 &Speed[RANGE::MAXIMUM],
		 &Accel[RANGE::MINIMUM], &Accel[RANGE::VALUE],
		 &Accel[RANGE::MAXIMUM],
		 &Time[RANGE::MINIMUM],  &Time[RANGE::VALUE],
		 &Time[RANGE::MAXIMUM]) < 10)
    return FALSE;

  PRO_END
  return TRUE;
}


bool JOINT::Write (FILE *fp /* = NULL */, int Indent /* = 0 */) const
{
  PRO_METHOD ("JOINT", "\n")

  if (fp == NULL)
    fp = stdout;

  Space   (fp, Indent);
  fprintf (fp, "[JOINT]\n");
  Indent += INDENT;

  if (! CAD_FILE::Write(fp))
    PRO_RETURN ("Write failed", FALSE)

  Space   (fp, Indent);
  fprintf (fp, "Axes: ");
  if (Axes & (1<<POS_X))
    fprintf (fp, "  POS_X");
  if (Axes & (1<<POS_Y))
    fprintf (fp, "  POS_Y");
  if (Axes & (1<<POS_Z))
    fprintf (fp, "  POS_Z");
  if (Axes & (1<<ROT_X))
    fprintf (fp, "  ROT_X");
  if (Axes & (1<<ROT_Y))
    fprintf (fp, "  ROT_Y");
  if (Axes & (1<<ROT_Z))
    fprintf (fp, "  ROT_Z");
  fprintf (fp, "\n");
  Space   (fp, Indent);
  fprintf (fp, "Speed:  %f %f %f\n", Speed[RANGE::MINIMUM], 
	       Speed[RANGE::VALUE], Speed[RANGE::MAXIMUM]);
  Space   (fp, Indent);
  fprintf (fp, "Accel:  %f %f %f\n", Accel[RANGE::MINIMUM], 
	       Accel[RANGE::VALUE], Accel[RANGE::MAXIMUM]);
  Space   (fp, Indent);
  fprintf (fp, "Time:   %f %f %f\n", Time[RANGE::MINIMUM], 
	       Time[RANGE::VALUE], Time[RANGE::MAXIMUM]);

  PRO_END
  return TRUE;
}



void JOINT::Scale (double ScaleX, double ScaleY, double ScaleZ)
{
  PRO_METHOD ("JOINT", "")

  if (Axes & (1 << POS_X))
    Range[POS_X].SetRange (Range[POS_X].GetMin() * ScaleX,
			   Range[POS_X].GetMax() * ScaleX);

  if (Axes & (1 << POS_Y))
    Range[POS_Y].SetRange (Range[POS_Y].GetMin() * ScaleY,
			   Range[POS_Y].GetMax() * ScaleY);

  if (Axes & (1 << POS_Z))
    Range[POS_Z].SetRange (Range[POS_Z].GetMin() * ScaleZ,
			   Range[POS_Z].GetMax() * ScaleZ);

  PRO_END
}


void JOINT::SetAxes (int Axes)
{
  this->Axes = Axes;
}


bool JOINT::SetValue (int Axis, double Value)
{
  assert (Axis >= POS_X && Axis <= ROT_Z);

  Range[Axis].SetValue (Value);

  if (! Range[Axis].Check(Value))
    PRO4 ("Out of Range %d: %.3f..%.3f: %.3f\n", Axis, 
	  Range[Axis].GetMin(), Range[Axis].GetValue(), Range[Axis].GetMax())

  return Range[Axis].Check (Value);
}


void JOINT::SetRange (int Axes, double Min, double Max)
{
  int count;

  for (count = POS_X; count <= ROT_Z; ++count)
    if (Axes & (1 << count))
      Range[count].SetRange (Min, Max);
}


void JOINT::SetSpeed (double Speed)
{
  this->Speed.SetRange (0.0, Speed);
}


void JOINT::SetAccel (double Accel)
{
  this->Accel.SetRange (0.0, Accel);
}


void JOINT::SetTime (double Time)
{
  this->Time.SetRange (0.0, Time);
}


int JOINT::GetAxes () const
{
  return Axes;
}


double JOINT::GetValue (int Axis) const
{
  assert (Axis >= POS_X && Axis <= ROT_Z);

  return Range[Axis].GetValue ();
}


double JOINT::GetMin (int Axis) const
{
  assert (Axis >= POS_X && Axis <= ROT_Z);

  return Range[Axis].GetMin ();
}


double JOINT::GetMax (int Axis) const
{
  assert (Axis >= POS_X && Axis <= ROT_Z);

  return Range[Axis].GetMax ();
}


double JOINT::GetSpeed () const
{
  return Speed.GetMax ();
}


double JOINT::GetAccel () const
{
  return Accel.GetMax ();
}


double JOINT::GetTime () const
{
  return Time.GetMax ();
}


bool JOINT::IsFree (int Axis) const
{
  assert (Axis >= POS_X && Axis <= ROT_Z);

  return (Axes & (1 << Axis));
}


bool JOINT::IsValid (int Axis, double Value) const
{
  assert (Axis >= POS_X && Axis <= ROT_Z);

  return Range[Axis].Check (Value);
}


double JOINT::GetValue (int Axis, double Value) const
{
  assert (Axis >= POS_X && Axis <= ROT_Z);

  return LIMIT (Range[Axis].GetMin(), Value, Range[Axis].GetMax());
}


H_MATRIX JOINT::GetFrame () const
{
  return GEORPY2Frame (Range[POS_X].GetValue(),
		       Range[POS_Y].GetValue(),
		       Range[POS_Z].GetValue(),
		       Range[ROT_X].GetValue(),
		       Range[ROT_Y].GetValue(),
		       Range[ROT_Z].GetValue());
}
