#include <assert.h>
#include <math.h>

#include "libcad3d.h"


static GLenum LightTable[] = 
{
  GL_LIGHT0, GL_LIGHT1, GL_LIGHT2, GL_LIGHT3,  
  GL_LIGHT4, GL_LIGHT5, GL_LIGHT6, GL_LIGHT7
};


LIGHT::LIGHT (NUMBER Number) :
  CAD_FILE  (),
  UNIT      (),
  LightNo   (Number),
  Intensity (1.0),
  Ambient   (COLOR(COLOR::BLACK)),
  Placement (STATIONARY), 
  Size      (0.0),
  HeadLight (NULL)
{
  PRO_METHOD ("LIGHT", "\n")

  if (Number == LIGHT1)  // Set OpenGL default-values 
  {
    Mode     = DIRECTIONAL;
    Diffuse  = COLOR (COLOR::WHITE);
    Specular = COLOR (COLOR::WHITE);
  }
  else
  {
    Mode     = OFF;
    Diffuse  = COLOR (COLOR::BLACK);
    Specular = COLOR (COLOR::BLACK);
  }

  Initialize ();

  PRO_END
}
    

LIGHT::LIGHT (      NUMBER  Number, 
	            MODE    Mode,
              const COLOR  &Ambient  /* = COLOR(COLOR::BLACK) */,
	      const COLOR  &Diffuse  /* = COLOR(COLOR::WHITE) */,
	      const COLOR  &Specular /* = COLOR(COLOR::WHITE) */) :
  CAD_FILE  (),
  UNIT      (),
  LightNo   (Number),
  Mode      (Mode),
  Intensity (1.0),
  Ambient   (Ambient),
  Diffuse   (Diffuse),
  Specular  (Specular),
  Placement (STATIONARY),
  Size      (0.0),
  HeadLight (NULL)
{
  PRO_METHOD ("LIGHT", "\n")

  Initialize ();

  PRO_END
}


LIGHT::~LIGHT ()
{
  PRO_METHOD ("LIGHT~", "")

  if (Shadow != NULL)
    delete Shadow;

  // Switch off light
  glDisable (LightTable[LightNo]);

  if (HeadLight != NULL)
    delete HeadLight;

  PRO_END
}


void LIGHT::Initialize ()
{
  if (Mode == SPOTLIGHT)
  {
    CutOff  = 45.0;
    SpotExp =  1.0;
  }
  else
  {
    CutOff  = 180.0;
    SpotExp =   0.0;
  }

  // Attenuation factor = 1 / (quadratic * d^2 + linear * d + constant)
  Attenuation[0] = 0.0;                 // quadratic
  Attenuation[1] = 0.0;                 // linear
  Attenuation[2] = 1.0;                 // constant

  // positional light parameter (w-component automatically selected in Set())
  Pos = GEOMakeVector (0.0, 0.0,  1.0);

  // Spotlight direction
  Dir = GEOMakeVector (0.0, 0.0, -1.0);

  Create  ();
  CalcWTL ();

  Shadow = new SHADOW (this);           // Create shadow for this light source
}


bool LIGHT::Read (DATAFILE &File, double Version)
{
  PRO_METHOD ("LIGHT", "\n")

  if (File.Parse("<\\[LIGHT\\]>\n") < 0)
    return FALSE;

  if (! CAD_FILE::Read(File, Version))
    return FALSE;

  if (File.Parse("  <No:>           (%d)\n"
		 "  <Mode:>         (OFF|DIRECTIONAL|POSITIONAL|SPOTLIGHT)\n"
		 "  <Intensity:>   (%f)\n"
		 "  <Ambient:>      (%f) (%f) (%f) (%f)\n"
		 "  <Diffuse:>      (%f) (%f) (%f) (%f)\n"
		 "  <Specular:>     (%f) (%f) (%f) (%f)\n"
		 "  <Position:>     (%f) (%f) (%f)\n"
		 "  <Attenuation:>  (%f) d + (%f) d + (%f)\n"
		 "  <Direction:>    (%f) (%f) (%f)\n"
		 "  <CutOff:>       (%f)\n"
		 "  <Exponent:>     (%f)\n",
		 &LightNo, &Mode, &Intensity,
		 &Ambient[RED], &Ambient[GREEN], &Ambient[BLUE], 
		 &Ambient[ALPHA],
		 &Diffuse[RED], &Diffuse[GREEN], &Diffuse[BLUE], 
		 &Diffuse[ALPHA],
		 &Specular[RED], &Specular[GREEN], &Specular[BLUE], 
		 &Specular[ALPHA],
		 &Pos.x, &Pos.y, &Pos.z,
		 &Attenuation[0], &Attenuation[1], &Attenuation[2],
		 &Dir.x, &Dir.y, &Dir.z, &CutOff, &SpotExp) < 26)
    return FALSE;
		 
  Create ();

  PRO_END
  return TRUE;
}


bool LIGHT::Write (FILE *fp /* = NULL */, int Indent /* = 0 */) const
{
  PRO_METHOD ("LIGHT", "\n")

  if (fp == NULL)
    fp = stdout;

  Space   (fp, Indent);
  fprintf (fp, "[LIGHT]\n");
  Indent += INDENT;

  if (! CAD_FILE::Write(fp))
    PRO_RETURN ("Writing failed", FALSE)

  Space   (fp, Indent);
  fprintf (fp, "  No:           %d\n", LightNo);

  Space   (fp, Indent);  
  fprintf (fp, "  Mode:         ");
  switch (Mode)
  {
    case         OFF : fprintf (fp, "OFF\n"); 
                       break;
    case DIRECTIONAL : fprintf (fp, "DIRECTIONAL\n");
                       break;
    case  POSITIONAL : fprintf (fp, "POSITIONAL\n");
                       break;
    case   SPOTLIGHT : fprintf (fp, "SPOTLIGHT\n");
                       break;
             default : fprintf (fp, "INVALID\n");
  }

  Space   (fp, Indent);
  fprintf (fp, "  Intensity:    %.3g\n", Intensity);

  Space   (fp, Indent);
  fprintf (fp, "  Ambient:      %.3g   %.3g   %.3g   %.3g\n", 
	   Ambient[RED], Ambient[GREEN], Ambient[BLUE], Ambient[ALPHA]);

  Space   (fp, Indent);
  fprintf (fp, "  Diffuse:      %.3g   %.3g   %.3g   %.3g\n", 
	   Diffuse[RED], Diffuse[GREEN], Diffuse[BLUE], Diffuse[ALPHA]);

  Space   (fp, Indent);
  fprintf (fp, "  Specular:     %.3g   %.3g   %.3g   %.3g\n", 
	   Specular[RED], Specular[GREEN], Specular[BLUE], Specular[ALPHA]);

  Space   (fp, Indent);
  fprintf (fp, "  Position:     %.3g %.3g %.3g\n", Pos.x, Pos.y, Pos.z);

  Space   (fp, Indent);
  fprintf (fp, "  Attenuation:  %.3g d + %.3g d + %.3g\n", 
	       Attenuation[0], Attenuation[1], Attenuation[2]);

  // spotlight parameters
  Space   (fp, Indent);
  fprintf (fp, "  Direction:    %.3g %.3g %.3g\n", Dir.x, Dir.y, Dir.z);

  Space   (fp, Indent);
  fprintf (fp, "  CutOff:       %.3f\n", CutOff);

  Space   (fp, Indent);
  fprintf (fp, "  Exponent:     %.3g\n", SpotExp);
 
  PRO_END
  return TRUE;
}


bool LIGHT::Create ()
{
  GLfloat Light_Position[4] = {0.0, 0.0,  1.0, 0.0};
  GLfloat Light_Direction[] = {0.0, 0.0, -1.0};
  GLfloat Light_Ambient[] =
          { 
            (GLfloat)Ambient[RED]   * Intensity,
	    (GLfloat)Ambient[GREEN] * Intensity,
	    (GLfloat)Ambient[BLUE]  * Intensity,
	    (GLfloat)Ambient[ALPHA]
          };
  GLfloat Light_Diffuse[] =
          { 
            (GLfloat)Diffuse[RED]   * Intensity,
	    (GLfloat)Diffuse[GREEN] * Intensity,
	    (GLfloat)Diffuse[BLUE]  * Intensity,
	    (GLfloat)Diffuse[ALPHA]  
          };
  GLfloat Light_Specular[] = 
          { 
            (GLfloat)Specular[RED]   * Intensity,
	    (GLfloat)Specular[GREEN] * Intensity,
	    (GLfloat)Specular[BLUE]  * Intensity,
	    (GLfloat)Specular[ALPHA]
          };

  PRO_METHOD ("LIGHT", "\n")

  SetID ("LIGHT");
  glNewList (Id, GL_COMPILE_AND_EXECUTE);
  glIndexi (Id);

  // Set Modellview to identity 
  glMatrixMode (GL_MODELVIEW);

  glPushMatrix ();                      // save model-view matrix

  if (Placement == STATIONARY)  
    glLoadIdentity ();                  // stationary light has absolute pos

  // Set color of light
  glLightfv (LightTable[LightNo], GL_AMBIENT,  Light_Ambient);
  glLightfv (LightTable[LightNo], GL_DIFFUSE,  Light_Diffuse);
  glLightfv (LightTable[LightNo], GL_SPECULAR, Light_Specular);

  // Set attenuation factor
  glLightf  (LightTable[LightNo], GL_QUADRATIC_ATTENUATION, Attenuation[0]);
  glLightf  (LightTable[LightNo], GL_LINEAR_ATTENUATION,    Attenuation[1]);
  glLightf  (LightTable[LightNo], GL_CONSTANT_ATTENUATION,  Attenuation[2]);

  if (Mode == POSITIONAL || (Mode == SPOTLIGHT && CutOff <= 90.0))
  {
    Light_Position[0] = (GLfloat)Pos.x;
    Light_Position[1] = (GLfloat)Pos.y;
    Light_Position[2] = (GLfloat)Pos.z;
    Light_Position[3] = 1.0;
    
    Light_Direction[0] = (GLfloat)Dir.x;
    Light_Direction[1] = (GLfloat)Dir.y;
    Light_Direction[2] = (GLfloat)Dir.z;

    glLightfv (LightTable[LightNo], GL_SPOT_DIRECTION, Light_Direction);

    glLightModeli (GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE);

    glLightf (LightTable[LightNo], GL_SPOT_CUTOFF, 
		                   (Mode==SPOTLIGHT)? CutOff:180.0);
    glLightf (LightTable[LightNo], GL_SPOT_EXPONENT, SpotExp);
  }
  elsif (Mode == DIRECTIONAL)
  {
    glLightModeli (GL_LIGHT_MODEL_LOCAL_VIEWER, GL_FALSE);

    glLightf  (LightTable[LightNo], GL_SPOT_CUTOFF,   180.0);
    glLightf  (LightTable[LightNo], GL_SPOT_EXPONENT,   0.0);
#if 0
    GLfloat LModel_Ambient[]  = {1.0, 1.0,  1.0, 1.0};
    glLightModelfv (GL_LIGHT_MODEL_AMBIENT, LModel_Ambient);
#endif
  }

  if (Mode == OFF)
    glDisable (LightTable[LightNo]);
  else
  {
    glLightfv (LightTable[LightNo], GL_POSITION, Light_Position);
    glEnable (LightTable[LightNo]);
  }

  glPopMatrix ();                       // restore model-view matrix

  glEndList ();

#if 0
  if (Placement == STATIONARY)
    Set ();
#endif

  PRO_END
  return TRUE;
}


void LIGHT::InitShadow ()
{
  if (Shadow != NULL)
    Shadow->SetLight (this);
}


void LIGHT::CalcWTL ()
{
  double Len;
  double Angles[3] = {0.0, 0.0, 0.0};

  PRO_METHOD ("LIGHT", "")

  if (Mode == SPOTLIGHT && CutOff <= 90.0)
  {
    /* Compute angles of direction-vector for headlight rotation */
    Len = sqrt ((Dir.x * Dir.x) + (Dir.y * Dir.y));
    Angles[0] = 180.0 - GRAD (atan2(Len,   Dir.z));
    Angles[2] =         GRAD (atan2(Dir.y, Dir.x)) - 90.0;
  
    WTL = GEORPY2Frame (Pos.x, Pos.y, Pos.z, Angles[0], Angles[1], Angles[2]);
  }
  elsif (Mode == POSITIONAL)
    WTL = GEOMakeTrans (Pos.x, Pos.y, Pos.z);
  else
    WTL = Matrix_1;

  PRO_END
}


void LIGHT::SetMode (MODE Mode /* = DIRECTIONAL */)
{
  PRO_METHOD ("LIGHT", "\n")

  this->Mode = Mode;

  if (Mode == SPOTLIGHT && CutOff == 180.0)
    CutOff = 90.0;

  Create ();
  Build  ();

  CalcWTL ();

  Shadow->Create ();                    // Update shadow matrix

  PRO_END
}


void LIGHT::SetIntensity (double Value /* = 1.0 */)
{
  PRO_METHOD ("LIGHT", "\n")

  Intensity = LIMIT (0.0, Value, 1.0);
  Create ();

  PRO_END
}


void LIGHT::SetPosition  (POINT3D Position)
{
  PRO_METHOD ("LIGHT", "\n")

  Pos = Position;

  Create ();

  CalcWTL ();                           // Compute World_to_Light frame
  Shadow->Create ();                    // Update shadow matrix

  PRO_END
}
 

void LIGHT::SetDirection (VECTOR3D Direction)
{
  PRO_METHOD ("LIGHT", "\n")

  Dir = Direction;
  Create ();

  CalcWTL ();

  PRO_END
}
 

void LIGHT::SetPlacement (PLACEMENT Place)
{
  PRO_METHOD ("LIGHT", "\n")

  Placement = Place;
  Create ();

  PRO_END
}


void LIGHT::SetCutOff (double Angle /* = 90.0 */)
{
  PRO_METHOD ("LIGHT", "\n")

  // Angle < 90: directional light 
  CutOff = LIMIT (0.0, Angle, 90.0);
  Create ();

  Build  ();

  PRO_END
}


void LIGHT::SetSpotExponent (double Exponent /* = 1.0 */)
{
  PRO_METHOD ("LIGHT", "\n")

  SpotExp = Exponent;
  Create ();

  PRO_END
}


void LIGHT::SetAmbient (const COLOR &Color /* = COLOR(COLOR::BLACK) */)
{
  PRO_METHOD ("LIGHT", "\n")

  Ambient = Color;
  Create ();

  PRO_END
}


void LIGHT::SetDiffuse (const COLOR &Color /* = COLOR(COLOR::WHITE) */)
{
  PRO_METHOD ("LIGHT", "\n")

  Diffuse = Color;
  Create ();

  PRO_END
}


void LIGHT::SetSpecular (const COLOR &Color /* = COLOR(COLOR::WHITE) */)
{
  PRO_METHOD ("LIGHT", "\n")

  Specular = Color;
  Create ();

  PRO_END
}


void LIGHT::SetAttenuation (double Quadratic /* = 0.0 */, 
			    double Linear    /* = 0.0 */,
			    double Constant  /* = 1.0 */)
{
  PRO_METHOD ("LIGHT", "\n")

  Attenuation[0] = Quadratic;
  Attenuation[1] = Linear;
  Attenuation[2] = Constant;
  Create ();

  PRO_END
}


void LIGHT::SetSize (double Size /* = 0.0 */)
{
  PRO_METHOD ("LIGHT", "\n")
  
  if ((this->Size = Size) > 0.0)
    Build ();

  // Will not work, when not set 
  //  Create ();

  PRO_END
}


void LIGHT::SetShadow (SHADOW *Shadow)
{
  if (this->Shadow != NULL)             // destroy old shadow
    delete this->Shadow;
  
  this->Shadow = Shadow;

  InitShadow ();
}


LIGHT::NUMBER LIGHT::GetNumber () const
{
  return LightNo;
}
    

LIGHT::MODE LIGHT::GetMode () const
{
  return Mode;
}


double LIGHT::GetIntensity () const
{
  return Intensity;
}


POINT3D LIGHT::GetPosition () const
{
  return Pos;
}
 

VECTOR3D LIGHT::GetDirection () const
{
  return Dir;
}


LIGHT::PLACEMENT LIGHT::GetPlacement () const
{
  return Placement;
}


double LIGHT::GetCutOff () const
{
  return CutOff;
}


double LIGHT::GetSpotExponent () const
{
  return SpotExp;
}


COLOR LIGHT::GetAmbient () const
{
  return Ambient;
}


COLOR LIGHT::GetDiffuse () const
{
  return Diffuse;
}


COLOR LIGHT::GetSpecular () const
{
  return Specular;
}


double *LIGHT::GetAttenuation ()
{
  return Attenuation;
}


double LIGHT::GetSize () const
{
  return Size;
}


SHADOW *LIGHT::GetShadow ()
{
  return Shadow;
}


// We doesn't use emitting colors for the headlight, in order not to keep
// the original light-souce and not to sophisticate them.
void LIGHT::Build ()
{
  char     Name[20];
  POINT2D  Points[6];
  OBJECT  *ObjList = NULL;
  PART    *Part;

  PRO_METHOD ("LIGHT", "\n")

  if (HeadLight != NULL)
  {
    delete HeadLight;
    HeadLight = NULL;
  }

  if (Size == 0.0)
    PRO_ABORT ("Not visible")

  PRO1 ("Size: %.2f\n", Size)

  if (Mode == SPOTLIGHT && CutOff <= 90.0)
  {
    // Create corpus
    ObjList = 
      OBJCreateCylinder (GEORPY2Frame(0.0, 0.0, 0.1*Size, 0.0, 0.0, 22.5),
			 2.0*Size, 3.0*Size, 360.0, 8, COLOR::BLACK);
    
    // Add front glas
    OBJAddObject (ObjList,
		  OBJCreateCylinder(Matrix_1, 1.5*Size, 0.1*Size, 360.0, 12, 
				    COLOR::YELLOW));
    // Coordinates of shutters
    Points[0].x =      Size;   Points[0].y = 0.0;
    Points[1].x =  2.0*Size;   Points[1].y =     Size;
    Points[2].x =  2.0*Size;   Points[2].y = 2.0*Size;
    Points[3].x = -2.0*Size;   Points[3].y = 2.0*Size;
    Points[4].x = -2.0*Size;   Points[4].y =     Size;
    Points[5].x = -    Size;   Points[5].y = 0.0;

    // Add shutter one
    OBJAddObject (ObjList, 
		  OBJCreateSweep(GEORPY2Frame(1.8*Size, 0.0, 0.0, 
					      CutOff-90.0, 0.0, 270.0),
				 6, Points, 0.1*Size, COLOR::MEDRED));

    // Add shutter two
    OBJAddObject (ObjList, 
		  OBJCreateSweep(GEORPY2Frame(-1.8*Size, 0.0, 0.0, 
					      CutOff-90.0, 0.0, 90.0),
				 6, Points, 0.1*Size, COLOR::MEDRED));
  }
  elsif (Mode == POSITIONAL)            // Draw sphere for positional light
  {
    // Create bulb
    ObjList = OBJCreateSphere (Matrix_1, 2.0*Size, 180.0, 360.0, 12, 
			       COLOR::YELLOW);
    // Add thread
    OBJAddObject (ObjList,
		  OBJCreateCylinder(GEORPY2Frame(0.0, 0.0, 2.0*Size, 
						 0.0, 0.0, 0.0),
				    Size, 2.0*Size, 360.0, 10, COLOR::SILVER));
  }

  if (ObjList != NULL)
  {
    sprintf (Name, "Light%02d", LightNo);
    Part = PARTCreate (Name, ObjList);
    HeadLight = new PART3D (Part);
  }

  PRO_END
}


void LIGHT::Draw ()
{
  if (HeadLight == NULL)
    return;

  // It makes no sense to draw directional light sources at infinity.
  if (Mode == POSITIONAL || Mode == SPOTLIGHT) 
  {
    // TODO: allgemeine Routine in Part (einschlielich Positionsmatrix ?)
    glPushMatrix ();
      glMultMatrixd ((GLdouble *)WTL.m);
      HeadLight->Draw ();
    glPopMatrix ();
  }
}


