/* **************************************************************** */
/* ** MATERIAL is a simple version of OpenGL material handling.  ** */
/* ** We use only one value for properties like ambient, diffuse ** */
/* ** specular light, which will be used for the red, green and  ** */
/* ** blue component. The Colorvalue (r,g,b or index into        ** */
/* ** predefined table) will be multiplied by these values       ** */
/* ** Also we have predefines for some materials (MaterialID)    ** */
/* **************************************************************** */
#include <assert.h>

#include <libtools.h>

#include "libcad3d.h"


typedef struct
{
  COLOR   Color;
  GLfloat Ambient;
  GLfloat Diffuse;
  GLfloat Specular;
  GLfloat Emission;
  GLfloat Shininess;
  char    TextureFile[PATH_MAX+1];
} MATERIAL_DATA;


char *MaterialNames[] =     // => Dialog oder fr Speicherung beibehalten ?
{
  "values",
  "plastic",
  "metal",
  "glass",
  "paper",
  "water",
  "wood",
  "stone",
  "cloth",
  "mirror",
};


GLfloat LightState[][4] =
{
  {  1.0,    1.0,    1.0,   1.0},   // Normal-Mask
  {  1.0,   -1.0,   -1.0,   1.0},   // Select-Mask
  { -1.0,    1.0,   -1.0,   1.0},   // Collision-Mask
  { -0.5,   -0.5,    1.0,   1.0}    // OOW-Mask
};


MATERIAL_DATA MaterialTable[] = 
{
  { // PLASTIC
    COLOR(COLOR::OBJECT),     // Color
    0.30,                     // Ambient
    0.75,                     // Diffuse
    0.80,                     // Specular
    0.00,                     // Emission
    50.0,                     // Shininess
    ""                        // Texture
  },
  { // METAL
    COLOR(COLOR::OBJECT),
    0.40,                     // e.g. Gold: * 1.0/0.84/0.0 = 0.40, 0.336, 0.0
    0.90,
    1.00,
    0.00,
    64.0, // (max)
    ""
  },
  { // GLASS
    COLOR(COLOR::OBJECT),
    0.00,
    0.00,
    0.00,
    1.00,
    64.0,
    ""
  },
  { // PAPER
    COLOR(COLOR::OBJECT),
    0.00,
    0.00,
    1.00,
    0.00,
    0.00,
    ""
  },
  { // WATER
    COLOR(COLOR::MEDBLUE),
    0.00,
    0.00,
    1.00,
    0.00,
    40.0,
    ""
  },
  { // WOOD
    COLOR(COLOR::TEXTURE),
    0.00,                     // Ambient
    1.00,                     // Diffuse
    0.00,                     // Specular
    0.00,                     // Emission
    0.00,                     // Shininess
    "wood.xpm",               // Texture
  },
  { // STONE
    COLOR(COLOR::GREY),
    0.00,
    0.00,
    1.00,
    0.00,
    0.00,
    ""
  },
  { // CLOTH
    COLOR(COLOR::OBJECT),
    0.00,
    0.00,
    1.00,
    0.00,
    0.00,
    ""
  },
  { // MIRROR
    COLOR(COLOR::OBJECT),
    0.00,
    0.00,
    1.00,
    0.00,
    64.0,
    ""
  }
};

#define MAX_MATERIAL (int)(sizeof(MaterialTable) / sizeof(MATERIAL_DATA))


MATERIAL::MATERIAL (COLOR::ID ColorID    /* = COLOR::OBJECT */, 
		    int       Mode       /* = MODE_SOLID    */) :
  CAD_FILE   (),
  UNIT       (),
  MaterialID (VALUES),
  Color      (ColorID),
  Mode       (Mode),
  Line       (0),
  Texture    (NULL)
{
  PRO_METHOD ("MATERIAL", "")

  Initialize ();

  PRO_END
};


MATERIAL::MATERIAL (COLOR Color) :
  CAD_FILE   (),
  UNIT       (),
  MaterialID (VALUES),
  Color      (Color),
  Mode       (SOLID),
  Line       (0),
  Texture    (NULL)
{
  PRO_METHOD ("MATERIAL", "")

  Initialize ();

  PRO_END
}


MATERIAL::MATERIAL (ID MaterialID) :
  CAD_FILE   (),
  UNIT       (),
  MaterialID (MaterialID),
  Color      (COLOR::RGB),
  Mode       (SOLID),
  Line       (0),
  Texture    (NULL)
{
  PRO_METHOD ("MATERIAL", "")

  assert (MaterialID >= VALUES && MaterialID < MAX_MATERIAL);

#if 1
  if (MaterialID != VALUES)
    if (MaterialTable[MaterialID].TextureFile[0] == '\0')
    {
      Ambient   = MaterialTable[MaterialID].Ambient;
      Diffuse   = MaterialTable[MaterialID].Diffuse;
      Specular  = MaterialTable[MaterialID].Specular;
      Emission  = MaterialTable[MaterialID].Emission;
      Shininess = MaterialTable[MaterialID].Shininess;
    }
    else
    {
#if 0
      Texture = new TEXTURE (MaterialTable[MaterialID].TextureFile);
      SetColorID (COLOR::TEXTURE);
#else
      LoadTexture (MaterialTable[MaterialID].TextureFile);
#endif
    }
#else
  Initialize ();
#endif

  PRO_END
}


MATERIAL::MATERIAL (const MATERIAL &Material) :
  CAD_FILE   (),
  UNIT       (),
  MaterialID (Material.MaterialID),
  Color      (Material.Color),
  Mode       (Material.Mode),
  Line       (Material.Line),
  Ambient    (Material.Ambient),
  Diffuse    (Material.Diffuse),
  Specular   (Material.Specular),
  Emission   (Material.Emission),
  Shininess  (Material.Shininess),
  Texture    (NULL)
{
  PRO_METHOD ("MATERIAL", "")

  if (Material.HasTexture())
    Texture = new TEXTURE (Material.Texture);

  PRO_END
}


MATERIAL::MATERIAL (const char *TextureFile) :
  CAD_FILE   (),
  UNIT       (),
  MaterialID (VALUES),
  Color      (COLOR::TEXTURE),
  Mode       (SOLID),
  Line       (0)
{
  PRO_METHOD ("MATERIAL", "")

  Initialize ();

  Texture = new TEXTURE (TextureFile);

  PRO_END
}


MATERIAL::~MATERIAL ()
{
  PRO_METHOD ("MATERIAL~", "")

  if (Texture != NULL)
    delete Texture;

  PRO_END
}


void MATERIAL::Initialize ()
{
  if (MaterialID != VALUES)
    if (MaterialTable[MaterialID].TextureFile[0] == '\0')
    {
      Ambient   = MaterialTable[MaterialID].Ambient;
      Diffuse   = MaterialTable[MaterialID].Diffuse;
      Specular  = MaterialTable[MaterialID].Specular;
      Emission  = MaterialTable[MaterialID].Emission;
      Shininess = MaterialTable[MaterialID].Shininess;
    }
    else
    {
#if 0
      Texture = new TEXTURE (MaterialTable[MaterialID].TextureFile);
      SetColorID (COLOR::TEXTURE);
#else
      LoadTexture (MaterialTable[MaterialID].TextureFile);
#endif
    }
  else
  {
    Ambient   = 0.0;
    Diffuse   = 1.0;
    Specular  = 0.0;
    Emission  = 0.0;
    Shininess = 0.0;
  }
}


bool MATERIAL::Read (DATAFILE &File, double Version)
{
  bool DoTexture;

  PRO_METHOD ("MATERIAL", "\n")

  if (File.Parse("<\\[MATERIAL\\]>\n") < 0)
    PRO_RETURN ("Key [MATERIAL] not found", FALSE)
  
  if (! CAD_FILE::Read(File, Version))
    PRO_RETURN ("Reading FILE failed", FALSE)

  if (File.Parse("  <ID:>     (%d)\n", &MaterialID) != 1)
    PRO_RETURN ("reading DATA 1 failed", FALSE)

  if (! Color.Read (File, Version))
    PRO_RETURN ("reading COLOR failed", FALSE)

  if (File.Parse("  <Mode:>       (NONE|WIRE|LINE|SOLID|INVALID)\n"
#if 0
		 "  <Render:>     \n"
#endif
		 "  <Line:>       (%d)\n"
		 "  <Ambient:>    (%f)\n"
		 "  <Diffuse:>    (%f)\n"
		 "  <Specular:>   (%f)\n"
		 "  <Emission:>   (%f)\n"
		 "  <Shininess:>  (%f)\n"
		 "  <Texture:>    (NO|YES)\n",
		 &Mode, 
		 &Line, &Ambient, &Diffuse, &Specular, &Emission,
		 &Shininess, &DoTexture) != 8)
    PRO_RETURN ("Reading DATA 2 failed", FALSE)
  
  if (DoTexture == TRUE)
  {
    if (Texture == NULL)
      Texture = new TEXTURE (NO_TEXTURE);
    if (! Texture->Read(File, Version))
      PRO_RETURN ("Reading TEXTURE failed", FALSE);
  }

  Initialize ();

  PRO_END
  return TRUE;
}


bool MATERIAL::Write (FILE *fp /* = NULL */, int Indent /* = 0 */) const
{
  PRO_METHOD ("MATERIAL", "\n")

  if (fp == NULL)
    fp = stdout;

  Space   (fp, Indent);
  fprintf (fp, "[MATERIAL]\n");
  Indent += INDENT;

  if (! CAD_FILE::Write(fp, Indent))
    PRO_RETURN ("Writing failed", FALSE)

  Space   (fp, Indent);
  fprintf (fp, "ID:         %d\n",   MaterialID);

  Color.Write (fp, Indent);

  Space   (fp, Indent);
  fprintf (fp, "Mode:         ");
  switch (Mode)
  {
    case NONE :
      fprintf (fp, "NONE\n");
      break;
    case WIRE :
      fprintf (fp, "WIRE\n");
      break;
    case LINE :
      fprintf (fp, "LINE\n");
      break;
    case SOLID :
      fprintf (fp, "SOLID\n");
      break;
    default :
      fprintf (fp, "INVALID\n");
  }

#if 0
  Space   (fp, Indent);
  fprintf (fp, "Render:       -\n");
#endif

  Space   (fp, Indent);
  fprintf (fp, "Line:       %d\n", Line);

  Space   (fp, Indent);
  fprintf (fp, "Ambient:    %.3f\n", Ambient);

  Space   (fp, Indent);
  fprintf (fp, "Diffuse:    %.3f\n", Diffuse);

  Space   (fp, Indent);
  fprintf (fp, "Specular:   %.3f\n", Specular);

  Space   (fp, Indent);
  fprintf (fp, "Emission:   %.3f\n", Emission);

  Space   (fp, Indent);
  fprintf (fp, "Shininess:  %.3f\n", Shininess);

  Space   (fp, Indent);
  fprintf (fp, "Texture:    %s\n", (Texture!=NULL)?"YES":"NO");
  
  if (Texture != NULL)
    Texture->Write (fp, Indent);

  PRO_END
  return TRUE;
}


void MATERIAL::Set () const
{
#if 1
  int     count;
  COLOR   Color;
  GLfloat Ambient[4];
  GLfloat Specular[4];
  GLfloat Diffuse[4];
  GLfloat Emission[4];
  GLfloat Shininess;
#endif

  PRO_METHOD ("MATERIAL", "")

#if 1
  if (HasTexture())
    Texture->Set (Color);
  else
  {
    // TODO: ColorID=Texture & Keine Textur -> Part-Farbe (COLOR_DEFAULT)
    if (MaterialID == VALUES)
      for (count = RED; count <= ALPHA; ++count)
      {
	Ambient[count]  = this->Ambient  * this->Color[count];
	Diffuse[count]  = this->Diffuse  * this->Color[count];
	Specular[count] = this->Specular * this->Color[count];
	Emission[count] = this->Emission * this->Color[count];
	Shininess       = this->Shininess;
      }
    else
    {
      Color = MaterialTable[MaterialID].Color;
      if (Color.GetID() == COLOR::OBJECT)
	Color = this->Color;
      for (count = RED; count <= ALPHA; ++count)
      {
	Ambient[count]  = MaterialTable[MaterialID].Ambient  * Color[count];
	Diffuse[count]  = MaterialTable[MaterialID].Diffuse  * Color[count];
	Specular[count] = MaterialTable[MaterialID].Specular * Color[count];
	Emission[count] = MaterialTable[MaterialID].Emission * Color[count];
	Shininess       = MaterialTable[MaterialID].Shininess;
      }
    }

    // GL_FONT_AND_BACK, falls nicht backface culling ?
    glMaterialfv (GL_FRONT, GL_AMBIENT,   Ambient); 
    glMaterialfv (GL_FRONT, GL_SPECULAR,  Specular);
    glMaterialfv (GL_FRONT, GL_DIFFUSE,   Diffuse);
    glMaterialfv (GL_FRONT, GL_EMISSION,  Emission);
    glMaterialf  (GL_FRONT, GL_SHININESS, Shininess);
  }
#else
  printf ("Set Material: %d\n", Id);
  //  Call ();
  glCallList (Id);
#endif

  PRO_END
}


bool MATERIAL::Create ()
{
  // Testen ob Display-List erstellt werden kann, oder ob sich diese
  // wegen Verschachtelung mit anderen Display-Lists nicht vertrgt (z.B. 
  // PART::Create - evtl. statt MATERIAL::Create dort MATERAIL::Set aufrufen.
  return TRUE;
}


void MATERIAL::SetColorID (COLOR::ID ColorID /* = COLOR::OBJECT */)
{
  PRO_METHOD ("MATERIAL", "")

  PRO1 ("\\%d...", ColorID);

  Color.SetID (ColorID);
  MaterialID = VALUES;

  PRO_END
}


void MATERIAL::SetColor (COLOR Color /* = COLOR(COLOR::OBJECT) */)
{
  PRO_METHOD ("MATERIAL", "")

  this->Color = Color;
  MaterialID = VALUES;

  PRO_END
}


void MATERIAL::SetColor (double r, double g, double b, 
			 double Alpha /* = 1.0 */)
{
  PRO_METHOD ("MATERIAL", "\n")

  Color.SetValues (r, g, b, Alpha);

  PRO_END
}


void MATERIAL::SetAlpha (double Alpha /* = 1.0 */)
{
  PRO_METHOD ("MATERIAL", "\n")

  Color.SetAlpha (Alpha);

  PRO_END
}


void MATERIAL::SetMode (int Mode /* = MODE_SOLID */)
{
  PRO_METHOD ("MATERIAL", "")

  assert (Mode >= (int)NONE && Mode <= (int)SOLID);

  this->Mode = Mode;

  PRO_END
}


void MATERIAL::SetLine (int Width /* = 0 */)
{
  PRO_METHOD ("MATERIAL", "")

  assert (Mode >= 0);

  Line = Width;

  PRO_END
}


/*
  -1: use Material Values (Ambient, Specular, Diffuse, ...)
   0: part has texture
  >0: index into predefiend material table
*/
void MATERIAL::SetMaterialID (ID MaterialID /* = VALUES */)
{
  PRO_METHOD ("MATERIAL", "")

  this->MaterialID = MaterialID;
  if (MaterialTable[MaterialID].TextureFile[0] == '\0')
  {
    Ambient   = MaterialTable[MaterialID].Ambient;
    Diffuse   = MaterialTable[MaterialID].Diffuse;
    Specular  = MaterialTable[MaterialID].Specular;
    Emission  = MaterialTable[MaterialID].Emission;
    Shininess = MaterialTable[MaterialID].Shininess;
    SetColorID (COLOR::RGB);
  }
  else
    LoadTexture (MaterialTable[MaterialID].TextureFile);

  PRO_END
}


bool MATERIAL::LoadTexture (const char *FileName)
{
  PRO_METHOD ("MATERIAL", "\n")

  try
  {
    if (Texture != NULL)
      Texture->SetTextureFile (FileName);
    else
      Texture = new TEXTURE (FileName);
    SetColorID (COLOR::TEXTURE);
  }
  catch (ERROR Error)
  {
    PRO_RETURN ("Texture Load Error", FALSE)
  }

  PRO_END
  return TRUE;
}


void MATERIAL::SetTextureID (int TextureID /* = NO_TEXTURE */)
{
  PRO_METHOD ("MATERIAL", "")

  if (TextureID == NO_TEXTURE)
    CleanTexture ();
  else
  {
    SetColorID (COLOR::TEXTURE);
    if (Texture != NULL)
      Texture->SetTextureID (TextureID);
    else
      Texture = new TEXTURE (TextureID);
  }

  PRO_END
}


void MATERIAL::SetAmbient (double Ambient /* = 1.0 */)
{
  PRO_METHOD ("MATERIAL", "")

  this->Ambient = LIMIT (0.0, Ambient, 1.0);

  PRO_END
}


void MATERIAL::SetDiffuse (double Diffuse /* = 0.0 */)
{
  PRO_METHOD ("MATERIAL", "")

  this->Diffuse = LIMIT (0.0, Diffuse, 1.0);

  PRO_END
}


void MATERIAL::SetSpecular (double Specular /* = 1.0 */)
{
  PRO_METHOD ("MATERIAL", "")

  this->Specular = LIMIT (0.0, Specular, 1.0);

  PRO_END
}


void MATERIAL::SetEmission (double Emission /* = 0.0 */)
{
  PRO_METHOD ("MATERIAL", "")

  this->Emission = LIMIT (0.0, Emission, 1.0);

  PRO_END
}


void MATERIAL::SetShininess (double Shininess /* = 0.0 */)
{
  PRO_METHOD ("MATERIAL", "")

  this->Shininess = Shininess;

  PRO_END
}


void MATERIAL::SetTexture (TEXTURE *NewTexture /* = NULL */)
{
  PRO_METHOD ("MATERIAL", "")

  if (Texture == NewTexture)
    return;

  CleanTexture ();
  Texture = NewTexture;

  if (Texture != NULL)
  {
    SetColorID (COLOR::TEXTURE);
    Texture->Create ();
  }

  PRO_END
}


GLfloat *CalcColor (int State, GLfloat *Color)
{
  int count;

  Color[0] = 0.0;
  Color[1] = 0.0;
  Color[2] = 0.0;
  Color[3] = 1.0;

  for (count = 0; count < 3; ++count)
    if (State & (1<<count))
    {
      Color[0] = LIMIT (0.0, Color[0]+LightState[count+1][0], 1.0);
      Color[1] = LIMIT (0.0, Color[1]+LightState[count+1][1], 1.0);
      Color[2] = LIMIT (0.0, Color[2]+LightState[count+1][2], 1.0);
    } 

  return Color;
} 


COLOR::ID MATERIAL::GetColorID () const
{
  return Color.GetID ();
}


COLOR &MATERIAL::GetColor ()
{
  return Color;
}


double MATERIAL::GetRed () const
{
  return Color[RED];
}


double MATERIAL::GetGreen () const
{
  return Color[GREEN];
}


double MATERIAL::GetBlue () const
{
  return Color[BLUE];
}


double MATERIAL::GetAlpha () const
{
  return Color[ALPHA];
}


int MATERIAL:: GetMode () const
{ 
  return Mode;
}


int MATERIAL::GetLine  () const
{
  return Line;
}


MATERIAL::ID MATERIAL::GetMaterialID () const
{ 
  return MaterialID;
}


double MATERIAL::GetAmbient () const
{ 
  return Ambient;
}


double MATERIAL::GetDiffuse () const 
{ 
  return Diffuse;
}


double MATERIAL::GetSpecular () const
{ 
  return Specular;
}


double MATERIAL::GetEmission () const
{ 
  return Emission;
}


double MATERIAL::GetShininess () const
{
  return Shininess;
}


TEXTURE *MATERIAL::GetTexture ()
{
  return Texture;
}


bool MATERIAL::HasTexture () const
{
  return (Color.GetID() == COLOR::TEXTURE && Texture != NULL);
}


void MATERIAL::CleanTexture ()
{
  if (Texture != NULL)
    delete Texture;
  Texture = NULL;

  SetColorID (COLOR::OBJECT);
}
