#include <math.h>
#include <assert.h>

#include "libcad3d.h"


// static double AlphaNo = 0;
static double GetAveragePolygonSize (PART *Part);


PART3D::PART3D (      PART     *Part     /* = NULL */, 
		      WORLD    *World    /* = NULL */,
		const MATERIAL &Material /* = MATERIAL(COLOR::OBJECT) */) :
  CAD_FILE (),
  UNIT     (),
  Part     (Part),
  Render   (WINDOW3D::FACES),
  Type     (TRANS | ROT),
  KinType  (TRANS | ROT),
  Joint    (0),
  Length   (0.0),
  Offset   (0.0),
  State    (STATE_NORMAL),
  Mode     (DRAW_PART),
  Material (Material),
  T        (Matrix_1),
  Rel      (Matrix_1),
  WTP      (Matrix_1),
  Box      (0.0, 0.0, 0.0),
  CS       (CS_ID),
  Parent   (NULL),
  Prev     (NULL),
  Chain    (NULL),
  Attach   (NULL),
  Next     (NULL),
  World    (World)
{
  PRO_METHOD ("PART3D", "\n")

  Initialize ();
 
  PRO_END
}


PART3D::PART3D (PART3D *Part, WORLD *World /* = NULL */) :
  CAD_FILE (),
  UNIT     (),
  Joint    (0),
  State    (STATE_NORMAL),
  Box      (0.0, 0.0, 0.0),
  CS       (CS_ID),
  Parent   (NULL),
  Prev     (NULL),
  Chain    (NULL),
  Attach   (NULL),
  Next     (NULL),
  World    (World)
{
  PRO_METHOD ("PART3D", "\n")

  if (Part == NULL)
    throw ERROR (ERROR::NO_DATA);

  if (Part != this)           // Copy part data
  {
    this->Part = PARTCopyPart (Part->Part);

    Render   = Part->Render;
    Type     = Part->Type;
    KinType  = Part->KinType;
    Joint    = Part->Joint;
    Length   = Part->Length;
    Offset   = Part->Offset;
    Mode     = Part->Mode;
    DHParam  = Part->DHParam;
    Material = Part->Material;
    T        = Part->T;
    Rel      = Part->Rel;
    WTP      = Part->WTP;
    Box      = Part->Box;
  }

  Initialize ();

  PRO_END
}


PART3D *PART3D::Copy (DEVICE3D *Parent)
{
  PART3D   *Chain;
  DEVICE3D *Attach;
  PART3D   *Result = new PART3D (this, World);

  Result->Parent = Parent;
  Result->T      = T;
  Result->Rel    = Rel;

  Chain = this->Chain;
  while (Chain != NULL)
  {
    Result->AddChain (Chain->Copy(Parent));
    Chain = Chain->Next;
  }

  Attach = this->Attach;
  while (Attach != NULL)
  {
    Result->AttachDev (Attach->Copy());
    Attach = Attach->GetNext ();
  }

  return Result;
}


void PART3D::Initialize ()
{
  if (Part != NULL)
  {
    Box.SetRange (Part->Info.Min.x, Part->Info.Max.x,
		  Part->Info.Min.y, Part->Info.Max.y,
		  Part->Info.Min.z, Part->Info.Max.z);
    SetNormalAngle ();
    Create ();
  }
}


PART3D::~PART3D () 
{
  PRO_METHOD ("~PART3D", "")

  if (Part != NULL)
    PARTFreePart (Part);

  PRO_END
}


PART3D *PART3D::operator + (PART3D &Part) // Union
{
  return new UNION (World, "union", Matrix_1, this, &Part, Material);
}


PART3D *PART3D::operator - (PART3D &Part) // Difference
{
  return new DIFFERENCE (World, "difference", Matrix_1, this, &Part, Material);
}


PART3D *PART3D::operator % (PART3D &Part) // Intersection
{
  return new INTERSECTION (World, "intersection", Matrix_1, this, &Part, 
			          Material);
}


bool PART3D::Read (DATAFILE &File, double Version)
{
  char FileName[4096] = "";

  PRO_METHOD ("PART3D", "\n")

  if (Part != NULL)                     // Remove an existing part
    PARTFreePart (Part);
  Part = NULL;

  if (File.Parse("<\\[PART\\]>\n") < 0)
    PRO_RETURN ("Key [PART] not found", FALSE)

  if (! CAD_FILE::Read(File, Version))
    PRO_RETURN ("Reading FILE failed", FALSE)

  if (File.Parse("  <File:>    '(%4096s)'\n"
		 "  <Render:>  (POINTS|WIRE_FRAME|HIDDEN_LINE|FACES)\n"
		 "  <Type:>    {ROT|TRANS}\n",
		 &FileName, &Render, &Type) < 3)
    PRO_RETURN ("Reading DATA 1 failed", FALSE)

  if (! Material.Read(File, Version))
    PRO_RETURN ("Reading MATERIAL failed", FALSE)

  if (! Joint.Read(File, Version))
    PRO_RETURN ("Reading JOINT failed", FALSE)

  if (File.Parse("  <Length:>  (%f)\n"
		 "  <Offset:>  (%f)\n"
		 "  <Mode:>    {PART|BOX|AXES}\n",
		 &Length, &Offset, &Mode) < 3)
    PRO_RETURN ("Reading DATA 2 failed", FALSE)

  if (FileName[0] != '\0')
    if ((Part = PARTLoad(FileName, "PART", TRUE)) == NULL)
      PRO_RETURN ("Reading PART failed", FALSE)

  Create ();

  PRO_END
  return TRUE;
}


bool PART3D::Write (FILE *fp /* = NULL */, int Indent /* = 0 */) const
{
  PART3D *ChainPtr;

  PRO_METHOD ("PART3D", "\n")

  if (fp == NULL)
    fp = stdout;

  Space   (fp, Indent);
  fprintf (fp, "[PART]\n");
  Indent += INDENT;

  if (! CAD_FILE::Write(fp))
    PRO_RETURN ("Write failed", FALSE)

  Space (fp, Indent);
  if (Part == NULL || Part->FileName == NULL)
    fprintf (fp, "File:     ''\n");
  else
    fprintf (fp, "File:     '%s'\n", Part->FileName);

  Space (fp, Indent);
  fprintf (fp, "Render:   ");
  switch (Render)
  {
    case      WINDOW3D::POINTS : fprintf (fp, "POINTS\n");
                                 break;
    case  WINDOW3D::WIRE_FRAME : fprintf (fp, "WIRE_FRAME\n");
                                 break;
    case WINDOW3D::HIDDEN_LINE : fprintf (fp, "HIDDEN_LINE\n");
                                 break;
    case       WINDOW3D::FACES : fprintf (fp, "FACES\n");
                                 break;
                       default : fprintf (fp, "UNKOWN\n");
  }

  Space   (fp, Indent);
  fprintf (fp, "Type:    ");
  if (Type & TRANS)
    fprintf (fp, " TRANS");
  if (Type & ROT)
    fprintf (fp, " ROT");
  fprintf (fp, "\n");

  if (! Material.Write(fp, Indent))
    throw ERROR (-2, "Writing MATERIAL failed");

  if (! Joint.Write(fp, Indent))
    PRO_RETURN ("Writing JOINT failed", FALSE)

  Space   (fp, Indent);
  fprintf (fp, "Length:   %f\n", Length);

  Space   (fp, Indent);
  fprintf (fp, "Offset:   %f\n", Offset);

  Space   (fp, Indent);
  fprintf (fp, "Mode:   ");
  if (Mode & DRAW_PART)
    fprintf (fp, "  PART");
  if (Mode & DRAW_BOX)
    fprintf (fp, "  BOX");
  if (Mode & DRAW_AXES)
    fprintf (fp, "  AXES");
  fprintf (fp, "\n");

  Space   (fp, Indent);
  fprintf (fp, "Chain:    %d\n", GetPartNumber());

  ChainPtr = Chain;
  while (ChainPtr != NULL)
  {
    ChainPtr->Write (fp, Indent+2);
    fprintf (fp, "\n");
    ChainPtr = ChainPtr->GetNext ();
  }
	   
  // TODO: Matrizen

  PRO_END
  return TRUE;
}


bool PART3D::Create ()
{
  if (! UNIT::Create())
    return FALSE;

  return Create (COLOR(COLOR::OBJECT));
}


bool PART3D::Create (const COLOR &Color)
{
  int       count, ColorID = -1;
  int       MatID;
  double    Stretch = 1.0;
  double    OffsX = 0.0, OffsY =0.0;
  H_MATRIX  CSMatr;
  POINT2D   TexCoord;
  MATERIAL *ObjMat = NULL;
  MATERIAL *Line = NULL;
  TEXTURE  *Texture = NULL;
  POINT3D   Point, CTP;
  VECTOR3D  Normal;
  POLYGON  *Polygon;
  OBJECT   *Obj;

  PRO_METHOD ("PART3D", "\n")

  if (Part == NULL)
    PRO_RETURN ("No Part defined", FALSE)

#if 0
  glPolygonMode (GL_FRONT, (GLenum)Render[0]);
  glPolygonMode (GL_BACK,  (GLenum)Render[1]);
#endif

  MatID = Material.GetColorID ();

  // Create new material if we shouldn't use the parts own material.
  // If we would create a copy of the part's material, the texture would
  // be destroyed at end of this method. 
  if (Color.GetID() == COLOR::OBJECT)
    if (Material.HasTexture())
      ObjMat = &Material;
    else
      ObjMat = new MATERIAL (Material);
  else
    ObjMat = new MATERIAL (Color);

  Id = OpenGL->GetID (Id, GetName());
  glNewList (Id, GL_COMPILE);
  glIndexi  (Id);

  ObjMat->Set ();

  if (ObjMat->GetLine() > 0)
  {
    Line = new MATERIAL (COLOR::BLACK);
    Line->SetLine  (ObjMat->GetLine());
    glEnable      (GL_STENCIL_TEST);
    glClear       (GL_STENCIL_BUFFER_BIT);
    glStencilFunc (GL_ALWAYS, 0, 1);
    glStencilOp   (GL_INVERT, GL_INVERT, GL_INVERT);
  }

  if (Color.GetID() == COLOR::OBJECT && ObjMat->HasTexture())
  {
    Texture = ObjMat->GetTexture ();
    Stretch = GetTextureFactor ();
    // TODO: Test using polar-coordinates to compute the texture-coordinates
    PRO1 ("STRETCH: %f\n", Stretch)
  }

  Obj = Part->ObjectList;
  while (Obj != NULL)
  {
    if (Color.GetID() == COLOR::OBJECT && State == STATE_NORMAL &&
	MatID == COLOR::OBJECT && ColorID != abs(Obj->ColorIndex))
    {
      ColorID = abs (Obj->ColorIndex);
      ObjMat->SetColorID ((COLOR::ID)ColorID);
      ObjMat->Set ();
    }

    switch (Obj->Type)
    {
      case PART_PLANE :
	Polygon = Obj->Data.Polyeder->List;
	while (Polygon != NULL)
	{
	  if (ObjMat->GetLine() > 0)        // Draw Outline...
	  {
	    DrawWireFrame (Obj, Polygon, Line);
	    ObjMat->Set ();
	    glStencilFunc (GL_EQUAL, 0, 1);
	    glStencilOp (GL_KEEP, GL_KEEP, GL_KEEP);
	  }

	  if (Color.GetID() == COLOR::OBJECT && Texture != NULL)
	  {
	    PARTCalcFaceCS (Part, Polygon, &CSMatr);
	    CSMatr = GEOMatrixInv (CSMatr);
	    OffsX = atan2 (CSMatr.m[P][Y], CSMatr.m[P][X]);
	    OffsY = atan2 (CSMatr.m[P][Z], CSMatr.m[P][X]);
	  }

#if 0
	  // Only for debug purpose to test tesselation
	  glBegin (GL_TRIANGLE_FAN);
#else
	  glBegin (GL_POLYGON);
#endif

	  Normal = Vector_0;
	  for (count = 0; count < Polygon->Number; ++count)
	  {
	    // Get the normal vector of each point
	    if (Polygon->Normals != NULL)
	      Normal = Polygon->Normals[count];

	    // Use the polygon normal if the normal vector is zero
	    if (Normal.x == 0.0 && Normal.y == 0.0 && Normal.z == 0.0)
	      Normal = Polygon->Info.n;	      

	    // Set the normal vector
	    glNormal3d (Normal.x, Normal.y, Normal.z);

	    Point = Obj->VertexList[Polygon->Vertex[count]];

	    // Define texture coordinates
	    if (Color.GetID() == COLOR::OBJECT && Texture != NULL)
	    {
#if 1
	      // Point relative to Face-CS: z-coordinate is zero:
	      CTP = GEOMatrixMult (CSMatr, Point);
	      TexCoord.x = (OffsX + CTP.y*Stretch) * Texture->GetStretchS ();
	      TexCoord.y = (OffsY + CTP.z*Stretch) * Texture->GetStretchT ();
	      glTexCoord2d (TexCoord.x, TexCoord.y);
#else
	      // Disable for testing generating of automatic texture coords
#endif
	    }

	    // set the polygons's vertex
	    glVertex3d (Point.x, Point.y, Point.z);
	  }
	  glEnd ();

	  if (ObjMat->GetLine() > 0)            // Draw Outline again...
	  {
	    glStencilFunc (GL_ALWAYS, 0, 1);
	    glStencilOp (GL_INVERT, GL_INVERT, GL_INVERT);
	    DrawWireFrame (Obj, Polygon, Line);
	  }

	  Polygon = Polygon->Next;
	}
	break;

      case PART_CURVE :
        glLineWidth (1.0);           // ObjMat.GetLine() ? 
	glBegin (GL_LINE_STRIP);
          for (count = 0; count < Obj->Data.Curve->Number; ++count)
	  {
	    Point = Obj->VertexList[Obj->Data.Curve->List[count]];
	    glVertex3d (Point.x, Point.y, Point.z);
	  }
	glEnd ();
	glLineWidth (1.0);

      case NONE :
      default :
	break;
    }
    Obj = Obj->Next;
  }

  if (ObjMat->GetLine() > 0)
  {
    glDisable (GL_STENCIL_TEST);
    delete Line;
  }

  if (Color.GetID() == COLOR::OBJECT && Texture != NULL)
    glDisable (GL_TEXTURE_2D);

  glEndList ();

  if (Color.GetID() != COLOR::OBJECT || ! Material.HasTexture())
    delete ObjMat;

  PRO_END
  return TRUE;
}


void PART3D::Draw ()
{
  int count;

  // Draw part
  if (Mode & DRAW_PART)
    Set ();

  // Draw bounding box
  if (Mode & DRAW_BOX)
    Box.Set ();

  if (CS >= 0)
  {
    // Draw part coordinate system
    if (State == STATE_SELECTED)
      glCallList (CS);

    // Draw user coordinate systems
    if ((Mode & DRAW_AXES) || State == STATE_SELECTED)
      for (count = 0; count < Part->CSNo; ++count)
      {
	glPushMatrix  ();
	glMultMatrixd ((GLdouble *)Part->CS[count].m);
	glCallList    (CS);
	glPopMatrix   ();
      }
  }
}


void PART3D::Scale (double ScaleX, double ScaleY, double ScaleZ, 
		    int Mask /* = 0 */)
{
  PRO_METHOD ("PART3D", "\n")

  if (Part != NULL)
    PARTScalePart (Part, ScaleX, ScaleY, ScaleZ);

  // Adjust translatoric ranges ? 
  T.m[P][X] *= ScaleX;
  T.m[P][Y] *= ScaleY;
  T.m[P][Z] *= ScaleZ;
  Joint.Scale (ScaleX, ScaleY, ScaleZ);

  Initialize ();

  if ((Mask & CHAIN_MASK) && Chain != NULL)
    Chain->Scale (ScaleX, ScaleY, ScaleZ, Mask);

  if ((Mask & ATTACH_MASK) && Attach != NULL)
    Attach->Scale (ScaleX, ScaleY, ScaleZ, Mask);

  if ((Mask & NEXT_MASK) && Next != NULL)
    Next->Scale (ScaleX, ScaleY, ScaleZ, Mask);

  UpdatePos ();          // Update positions of all following parts

  PRO_END
}


char *PART3D::GetName ()
{
  if (Part != NULL)
    return Part->PartName;
  else
    return NULL;
}


int PART3D::GetPartNumber (int Mask /* = 0 */) const
{
  int Result = 1;

  if ((Mask & CHAIN_MASK) && Chain != NULL)
    Result += Chain->GetPartNumber (Mask);

  if ((Mask & ATTACH_MASK) && Attach != NULL)
    Result += Attach->GetPartNumber (Mask);

  if ((Mask & NEXT_MASK) && Next != NULL)
    Result += Next->GetPartNumber (Mask);

  return Result;
}


int PART3D::GetDevNumber (int Mask /* = 0 */)
{
  int Result = 0;

  if ((Mask & CHAIN_MASK)  && Chain  != NULL)
    Result += Chain->GetDevNumber (Mask);

  if ((Mask & ATTACH_MASK) && Attach != NULL)
    Result += Attach->GetDevNumber (Mask);

  if ((Mask & NEXT_MASK)   && Next   != NULL)
    Result += Next->GetDevNumber (Mask);

  return Result;
}



MATERIAL *PART3D::GetMaterial ()
{
  return &Material;
}


BOX *PART3D::GetBox ()
{
  return &Box;
}


int PART3D::GetState () const
{
  return State;
}


int PART3D::GetMode () const
{
  return Mode;
}


int PART3D::GetRender () const
{ 
  return Render;
}


int PART3D::GetType () const
{ 
  return Type;
}


double PART3D::GetVal (int No) const
{
  return Joint.GetValue (No);
}


JOINT &PART3D::GetJoint ()
{
  return Joint;
}


double PART3D::GetSpeed () const
{ 
  return Joint.GetSpeed ();
}


double PART3D::GetAccel () const
{ 
  return Joint.GetAccel ();
}


double PART3D::GetTime () const
{ 
  return 0.0;
}


H_MATRIX PART3D::GetMatrix () const
{ 
  return T;
}


H_MATRIX PART3D::GetRel () const
{ 
  return Rel;
}


H_MATRIX PART3D::GetWTP () const
{
  return WTP;
}


PART *PART3D::GetPart ()
{ 
  return Part;
}


PART3D *PART3D::GetPart (const char *PartName)
{
  PART3D *Ptr, *Result;

  if (Part != NULL && strcmp(Part->PartName, PartName) == 0)
    return this;

  Ptr = Chain;
  while (Ptr != NULL)
  {
    if ((Result = Ptr->GetPart(PartName)) != NULL)
      return Result;
    Ptr = Ptr->Next;
  }

  // This method searches only in the chain-list; not the next parts !

  return NULL;
}    


PART3D *PART3D::GetPrev ()
{ 
  return Prev;
}


PART3D *PART3D::GetNext ()
{ 
  return Next;
}


PART3D *PART3D::GetChain ()
{ 
  return Chain;
}


DEVICE3D *PART3D::GetParent ()
{ 
  return Parent;
}


DEVICE3D *PART3D::GetAttach ()
{ 
  return Attach;
}


double PART3D::GetNormalAngle () const
{
  if (Part == NULL)
    return 0.0;

  return Part->NormalAngle;
}


void PART3D::GetSize (double *MinX, double *MaxX, 
		      double *MinY, double *MaxY,
		      double *MinZ, double *MaxZ) const
{
  if (Part != NULL)
  {
    if (MinX != NULL)
      *MinX = Part->Info.Min.x;
    if (MaxX != NULL)
      *MaxX = Part->Info.Max.x;
    if (MinY != NULL)
      *MinY = Part->Info.Min.y;
    if (MaxY != NULL)
      *MaxY = Part->Info.Max.y;
    if (MinZ != NULL)
      *MinZ = Part->Info.Min.z;
    if (MaxZ != NULL)
      *MaxZ = Part->Info.Max.z;
  }
}


static void PART3D_GetRange (PART3D *Part,
			     double *MinX, double *MaxX, 
			     double *MinY, double *MaxY,
			     double *MinZ, double *MaxZ, int Mask /* = 0 */,
			     bool first /* = TRUE */)
{
  POINT3D Point;

  if (Part->GetPart() != NULL)
  {
    Point = GEOMatrixMult (Part->GetPart()->WTP, Part->GetPart()->Info.Min);
    if (first)
    {
      if (MinX != NULL)
	*MinX = Point.x;
      if (MinY != NULL)
	*MinY = Point.y;
      if (MinZ != NULL)
	*MinZ = Point.z;
    }
    else
    {
      if (MinX != NULL)
	*MinX = MIN (*MinX, Point.x);
      if (MinY != NULL)
	*MinY = MIN (*MinY, Point.y);
      if (MinZ != NULL)
	*MinZ = MIN (*MinZ, Point.z);
    }
    Point = GEOMatrixMult (Part->GetPart()->WTP, Part->GetPart()->Info.Min);
    if (first)
    {
      if (MaxX != NULL)
	*MaxX = Point.x;
      if (MaxY != NULL)
	*MaxY = Point.y;
      if (MaxZ != NULL)
	*MaxZ = Point.z;
    }
    else
    {
      if (MaxX != NULL)
	*MaxX = MAX (*MaxX, Point.x);
      if (MaxY != NULL)
	*MaxY = MAX (*MaxY, Point.y);
      if (MaxZ != NULL)
	*MaxZ = MAX (*MaxZ, Point.z);
    }
    first = FALSE;
  }

  if ((Mask & CHAIN_MASK) && Part->GetChain() != NULL)
    Part->GetChain()->GetRange (MinX, MaxX, MinY, MaxY, MinZ, MaxZ, Mask);

  if ((Mask & ATTACH_MASK) && Part->GetAttach() != NULL)
    Part->GetAttach()->GetRange (MinX, MaxX, MinY, MaxY, MinZ, MaxZ);

  if ((Mask & NEXT_MASK) && Part->GetNext() != NULL)
    Part->GetNext()->GetRange (MinX, MaxX, MinY, MaxY, MinZ, MaxZ, Mask);
}


void PART3D::GetRange (double *MinX, double *MaxX, 
		       double *MinY, double *MaxY,
		       double *MinZ, double *MaxZ, int Mask /* = 0 */) 
{
  PART3D_GetRange (this, MinX, MaxX, MinY, MaxY, MinZ, MaxZ, Mask, TRUE);
}


void PART3D::SetName (const char *Name)
{
  if (Part != NULL)
    PARTSetName (Part, Name);
}


void PART3D::SetMaterial (const MATERIAL &Material)
{
  this->Material = Material;
  Create ();
}


void PART3D::SetColorID (COLOR::ID ColorID /* = COLOR::OBJECT */)
{
  Material.SetColorID (ColorID);
  Create ();
}


void PART3D::SetColor (COLOR Color)
{
  Material.SetColor (Color);
  Create ();
}


void PART3D::SetBox (const BOX &Box)
{
  this->Box = Box;
  this->Box.Create ();
}


void PART3D::SetState (int NewState)
{
  if (State & NewState)
    return;

  switch (NewState)
  {
    case STATE_NORMAL    : State &= (STATE_COLLISION | STATE_OOW);
                           break;
    case STATE_SELECTED  : State &= (STATE_COLLISION | STATE_OOW);
                           State |= STATE_SELECTED;
			   break;
    case STATE_COLLISION : 
    case STATE_OOW       :
    default              : if (State & NewState)
                             State &= ~(NewState);
                           else
			     State |= NewState;
  }

  Create ();
}


void PART3D::SetMode (int Mode /* = DRAW_PART */)
{
  this->Mode = Mode;
}


void PART3D::SetRender (int Mode)
{
  Render = Mode;

  Create ();
}


void PART3D::SetTextureID (int TextureID /* = NO_TEXTURE */)
{
  Material.SetTextureID (TextureID);
  Create ();
}


void PART3D::SetAlpha (double Alpha /* = 1.0 */)
{ 
  Material.SetAlpha(Alpha); 
}


void PART3D::SetType (int Type)
{ 
  this->Type = Type; 
}


void PART3D::SetVal (double *Values)
{
  int count;

  PRO_METHOD ("PARt3D", "")

  for (count = POS_X; count <= ROT_Z; ++count)
    Joint.SetValue (count, Values[count]);

  UpdatePos ();

  PRO_END
}


void PART3D::SetVal (int No, double Value)
{
  PRO_METHOD ("PART3D", "")

  assert (No >= PX && No <= RZ);

  Joint.SetValue (No, Value);

  // Set or clear OOW-flag:
  if (Joint.IsValid(No, Value))
    State |= STATE_OOW;
  elsif (State & STATE_OOW)
    State &= (~ STATE_OOW);

  UpdatePos ();

  PRO_END
}


void PART3D::SetSpeed (double Speed)
{ 
  Joint.SetSpeed (Speed);
}


void PART3D::SetAccel (double Accel)
{ 
  Joint.SetAccel (Accel);
}


void PART3D::SetTime (double Time)
{
  Joint.SetTime (Time);
}


void PART3D::SetPos (double px, double py, double pz)
{
  Joint.SetValue (POS_X, px);
  Joint.SetValue (POS_Y, py);
  Joint.SetValue (POS_Z, pz);  

  if (! Joint.IsValid(POS_X, px) ||
      ! Joint.IsValid(POS_Y, py) ||
      ! Joint.IsValid(POS_Z, pz))
    State |= STATE_OOW;
  else
    State &= (~ STATE_OOW);

  UpdatePos ();
}


void PART3D::SetRot (double rx, double ry, double rz)
{
  Joint.SetValue (ROT_X, rx);
  Joint.SetValue (ROT_Y, ry);
  Joint.SetValue (ROT_Z, rz);  

  if (! Joint.IsValid(ROT_X, rx) ||
      ! Joint.IsValid(ROT_Y, ry) ||
      ! Joint.IsValid(ROT_Z, rz))
    State |= STATE_OOW;
  else
    State &= (~ STATE_OOW);  

  UpdatePos ();
}


void PART3D::SetMatrix (H_MATRIX T)
{
  this->T = T;
  UpdatePos ();
}


void PART3D::SetRel (H_MATRIX Rel)
{
  this->Rel = Rel;
  UpdatePos ();
}


void PART3D::Place (H_MATRIX WTP)
{
  /*
    The parts position is given by the T matrix multiplied with the Rel 
    matrix, wich is given by the current joint values. This method 
    does not reset these joint-values. So, the part might not have the 
    specified WTP-frame.  To place the part at a desired
    place in the world, we need to subtract the inverse of the current
    relative-matrix:
      SetMatrix (GEOMatrixProd(WTP, GEOMatrixInv(Rel)));
  */
  SetMatrix (GEOMatrixProd(WTP, GEOMatrixInv(Rel)));
}


int PART3D::SetJoint (double *Value)
{
  int    count, No = 0;
  double RPY[6] = {0.0, 0.0, 0.0, 0.0, 0.0, 0.0};

  for (count = 0; count < 3; ++count)    // translatoric
    if (Joint.IsFree(count))
      RPY[count] = Value[No++];
	    
  for (count = 0; count < 3; ++count)    // rotatoric
    if (Joint.IsFree(count+3))
      RPY[count+3] = Value[No++];

  for (count = POS_X; count <= ROT_Z; ++count)
    Joint.SetValue (count, RPY[count]);

  UpdatePos ();

  return No;
}


void PART3D::SetPart (PART *Part)
{ 
  this->Part = Part;
  Initialize ();
}


void PART3D::SetPrev (PART3D *PrevPart)
{
  Prev = PrevPart;

#if 0
  // TODO: Prfen, ob so Endlosschleife mglich ist
  if (PrevPart->Chain == NULL)
    PrevPart->AddChain (this);
  else
  {
    // Prfen ,ob dieses Teil bereits in der Chain ist.
    // PrevPart->Chain->AddPart (this); 
  }
#endif

  UpdatePos ();
}


void PART3D::SetNext (PART3D *NextPart)
{ 
  Next = NextPart; 

  if (NextPart != NULL)
  {
    // Update reference-pointers and positions
    NextPart->Parent = Parent;
    NextPart->Prev   = Prev;
    NextPart->UpdatePos ();

    // Update counters
    if (Parent != NULL)
    {
      // Parent->IncPartNo (NextPart->GetPartNo());
      // TODO: Parts in attached-ten Gerten bercksichtigen
    }
  }
}


void PART3D::AddNext (PART3D *NewPart)
{
  PART3D *Ptr;

  PRO_METHOD ("PART3D", "")

  if ((Ptr = Next) == NULL)
    SetNext (NewPart);
  else
  {
    while (Ptr->Next != NULL)
      Ptr = Ptr->Next;
    Ptr->SetNext (NewPart);
  }

  PRO_END
}


void PART3D::AddChain (PART3D *ChainPart)
{ 
  PART3D *Ptr;

  PRO_METHOD ("PART3D", "")

  if (ChainPart == NULL)
    PRO_ABORT ("No part")

  if ((Ptr = Chain) == NULL)            // No next part in chain
    Chain = ChainPart;
  else                                  // Add to next chain-list
  {
    while (Ptr->Next != NULL)
      Ptr = Ptr->Next;
    Ptr->Next = ChainPart;
  }

  ChainPart->Prev   = this;
  ChainPart->Parent = Parent;
  ChainPart->UpdatePos ();

  PRO_END
}


void PART3D::SetParent (DEVICE3D *Dev)
{
  Parent = Dev;

  UpdatePos ();
}


void PART3D::SetAttach (DEVICE3D *Dev)
{
  Attach = Dev;

  if (Dev != NULL)
  {
    Dev->SetFix (this);
    Dev->UpdatePos ();
  }
}


bool PART3D::AttachDev (DEVICE3D *Dev)
{
  PRO_METHOD ("PART3D", "")

  if (Dev == NULL)
    PRO_RETURN ("Empty device specified", TRUE)

  if (Attach == NULL)                         // No attached devices up to now
    SetAttach (Dev);
  else                                       // Add to attach-list
    Attach->AddDevice (Dev);

  if (Parent != NULL)
    Parent->IncPartNo (Dev->GetPartNo());    // Update part-counter

  PRO_END
  return TRUE;
}


// Updates the absolute world postion of the part and all attached devices 
// The parameter WTP specifies the WorldToPart frame of the previous part
// in the chain. WTP of this part considers also it's relative placement.
void PART3D::UpdatePos ()
{
  PART3D *Fix;

  if (Prev != NULL)                         // Not first part in chain
    UpdatePos (Prev->GetWTP());
  elsif (Parent != NULL)                    // First part of device
    if  ((Fix = Parent->GetFix()) != NULL)  // dev is attached to another part
      UpdatePos (Fix->GetWTP());
    else                                    // device is base-device
      UpdatePos (Parent->GetWTD());
  else
    UpdatePos (Matrix_1);                   // unknown parent device
}


void PART3D::UpdatePos (H_MATRIX Frame)
{
  H_MATRIX  M;
  PART3D   *Ptr;
  DEVICE3D *Dev;

  // Update relative transformation matrix
  M = Joint.GetFrame ();
  Rel = GEOMatrixProd (T,     M);
  WTP = GEOMatrixProd (Frame, Rel);

  // Set the part's wtp frame
  if (Part != NULL)
    Part->WTP = WTP;

  // Recursively upate all chain parts
  Ptr = Chain;
  while (Ptr != NULL)
  {
    Ptr->UpdatePos (WTP);
    Ptr = Ptr->GetNext ();
  }
    
  Dev = Attach;
  while (Dev != NULL)
  {
    Dev->UpdatePos (WTP);
    Dev = Dev->GetNext ();
  }

  // Do not update next part. Create loop if following parts need update
}


void PART3D::SetNormalAngle (double Angle /* = 30.0 */)
{
  if (Part != NULL)
  {
    PARTCalcNormals (Part, Angle);
    Create ();
  }
}


void PART3D::SetWorld (WORLD *World)
{
  this->World = World;

  // TODO: 
  // Material.SetWorld, Box.SetWorld, CS.SetWorld, chain/attached ?
}


void PART3D::DrawWireFrame (OBJECT *Obj, POLYGON *Polygon, 
			    const MATERIAL *Material)
{
  int     count;
  POINT3D Point;

  if (Material == NULL)
    return;

  // Draw Outline:
  glLineWidth (Material->GetLine());
  Material->Set ();
  glBegin (GL_LINE_LOOP);
    for (count = 0; count < Polygon->Number; ++count)
    {
      Point = Obj->VertexList[Polygon->Vertex[count]];
      glVertex3d (Point.x, Point.y, Point.z);
    }
  glEnd ();
}


double PART3D::GetTextureFactor ()
{
  // Compute the factor used to stretch the texture-coordinate approximily 
  // to the range 0..1. We use a global stretching factor in order to
  // a) reduce computation time
  // b) get a uniform magnification of the texture
  
  return 1.0 / sqrt(GetAveragePolygonSize(Part));
}


// Remove part from chain or attached devices. Does not process part itself,
// because this must be done from calling routine in order to update next-
// pointer. This Method only unregisters a part - it does not delete it !
bool PART3D::Remove (PART3D *RemovePart)
{
  bool      Result = FALSE;
  PART3D   *Ptr;
  DEVICE3D *Ptr2;

  if ((Ptr = Chain) != NULL)
  {
    if ((Result = (Chain == RemovePart)))
      Chain = RemovePart->Next;

    while (! Result && Ptr->Next != NULL)
    {
      if ((Result = (Ptr->Next == RemovePart)))
	Ptr->SetNext (RemovePart->Next);
      else
	Result = Ptr->Remove (RemovePart);

      Ptr = Ptr->Next;
    }
  }

  Ptr2 = Attach;
  while (! Result && Ptr2 != NULL)
  {
    Result = Ptr2->Remove (RemovePart);
    Ptr2 = Ptr2->GetNext ();
  }

  return Result;
}


bool PART3D::Remove (DEVICE3D *Dev)
{
  bool      Result = FALSE;
  PART3D   *Ptr = Chain;
  DEVICE3D *Ptr2 = Attach;

  if (Ptr2 != NULL)
  {
    if ((Result = (Attach == Dev)))
      Attach = Dev->GetNext ();
    else
      Result = Ptr2->Remove (Dev);

    while (! Result && Ptr2->GetNext() != NULL)
    {
      if ((Result = (Ptr2->GetNext() == Dev)))
	Ptr2->SetNext (Dev->GetNext());
      else
	Result = Ptr2->Remove (Dev);

      Ptr2 = Ptr2->GetNext ();
    }

    if (Result)
    {
      Dev->SetFix (NULL);
      World->IncPartNum (- Dev->GetPartNumber());
    }
  }

  while (! Result && Ptr != NULL)
  {
    Result = Ptr->Remove (Dev);
    Ptr = Ptr->Next;
  }

  return Result;
}  

    
/* ********************************************************************** */
/* ** Compute the average area of all part's polygons. The smaller the ** */
/* ** the value, the oftener we have to repeat the texture for larger  ** */
/* ** polygons.                                                        ** */
/* ********************************************************************** */
double GetAveragePolygonSize (PART *Part)
{
  int     count = 0;
  double  Sum = 0.0;
  OBJECT *Obj;

  // TODO:
  // find out a formula which respects parts having a large amount of small
  // polygons: 
  // Ideas of computing a part's compactness, using
  //  o Volume(Circle)/Area(Circle) == 1
  //  o Volume(Part)/Area(Part) < 1
  //  o Min,Max,Avg(Area)
  //  o # objects
  //  o variance to optimum radius
  //      vol(Part) = pi * r^2   <=>  radius = sqrt(Volume(part) / pi)

  Obj = Part->ObjectList;
  while (Obj != NULL)
  {
    if (Obj->Type == PART_PLANE)
    {
      Sum += Obj->Info.Area;
      count += Obj->Data.Polyeder->Number;
    }      
    Obj = Obj->Next;
  }

  return (Sum / (double)count);
}

