#include <assert.h>
#include <math.h>

#include "libcad3d.h"



PROJECTION::PROJECTION () :
  CAD_FILE ()
{
  PRO_METHOD ("PROJECTION", "")

  SetRange (NULL);

  PRO_END
}


PROJECTION::PROJECTION (PROJECTION *Projection) :
  CAD_FILE ()
{
  PRO_METHOD ("PROJECTION", "")

  if (Projection != this)
    memcpy (Range, Projection->Range, sizeof(double) * 6);

  PRO_END
}


void PROJECTION::Set (double PosX  /* = -1.0 */, double PosY  /* = -1.0 */,
		      double SizeX /* =  1.0 */, double SizeY /* =  1.0 */)
{
  GLint ViewPort[4];
  
  glMatrixMode (GL_PROJECTION);
  glLoadIdentity ();

  if (PosX >= 0 && PosY >= 0)
  {
    glGetIntegerv (GL_VIEWPORT, ViewPort);
    gluPickMatrix ((GLdouble)PosX,  (GLdouble)PosY, 
		   (GLdouble)SizeX, (GLdouble)SizeY, ViewPort);
  }
}


bool PROJECTION::Read (DATAFILE &File, double Version)
{
  PRO_METHOD ("PROJECTION", "\n")

  if (File.Parse("<\\[PROJECTION\\]>\n") < 0)
    return FALSE;

  if (! CAD_FILE::Read(File, Version))
    return FALSE;

  if (File.Parse("  <Range:>  (%f) ... (%f)\n"
		 "            (%f) ... (%f)\n"
		 "            (%f) ... (%f)\n",
		 &Range[MIN_X], &Range[MAX_X],
		 &Range[MIN_Y], &Range[MAX_Y],
		 &Range[MIN_Z], &Range[MAX_Z]) < 6)
    return FALSE;

  PRO_END
  return TRUE;
}


bool PROJECTION::Write (FILE *fp /* = NULL */, int Indent /* = 0 */) const
{
  PRO_METHOD ("PROJECTION", "\n")

  if (fp == NULL)   
    fp = stdout;

  Space   (fp, Indent);
  fprintf (fp, "[PROJECTION]\n");
  Indent += INDENT;

  if (! CAD_FILE::Write(fp, Indent))
    PRO_RETURN ("Writing failed", FALSE)

  Space   (fp, Indent);
  fprintf (fp, "Range:   %.2f ... %.2f\n", Range[MIN_X], Range[MAX_X]);
  Space   (fp, Indent);
  fprintf (fp, "         %.2f ... %.2f\n", Range[MIN_Y], Range[MAX_Y]);
  Space   (fp, Indent);
  fprintf (fp, "         %.2f ... %.2f\n", Range[MIN_Z], Range[MAX_Z]);

  PRO_END
  return TRUE;
}


void PROJECTION::Pan (double DirX, double DirY)
{
  DirX = DirX * (Range[MAX_X] - Range[MIN_X]);
  DirY = DirY * (Range[MAX_Y] - Range[MIN_Y]);

  Range[MIN_X] += DirX;
  Range[MAX_X] += DirX;
  Range[MIN_Y] += DirY;
  Range[MAX_Y] += DirY;
}


double *PROJECTION::GetRange ()
{
  return Range;
}


// Set absolute range
void PROJECTION::SetRange (double *NewRange)
{
  PRO_METHOD ("PROJECTION", "\n")

  if (NewRange != NULL)          // Set to default-values, if not defined
    memcpy (Range, NewRange, sizeof(double) * 6);

  PRO6 ("=>            %.1f %.1f %.1f - %.1f %.1f %.1f\n", 
	Range[MIN_X], Range[MIN_Y], Range[MIN_Z],
	Range[MAX_X], Range[MAX_Y], Range[MAX_Z])

  PRO_END
}


// Set relative range
void PROJECTION::SetRange (double RelX1, double RelX2, 
			   double RelY1 /* = 0.0 */, double RelY2 /* = 1.0 */,
			   double RelZ1 /* = 0.0 */, double RelZ2 /* = 1.0 */)

{
  double SizeX, SizeY, SizeZ;

  SizeX = Range[MAX_X] - Range[MIN_X];
  SizeY = Range[MAX_Y] - Range[MIN_Y];
  SizeZ = Range[MAX_Z] - Range[MIN_Z];

  Range[MAX_X] = Range[MIN_X] + (SizeX * RelX2);
  Range[MIN_X] = Range[MIN_X] + (SizeX * RelX1);
  Range[MAX_Y] = Range[MIN_Y] + (SizeY * RelY2);
  Range[MIN_Y] = Range[MIN_Y] + (SizeY * RelY1);
  Range[MAX_Z] = Range[MIN_Z] + (SizeZ * RelZ2);
  Range[MIN_Z] = Range[MIN_Z] + (SizeZ * RelZ1);
}


void PROJECTION::GetCOI (double *x, double *y, double *z)
{
  PRO_METHOD ("PROJECTION", "")

  if (x != NULL)
    *x = 0.5 * (Range[MIN_X] + Range[MAX_X]);
  if (y != NULL)
    *y = 0.5 * (Range[MIN_Y] + Range[MAX_Y]);
  if (z != NULL)
    *z = 0.5 * (Range[MIN_Z] + Range[MAX_Z]);

  PRO_END
}



ORTHOGRAPHIC::ORTHOGRAPHIC () :
  PROJECTION ()
{
};


ORTHOGRAPHIC::ORTHOGRAPHIC (ORTHOGRAPHIC *Projection) :
  PROJECTION (Projection)
{
};


void ORTHOGRAPHIC::Set (double PosX  /* = -1.0 */, double PosY  /* = -1.0 */,
                        double SizeX /* =  1.0 */, double SizeY /* =  1.0 */)
{
  PROJECTION::Set (PosX, PosY, SizeX, SizeY);    // Call inherited function

  glOrtho (Range[MIN_X], Range[MAX_X], 
	   Range[MIN_Y], Range[MAX_Y],
	   Range[MIN_Z], Range[MAX_Z]);

  glMatrixMode (GL_MODELVIEW);  // TODO: in VIEW verschieben 
}


void ORTHOGRAPHIC::SetRange (double *NewRange /* = NULL */)
{
  if (NewRange != NULL)
    PROJECTION::SetRange (NewRange);
  else
  {
    Range[MIN_X] = -1.0;  Range[MAX_X] = 1.0; 
    Range[MIN_Y] = -1.0;  Range[MAX_Y] = 1.0; 
    Range[MIN_Z] = -1.0;  Range[MAX_Z] = 1.0; 
  }

  Set ();
}


void ORTHOGRAPHIC::GetCOI (double *x, double *y, double *z)
{
  PROJECTION::GetCOI (x, y, z);
};



PERSPECTIVE::PERSPECTIVE () :
  PROJECTION ()
{
}


PERSPECTIVE::PERSPECTIVE (PERSPECTIVE *Projection) :
  PROJECTION (Projection)
{
}


void PERSPECTIVE::Set (double PosX  /* = -1.0 */, double PosY  /* = -1.0 */,
		       double SizeX /* =  1.0 */, double SizeY /* =  1.0 */)
{
  PROJECTION::Set (PosX, PosY, SizeX, SizeY);     // Call inherited function

    // Size = MAX (Ortho[MAX_X]-Ortho[MIN_X], Ortho[MAX_Y]-Ortho[MIN_Y]);
    // TODO: Entferung von Kameraposition fr near ermittel/uebergeben 
    /*
    glFrustum (Ortho[MIN_X], Ortho[MAX_X],      // left, right
               Ortho[MIN_Y], Ortho[MAX_Y],      // bottom, top
	       fabs(Ortho[MIN_Z]),                                 // near
	       fabs(Ortho[MIN_Z]) + (Ortho[MAX_Z]-Ortho[MIN_Z]));  // far
    */
  glFrustum (     Range[MIN_X],       Range[MAX_X],     // left, right
		  Range[MIN_Y],       Range[MAX_Y],     // bottom, top
	     fabs(Range[MIN_Z]), fabs(Range[MAX_Z]));   // near, far

  glMatrixMode (GL_MODELVIEW);   // TODO: raus in VIEW
}


void PERSPECTIVE::SetRange (double *NewRange /* = NULL */)
{
  if (NewRange != NULL)
    PROJECTION::SetRange (NewRange);
  else
  {
    Range[MIN_X] = -1.0;  Range[MAX_X] = 1.0; 
    Range[MIN_Y] = -1.0;  Range[MAX_Y] = 1.0; 
    Range[MIN_Z] =  1.0;  Range[MAX_Z] = 1000.0; 
  }
  Set ();
}


void PERSPECTIVE::GetCOI (double *x, double *y, double *z)
{
  PROJECTION::GetCOI (x, y, z);
};



