#include <assert.h>

#include "libcad3d.h"


SHADOW::SHADOW (LIGHT *Light, FLOOR *Floor /* = NULL */,
		MODE  Mode  /* = OFF */, 
		COLOR Color /* = COLOR(0.0, 0.0, 0.0, 0.5) */) :
  CAD_FILE (),
  UNIT     (),
  Mode     (Mode),
  Offset   (0.1),
  Color    (Color),
  Floor    (Floor),
  Light    (Light)
{
  PRO_METHOD ("SHADOW", "\n")

  Create ();

  PRO_END
}


bool SHADOW::Read (DATAFILE &File, double Version)
{
  bool HasFloor, HasLight;

  PRO_METHOD ("SHADOW", "\n")

  if (File.Parse("<\\[SHADOW\\]>\n") < 0)
    return FALSE;

  if (! CAD_FILE::Read(File, Version))
    return FALSE;

  if (File.Parse("  <Mode:>    (OFF|ON)\n"
		 "  <Offset:>  (%f)\n"
		 "  <Floor:>   (NO|YES)\n"
		 "  <Light:>   (NO|YES)\n", 
		 &Mode, &Offset, 
		 &HasFloor, &HasLight) < 4)
    return FALSE;

  if (! Color.Read(File, Version))
    return FALSE;

  if (HasFloor)
  {
    if (Floor == NULL)
      Floor = new FLOOR ();
    if (! Floor->Read(File, Version))
      return FALSE;
  }

  if (HasLight)
  {
    if (Light == NULL)
      Light = new LIGHT (LIGHT::LIGHT1);
    if (! Light->Read(File, Version))
      return FALSE;
  }

  Create ();

  PRO_END
  return TRUE;
}


bool SHADOW::Write (FILE *fp /* = NULL */, int Indent /* = 0 */) const
{
  PRO_METHOD ("SHADOW", "\n")

  if (fp == NULL)
    fp = stdout;

  Space   (fp, Indent);
  fprintf (fp, "[SHADOW]\n");
  Indent += INDENT;

  if (! CAD_FILE::Write(fp, Indent))
    PRO_RETURN ("Writing failed", FALSE)

  Space   (fp, Indent);
  fprintf (fp, "Mode:    %s\n", (Mode==ON)? "ON":"OFF");

  Space   (fp, Indent);
  fprintf (fp, "Offset:  %.3g\n", Offset);

  Space   (fp, Indent);
  fprintf (fp, "Floor:   %s\n", (Floor!=NULL)?"YES":"NO");

  Space   (fp, Indent);
  fprintf (fp, "Light:   %s\n", (Light!=NULL)?"YES":"NO");

  Color.Write (fp, Indent);

  if (Floor != NULL)
    Floor->Write (fp, Indent);

  if (Light != NULL)
    Light->Write (fp, Indent);

  PRO_END
  return TRUE;
}


bool SHADOW::Create ()
{
  double Product;
  double Plane[4], LightPos[4];

  PRO_METHOD ("SHADOW", "\n")

  ShadowMatrix = Matrix_1;

  // No shadow if no light defined or light is turned off
  if (Light == NULL || Light->GetMode() == LIGHT::OFF)
    PRO_RETURN ("Mo light defined or light turned off", FALSE)

  if (Light->GetMode() == LIGHT::DIRECTIONAL)    // Light comes from infinity
  {
    LightPos[X] = 0.0;
    LightPos[Y] = 0.0;
    LightPos[Z] = 1.0;
    LightPos[W] = 0.0;
  }
  else                                           // Light has position
  {
    LightPos[X] = Light->GetPosition().x;
    LightPos[Y] = Light->GetPosition().y;
    LightPos[Z] = Light->GetPosition().z;
    LightPos[W] = 1.0;
  }

  CalcPlane (Plane);                    // Compute equation of projection plane

  /* Find dot product between light position vector and ground plane normal. */
  Product =   Plane[X] * LightPos[X] 
            + Plane[Y] * LightPos[Y] 
            + Plane[Z] * LightPos[Z] 
            + Plane[W] * LightPos[W];
 
  ShadowMatrix.m[N][X] = Product - LightPos[X] * Plane[X];
  ShadowMatrix.m[O][X] =         - LightPos[X] * Plane[Y];
  ShadowMatrix.m[A][X] =         - LightPos[X] * Plane[Z];
  ShadowMatrix.m[P][X] =         - LightPos[X] * Plane[W];
 
  ShadowMatrix.m[N][Y] =         - LightPos[Y] * Plane[X];
  ShadowMatrix.m[O][Y] = Product - LightPos[Y] * Plane[Y];
  ShadowMatrix.m[A][Y] =         - LightPos[Y] * Plane[Z];
  ShadowMatrix.m[P][Y] =         - LightPos[Y] * Plane[W];

  ShadowMatrix.m[N][Z] =         - LightPos[Z] * Plane[X];
  ShadowMatrix.m[O][Z] =         - LightPos[Z] * Plane[Y];
  ShadowMatrix.m[A][Z] = Product - LightPos[Z] * Plane[Z];
  ShadowMatrix.m[P][Z] =         - LightPos[Z] * Plane[W];
 
  ShadowMatrix.m[N][W] =         - LightPos[W] * Plane[X];
  ShadowMatrix.m[O][W] =         - LightPos[W] * Plane[Y];
  ShadowMatrix.m[A][W] =         - LightPos[W] * Plane[Z];
  ShadowMatrix.m[P][W] = Product - LightPos[W] * Plane[W];     

  PRO_END
  return TRUE;
}


void SHADOW::SetMode (MODE Mode)
{
  this->Mode = Mode;
}


void SHADOW::SetFloor (FLOOR *Floor)
{
  this->Floor = Floor;
  Create ();
}


void SHADOW::SetLight (LIGHT *Light)
{
  this->Light = Light;
  Create ();
}


void SHADOW::SetOffset (double Offset)
{
  this->Offset = Offset;
}


void SHADOW::SetColor (const COLOR &Color /* = COLOR(COLOR::BLACK */)
{
  this->Color = Color;
}


SHADOW::MODE SHADOW::GetMode () const
{
  return Mode;
}


H_MATRIX SHADOW::GetMatrix () const
{
  return ShadowMatrix;
}


double SHADOW::GetOffset () const
{
  return Offset;
}


COLOR &SHADOW::GetColor ()
{
  return Color;
}


void SHADOW::CalcPlane (double Plane[4])
{
  double   SizeX, SizeY;
  double   Height;
  double   V0[3];
  VECTOR3D Vec[2];

  PRO_METHOD ("SHADOW", "")

  if (Floor == NULL)
  {
    SizeX  = 1.0;
    SizeY  = 1.0;
    Height = 0.0;
  }
  else
  {
    SizeX  = Floor->GetSizeX() * Floor->GetSpaceX();
    SizeY  = Floor->GetSizeY() * Floor->GetSpaceY();
    Height = (*Floor)[PZ];
  }

  V0[X] = 0.5 * SizeX;
  V0[Y] = 0.5 * SizeY;
  V0[Z] = Height + Offset;

  Vec[0] = GEOMakeVector (   0.0, SizeY, 0.0);
  Vec[1] = GEOMakeVector (-SizeX, SizeY, 0.0);

  /* find cross product to get A, B, and C of plane equation */
  Plane[0] =   Vec[0].y * Vec[1].z - Vec[0].z * Vec[1].y;
  Plane[1] = -(Vec[0].x * Vec[1].z - Vec[0].z * Vec[1].x);
  Plane[2] =   Vec[0].x * Vec[1].y - Vec[0].y * Vec[1].x;

  Plane[3] = -(  Plane[0] * V0[X] 
               + Plane[1] * V0[Y]
               + Plane[2] * V0[Z]);

  PRO_END
}


