#include <assert.h>

#include "libcad3d.h"


// Mapping between class- and OpenGL-constants
static GLint WrapMode[]      = {GL_CLAMP,    GL_REPEAT                      };
static GLint TextureFilter[] = {GL_NEAREST,  GL_LINEAR                      };
static GLint TextureMode[]   = {GL_MODULATE, GL_REPLACE, GL_DECAL,  GL_BLEND};


TEXTURE::TEXTURE (const char *FileName) :
  CAD_FILE   (),
  UNIT       (),
  TextureID  (NO_TEXTURE),
  Mapping    (REPEAT),
  Filter     (NEAREST),
  Mode       (DECAL),
  StretchS   (1.0),
  StretchT   (1.0),
  Data       (NULL),
  TextureImg (NULL)
{
  PRO_METHOD ("TEXTURE", "\n")

  if (! SetTextureFile(FileName))
    throw ERROR (ERROR::FILE_ERROR);

  PRO_END  
}


TEXTURE::TEXTURE (int TextureID) :
  CAD_FILE   (),
  UNIT       (),
  Mapping    (REPEAT),
  Filter     (NEAREST),
  Mode       (DECAL),
  StretchS   (1.0),
  StretchT   (1.0),
  Data       (NULL),
  TextureImg (NULL)
{
  PRO_METHOD ("TEXTURE", "")
    
  SetTextureID (TextureID);

  PRO_END
}


TEXTURE::TEXTURE (XImage *Image) :
  CAD_FILE   (),
  UNIT       (),
  Mapping    (REPEAT),
  Filter     (NEAREST),
  Mode       (DECAL),
  StretchS   (1.0),
  StretchT   (1.0),
  Data       (NULL),
  TextureImg (NULL)  
{
  PRO_METHOD ("TEXTURE", "\n")
   
  if (Image == NULL)
    throw ERROR (ERROR::NO_DATA);

  Initialize ();

  TextureImg = TOOLCopyImage (Image);
  Create ();

  PRO_END
}


TEXTURE::TEXTURE (TEXTURE *Texture) :
  CAD_FILE   (),
  UNIT       ()
{
  PRO_METHOD ("TEXTURE", "\n")

  if (Texture == NULL)
    throw ERROR (ERROR::NO_DATA);

  Initialize ();

  Mapping  = Texture->Mapping;
  Filter   = Texture->Filter;
  Mode     = Texture->Mode;
  StretchS = Texture->StretchS;
  StretchT = Texture->StretchT;
  strcpy (TextureFile, Texture->TextureFile);

  // Use TOOLCopyImage because XSubImage seems to be buggy when copying
  // the whole image !
  TextureImg = TOOLCopyImage (Texture->TextureImg);
  Create ();

  PRO_END
}


TEXTURE::~TEXTURE ()
{
  GLuint Texture = TextureID;

  PRO_METHOD ("TEXTURE~", "")

  if (Data != NULL)
    free (Data);

  if (TextureImg != NULL)
    TOOLDestroyImage (TextureImg);

  if (TextureID >= 0)
    glDeleteTextures (1, &Texture);

  PRO_END
}


void TEXTURE::Initialize ()
{
  TextureID      = NO_TEXTURE;
  Mapping        = REPEAT;
  Filter         = NEAREST;
  Mode           = DECAL;
  StretchS       = 1.0;
  StretchT       = 1.0;
  Data           = NULL;
  TextureImg     = NULL;
  TextureFile[0] = '\0';
}


TEXTURE &TEXTURE::operator = (const TEXTURE &Source)
{
  GLuint Texture = TextureID;

  PRO_METHOD ("TEXTURE", "")

  if (this != &Source)                    // check for self-assignment
  {
    if (Data != NULL)
      free (Data);

    if (TextureImg != NULL)
      TOOLDestroyImage (TextureImg);

    if (TextureID >= 0)
      glDeleteTextures (1, &Texture);

    Initialize ();

    Mapping  = Source.Mapping;
    Filter   = Source.Filter;
    Mode     = Source.Mode;
    StretchS = Source.StretchS;
    StretchT = Source.StretchT;
    strcpy (TextureFile, Source.TextureFile);

    // Use TOOLCopyImage because XSubImage seems to be buggy or I haven't 
    // understand it :-(
    TextureImg = TOOLCopyImage (Source.TextureImg);
  }
  Create ();

  PRO_END
  return *this;
}


bool TEXTURE::Read (DATAFILE &File, double Version)
{
  PRO_METHOD ("TEXTURE", "\n")

  if (File.Parse("<\\[TEXTURE\\]>\n") < 0)
    return FALSE;

  if (! CAD_FILE::Read(File, Version))
    return FALSE;

  if (File.Parse("  <File:>       '(%4096s)'\n"
		 "  <Mapping:>    (CLAMP|REPEAT)\n"
		 "  <Filter:>     (NEAREST|LINEAR)\n"
		 "  <Mode:>       (MODULATE|REPLACE|DECAL|BLEND)\n"
		 "  <Stretch:>    (%f) x (%f)\n",
#if 0
		 "  <Data:>       (NO|YES)\n"
		 "  <Image:>      (NO|YES)\n"
		 "  <Size:>       (%d) x (%d)\n",
#endif
		 &TextureFile, 
		 &Mapping, &Filter, &Mode,
		 &StretchS, &StretchT) < 6)
    return FALSE;

  Create ();

  PRO_END
  return TRUE;
}


bool TEXTURE::Write (FILE *fp /* = NULL */, int Indent /* = 0 */) const
{
  PRO_METHOD ("TEXTURE", "\n")

  if (fp == NULL)
    fp = stdout;

  Space (fp, Indent);
  fprintf (fp, "[TEXTURE]\n");
  Indent += INDENT;

  if (! CAD_FILE::Write(fp, Indent))
    PRO_RETURN ("Writing failed", FALSE)

  Space   (fp, Indent);
  fprintf (fp, "  File:     '%s'\n", TextureFile);

  Space   (fp, Indent);
  fprintf (fp, "  Mapping:  %s\n",   (Mapping==REPEAT)? "REPEAT":"CLAMP");

  Space   (fp, Indent);
  fprintf (fp, "  Filter:   %s\n",   (Filter==LINEAR)? "LINEAR":"NEAREST");

  Space   (fp, Indent);
  fprintf (fp, "  Mode:     %s\n",
	       (Mode==DECAL)? "DECAL":((Mode==BLEND)? "BLEND":"MODULATE"));

  Space   (fp, Indent);
  fprintf (fp, "  Stretch:  %.3f x %.3f\n", StretchS, StretchT);

#if 0
  Space   (fp, Indent);
  fprintf (fp, "  Data:     %s\n",   (Data==NULL)? 'NO':'YES');

  Space   (fp, Indent);
  fprintf (fp, "  Image:    %s\n",   (TextureImg==NULL)? 'NO':'YES');

  Space   (fp, Indent);
  if (TextureImg != NULL)
    fprintf (fp, "  Size:   %d x %d\n",
                 TextureImg->width, TextureImg->height);
  else
    fprintf (fp, "  Size:   0 x 0\n");
#endif

  // ...

  PRO_END
  return TRUE;
}


void TEXTURE::Set (const COLOR &BlendColor /* = COLOR(COLOR::BLACK) */) const
{
  GLfloat Color[4];

  PRO_METHOD ("TEXTURE", "\n")

  if (TextureID >= 0)
  {
    glEnable (GL_TEXTURE_2D);
    glBindTexture (GL_TEXTURE_2D, (GLuint)TextureID);

    glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, WrapMode[Mapping]);
    glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, WrapMode[Mapping]);
    glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, 
		                      TextureFilter[Filter]);
    glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER,
		                      TextureFilter[Filter]);

    glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, TextureMode[Mode]);
    if (Mode == BLEND)                  // When in blend-mode set blend-color
    {
      Color[RED]   = BlendColor[RED];
      Color[GREEN] = BlendColor[GREEN];
      Color[BLUE]  = BlendColor[BLUE];
      Color[RED]   = BlendColor[ALPHA];
      //       glTexEnvi  (GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, GL_RGBA);
      glTexEnvfv (GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, Color);
    }

    // TODO: Border(-color) if filter is linear
#if 0
    // Test of generating automatic texture coordinates
    // see also CAD_PART::Create
    GLfloat XZero[] = {1.0, 0.0, .0, 0.0};
    glTexGeni  (GL_S, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
    glTexGenfv (GL_S, GL_OBJECT_PLANE,     XZero);
    glEnable (GL_TEXTURE_GEN_S);

    GLfloat YZero[] = {0.0, 1.0, .0, 0.0};
    glTexGeni  (GL_T, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
    glTexGenfv (GL_T, GL_OBJECT_PLANE,     YZero);
    glEnable (GL_TEXTURE_GEN_T);
#endif
  }

  PRO_END
}


int TEXTURE::CalcSize (int Size)
{
  int Value = 1;

  while (2*Value <= Size)
    Value *= 2;

  return Value;
} 


bool TEXTURE::Create ()
{
  int     SizeX, SizeY;
  int     RasterX, RasterY;
  double  Alpha = -1.0;
  GLuint  Texture = TextureID;
  GLint   Format;
  XImage *Image;

  PRO_METHOD ("TEXTURE", "\n")

  if (Data != NULL)                     // Free old texture data
  {
    free (Data);
    Data = NULL;
  }

  if (TextureID != NO_TEXTURE)
  {
    glDeleteTextures (1, &Texture);
    TextureID = NO_TEXTURE;
  }

  if (TextureImg == NULL)
    PRO_RETURN ("No Texture-Image", FALSE)

  // If the image size is not a power to two, scale size of image to size 2^n 
  RasterX = CalcSize (TextureImg->width);
  RasterY = CalcSize (TextureImg->height);
  if (RasterX < TextureImg->width || RasterY < TextureImg->height)
  {
    PRO4 ("Scale Texture-Image: %d x %d -> %d x %d\n", 
	  TextureImg->width, TextureImg->height, RasterX, RasterY)
    Image = TOOLScaleImage (TextureImg, RasterX, RasterY);
    TOOLDestroyImage (TextureImg);      // Destroy old image
    TextureImg = Image;
  }

  if ((Data = TOOLMakeRGBA(TextureImg, Alpha, &SizeX, &SizeY)) == NULL)
    PRO_RETURN ("Converting texture failed", FALSE)

  Format = (Alpha >= 0.0 && Alpha <= 1.0)? GL_RGBA : GL_RGB;

  glGenTextures (1,             &Texture);
  glBindTexture (GL_TEXTURE_2D,  Texture);
  TextureID = Texture;
  glTexImage2D  (GL_TEXTURE_2D, 
		 0,                         // LOD                
		 Format,                    // # internal Colors-components
		 SizeX, SizeY,              // Texture size
		 0,                         // Border
		 (GLenum)Format,            // Format
		 GL_UNSIGNED_BYTE,          // Type
		 Data);                     // Image data (pixels)

  PRO_END
  return (glGetError() == 0);
}


bool TEXTURE::SetTextureFile (const char *FileName)
{
  PRO_METHOD ("TEXTURE", "\n")

  if (TextureImg != NULL)
  {
    TOOLDestroyImage (TextureImg); 
    TextureImg = NULL;
  }

  TextureFile[0] = '\0';

  if (FileName == NULL)
    PRO_RETURN ("bad texture filename", FALSE)

  if ((TextureImg = TOOLLoadImage(FileName)) == NULL)
    PRO_RETURN1 ("reading texture-file failed: %s", FileName, FALSE)

  strncpy (TextureFile, FileName, PATH_MAX);
  TextureFile[PATH_MAX] = '\0';

  Create ();

  PRO_END
  return TRUE;
}


void TEXTURE::SetTextureID (int TextureID)
{
  // Set Texture to predefined id: No cleanup and creation needed.
  this->TextureID = TextureID;

  Create ();
}


void TEXTURE::SetMapping (MAPPING Mapping)
{
  this->Mapping = Mapping;
}


void TEXTURE::SetFilter (FILTER Filter)
{
  this->Filter = Filter;
}


void TEXTURE::SetMode (MODE Mode)
{
  this->Mode = Mode;
}


void TEXTURE::SetStretch (double s /* = 1.0 */, double t /* = 1.0 */)
{
  assert (s > 0.0 && t > 0.0);

  StretchS = s;
  StretchT = t;
}


char *TEXTURE::GetFileName ()
{
  return TextureFile;
}


int TEXTURE::GetTextureID () const
{
  return TextureID;
}


TEXTURE::MAPPING TEXTURE::GetMapping () const
{
  return Mapping;
}


TEXTURE::FILTER TEXTURE::GetFilter () const
{
  return Filter;
}


TEXTURE::MODE TEXTURE::GetMode () const
{
  return Mode;
}


double TEXTURE::GetStretchS () const
{
  return StretchS;
}


double TEXTURE::GetStretchT () const
{
  return StretchT;
}


XImage *TEXTURE::GetImage ()
{
  return TextureImg;
}


