#include <assert.h>

#include "libcad3d.h"


VIEW::VIEW (int SizeX, int SizeY,
	    RPY  Camera /* = RPY_0 */, double *Clipping /* = NULL  */, 
	    MODE Mode   /* = ORTHO */, double *Range    /* = NULL  */) :
  CAD_FILE (),
  Scale    (1.0),
  Border   (COLOR(COLOR::WHITE)),
  PrjNo    (Mode == PERSPECT),
  Camera   (Camera),
  WinX     (SizeX),
  WinY     (SizeY)
{
  PRO_METHOD ("VIEW", "\n")

  if (Clipping == NULL)
  {
    this->Clipping[0] = 0.0;     // x
    this->Clipping[1] = 0.0;     // y
    this->Clipping[2] = 1.0;     // width
    this->Clipping[3] = 1.0;     // heigth
  }
  else
    memcpy (this->Clipping, Clipping, sizeof(double) * 4);

  Projection[0] = new ORTHOGRAPHIC ();
  Projection[1] = new PERSPECTIVE  ();
  Projection[0]->SetRange (Range);
  Projection[1]->SetRange (Range);

  CalcViewPort ();

  PRO_END
}


VIEW::VIEW (VIEW *View) :
  CAD_FILE (),
  Scale    (View->Scale),
  Border   (View->Border),
  PrjNo    (View->PrjNo),
  Camera   (View->Camera.GetLocation()),
  WinX     (View->WinX),
  WinY     (View->WinY)
{
  PRO_METHOD ("VIEW", "")

  Projection[0] = new ORTHOGRAPHIC ((ORTHOGRAPHIC *)View->Projection[0]);
  Projection[1] = new PERSPECTIVE  ((PERSPECTIVE  *)View->Projection[1]);

  memcpy (Clipping, View->Clipping, sizeof(double) * 4);
  memcpy (ViewPort, View->ViewPort, sizeof(int)    * 4);

  PRO_END
}


VIEW::~VIEW ()
{
}


bool VIEW::Read (DATAFILE &File, double Version)
{
  PRO_METHOD ("VIEW", "\n")

  if (File.Parse("<\\[VIEW\\]>\n") < 0)
    return FALSE;

  if (! CAD_FILE::Read(File, Version))
    return FALSE;

  if (File.Parse("  <Clipping:>  (%f) / (%f) \\[(%f) x (%f)\\]\n"
		 "  <ViewPort:>  (%d) / (%d) ... (%d) / (%d)\n"
		 "  <Scale:>     (%f)\n"
		 "  <Size:>      (%d) x (%d)\n",
		 &Clipping[0], &Clipping[1], &Clipping[2], &Clipping[3], 
		 &ViewPort[0], &ViewPort[1], &ViewPort[2], &ViewPort[3], 
		 &Scale, &WinX, &WinY) < 11)
    return FALSE;

  if (! Camera.Read(File, Version))
    return FALSE;

  if (File.Parse("  <Type:>    (ORTHOGRAPHIC|PERSPECTIVE)\n", &PrjNo) < 1)
    return FALSE;

  if (Projection[0] == NULL)
    Projection[0] = new ORTHOGRAPHIC ();
  if (! Projection[0]->Read(File, Version))
    return FALSE;

  if (Projection[1] == NULL)
    Projection[1] = new PERSPECTIVE ();
  if (! Projection[1]->Read(File, Version))
    return FALSE;

  CalcViewPort ();

  PRO_END
  return TRUE;
}


bool VIEW::Write (FILE *fp /* = NULL */, int Indent /* = 0 */) const
{
  PRO_METHOD ("VIEW", "\n")

  if (fp == NULL)
    fp = stdout;

  Space   (fp, Indent);
  fprintf (fp, "[VIEW]\n");
  Indent += INDENT;

  if (! CAD_FILE::Write(fp, Indent))
    PRO_RETURN ("Writing failed", FALSE)

  Space   (fp, Indent);
  fprintf (fp, "Clipping: %4.2f / %4.2f  [%.2f x %.2f]\n", 
	       Clipping[0], Clipping[1], Clipping[2], Clipping[3]);

  Space   (fp, Indent);
  fprintf (fp, "ViewPort: %d / %d ... %d / %d\n",
	       ViewPort[0], ViewPort[1], ViewPort[2], ViewPort[3]);

  Space   (fp, Indent);
  fprintf (fp, "Scale:  %7.3f\n", Scale);

  Space   (fp, Indent);
  fprintf (fp, "Size:   %d x %d\n", WinX, WinY);

  Camera.Write (fp, Indent);

  Space   (fp, Indent);
  fprintf (fp, "Type:   %s\n", (PrjNo==0)? "ORTHOGRAPHIC":"PERSPECTIVE");

  Projection[0]->Write (fp, Indent);
  Projection[1]->Write (fp, Indent);

  PRO_END
  return TRUE;
}


void VIEW::Set ()
{
  PRO_METHOD ("VIEW", "\n")

  glMatrixMode (GL_MODELVIEW);

  glViewport (ViewPort[0], ViewPort[1], ViewPort[2], ViewPort[3]);

  if (! IsFullScreen())
  {
    Projection[PrjNo]->Set ();
    glScissor (ViewPort[0], ViewPort[1], ViewPort[2], ViewPort[3]);
    glEnable (GL_SCISSOR_TEST);
      glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glDisable (GL_SCISSOR_TEST);
  }
  else
    glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  glLoadIdentity ();

  PRO_END
}  


void VIEW::SetBorder (COLOR Color /* = COLOR(COLOR::WHITE) */)
{ 
  Border = Color;
  DrawBorder ();
}


// TODO: Funktion in VIEW-Class verschieben
void VIEW::Rotate (double rx, double ry, double rz)
{
#if 0
  VECTOR Center;
  MATRIX M;
#endif

  PRO_METHOD ("VIEW", "")
 
  // TODO um den COI drehen, so dass dieser konstant bleibt ! 
  // Problem: Bestimmung des COI von 2D-Screen auf 3D-Koordinaten
  // aktuelle Kamera-Pos bercksichtigen ?
  // Falls Obj in Fenster-Mitte: Koordinaten bestimmen

#if 0
  GetCOI (&Center.x, &Center.y, &Center.z);

  RPY Location = Camera.GetLocation ();
  // Translate camera-position about center->world
  MATRIX CTC = GEOMakeTrans (Location.px - Center.x, 
			     Location.py - Center.y, 
			     Location.pz - Center.z);
  // Rotate the camera around the origin
  MATRIX Rot = GEORPY2Frame (0.0, 0.0, 0.0, rx, ry, rz);
  M = GEOMatrixProd (Rot, CTC);
  // Move the camera about the difference
  Location.px += (M.m[P][X] - CTC.m[P][X]);
  Location.py += (M.m[P][Y] - CTC.m[P][Y]);
  Location.pz += (M.m[P][Z] - CTC.m[P][Z]);
  Location.y = MOD (Location.y+rx, -180.0, 180.0, 360.0);
  Location.p = MOD (Location.p+ry, -180.0, 180.0, 360.0);
  Location.r = MOD (Location.r+rz, -180.0, 180.0, 360.0);
  Camera.Place (Location);

Write ();
#endif  

#if 0
  /*
    // braucht World -> Parameter ?
  int No = SelNo (0.5*WinX, 0.5*WinY, 0.0, &Center.x, &Center.y, &Center.z);
  printf ("Rotate: %d\n", No);
  if (No < 0)
  */

  M =                   GEOMakeRot(Vector_X, Camera.y);
  M = GEOMatrixProd (M, GEOMakeRot(Vector_Y, Camera.p));
  M = GEOMatrixProd (M, GEOMakeRot(Vector_Z, Camera.r));
  M = GEOMatrixProd (M, GEOMakeTrans(Center.x, Center.y, Center.z));

  Camera.px = -M.m[P][X] + Center.x;
  Camera.py = -M.m[P][Y] + Center.z;
  Camera.pz = -M.m[P][Z] + Center.y;
#endif

#if 1
  Camera.Rotate (rx, ry, rz);
#endif

  PRO_END
}


#if 0
void VIEW::Fly (double px, double py, double pz, 
		double rx, double ry, double rz)
{
}
#endif


void VIEW::Pan (double DirX, double DirY)
{
  PRO_METHOD ("VIEW", "")

  Projection[PrjNo]->Pan (DirX / double(WinX), DirY / double(WinY));

  PRO_END
}


void VIEW::Zoom (int x1, int y1, int x2, int y2)
{
  PRO_METHOD ("VIEW", "")

  GetCoordinates (x1, y1, &x1, &y1);
  GetCoordinates (x2, y2, &x2, &y2);

  // Y-coords will be swapped because of inverse screen-coordinates (origin) ! 
  Projection[PrjNo]->SetRange (x1/(double)ViewPort[2], x2/(double)ViewPort[2], 
			       y2/(double)ViewPort[3], y1/(double)ViewPort[3]);

  PRO_END
}


bool VIEW::SelectPos (double px, double py, double pz, 
		      double *x, double *y, double *z)
{
  double   mx, my, mz;
  GLdouble Proj[16], Model[16];

  PRO_METHOD ("VIEW", "\n")

  PRO3 ("SelectPos: %f / %f / %f\n", px, py, pz)

  glGetDoublev  (GL_PROJECTION_MATRIX, Proj);
  glGetDoublev  (GL_MODELVIEW_MATRIX,  Model);

  if (gluUnProject((GLdouble)px, (GLdouble)py, (GLdouble)pz, 
		   Model, Proj, ViewPort, &mx, &my, &mz) != GL_TRUE)
    PRO_RETURN ("UnProject failed", FALSE)

  PRO3 ("\\%.3f %.3f %.3f\n", mx, my , mz)

  if (x != NULL)
    *x = mx;
  if (y != NULL)
    *y = my;
  if (z != NULL)
    *z = mz;

  PRO_END
  return TRUE;
}


int VIEW::SelNo (WORLD *World, int     px, int     py, double  pz,
		               double *x,  double *y,  double *z)
{
  bool    first = TRUE;
  int     count, Index;
  int     Size, No = -1;
  GLuint  ScrZ = 0;
  GLuint  NameNo;
  GLuint *Buffer, *Ptr;
  GLint   Hits;

  PRO_METHOD ("VIEW", "\n")

  if (World->GetPartNum() <= 0)
    PRO_RETURN ("No Objects", -1)

  Size = sizeof (GLuint) * 16 * World->GetPartNum();
  Buffer = (GLuint *)malloc (Size);

  Set ();                               // Activate this current view

  glSelectBuffer (Size, Buffer);
  glRenderMode   (GL_SELECT);

  glInitNames ();
  glPushName  ((GLuint)-1);

  py = WinY - py;          /* flip y-coordinate */

  Projection[PrjNo]->Set (px, py, 1, 1);

  // Draw workcell...
  glMatrixMode   (GL_MODELVIEW);
  glViewport     (ViewPort[0], ViewPort[1], ViewPort[2], ViewPort[3]);
  glLoadIdentity ();
  Camera.Set     ();
  World->Draw    (this, GL_SELECT);

  glMatrixMode (GL_PROJECTION);
  
  if ((Hits = glRenderMode(GL_RENDER)) <= 0)
    PRO0 ("No hits !\n")
  else
  {
    PRO1 ("Hits: %d\n", Hits)
    Ptr = (GLuint *)Buffer;

    for (count = 0; count < Hits; ++count)
    {
      NameNo = *Ptr;
      PRO2 ("# Names(%d): %d\n", count+1, NameNo)
      Ptr++;

      // Search for nearest(pz=0) or farest(pz=1) part-coordinates.
      // z-Values are multiplied by 2^32-1
      if (first || (pz == 0 && *Ptr < ScrZ) || (pz == 1 && *Ptr > ScrZ))
      {
	PRO2 ("MIN.-DEPTH: %u -> %g\n", *Ptr, (float)*Ptr/0xffffffff) 
	if (pz == 0.0)
	  ScrZ = *Ptr;                       // Min. depth of selection
	Ptr++;

	PRO2 ("MAX.-DEPTH: %u -> %g\n", *Ptr, (float)*Ptr/0xffffffff)
	if (pz == 1.0)
	  ScrZ = *Ptr;                       // Min. depth of selection
	Ptr++;                               // Max. depth of selection

	No = *Ptr;
	first = FALSE;
      }
      else
	Ptr += 2;

      for (Index = 0; Index < (int)NameNo; ++Index)
      {
	PRO1 ("NAME: %d\n", *Ptr)
	Ptr++;
      }
    }
    PRO1 (" Part-Number: %d\n", No)
    pz = (double)ScrZ / 0xffffffff;     // divide through 2^32-1
  }

  SelectPos ((double)px, (double)py, pz, x, y, z);
  free (Buffer);                        // free selection buffer

  PRO1 ("No: %d\n", No)

  PRO_END
  return No;
}


int VIEW::SelectRegion (WORLD *World, int px, int py, double pz, int **NoList,
			int SizeX /* = 1 */, int SizeY /* = 1 */)
{
  int     count, Size;
  GLuint  NameNo;
  GLuint *Buffer, *Ptr;
  GLint   Hits;

  PRO_METHOD ("VIEW", "\n")

  if (World->GetPartNum() <= 0)
    PRO_RETURN ("No Objects", -1)

  Size = sizeof (GLuint) * 16 * World->GetPartNum();
  Buffer = (GLuint *)malloc (Size);

  Set ();                               // Activate this current view

  glSelectBuffer (Size, Buffer);
  glRenderMode   (GL_SELECT);

  glInitNames ();
  glPushName  ((GLuint)-1);

  py = WinY - py;          /* flip y-coordinate */

  Projection[PrjNo]->Set (px, py, SizeX, SizeY);

  // Draw workcell...
  glMatrixMode   (GL_MODELVIEW);
  glViewport     (ViewPort[0], ViewPort[1], ViewPort[2], ViewPort[3]);
  glLoadIdentity ();
  Camera.Set     ();
  World->Draw    (this, GL_SELECT);

  glMatrixMode (GL_PROJECTION);
  
  if ((Hits = glRenderMode(GL_RENDER)) <= 0)
    PRO0 ("No hits !\n")
  else
  {
    PRO1 ("Hits: %d\n", Hits)
    *NoList = new int [Hits];

    Ptr = (GLuint *)Buffer;
    for (count = 0; count < Hits; ++count)
    {
      NameNo = *Ptr;
      Ptr += 3;                    // # names, min + max depth of selection
      (*NoList)[count] = *Ptr;
      Ptr += (int)NameNo;
    }
  }

  free (Buffer);                        // free selection buffer

  PRO_END
  return Hits;
}


void VIEW::DrawBorder ()
{
  PRO_METHOD ("VIEW", "")

  if (IsFullScreen())
    PRO_ABORT ("Full-Screen: No border")

  glMatrixMode (GL_PROJECTION);
  glPushMatrix   ();                            // save projection matrix
  glLoadIdentity ();

  glMatrixMode   (GL_MODELVIEW);
  glPushMatrix   ();                              // save modelview matrix 
  glLoadIdentity ();

  glPushAttrib (GL_LIGHTING_BIT);    
  glColorMaterial (GL_FRONT, GL_EMISSION);
  glEnable (GL_COLOR_MATERIAL);

  Border.Set ();

  gluOrtho2D (0.0,  WinX,  0.0, WinY);   // = glOrtho (0, 1, 0, 1, -1, 1)
  glViewport (0,       0, WinX, WinY);

  if (Clipping[0]+Clipping[2] < 1.0)    // vertical separator-line
  {
    glBegin (GL_LINES);
      glVertex2i (ViewPort[0]+ViewPort[2], ViewPort[1]);
      glVertex2i (ViewPort[0]+ViewPort[2], ViewPort[1]+ViewPort[3]);
    glEnd ();
  }

  if (Clipping[1]+Clipping[3] < 1.0)    // horizontal separator-line
  {
    glBegin (GL_LINES);
      glVertex2i (ViewPort[0],             ViewPort[1]+ViewPort[3]);
      glVertex2i (ViewPort[0]+ViewPort[2], ViewPort[1]+ViewPort[3]);
    glEnd ();
  }

  glDisable (GL_COLOR_MATERIAL);   
  glPopAttrib ();

  glPopMatrix ();                  

  glMatrixMode (GL_PROJECTION);
  glPopMatrix ();
  
  PRO_END
}


void VIEW::SetWinSize (int SizeX, int SizeY)
{
  PRO_METHOD ("VIEW", "")

  WinX = SizeX;
  WinY = SizeY;

  CalcViewPort ();

  PRO_END
}


void VIEW::SetCamera (RPY Frame)
{
  Camera.Place (Frame);
}


void VIEW::SetPrjMode (MODE Mode /* = ORHTO */)
{
  PrjNo = (Mode == PERSPECT);
}


void VIEW::SetClipping (double *Clipping)
{
  PRO_METHOD ("VIEW", "")

  memcpy (this->Clipping, Clipping, sizeof(double) * 4);

  CalcViewPort ();

  PRO_END
}


void VIEW::SetRange (double *Range)
{
  Projection[PrjNo]->SetRange (Range);
}


void VIEW::SetScale (double Scale /* = 1.0 */) 
{
  this->Scale = Scale;
}


void VIEW::GetSize (int *SizeX, int *SizeY) const
{
  if (SizeX != NULL)
    *SizeX = ViewPort[2];
  if (SizeY != NULL)
    *SizeY = ViewPort[3];
}


double VIEW::GetScale () const
{
  return Scale;
}


CAMERA *VIEW::GetCamera ()
{
  return &Camera;
}


double *VIEW::GetClipping ()
{
  return Clipping;
}


double *VIEW::GetRange ()
{
  return Projection[PrjNo]->GetRange ();
}


COLOR &VIEW::GetBorder ()
{
  return Border;
}


void VIEW::GetCOI (double *x, double *y, double *z)
{
  Projection[PrjNo]->GetCOI (x, y, z);
}


bool VIEW::IsInView (int MouseX, int MouseY) const
{
  MouseY = WinY - MouseY;              // Swap window origin's y-coordinate

  return (ViewPort[0]               <= MouseX &&
	  ViewPort[0] + ViewPort[2] >  MouseX &&
	  ViewPort[1]               <= MouseY &&
	  ViewPort[1] + ViewPort[3] >  MouseY);
}


// Computes the Coordinates relative to this view - even, if the coordinates
// are outside (return-value = FALSE)
bool VIEW::GetCoordinates (int MouseX, int MouseY, int *px, int *py) const
{
  PRO_METHOD ("VIEW", "")

  if (px != NULL)
    *px = MouseX - ViewPort[0];
  if (py != NULL)
    *py = (WinY - MouseY) - ViewPort[1];

  PRO_END
  return IsInView (MouseX, MouseY);
}


VIEW::MODE VIEW::GetPrjMode ()
{
  return (PrjNo==0)? ORTHO : PERSPECT;
}


bool VIEW::IsFullScreen ()
{
  // Statisch speichern -> Schneller
  return (Clipping[0] == 0.0 && Clipping[1] == 0.0 &&
	  Clipping[2] == 1.0 && Clipping[2] == 0.0);
}


void VIEW::CalcViewPort ()
{
  ViewPort[0] = int (Clipping[0] * WinX);    // x-position
  ViewPort[1] = int (Clipping[1] * WinY);    // y-postion
  ViewPort[2] = int (Clipping[2] * WinX);    // width
  ViewPort[3] = int (Clipping[3] * WinY);    // height
}


// Auslagern in Methode von Camera-Klasse ?
void VIEW::SetModelView ()
{
  PRO_METHOD ("VIEW", "")

  glLoadIdentity ();
  Camera.Set     ();

  PRO_END
}


void VIEW::ShowObject (int ID, H_MATRIX Matrix)
{
  glViewport (ViewPort[0], ViewPort[1], ViewPort[2], ViewPort[3]);
 
  if (! IsFullScreen())
    Projection[PrjNo]->Set ();

  SetModelView ();

  glPushMatrix ();
  glLoadMatrixd ((GLdouble*)Matrix.m);
  glCallList (ID);
  glPopMatrix ();
  glFlush ();
}
