#include <assert.h>
#include <stdarg.h>

#include <libtools.h>

#include "libcad3d.h"


static void WinHandle (Widget widget, XtPointer Data, XEvent *Event)
{
  if (Data == NULL)
    return;

  ((WINDOW *)Data)->EventHandler (widget, Event);
}


static void WinExitCB (Widget widget, WINDOW *Data, 
		                      XmAnyCallbackStruct *Ptr)
{
  XEvent Event;

  if (Data == NULL)
    return;

  Event.type = DestroyNotify;
  ((WINDOW *)Data)->EventHandler (widget, &Event);
}


WINDOW::WINDOW (Display *Displ, int Screen, Widget Application,
		const char *Title, double Width, double Height) :
  XDisplay (Displ),
  XScreen  (Screen),
  Shell    (NULL),
  PosX     (0),
  PosY     (0)
{
  PRO_METHOD ("WINDOW", "")

  if (Displ == NULL || Application == NULL)
    return;

  SetTitle (Title);

  TOOLGetScrValue (XDisplay, XScreen, Width, Height, &SizeX, &SizeY);

  Shell = XtVaCreatePopupShell ("window", 
				transientShellWidgetClass, Application, 
				XmNtitle,             this->Title,
				XmNdeleteResponse,    XmDESTROY,
				XmNtransient,         True,
				XmNmappedWhenManaged, True,
				XmNwidth,             SizeX,
				XmNheight,            SizeY,
				NULL);
  XtAddEventHandler (Shell, PointerMotionMask   | ExposureMask        |  
   		            EnterWindowMask     | LeaveWindowMask     |
	                    ButtonPressMask     | ButtonReleaseMask   |
	                    KeyPressMask        | KeyReleaseMask      |
 		            StructureNotifyMask | VisibilityChangeMask,
		            FALSE, 
	                    (XtEventHandler)WinHandle, (void *)this);

  XtAddCallback (Shell, XmNdestroyCallback, (XtCallbackProc)WinExitCB, 
		                            (void *)this);

  XtManageChild   (Shell);
  XtRealizeWidget (Shell);
  XtPopup         (Shell, XtGrabNonexclusive);
  XFlush          (XDisplay);

  Win = XtWindow  (Shell);

  SetMouseCursor ();

  PRO_END
}


WINDOW::WINDOW (Display *Displ, int Screen, Widget Parent) :
  XDisplay (Displ),
  XScreen  (Screen),
  Shell    (Parent),
  PosX     (0),
  PosY     (0)
{
  PRO_METHOD ("WINDOW", "")

  assert (Shell != NULL);

  Title[0] = '\0';
  XtAddEventHandler (Shell, PointerMotionMask   | ExposureMask        |
		            EnterWindowMask     | LeaveWindowMask     |
	                    ButtonPressMask     | ButtonReleaseMask   |
	                    KeyPressMask        | KeyReleaseMask      |
 		            StructureNotifyMask | VisibilityChangeMask,
		            FALSE,
		            (XtEventHandler)WinHandle, (void *)this);

  XtManageChild   (Shell);
  XtRealizeWidget (Shell);
  XFlush (XDisplay);
  Win = XtWindow (Shell);

  SetMouseCursor ();

  PRO_END
}


WINDOW::~WINDOW ()
{
  PRO_METHOD ("WINDOW~", "")

  if (Shell == NULL)
    return;

  XtRemoveEventHandler (Shell, PointerMotionMask, FALSE, 
			       (XtEventHandler)WinHandle, (void *)this);

  XtUnmapWidget   (Shell);
  XtDestroyWidget (Shell);

  PRO_END
}


COLOR &WINDOW::GetBackground ()
{
  return bg;
}


COLOR &WINDOW::GetForeground ()
{
  return fg;
}


int WINDOW::GetSizeX () const
{
  return SizeX;
}


int WINDOW::GetSizeY () const
{
  return SizeY;
}


// TODO: Methode in WINDOW-Klasse verlegen
int WINDOW::GetMouse (XtAppContext AppCon, int &MouseX, int &MouseY)
{
  XEvent Event;

  PRO_METHOD ("WINDOW", "")

  XtRemoveEventHandler (Shell, PointerMotionMask, FALSE, 
			       (XtEventHandler)WinHandle, (void *)this);

  for (;;)
  {
    if (XtAppPeekEvent (AppCon, &Event))         // is there an event ?
    {
      XtAppNextEvent (AppCon, &Event);           // yes: get it

      if (Event.type == ButtonPress &&
	  Event.xbutton.window == XtWindow(Shell))
      {
	MouseX = Event.xbutton.x;
	MouseY = Event.xbutton.y;

	// Continue handling Motion-Events
	XtAddEventHandler (Shell, PointerMotionMask, FALSE, 
			          (XtEventHandler)WinHandle, (void *)this);

        PRO2 ("Button pressed: %d / %d\n", MouseX, MouseY)
	PRO_RETURN ("Button pressed", Event.xbutton.button)
      }
      else
        XtDispatchEvent (&Event);                // evaluate event
    }
    else
      XtAppProcessEvent (AppCon, XtIMTimer);  
  }

  PRO_END
  return -1;
}


void WINDOW::SetPos (double PosX /* = 0.0 */, double PosY /* = 0.0 */)
{
  TOOLGetScrValue (XDisplay, XScreen, PosX, PosY, &this->PosX, &this->PosY);  

  XtVaSetValues (Shell, XmNx, this->PosX,
		        XmNy, this->PosY,
		        NULL);
}


void WINDOW::SetSize (double SizeX /* = 250.0 */, double /* SizeY = 250.0 */)
{
  TOOLGetScrValue (XDisplay, XScreen, 
		   SizeX, SizeY, &this->SizeX, &this->SizeY);

  XtVaSetValues (Shell, XmNwidth,  this->SizeX,
		        XmNheight, this->SizeY,
		        NULL);
}


void WINDOW::SetMouseCursor (int Shape /* = ToolNORMAL */)
{
  PRO_METHOD ("WINDOW", "")

  TOOLSetMouseCursor (Shell, Shape, NULL, NULL);

  PRO_END
}


void WINDOW::SetTitle (const char *Title, ...)
{
  va_list Arguments;

  PRO_METHOD ("WINDOW", "")

  if (Title != NULL && strlen(Title) < 256)
  {
    va_start (Arguments, Title);
    vsprintf (this->Title, Title, Arguments);
    va_end   (Arguments);
    this->Title[255] = '\0';
  }
  else
    strcpy (this->Title, "window");

  if (Shell != NULL)
    XtVaSetValues (Shell, XmNtitle, this->Title, NULL);

  PRO_END
}  


Display *WINDOW::GetDisplay () const
{
  return XDisplay;
}


Widget WINDOW::GetShell () const
{
  return Shell;
}


Window WINDOW::GetWindow () const
{
  return Win;
}


void WINDOW::SetBackground (const COLOR &Color /* = COLOR(COLOR::BLACK)*/)
{
  bg = Color;
}


void WINDOW::SetForeground (const COLOR &Color /* = COLOR(COLOR::WHITE)*/)
{
  fg = Color;
}


void WINDOW::EventHandler (Widget widget, XEvent *Event)
{
  // No events needs to be handled, yet.
  switch (Event->type)
  {
    case MapNotify :
      PRO0 ("MapNotify\n")
      break;

    case ReparentNotify :
      PRO0 ("Reparent\n")
      break;

    case LeaveNotify :
      break;

    default :
      PRO1 ("WINDOW.EventHandler: Event %d not handled\n", Event->type)
  }
}


