#include <stdarg.h>
#include <assert.h>

#include <libtools.h>

#include "libcad3d.h"


// #define USE_MESA


enum MOUSE_MODES
{
  MOUSE_NONE = 0,
  MOUSE_PRESS,
  MOUSE_START,
  MOUSE_DRAG,
  MOUSE_DCLICK
};


typedef struct
{
  unsigned int  Button;
  int           Mode;
  int           MouseX, MouseY;
  WORLD        *World;
  XtIntervalId  Timer;
} MOUSE_DATA;


static Window Current         = None;
static bool   locked          = FALSE;
static int    Culling[4]      = {0, GL_FRONT, GL_BACK, GL_FRONT_AND_BACK};
static int    RenderTab[]     = {GL_POINT, GL_LINE, GL_LINE, GL_FILL};
static int    Shade[]         = {0, GL_FLAT, GL_SMOOTH};
static double SplitTable[][4] =
{ // x   y       w     l
  {0.00, 0.00, 1.00, 1.00},           //  0: full window     (   1)

  {0.00, 0.00, 0.50, 1.00},           //  1: left half       (   2)
  {0.50, 0.00, 0.50, 1.00},           //  2: right half      (   4)

  {0.00, 0.50, 0.50, 0.50},           //  3: upper left      (   8)
  {0.50, 0.50, 0.50, 0.50},           //  4: upper right     (  16)
  {0.00, 0.00, 0.50, 0.50},           //  5: lower left      (  32)
  {0.50, 0.00, 0.50, 0.50},           //  6: lower right     (  64)

  {0.00, 0.00, 0.66, 1.00},           //  7: 2/3 left         ( 128)
  {0.66, 0.66, 0.34, 0.33},           //  8: 1/3 upper right  ( 256)
  {0.66, 0.33, 0.34, 0.34},           //  9: 1/3 middle right ( 512)
  {0.66, 0.00, 0.34, 0.33}            // 10: 1/3 lower right  (1024)
};


#define SPLIT_MODES ((int)(sizeof(SplitTable) / (sizeof(double)*4)))
#define DOUBLE_CLICK_SPEED 200  // [Milliseconds]

#define RESET_TIMER(Timer) \
	  if (Timer != 0) \
          { \
	    XtRemoveTimeOut (Timer); \
	    Timer = 0; \
          }

// =========================================================================


WINDOW3D::WINDOW3D (WORLD *World, 
		    Display *Displ, int Screen, Widget Application, 
		    const char *Name, 
		    double Width /* = 250 */, double Height /* = 250 */) :
  WINDOW (Displ, Screen, Application, Name, Width, Height),
  World  (World),
  WinGL  (None)
{
  PRO_METHOD ("WINDOW3D", "\n")

  Initialize ();

  PRO_END
}


WINDOW3D::WINDOW3D (WORLD *World, Display *Displ, int Screen, Widget Parent) :
  WINDOW (Displ, Screen, Parent),
  World  (World),
  WinGL  (None)
{
  PRO_METHOD ("WINDOW3D", "\n")

  Initialize ();

  PRO_END
}


WINDOW3D::~WINDOW3D ()
{
  int count;

  PRO_METHOD ("WINDOW3D~", "")

#ifdef USE_MESA
  glXReleaseBuffersMESA (XDisplay, Win);
#endif

  if (Fog != NULL)
    delete Fog;

  for (count = 0; count < GetViewNo(); ++count)
    delete Views[count];

  PRO_END
}


void WINDOW3D::Initialize ()
{
  XSetWindowAttributes  Attr;
  Colormap              ColorMap;

  PRO_METHOD ("WINDOW3D", "\n")

  WinGL = Win;
  ColorMap = XCreateColormap (XDisplay, 
			      RootWindow(XDisplay, 
					 OpenGL->GetVisualInfo()->screen),
			      OpenGL->GetVisualInfo()->visual, AllocNone);
  Attr.colormap     = ColorMap;
  Attr.border_pixel = 0;
  XChangeWindowAttributes (XDisplay, WinGL, CWBackPixel | CWColormap, &Attr);

  if (! Select())
    PRO_ABORT ("MakeCurrent failed")

  Views[0] = new VIEW (SizeX, SizeY);
  Views[1] = NULL;
  Views[2] = NULL;
  Views[3] = NULL;

  Data.Render[0] = WIRE_FRAME;
  Data.Render[1] = FACES;

  Data.Aliasing = ANTIALIASING_OFF;

  InitGL ();

  Fog = new FOG ();

  PRO_END
}


bool WINDOW3D::InitGL ()
{
  PRO_METHOD ("WINDOW3D", "\n")

  // Use Open-GL default values of light 0
  GLfloat Light_Ambient[]  = {0.0, 0.0,    0.0,  1.0};
  GLfloat Light_Diffuse[]  = {1.0, 1.0,    1.0,  1.0};   // Light-Mask
  GLfloat Light_Specular[] = {1.0, 1.0,    1.0,  1.0};
  GLfloat Light_Position[] = {0.0, 0.0,    1.0,  0.0};
  // GLfloat Light_Direction[] = { 1,   1,   -0.5};

  // Use default OpenGL values for ambient light model
  GLfloat LModel_Ambient[] = {0.2, 0.2, 0.2, 1.0};

#if 0
  glColorMaterial (GL_FRONT_AND_BACK, GL_AMBIENT);
  glEnable        (GL_COLOR_MATERIAL);
#endif

  // drawing-direction for front faces (default is counterclockwise)
  SetRotary (COUNTERCLOCKWISE);

  // back-face drawing even if you're inside an object
  SetCulling (BACK);

  // enable lighting
  glEnable (GL_LIGHTING);      // It's make no sense to turn off lighting

  // Set ambient component for basic lightning
  glLightModelfv (GL_LIGHT_MODEL_AMBIENT, LModel_Ambient);

  // Set default parameter for light 0
  glLightfv (GL_LIGHT0, GL_AMBIENT,  Light_Ambient);
  glLightfv (GL_LIGHT0, GL_DIFFUSE,  Light_Diffuse);
  glLightfv (GL_LIGHT0, GL_SPECULAR, Light_Specular);
  glLightfv (GL_LIGHT0, GL_POSITION, Light_Position);
  glEnable  (GL_LIGHT0);

  // Vector normalization
  glEnable (GL_NORMALIZE);     // Normalize normal vectors when scaling

  // enable testing Alpha channel: (TRANSPARENT, SOLID)
  SetBlending (SOLID);

  //  set shade model (GL_FLAT, GL_SMOOTH, ...)
  SetShading (SMOOTH);

  // polygon-mode for front- and back-side
  SetPolygonMode (Data.Render[0], BACK_SIDE);
  SetPolygonMode (Data.Render[1], FRONT_SIDE);

  // depth-function
  glDepthFunc (GL_LEQUAL);
  glEnable    (GL_DEPTH_TEST);

  // set background (clear) color
  SetBackground ();
  glClear       (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT |
		 GL_STENCIL_BUFFER_BIT);

  glMatrixMode (GL_MODELVIEW);

  PRO_END
  return TRUE;
}


bool WINDOW3D::Select (SELECT_MODE lock /* = LOCK */)
{
  PRO_METHOD ("WINDOW3D", "")

  if (Current != WinGL && ! locked)
  {
    PRO0 ("Make current...\n")
    if (! glXMakeCurrent(XDisplay, WinGL, OpenGL->GetContext()))
      PRO_RETURN ("\\glxMakeCurrent failed...", FALSE)
    Current = WinGL;
  }

  /* This method attaches a GLX context to the X window stored in this class.
     It will be called primary on expose-, visibility-events or when the mouse
     enters a CAD-window. This behaviour still causes problems when loading 
     workcells: If a new workcell is created, first the OpenGL window is
     opened, because all parts need to create a appropriate display list-ID.
     Well, while loading needs some time, we need to take care, that the GLX
     context won't be swapped to another window while we are loading and 
     creating the parts. E.g. this could happened because of a closed file 
     selector, which might create an expose event for another TuxCAD-window. 
     So, when starting such a process, we lock the context here.
  */
  locked = (lock == LOCK || (locked && lock == CHECK));

  PRO_END
  return TRUE;
}


void WINDOW3D::Flush ()
{
  PRO_METHOD ("WINDOW3D", "")

  glXSwapBuffers (XDisplay, WinGL);

  PRO_END
}


WINDOW3D::POLYGON_MODE WINDOW3D::GetPolygonMode (SIDE Side /* =FRONT_SIDE */) 
                                                 const
{
  return Data.Render[(int)Side-1]; 
}


WINDOW3D::SHADING WINDOW3D::GetShading () const
{ 
  return Data.Shade; 
}


WINDOW3D::BLENDING WINDOW3D::GetBlending () const 
{
  return Data.Blend; 
}


WINDOW3D::CULLING WINDOW3D::GetCulling () const
{
  return Data.Culling;
}


WINDOW3D::ROTARY WINDOW3D::GetRotary () const
{
  return Data.Rotary;
}


WINDOW3D::ALIASING WINDOW3D::GetAliasing () const
{
  return Data.Aliasing;
}


VIEW *WINDOW3D::GetView (int No)
{
  assert (No >= 0 && No < 4);

  return Views[No];
}


int WINDOW3D::GetViewNo () const
{
  int count, Result = 0;

  for (count = 0; count < 4; ++count)
    if (Views[count] != NULL)
      Result++;

  return Result;
}


int WINDOW3D::GetViewNo (int px, int py) const
{
  int count;

  for (count = 0; count < GetViewNo(); ++count)
    if (Views[count]->IsInView(px, py))
      return count;

  return -1;
}


FOG *WINDOW3D::GetFog () const
{
  return Fog;
}


void WINDOW3D::SetBackground (const COLOR &Color /* = COLOR(COLOR::BLACK)*/)
{
  WINDOW::SetBackground (Color);

  glClearColor (Color[RED], Color[GREEN], Color[BLUE], Color[ALPHA]);
}


void WINDOW3D::SetPolygonMode (POLYGON_MODE Mode, int Side /* = FRONT_SIDE */)
{
  PRO_METHOD ("WINDOW3D", "")

  if (Side & BACK_SIDE)
    Data.Render[0] = Mode;
  if (Side & FRONT_SIDE)
    Data.Render[1] = Mode;

  if (Mode != HIDDEN_LINE)
  {
    glPolygonMode (GL_BACK,  (GLenum)RenderTab[(int)Data.Render[0]]);
    glPolygonMode (GL_FRONT, (GLenum)RenderTab[(int)Data.Render[1]]);
  }

  PRO_END
}


void WINDOW3D::SetShading (SHADING Shading)
{
  PRO_METHOD ("WINDOW3D", "")

  Data.Shade = Shading;
 
  glShadeModel ((GLenum)Shade[(int)Data.Shade]);

  PRO_END
}
  

void WINDOW3D::SetBlending (BLENDING Blending)
{
  PRO_METHOD ("WINDOW3D", "")

  Data.Blend = Blending;

  if (Data.Blend == TRANSPARENT)
  {
    glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glEnable (GL_BLEND);
    glEnable (GL_ALPHA_TEST);
  }
  else
  {
    glDisable (GL_BLEND);
    glDisable (GL_ALPHA_TEST);
  }

  PRO_END
}


void WINDOW3D::SetCulling (CULLING Cull)
{
  PRO_METHOD ("WINDOW3D", "")

  Data.Culling = Cull;

  if (Data.Culling == OFF)
    glDisable (GL_CULL_FACE);
  else
  {
    glCullFace ((GLenum)Culling[(int)Data.Culling]);
    glEnable (GL_CULL_FACE);
  }
 
  PRO_END
}


void WINDOW3D::SetRotary (ROTARY Rotary)
{
  PRO_METHOD ("WINDOW3D", "")  

  Data.Rotary = Rotary;

  glFrontFace ((Data.Rotary == CLOCKWISE)? GL_CW : GL_CCW);

  PRO_END
}


void WINDOW3D::SetAliasing (ALIASING Aliasing)
{
  PRO_METHOD ("WINDOW3D", "")

  Data.Aliasing = Aliasing;

  if (Data.Aliasing != ANTIALIASING_OFF)
  {
    glHint (GL_LINE_SMOOTH_HINT, 
	    (Data.Aliasing == ANTIALIASING_FASTEST)? GL_FASTEST :  GL_NICEST);
    glHint (GL_POLYGON_SMOOTH_HINT, 
	    (Data.Aliasing == ANTIALIASING_FASTEST)? GL_FASTEST :  GL_NICEST);
    glEnable (GL_LINE_SMOOTH);
    glEnable (GL_POLYGON_SMOOTH);
    glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glEnable (GL_BLEND);
    glEnable (GL_ALPHA_TEST);
  }
  else
  {
    glDisable (GL_LINE_SMOOTH);
    glDisable (GL_POLYGON_SMOOTH);
    SetBlending (Data.Blend);
  }

  PRO_END
}


void WINDOW3D::SetProjection (int Mode)
{
  int count;

  PRO_METHOD ("WINDOW3D", "")

  for (count = 0; count < GetViewNo(); ++count)
    Views[count]->SetPrjMode ((VIEW::MODE)Mode);

  PRO_END
}


void WINDOW3D::SetFog (FOG *Fog)
{
  
  PRO_METHOD ("WINDOW3D", "")

  if (this->Fog != NULL)
    delete this->Fog;

  this->Fog = Fog;

  Fog->Set ();

  PRO_END
}


void WINDOW3D::SetViews (int Mask)
{
  int   count, Number = 0;
  RPY   Camera;

  PRO_METHOD ("WINDOW3D", "\n")

  for (count = 0; count < SPLIT_MODES; ++count)
    if (Mask & (1 << count))
    {
      if (Views[Number] == NULL)            // Add new view if not existant
	Views[Number] = new VIEW (Views[0]);

      Views[Number]->SetWinSize  (SizeX, SizeY);
      Views[Number]->SetClipping (SplitTable[count]);

      Number++;
    }

  // When reduce amount of views, delete the unused ones
  for (count = Number; count < 4; ++count)
    if (Views[count] != NULL)
    {
      delete Views[count];
      Views[count] = NULL;
    }

  Camera = Views[0]->GetCamera()->GetLocation ();
  for (count = 1; count < Number; ++count)
  {
    Camera.r += (360.0 / Number);
    Views[count]->SetCamera (Camera);
  }

  PRO_END
}


void WINDOW3D::Print (double x, double y, double z, const char *Msg)
{
  PRO_METHOD ("WINDOW3D", "")

  glRasterPos3d (x, y, z);
  glPushAttrib  (GL_LIST_BIT);
  glCallLists   (strlen(Msg), GL_UNSIGNED_BYTE, (GLubyte *)Msg);
  glPopAttrib   ();

  PRO_END
}


/* ********************************************************************** */
/* ** There are five interesting events using the mouse buttons:       ** */
/* ** 1:  Clicking a mouse button (MB)                                 ** */
/* ** 2a: Start dragging mouse (e.g. for changing the mouse cursor)    ** */
/* ** 2b: moving the mouse while pressing a button (dragging)          ** */
/* ** 2c: releasing MB after dragging (resetting mouse cursor)         ** */
/* ** 3:  double click on a mouse-button (within a specified time)     ** */
/* ********************************************************************** */
static void MouseTimer (MOUSE_DATA *Mouse, XtIntervalId ID)
{
  Mouse->Timer = 0;           // reset timer first, to avoid timer conflicts

  switch (Mouse->Mode)
  {
    case MOUSE_PRESS :
      Mouse->Mode = MOUSE_NONE;
      Mouse->World->ClickHandle (Mouse->Button, Mouse->MouseX, Mouse->MouseY);
      PRO3 ("-> MOUSE_PRESS #%d: %d / %d\n", 
	    Mouse->Button, Mouse->MouseX, Mouse->MouseY)
      break;

    case MOUSE_START :
      Mouse->World->ButtonHandle (Mouse->Button, Mouse->MouseX, Mouse->MouseY);
      Mouse->Mode = MOUSE_DRAG;
      PRO3 ("-> MOUSE DRAGGING STARTED #%d: %d / %d\n", 
	    Mouse->Button, Mouse->MouseX, Mouse->MouseY)
      break;

    case MOUSE_DCLICK :
      // Should never happens, because of deleting timer on button-release
      break;

    case MOUSE_DRAG :
      // Should never happens, because the timer expires with MOUSE_START in
      // order to switch to dragging-mode and no other timer will be started.
      break;

    /* Dragging and double click are NOT processed by timeout... */
    case MOUSE_NONE :
    default :
      PRO2 ("-> MOUSE #%x: %d\n", Mouse->Button, Mouse->Mode)
  }
}


void WINDOW3D::EventHandler (Widget widget, XEvent *Event)
{
  static MOUSE_DATA Mouse = {0, MOUSE_NONE, 0, 0, NULL, 0};

  if (World == NULL || widget == NULL)
    return;

  PRO1 (">>>>> Event: %d\n", Event->type);

  Mouse.World = World;

  switch (Event->type)
  {
    case Expose :
      XEvent Ev;

      // Remove all other expose events for this window, because we refresh 
      // the whole window only once.
      PRO0 ("Expose\n")
      Select ();
      while (XCheckWindowEvent(XDisplay, Event->xexpose.window, 
			       ExposureMask, &Ev))
      {
      };
	
      //      Select ();
      World->Redraw ();
      break;

    case EnterNotify :
      PRO0 ("EnterNotify\n")
      if (XtIsRealized(widget))
	Select ();
      break;

    case MotionNotify :
      {
	XMotionEvent *Motion = (XMotionEvent *)Event;

        if (Mouse.Mode == MOUSE_DRAG)
	  if (Motion->state == 0)
	  {
	    // When button is released while mouse motion, the release event
	    // may be suppressed. So, stop dragging when no button is pressed
	    Mouse.Mode = MOUSE_NONE;
	  }
	  else
	  {
	    World->MouseHandle (Motion->state, Motion->x, Motion->y);
	    PRO3 ("-> MOUSE DRAGGING #%x: %d / %d\n",
		  Motion->state, Motion->x, Motion->y)
	  }
        elsif (Mouse.Mode == MOUSE_START && 
	       (abs(Motion->x-Mouse.MouseX) > 2 || 
	        abs(Motion->y-Mouse.MouseY) > 2))
        {
          // if mouse will be moved for more than 2 pixel while timer is 
          // running, switch to dragging mode.
	  RESET_TIMER (Mouse.Timer);
          Mouse.Mode  = MOUSE_DRAG;
	  World->ButtonHandle (Mouse.Button, Mouse.MouseX, Mouse.MouseY);
          PRO3 ("-> MOUSE DRAGGING2 STARTED #%x: %d / %d\n",
		Motion->state, Mouse.MouseX, Mouse.MouseY)
        }
	else
	  World->MouseHandle (0, Motion->x, Motion->y);
      }
      break;

    case ButtonPress :
      {
	XButtonEvent *Button = (XButtonEvent *)Event;

	PRO0 ("ButtonPress\n")

	if (Mouse.Mode != MOUSE_NONE && Mouse.Button != Button->button)
	{
	  RESET_TIMER (Mouse.Timer);
	  Mouse.Mode = MOUSE_NONE;
	  PRO3 ("-> MOUSE CHANGED #%d: %d / %d\n",
		Mouse.Button, Mouse.MouseX, Mouse.MouseY)
	}

	if (Mouse.Mode == MOUSE_NONE && Mouse.Timer == 0)
	{
	  // First time mouse button is pressed 
	  Mouse.Timer =
	    XtAppAddTimeOut (AppCon, DOUBLE_CLICK_SPEED, 
			     (XtTimerCallbackProc)MouseTimer, &Mouse);
	  Mouse.Mode   = MOUSE_START;
	  Mouse.Button = (int)Button->button;
	  Mouse.MouseX = Button->x;
	  Mouse.MouseY = Button->y;
#if 0
	  // Button handle will be called on timeout due to decide 
	  // wheter we have a single or a double click. This will cause a
	  // delayed reaction, but is safer to recognize single clicks.
	  World->ButtonHandle (Mouse.Button, Mouse.MouseX, Mouse.MouseY);
#endif
	}
	elsif (Mouse.Mode == MOUSE_PRESS && Mouse.Button == Button->button)
	{
	  // same button pressed within timeout
	  RESET_TIMER (Mouse.Timer);
	  Mouse.Mode = MOUSE_DCLICK;
	  // Don't start user select handle here, because of pending 
	  // release event ! Execution when button released. Otherwise, 
	  // if the user creates a popup-menu using the same shell as parent, 
	  // this shell would get the release event.
	}
      }
      break;

    case ButtonRelease :
      {
	XButtonEvent *Button = (XButtonEvent *)Event;

	PRO0 ("ButtonRelease\n")

	if (Mouse.Mode == MOUSE_START)
	{
	  // start within timeout -> single button press
	  Mouse.Mode = MOUSE_PRESS;
          // If no button will be pressed during timout, MouseHandle will be 
	  // called in timer-function
	}
	elsif (Mouse.Mode == MOUSE_DRAG)  /* in dragging: stop it */
	{
	  // Ther should be no timer active, because dragging will be started
	  // when the timer expires.
	  Mouse.Mode = MOUSE_NONE;
	  World->ButtonHandle (-Mouse.Button, Button->x, Button->y);
	  PRO3 ("-> MOUSE DRAGGING STOPPED #%x: %d / %d\n", 
		Mouse.Button, Button->x, Button->y)
	  Mouse.Button = 0;
	}
	elsif (Mouse.Mode == MOUSE_DCLICK)
	{
	  RESET_TIMER (Mouse.Timer);
	  Mouse.Mode = MOUSE_NONE;
	  World->SelectHandle (Mouse.Button, Mouse.MouseX, Mouse.MouseY);
	  PRO3 ("-> MOUSE DOUBLE CLICK #%x: %d / %d\n",
		Mouse.Button, Mouse.MouseX, Mouse.MouseY)
	  Mouse.Button = 0;
	}
      }
      break;

    case VisibilityNotify :
      {
	XVisibilityEvent *Visibility = (XVisibilityEvent *)Event;

	PRO0 ("VisibilityNotify\n")

        if (Visibility->window == WinGL)
	  Select ();
      }
      break;

    case ConfigureNotify :
      {
	int              count;
	XConfigureEvent *Configure = (XConfigureEvent *)Event;

	PRO0 ("ConfigureNotify\n")
        Select ();

	SizeX = Configure->width;
	SizeY = Configure->height;

	World->ResizeHandle (Configure->width, Configure->height);
	// XMoveResizeWindow (...)
	XResizeWindow (XDisplay, WinGL, Configure->width, Configure->height);

	for (count = 0; count < GetViewNo(); ++count)
	  Views[count]->SetWinSize (Configure->width, Configure->height);
      }
      break;

    case KeyPress :
    case KeyRelease :
      {
	int               Modifier;
	KeySym            Symbol;
	XKeyPressedEvent *Key = (XKeyPressedEvent *)Event;

	PRO0 ("Key\n")

	Symbol   = XLookupKeysym (Key, 0);
	Modifier = Key->state;
	if (Event->type == KeyRelease)
	  Modifier *= -1;

	World->KeyHandle (Symbol, Modifier);
      }
      break;

    case DestroyNotify :
      PRO0 ("Destroy-Notify\n")
      if (! World->DestroyHandle())
	delete World;
      break;

    default :
      WINDOW::EventHandler (widget, Event);
  }
  PRO0 ("<<<<< Event DONE\n")
}
