#include <assert.h>
#include <stdarg.h>
#include <math.h>

#include <libtools.h>

#include "libcad3d.h"


#define DRAW_ALPHA   (1<<0)
#define DRAW_SOLID   (1<<1)
#define DRAW_SHADOW  (1<<2)


#if 0
static void       WORLD_PrintDev     (DEVICE3D *DevList, int Indent);
#endif
static void       WORLD_DrawDevList  (DEVICE3D *DevList, H_MATRIX M, 
				      GLenum      Mode,    int Type);
static PART3D    *WORLD_GetPart      (DEVICE3D *DevList, 
				      char *DevName, char *PartName);
static PART3D    *WORLD_GetPartID    (DEVICE3D *DevList, int ID);
static bool       WORLD_GetPartPos   (DEVICE3D *DevList, PART3D *Part, 
				      H_MATRIX &WTP);
static DEVICE3D  *WORLD_GetDev       (DEVICE3D *DevList, PART3D *Part);
static DEVICE3D  *WORLD_GetDev       (DEVICE3D *DevList, char *DevName);
static bool       WORLD_GetDevPos    (DEVICE3D *DevList, 
				      DEVICE3D *Dev, H_MATRIX &WTD);


WORLD::WORLD (const char *Name  /* = NULL */, 
                    int   SizeX /* = 250  */, int SizeY /* = 250 */) :
  CAD_FILE          (),
  DevList           (NULL),
  PartNum           (0),
  DevNum            (0),
  AlphaNum          (0),
  ViewNo            (0),
  Name              (NULL),
  PartHighLight     (COLOR::MEDRED),
  DevHighLight      (COLOR::LIME),
  UserResizeHandle  (NULL),
  UserMouseHandle   (NULL),
  UserButtonHandle  (NULL),
  UserSelectHandle  (NULL),
  UserClickHandle   (NULL),
  UserKeyHandle     (NULL),
  UserDestroyHandle (NULL)
{
  int count;

  PRO_METHOD ("WORLD", "\n")

  // Create CAD-window:
  Win = new WINDOW3D (this, XDisplay, XScreen, AppShell, Name, SizeX, SizeY);

  SetName (Name);

  for (count = MIN_X; count <= MAX_Z; ++count)
    Range[count] = 0.0;

  Floor = new FLOOR ();                 // Create floor

  // Initialize Lights:
  for (count = 0; count < OpenGL->GetMaxLights(); ++count)
  {
    Lights[count] = new LIGHT ((LIGHT::NUMBER)count);
    Lights[count]->GetShadow()->SetFloor (Floor);
  }

  glPolygonMode (GL_BACK, GL_LINE);

  PRO_END
}


WORLD::WORLD (Widget Parent, const char *Name /* = NULL */) :
  CAD_FILE          (),
  DevList           (NULL),
  PartNum           (0),
  DevNum            (0),
  AlphaNum          (0),
  ViewNo            (0),
  Name              (NULL),
  PartHighLight     (COLOR::MEDRED),
  DevHighLight      (COLOR::LIME),
  UserResizeHandle  (NULL),
  UserMouseHandle   (NULL),
  UserButtonHandle  (NULL),
  UserSelectHandle  (NULL),
  UserClickHandle   (NULL),
  UserKeyHandle     (NULL),
  UserDestroyHandle (NULL)
{
  int count;

  PRO_METHOD ("WORLD", "\n")

  // Create CAD-window:
  Win = new WINDOW3D (this, XDisplay, XScreen, Parent);

  SetName (Name);

  for (count = MIN_X; count <= MAX_Z; ++count)
    Range[count] = 0.0;

  Floor = new FLOOR ();                 // Create floor

  // Initialize Lights:
  for (count = 0; count < OpenGL->GetMaxLights(); ++count)
  {
    Lights[count] = new LIGHT ((LIGHT::NUMBER)count);
    Lights[count]->GetShadow()->SetFloor (Floor);
  }

  glPolygonMode (GL_BACK, GL_LINE);

  PRO_END
}


WORLD::~WORLD ()
{
  int       count;
  DEVICE3D *Dev, *Next;

  PRO_METHOD ("~WORLD", "\n")

  Dev = DevList;
  while (Dev != NULL)
  {
    Next = Dev->GetNext();
    delete (Dev);
    Dev = Next;
  }

  if (Name != NULL)
    delete Name;

  if (Floor != NULL)
    delete Floor;

  for (count = 0; count < OpenGL->GetMaxLights(); ++count)
    if (Lights[count] != NULL)
      delete Lights[count];

  if (Win != NULL)
    delete Win;

  PRO_END
}


char *WORLD::GetName ()
{
  return Name;
}


void WORLD::GetWinSize (int *SizeX, int *SizeY)
{
  if (SizeX != NULL)
    *SizeX = Win->GetSizeX();

  if (SizeY != NULL)
    *SizeY = Win->GetSizeY();
}


WINDOW3D *WORLD::GetWin ()
{
  return Win;
}


VIEW *WORLD::GetView (int No /* = -1 */)
{
  // if requested view-no is negative return current view
  if (No < 0)
    return Win->GetView (ViewNo);
  else
    return Win->GetView (No);
}


double WORLD::GetScale ()
{ 
  return GetView()->GetScale ();
}


RPY WORLD::GetCamera ()
{
  CAMERA *Camera;

  Camera = GetView()->GetCamera();

  return Camera->GetLocation ();
}


double *WORLD::GetRange ()
{  
  return GetView()->GetRange ();
}


void WORLD::GetCOI (double *x, double *y, double *z)
{
  GetView()->GetCOI (x, y, z);
}


COLOR WORLD::GetPartHighLight ()
{
  return PartHighLight;
}


COLOR WORLD::GetDevHighLight ()
{
  return DevHighLight;
}


LIGHT *WORLD::GetLight (LIGHT::NUMBER Number)
{
  if (Number < LIGHT::LIGHT1 || Number > (LIGHT::NUMBER)OpenGL->GetMaxLights())
    return NULL;

  return Lights[int(Number)];
}


bool WORLD::SetLight (LIGHT::NUMBER Number, LIGHT *Light /* = NULL */)
{
  if (Number < LIGHT::LIGHT1 || Number > (LIGHT::NUMBER)OpenGL->GetMaxLights())
    return FALSE;

  if (Lights[Number] != Light)
  {
    if (Lights[Number] != NULL) 
      delete Lights[Number];
    Lights[Number] = Light;
  }

  return TRUE;
}


FLOOR *WORLD::GetFloor () const
{
  return Floor;
}


void WORLD::SetFloor (FLOOR *Floor)
{
  if (this->Floor != NULL)
    delete this->Floor;

  this->Floor = Floor;
}


void WORLD::SetName (const char *Name, ...)
{
  char WorldName[513];  

  va_list Arguments;

  if (this->Name != NULL)
    delete this->Name;

  if (Name != NULL && strlen(Name) < 512)
  {
    va_start (Arguments, Name);
    vsprintf (WorldName, Name, Arguments);
    va_end   (Arguments);
    WorldName[512] = '\0';

    this->Name = new char[strlen(WorldName)+1];
    strcpy (this->Name, WorldName);
  }
  else
  {
    this->Name = new char[6];
    strcpy (this->Name, "World");
  }

  Win->SetTitle (this->Name);
}


void WORLD::SetScale (double Scale)
{
  int     count;
  double  Center;
  double *Range;
  VIEW   *View = GetView ();

  PRO_METHOD ("WORLD", "\n")

  assert (Scale > 0.0);

  Range = View->GetRange();

  // Sccale Viewport arround Center Of Interest (COI):  Change only 
  // ViewPort[0]..[3] will not walk through parts 
  for (count = 0; count < 6; count+=2)
  {
    Center = 0.5 * (Range[count] + Range[count+1]);
    Range[count] =   Center + Scale * (Range[count]-Center);
    Range[count+1] = Center + Scale * (Range[count+1]-Center);
  }
  Scale = View->GetScale () / Scale;

  // update scale: divide to have greater value for larger representation
  View->SetScale (Scale);
  View->SetRange (Range);

  Redraw (ViewNo);

  PRO_END
}


// Scale world, so that it fits in the window
void WORLD::Fill (double Factor /* = 1.0 */)
{
  double   MinX, MaxX;
  double   MinY, MaxY;
  double   MinZ, MaxZ;
  double   Radius;
  double  *Range;
  POINT3D  Center;
  VIEW    *View = GetView ();

  PRO_METHOD ("WORLD", "\n")

  Range = View->GetRange ();

  GetSize (&MinX, &MaxX, &MinY, &MaxY, &MinZ, &MaxZ);
  Center.x = 0.5 * (MinX + MaxX);
  Center.y = 0.5 * (MinY + MaxY);
  Center.z = 0.5 * (MinZ + MaxZ);

  Radius = 0.75 * sqrt (SQR(MaxX-MinX) + SQR(MaxY-MinY) + SQR(MaxZ-MinZ));

#if 0
  printf ("Fill-Range: X: %.1f...%.1f  Y: %.1f...%.1f  Z: %.1f...%.1f\n",
	  MinX, MaxX, MinY, MaxY, MinZ, MaxZ);
#endif

  Range[MIN_X] = Center.x - Radius;
  Range[MAX_X] = Center.x + Radius;

  Range[MIN_Y] = Center.y - Radius;
  Range[MAX_Y] = Center.y + Radius;

  Range[MIN_Z] = Center.z - 5.0 * Radius;
  Range[MAX_Z] = Center.z + 5.0 * Radius;

  View->SetScale (1.0);                 // Set scale to 1
  View->SetRange (Range);

  Redraw (ViewNo);

  PRO_END
}


void WORLD::SetRange (double *Range)
{
  GetView ()->SetRange (Range);
  Redraw (ViewNo);
}


void WORLD::SetCamera (RPY Position)
{
  GetView ()->SetCamera (Position);
  Redraw (ViewNo);
}


void WORLD::RotateView (double rx, double ry, double rz)
{
  GetView ()->Rotate (rx, ry, rz);
  Redraw (ViewNo);
}


void WORLD::PanView (double DirX, double DirY)
{
  GetView ()->Pan (DirX, DirY);
  Redraw (ViewNo);
}


void WORLD::ZoomView (int x1, int y1, int x2, int y2)
{
  GetView ()->Zoom (x1, y1, x2, y2);
  Redraw (ViewNo);
}


void WORLD::SetViews (int Mask)
{
  Win->SetViews (Mask);
  Redraw ();
}


void WORLD::SetMouseCursor (int Shape)
{
  Win->SetMouseCursor (Shape);
}


void WORLD::SetPartHighLight (COLOR Color)
{
  PartHighLight = Color;
}


void WORLD::SetDevHighLight (COLOR Color)
{
  DevHighLight = Color;
}


void WORLD::SetUserResizeHandle (RESIZE_HANDLE ResizeHandle)
{
  UserResizeHandle = ResizeHandle;  
}


void WORLD::SetUserMouseHandle (MOUSE_HANDLE  MouseHandle)
{
  UserMouseHandle = MouseHandle;  
}


void WORLD::SetUserButtonHandle (BUTTON_HANDLE ButtonHandle)
{
  UserButtonHandle = ButtonHandle;
}


void WORLD::SetUserSelectHandle (SELECT_HANDLE SelectHandle)
{
  UserSelectHandle = SelectHandle;
}


void WORLD::SetUserClickHandle (CLICK_HANDLE ClickHandle)
{
  UserClickHandle = ClickHandle;
}


void WORLD::SetUserKeyHandle (KEY_HANDLE KeyHandle)
{
  UserKeyHandle = KeyHandle;
}


void WORLD::SetUserDestroyHandle (DESTROY_HANDLE DestroyHandle)
{
  UserDestroyHandle = DestroyHandle;
}


void WORLD::IncAlpha (int Value /* = 1 */)
{
  AlphaNum += Value;

  if (AlphaNum > 0)
    Win->SetBlending (WINDOW3D::TRANSPARENT);
}


void WORLD::DecAlpha (int Value /* = 1 */)
{
  AlphaNum = MAX (0, AlphaNum-Value);

  if (AlphaNum == 0)
    Win->SetBlending (WINDOW3D::SOLID);
}


void WORLD::Clear ()
{
  if (DevList != NULL)
    delete (DevList);
  DevList = NULL;

  Redraw ();
}


// ============================== DRAW ==================================
// TODO: In Part verlegen
static void WORLD_DrawPartList (PART3D *PartList, H_MATRIX M, GLenum Mode, 
				int Type)
{
  double    Alpha;
  PART     *Part;
  PART3D   *Chain, *Ptr;
  DEVICE3D *Attach;

  Ptr = PartList;
  while (Ptr != NULL)
  {
    glPushMatrix ();                    // Save ViewPort (parts have same base)

    Part = Ptr->GetPart ();

    if (Mode == GL_SELECT)
    {
      PRO2 ("Part %d -> %s\n", Ptr->GetID(), Part->PartName)
      glLoadName (Ptr->GetID());
    }

    // Compute actual drawing frame
    glMultMatrixd ((GLdouble *)Ptr->GetRel().m);

    // Draw the part and associated data
    if ((Alpha = Ptr->GetMaterial()->GetAlpha()) > 0.0 && 
        (((Type & DRAW_SOLID) && Alpha == 1.0) || 
	 ((Type & DRAW_ALPHA) && Alpha <  1.0)))
      Ptr->Draw ();

    // Draw part in chain list
    if ((Chain = Ptr->GetChain()) != NULL)
      WORLD_DrawPartList (Chain, Part->WTP, Mode, Type);

    // Draw attached devices
    if ((Attach = Ptr->GetAttach()) != NULL)
      WORLD_DrawDevList (Attach, Part->WTP, Mode, Type);

    glPopMatrix ();                     // Restore ViewPort matrix

    Ptr = Ptr->GetNext ();
  }
}


static void WORLD_DrawDevList (DEVICE3D *DevList, H_MATRIX M, 
			       GLenum Mode, int Type)
{
  DEVICE3D *Dev;

  Dev = DevList;
  while (Dev != NULL)
  {
    glPushMatrix ();
    glMultMatrixd ((GLdouble*)Dev->GetWTD().m);

    if (Dev->GetAlpha() > 0.0)
      WORLD_DrawPartList (Dev->GetPartList(), M, Mode, Type);

    glPopMatrix ();

    Dev = Dev->GetNext ();
  }
}


void WORLD::DrawShadow ()
{
  H_MATRIX  Base;
  SHADOW   *Shadow;

  // TODO: Stencil-Buffer, if shadow-color has alpha < 1.0

#ifdef HAVE_STENCIL_BUFFER
  // To avoid redrawing the shadow, draw only if stencil buffer is < 2.
  glStencilFunc (GL_LESS, 2, 0xffffffff);    
  glStencilOp   (GL_REPLACE, GL_REPLACE, GL_REPLACE);
#endif

  if (AlphaNum == 0)   // Activate Alpha blending when not done before
  {
    glEnable    (GL_BLEND);
    glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  }

  glDisable (GL_LIGHTING);
  glEnable  (GL_COLOR_MATERIAL);

  for (int count = 0; count < OpenGL->GetMaxLights(); ++count)
  {
    Shadow = Lights[count]->GetShadow ();
    if (Lights[count]->GetMode() != LIGHT::OFF && 
	Shadow->GetMode() == SHADOW::ON)
    {
      Shadow->GetColor().Set ();
      // TODO: Bei parts mit Textur wird der Schatten auch mit Textur
      //       gezeichnet, da disable lighting nicht wirkt 
      //       Bei durchsichtigen parts msste die Shadow-Farbe ebenfalls
      //       transparenter (evtl. Objektfarbe) sein. 
      // => Lsung: Trennung von Farbe und Bauteildaten (CAD_PART::Create)
      // berlegen: Enable/Disable Texture aus TEXTURE::Set heraus und hier
      // auf disable setzen ???
      glPushMatrix ();
        // Set the matrix that maps the workcell on the specified plane
        Base = Shadow->GetMatrix ();
        glMultMatrixd ((GLdouble *)Base.m);
        WORLD_DrawDevList (DevList, Matrix_1, GL_RENDER, DRAW_SOLID);

      glPopMatrix ();
    }
  }

  glDisable (GL_COLOR_MATERIAL);
  glEnable  (GL_LIGHTING);

  if (AlphaNum == 0)
    glDisable (GL_BLEND);  

#ifdef HAVE_STENCIL_BUFFER
  glDisable (GL_STENCIL_TEST);
#endif
}


void WORLD::Draw (VIEW *View, GLenum Mode /* = GL_RENDER */)
{
  int count;

  PRO_METHOD ("WORLD", "\n")

  if (Mode == GL_RENDER)
  {
    // Set current view parameter
    View->Set ();

    // Show stationary light-sources
    for (count = 0; count < OpenGL->GetMaxLights(); ++count)
      if (Lights[count] != NULL && Lights[count]->GetSize() > 0 &&
	  Lights[count]->GetPlacement() == LIGHT::STATIONARY)
	Lights[count]->Draw ();

    // Set Camera positon and orientation
    View->GetCamera()->Set ();

    // Draw the floor
    Floor->Draw ();
  }

#ifdef HAVE_STENCIL_BUFFER
  // Prepare for drawing shadow usign stencil buffer
  if (Shadow->GetMode() == SHADOW::ON && Mode == GL_RENDER)
  {
    glEnable      (GL_STENCIL_TEST);
    glStencilFunc (GL_ALWAYS, 3,       0xffffffff);
    glStencilOp   (GL_KEEP,   GL_KEEP, GL_REPLACE);
  }
#endif

  // Set and draw relative lights
  if (Mode == GL_RENDER)
    for (count = 0; count < OpenGL->GetMaxLights(); ++count)
      if (Lights[count] != NULL &&
	  Lights[count]->GetPlacement() == LIGHT::RELATIVE)
      {
	Lights[count]->Set ();
	if (Lights[count]->GetSize() > 0)
	  Lights[count]->Draw ();
      }

  // Draw all devices in the workcell...
  if (AlphaNum == 0 || Mode == GL_SELECT)    // selection mode or no blending 
    WORLD_DrawDevList (DevList, Matrix_1, Mode, DRAW_ALPHA | DRAW_SOLID);
  else
  {
    WORLD_DrawDevList (DevList, Matrix_1, Mode, DRAW_SOLID);
    WORLD_DrawDevList (DevList, Matrix_1, Mode, DRAW_ALPHA);
  }

  // Draw shadow
  if (Mode == GL_RENDER)
    DrawShadow ();

  PRO_END
}


void WORLD::Redraw (int ViewNo /* = -1 */)
{
  int   count;
  VIEW *View;

  PRO_METHOD ("WORLD", "\n")

  Select ();
  if (ViewNo < 0)                       // Redraw all views
  {
    for (count = 0; count < Win->GetViewNo(); ++count)
      if ((View = Win->GetView(count)) != NULL)
	Draw (View);

    for (count = 0; count < Win->GetViewNo(); ++count)
      if ((View = Win->GetView(count)) != NULL)
	View->DrawBorder ();
  }
  elsif ((View = Win->GetView(ViewNo)) != NULL)
  {
    Draw (View);
    View->DrawBorder ();
  }

  Win->Flush ();

  PRO_END
}


// ========================== PRINT =====================================
#if 0
static void WORLD_PrintPart (PART3D *PartList, int Indent)
{
  int       count;
  PART     *Part;
  PART3D   *Ptr, *Next;
  DEVICE3D *Attach;

  Next = PartList;
  while (Next != NULL)
  {
    for (count = 0; count < Indent; ++count)
      printf ("  ");

    if ((Part = Next->GetPart()) != NULL)
      printf ("  Part '%s': ", Part->PartName);
    printf ("#%d\n", Next->GetID());

    if ((Ptr = Next->GetChain()) != NULL)
      WORLD_PrintPart (Ptr, Indent+2);
    
    if ((Attach = Next->GetAttach()) != NULL)
    {
      for (count = 0; count < Indent; ++count)
        printf ("  ");
      printf ("  ATTACHED:\n");
      WORLD_PrintDev (Attach, Indent+1);
    }

    Next = Next->GetNext ();
  }
}

static void WORLD_PrintDev (DEVICE3D *DevList, int Indent)
{
  int       count;
  DEVICE3D *Dev;

  Dev = DevList;
  while (Dev != NULL)
  {
    for (count = 0; count < Indent; ++count)
      printf ("  ");
    printf ("Dev '%s': [%d]\n", Dev->GetName(), Dev->GetPartNo());

    WORLD_PrintPart (Dev->GetPartList(), Indent);
    Dev = Dev->GetNext ();
  }
}
#endif


void WORLD::Print ()
{
  PRO_METHOD ("WORLD", "")

#if 1
  Write ();
#else
  printf ("*** W O R L D - D A T A ***\n");
  printf ("Name:        %s\n",   GetName());
  printf ("PolygonMode  ");
  switch (Win->GetPolygonMode())
  {
    case WINDOW3D::POINTS :
      printf ("POINTS\n");
      break;

    case WINDOW3D::WIRE_FRAME :
      printf ("WIRE-FRAME\n");
      break;

    case WINDOW3D::HIDDEN_LINE :
      printf ("HIDDEN-LINE\n");
      break;

    case WINDOW3D::FACES :
      printf ("FACES\n");
      break;

    default :
      printf ("INVALID Mode\n");
  }
		    
  printf ("Blend:      %s\n",
	  (Win->GetBlending()==WINDOW3D::SOLID)? "SOLID":"TRANSPARENT");
  printf ("Devices:    %d\n", GetDevNum());
  printf ("Parts:      %d\n", GetPartNum());
  printf ("\n");

  WORLD_PrintDev (DevList, 0);
#endif

  PRO_END
}


DEVICE3D *WORLD::GetData ()
{
  return DevList; 
};


void WORLD::SetData (DEVICE3D *DevList)
{
  this->DevList = DevList;
}


static void WORLD_GetSize (DEVICE3D *DevList, 
			   double *MinX, double *MaxX,
			   double *MinY, double *MaxY,
			   double *MinZ, double *MaxZ,
			   bool First)
{
  POINT3D   Min, Max, Point;
  PART     *Part;
  PART3D   *Ptr, *Next;
  DEVICE3D *Dev;

  Dev = DevList;
  while (Dev != NULL)
  {
    Ptr = Dev->GetPartList ();
    while (Ptr != NULL)
    {
      Next = Ptr;
      while (Next != NULL)
      {
	Part = Next->GetPart ();
	if (First)
	{
	  Next->GetBox()->GetRange (&Min.x, &Max.x, 
				    &Min.y, &Max.y, 
				    &Min.z, &Max.z);
	  Point = GEOMatrixMult (Part->WTP, Min);
	  *MinX = Point.x;
	  *MinY = Point.y;
	  *MinZ = Point.z;
	  Point = GEOMatrixMult (Part->WTP, Max);
	  *MaxX = Point.x;
	  *MaxY = Point.y;
	  *MaxZ = Point.z;

	  First = FALSE;
	}
	else
	{
	  Next->GetBox()->GetRange (&Min.x, &Max.x, 
				    &Min.y, &Max.y, 
				    &Min.z, &Max.z);
	  Point = GEOMatrixMult (Part->WTP, Min);
	  *MinX = MIN (*MinX, Point.x);
	  *MinY = MIN (*MinY, Point.y);
	  *MinZ = MIN (*MinZ, Point.z);
	  Point = GEOMatrixMult (Part->WTP, Max);
	  *MaxX = MAX (*MaxX, Point.x);
	  *MaxY = MAX (*MaxY, Point.y);
	  *MaxZ = MAX (*MaxZ, Point.z);

	  Next = Next->GetNext ();
	}
      }

      if (Ptr->GetAttach() != NULL)
	WORLD_GetSize (Ptr->GetAttach(), 
		       MinX, MaxX, MinY, MaxY, MinZ, MaxZ, First);

      Ptr = Ptr->GetChain ();
    }
    Dev = Dev->GetNext ();
  }
}

void WORLD::GetSize (double *MinX, double *MaxX,
		     double *MinY, double *MaxY,
		     double *MinZ, double *MaxZ)
{
  WORLD_GetSize (DevList, &Range[MIN_X], &Range[MAX_X], 
		          &Range[MIN_Y], &Range[MAX_Y],
		          &Range[MIN_Z], &Range[MAX_Z],
		          TRUE);
  
  if (MinX != NULL)
    *MinX = Range[MIN_X];
  if (MaxX != NULL)
    *MaxX = Range[MAX_X];
  if (MinY != NULL)
    *MinY = Range[MIN_Y];
  if (MaxY != NULL)
    *MaxY = Range[MAX_Y];
  if (MinZ != NULL)
    *MinZ = Range[MIN_Z];
  if (MaxZ != NULL)
    *MaxZ = Range[MAX_Z];

}


bool WORLD::Select ()
{
  return Win->Select ();
}


bool WORLD::Get3DPos (int px, int py, double pz /* = 0.0 */, 
		      double *x /* = NULL */, double *y /* = NULL */,
		      double *z /* = NULL */)
{
  int     No;
  double  mx, my, mz;
  VIEW   *View;

  PRO_METHOD ("WORLD", "")

  if (PartNum <= 0)
    PRO_RETURN ("No parts in world", FALSE)

  View = Win->GetView (Win->GetViewNo(px, py));
  No = View->SelNo (this, px, py, pz, &mx, &my, &mz);

  if (x != NULL)
    *x = mx;
  if (y != NULL)
    *y = my;
  if (z != NULL)
    *z = mz;

  PRO_END
  return (No >= 0);
}


H_MATRIX WORLD::GetPos (DEVICE3D *Dev)
{
  H_MATRIX WTD = Matrix_1;

  if (WORLD_GetDevPos(DevList, Dev, WTD))
    return WTD;

  return Matrix_1;
}


H_MATRIX WORLD::GetPos (PART3D *Part)
{
  H_MATRIX WTP = Matrix_1;

  if (WORLD_GetPartPos(DevList, Part, WTP))
    return WTP;

  return Matrix_1;
}


H_MATRIX WORLD::GetPos (DEVICE3D *Dev1, DEVICE3D *Dev2)
{
  if (Dev1 == NULL)
    if (Dev2 == NULL)
      return Matrix_1;
    else
      return Dev2->GetWTD ();
  elsif (Dev2 == NULL)
    return Dev1->GetWTD ();
  else
    return GEOMatrixProd (GEOMatrixInv(Dev1->GetWTD()), Dev2->GetWTD());
}


H_MATRIX WORLD::GetPos (DEVICE3D *Dev, PART3D *Part)
{
  if (Dev == NULL)
    if (Part == NULL)
      return Matrix_1;
    else
      return Part->GetWTP ();
  elsif (Part == NULL)
    return Dev->GetWTD();
  else
    return GEOMatrixProd (GEOMatrixInv(Dev->GetWTD()), Part->GetWTP());
}


H_MATRIX WORLD::GetPos (PART3D *Part1, PART3D *Part2)
{
  if (Part1 == NULL)
    if (Part2 == NULL)
      return Matrix_1;
    else
      return Part2->GetWTP ();
  elsif (Part2 == NULL)
    return Part1->GetWTP ();
  else
    return GEOMatrixProd (GEOMatrixInv(Part1->GetWTP()), Part2->GetWTP());
}


H_MATRIX WORLD::GetPos (char *DevName, char *PartName)
{
  PART     *Part;
  PART3D   *Ptr;
  DEVICE3D *Dev;

  PRO_METHOD ("WORLD", "")

  if (DevName == NULL)
    return Matrix_1;

  if (PartName == NULL)
  {
    Dev = WORLD_GetDev (DevList, DevName);
    Ptr = Dev->GetPartList ();
    Part = Ptr->GetPart ();
    return Part->WTP;
  }
  else
  {
    Ptr = WORLD_GetPart (DevList, DevName, PartName);
    Part = Ptr->GetPart ();
    return Part->WTP;
  }

  PRO_END
  return Matrix_1;
}


static PART3D *WORLD_GetPartID (PART3D *PartList, int ID)
{
  PART3D   *Part, *Chain;
  PART3D   *Ptr;
  DEVICE3D *Attach;

  Part = PartList;
  while (Part != NULL)
  {
    if (Part->GetID() == ID)
      return Part;

    if ((Chain = Part->GetChain()) != NULL &&
        (Ptr = WORLD_GetPartID(Chain, ID)) != NULL)
      return Ptr;

    if ((Attach = Part->GetAttach()) != NULL &&
	(Ptr = WORLD_GetPartID(Attach, ID)) != NULL)
      return Ptr;

    Part = Part->GetNext ();
  }
  return NULL;
}

static PART3D *WORLD_GetPartID (DEVICE3D *DevList, int ID)
{
  DEVICE3D *Dev;
  PART3D   *Ptr = NULL;

  Dev = DevList;
  while (Dev != NULL)
  {
    if ((Ptr = WORLD_GetPartID (Dev->GetPartList(), ID)) != NULL)
      return Ptr;
    Dev = Dev->GetNext ();
  }

  return NULL;
}


static PART3D *WORLD_GetPart (DEVICE3D *DevList, 
			      char *DevName, char *PartName)
{
  bool      found;
  PART     *Part;
  PART3D   *Ptr, *Next;
  PART3D   *Result = NULL;
  DEVICE3D *Dev, *Attach;

  if ((Dev = DevList) == NULL)
    return NULL;

  while (Dev != NULL)
  {
    found = (DevName == NULL || strcmp(Dev->GetName(), DevName) == 0);

    Ptr = Dev->GetPartList ();
    while (Ptr != NULL)
    {
      Next = Ptr;
      while (Next != NULL)
      {
	if (found)
	{
	  if ((Part = Next->GetPart()) != NULL && 
	      strcmp(Part->PartName, PartName) == 0)
	    return Next;
	}
	elsif ((Attach = Next->GetAttach()) != NULL && 
	       (Result = WORLD_GetPart(Attach, DevName, PartName)) != NULL)
	  return Result;

	Next = Next->GetNext ();
      }

      Ptr = Ptr->GetChain ();
    }
    if (found)
      return NULL;

    Dev = Dev->GetNext ();
  }
  return NULL;
}


static bool WORLD_GetPartPos (PART3D *PartList, PART3D *Part, H_MATRIX &WTP)
{
  H_MATRIX  M = WTP;
  PART3D   *Ptr, *Next;
  DEVICE3D *Attach;

  Next = PartList;
  while (Next != NULL && Next != Part)
  {
    if ((Ptr = Next->GetChain()) != NULL &&
	WORLD_GetPartPos(Ptr, Part, M))
    {
      WTP = GEOMatrixProd (Next->GetRel(), M);
      return TRUE;
    }

    if ((Attach = Next->GetAttach()) != NULL &&
	WORLD_GetPartPos(Attach, Part, M))
    {
      WTP = GEOMatrixProd (Next->GetRel(), M);
      return TRUE;
    }
    Next = Next->GetNext ();
  }
  if (Next == Part)
  {
    WTP = Next->GetRel();
    return TRUE;
  }

  return FALSE;
}

static bool WORLD_GetPartPos (DEVICE3D *DevList, PART3D *Part, H_MATRIX &WTP)
{
  H_MATRIX  M = WTP;
  DEVICE3D *Ptr;

  Ptr = DevList;
  while (Ptr != NULL)
  {
    if (WORLD_GetPartPos(Ptr->GetPartList(), Part, M))
    {
      WTP = GEOMatrixProd (WTP, M);
      return TRUE;
    }
    Ptr = Ptr->GetNext ();
  }
  return FALSE;
}


bool WORLD::AddPart (char *Name, PART *Part)
{
  bool      Result;
  PART3D   *NewPart;
  DEVICE3D *Dev;

  PRO_METHOD ("WORLD", "")

  if (Part == NULL)
    return FALSE;

  /* create a new device for the part */
  NewPart = new PART3D   (Part, this);
  Dev     = new DEVICE3D (Name, NewPart, this);
  Result  = AddDevice (Dev);

  PRO_END
  return Result;
}


void WORLD::HighLight (PART3D *Part, bool highlight /* = FALSE */)
{
  PRO_METHOD ("WORLD", "")

  if (((Part->GetState() & STATE_SELECTED) != 0) == highlight)
    PRO_ABORT ("Already highlighted")

  if (highlight)
  {
    Part->SetState (Part->GetState() | STATE_SELECTED);
    Part->Create (PartHighLight);
  }
  else
  {
    Part->SetState (Part->GetState() & (~STATE_SELECTED));
    Part->Create ();
  }

  PRO_END
}


bool WORLD::MergeParts (PART3D *Part1, PART3D *Part2, int Mask /* = 0 */)
{
  bool      Result;
  PART     *Part;
  PART3D   *Chain, *Next;
  DEVICE3D *Attach;

  PRO_METHOD ("WORLD", "")

  Part = PARTMergeParts (Part1->GetPart(), Part2->GetPart(), FALSE);

  if (Part == NULL)
    PRO_RETURN ("Merge failed", FALSE)

  if ((Chain = Part2->GetChain()) != NULL)
    if (Mask & CHAIN_MASK)
      while (Chain != NULL)
      {
	PARTMergeParts (Part, Chain->GetPart(), FALSE);
	Chain = Chain->GetChain ();
      }
    else  // move chain-parts from part2 to part1 (update transformation)
    {
      Part1->AddChain (Chain);
      // TODO: remove chain from part2
    }

  if ((Attach = Part2->GetAttach()) != NULL)
    if (Mask & ATTACH_MASK)
    {
    }
    else // Devices attached to 2nd part will be attached to the first one.
    {
      Part1->AttachDev (Attach);
      // TODO: Detach device from part2
    }
  
  if ((Next = Part2->GetNext()) != NULL)
    if (Mask & NEXT_MASK)               // Merge next parts if desired
    {
      Next = Part2->GetNext ();
      while (Next != NULL)
      {
	PARTMergeParts (Part, Next->GetPart(), FALSE);
	Next = Next->GetNext ();
      }
    }
    else  // Following parts will be added to next-list of first part
    {
      Part1->AddNext (Next);
      // TODO: remove next from part2
    }

  // Update Part-counter of 1st part's device

  Part1->SetMaterial (MATERIAL(COLOR::OBJECT));
  Part1->SetPart     (Part);
  Part1->Create      ();

  /* TODO:
     Part2->GetPart() ohne Objektliste lschen  
  */

  Part2->SetPart (NULL);
#if 0
  Result = Delete (Part2);
#else
  Result = TRUE;
#endif

  PRO_END
  return Result;
}


bool WORLD::Smash (PART3D *Part)
{
  PART     *PartList;
  PART     *Ptr, *Next; 
  PART3D   *NewPart;
  DEVICE3D *Dev;

  PRO_METHOD ("WORLD", "\n")

  if (Part == NULL)
    PRO_RETURN ("No part to smash", TRUE)

  // Smash part into subparts
  PartList = PARTSmashPart (Part->GetPart());

  // First subpart will replace the whole part
  if ((Ptr = PartList) != NULL)
  {
    Next = Ptr->Next;
    Ptr->Next = NULL;
    Part->SetPart (Ptr);
    Ptr = Next;
  }

  Dev = Part->GetParent ();          // Fetch the parts' device 

  while (Ptr != NULL)
  {
    NewPart = new PART3D (Ptr, this);
    NewPart->SetMatrix (Part->GetWTP());
    Dev->AddPart (NewPart);
    Ptr = Ptr->Next;
  }

  PRO_END
  return TRUE;
}


bool WORLD::Smash (DEVICE3D *Dev)
{
  int       count = 1;
  char      DevName[255];
  PART3D   *Ptr, *Next;
  DEVICE3D *NewDev;

  PRO_METHOD ("WORLD", "\n")

  if (Dev == NULL)
    PRO_RETURN ("No device to smash", TRUE)

  if ((Ptr = Dev->GetPartList()) == NULL)
    PRO_RETURN ("Empty device", FALSE)

  Next = Ptr->GetNext ();
  Ptr->SetNext (NULL);
  Ptr->Create  ();
  Dev->SetPartNo (1);

  Ptr = Next;
  while (Ptr != NULL)
  {
    sprintf (DevName, "%s_%d", Dev->GetName(), count);
    Next = Ptr->GetNext ();
    Ptr->SetNext (NULL);
    Ptr->Create ();
    NewDev = new DEVICE3D (DevName, Ptr, this);
    AddDevice (NewDev);

    Ptr = Next;
    count++;
  }

  PRO_END
  return TRUE;
}


PART3D *WORLD::GetPart (char *DevName, char *PartName)
{
  DEVICE3D *Dev;

#if 1
  if ((Dev = GetDev(DevName)) != NULL)
    return Dev->GetPart (PartName);

  return NULL;
#else
  return WORLD_GetPart (DevList, DevName, PartName);
#endif
}


PART3D *WORLD::GetPart (int px, int py, 
			double *x /* = NULL */, double *y /* = NULL */, 
			double *z /* = NULL */)
{
  int       ID;
  double    mx, my, mz;
  H_MATRIX  WTM = Matrix_1, WTP;
  VIEW     *View;
  PART3D   *Part = NULL;

  PRO_METHOD ("WORLD", "\n")

  if (PartNum <= 0)
    PRO_RETURN ("No part to select", NULL)

  View = Win->GetView (Win->GetViewNo(px, py));
  if ((ID = View->SelNo(this, px, py, 0.0, &mx, &my, &mz)) < 0)
    PRO_RETURN ("Nothing selected", NULL)

  if ((Part = WORLD_GetPartID(DevList, ID)) != NULL)         // Locate part
    PRO2 ("Part: '%s:%s'\n", Part->GetParent()->GetName(), Part->GetName())

  if (x != NULL)
    *x = mx;
  if (y != NULL)
    *y = my;
  if (z != NULL)
    *z = mz;

  WTP = Part->GetWTP ();
  PRO_HMATRIX (WTP, "WORLD_T_PART")

#if 0
  // Compute Mouse-Coordiantes relative to part 
  WTM.m[P][X] = mx;
  WTM.m[P][Y] = my;
  WTM.m[P][Z] = mz;
  PRO3 ("WTM: %.3f / %.3f / %.3f\n", mx, my, mz)   // global position

  PRO0 ("DRAW SELECTED POINT\n")
  PART3D *Mouse = new SPHERE ("mouse", 3.0, 180.0, 360.0, 8, 
			      MATERIAL(COLOR::YELLOW));
  DEVICE3D *Dev = new DEVICE3D ("mouse", Mouse, this);
  Dev->SetPos (GEOMakeTrans(mx, my, mz));
  AddDevice (Dev);  
#endif

  PRO_END
  return Part;
}


int WORLD::SearchParts (int x1, int y1, int x2, int y2, PART3D ***PartList)
{
  int     count, Hits;
  int     px, py;
  int     SizeX, SizeY;
  int    *NoList = NULL;
  VIEW   *View;
  PART3D *Part = NULL;

  PRO_METHOD ("WORLD", "\n")

  if (PartNum <= 0)
    PRO_RETURN ("No part to select", 0)

  px = int (0.5 * (x1 + x2));
  py = int (0.5 * (y1 + y2));
  SizeX = abs (x2-x1)+1;
  SizeY = abs (y2-y1)+1;

  View = Win->GetView (Win->GetViewNo(px, py));
  if ((Hits = View->SelectRegion(this, px, py, 0.0, &NoList,
				 SizeX, SizeY)) == 0)
    PRO_RETURN ("Nothing selected", 0)

  *PartList = new PART3D *[Hits];
  for (count = 0; count < Hits; ++count)
    if ((Part = WORLD_GetPartID(DevList, NoList[count])) != NULL) 
    {
      printf ("Part: '%s:%s'\n", 
	      Part->GetParent()->GetName(), Part->GetName());
      (*PartList)[count] = Part;
    }
    else
      (*PartList)[count] = NULL;

  delete NoList;

  PRO_END
  return Hits;
}


void WORLD::Select (PART3D *Part, bool highlight /* = TRUE */)
{
  PRO_METHOD ("WORLD", "\n")

  if (Part == NULL)
    PRO_ABORT ("Noting selected")

  HighLight (Part, highlight);
  Redraw ();

  PRO_END
}


PART3D *WORLD::PickPart (double *x /* = NULL */, double *y /* = NULL */, 
			 double *z /* = NULL */)
{
  int     Button, px, py;
  PART3D *Part = NULL;

  do
    Button = Win->GetMouse (AppCon, px, py);
  while (Button != 1 && Button != 3);

  if (Button == 1)
  {
    PRO0 ("Select Part...\n")
    Part = GetPart (px, py, x, y, z);
  }

  return Part;
}


static DEVICE3D *WORLD_GetBase (DEVICE3D *DevList, int ID)
{
  PART3D   *Part, *Next;
  DEVICE3D *Dev, *Base, *Attach;

  if ((Dev = DevList) == NULL)
    return NULL;

  while (Dev != NULL)
  {
    Part = Dev->GetPartList ();
    while (Part != NULL)
    {
      if (Part->GetID() == ID)
	return Dev;

      Next = Part->GetNext ();
      while (Next != NULL)
      {
	if (Next->GetID() == ID)
	  return Dev;
	Next = Next->GetNext ();
      }

      // Do not return the base device on which the searched id is attached
      // because if we want to detach such a device we need it themself
      if ((Attach = Part->GetAttach()) != NULL && 
	  (Base = WORLD_GetBase(Attach, ID)) != NULL)
	return Base;

      Part = Part->GetChain ();
    }
    Dev = Dev->GetNext ();
  }
  return NULL;
}


static bool WORLD_GetDevPos (PART3D *PartList, DEVICE3D *Dev, H_MATRIX &WTD)
{
  H_MATRIX  M;
  PART3D   *Ptr, *Next;
  DEVICE3D *Attach;

  Next = PartList;
  while (Next != NULL)
  {
    if ((Ptr = Next->GetChain()) != NULL &&
	WORLD_GetDevPos (Ptr, Dev, M))
    {
      WTD = GEOMatrixProd (Next->GetRel(), M);
      return TRUE;
    }

    if ((Attach = Next->GetAttach()) != NULL &&
	WORLD_GetDevPos (Attach, Dev, M))
    {
      WTD = GEOMatrixProd (Next->GetRel(), M);
      return TRUE;
    }
    Next = Next->GetNext ();
  }
  return FALSE;
}

static bool WORLD_GetDevPos (DEVICE3D *DevList, DEVICE3D *Dev, H_MATRIX &WTD)
{
  H_MATRIX  M;
  DEVICE3D *Ptr;

  Ptr = DevList;
  while (Ptr != NULL && Ptr != Dev)
  {
    if (WORLD_GetDevPos(Ptr->GetPartList(), Dev, M))
    {
      WTD = GEOMatrixProd (WTD, M);
      return TRUE;
    }

    Ptr = Ptr->GetNext();
  }
  if (Ptr == Dev)
  {
    WTD =  Ptr->GetPartList()->GetRel ();
    return TRUE;
  }

  return FALSE;
}


int WORLD::GetPartNum (int Mask /* = ALL_MASK */) const
{
  if (DevList == NULL)
    return 0;
  elsif (Mask == ALL_MASK)
    return PartNum;
  else
    return DevList->GetPartNumber (Mask);
}


void WORLD::IncPartNum (int Number /* = 1 */)
{
  PartNum += Number;
}


int WORLD::GetDevNum (int Mask /* = NEXT_MASK */) const
{
  if (DevList == NULL)
    return 0;
  else
    return DevList->GetDevNumber (Mask);
}



bool WORLD::AddDevice (DEVICE3D *Dev)
{
  PRO_METHOD ("WORLD", "\n")

  if (Dev == NULL)
    PRO_RETURN ("No Device defined", TRUE)

  if (DevList == NULL)                  // No other devices up to now
  {
    DevList = Dev;
    Dev->SetFix    (NULL);
    Dev->UpdatePos ();
  }
  else
    DevList->AddDevice (Dev);           // Add device to list

  while (Dev != NULL)                   // Update counters
  {
    (DevNum)++;
    PartNum += Dev->GetPartNo ();

    Dev = Dev->GetNext ();
  }

  PRO_END
  return TRUE;
}


void WORLD::HighLight (DEVICE3D *Dev, bool highlight /* = FALSE */)
{
  PRO_METHOD ("WORLD", "")

  if (((Dev->GetState() & STATE_SELECTED) != 0) == highlight)
    PRO_ABORT ("Already highlighted")

  if (highlight)
  {
    Dev->SetState (Dev->GetState() | STATE_SELECTED);
    Dev->Create (DevHighLight);
  }
  else
  {
    Dev->SetState (Dev->GetState() & (~ STATE_SELECTED));
    Dev->Create ();
  }

  PRO_END
}


bool WORLD::Detach (DEVICE3D *Dev)
{
  PART3D *Part;

  PRO_METHOD ("WORLD", "")

  if (Dev == NULL || (Part = Dev->GetFix()) == NULL)
    PRO_RETURN ("Device is not attached", TRUE)

  if (! Remove(Dev))
    PRO_RETURN ("Device cannot be removed", FALSE)

  AddDevice (Dev);

  PRO_END
  return TRUE;
}

  
static DEVICE3D *WORLD_GetDev (DEVICE3D *DevList, PART3D *Part)
{
  PART3D   *PartPtr;
  DEVICE3D *Ptr, *Dev;

  if ((Dev = DevList) == NULL)
    return NULL;

  while (Dev != NULL)
  {
    PartPtr = Dev->GetPartList ();
    while (PartPtr != NULL)
    {
      if (PartPtr == Part)
	return Dev;

      if ((Ptr = PartPtr->GetAttach()) != NULL)
	Ptr = WORLD_GetDev (Ptr, Part);
      if (Ptr != NULL)
	return Ptr;

      PartPtr = PartPtr->GetChain ();
    }
    Dev = Dev->GetNext ();
  }

  return Dev;
}

static DEVICE3D *WORLD_GetDev (DEVICE3D *DevList, char *DevName)
{
  PART3D   *Part;
  DEVICE3D *Ptr, *Dev;

  if (DevName == NULL || (Dev = DevList) == NULL)
    return NULL;

  while (Dev != NULL && strcmp(Dev->GetName(), DevName) != 0)
  {
    Part = Dev->GetPartList ();
    while (Part != NULL)
    {
      if ((Ptr = Part->GetAttach()) != NULL)
	Ptr = WORLD_GetDev (Ptr, DevName);
      if (Ptr != NULL)
        return Ptr;
      Part = Part->GetChain ();
    }
    Dev = Dev->GetNext ();
  }

  return Dev;
}


DEVICE3D *WORLD::GetDev (char *DevName)
{
  return WORLD_GetDev (DevList, DevName);
}


#if 0
DEVICE3D *WORLD::GetDev (int No /* = 0 */)
{
  if (No == 0)
    if (SelDev.Number() > 0)
      return (DEVICE3D *)SelDev.Get (0);
    else
      return NULL;
  else
    return NULL;
}
#endif


DEVICE3D *WORLD::GetDev (PART3D *Part)
{
  return WORLD_GetDev (DevList, Part);
}


DEVICE3D *WORLD::GetDev (int px, int py,         double *x /* = NULL */, 
			 double *y /* = NULL */, double *z /* = NULL */)
{
  int   ID;
  VIEW *View;

  if (PartNum <= 0)
    PRO_RETURN ("No part to select", NULL)

  View = Win->GetView (Win->GetViewNo(px, py));
  if ((ID = View->SelNo(this, px, py, 0.0, x, y, z)) < 0)
    return NULL;

  return WORLD_GetBase (DevList, ID);   // Locate device
}


void WORLD::Select (DEVICE3D *Dev, bool highlight /* = TRUE  */)
{
  PRO_METHOD ("WORLD", "\n")

  if (Dev == NULL)
    PRO_ABORT ("No Device")

  HighLight (Dev, highlight);
  Redraw ();

  PRO_END
}


DEVICE3D *WORLD::PickDev ()
{
  int       Button, px, py;
  double    x, y, z;
  DEVICE3D *Dev = NULL;

  do
  {
    Button = Win->GetMouse (AppCon, px, py);
  } while (Button != 1 && Button != 3);

  if (Button == 1)
  {
    PRO0 ("Select Device...\n")
    Dev = GetDev (px, py, &x, &y, &z);
  }

  return Dev;
}


bool WORLD::ResizeHandle (int SizeX, int SizeY)
{
#ifdef irix
  XResizeWindow (Win->GetDisplay(), Win->GetWindow(), SizeX, SizeY);
#endif

  if (UserResizeHandle != NULL)
    UserResizeHandle (this, SizeX, SizeY);

  return FALSE;
}


bool WORLD::MouseHandle (int Button, int MouseX, int MouseY)
{
  if (UserMouseHandle != NULL)
    UserMouseHandle (this, Button, MouseX, MouseY);

  return FALSE;
}


bool WORLD::ButtonHandle (int Button, int MouseX, int MouseY)
{
  if (UserButtonHandle != NULL)
    UserButtonHandle (this, Button, MouseX, MouseY);
 
  ViewNo = Win->GetViewNo (MouseX, MouseY);
  return TRUE;
}


bool WORLD::SelectHandle (int Button, int MouseX, int MouseY)
{
  if (UserSelectHandle != NULL)
    UserSelectHandle (this, Button, MouseX, MouseY);

  ViewNo = Win->GetViewNo (MouseX, MouseY);
  return FALSE;
}


bool WORLD::ClickHandle (int Button, int MouseX, int MouseY)
{
  if (UserClickHandle != NULL)
    UserClickHandle (this, Button, MouseX, MouseY);

  ViewNo = Win->GetViewNo (MouseX, MouseY);
  return FALSE;
}


bool WORLD::KeyHandle (KeySym Key, int Mask)
{
  if (UserKeyHandle != NULL)
    UserKeyHandle (this, Key, Mask);
  
  return FALSE;
}


bool WORLD::DestroyHandle ()
{
  if (UserDestroyHandle != NULL)
    UserDestroyHandle (this);

  return (UserDestroyHandle != NULL);
}


Widget WORLD::GetShell ()
{ 
  return Win->GetShell();
}


Display *WORLD::GetDisplay () 
{ 
  return Win->GetDisplay(); 
}

