#ifndef _LIBCAD3D_H_
#define _LIBCAD3D_H_

#include <limits.h>

#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glx.h>

#include <libpart.h>
#include <libtools.h>

#include <libsim3d.h>


#define CS_ID  10001

#define ROLL  5
#define PITCH 4
#define YAW   3

#ifdef RED
  #undef RED
#endif
#define RED   0

#ifdef GREEN
  #undef GREEN
#endif
#define GREEN 1

#ifdef BLUE
  #undef BLUE
#endif
#define BLUE  2

#ifdef ALPHA
  #undef ALPHA
#endif
#define ALPHA 3

#define POS_X 0
#define POS_Y 1
#define POS_Z 2
#define ROT_X 3
#define ROT_Y 4
#define ROT_Z 5

#define MIN_X 0
#define MAX_X 1
#define MIN_Y 2
#define MAX_Y 3
#define MIN_Z 4
#define MAX_Z 5 

#define STATE_NORMAL     (0<<0)
#define STATE_SELECTED   (1<<0)
#define STATE_COLLISION  (1<<1)
#define STATE_OOW        (1<<2)

#define SINGLE_MASK      (0<<0)
#define NEXT_MASK        (1<<0)
#define CHAIN_MASK       (1<<1)
#define ATTACH_MASK      (1<<2)
#define ALL_MASK         (0xFF)


#define MOD(Val,Min,Max,Mod) \
          ((Val) > Max)? ((Val)-Mod) : ((Val<Min)? ((Val)+Mod):(Val))
#define TO_FRAME(Matrix) \
          (FRAME(Matrix.m[N][X], Matrix.m[N][Y], Matrix.m[N][Z], \
		 Matrix.m[O][X], Matrix.m[O][Y], Matrix.m[O][Z], \
		 Matrix.m[A][X], Matrix.m[A][Y], Matrix.m[A][Z], \
		 Matrix.m[P][X], Matrix.m[P][Y], Matrix.m[P][Z]))
#define TO_MATRIX(Frame) \
          (GEOCreateMatrix(Frame[N][X], Frame[N][Y], Frame[N][Z], \
		           Frame[O][X], Frame[O][Y], Frame[O][Z], \
		           Frame[A][X], Frame[A][Y], Frame[A][Z], \
		           Frame[P][X], Frame[P][Y], Frame[P][Z]))


// Forward declarations:
class OPENGL;
class FLOOR;
class SHADOW;
class PART3D;
class DEVICE3D;
class WORLD;


// Function prototypes
typedef void (*MOUSE_HANDLE)   (WORLD *World, 
			        int Button, int MouseX, int MouseY);
typedef void (*BUTTON_HANDLE)  (WORLD *World, 
			        int Button, int MouseX, int MouseY);
typedef void (*SELECT_HANDLE)  (WORLD *World,
			        int Button, int MouseX, int MouseY);
typedef void (*CLICK_HANDLE)   (WORLD *World,
				int Button, int MouseX, int MouseY);
typedef void (*RESIZE_HANDLE)  (WORLD *World, int SizeX, int SizeY);
typedef void (*KEY_HANDLE)     (WORLD *World, KeySym Key, int Modifier);
typedef void (*DESTROY_HANDLE) (WORLD *World);

typedef void (*TRANSFORM_FUNC) (POINT3D *Point, VECTOR3D *Normal);
typedef void (*ACTION_FUNC)    (DEVICE3D *Dev, PART3D *Part, void *Ptr);


extern    char    *ColorNames[];        // Array of color names
extern    char    *MaterialNames[];     // Array of matrerial names
extern    GLfloat  ColorTable[][4];
extern    OPENGL  *OpenGL;

GLfloat  *CalcColor     (int State, GLfloat *Color);
int       Indent        (FILE *fp = NULL, int Indent = 0);


namespace EXCEPTIONS
{
  class BREAK
  {
    public :
      // Constructor/Destructor:
             BREAK (int Reason, const char *Level, const char *Msg = NULL);
            ~BREAK ();
    protected :
      // Methods:
      char  *GetMsg ();

    // Attributes:
    public :
      int   No;

    protected :
      char *Message;
  };

  class ERROR : public BREAK
  {
    public :
      enum REASON
      {
	RANGE_ERROR = -1,
	FILE_ERROR  = -2,
	NO_DATA     = -3
      };

      // Constructor:
            ERROR  (int Reason, const char *Msg = NULL); 

      // Methods:
      char *GetErrorMsg  ();
  };

  class WARNING : public BREAK
  {
    public :
      // Constructor:
            WARNING (int Reason, const char *Msg = NULL);

      // Methods:
      char *GetWarningMsg  ();
  };

  class HINT : public BREAK
  {
    public :
      // Constructor:
            HINT (int Reason, const char *Msg = NULL);

      // Methods:
      char *GetHintMsg  ();
  };
};  


/* ********************************************************************** */
/* **                              O P E N G L                         ** */
/* ********************************************************************** */
class OPENGL
{
  public :
                  OPENGL  ();
                 ~OPENGL  ();

    void          Initialize    ();

    bool          Write          (FILE *fp = NULL, int Indent = 0) const;
    int           GetID          (int ID = -1, const char *Name = NULL);
    void          FreeID         (int ID);
    int           GetMaxLights   () const;
    GLuint        GetFontBase    ();
    GLXContext    GetContext     () const;
    XVisualInfo  *GetVisualInfo  ();
    GLuint        MakeFont       (const char *FontSpec /* = "fixed" */, ...);

  private :
    // Attributs:
    XVisualInfo  *VisualInfo;
    GLXContext    Context;
    char          Vendor[40];
    int           Version[2];
    int           Extension[2];
    int           StencilSize;
    int           TextureSize;
    int           MaxLights;                   // max amount of lights;
    GLuint        FontBase;
    // Stencil-Buffer, Farbtiefe, Textur-Gre/Buffer
};


/* ********************************************************************** */
/* **                          D A T A F I L E                         ** */
/* ********************************************************************** */
class DATAFILE
{
  public:
                   DATAFILE   (const char *FileName, char Comment = '#');
                  ~DATAFILE   ();

    int            ReadLine   ();
    int            Parse      (const char *Fmt, ...);

    char           GetChar    (bool Skip = FALSE);
    int            GetInt     (bool Skip);
    unsigned long  GetHex     (bool Skip);
    double         GetDouble  (bool Skip);
    char          *GetString  (char *Result, int MaxLen, char StopChar, 
			       bool EOL);
    char          *GetLine    (char *Result, int MaxLen);


  // TODO: ReadInt, ReadDouble, ReadString, ReadChar, 
  //       Write, Indent, ...

  //       Write, Indent, ...
  //       Write, Indent, ...

  private:
    // Methods:
    bool   IsSpace     (char Letter);
    char   SkipSpace   (bool EOL);
    bool   CheckData   ();

   // Attributes:
   char  *Buffer;                        // Buffer for the data
   int    LineNo;                        // current line number
   char  *Ptr;                           // current position in the data
};


/* ********************************************************************** */
/* **                           C A D _ F I L E                        ** */
/* ********************************************************************** */
class CAD_FILE
{
  public :
    #define INDENT 2

    // Constructor:
                  CAD_FILE  ();

    // Methods:
    virtual bool  Read      (DATAFILE &File,      double Version);
    virtual bool  Write     (FILE     *fp = NULL, int    Indent = 0) const;

  protected :
    // Methods:
            void   Space    (FILE *fp = NULL, int Indent = 0) const;
};


/* ********************************************************************** */
/* **                                U N I T                           ** */
/* ********************************************************************** */
class UNIT
{
  public :
    // Constructor / Destructor:
                   UNIT    (int Id = -1);
    virtual       ~UNIT    ();

    // Methods:
    virtual bool   Create  ();
            void   Set     () const;

            void   SetID   (const char *Name = NULL);
            int    GetID   () const;

    // Attributes:
    int     Id;                         // OpenGL display list ID
};


/* ********************************************************************** */
/* **                           J O I N T                              ** */
/* ********************************************************************** */
class JOINT : CAD_FILE
{
  public :
    // Constructor:
                     JOINT       (int Axes);

    // Methods:
    virtual bool     Read        (DATAFILE &File, double Version);
    virtual bool     Write       (FILE     *fp = NULL, int Indent = 0) const;

           void      Scale       (double ScaleX, double ScaleY, double ScaleZ);

           void      SetAxes     (int    Axes);
           bool      SetValue    (int    Axis, double Value);
           void      SetRange    (int    Axis, double Min, double Max);
           void      SetSpeed    (double Speed);
           void      SetAccel    (double Accel);
           void      SetTime     (double Time);

           int       GetAxes     () const;
           double    GetValue    (int Axis) const;
           double    GetMin      (int Axis) const;
           double    GetMax      (int Axis) const;
           double    GetSpeed    () const;
           double    GetAccel    () const;
           double    GetTime     () const;

           bool      IsFree      (int Axis) const;
           bool      IsValid     (int Axis, double Value) const;
           double    GetValue    (int Axis, double Value) const;
           H_MATRIX  GetFrame    () const;

  private :
    // Attributes:
    int     Axes;                       // Variable axes
    RANGE   Range[6];                   // Ranges for all axes
    RANGE   Speed, Accel, Time;         // Mobility parameter
};


/* ********************************************************************** */
/* **                              C O L O R                           ** */
/* ********************************************************************** */
class COLOR : public CAD_FILE, public UNIT
{
  public :
    enum ID
    {
      OBJECT = -1,                      // Use object colors
      RGB,                              // Use RGB-values (0)
      BLACK,
      GREY50,
      SILK,
      GREY,
      SILVER,                           // ( 5)
      WHITE,
      DARKRED,
      MEDRED,
      LIGHTRED,
      DARKORANGE,                       // (10)
      ORANGE,
      GOLD,
      YELLOW,
      LEMONCHIFFON,
      DARKGREEN,                        // (15)
      LIME,
      MEDGREEN,
      SEAGREEN,
      TURQUOISE,
      STEELBLUE,                        // (20)
      DARKBLUE,
      MEDBLUE,
      DODGERBLUE,
      LIGHTBLUE,
      CYAN,                             // (25)
      PURPLE,
      ORCHID,
      PINK,
      MAGENTA,
      CHERRY,                           // (30)
      BROWN,
      SIENNA,
      PERU,
      TAN,
      PEACHPUFF,                        // (35)
      NONE,                                   // Separator
      TEXTURE                                 // Texture
    };

    static int COLORS;                        // Amount of colors

    // Constructors:
                      COLOR       (ID ColorID = OBJECT);
                      COLOR       (double Red, double Green, double Blue, 
				   double Alpha = 1.0);

    // Operators:
            COLOR    &operator =  (const COLOR &Source);
            double   &operator [] (const int &i);
            double    operator [] (const int &i) const;

    // Methods:
    virtual bool      Read        (DATAFILE &File, double Version);
    virtual bool      Write       (FILE *fp = NULL, int Indent = 0) const;

            COLOR     Complement  () const;

    virtual bool      Create      ();
            void      Set         () const;
            void      SetID       (ID ColorID = OBJECT);
            void      SetValues   (double Red, double Green, double Blue,
				   double Alpha = 1.0);
            void      SetRed      (double Red);
            void      SetGreen    (double Green);
            void      SetBlue     (double Blue);
            void      SetAlpha    (double Alpha);

            double    GetRed      () const;
            double    GetGreen    () const;
            double    GetBlue     () const;
            double    GetAlpha    () const;
            ID        GetID       () const;

            char     *GetName     ();

  private :
    // Attributes:
    ID      ColorID;
    double  Color[4];
};


/* ********************************************************************** */
/* **                       L I N E _ S T Y L E                        ** */
/* ********************************************************************** */
class LINE_STYLE : public CAD_FILE, public UNIT
{
  public :
    // Constructor:
    LINE_STYLE (COLOR Color = COLOR(COLOR::BLACK), 
		int Width = 1, int Pattern = 0xFFFF);

    // Methods:
    virtual bool    Read             (DATAFILE &File, double Version);
    virtual bool    Write            (FILE *fp = NULL, int Indent = 0) const;

    virtual bool    Create           ();
            void    Set              () const;

    void       SetColor         (COLOR Color = COLOR(COLOR::BLACK));
    void       SetWidth         (int Width = 1);
    void       SetPattern       (int Patter = 0xFFFF);
    void       SetAntialiasing  (bool Mode);

    COLOR     &GetColor         ();
    int        GetWidth         () const;
    int        GetPattern       () const;
    bool       GetAntiAliasing  () const;

  private :
    // Attributes:
    COLOR Color;
    int   Width;
    int   Pattern;
    bool  Antialiasing;
};


/* ********************************************************************** */
/* **                           T E X T U R E                          ** */
/* ********************************************************************** */
class TEXTURE : public CAD_FILE, public UNIT
{
  public :
    enum MAPPING
    {
      CLAMP = 0,
      REPEAT
    };

    enum FILTER
    {
      NEAREST = 0,
      LINEAR
    };

    enum MODE
    {
      MODULATE = 0,
      REPLACE,
      DECAL,
      BLEND
    };

    #define NO_TEXTURE -1

    // TODO: wrapping methods: planar, spherical, cylindrical, auto

  public :
    // Constructors/destructor:
              TEXTURE          (const char *FileName);
              TEXTURE          (int TextureID);
	      TEXTURE          (XImage *Image);
	      TEXTURE          (TEXTURE *Texture);
             ~TEXTURE          ();

    // Operators:
    TEXTURE    &operator =     (const TEXTURE &Source);

    // Methods:
    bool       Read            (DATAFILE &File, double Version);
    bool       Write           (FILE *fp = NULL, int Indent = 0) const;

    void       Set             (const COLOR &BlendColor = 
				COLOR(COLOR::BLACK)) const;
    virtual bool       Create          ();

    bool       SetTextureFile  (const char *FileName);
    void       SetTextureID    (int TextureID);
    void       SetMapping      (MAPPING Mapping);
    void       SetFilter       (FILTER  Filter);
    void       SetMode         (MODE    Mode);
    void       SetStretch      (double s = 1.0, double v = 1.0);

    char      *GetFileName     ();
    int        GetTextureID    () const;
    MAPPING    GetMapping      () const; 
    FILTER     GetFilter       () const;
    MODE       GetMode         () const;  
    double     GetStretchS     () const;
    double     GetStretchT     () const;
    XImage    *GetImage        ();

  private :
    // Methods:
    void       Initialize      ();
    int        CalcSize        (int Size);
 
    // Attributes:
    int       TextureID;
    MAPPING   Mapping;
    FILTER    Filter;
    MODE      Mode;
    double    StretchS, StretchT;
    char      TextureFile[PATH_MAX+1];
    byte     *Data;
    XImage   *TextureImg;
}; 


/* ********************************************************************** */
/* **                          M A T E R I A L                         ** */
/* ********************************************************************** */
class MATERIAL : public CAD_FILE, public UNIT
{
  public :
    enum MODE
    {
      NONE = 0,
      WIRE,
      LINE,
      SOLID
    };

    // SHADE: NO/FLAT/SMOOTH ?

    enum ID
    {
      VALUES = 0,
      WOOD,
      PLASTIC,
      METAL,
      STONE,
      GLASS,
      PAPER,
      CLOTH,
      WATER,
      MIRROR
    };

    // Constructors/destructor
                     MATERIAL         (COLOR::ID ColorID = COLOR::OBJECT, 
				       int Mode = SOLID);
		     MATERIAL         (COLOR Color);
		     MATERIAL         (ID MaterialID);
		     MATERIAL         (const MATERIAL &Material);
		     MATERIAL         (const char *TextureFile);
		    ~MATERIAL         ();

    // Methods:
    virtual bool       Read           (DATAFILE &File, double Version);
    virtual bool       Write          (FILE     *fp = NULL, 
				       int       Indent = 0) const;

            void       Set            () const;

	    void       SetColorID     (COLOR::ID ColorID = COLOR::OBJECT);
	    void       SetColor       (COLOR     Color = COLOR(COLOR::OBJECT));
	    void       SetColor       (double r, double g, double b, 
				       double Alpha = 1.0);
	    void       SetAlpha       (double Alpha      = 1.0);
	    void       SetMode        (int    Mode       = SOLID);
	    void       SetLine        (int    Width      = 0);
	    void       SetMaterialID  (ID     MaterialID = VALUES);
	    void       SetTextureID   (int    TextureID  = NO_TEXTURE);
	    void       SetAmbient     (double Ambient    = 1.0);
	    void       SetDiffuse     (double Diffuse    = 0.0);
	    void       SetSpecular    (double Specular   = 1.0);
	    void       SetEmission    (double Emission   = 0.0);
	    void       SetShininess   (double Shininess  = 0.0);
	    void       SetTexture     (TEXTURE *NewTexture = NULL);

	    bool       LoadTexture    (const char *FileName);

	    COLOR::ID  GetColorID     () const;
	    COLOR     &GetColor       ();
	    double     GetRed         () const;
	    double     GetGreen       () const;
	    double     GetBlue        () const;
	    double     GetAlpha       () const;
	    int        GetMode        () const;
	    int        GetLine        () const;
	    ID         GetMaterialID  () const;
	    double     GetAmbient     () const;
	    double     GetDiffuse     () const;
	    double     GetSpecular    () const;
	    double     GetEmission    () const;
	    double     GetShininess   () const;
	    TEXTURE   *GetTexture     ();

	    bool       HasTexture     () const;

  protected :
    // Methods:
            void       CleanTexture   ();

  private :
    // Methods:
            void       Initialize     ();
    virtual bool       Create         ();

    // Attributes:
    ID        MaterialID;                  // predefined material ID
    COLOR     Color;
    int       Render;                      // POINT, WIRE, HIDDEN, FILL
    int       Mode;                        // FLAT, SMOOTH    
    int       Line;                        // Line width

    double    Ambient;                     // simplified color components
    double    Diffuse;
    double    Specular;
    double    Emission; 
    double    Shininess;   

    TEXTURE  *Texture;                     // Texture data
};


/* ********************************************************************** */
/* **                            L I G H T                             ** */
/* ********************************************************************** */
class LIGHT : public CAD_FILE, public UNIT
{
  public :
    enum MODE
    {
      OFF = 0,
      DIRECTIONAL,
      POSITIONAL,
      SPOTLIGHT
    };

    enum NUMBER
    {            
      LIGHT1 = 0,
      LIGHT2,
      LIGHT3,
      LIGHT4,
      LIGHT5,
      LIGHT6,
      LIGHT7,
      LIGHT8
    };

    enum PLACEMENT
    {
      STATIONARY = 0,
      RELATIVE
    };

    // Constructor/Destructor:
            LIGHT            (      NUMBER Number);
            LIGHT            (      NUMBER Number, 
			            MODE   Mode,
		              const COLOR &Ambient  = COLOR(COLOR::BLACK),
		              const COLOR &Diffuse  = COLOR(COLOR::WHITE),
	                      const COLOR &Specular = COLOR(COLOR::WHITE));
           ~LIGHT            ();
    
    // Methods:
    virtual bool       Read             (DATAFILE &File, double Version);
    virtual bool       Write            (FILE *fp = NULL, 
					 int Indent = 0) const;

    void       Draw             ();

    void       InitShadow        ();

    void       SetMode          (MODE   Mode = DIRECTIONAL);
    void       SetIntensity     (double Value = 1.0);
    void       SetPosition      (POINT3D  Position);
    void       SetDirection     (VECTOR3D Direction = Vector_0);
    void       SetPlacement     (PLACEMENT Place);
    void       SetCutOff        (double Angle = 90.0);
    void       SetSpotExponent  (double Exponent = 1.0);
    void       SetAmbient       (const COLOR &Color = COLOR(COLOR::BLACK));
    void       SetDiffuse       (const COLOR &Color = COLOR(COLOR::WHITE));
    void       SetSpecular      (const COLOR &Color = COLOR(COLOR::WHITE));
    void       SetAttenuation   (double Quadratic = 0, 
				 double Linear    = 0, 
				 double Constant  = 1);
    void       SetSize          (double Size = 0.0);
    void       SetShadow        (SHADOW *Shadow);

    NUMBER     GetNumber        () const;
    MODE       GetMode          () const;
    double     GetIntensity     () const;
    POINT3D    GetPosition      () const;
    VECTOR3D   GetDirection     () const;
    PLACEMENT  GetPlacement     () const;
    double     GetCutOff        () const;
    double     GetSpotExponent  () const;
    COLOR      GetAmbient       () const;
    COLOR      GetDiffuse       () const;
    COLOR      GetSpecular      () const;
    double    *GetAttenuation   ();
    double     GetSize          () const;
    SHADOW    *GetShadow        ();

  private :
    // Methods:
            void      Initialize       ();
    virtual bool      Create           ();
            void      Build            ();
            void      CalcWTL          ();

    // Attributes:
    NUMBER     LightNo;                 // Number of light
    MODE       Mode;                    // Light-mode
    double     Intensity;               // Intensity factor
    COLOR      Ambient;                 // Scattered/bounced light
    COLOR      Diffuse;                 // Light comming from one direction
    COLOR      Specular;                // Shininess
    double     Attenuation[3];          // Decreasing of light intensity
    POINT3D    Pos;                     // Pos of spotlight
    VECTOR3D   Dir;                     // direction of spotlights
    PLACEMENT  Placement;               // Placement of spotlights
    double     CutOff;                  // Cuttoff angle (50 %)
    double     SpotExp;                 // Attenuation of spotlights
    double     Size;                    // Draw light source as headlight
    PART3D    *HeadLight;               // Part for visualization
    H_MATRIX   WTL;                     // World-To-Light frame
    SHADOW    *Shadow;                  // Shadow pointer
};


/* ********************************************************************** */
/* **                            S H A D O W                           ** */
/* ********************************************************************** */
class SHADOW : public CAD_FILE, public UNIT
{
  public :
    enum MODE
    {
      OFF = 0,
      ON
    };

    // Constructor:
              SHADOW          (LIGHT *Light, FLOOR *Floor = NULL,
			       MODE Mode = OFF,
			       COLOR Color = COLOR(0.0, 0.0, 0.0, 0.5));

    // Methods:
    virtual bool      Read            (DATAFILE &File, double Version);
    virtual bool      Write           (FILE *fp = NULL, int Indent = 0) const;
    virtual bool      Create          ();

    void      SetMode         (MODE    Mode);
    void      SetFloor        (FLOOR  *Floor);
    void      SetLight        (LIGHT  *Light);
    void      SetOffset       (double  Offset);
    void      SetColor        (const COLOR &Color = COLOR(COLOR::BLACK));

    MODE      GetMode         () const;
    H_MATRIX  GetMatrix       () const;
    double    GetOffset       () const;
    COLOR    &GetColor        ();

  private :
    // Methods:
    void      CalcPlane       (double Plane[4]);

    // Attributes:
    MODE      Mode;                     // shadow mode
    H_MATRIX  ShadowMatrix;             // projection matrix for shadow
    double    Offset;                   // shadow offset
    COLOR     Color;                    // color of shadow
    FLOOR    *Floor;                    // pointer to projection area
    LIGHT    *Light;                    // pointer to light source
};


/* ********************************************************************** */
/* **                                 F O G                            ** */
/* ********************************************************************** */
class FOG : public CAD_FILE, public UNIT
{
  public:
    enum MODE 
    {
      OFF = 0,
      LINEAR,
      EXP,
      EXP2
    };

    enum HINT
    {
      DONT_CARE = 0,
      FASTEST,
      NICEST
    };

    // Constructor:
            FOG         (MODE Mode = OFF, double Density = 0.5);

    // Methods:
    virtual bool    Read        (DATAFILE &File, double Version);
    virtual bool    Write       (FILE *fp = NULL, int Indent = 0) const;

    void    Set         () const;

    void    SetMode     (MODE Mode = OFF);
    void    SetHint     (HINT Hint = FASTEST);
    void    SetColor    (COLOR Color = COLOR(COLOR::WHITE));
    void    SetDensity  (double Density);
    void    SetRange    (double Start = 0.0, double End = 1.0);

    MODE    GetMode     () const;
    HINT    GetHint     () const;
    COLOR  &GetColor    ();
    double  GetDensity  () const;
    double  GetStart    () const;
    double  GetEnd      () const;
 
  private :
    // Attributes:
    MODE    Mode;
    HINT    Hint;
    COLOR   Color;
    double  Density;
    double  Start, End;
};


/* ********************************************************************** */
/* **                       P R O J E C T I O N                        ** */
/* ********************************************************************** */
class PROJECTION : public CAD_FILE
{
  public :
            PROJECTION        ();
            PROJECTION        (PROJECTION *Projection);

    virtual bool    Read      (DATAFILE &File, double Version);
    virtual bool    Write     (FILE *fp = NULL, int Indent = 0) const;  

            double *GetRange  ();

    virtual void    Set       (double PosX  = -1.0, double PosY  = -1.0,
			       double SizeX =  1.0, double SizeY =  1.0);
    virtual void    SetRange  (double *Range);
            void    SetRange  (double RelX1,       double RelX2,
			       double RelY1 = 0.0, double RelY2 = 1.0,
			       double RelZ1 = 0.0, double RelZ2 = 1.0);

    virtual void    Pan       (double DirX, double DirY);
    virtual void    GetCOI    (double *x, double *y, double *z);
  
  protected :
    double  Range[6];         // Projection parameter
};

class ORTHOGRAPHIC : public PROJECTION   // Orthographic projection
{
  public :
            ORTHOGRAPHIC    ();
            ORTHOGRAPHIC    (ORTHOGRAPHIC *Projection);
    
    virtual void  Set       (double PosX  = -1.0, double PosY  = -1.0,
			     double SizeX =  1.0, double SizeY =  1.0);
    virtual void  SetRange  (double *Range = NULL);
    virtual void  GetCOI    (double *x, double *y, double *z);
};

class PERSPECTIVE : public PROJECTION     // Perspective projection
{ 
  public :
            PERSPECTIVE     ();
            PERSPECTIVE     (PERSPECTIVE *Projection);
    
    virtual void  Set       (double PosX  = -1.0, double PosY  = -1.0,
			     double SizeX =  1.0, double SizeY =  1.0);
    virtual void  SetRange  (double *Range = NULL);
    virtual void  GetCOI    (double *x, double *y, double *z);
};


/* ********************************************************************** */
/* **                            C A M E R A                           ** */
/* ********************************************************************** */
class CAMERA : public CAD_FILE,  public UNIT
{
  public :
                   CAMERA       (RPY Frame = RPY_0);
                  ~CAMERA       ();

    // operator [] ?

    virtual bool   Read         (DATAFILE &File, double Version);
    virtual bool   Write        (FILE *fp = NULL , int Indent = 0) const;

            void   Place        (RPY Frame = RPY_0);
            void   Rotate       (double rx, double ry, double rz);

            RPY    GetLocation  () const;
  
  private :
    virtual bool   Create       ();

    RPY  Location;                      // Camera position / orientation
};


/* ********************************************************************** */
/* **                             V I E W                              ** */
/* ********************************************************************** */
class VIEW : public CAD_FILE
{
  public :
    enum MODE
    {
      ORTHO = 0,
      PERSPECT
    };

  public :
                VIEW            (int SizeX, int SizeY, RPY Camera = RPY_0, 
			         double *Clipping = NULL, MODE Mode = ORTHO,
			         double *Range = NULL);
                VIEW            (VIEW *View);
    virtual    ~VIEW            ();

    virtual bool        Read        (DATAFILE &File, double Version);
    virtual bool        Write       (FILE *fp = NULL, int Indent = 0) const;

    void        Set             ();
    void        SetBorder       (COLOR Color = COLOR(COLOR::WHITE));
    void        DrawBorder      ();

    void        Rotate          (double rx, double ry, double rz);
    void        Pan             (double DirX, double DirY);
    void        Zoom            (int x1, int y1, int x2, int y2);
    // TODO: Translate, Fly, ...

    void        SetWinSize      (int SizeX, int SizeY);
    void        SetClipping     (double *Clipping);
    void        SetRange        (double *Range);
    void        SetScale        (double Scale = 1.0);
    void        SetCamera       (RPY  Frame);
    void        SetPrjMode      (MODE Mode = ORTHO);

    void        GetSize         (int *SizeX, int *SizeY) const;
    double      GetScale        () const;
    CAMERA     *GetCamera       ();
    double     *GetClipping     ();
    double     *GetRange        ();
    COLOR      &GetBorder       ();
    bool        GetCoordinates  (int MouseX, int MouseY, 
				 int *px, int *py) const;
    MODE        GetPrjMode      ();
    void        GetCOI          (double *x, double *y, double *z);

    bool        GetPos          (double  px, double  py, double  pz, 
				 double *x,  double *y,  double *z);
    int         SelNo           (WORLD *World,
				 int    px,  int     py, double  pz, 
				 double *x,  double *y,  double *z);
    int         SelectRegion    (WORLD *World,
				 int    px, int     py, double  pz, 
				 int **NoList, int SizeX = 1, int SizeY = 1);

    bool        IsInView        (int MouseX, int MouseY) const;

    void        ShowObject      (int ID, H_MATRIX Matrix);

  protected :
    void        CalcViewPort    ();
    void        SetModelView    ();

  private :
    // Methods:
    bool        IsFullScreen    ();
    bool        SelectPos       (double  px, double  py, double  pz, 
				 double *x,  double *y,  double *z);

    // Attributes:
    double      Clipping[4];            // clipping area of whole window
    int         ViewPort[4];            // viewport at current size
    double      Scale;                  // scaling-factor (neccessary ???)
    COLOR       Border;                 // border-color
    // TODO: Clipping Planes

    int         PrjNo;                  // Index of Projection
    PROJECTION *Projection[2];          // ortho & perspective projection
    CAMERA      Camera;                 // Camera location

    int         WinX, WinY;             // Size of window
};


/* ********************************************************************** */
/* **                           W I N D O W                            ** */
/* ********************************************************************** */
class WINDOW
{
  public :
    // Constructor / Destructor:
               WINDOW   (Display *Displ, int Screen, Widget Application,
			 const char *Title, double Width, double Height);
               WINDOW   (Display *Displ, int Screen, Widget Parent);
    virtual   ~WINDOW   ();

    // Methods:
    int      GetMouse        (XtAppContext AppCon, int &MouseX, int &MouseY);
    int      GetSizeX        () const;
    int      GetSizeY        () const;

    void     SetPos          (double PosX  =   0.0, double PosY  =   0.0);
    void     SetSize         (double SizeX = 250.0, double SizeY = 250.0);

    void     SetMouseCursor  (int Shape = ToolNORMAL);
    void     SetTitle        (const char *Title, ...);

    Display *GetDisplay      () const;
    Widget   GetShell        () const; 
    Window   GetWindow       () const;

    COLOR   &GetBackground   ();
    COLOR   &GetForeground   ();

    // Virtual Methods:
    virtual void  SetBackground  (const COLOR &Color = COLOR(COLOR::BLACK));
    virtual void  SetForeground  (const COLOR &Color = COLOR(COLOR::WHITE));

    virtual void  EventHandler   (Widget widget, XEvent *Event);

  protected :
    // Attributes:
    COLOR    bg, fg;
    Display *XDisplay;
    Window   Win;
    int      XScreen;
    Widget   Shell;
    int      PosX,   PosY;
    int      SizeX,  SizeY;

  private :
    // Attributes:
    char     Title[256];
};


/* ********************************************************************** */
/* **                          W I N D O W 3 D                         ** */
/* ********************************************************************** */
class WINDOW3D : public WINDOW
{
  public :
    enum SIDE
    {
      BACK_SIDE  = (1<<0),
      FRONT_SIDE = (1<<1)
    };

    enum CULLING
    {
      OFF = 0,
      FRONT,
      BACK,
      BOTH
    };

    enum POLYGON_MODE
    {
      POINTS = 0,
      WIRE_FRAME,
      HIDDEN_LINE,
      FACES
    };

    enum BLENDING
    {
      SOLID,
      TRANSPARENT
    };

    enum SHADING
    {
      NO_SHADE,
      FLAT,
      SMOOTH
    };

    enum ROTARY
    {
      COUNTERCLOCKWISE = 0,
      CLOCKWISE
    };

    enum SELECT_MODE
    {
      UNLOCK = -1,
      CHECK,
      LOCK
    };

    enum ALIASING
    {
      ANTIALIASING_OFF = 0,
      ANTIALIASING_FASTEST,
      ANTIALIASING_NICEST
    };

                  WINDOW3D        (WORLD *World,
				   Display *Displ,     int Screen, 
				   Widget Application, const char *Name, 
				   double Width = 250, double Height = 250);
	          WINDOW3D        (WORLD *World,
				   Display *Displ, int Screen, Widget Parent);
    virtual      ~WINDOW3D        ();

    bool          Select          (SELECT_MODE lock = CHECK);
    void          Close           ();
    void          Flush           ();

    POLYGON_MODE  GetPolygonMode  (SIDE Side = FRONT_SIDE) const;
    SHADING       GetShading      () const;
    BLENDING      GetBlending     () const;
    CULLING       GetCulling      () const;
    ROTARY        GetRotary       () const;
    ALIASING      GetAliasing     () const;
    int           GetViewNo       () const;
    int           GetViewNo       (int px, int py) const;
    VIEW         *GetView         (int No);
    FOG          *GetFog          () const;

    void          SetPolygonMode  (POLYGON_MODE RenderMode, 
				   int Side = (int)FRONT_SIDE);
    void          SetShading      (SHADING   Shading);
    void          SetBlending     (BLENDING  Blending);
    void          SetCulling      (CULLING   Cull);
    void          SetRotary       (ROTARY    Rotary);
    void          SetAliasing     (ALIASING  Aliasing);
    void          SetFog          (FOG      *Fog);
    void          SetViews        (int       Mask);
    void          SetProjection   (int Mode);

    void          Print           (double x, double y, double z, 
				   const char *Msg);

    // Virtual Methods:
    virtual void  SetBackground   (const COLOR &Color = COLOR(COLOR::BLACK));

  protected :
    // Methods:
    virtual void  EventHandler    (Widget widget, XEvent *Event);

    // Attributes:
    VIEW   *Views[4];

  private :
    typedef struct
    {
      CULLING      Culling;          // off / front / back / both
      POLYGON_MODE Render[2];        // back|front: point/wire/hidden-line/face
      SHADING      Shade;            // no/flat/smooth
      BLENDING     Blend;            // solid/transparent
      ROTARY       Rotary;           // clockwise/counterclockwise
      ALIASING     Aliasing;         // Antialiasing mode
    } GL_PARAMETERS;

    // Methods:
    void   Initialize     ();
    bool   InitGL         ();

    // Attributes:
    GL_PARAMETERS   Data;
    WORLD          *World;
    FOG            *Fog;
    Window          WinGL;
};


/* ********************************************************************** */
/* **                               B O X                              ** */
/* ********************************************************************** */
class BOX : public CAD_FILE, public UNIT
{
  public :
    // Constructor/Destructor;
                      BOX        (double SizeX, double SizeY, double SizeZ,
				  H_MATRIX Center = Matrix_1);
                      BOX        (double MinX, double MaxX,
				  double MinY, double MaxY, 
				  double MinZ, double MaxZ);
                     ~BOX        ();

    // Operators:
            BOX      &operator  = (const BOX &Box);

    // Methods:
    virtual bool      Read        (DATAFILE &File, double Version);
    virtual bool      Write       (FILE *fp = NULL, int Indent = 0) const;

    virtual bool      Create      ();

            void      Place       (H_MATRIX Center);

            void      SetRange    (double MinX, double MaxX, 
				   double MinY, double MaxY, 
				   double MinZ, double MaxZ);
            void      SetSize     (double SizeX, double SizeY, double SizeZ);
            void      SetLen      (double Length = 0.2); 
            void      SetColor    (COLOR  Color = COLOR(COLOR::GOLD));

            POINT3D   GetSize     () const;
            double    GetLen      () const;
            void      GetRange    (double *MinX, double *MaxX,
				   double *MinY, double *MaxY,
				   double *MinZ, double *MaxZ);
            COLOR    &GetColor    ();

  private :
    // Attributes:
    double   Len;                       // Length of lines (0..1)
    H_MATRIX PTB;                       // Part_to_Box frame
    POINT3D  Size;                      // size of bounding box
    COLOR    Color;                     // Color of box
};


/* ********************************************************************** */
/* **                              A X E S                             ** */
/* ********************************************************************** */
class AXES : public CAD_FILE, public UNIT
{
  public :
    // Constructor/Destructor:
                   AXES         (double Size, int FontBase = -1);
                   AXES         (double Size, Display *Displ, 
				 const char *FontSpec /* = "fixed" */, ...);
                  ~AXES         ();

    // Methods:
    virtual bool   Read         (DATAFILE &File, double Version);
    virtual bool   Write        (FILE *fp = NULL, int Indent = 0) const;

            void   Draw         () const;
            void   Draw         (H_MATRIX Matrix) const;
            void   SetSize      (double   Size);
    // SetFont...

  private :
    // Methods:
    virtual bool   Create       ();
            void   Print        (double x, double y, double z, 
				 const char *Msg);

    // Attributes:
    double Size;
    int    FontBase;
}; 


/* ********************************************************************** */
/* **                            F L O O R                             ** */
/* ********************************************************************** */
class FLOOR : public CAD_FILE, public UNIT
{
  public :
    enum MODE
    {
      OFF = 0,
      ON
    };

    // Constructor / Destructor:
                      FLOOR          (MODE Mode = ON, int LineWidth = 1);
                     ~FLOOR          ();

    // Operators:
            double   &operator []    (const int &i);
            double    operator []    (const int &i) const;

    // Methods:
    virtual bool      Read           (DATAFILE &File, double Version);
    virtual bool      Write          (FILE *fp = NULL, int Indent = 0) const;
    virtual bool      Create         ();

            void      Draw           ();

            void      SetMode        (MODE   NewMode = ON);
            void      SetSize        (int    Width, int Height);
            void      SetPos         (double px, double py, double pz);
            void      SetRot         (double rx, double ry, double rz);
            void      SetSpacing     (double DistX, double DistY);
            void      SetMaterial    (const  MATERIAL &NewMaterial);
            void      SetLineWidth   (int    Width = 1);
            void      SetGridColor   (const  COLOR &Color);

            int       GetMode        () const;
            int       GetSizeX       () const;
            int       GetSizeY       () const;
            double    GetSpaceX      () const;
            double    GetSpaceY      () const;
            MATERIAL *GetMaterial    ();
            int       GetLineWidth   () const;
            COLOR    &GetGridColor   ();

  private :
    // Attributes:
    MODE      Mode;                // Grid-mode
    int       SizeX, SizeY;        // Number of cells
    double    SpaceX, SpaceY;      // Cell-size
    RPY       Place;               // Position and rotation of floor
    MATERIAL  Material;            // Material of floor
    int       LineWidth;           // Width of grid-lines
    COLOR     Grid;                // grid color
    int       ID;                  // OpenGL-ID
};


/* ********************************************************************** */
/* **                              P A R T 3 D                         ** */
/* ********************************************************************** */
class PART3D : public CAD_FILE, public UNIT
{
  public:
    enum JOINT_TYPE
    {
      UNKNOWN = -1,
      FIX     = 0,
      ROT     = (1<<0),
      TRANS   = (1<<1)
    };

    enum DRAW_MODE
    {
      NOTHING   = 0,
      DRAW_PART = (1<<0),
      DRAW_BOX  = (1<<1),
      DRAW_AXES = (1<<2)
    };

    // Constructor / destructor:
                PART3D  (      PART     *Part     = NULL,
			       WORLD    *World    = NULL,
			 const MATERIAL &Material = MATERIAL(COLOR::OBJECT));
                PART3D  (PART3D *Part, WORLD *World = NULL);
               ~PART3D  ();

    // CSG-Operations:
    PART3D     *operator    +   (PART3D &Part);   // Union
    PART3D     *operator    -   (PART3D &Part);   // Difference
    PART3D     *operator    %   (PART3D &Part);   // Intersection

    // Methods:
            PART3D     *Copy           (DEVICE3D *Parent);
            void        Draw           ();
            void        Place          (H_MATRIX WTP);
            void        Scale          (double ScaleX, double ScaleY, 
					double ScaleZ, int    Mask = 0);
            void        AddNext        (PART3D   *Part);
            bool        AttachDev      (DEVICE3D *Dev);
    virtual bool        Create         ();
            bool        Create         (const COLOR &Color);

    virtual bool        Read           (DATAFILE &File, double Version);
    virtual bool        Write          (FILE *fp = NULL, int Indent = 0) const;

            char       *GetName        ();
            int         GetPartNumber  (int Mask = 0) const;
            int         GetDevNumber   (int Mask = 0);
            MATERIAL   *GetMaterial    ();
            BOX        *GetBox         ();

            PART3D     *GetPart        (const char *PartName);

            int         GetState       () const;
            int         GetRender      () const;
            int         GetType        () const;
            int         GetMode        () const;
            JOINT      &GetJoint       ();
            double      GetVal         (int No) const;
            double      GetSpeed       () const;
            double      GetAccel       () const;
            double      GetTime        () const;
            H_MATRIX    GetMatrix      () const;
            H_MATRIX    GetRel         () const;
            H_MATRIX    GetWTP         () const;
            PART       *GetPart        ();
            PART3D     *GetPrev        ();
            PART3D     *GetNext        ();
            PART3D     *GetChain       ();
            DEVICE3D   *GetParent      ();
            DEVICE3D   *GetAttach      ();
            double      GetNormalAngle () const;
            void        GetSize        (double *MinX, double *MaxX, 
					double *MinY, double *MaxY,
					double *MinZ, double *MaxZ) const;
	    void        GetRange       (double *MinX, double *MaxX, 
					double *MinY, double *MaxY,
					double *MinZ, double *MaxZ, 
					int Mask = 0);
	    bool        Remove         (PART3D   *RemovePart);
	    bool        Remove         (DEVICE3D *Dev);

	    void        SetName        (const char *Name);
	    void        SetMaterial    (const      MATERIAL &Material);
	    void        SetColorID     (COLOR::ID  ColorID = COLOR::OBJECT);
	    void        SetColor       (COLOR Color = COLOR(COLOR::OBJECT));
	    void        SetBox         (const      BOX &Box);
	    void        SetState       (int        State);
	    void        SetMode        (int        Mode = DRAW_PART);
	    void        SetRender      (int        Render);
	    void        SetTextureID   (int        TextureID = -1);
	    void        SetAlpha       (double     Alpha     = 1.0);
	    void        SetType        (int        Type);
	    void        SetVal         (double    *Values);
	    void        SetVal         (int        No, double Value);
	    void        SetSpeed       (double     Speed);
	    void        SetAccel       (double     Accel);
	    void        SetTime        (double     Time);

	    void        SetPos         (double px, double py, double pz);
	    void        SetRot         (double rx, double ry, double rz);
	    void        SetMatrix      (H_MATRIX    T);
	    void        SetRel         (H_MATRIX    Rel);
	    int         SetJoint       (double     *Value);

	    void        SetPart        (PART       *Part);
	    void        SetPrev        (PART3D     *PrevPart);
	    void        SetNext        (PART3D     *NextPart);
	    void        AddChain       (PART3D     *ChainPart);
	    void        SetParent      (DEVICE3D   *Dev);
	    void        SetAttach      (DEVICE3D   *Dev);

	    void        UpdatePos      ();
	    void        UpdatePos      (H_MATRIX    Frame);
	    void        SetNormalAngle (double      Angle = 30.0);

   virtual  void        SetWorld       (WORLD *World);

  protected :
    // Methods
    void        Initialize      ();

    // Attributes:
    PART       *Part;

  private : 
    // Methods
    void        DrawWireFrame      (      OBJECT   *Obj,    
					  POLYGON  *Polygon,
				    const MATERIAL *Material);
    double      GetTextureFactor   ();

    // Attributes :
    int         Render;                  // weg ???
    int         Type, KinType;
    JOINT       Joint;
    double      Length, Offset;
    int         State;                   // NORMAL, SELECTED, RANGE, COLL
    int         Mode;                    // Drawing mode
    DHP         DHParam;
    MATERIAL    Material;                // Material-information
    H_MATRIX    T;                       // A-Matrix in home-position
    H_MATRIX    Rel;                     // actual relative transformation
    H_MATRIX    WTP;                     // absolute world->part frame

    BOX         Box;                     // Bounding box
    int         CS;                      // Displaylist of coord-system
  // TODO: AXES *CS; (Gre abhngig von Bauteilgre)

    DEVICE3D   *Parent;                  // Device containing this part
    PART3D     *Prev;                    // Preveious part in chain or NULL
    PART3D     *Chain;                   // Next part in chain
    DEVICE3D   *Attach;                  // Attached device
    PART3D     *Next;                    // Next part in the device
    WORLD      *World;                   // Reference to the world
};


#if 0
  // Operatoren:  kopieren =,
  // Objekte zusammenfgen */
  PART3D  *Merge           (PART3D *Part, MATRIX M);
  // Additionsoperator fr alle enthaltenen Objekte
  PART3D  *Add             (PART3D *Part, MATRIX M);
  // Subtraktion this - Part
  PART3D  *Diff            (PART3D *Part, MATRIX M);
  // Schnittmenge zweier Parts
  PART3D  *Cut             (PART3D *Part, MATRIX M);
  // Austausch der Part-Daten
  void     Exchange        (PART *Part);
#endif


class ARROW : public PART3D
{
  public :
    ARROW (WORLD *World,    const char *Name,     H_MATRIX Frame,
	   double Radius1,  double      Len1,     
	   double Radius2,  double      Len2, 
	   double Angle,    int         Facettes, double   Segment, 
	   const  MATERIAL &Material);
};


class BOLID : public PART3D
{
  public :
    BOLID (WORLD  *World,   const char *Name,   H_MATRIX Frame,
	   double  Radius1, double      Radius2,    double   Radius3,
	   double  Height,  int         Facettes1,  int      Facettes2, 
	   double  Segment, const       MATERIAL   &Material);
};


class CONE : public PART3D
{
  public :
    CONE (WORLD  *World,    const char *Name,     H_MATRIX Frame,
	  double  Radius1,  double      Radius2,  double   Radius3,
	  double  SizeZ,    double      Segment,  int      Facettes, 
	  const   MATERIAL &Material);
};


class CYLINDER : public PART3D
{
  public :
    CYLINDER (WORLD  *World,  const char *Name,     H_MATRIX Frame,
	      double  Radius, double      Depth,    double   Segment, 
	      int Facettes,   const       MATERIAL &Material);
};


class GEAR : public PART3D
{
  public :
    GEAR (WORLD  *World,   const char *Name,     H_MATRIX Frame,
	  double  Radius1, double      Radius2,  double   Radius3,
	  double  Depth1,  double      Depth2,   int      Teeth,  
	  double  Skew,    const       MATERIAL &Material);
};


class ICOSAHEDRON : public PART3D
{
  public :
    ICOSAHEDRON (WORLD  *World,    const char *Name,     H_MATRIX Frame,
		 double  Radius1,  double      Radius2, 
		 int     Density,  int         Iterations, 
		 const   MATERIAL &Material);
};


class PIPE : public PART3D
{
  public :
    PIPE (WORLD  *World,      const char *Name,       H_MATRIX Frame,
	  double  Radius1,    double      Radius2, 
	  double  Thickness1, double      Thickness2,
	  double  SizeZ,      double      Segment,
	  int     Facettes,   const       MATERIAL &Material);
};


class PRISM : public PART3D
{
  public :
    PRISM (WORLD  *World,    const char *Name,    H_MATRIX Frame,
	   double  Width1,   double      Width2,  double   OffsX,
	   double  Depth1,   double      Depth2,  double   OffsY,
	   double  Height1,  double      Height2, double   OffsZ,
	   const   MATERIAL &Material);
};


class QUADER : public PART3D
{
  public :
    QUADER (WORLD  *World,    const char *Name,  H_MATRIX Frame,
	    double  SizeX,    double      SizeY, double   SizeZ, 
	    const   MATERIAL &Material);
};


class ROTATIONAL : public PART3D
{
  public :
    ROTATIONAL (WORLD  *World,   const char *Name,    H_MATRIX Frame,
		int     Number,  POINT2D    *Points, 
		double  Segment, int         Facettes, 
		double  Radius,  const       MATERIAL &Material);
};


class SCREW : public PART3D
{
  public :
    enum THREAD
    {
      LEFT_THREAD,
      RIGHT_THREAD
    };

    SCREW (WORLD  *World,     const char *Name,      H_MATRIX Frame,
	   double  Radius1,   double      Radius2,
	   double  Height,    double      Increase,  double   Offset,
	   int     Facettes,  THREAD      Direction,
	   const   MATERIAL  &Material);
};


class SLICE : public PART3D
{
  public :
    SLICE (WORLD  *World,   const char *Name,       H_MATRIX Frame,
	   double  Radius1, double      Radius2,    double   Radius3,
	   double  SizeZ,   int         Facettes1,  int      Facettes2,
	   double  Segment, const       MATERIAL   &Material);
};


class SPHERE : public PART3D
{
  public :
    SPHERE (WORLD  *World,    const char *Name,      H_MATRIX Frame,
	    double  Radius,   double      SegmentX,  double   SegmentZ,
	    int     Facettes, const       MATERIAL  &Material);
};


class SPRING : public PART3D
{
  public :
    enum THREAD
    {
      LEFT_THREAD,
      RIGHT_THREAD
    };

    SPRING (WORLD  *World,      const char *Name,      H_MATRIX Frame,
	    double  Radius1,    double      Radius2,   double   Radius3,
	    double  Height,     double      Increase,
	    int     Facettes1,  int         Facettes2, THREAD   Direction,
	    const   MATERIAL   &Material);
};


class SWEEP : public PART3D
{
  public :
    SWEEP (WORLD *World,    const char *Name,   H_MATRIX Frame,
	   int    Number,   POINT2D    *Points, double   Height, 
	   const  MATERIAL &Material);
};


class TORUS : public PART3D
{
  public :
    TORUS (WORLD  *World,      const char *Name,     H_MATRIX Frame,
	   double  Radius1,   double      Radius2,  
	   int     Facettes1, int         Facettes2, 
	   double  Segment,   const       MATERIAL &Material);
};


class WEDGE : public PART3D
{
  public :
    WEDGE (WORLD  *World,    const char *Name,    H_MATRIX Frame,
	   double  Width1,   double      Width2, 
	   double  Height1,  double      Height2, 
	   double  Depth1,   double      Depth2, 
	   const   MATERIAL &Material);
};


/* CSG: Union, Difference, Intersection */
class INTERSECTION : public PART3D
{
  public :
    INTERSECTION (      WORLD    *World,   const char   *Name,  H_MATRIX Frame,
		        PART3D   *Part1,         PART3D *Part2, 
		  const MATERIAL &Material);
};


class DIFFERENCE : public PART3D
{
  public :
    DIFFERENCE (      WORLD    *World,    const char   *Name,  H_MATRIX Frame, 
		      PART3D   *Part1,          PART3D *Part2, 
		const MATERIAL &Material);
};


class UNION : public PART3D
{
  public :
    UNION (      WORLD    *World,    const char   *Name,  H_MATRIX Frame, 
	         PART3D   *Part1,          PART3D *Part2, 
	   const MATERIAL &Material);
};


/* ********************************************************************** */
/* **                         D E V I C E 3 D                          ** */
/* ********************************************************************** */
class DEVICE3D : public CAD_FILE, public UNIT
{
  public:
     // Constructor/Destructor:
                         DEVICE3D      (const char   *DevName = NULL, 
					PART3D *Part    = NULL,
					WORLD  *World   = NULL);
			 DEVICE3D      (DEVICE3D *Dev, WORLD *World = NULL);
     virtual            ~DEVICE3D      ();

     // Methods:
             DEVICE3D   *Copy          ();
     virtual bool        Create        ();
             bool        Create        (COLOR Color);

             void        Place         (H_MATRIX WTD);

             bool        Write         (FILE *fp = NULL, int Indent = 0);
	     bool        Save          (const char *FileName);

             int         GetPartNumber (int Mask = 0);
             int         GetDevNumber  (int Mask = 0);
             char       *GetName       ();
             int         GetConfigNo   ();
             int         GetMountNo    ();
             H_MATRIX    GetOffset     ();
             double      GetAlpha      ();
             int         GetHomeNo     ();
             H_MATRIX    GetWTD        ();
             H_MATRIX    GetTCP        ();
             int         GetDOF        ();
             int         GetState      ();
             int         GetPartNo     ();
             ROBOT      *GetRobot      ();
             PART3D     *GetPartList   ();
             PART3D     *GetFix        ();
             DEVICE3D   *GetNext       ();

             void        SetName       (char       *Name);
             void        SetPos        (H_MATRIX    M);
             void        SetConfigNo   (int         ConfigNo);
             void        SetMountNo    (int         MountNo);
             void        SetOffset     (H_MATRIX    Offset);
             void        SetHomeNo     (int         HomeNo);
             void        SetWTD        (H_MATRIX    WTD);
             void        SetDOF        (int         DOF);
             void        SetState      (int         State);
             void        SetPartNo     (int         PartNo);
             void        SetPart       (PART3D     *Part);
             void        SetFix        (PART3D     *Fix);
             void        SetNext       (DEVICE3D   *Dev);
             void        SetAlpha      (double      Alpha);
             void        SetRobot      (ROBOT      *Robot);
             void        IncPartNo     (int         Value);

             PART3D     *GetPart       (const char *PartName);
             void        AddDevice     (DEVICE3D   *Dev);
             void        AddPart       (PART3D     *Part, PART3D *Prev = NULL);
             bool        DetachPart    (PART3D     *Part);
             bool        Remove        (PART3D     *Part);
             bool        Remove        (DEVICE3D   *Dev);
             PART3D     *GetMount      ();

             void        UpdatePos     ();
             void        UpdatePos     (H_MATRIX WTD);

             void        GetRange      (double *MinX, double *MaxX, 
					double *MinY, double *MaxY,
					double *MinZ, double *MaxZ);
             void        Scale         (double ScaleX, double ScaleY, 
					double ScaleZ, int Mask = 0);
             void        SetColorID    (COLOR::ID ColorID = COLOR::OBJECT);
             void        SetColor      (COLOR Color = COLOR(COLOR::OBJECT));
             int         GetRange      (int No, 
					double &Min, double &Val, double &Max);
             int         SetJoint      (double *Joint, int No, 
					bool attached = FALSE);
             void        SetAllJoints  (double Value = 0.0); 
             int         GetJoints     (double **Joints);

             bool        MoveTCP       (H_MATRIX Rel);

             void        CalcDHP       ();

  private:
    typedef struct HOME_DATA
    {
      char   Name[20];
      double Pos[6];
    } HOME_DATA;

    int         No;
    char       *Name;
    char       *PrgName;
    int         ConfigNo;
    int         KinType;                  // Move_By_Links/kin_generic
    int         MountNo;
    H_MATRIX    Offset;
    int         HomeNo;
    HOME_DATA  *Home;
    double      Alpha;
    ROBOT      *Robot;                    // Kinematic data

    H_MATRIX    WTD;
    int         DOF;
    int         State;
    int         PartNo;
    PART3D     *PartList;
    PART3D     *Fix;                    // Part at which the device is attached
    DEVICE3D   *Next;                   // Next device of the world
    WORLD      *World;                  // Reference to the world
};


/*
  void   AddHome       (char *Name, double *ValueList);
  void   MoveTCP       (MATRIX M, Speed=, Steps=, execute=true);
  void   TransTCP      (double dx, double dy, double dz, int Steps=);
  void   RotTCP        (double rx, double ry, double rz, int Steps=);
  void   Exchange      (DEVICE3D *Dev);
  void   XML           (FILE *fp);
*/


/* ********************************************************************** */
/* **                             W O R L D                            ** */
/* ********************************************************************** */
class WORLD : public CAD_FILE
{
  public :
    // Constructor/Destructor:
                   WORLD                 (const char *Name = NULL, 
					  int SizeX = 250, int SizeY = 250);
                   WORLD                 (Widget Parent, 
					  const char *Name /* = NULL */);
    virtual       ~WORLD                 ();

    // Methods:
    virtual void   Draw                  (VIEW *View, GLenum Mode = GL_RENDER);
    virtual void   Redraw                (int ViewNo = -1);

    void           Clear                 ();
    void           Print                 ();

    DEVICE3D      *GetData               ();
    void           SetData               (DEVICE3D *DevList);

    virtual void   GetSize               (double *MinX, double *MaxX,
					  double *MinY, double *MaxY,
					  double *MinZ, double *MaxZ);

    LIGHT         *GetLight              (LIGHT::NUMBER Number);
    bool           SetLight              (LIGHT::NUMBER Number, 
					  LIGHT *Light = NULL);

    FLOOR        *GetFloor               () const;
    void          SetFloor               (FLOOR *Floor);

    DEVICE3D      *LoadDevice            (const char *FileName,
					  const char *DataPath = NULL);
    DEVICE3D      *LoadPart              (const char *FileName);
    virtual bool   LoadWorld             (const char *FileName,
					  const char *DataPath = NULL);

    virtual bool   Write                 (FILE *fp = NULL) const;
    bool           Save                  (const char *FileName);  // own format

    bool           Select                ();
    bool           Get3DPos              (int px, int py, double pz = 0.0,
					  double *x = NULL, double *y = NULL,
					  double *z = NULL);
    H_MATRIX       GetPos                (DEVICE3D  *Dev);
    H_MATRIX       GetPos                (PART3D    *Part);
    H_MATRIX       GetPos                (DEVICE3D  *Dev1,  DEVICE3D *Dev2);
    H_MATRIX       GetPos                (DEVICE3D  *Dev,   PART3D   *Part);
    H_MATRIX       GetPos                (PART3D    *Part1, PART3D   *Part2);
    H_MATRIX       GetPos                (char *DevName, char *PartName);

    // *** PARTS:
    int            GetPartNum            (int Mask = ALL_MASK) const;
    void           IncPartNum            (int Number = 1);
    bool           AddPart               (char *Name, PART *Part);
    void           HighLight             (PART3D *Part, 
					  bool highlight = FALSE);
    bool           MergeParts            (PART3D *Part1, PART3D *Part2,
					  int Mask = 0);
    bool           Smash                 (PART3D *Part);

    PART3D        *GetPart               (int px, int py, 
					  double *x = NULL, double *y = NULL, 
					  double *z = NULL);
    int            SearchParts           (int x1, int y1, int x2, int y2,
					  PART3D ***PartList);
    PART3D        *GetPart               (char *DevName, char *PartName);

    void           Select                (PART3D *Part, bool highlight = TRUE);
    PART3D        *PickPart              (double *x = NULL, double *y = NULL, 
					  double *z = NULL);

    // *** DEVICES:
    int            GetDevNum             (int Mask = NEXT_MASK) const;
    bool           AddDevice             (DEVICE3D *Dev);
    void           HighLight             (DEVICE3D *Dev, 
					  bool highlight = FALSE);
    //  bool           AttachDevice          (DEVICE3D *Dev, PART3D *Part);
    bool           Detach                (DEVICE3D *Dev);
    bool           Smash                 (DEVICE3D *Dev);
    DEVICE3D      *GetDev                (int px, int py, 
					  double *x = NULL, double *y = NULL, 
					  double *z = NULL);
    DEVICE3D      *GetDev                (char *DevName);
#if 0
    DEVICE3D      *GetDev                (int No = 0);
#endif
    DEVICE3D      *GetDev                (PART3D *Part);
    void           Select                (DEVICE3D *Dev, 
					  bool highlight = TRUE);
    DEVICE3D      *PickDev               ();
    DEVICE3D      *FindNearest           (H_MATRIX TCP);

    DEVICE3D      *CreateText            (const char *FontName, int FontSize, 
					  double Depth, double Accuracy, 
					  const MATERIAL &Material,
					  TRANSFORM_FUNC Transform, 
					  bool relative, 
					  const char *Text, ...);
    double         GetTextWidth          (char *FontName, int FontSize, 
				          char *Text, ...);

    char          *GetName               ();
    void           GetWinSize            (int *SizeX, int *SizeY);
    WINDOW3D      *GetWin                ();
    VIEW          *GetView               (int ViewNo = -1);
    double         GetScale              ();
    RPY            GetCamera             ();
    double        *GetRange              ();
    void           GetCOI                (double *x, double *y, double *z);

    COLOR          GetPartHighLight      ();
    COLOR          GetDevHighLight       ();

    void           SetName               (const char *Name, ...);
    void           SetScale              (double Scale);
    void           Fill                  (double Factor = 1.0);
    void           SetRange              (double *Range);
    void           SetCamera             (RPY Position);
    void           SetViews              (int Mask);
    void           RotateView            (double rx, double ry, double rz);
    void           PanView               (double DirX, double DirY);
    void           ZoomView              (int x1, int y1, int x2, int y2);

    void           IncAlpha              (int Value = 1);
    void           DecAlpha              (int Value = 1);

    void           SetMouseCursor        (int Shape);

    void           SetPartHighLight      (COLOR Color);
    void           SetDevHighLight       (COLOR Color);

    void           SetUserResizeHandle   (RESIZE_HANDLE ResizeHandle);
    void           SetUserMouseHandle    (MOUSE_HANDLE  MouseHandle);
    void           SetUserButtonHandle   (BUTTON_HANDLE ButtonHandle);
    void           SetUserSelectHandle   (SELECT_HANDLE SelectHandle);
    void           SetUserClickHandle    (CLICK_HANDLE  ClickHandle);
    void           SetUserKeyHandle      (KEY_HANDLE    KeyHandle);
    void           SetUserDestroyHandle  (DESTROY_HANDLE DestroyHandle);

    bool           ReadLine              (FILE *fp, bool show = FALSE, 
					  char *Result = NULL, int Len = 0);

    bool           Remove                (DEVICE3D *Dev);
    bool           Delete                (DEVICE3D *Dev);
    bool           Remove                (PART3D *Part);
    bool           Delete                (PART3D *Part);

    void           CalcDHP               ();

    void           UserAction            (ACTION_FUNC Func, void *DataPtr);

    virtual bool   ResizeHandle          (int SizeX,  int SizeY);
    virtual bool   MouseHandle           (int Button, int MouseX, int MouseY);
    virtual bool   ButtonHandle          (int Button, int MouseX, int MouseY);
    virtual bool   SelectHandle          (int Button, int MouseX, int MouseY);
    virtual bool   ClickHandle           (int Button, int MouseX, int MouseY);
    virtual bool   KeyHandle             (KeySym Key, int Mask);
    virtual bool   DestroyHandle         ();

    Widget        GetShell               ();
    Display      *GetDisplay             ();

  protected :
    // Methods:
    void           DrawShadow            ();
    DEVICE3D      *ReadDev               (FILE *fp, int &ListNo, int Version,
					  const char *DataPath);

    // Attributes:
    WINDOW3D      *Win;             
    DEVICE3D      *DevList;
    int            PartNum;
    int            DevNum;
    int            AlphaNum;
    int            ViewNo;
    FLOOR         *Floor;    

  private :
    // Methods:
            bool   LoadDXF              (const char *FileName);
    virtual bool   Load                 (const char *FileName);

    // Attributes:
    char           *Name;                // World-Name
    LIGHT          *Lights[8];           // Lights 
    double          Range[6];            // world expansion

    COLOR           PartHighLight;       // Part highlight color
    COLOR           DevHighLight;        // Device highlight color

    RESIZE_HANDLE   UserResizeHandle;
    MOUSE_HANDLE    UserMouseHandle;
    BUTTON_HANDLE   UserButtonHandle;
    SELECT_HANDLE   UserSelectHandle;
    CLICK_HANDLE    UserClickHandle;
    KEY_HANDLE      UserKeyHandle;
    DESTROY_HANDLE  UserDestroyHandle;
};


/* ********************************************************************** */
/* **                           W O R K C E L L                        ** */
/* ********************************************************************** */
class WORKCELL : public WORLD
{
  public :
    // Constructor/Destructor:
                 WORKCELL         (char *Name = NULL, 
				   int SizeX = 250, int SizeY = 250);
                 WORKCELL         (Widget Parent, const char *Name = NULL);
                ~WORKCELL         ();

    // Methods:
    virtual void  Draw            (VIEW *View, GLenum Mode = GL_RENDER);
    void          DrawWorld       (int View, GLenum Mode);
    void          DrawFrame       (H_MATRIX Matrix);
#if 0
    virtual void  Print           ();

    bool          Save            (char *FileName);  // own format
#endif
    bool          LoadWorld       (const char *FileName, 
				   const char *DataPath = "~/tuxcad/data");

    PART3D       *GetMount        (char *DevName);

    int           GetDevList      (DEVICE3D ***List = NULL);
    int           GetPartList     (PART3D   ***List = NULL);

    void          DrawText        ();

    virtual bool  KeyHandle       (KeySym Key, int Mask);

  private :
    // Attributes:
    AXES  Axes;                         // CoordinateSystem
};


/* ********************************************************************** */
/* **                             P L O T 3 D                          ** */
/* ********************************************************************** */
class PLOT3D : public WORLD
{
  public :
             PLOT3D        (const char *Name, 
			    int SizeX = 250, int SizeY = 250);
            ~PLOT3D        ();

    void     Draw          (VIEW *View, GLenum Mode = GL_RENDER);
    void     Redraw        (int ViewNo = -1);
    void     GetSize       (double *MinX, double *MaxX,
			    double *MinY, double *MaxY,
			    double *MinZ, double *MaxZ);

    void     SetValues     (int Rows, int Cols, POINT3D **Points);

  private :
    void           DeleteValues  ();

    int            Rows, Cols;
    POINT3D      **Values;
};


/*
class CIRCUIT : public WORLD
{
  // ...
}
*/

/*
  void WORLD::TransPart (PART3D *Part,
                         double dx, double dy, double dz, int Steps=1);
  void WORLD::MovePart (PART3D *Part, MATRIX M, Speed, Steps, execute=true);
  void WORLD::RotPart  (PART3D *Part, 
                        double rx, double ry, double rz, int Steps=1);
*/


// use this namespace in all units
using namespace EXCEPTIONS;





#endif








