#include <unistd.h>   /* sleep */

#include "libgeo.h"


typedef enum
{
  GeoInPoly2 = -1,
  GeoUnknown,
  GeoInPoly1
} POLYGON_INCLUDE;


/* Twice the Area of the triangle in the xy-area. Positive if oriented CCW, 
   negative, if CW */
static double GEO2D_DoubleArea (POINT2D p1, POINT2D p2, POINT2D p3)
{
  return p1.x * p2.y - p1.y * p2.x +
         p1.y * p3.x - p1.x * p3.y + 
         p2.x * p3.y - p3.x * p2.y;
}



/* Check the orientation of a polyon in the xy-area walking through the 
  pointlist straight forward. 
  Result: -1 (GeoDirCW) :  Orientation is CW 
           0 (GeoDirNO) :  polygon is invalid
           1 (GeoDirCCW):  Orientation is CCW 
*/
int GEO2DPolygonOrientation (POLYGON2D *Poly)
{
  int    count;
  double Sum = 0.0;

  for (count = 1; count < Poly->Number-1; ++count)
    Sum += 
      GEO2D_DoubleArea (Poly->Points[0], 
			Poly->Points[count], 
			Poly->Points[count+1]);

  return SIGN (Sum);
}


POLYGON2D *GEO2DNewPolygon ()
{
  POLYGON2D *Polygon;

  Polygon = (POLYGON2D *)malloc (sizeof(POLYGON2D));
  Polygon->Number = 0;
  Polygon->Points = NULL;

  return Polygon;
}


void GEO2DFreePolygon (POLYGON2D *Polygon)
{
  if (Polygon == NULL)
    return;

  if (Polygon->Points != NULL)
    free (Polygon->Points);

  free (Polygon);
}


POLYGON2D *GEO2DPolygonTransformXY (POLYGON *Polygon)
{
  int        count;
  POINT3D    Point;
  H_MATRIX   M;
  POLYGON2D *Poly;

  M = GEOGetTransformXY (Polygon);

  Poly = (POLYGON2D *)malloc (sizeof(POLYGON2D));

  Poly->Number = Polygon->Number;
  Poly->Points = (POINT2D *)malloc (sizeof(POINT2D) * Polygon->Number);
  for (count = 0; count < Polygon->Number; ++count)
  {
    Point = GEOMatrixMult (M, Polygon->Points[Polygon->Vertex[count]]);
    Poly->Points[count].x = Point.x;
    Poly->Points[count].y = Point.y;
    PRO4 ("POLY %d:  [%5.1f %5.1f %5.1f]\n", count, Point.x, Point.y, Point.z)
  }
  return Poly;
}


void GEO2DSwapOrientation (POLYGON2D *Polygon)
{
  int     count, Index;
  POINT2D Swap;

  for (count = 0; count < Polygon->Number/2; ++count)
  {
    Index = Polygon->Number-count-1;
    Swap                   = Polygon->Points[count];
    Polygon->Points[count] = Polygon->Points[Index];
    Polygon->Points[Index] = Swap;
  }
}

static dual GEO2D_Left (POINT2D p1, POINT2D p2, POINT2D p3)
{
  return GEO2D_DoubleArea (p1, p2, p3) > 0.0;
}

static dual GEO2D_LeftOn (POINT2D p1, POINT2D p2, POINT2D p3)
{
  return GEO2D_DoubleArea (p1, p2, p3) >= 0.0;
}

/* are the vectors p1-p2 and p1-p3 colinear */
static dual GEO2D_Collinear (POINT2D p1, POINT2D p2, POINT2D p3)
{
  return GEO2D_DoubleArea (p1, p2, p3) == 0.0;
}


dual GEO2DInCone (int p1, int p3, POLYGON2D *Poly)
{
  int p2, p0; 

  p2 = (p1 + 1)                % Poly->Number;
  p0 = (p1 + Poly->Number - 1) % Poly->Number;

  /* if Polygon[p1] is a convex vertex [p2 left or on (p0,p1)]. */
  if (GEO2D_LeftOn(Poly->Points[p0], Poly->Points[p1], Poly->Points[p2]))
    return (GEO2D_Left(Poly->Points[p1], Poly->Points[p3], Poly->Points[p0]) &&
	    GEO2D_Left(Poly->Points[p3], Poly->Points[p1], Poly->Points[p2]));

  return 
    ! (GEO2D_LeftOn(Poly->Points[p1], Poly->Points[p3], Poly->Points[p2]) &&
       GEO2D_LeftOn(Poly->Points[p3], Poly->Points[p1], Poly->Points[p0]));
}


static dual GEO2D_Between (POINT2D p1, POINT2D p2, POINT2D p3)
{
  if (! GEO2D_Collinear(p1, p2, p3))
    return FALSE;

  if (p1.x != p2.x)
    return ((p1.x <= p3.x && p3.x <= p2.x) ||
	    (p1.x >= p3.x && p3.x >= p2.x));
  else
    return ((p1.y <= p3.y && p3.y <= p2.y) ||
	    (p1.y >= p3.y && p3.y >= p2.y));
}

static dual GEO2D_IntersectProp (POINT2D p1, POINT2D p2, 
				 POINT2D p3, POINT2D p4)
{
  if (GEO2D_Collinear(p1, p2, p3) || GEO2D_Collinear(p1, p2, p4) ||
      GEO2D_Collinear(p3, p4, p1) || GEO2D_Collinear(p3, p4, p2))
    return FALSE;

  return ((GEO2D_Left(p1, p2, p3) ^ GEO2D_Left(p1, p2, p4)) &&
	  (GEO2D_Left(p3, p4, p1) ^ GEO2D_Left(p3, p4, p2)));
}

static dual GEO2D_Intersect (POINT2D p1, POINT2D p2, POINT2D p3, POINT2D p4)
{
  if (GEO2D_IntersectProp(p1, p2, p3, p4))
    return TRUE;

  return (GEO2D_Between(p1, p2, p3) || GEO2D_Between(p1, p2, p4) ||
	  GEO2D_Between(p3, p4, p1) || GEO2D_Between(p3, p4, p2));
}

dual GEO2DDiagonalie (int p1, int p3, POLYGON2D *Poly)
{
  int count, count1;

  for (count = 0; count < Poly->Number; ++count)
  {
    count1 = (count + 1) % Poly->Number;
    if (!(count == p1 || count1 == p1 || count == p3 || count1 == p3))
      if (GEO2D_Intersect(Poly->Points[p1],    Poly->Points[p3],
			  Poly->Points[count], Poly->Points[count1]))
	return FALSE;
  }
  return TRUE;
}


/* *********************************************************************** */
/* **                          Polygon Intersection                     ** */
/* *********************************************************************** */
static void GEO2D_AddPolygonPoint (POLYGON **Polygon, POINT2D Point)
{
  POLYGON *Poly;

  if (*Polygon == NULL)
    *Polygon = Poly = GEONewPolygon ();
  else
  {
    Poly = *Polygon;
    if (Poly->Number > 0 && Poly->Points[Poly->Number-1].x == Point.x &&
 	                    Poly->Points[Poly->Number-1].y == Point.y)
    return;
  }

  Poly->Points = 
    (POINT3D *)realloc (Poly->Points, sizeof(POINT3D) * (Poly->Number+1));
  Poly->Points[Poly->Number].x = Point.x;
  Poly->Points[Poly->Number].y = Point.y;
  Poly->Points[Poly->Number].z = 0.0;

  Poly->Vertex = (int *)realloc (Poly->Vertex, sizeof(int) * (Poly->Number+1));
  Poly->Vertex[Poly->Number] = Poly->Number;
  (Poly->Number)++;
}

static dual GEO2D_EdgeIntersection (POINT2D p1, POINT2D p2, 
				    POINT2D p3, POINT2D p4, POINT2D *p)
{
  double s, t;
  double denom;

  denom = p1.x * (p4.y - p3.y) +
          p2.x * (p3.y - p4.y) +
          p4.x * (p2.y - p1.y) +
          p3.x * (p1.y - p2.y);

  if (denom == 0.0)
    return FALSE;

  s = (p1.x * (p4.y - p3.y) + 
       p3.x * (p1.y - p4.y) +
       p4.x * (p3.y - p1.y)) / denom;

  t = - (p1.x * (p3.y - p2.y) +
         p2.x * (p1.y - p3.y) +
	 p3.x * (p2.y - p1.y)) / denom;

  p->x = p1.x + s * (p2.x - p1.x);
  p->y = p1.y + s * (p2.y - p1.y);

  return (0.0 <= s && s <= 1.0 && 0.0 <= t && t <= 1.0);
}

static int GEO2D_InOut (POINT2D p, dual aHB, POLYGON **Polygon)
{
  GEO2D_AddPolygonPoint (Polygon, p);
  
  return aHB? GeoInPoly1 : GeoInPoly2;
}

static int GEO2D_Advance (int a, int *Count, int Number, dual inside, 
			  POINT2D p, POLYGON **Polygon)
{
  if (inside)
    GEO2D_AddPolygonPoint (Polygon, p);

  (*Count)++;
  return (a+1) % Number;
}


dual GEO2DPointInPolygon (POLYGON2D *Poly, POINT2D Point)
{
  int      count, count1;
  int      Crossings = 0;
  POINT2D *PointList;

  PRO_START ("")

  PointList = (POINT2D *)malloc (sizeof(POINT2D) * Poly->Number);

  /* Transform polygon, so that point lies in origin */
  for (count = 0; count < Poly->Number; ++count)
  {
    PointList[count].x = Poly->Points[count].x - Point.x;
    PointList[count].y = Poly->Points[count].y - Point.y;
  }
  
  for (count = 0; count < Poly->Number; ++count)
  {
    count1 = ABS_MOD (count+1, Poly->Number);
    if ((PointList[count].y  > 0 && PointList[count1].y <= 0) ||
	(PointList[count1].y > 0 && PointList[count].y <= 0))
      if ((PointList[count].x  * PointList[count1].y - 
           PointList[count1].x * PointList[count].y) /
	  (PointList[count1].y - PointList[count].y) > 0)
        Crossings++;
  }

  free (PointList);

  PRO_END
  return (Crossings & 1);
}

static POLYGON *GEO2D_Make3DPolygon (POLYGON2D *Poly)
{
  int      count;
  POLYGON *Polygon = NULL;

  Polygon = GEONewPolygon ();
  Polygon->Number = Poly->Number;
  Polygon->Points = (POINT3D *)malloc (sizeof(POINT3D) * Poly->Number);
  Polygon->Vertex = (int *)malloc (sizeof(int) * Poly->Number);
  for (count = 0; count < Poly->Number; ++count)
  {
    Polygon->Points[count].x = Poly->Points[count].x;
    Polygon->Points[count].y = Poly->Points[count].y;
    Polygon->Points[count].z = 0.0;
    Polygon->Vertex[count] = count;
  }
  return Polygon;
}



POLYGON *GEO2DConvexIntersection (POLYGON2D *Poly1, POLYGON2D *Poly2)
{
  dual     bHA, aHB;
  int      a = 0, b = 0;
  int      a1, b1;
  int      inflag = GeoUnknown;
  int      countA = 0, countB = 0;
  double   Cross;
  POINT2D  Origin = {0.0, 0.0};
  POINT2D  Point, P1E, P2E;
  POLYGON *Polygon = NULL;

  PRO_START ("\n")

  if (Poly1 == NULL || Poly1->Number == 0 || 
      Poly2 == NULL || Poly2->Number == 0)
    PRO_RETURN ("Empty Polygon", NULL)

  do
  {
    a1 = (a + Poly1->Number - 1) % Poly1->Number;
    b1 = (b + Poly2->Number - 1) % Poly2->Number;
    P1E.x = Poly1->Points[a].x - Poly1->Points[a1].x;
    P1E.y = Poly1->Points[a].y - Poly1->Points[a1].y;
    P2E.x = Poly2->Points[b].x - Poly2->Points[b1].x;
    P2E.y = Poly2->Points[b].y - Poly2->Points[b1].y;
    
    Cross = GEO2D_DoubleArea (Origin, P1E, P2E);
    bHA = GEO2D_LeftOn (Poly1->Points[a1], Poly1->Points[a], Poly2->Points[b]);
    aHB = GEO2D_LeftOn (Poly2->Points[b1], Poly2->Points[b], Poly1->Points[a]);
    
    /* if A and B intersect, update inflag */
    if (GEO2D_EdgeIntersection(Poly1->Points[a1], Poly1->Points[a],
			       Poly2->Points[b1], Poly2->Points[b], &Point))
    {
      if (inflag == GeoUnknown)
        countA = countB = 0;
      inflag = GEO2D_InOut (Point, aHB, &Polygon);
    }

    if ((Cross >= 0.0 && bHA) || (Cross < 0.0 && ! aHB))
      a = GEO2D_Advance (a, &countA, Poly1->Number, 
			 inflag == GeoInPoly1, Poly1->Points[a], &Polygon);
    else
      b = GEO2D_Advance (b, &countB, Poly2->Number,
			 inflag == GeoInPoly2, Poly2->Points[b], &Polygon);
  } while ((countA < Poly1->Number) && (countB < Poly2->Number));

  PRO2 ("countA: %d   countB: %d\n", countA, countB)

  /* If no intersections was found: Test, if one polygon includes the other. */
  if (Polygon == NULL)
  {
    if (GEO2DPointInPolygon(Poly1, Poly2->Points[0]))
      Polygon = GEO2D_Make3DPolygon (Poly2);
    elsif (GEO2DPointInPolygon(Poly2, Poly1->Points[0]))
      Polygon = GEO2D_Make3DPolygon (Poly1);
  }

  PRO_END
  return Polygon;
}

