#include <math.h>

#include "libgeo.h"


#define COPLANAR_RANGE 1e-6
#define COPLANAR_DIST  1e-3


INTERSECT *GEONewIntersection (POLYGON *Poly1, POLYGON *Poly2)
{
  int        Side;
  INTERSECT *Result;

  Result = (INTERSECT *)malloc (sizeof(INTERSECT));
  Result->ID      = 0;
  Result->Type    = -1;
  Result->Poly[0] = Poly1;
  Result->Poly[1] = Poly2;
  for (Side = 0; Side < 2; ++Side)
  {
    Result->Number[Side]    = 0;
    Result->Points[Side]    = NULL;
    Result->Index[Side][0]  = -1;
    Result->Index[Side][1]  = -1;
    Result->Lambda[Side][0] = -1.0;
    Result->Lambda[Side][1] = -1.0;
  }
  Result->done = FALSE;
  Result->Next = NULL;

  return Result;
}


void GEOFreeIntersection (INTERSECT *Intersect)
{
  if (Intersect == NULL)
    return;

  if (Intersect->Points[0] != NULL)
    free (Intersect->Points[0]);
  if (Intersect->Points[1] != NULL)
    free (Intersect->Points[1]);
}


/* ************************************************************************ *
   Compute the instersection of two convex polygons. There are 11 main cases:
   1. Polygons are collinear but not in the same area      -> no intersection
   2. Polygons have same area but are disjunct             -> no intersection
   3. Polygons lies in different areas but do not collide  -> no intersection
   4. Polygons have one common vertex                      -> Vertex
   5. One point of a polygon lies on the edge of the other -> Vertex
   6. One point of a polygon lies in the area of the other -> Vertex
   7. Both polygons have a common edge                     -> Edge
   8. An Edge of one polygon lies in the area of the other -> Edge
   9. The polygons have different areas and cut each other -> Edge
  10. One polygon is included by the other                 -> Polygon (convex)
  11. Both polygons intersect each other                   -> Polygon (convex)
  ************************************************************************** */

/* Computer Intersection of two convex polygons which are coplanar. */
static POLYGON *GEO_PolygonIntersection (POLYGON *Poly1, POLYGON *Poly2)
{
  int        count;
  POINT3D    Point;
  H_MATRIX   M;
  POLYGON2D *PolyA, *PolyB;
  POLYGON   *Polygon;

  PRO_START ("\n")

#if 0
  M = GEOGetTransformXY (Poly1);
  WRITE_MATRIX (stdout, M, "TransformXY")
#else
  M = Matrix_1;
#endif

  /* Transform both polygons into xy-area using matrix of first polygon */
  PolyA = GEO2DNewPolygon ();
  PolyA->Number = Poly1->Number;
  PolyA->Points = (POINT2D *)malloc (sizeof(POINT2D) * Poly1->Number);
  for (count = 0; count < Poly1->Number; ++count)
  {
    Point = GEOMatrixMult (M, Poly1->Points[Poly1->Vertex[count]]);
    PolyA->Points[count].x = Point.x;
    PolyA->Points[count].y = Point.y;
    PRO4 ("POLY %d:  [%5.1f %5.1f %5.1f]\n", count, Point.x, Point.y, Point.z)
  }

  PolyB = GEO2DNewPolygon ();
  PolyB->Number = Poly2->Number;
  PolyB->Points = (POINT2D *)malloc (sizeof(POINT2D) * Poly2->Number);
  for (count = 0; count < Poly2->Number; ++count)
  {
    Point = GEOMatrixMult (M, Poly2->Points[Poly2->Vertex[count]]);
    PolyB->Points[count].x = Point.x;
    PolyB->Points[count].y = Point.y;
    PRO4 ("POLY %d:  [%5.1f %5.1f %5.1f]\n", count, Point.x, Point.y, Point.z)
  }

  if (GEO2DPolygonOrientation(PolyA) == GeoDirCW)
  {
#ifdef DEBUG_CSG
    printf ("Swap PolygonA\n");
#endif
    GEO2DSwapOrientation (PolyA);
  }

  if (GEO2DPolygonOrientation(PolyB) == GeoDirCW)
  {
#ifdef DEBUG_CSG
    printf ("Swap PolygonB\n");
#endif
    GEO2DSwapOrientation (PolyB);
  }

  /* Calc intersection and retransform computed polygon */
  if ((Polygon = GEO2DConvexIntersection(PolyA, PolyB)) != NULL)
    GEOTransformPolygon (Polygon, GEOMatrixInv(M));

  GEO2DFreePolygon (PolyA);
  GEO2DFreePolygon (PolyB);

  PRO_END
  return Polygon;
}


INTERSECT *GEO_CoplanarIntersection (POLYGON *Poly1, POLYGON *Poly2)
{
  int        count1, count2;
  double     Lambda, mu;
  POINT3D    Point, *Contact;
  POINT3D   *Points1 = Poly1->Points;
  POINT3D   *Points2 = Poly2->Points;
  VECTOR3D   Dir1, Dir2;
  POLYGON   *Actual = NULL;
  INTERSECT *Result = NULL;

#if 0
  Result = GEONewIntersection (Poly1, Poly2);
  Result->Type = geoCOPLANAR;
  Result->Points[0] = (POINT3D *)malloc (sizeof(POINT3D) * ...);
  Result->Points[1] = (POINT3D *)malloc (sizeof(POINT3D) * ...);
#endif

  count1 = 0;
  while (count1 < Poly1->Number && Actual == NULL)
    if (GEOPointInPolygon(Poly2, Points1[Poly1->Vertex[count1]]))
      Actual = Poly1;
    else
      ++count1;

  if (Actual == NULL)
  {
    count1 = 0;
    while (count1 < Poly2->Number && Actual == NULL)
      if (GEOPointInPolygon(Poly1, Points2[Poly2->Vertex[count1]]))
	Actual = Poly2;
      else
	++count1;
  }

  if (Actual == NULL)
  {
    count1 = 0;
    while (count1 < Poly1->Number-1)
    {
      Dir1 = GEOVectorSub (Points1[Poly1->Vertex[(count1+1) % Poly1->Number]],
			   Points1[Poly1->Vertex[count1]]);
      count2 = 0;
      while (count2 < Poly2->Number-1)
      {
	Dir2 = GEOVectorSub (Points2[Poly2->Vertex[count2+1]],
			     Points2[Poly2->Vertex[count2]]);
	if (GEOLineIntersection(Points1[Poly1->Vertex[count1]], Dir1,
				Points2[Poly2->Vertex[count2]], Dir2,
				&Point, &Lambda, &mu) && 
	    Lambda >= 0.0 && Lambda <= 1.0)
	  Actual = Poly1;
	else
	  ++count2;
      }
      if (Actual == NULL)
	++count1;
    }
  }
  
  if (Actual == NULL)
    return NULL;

  Contact = (POINT3D *)malloc (sizeof(POINT3D) * 
			       (Poly1->Number+Poly2->Number));

#if 0
    /* IDEE: 
       1. Suche Punkt der ausserhalb des anderen Polygons liegt
       2. Iteriere ber die Kanten, bis eine Kante mit Schnittpunkt gefunden
       wird, oder alle Kanten durchlaufen sind.
       3. Falls alle Kanten durchlaufen: anderes Polygon ist included
       sonst: Wechsle zum anderen Polygon und verfolge dieses, bis nchsten
       Schnittpunkt, dann wieder Wechsel.
    */

    for (count1 = 0; count1 < Poly1->Number-1; ++count1)
    {
      Dir1 = GEOVectorSub (Poly1->Points[Poly1->Vertex[count1+1]],
			   Poly1->Points[Poly1->Vertex[count1]]);
      for (count2 = 0; count2 < Poly2->Number-1; ++count2)
      {
	Dir2 = GEOVectorSub (Poly1->Points[Poly1->Vertex[count2+1]],
			     Poly1->Points[Poly1->Vertex[count2]]);
	if (GEOLineIntersection(Poly1->Points[Poly1->Vertex[count1]], Dir1,
				Poly2->Points[Poly2->Vertex[count2]], Dir2,
				&Point, &Lambda, &mu))
	{
	  /* TODO: gemeinsame Flche ermitteln
	     Prfen, ob nchster Punkt(Poly1) innerhalb oder ausserhalb von 
	     Poly2 ist, entsprechend weiterlaufen (dabei speichern) bis zum
	     nchsten Schnittpunkt, dann anderes Polygon */
	}
      }
    }
  }
#endif

  return Result;
}


INTERSECT *GEOPolygonIntersection (POLYGON *Poly1, POLYGON *Poly2)
{
  int        Contacts1, Contacts2 = 0;
  int        Index[4];
  double     Distance;
  double     LLambda[4] = {0.0, 0.0};   /* lambda of intersection line */ 
  double     ELambda[4] = {0.0, 0.0};   /* Lambda of intersection edge */
  POINT3D    Point;
  POINT3D    Point1a, Point1b;
  POINT3D    Point2a, Point2b;
  VECTOR3D   Dir, Normal;
  POLYGON   *Poly;
  INTERSECT *Result = NULL;

  PRO_START ("\n")

  if (Poly1 == NULL || Poly2 == NULL)
    PRO_RETURN ("Empty Polygon", NULL)

  /* Test, if both polygons are coplanar */
  Normal = GEOVectorCross (Poly1->Info.n, Poly2->Info.n);
  if (GEOVectorLength(Normal) < COPLANAR_RANGE)
  {
#ifdef DEBUG_CSG
printf ("COPLANAR intersection\n");
#endif
    /* Test the distance between both area */
    Distance = GEOAreaPointDist (Poly1, Poly2->Points[Poly2->Vertex[0]]);
    if (fabs(Distance) > COPLANAR_DIST)
      PRO_RETURN ("Polygons coplanar but disjunct", NULL)

    if ((Poly = GEO_PolygonIntersection(Poly1, Poly2)) != NULL)
    {
      Result = GEO_CoplanarIntersection (Poly1, Poly2);
      PRO_RETURN ("Plane intersection", Result)
    }
  }
  /* Compute the intersection line of both polygons' areas */
  elsif (GEOAreaIntersection( Poly1->Info.n,  Poly1->Info.a, 
	                      Poly2->Info.n,  Poly2->Info.a, 
	                     &Point,         &Dir))
  {
#ifdef DEBUG_CSG
printf ("Computing Obj1: Poly <-> Obj2: Poly\n"
        "%.3f %.3f %.3f / %.3f %.3f %.3f\n",
        Point.x, Point.y, Point.z, Dir.x, Dir.y, Dir.z);
#endif

    /* Compute intersection of this line and both polygons. If at least one
       polygon has no intersection points with this line, both polygons are
       disjunct. */ 
    if ((Contacts1 =
         GEOPolygonLineIntersection(Poly1,  Point,       Dir, 
	                                   &Point1a,    &Point1b,
				           &LLambda[0], &LLambda[1],
				           &Index[0],   &Index[1],
				           &ELambda[0], &ELambda[1])) == 0 ||
        (Contacts2 =
	   GEOPolygonLineIntersection(Poly2,  Point,       Dir, 
				             &Point2a,    &Point2b,
				             &LLambda[2], &LLambda[3],
				             &Index[2],   &Index[3],
				             &ELambda[2], &ELambda[3])) == 0)
      PRO_RETURN ("Polygons are disjunct", NULL)

    PRO0 ("TRANSVERSAL intersection\n")

    if (Contacts1 == 1)
    {
      LLambda[1] = LLambda[0];
      ELambda[1] = ELambda[0];
      Point1b = Point1a;
    }
    if (Contacts2 == 1)
    {
      LLambda[3] = LLambda[2];
      ELambda[3] = ELambda[2];
      Point2b = Point2a;
    }

    /* TODO:
       berlegen, was zu tun ist, wenn beiden gleichen Bereich haben
       Fr beide Polygon (doppelt) eintragen ?
       ...-------+-------...
                 | 
         Poly1   |  Poly2
	         | 
       ...-------+-------...
    */

    Result = GEONewIntersection (Poly1, Poly2);
    Result->Type      = geoTRANSVERSAL;
    Result->Points[0] = (POINT3D *)malloc (sizeof(POINT3D) * 2);
    Result->Points[1] = (POINT3D *)malloc (sizeof(POINT3D) * 2);

    /* Compare the polygons' regions on the intersection line. We're 
       only interested in the common range */
    if (LLambda[0] >= LLambda[2] && LLambda[0] <= LLambda[3])
    {
      PRO_VECTOR3D (Point1a, "POLY-CONTACT#1a")
      Result->Points[0][0] = Point1a;
      Result->Index[0][0]  = Index[0];
      Result->Lambda[0][0] = ELambda[0];
      (Result->Number[0])++;
    }
    if (LLambda[1] >= LLambda[2] && LLambda[1] <= LLambda[3])
    {
      PRO_VECTOR3D (Point1b, "POLY-CONTACT#1b")
      Result->Points[0][Result->Number[0]] = Point1b;
      Result->Index[0][Result->Number[0]]  = Index[1]; 
      Result->Lambda[0][Result->Number[0]] = ELambda[1];
      (Result->Number[0])++; 
    }
    if (LLambda[2] >= LLambda[0] && LLambda[2] <= LLambda[1])
    {
      PRO_VECTOR3D (Point2a, "POLY-CONTACT#2a")
      Result->Points[1][0] = Point2a;
      Result->Index[1][0]  = Index[2];
      Result->Lambda[1][0] = ELambda[2];
      (Result->Number[1])++; 
    }
    if (LLambda[3] >= LLambda[0] && LLambda[3] <= LLambda[1])
    {
      PRO_VECTOR3D (Point2b, "POLY-CONTACT#2b")
      Result->Points[1][Result->Number[1]] = Point2b;
      Result->Index[1][Result->Number[1]]  = Index[3]; 
      Result->Lambda[1][Result->Number[1]] = ELambda[3];
      (Result->Number[1])++;
    }

    /* Test, if the contacts of both polygons have disjunct ranges on the 
       intersection line */
    if (Result->Number[0] == 0 && Result->Number[1] == 0)
    {
      GEOFreeIntersection (Result);
      PRO_RETURN ("No contact", NULL)
    }

    /* TODO: "Vertex in Polygon" und "Polygon in Polygon" testen */
  }
  else
    PRO0 ("Polygons do not intersect")

  PRO_END
  return Result;
}


#if 0
dual GEOPointInTetraeder (...)
/* TO BE DONE: Test sign of distance to all four faces */
#endif


dual GEOPointInPolyeder (POLYEDER *Polyeder, POINT3D Point)
{
  /* TODO... */

  return TRUE;
}


void GEOPolyederIntersection (POLYEDER *Polyeder1, POLYEDER *Polyeder2)
{
  int      count;
  POINT3D *Points;
  POLYGON *Poly;

  /* 
     Erster Test:
     Ermittle die Punkte des 2. Polyeders, die im ersten Polyeder liegen 
  */

  Poly = Polyeder2->List;
  while (Poly != NULL)
  {
    Points = Poly->Points;  
    /* 
       TODO: 
       Punktliste kopieren und an entsprechende Position transformieren.
       -> besser Punktliste des Objekts benutzen statt fr jedes Polygon 
          einzeln
    */

    for (count = 0; count < Poly->Number; ++count)
      if (GEOPointInPolyeder(Polyeder1, Points[Poly->Vertex[count]]))
	printf ("Polyeder2 Point %d in Polyeder 1\n", count);
    Poly = Poly->Next;
  }
}
