#include <math.h>

#include "libgeo.h"


#define LINE_CONTACT  1E-5


int GEOPointOnEdge (POINT3D p1, POINT3D p2, POINT3D p)
{
  double   x, t1, t2;
  VECTOR3D Edge, Diff;

  Edge = GEOVectorSub (p2, p1);
  Diff = GEOVectorSub (p,  p1);

  if (Edge.x != 0.0)
  {
    x  = Diff.x / Edge.x;
    t1 = x * Edge.y;
    t2 = x * Edge.z;

    if (fabs(t1 - Diff.y) < D_DIST && fabs(t2 - Diff.z) < D_DIST)
      if (x == 0.0 || x == 1.0 || (x > -D_DIST && x < 1.0+D_DIST))
	return 1;                       /* Point lies on the edge */
      elsif (x > 0.0)
	return 3;                       /* Point lies outside p2  */
      else
	return 2;                       /* Point lies outside p1  */
    else
      return 0;                         /* Point lies not the edge's line */
  }
  elsif (Edge.y != 0.0)                 /* edge is vertical (same x-coord) */
  {
    x = Diff.y / Edge.y;
    t1 = x * Edge.x;
    t2 = x * Edge.z;

    if (fabs(t1 - Diff.x) < D_DIST && fabs(t2 - Diff.z) < D_DIST)
      if (x == 0.0 || x == 1.0 || (x > -D_DIST && x < 1.0+D_DIST))
	return 1;
      elsif (x >= 0.0)
	return 3;
      else
	return 2;
    else
      return 0;
  }
  elsif (Edge.z != 0.0)
  {
    x = Diff.z / Edge.z;
    t1 = x * Edge.x;
    t2 = x * Edge.y;

    if (fabs(t1 - Diff.x) < D_DIST && fabs(t2 - Diff.y) < D_DIST)
      if (x == 0.0 || x == 1.0 || (x > -D_DIST && x < 1.0+D_DIST))
	return 1;
      elsif (x > 0.0)
	return 3;
      else
	return 2;
    else
      return 0;
  }
  else
    return 0;
}


/* Compute contact of two edges. Result is:
  -1. Error while computation
   1. Insersection of both edges
   2. 
   3. vectors on same line, but disjunct
   4. Intersection of two collinear edges
   5. Vectors are collinear with distance > 0
   6. Both edges are identically
   7.
*/
int GEOEdgeContact (POINT3D   p1, POINT3D   p2, POINT3D p3, POINT3D p4, 
                    VECTOR3D *pA, VECTOR3D *pB)
{
  int      Index;
  double   Det, DetX, DetY;
  double   x1, x2, Test;
  VECTOR3D Dir12, Dir34;
  VECTOR3D LHS[2], RHS, Cross;

  if (pA != NULL)
    *pA = Vector_0;
  if (pB != NULL)
    *pB = Vector_0;

  if ((VEC_SAME(p1,p3) && VEC_SAME(p2,p4)) ||
      (VEC_SAME(p1,p4) && VEC_SAME(p2,p3)))
  {
    if (pA != NULL) 
      *pA = p1;
    if (pB != NULL) 
      *pB = p2;
    return 6;
  }

  Dir12 = GEOVectorSub (p2, p1);
  Dir34 = GEOVectorSub (p4, p3);
  RHS   = GEOVectorSub (p3, p1);

  Cross = GEOVectorCross (Dir12, Dir34);
  if (VEC_LEN(Cross) == 0.0)                 /* collinear lines           */
  {
     Cross = GEOVectorCross (Dir12, RHS);     /* same vectors              */
    if (VEC_LEN(Cross) == 0.0)
    {
      if (GEOPointOnEdge(p3, p1, p2) == 1 && 
	  GEOPointOnEdge(Dir34, p1, p2) == 1)
      {
	if (pB != NULL)
	  *pB = p2;

	return 4;
      }
      else
      {
	
	return 3;
      }
    }
    else
      return 5;
  }

  LHS[0] = Dir12;
  LHS[1] = GEOVectorMult (-1.0, Dir34);

  Det = LHS[0].x * LHS[1].y - LHS[1].x * LHS[0].y;
  Index = 0;
  if (fabs(Det) < D_DIST)
  {
    Det = LHS[0].x * LHS[1].z - LHS[1].x * LHS[0].z;
    Index = 1;
    if (fabs(Det) < D_DIST)
    {
      Det = LHS[0].y * LHS[1].z - LHS[1].y * LHS[0].z;
      Index = 2;
      if (fabs(Det) < D_DIST)
	return 0;
    }
  }

  switch (Index)
  {
    case 0 :
      DetX = RHS.x    * LHS[1].y - LHS[1].x * RHS.y;
      DetY = LHS[0].x * RHS.y    - RHS.x    * LHS[0].y;
      break ;
    case 1 :
      DetX = RHS.x    * LHS[1].z - LHS[1].x * RHS.z;
      DetY = LHS[0].x * RHS.z    - RHS.x    * LHS[0].z;
      break ;
    case 2 :
      DetX = RHS.y    * LHS[1].z - LHS[1].y * RHS.z;
      DetY = LHS[0].y * RHS.z    - RHS.y    * LHS[0].z;
      break ;
    default :
      PRO0 ("Error computing cramer-rule\n")
      return -1;
  }
  
  x1 = DetX / Det;
  x2 = DetY / Det;

  /* Check correctness with 3rd equation                                  */
  switch (Index)
  {
    case 0 :
      Test = x1 * LHS[0].z + x2 * LHS[1].z;
      if (Test == 0.0 && RHS.z == 0.0)
	break;
      if (fabs(Test-RHS.z) > EPSILON)
	return 0;
      break;

    case 1 : 
      Test = x1 * LHS[0].y + x2 * LHS[1].y;
      if (Test == 0.0 && RHS.y == 0.0)
	break;
      if (fabs(Test-RHS.y) > EPSILON)
	return 0;
      break;

    case 2 : 
      Test = x1 * LHS[0].x + x2 * LHS[1].x;
      if (Test == 0.0 && RHS.x == 0.0)
	break;
      if (fabs(Test-RHS.x) > EPSILON)
	return 0;
      break ;

    default : 
      PRO0 ("Error while checking parameter\n")
      return -1;
  }

  /* Computate intersection-point */
  if (pA != NULL)
    *pA = GEOVectorAdd (p1, GEOVectorMult(x1, Dir12));

  if (! (x1 < D_DIST || x1 > 1.0-D_DIST || x2 < D_DIST || x2 > 1.0-D_DIST))
    return 1;
  elsif (x1 >= 0.0 && ! (x2 < D_DIST || x2 > 1.0-D_DIST))
    return 7;
  else
    return 2;
}


/* ********************************************************************** */
/* ** Test, if a pooint p is included in a polygon.                    ** */
/* ** INPUT:  Polygon  Pointer to the polygon                          ** */
/* **         p        the interesting point                           ** */
/* ** OUTPUT: (bool)   TRUE: point is included                         ** */
/* **                  FALSE: point is outside                         ** */
/* ********************************************************************** */
dual GEOPointInPolygon (POLYGON *Polygon, VECTOR3D Point)
{
  int       count;
#if 0
  VECTOR3D  Cross1, Cross2;
  VECTOR3D  Diff1, Diff2;
#endif
  POINT3D  *Points;

  if (Polygon == NULL)
    return FALSE;

  /* Abort, if the point lies not in the same area */
  if (GEOAreaPointDist(Polygon, Point) > 1e-2)
    return FALSE;

#if 1
  if (Polygon->Triangle == NULL)
  {
    PRO0 ("Polygon not triangulated !\n")
    return FALSE;
  }

  Points = Polygon->Points;
  for (count = 0; count < Polygon->Number-2; ++count)
    if (GEOPointInTriangle(Points[Polygon->Triangle[count][0]],
			   Points[Polygon->Triangle[count][1]],
			   Points[Polygon->Triangle[count][2]], Point))
      return TRUE;

  return FALSE;
#else
  Points = Polygon->Points;

  Diff1 = GEOVectorSub (Points[Polygon->Vertex[1]],
			Points[Polygon->Vertex[0]]);
  Diff2 = GEOVectorSub (Point, Points[Polygon->Vertex[0]]);
  Cross1 = GEOVectorCross (Diff1, Diff2);

  for (count = 1; count < Polygon->Number-2; ++count)
  {
    Diff1 = GEOVectorSub (Points[Polygon->Vertex[count+1]].Pos,
			  Points[Polygon->Vertex[count]].Pos);
    Diff2 = GEOVectorSub (Point, Points[Polygon->Vertex[count]].Pos);
    Cross2 = GEOVectorCross (Diff1, Diff2);

    if (((Cross1.x > 0.0 && Cross2.x < 0.0) ||
	 (Cross1.x < 0.0 && Cross2.x > 0.0)) &&
	((Cross1.y > 0.0 && Cross2.y < 0.0) ||
	 (Cross1.y < 0.0 && Cross2.y > 0.0)) &&
	((Cross1.z > 0.0 && Cross2.z < 0.0) ||
	 (Cross1.z < 0.0 && Cross2.z > 0.0)))
      return FALSE;
  }

  return TRUE;
#endif
}


#if 0
/* wird nicht bentigt (Aufruf von Compute_Contact) */
int GEOPolygonContact2 (POLYGON *Polygon1, POLYGON *Polygon2, POLYGON *Contact)
{
  int      count;
  int      count1 = 0, count2 = 0;
  VECTOR   Point, p1, p2, p3, p4;
  VECTOR  *Points1, *Points2;
  POLYGON *Polygon = NULL;

  PRO_START (" ")

  if (Polygon1 == NULL || Polygon2 == NULL)
    PRO_RETURN ("Invalid Polygon(s)", 0)

  if (Polygon1->Number < 3 && Polygon2->Number < 3)
    PRO_RETURN ("Less than three Points", 0)

  Points1 = Polygon1->Points;
  Points2 = Polygon2->Points;

  while (Polygon == NULL && count2 < Polygon2->Number)
    if (GEOPointInPolygon(Polygon1, Points2[Polygon2->Vertex[count2]]))
    {
      Point = Points2[Polygon2->Vertex[count2]];
      Polygon = Polygon2;
    }
    else
      count2++;

  while (Polygon == NULL && count1 < Polygon1->Number)
    if (GEOPointInPolygon(Polygon2, Points1[Polygon1->Vertex[count1]]))
    {
      Point = Points1[Polygon1->Vertex[count1]];
      Polygon = Polygon1;
    }
    else
    count1++;

  if (Polygon != NULL)
  {
    for (count = 0; count < Polygon->Number; ++count)
      if (VEC_SAME(Polygon->Points[Polygon->Vertex[count]], Point))
	PRO_RETURN ( "vector equal", 1)
    
    for (count = 0; count < Polygon->Number; ++count)
      if (GEOPointOnPolygon (Polygon, Point))
	PRO_RETURN ("Point on Polygon", 2)
  }

  for (count1 = 0; count1 < Polygon1->Number; ++count1)
    for (count2 = 0; count2 < Polygon2->Number; ++count2)
    {
      p1 = Points1[Polygon1->Vertex[count1]];
      p2 = Points1[Polygon1->Vertex[(count1+1) % Polygon1->Number]];
      p3 = Points2[Polygon2->Vertex[count2]];
      p4 = Points2[Polygon2->Vertex[(count2+1) % Polygon2->Number]];

      if (GEOEdgeIntersection (p1, p2, p3, p4, NULL))
	PRO_RETURN ("Edge intersection", 3)
    }
  
  PRO_END
  return 4;
}
#endif	       


static void GEO_AddPolygonPoint (POLYGON *Polygon, POINT3D p)
{
  Polygon->Points = (POINT3D *)realloc (Polygon->Points, 
				       sizeof(POINT3D) * (Polygon->Number+1));
  Polygon->Points[Polygon->Number] = p;

  Polygon->Vertex = (int*)realloc (Polygon->Vertex, 
				   sizeof(int) * (Polygon->Number+1));
  Polygon->Vertex[Polygon->Number] = Polygon->Number;

  (Polygon->Number)++;
}


POLYGON *GEOPolygonContact (POLYGON *Polygon1, POLYGON *Polygon2)
{
  dual      included, new = TRUE;
  int       count;
  int       count1, count2;
  VECTOR3D  v1, v2, v3, v4, v;
  POINT3D  *Points1, *Points2;
  POINT3D  *Points = NULL;
  POLYGON  *Contact = NULL;

  PRO_START (" ")

  if (Polygon1 == NULL || Polygon2 == NULL)
    PRO_RETURN ("Empty Polygon", NULL)

  if (Polygon1->Number < 3 && Polygon2->Number < 3)
    PRO_RETURN ("Less than three Points", 0)

  Points1 = Polygon1->Points;
  Points2 = Polygon2->Points;

  PRO0 ("\\1.")

  /* locate points of polygon 2 which are included in polygon 1           */
  included = TRUE;
  count2 = 0;
  while (count2 < Polygon2->Number)
  {
    if (! GEOPointInPolygon(Polygon1, Points2[Polygon2->Vertex[count2]]))
      included = FALSE;
    else
    {
      if (Contact == NULL)
	Contact = GEONewPolygon ();

      new = TRUE;
      Points = Contact->Points;
      count = 0;
      while (count < Contact->Number && new)
      {
	if (VEC_SAME(Points[count], Points2[Polygon2->Vertex[count2]]))
	  new = FALSE;
	else
	  count++;
      }

      if (new)
	GEO_AddPolygonPoint (Contact, Points2[Polygon2->Vertex[count2]]);
    }
    count2++;
  }

  PRO0 ("\\2.")

  /* Now the same procedure for polygon 1 included in polygon 2           */
  if (! included)             /* polygon 2 is NOT completely included     */
  {
    included = TRUE;
    count1 = 0;
    while (count1 < Polygon1->Number)
    {
      if (! GEOPointInPolygon(Polygon2, Points1[Polygon1->Vertex[count1]]))
        included = FALSE;
      else
      {
	if (Contact == NULL)
	  Contact = GEONewPolygon ();

	new = TRUE;
	Points = Contact->Points;
	count = 0;
	while (count < Contact->Number && new)
	{
	  if (VEC_SAME(Points[count], Points1[Polygon1->Vertex[count1]]))
	    new = FALSE;
	  else
	    count++;
	}

        if (new)
	  GEO_AddPolygonPoint (Contact, Points1[Polygon1->Vertex[count1]]);
      }
      count1++;
    }
  }

  if (included)
    PRO_RETURN ("<+>", Contact)

  PRO0 ("\\3.")

  /* Neither poly1 is included in poly2 nor poly2 is included in poly1:   */
  /* -> let's search for cutting-points                                   */
  count1 = 0;
  while (count1 < Polygon1->Number)
  {
    v1 = Points1[Polygon1->Vertex[count1]];
    v2 = Points1[Polygon1->Vertex[(count1+1) % Polygon1->Number]];

    count2 = 0;
    while (count2 < Polygon2->Number)
    {
      v3 = Points2[Polygon2->Vertex[count2]];
      v4 = Points2[Polygon2->Vertex[(count2+1) % Polygon2->Number]];

      if (GEOEdgeContact(v1, v2, v3, v4, &v, NULL) == 1)
      {
	if (Contact == NULL)
	  Contact = GEONewPolygon ();

	GEO_AddPolygonPoint (Contact, v);
      }
      count2++;
    }
    count1++;
  }

  PRO1 ("\\[%c].", (Contact != NULL)? '+':'-')

  PRO_END
  return Contact;
}


/* ********************************************************************** */
/* ** Test for intersection between a polygon and a line. Returns the  ** */
/* ** amount of intersections: 0,1 or 2                                ** */
/* ** The polygon must be convex and the specified line must lie in    ** */
/* ** the same area as the polygon.                                    ** */
/* ** Algorithm: Compute intersecton between line and all edges.       ** */
/* ** After computing the intersection of an edge with the given line  ** */
/* ** there will be checked, if the intersection point is in the       ** */
/* ** common segment. It returns eighter the intersection points and   ** */
/* the lambda values for this points needed for the point-direction    ** */
/* ** form.                                                            ** */
/* ********************************************************************** */
int GEOPolygonLineIntersection (POLYGON *Polygon,  POINT3D  Pos, VECTOR3D Dir,
				POINT3D *Point1,   POINT3D *Point2,
				double  *LLambda1, double  *LLambda2,
				int     *Index1,   int     *Index2,
				double  *ELambda1, double  *ELambda2)
{
  int       count, No = 0, I[2];
  int       first = 0, second = 1;
  double    Lambda[2], L[2], E[2];
  VECTOR3D  P1P2;
  POINT3D   P1, P2, C[2];
  POINT3D   Result = Vector_0;
  POINT3D  *Points;

  PRO_START (" ")

  Points = Polygon->Points;
  P2 = Points[Polygon->Vertex[0]];
  for (count = 0; count < Polygon->Number; ++count)
  {
    P1 = P2;
    P2 = Points[Polygon->Vertex[ABS_MOD(count+1, Polygon->Number)]];
    P1P2 = GEOVectorSub (P2, P1);

#if 1
    /* There have been inaccuracies when computing the intersection line 
       between two faces. As result, we might *not* find the intersection
       points between this line and the edges of the polygon because of
       a very small distance between both lines. */
    if (GEOLineLineDist(P1, P1P2, Pos, Dir) < LINE_CONTACT &&
	GEOLineLineLot(P1, P1P2, Pos, Dir, &Result, NULL) &&
	GEOVectorAngle(GEOVectorSub(Result, P1), P1P2) < 90.0)
    {
      Lambda[0] = GEOVectorLength(GEOVectorSub(Result, P1)) / 
	          GEOVectorLength(P1P2);

      if (Lambda[0] >= 0.0 && Lambda[0] <= 1.0)
      {
	/* There are maximal two intersection point in a convex polygon */
	if (No < 2)
	{
	  Lambda[1] = GEOVectorLength(GEOVectorSub(Result, Pos)) / 
	              GEOVectorLength(Dir);
	  if (GEOVectorAngle(GEOVectorSub(Result, Pos), Dir) > 90.0)
	    Lambda[1] *= -1.0;

	  C[No] = Result;               /* contact point */
	  L[No] = Lambda[1];            /* line lambda   */
	  I[No] = count;                /* index no      */
	  E[No] = Lambda[0];            /* edge lambda   */
	}
	No++;
      }
    }
#else
    if (GEOLineIntersection(P1, P1P2, Pos, Dir, 
			    &Result, &Lambda[0], &Lambda[1]))
      if (Lambda[0] >= 0.0 && Lambda[0] <= 1.0)
      {
	if (No < 2)
	{
	  C[No] = Result;               /* contact point */
	  L[No] = Lambda[1];            /* line lambda   */
	  I[No] = count;                /* index no      */
	  E[No] = Lambda[0];            /* edge lambda   */
	}
	No++;
      }
#endif
  }

  PRO1 ("Line-Intersections: %d\n", No)

  /* Sort results for line lambda values */
  if (No > 1 && L[0] > L[1])
  {                                     /* The values have to be swapped */
    first  = 1;
    second = 0;
  }

  /* Assign result values if specified */
  if (Point1 != NULL)
    *Point1 = C[first];
  if (Point2 != NULL)
    *Point2 = C[second];
  if (LLambda1 != NULL)
    *LLambda1 = L[first];
  if (LLambda2 != NULL)
    *LLambda2 = L[second];
  if (Index1 != NULL)
    *Index1 = I[first];
  if (Index2 != NULL)
    *Index2 = I[second];
  if (ELambda1 != NULL)
    *ELambda1 = E[first];
  if (ELambda2 != NULL)
    *ELambda2 = E[second];

  PRO_END
  return No;
}


dual GEOLineIntersection (POINT3D Pos1,   VECTOR3D Dir1,
			  POINT3D Pos2,   VECTOR3D Dir2, 
			  POINT3D *Point, double *Lambda, double *mu)
{
  LGS lp;

#if 0
  WRITE_VECTOR3D (stdout, Pos1, "Pos1");
  WRITE_VECTOR3D (stdout, Dir1, "Dir1");
  WRITE_VECTOR3D (stdout, Pos2, "Pos2");
  WRITE_VECTOR3D (stdout, Dir2, "Dir2");
#endif

  lp.Size = 3;
  lp.Field[0][0] =  Dir1.x;
  lp.Field[0][1] = -Dir2.x;
  lp.Field[0][2] =  0.0;
  lp.Field[0][3] =  Pos2.x - Pos1.x;
  lp.Field[1][0] =  Dir1.y;
  lp.Field[1][1] = -Dir2.y;
  lp.Field[1][2] =  0.0;
  lp.Field[1][3] =  Pos2.y - Pos1.y;
  lp.Field[2][0] =  Dir1.z;
  lp.Field[2][1] = -Dir2.z;
  lp.Field[2][2] =  0.0;
  lp.Field[2][3] =  Pos2.z - Pos1.z;

#if 0
  GEOShowLGS ("LGS", lp);
#endif

  lp = GEOSolveLGS (lp);

  if (lp.State >= 0)
  {
    if (Point != NULL)
      *Point = GEOVectorAdd (Pos1, GEOVectorMult(lp.Field[0][1], Dir1));
#if 0
    WRITE_VECTOR3D (stdout, *Point, "Contact");
#endif

    if (Lambda != NULL)
      *Lambda = lp.Field[0][1];

    if (mu != NULL)
      *mu = lp.Field[0][2];

    PRO5 ("-> %.3f %.3f %.3f  %.3f  %.3f\n",
	  Point->x, Point->y, Point->z, *Lambda, *mu)
  }

  return (lp.State >= 0);
}


/* ********************************************************************** */
/* ** Compute intersection of two areas if it exist. Returns TRUE on   ** */
/* ** success and FALSE, if both both areas do not intersect           ** */
/* ** Idea: solve intersection for the areas projected into each of    ** */
/* ** the three areas of the coordinate-system (xy, xz and yz) by      ** */
/* ** setting the appropriate coordinate to zero. If there are         ** */
/* ** intersections in two areas, both points describe the             ** */
/* ** intersection-line of the areas. If only one intersection was     ** */
/* ** found, the direction of the intersection-line is perpendicular   ** */
/* ** to the area the intersection was found in.                       ** */
/* ********************************************************************** */
dual GEOAreaIntersection (VECTOR3D n1, double a1, VECTOR3D n2, double a2,
			  POINT3D *Point, VECTOR3D *Dir)
{
  int      No = 0;
  VECTOR3D p[2], Plumb = Vector_0;

  PRO_START ("")

  p[0] = Vector_0;
  p[1] = Vector_0;

  /* Compute intersection projected in xy-area */
  if (n1.x != n2.x || n1.y != n2.y)
  {
    /* 
       Set z-coordinate to zero and solve the linear program
         x1 + y1 + a1 = 0      <=>  x1 + y1 = -a1
         x2 + y2 + a2 = 0      <=>  x2 + y2 = -a2
    */
    if (GEOSolve2DimLGS(n1.x, n1.y, -a1, n2.x, n2.y, -a2, 
			&p[No].x, &p[No].y) == 1)
    {
      No++;
      Plumb = GEOMakeVector (0.0, 0.0, 1.0);
    }
#ifdef DEBUG_CSG
    else
      printf ("ERROR SOLVING XY-SYSTEM\n");
#endif
  }

  /* Same for intersection in xz-area */
  if (n1.x != n2.x || n1.z != n2.z)
  {
    if (GEOSolve2DimLGS(n1.x, n1.z, -a1, n2.x, n2.z, -a2,
			&p[No].x, &p[No].z) == 1)
    {
      No++;
      Plumb = GEOMakeVector (0.0, 1.0, 0.0);
    }
#ifdef DEBUG_CSG
    else
      printf ("ERROR SOLVING XZ-SYSTEM\n");
#endif
  }

  if (No < 2)    /* repeat for intersection in xz-area */
    if (n1.y != n2.y || n1.z != n2.z)
    {
      if (GEOSolve2DimLGS(n1.y, n1.z, -a1, n2.y, n2.z, -a2,
			  &p[No].y, &p[No].z) == 1)
      {
	No++;
	Plumb = GEOMakeVector (1.0, 0.0, 0.0);
      }
#ifdef DEBUG_CSG
      else
	printf ("ERROR SOLVING XZ-SYSTEM\n");
#endif
    }

  *Point = p[0];

  PRO1 ("Area-Intersections: %d\n", No)

  if (No == 1)
    *Dir = Plumb;
  elsif (No == 2)
    *Dir = GEOVectorNorm (GEOVectorSub(p[1], p[0]));

  PRO_END
  return (No > 0);
}


/* ********************************************************************** */
/* GEO_PolygonEdgeIntersection                                         ** */
/* ** Computes intersection of an polygon and an edge if both objects  ** */
/* ** lies in the same plan                                            ** */
/* ********************************************************************** */
static int GEO_PolygonEdgeIntersection (POLYGON *Polygon, 
					POINT3D Pos1,    POINT3D  Pos2,
					POINT3D *Point1, POINT3D *Point2)
{
  int      No = 0, count;
  double   lambda, mu;
  POINT3D *Points;
  POINT3D  P1, P2;
  POINT3D  Result = Vector_0;

  Points = Polygon->Points;
  P2 = Points[Polygon->Vertex[0]];
  for (count = 0; count < Polygon->Number; ++count)
  {
    P1 = P2;
    P2 = Points[Polygon->Vertex[ABS_MOD(count+1, Polygon->Number)]];
    if (GEOLineIntersection(P1,   GEOVectorSub(P2, P1),
			    Pos1, GEOVectorSub(Pos2, Pos1),
			    &Result, &lambda, &mu))
    {
      if (lambda >= 0.0 && lambda <= 1.0 && mu >= 0.0 && mu <= 1.0)
      {
	No++;
	if (No == 1 && Point1 != NULL)
	  *Point1 = Result;
	elsif (No == 2 && Point2 != NULL)
	  *Point2 = Result;
      }
      else
      {
	WRITE_VECTOR3D (stdout, P1,   "P1");
	WRITE_VECTOR3D (stdout, P2,   "P2");
	WRITE_VECTOR3D (stdout, Pos1, "P3");
	WRITE_VECTOR3D (stdout, Pos2, "P4");
	PRO2 ("LAMBDA: %.4f  MU: %.4f\n", lambda, mu)
      }
    }
  }
#ifdef DEBUG_CSG
  printf ("Edge-Intersections: %d\n", No);
#endif

  return No;
}


/* ********************************************************************** */
/* ** GEOPolygonEdgeIntersection                                       ** */
/* ** Compute intersection between a polygon and an edge. The result   ** */
/* ** can be eighter 0 (no intersection), 1 (the edge is pushing       ** */
/* ** through the polygon) or 2 (both objects lies in the same plane)  ** */
/* ** Algorithm:                                                       ** */
/* ** 1. Compute intersection between plane and line                   ** */
/* ** 2. Is there a cutton point:                                      ** */
/* **    a) Yes: Test, if the cutting-point lies on the edge and in    ** */
/* **            the polygon                                           ** */
/* **    b) No:  The objects are collinear. Test, if they have the     ** */
/* **            same plane. If so, compute the intersection points.   ** */
/* ********************************************************************** */
int GEOPolygonEdgeIntersection (POLYGON *Polygon, POINT3D v1, POINT3D v2,
				POINT3D *Point1,  POINT3D *Point2)
{
  int      No;
  POINT3D  Point;
  VECTOR3D Dir;

  /* Algorthm:
     1. berechne Schnitt zwischen Ebene und Geraede
     2. Pruefe, ob ein evtl. Schnittpunkt im Polygon liegt.
  */

  /* Is the edge collinear to the polygon's plane ? */
  Dir = GEOVectorSub (v2, v1);
  if (GEOFaceLinePoint(Polygon->Info.n, Polygon->Info.a, v1, Dir, &Point))
  {
    if (GEOPointOnEdge(v1, v2, Point) == 1 && 
	GEOPointInPolygon(Polygon, Point))
    {
      if (Point1 != NULL)
	*Point1 = Point;
      No = 1;
    }
    else
      No = 0;

    if (Point2 != NULL)
      *Point2 = Vector_0;
  }
  else                                      /* Edge and plane are complanar  */
  {
    if (GEOAreaPointDist(Polygon, v1) > 0.0) /* complanar but not same plane */
      No = 0;
    else
      No = GEO_PolygonEdgeIntersection (Polygon, v1, v2, Point1, Point2);
  }

  return No;
}

