#include <math.h>

#include "libgeo.h"


/* ********************************************************************** */
/* ** Prints the elements of the given linear program onto the screen  ** */
/* ** INPUT:  Name  Title of output                                    ** */
/* **         M     The matrix of the linear program                   ** */
/* ** OUTPUT: ---                                                      ** */
/* ********************************************************************** */ 
void GEOShowLGS (char *Name, LGS M)
{
  int row, col;

  printf ("%s:\n", Name);
  for (row = 0; row < M.Size; ++row)
  {
    for (col = 0; col < M.Size; ++col)
      printf ("%9.4e  ", M.Field[row][col]);
    printf ("| %9.4e\n", M.Field[row][M.Size]);
  } 
}


/* *********************************************************************** */
/* ** Corrects the values of the linear program. We suppose, that all   ** */
/* ** values smaller than EPSILON should be zero.                       ** */
/* ** INOUT:  M  Matrix, whose elements should be corrected.            ** */
/* *********************************************************************** */
static void GEO_CorrectLGS (LGS *M)
{
  int row, col;

  for (row = 0; row < M->Size; ++row)
    for (col = 0; col <= M->Size; ++col)
      if (fabs(M->Field[row][col]) < LGS_EPSILON)
	M->Field[row][col] = 0.0;
}


/* ********************************************************************** */
/* ** Performs a gaussian elemination on the elements of the specified ** */
/* ** linear program.                                                  ** */
/* ********************************************************************** */
static void GEO_Gauss (LGS *M)
{
  int     k, i, j, l, c;
  double  swap;
  double  Q[MAX_LGS];

  k = 0;
  do
  {
    l = k;
    while (M->Field[k][k] == 0.0)
    {
      l++;
      if (l > M->Size-1)
	return;

      /* swap line l and k */
      for (c = 0; c <= M->Size; ++c)
      {
	swap           = M->Field[k][c];
	M->Field[k][c] = M->Field[l][c];
	M->Field[l][c] = swap;
      }
    }

    for (i = k+1; i <= M->Size; ++i)
    {
      Q[i] = M->Field[i][k] / M->Field[k][k];
      M->Field[i][k] = 0.0;
      for (j = k+1; j <= M->Size; ++j)
	M->Field[i][j] -= Q[i] * M->Field[k][j];
    }
    k++;
  } while (k < M->Size);
}


/* ********************************************************************** */
/* ** Check the reduced linear program for valid solutions.            ** */
/* ********************************************************************** */
static int GEO_CheckLGS (LGS M)
{
  int Result = 0;
  int row, col;

  if (M.Field[M.Size-1][M.Size-1] == 0.0)
  {
    for (row = 0; row < M.Size; ++row)
    {
      /* check, if whole line is zero */
      col = 0;
      while (col <= M.Size && M.Field[row][col] == 0.0)
	col++;

      if (col == M.Size+1)
	Result++;
      elsif (col == M.Size)            /* There's no x like: 0 * x != 0  */
	return -1;
    }
    if (Result == 0)
      return -1;
  }

  return Result;
}


/* ********************************************************************** */
/* ** Solves the linear program which was defined by the parameter     ** */
/* ** 'M'. The computed function corresponds to M * x = y              ** */
/* ** The returned matrix contains the solution in the form            ** */
/* ** x = v1 + r * v2 + s * v3 + ...                                   ** */
/* ** The 'State'-parameter inside this structure specifies the amount ** */
/* ** of valid results:                                                ** */
/* **  0: exactly one solution,                                        ** */
/* **  2: no valid results                                             ** */
/* ** <0: corresponds to the number of valid vectors. The vectors are  ** */
/* **     stored in the rows of the matrix.                            ** */
/* ** ATTENTION: DO NOT FORGET TO INITIALIZE THE MATRIX-SIZE !         ** */
/* ** INPUT:  M     Matrix of the linear program                       ** */
/* ** OUTPUT: (LGS) Solution of the linear program                     ** */
/* ********************************************************************** */
LGS GEOSolveLGS (LGS M)
{
  int    row, col, k, Index, count;
  double s;
  LGS    V;

  V.Size = M.Size;
  for (row = 0; row < M.Size; ++row)    /* Initialize solution-matrix     */
    for (col = 0; col <= M.Size; ++col)
      V.Field[row][col] = 0.0;

  GEO_Gauss (&M);                       /* Reduce matrix                  */
  GEO_CorrectLGS (&M);

  V.State = GEO_CheckLGS (M);           /* Check for valid results        */

  if (V.State > 0)
  {
    Index = V.State;
    count = 0;
    for (col = M.Size; col >= M.Size-Index+1; col--)
    {
      for (row = 0; row <= Index; ++row)
	V.Field[row][col] = 0.0;
      count++;
      V.Field[count][col] = 1.0;
    }

    for (col = M.Size-Index; col >= 1; col--)
    {
      for (k = 0; k <= Index; ++k)
      {
	V.Field[k][col] = 0.0;
	for (row = col+1; row <= M.Size; ++row)
	  V.Field[k][col] -= 
	    M.Field[col-1][row-1] * V.Field[k][row] / M.Field[col-1][col-1];
      }
      V.Field[0][col] += M.Field[col-1][M.Size] / M.Field[col-1][col-1];
    }
  }    
  elsif (V.State != -1)
  {
    V.Field[0][M.Size-1] = 
      M.Field[M.Size-1][M.Size] / M.Field[M.Size-1][M.Size-1];
    for (col = M.Size-2; col >= 0; col--)
    {
      s = 0;
      for (row = 0; row < M.Size-col+1; ++row)
	s += M.Field[col][col+row] * V.Field[0][col+row];
      V.Field[0][col] = (1.0 / M.Field[col][col]) * (M.Field[col][M.Size]-s);
    }
  }

  return V;
}


/* ********************************************************************** */
/* ** Prints the solution of a solved linear program.                  ** */
/* ** INPUT:  M    Matrix of a linear program.                         ** */
/* ** OUTPUT  ---                                                      ** */
/* ********************************************************************** */
void GEOPrintLGS (LGS M)
{
  int  c, i, j;

  if (M.State == -1)
    printf ("There's no valid solution: %d !\n", M.State);
  else if (M.State > 0)
  {
    printf ("Solutions are:\n");
    for (i = 0; i <= M.State; ++i)
      printf (" %10.6f     ", M.Field[i][1]);
    printf ("\n");

    for (i = 0; i <= M.State-1; ++i)
      printf (" %10.6f + %c ", M.Field[i][2], 'r'+i);
    printf (" %10.6f\n", M.Field[-M.State][2]);

    for (i = 3; i <= M.Size; ++i)
    {
      for (j = 0; j <= M.State; ++j)
	printf (" %10.6f     ", M.Field[j][i]);
      printf ("\n");
    } 
  }
  else if (M.State == 0)
  {
    printf ("Solution is:\n");
    for (c = 0; c < M.Size; ++c)
      printf ("%10.6f   ", M.Field[0][c]);  
    printf ("\n");
  }
}


/* Solves the two-dimensional linear program. The result will be returned
   in x and y. The return-value has the following meaning:
   -1: there's no resolution for the linear program
    0: any values for x/y fits the program
    1: exactly the returned values will solve the program 
    (see J.Khler, R.Hwelmann, H.Krmer: 
      'Analytische Geometrie in vekt. Darstellung', p12/13)
*/
int GEOSolve2DimLGS (double a1, double b1, double c1,
		     double a2, double b2, double c2,
		     double *a, double *b)
{
  double DetAB;
  double DetBC, DetAC;

  /* Correct values */
#if 1
  if (fabs(a1) < LGS_EPSILON)
    a1 = 0.0;
  if (fabs(b1) < LGS_EPSILON)
    b1 = 0.0;
  if (fabs(c1) < LGS_EPSILON)
    c1 = 0.0;
  if (fabs(a2) < LGS_EPSILON)
    a2 = 0.0;
  if (fabs(b2) < LGS_EPSILON)
    b2 = 0.0;
  if (fabs(c2) < LGS_EPSILON)
    c2 = 0.0;
#endif

  PRO3 ("%.4e + %.4e = %.4e\n", a1, b1, c1)
  PRO3 ("%.4e + %.4e = %.4e\n", a2, b2, c2)

  DetAB = a1 * b2 - a2 * b1;
  DetBC = c1 * b2 - c2 * b1;
  DetAC = a1 * c2 - a2 * c1;

  if (DetAB == 0.0)
  {
    if (DetBC != 0.0 || DetAC != 0.0) 
      return -1;    /* LGS 'berbestimmt' */
    else
      return 0;     /* LGS 'allgemeingltig' */
  }

  if (a != NULL)
    *a = DetBC / DetAB;

  if (b != NULL)
    *b = DetAC / DetAB;

  return 1;
}
