#include <math.h>

#include "libgeo.h"


/* ************************************************************************* */
/* geoMATRIX.C                                                               */
/* stellt Funktionen zur Erzeugung und Manipulation homogener Transforma-    */
/* tionsmatrizen zur Verfgung.                                              */
/* ------------------------------------------------------------------------- */
/* Von: Stefan Vietinghoff und Heike Ochs-Vietinghoff                        */
/* ************************************************************************* */

H_MATRIX Matrix_0 = 
{
  {
    { 0.0,  0.0,  0.0,  0.0},         /* homogene Nullmatrix */
    { 0.0,  0.0,  0.0,  0.0},
    { 0.0,  0.0,  0.0,  0.0},
    { 0.0,  0.0,  0.0,  1.0}
  }
};

H_MATRIX Matrix_1 = 
{
  {
    { 1.0,  0.0,  0.0,  0.0},         /* Einheitsmatrix      */
    { 0.0,  1.0,  0.0,  0.0},
    { 0.0,  0.0,  1.0,  0.0},
    { 0.0,  0.0,  0.0,  1.0}
  }
};

H_MATRIX REL_M = 
{
  {
    {-1.0,  0.0,  0.0,  0.0},         /* N */
    { 0.0,  1.0,  0.0,  0.0},         /* O */
    { 0.0,  0.0, -1.0,  0.0},         /* A */
    { 0.0,  0.0,  0.0,  1.0}          /* P */
  }
};


RPY RPY_0 = {0.0, 0.0, 0.0, 0.0, 0.0, 0.0};


/* ************************************************************************* */
/* GEOMakeMatrix                                                             */
/* erzeugt aus den Vektoren "n", "o", "a" und "p" die Matrix                 */
/*                    ( n.x  o.x  a.x  p.x )                                 */
/*                    ( n.y  o.y  a.y  p.y )                                 */
/*                    ( n.z  o.z  a.z  p.z )                                 */
/*                    (  0    0    0    1  ).                                */
/* ************************************************************************* */
H_MATRIX GEOMakeMatrix (VECTOR3D n, VECTOR3D o, VECTOR3D a, VECTOR3D p)
{
  H_MATRIX Result = Matrix_0;

  Result.m[N][X] = n.x;
  Result.m[N][Y] = n.y;
  Result.m[N][Z] = n.z;
  Result.m[N][W] = 0.0;

  Result.m[O][X] = o.x;
  Result.m[O][Y] = o.y;
  Result.m[O][Z] = o.z;
  Result.m[O][W] = 0.0;

  Result.m[A][X] = a.x;
  Result.m[A][Y] = a.y;
  Result.m[A][Z] = a.z;
  Result.m[A][W] = 0.0;

  Result.m[P][X] = p.x;
  Result.m[P][Y] = p.y;
  Result.m[P][Z] = p.z;
  Result.m[P][W] = 1.0;

  return Result;
}

/* ************************************************************************* */
/* GEOCrateMatrix                                                            */
/* erzeugt aus uebergegenen Koponentnen die Matrix                           */
/*                    ( nx  ox  ax  px )                                     */
/*                    ( ny  oy  ay  py )                                     */
/*                    ( nz  oz  az  pz )                                     */
/*                    (  0   0   0   1 ).                                    */
/* ************************************************************************* */
H_MATRIX GEOCreateMatrix (double nx, double ny, double nz,
			  double ox, double oy, double oz,
			  double ax, double ay, double az,
			  double px, double py, double pz)
{
  H_MATRIX Result = Matrix_0;

  Result.m[N][X] = nx;
  Result.m[N][Y] = ny;
  Result.m[N][Z] = nz;
  Result.m[N][W] = 0.0;

  Result.m[O][X] = ox;
  Result.m[O][Y] = oy;
  Result.m[O][Z] = oz;
  Result.m[O][W] = 0.0;

  Result.m[A][X] = ax;
  Result.m[A][Y] = ay;
  Result.m[A][Z] = az;
  Result.m[A][W] = 0.0;

  Result.m[P][X] = px;
  Result.m[P][Y] = py;
  Result.m[P][Z] = pz;
  Result.m[P][W] = 1.0;

  return Result;
}


H_MATRIX GEOPtrMatrix (double *Values)
{
  H_MATRIX Result = Matrix_0;

  if (Values == NULL)
    return Result;

  memcpy (Result.m, Values, sizeof(double) * 16);

  return Result;
}


/* ********************************************************************** */
/* ** Normalizes the vectors of the given frame and rearange them, so  ** */
/* ** that they are orthogonal.                                        ** */
/* ********************************************************************** */
H_MATRIX GEONormMatrix (H_MATRIX M)
{
  VECTOR3D n, o, p;

  n = GEOMatrixVector (M, N);
  o = GEOMatrixVector (M, O);
  p = GEOMatrixVector (M, P);

  return GEOMakeRightSys (n, o, p);
}


/* ********************************************************************** */
/* ** Returns a matrix's component (N, O, A or P) as vector            ** */
/* ********************************************************************** */
VECTOR3D GEOMatrixVector (H_MATRIX M, int Component)
{
  VECTOR3D Result;

  if (Component < N || Component > P)
    return Vector_0;

  Result.x = M.m[Component][X];
  Result.y = M.m[Component][Y];
  Result.z = M.m[Component][Z];

  return Result;
}


/* ********************************************************************** */
/* ** GEOSetMatrixVector changes a matrix's component-vector           ** */
/* ********************************************************************** */
H_MATRIX GEOSetMatrixVector (H_MATRIX M, VECTOR3D V, int Component)
{
  H_MATRIX Result;

  if (Component < N || Component > P)
    return Matrix_0;

  Result = M;
  Result.m[Component][X] = V.x;
  Result.m[Component][Y] = V.y;
  Result.m[Component][Z] = V.z;

  return Result;
}


/* ********************************************************************** */
/* ** GEOMatrixMult                                                    ** */
/* ** transformiert den Vektor "V" mit der Matrix "M".                 ** */
/* ********************************************************************** */
VECTOR3D GEOMatrixMult (H_MATRIX M, VECTOR3D Vec)
{
  VECTOR3D Result;

  Result.x =   M.m[N][X] * Vec.x + M.m[O][X] * Vec.y 
             + M.m[A][X] * Vec.z + M.m[P][X];
  Result.y =   M.m[N][Y] * Vec.x + M.m[O][Y] * Vec.y 
             + M.m[A][Y] * Vec.z + M.m[P][Y];
  Result.z =   M.m[N][Z] * Vec.x + M.m[O][Z] * Vec.y 
             + M.m[A][Z] * Vec.z + M.m[P][Z];

  return Result;
}


/* ********************************************************************** */
/* ** GEOMatrixProd                                                    ** */
/* ** berechnet das Produkt der homogenen Matrizen "M1" und "M2".      ** */
/* ********************************************************************** */
H_MATRIX GEOMatrixProd (H_MATRIX M1, H_MATRIX M2)
{
  int      row, col, count;
  H_MATRIX Result = Matrix_0;
  
  for (row = X; row <= Z; ++row)
    for (col = N; col <= P; ++col)
      for (count = 0; count < 3; ++count)  /* 0..3 wg. 0-en in letzter Zeile */
	Result.m[col][row] += M1.m[count][row] * M2.m[col][count];

  Result.m[P][X] += M1.m[P][X];            /* weil Skal.-Faktor = 1          */
  Result.m[P][Y] += M1.m[P][Y];
  Result.m[P][Z] += M1.m[P][Z];

  Result.m[N][W] = 0.0;
  Result.m[O][W] = 0.0;
  Result.m[A][W] = 0.0;
  Result.m[P][W] = 1.0;

  return Result;
}

  
/* ********************************************************************** */
/* ** GEOMatrixInv                                                     ** */
/* ** invertiert die Matrix "M".                                       ** */
/* ********************************************************************** */
H_MATRIX GEOMatrixInv (H_MATRIX M)
{
  H_MATRIX Result = Matrix_0;

  Result.m[N][X] = M.m[N][X];               /* Rotationsanteil transponieren */
  Result.m[N][Y] = M.m[O][X];
  Result.m[N][Z] = M.m[A][X];
  Result.m[N][W] = 0.0;

  Result.m[O][X] = M.m[N][Y];
  Result.m[O][Y] = M.m[O][Y];
  Result.m[O][Z] = M.m[A][Y];
  Result.m[O][W] = 0.0;

  Result.m[A][X] = M.m[N][Z];
  Result.m[A][Y] = M.m[O][Z];
  Result.m[A][Z] = M.m[A][Z];
  Result.m[A][W] = 0.0;

  Result.m[P][X] = - (M.m[P][X] * M.m[N][X] +  /* Positionsanteil berechnen */
		      M.m[P][Y] * M.m[N][Y] +
		      M.m[P][Z] * M.m[N][Z]);
  Result.m[P][Y] = - (M.m[P][X] * M.m[O][X] + 
		      M.m[P][Y] * M.m[O][Y] +
		      M.m[P][Z] * M.m[O][Z]);
  Result.m[P][Z] = - (M.m[P][X] * M.m[A][X] + 
		      M.m[P][Y] * M.m[A][Y] +
		      M.m[P][Z] * M.m[A][Z]);
  Result.m[P][W] = 1.0;

  return Result;
}


/* ********************************************************************** */
/* ** Compare two matrixes. They are assumed as equal, if the N-, O-,  ** */
/* ** A- and P-vectors are very similar (differ in a small value)      ** */
/* ********************************************************************** */
dual GEOMatrixCompare (H_MATRIX M1, H_MATRIX M2)
{
  return (sqrt(SQR(M2.m[N][X] - M1.m[N][X]) +
	       SQR(M2.m[N][Y] - M1.m[N][Y]) +
	       SQR(M2.m[N][Z] - M1.m[N][Z]) < D_DIST) &&
	  sqrt(SQR(M2.m[O][X] - M1.m[O][X]) +
	       SQR(M2.m[O][Y] - M1.m[O][Y]) +
	       SQR(M2.m[O][Z] - M1.m[O][Z]) < D_DIST) &&
	  sqrt(SQR(M2.m[A][X] - M1.m[A][X]) +
	       SQR(M2.m[A][Y] - M1.m[A][Y]) +
	       SQR(M2.m[A][Z] - M1.m[A][Z]) < D_DIST) &&
	  sqrt(SQR(M2.m[P][X] - M1.m[P][X]) +
	       SQR(M2.m[P][Y] - M1.m[P][Y]) +
	       SQR(M2.m[P][Z] - M1.m[O][Z]) < D_DIST));
}


/* ********************************************************************** */
/* ** Creates a rotation-matrix with the given rotation-axis and the   ** */
/* ** specified angle                                                  ** */
/* ** INPUT:  Axis     rotation-axis (X, Y or Z)                       ** */
/* **         Angle    rotation-angle                                  ** */
/* ** OUTPUT: (H_MATRIX) Rotation-Matrix                                 ** */
/* ********************************************************************** */
H_MATRIX GEORotMatrix (int Axis, double Angle)
{
  double c, s;
  H_MATRIX M;

  s = sin(RAD(Angle));
  c = cos(RAD(Angle));

  M = Matrix_1;

  switch (Axis)
  {
    case X : M.m[O][Y] =  c;
             M.m[O][Z] =  s;
	     M.m[A][Y] = -s;
             M.m[A][Z] =  c;
	     break;
    case Y : M.m[N][X] =  c;
             M.m[N][Z] = -s;
             M.m[A][X] =  s;
             M.m[A][Z] =  c;
	     break;
    case Z : M.m[N][X] =  c;
             M.m[N][Y] =  s;
	     M.m[O][X] = -s;
	     M.m[O][Y] =  c;
    default : break;
  }

  return M;
}


/* ********************************************************************** */
/* ** Create a righhand-system from a given N-, O- and P-vector by     ** */
/* ** calculating the appropriate A-vector and norm all vectors        ** */
/* ********************************************************************** */
H_MATRIX GEOMakeRightSys (VECTOR3D n, VECTOR3D o, VECTOR3D p)
{
  VECTOR3D n_norm, o_norm, a_norm;

  n_norm = GEOVectorNorm (n);
  o_norm = GEOVectorNorm (o);

  a_norm = GEOVectorNorm (GEOVectorCross(n_norm, o_norm));
  o_norm = GEOVectorNorm (GEOVectorCross(a_norm, n_norm));

  return GEOMakeMatrix (n_norm, o_norm, a_norm, p);
}


/* ********************************************************************** */
/* ** Needed for symbolic-spartial-relations, because the n-(x-)vector ** */
/* ** have to stay upright onto the plane. So we know the a- and       ** */
/* ** o-vector of this plane                                           ** */
/* ********************************************************************** */
H_MATRIX GEOMakeRightSys2 (VECTOR3D o, VECTOR3D a, VECTOR3D p)
{
  VECTOR3D n_norm, o_norm, a_norm;

  o_norm = GEOVectorNorm (o);
  a_norm = GEOVectorNorm (a);

  n_norm = GEOVectorNorm (GEOVectorCross(o_norm, a_norm));
  a_norm = GEOVectorNorm (GEOVectorCross(n_norm, o_norm));

  return GEOMakeMatrix (n_norm, o_norm, a_norm, p);
}


/* Same as GEOMakeRightSys2, but here, the open-vector will be corrected */
H_MATRIX GEOMakeAOFrame (VECTOR3D o, VECTOR3D a, VECTOR3D p)
{
  VECTOR3D n_norm, o_norm, a_norm;

  o_norm = GEOVectorNorm (o);
  a_norm = GEOVectorNorm (a);

  n_norm = GEOVectorNorm (GEOVectorCross(o_norm, a_norm));
  o_norm = GEOVectorNorm (GEOVectorCross(a_norm, n_norm));

  return GEOMakeMatrix (n_norm, o_norm, a_norm, p);
}


/* ********************************************************************** */
/* ** Creates a valid frame by computing an open- and an               ** */
/* ** approach-vector to the given normal-vector. Then the vectors     ** */
/* ** will be normalized. The Position will be set to origin.          ** */
/* ** INPUT:  n         a normal vector                                ** */
/* ** OUTPUT: (H_MATRIX)  the resulting frame or Matrix_1, if all        ** */
/* **                   componets are zero.                            ** */
/* ********************************************************************** */
H_MATRIX GEOMakeFrame (VECTOR3D n)
{
  VECTOR3D n_norm, o_norm, a_norm;
  VECTOR3D o = {1.0, 1.0, 1.0};

  /* find a vector which stands upright to the given approach-vector      */
  if (n.x != 0.0)
    o.x = - (n.y + n.z) / n.x;
  elsif (n.y != 0.0)
    o.y = - (n.x + n.z) / n.y;
  elsif (n.z != 0.0)
    o.z = - (n.x + n.y) / n.z;
  else
  {
    PRO0 ("Error Invalid normal-vector\n")
    return Matrix_1;
  }

  n_norm = GEOVectorNorm (n);
  o_norm = GEOVectorNorm (o);

  a_norm = GEOVectorNorm (GEOVectorCross(n_norm, o_norm));
  o_norm = GEOVectorNorm (GEOVectorCross(a_norm, n_norm));

  return GEOMakeMatrix (n_norm, o_norm, a_norm, Vector_0);
}


H_MATRIX GEOMakeAFrame (VECTOR3D a)
{
  VECTOR3D n_norm, o_norm, a_norm;
  VECTOR3D o = {1.0, 1.0, 1.0};

  /* find a vector which stands upright to the given approach-vector      */
  if (a.x != 0.0)
    o.x = - (a.y + a.z) / a.x;
  elsif (a.y != 0.0)
    o.y = - (a.x + a.z) / a.y;
  elsif (a.z != 0.0)
    o.z = - (a.x + a.y) / a.z;
  else
  {
    PRO0 ("Error Invalid normal-vector\n")
    return Matrix_1;
  }

  a_norm = GEOVectorNorm (a);
  o_norm = GEOVectorNorm (o);

  n_norm = GEOVectorNorm (GEOVectorCross(a_norm, o_norm));
  o_norm = GEOVectorNorm (GEOVectorCross(n_norm, a_norm));

  return GEOMakeMatrix (n_norm, o_norm, a_norm, Vector_0);
}


/* ********************************************************************** */
/* ** Calculate homogenous transformation-matrix from euler            ** */
/* ** representation Roll, Pitch and Yaw must be given in degrees      ** */
/* ** Fu, Gozales, Li: page 25                                         ** */
/* ********************************************************************** */
H_MATRIX GEORPY2Frame (double px, double py, double pz, 
		       double y,  double p,  double r)
{
  H_MATRIX Result = Matrix_0;
  double   sr, sp, sy;
  double   cr, cp, cy;

  r = RAD (r);
  p = RAD (p);
  y = RAD (y);

  sr = sin (r);
  sp = sin (p);
  sy = sin (y);
  cr = cos (r);
  cp = cos (p);
  cy = cos (y);

  Result.m[N][X] = cr * cp;
  Result.m[N][Y] = sr * cp;
  Result.m[N][Z] = - sp;
  Result.m[N][W] = 0.0;

  Result.m[O][X] = cr * sp * sy - sr * cy;
  Result.m[O][Y] = sr * sp * sy + cr * cy;
  Result.m[O][Z] = cp * sy;
  Result.m[O][W] = 0.0;

  Result.m[A][X] = cr * sp * cy + sr * sy;
  Result.m[A][Y] = sr * sp * cy - cr * sy;
  Result.m[A][Z] = cp * cy;
  Result.m[A][W] = 0.0;

  Result.m[P][X] = px;
  Result.m[P][Y] = py;
  Result.m[P][Z] = pz;
  Result.m[P][W] = 1.0;

  return Result;
}


/* ********************************************************************** */
/* ** compute RPY-notation from transformation-matrix,                 ** */
/* ** R, P and Y in degrees                                            ** */
/* ********************************************************************** */
void GEOFrame2RPY (H_MATRIX M, double *px, double *py, double *pz,
		               double *y,  double *p,  double *r)
{
  double cr, sr;

  if (px != NULL)
    *px = M.m[P][X];
  if (py != NULL)
    *py = M.m[P][Y];
  if (pz != NULL)
    *pz = M.m[P][Z];
  
#if 1
  *r = atan2 (M.m[N][Y], M.m[N][X]);
  sr = sin (*r);
  cr = cos (*r);

  *p = GRAD (atan2(-1.0 * M.m[N][Z], M.m[N][X]*cr + M.m[N][Y]*sr));
  *y = GRAD (atan2(M.m[A][X]*sr - M.m[A][Y]*cr,
		   M.m[O][Y]*cr - M.m[O][X]*sr));
  *r = GRAD (*r);
#else
  if (M.m[N][Y] != 0.0 && M.m[N][X] != 0)
    *r = atan2 (M.m[N][Y], M.m[N][X]);
  else
    *r = PI/2.0;

  if (M.m[N][Z] != 0.0 && M.m[N][X]*cos(*r) + M.m[N][Y]*sin(*r) != 0.0)
    *p = GRAD (atan2(-1.0 * M.m[N][Z], 
		     M.m[N][X]*cos(*r) + M.m[N][Y]*sin(*r)));
  else
    *p = PI/2.0;

  if (M.m[A][X]*sin(*r) - M.m[A][Y]*cos(*r) != 0.0 &&
      M.m[O][Y]*cos(*r) - M.m[O][X]*sin(*r) != 0.0)
    *y = GRAD (atan2(M.m[A][X]*sin(*r) - M.m[A][Y]*cos(*r),
		     M.m[O][Y]*cos(*r) - M.m[O][X]*sin(*r)));
  else
    *y = PI/2.0;

  *r = GRAD (*r);
#endif
}


H_MATRIX GEOEuler2Frame (double px, double py, double pz, 
			 double z1, double y,  double z2)
{
  H_MATRIX Result = Matrix_0;
  double   sz1, sy, sz2;
  double   cz1, cy, cz2;

  z1 = RAD (z1);
  y  = RAD (y);
  z2 = RAD (z2);

  sz1 = sin (z1);
  sy  = sin  (y);
  sz2 = sin (z2);
  cz1 = cos (z1);
  cy  = cos  (y);
  cz2 = cos (z2);

#if 0
  Result.m[N][X] = cz1 * cz2 - sz1 * cy * sz2;
  Result.m[N][Y] = sz1 * cz2 + cz1 * cy * sz2;
  Result.m[N][Z] = sy * sz2;
  Result.m[N][S] = 0.0;

  Result.m[O][X] = -cz1 * sz2 - sz1 * cy * cz2;
  Result.m[O][Y] = -sz1 * sz2 + cz1 * cy * cz2;
  Result.m[O][Z] = sy * cz2;
  Result.m[O][W] = 0.0;

  Result.m[A][X] =  sz1 * sy;
  Result.m[A][Y] = -cz1 * sy;
  Result.m[A][Z] = cy;
  Result.m[A][W] = 0.0;
#else
  Result.m[N][X] = cz1 * cy * cz2 - sz1 * sz2;
  Result.m[N][Y] = sz1 * cy * cz2 + cz1 * sz2;
  Result.m[N][Z] = -sy * cz2;
  Result.m[N][W] = 0.0;

  Result.m[O][X] = -cz1 * cy * sz2 - sz1 * cz2;
  Result.m[O][Y] = -sz1 * cy * sz2 + cz1 * cz2;
  Result.m[O][Z] = sy * sz2;
  Result.m[O][W] = 0.0;

  Result.m[A][X] = cz1 * sy;
  Result.m[A][Y] = sz1 * sy;
  Result.m[A][Z] = cy;
  Result.m[A][W] = 0.0;
#endif

  Result.m[P][X] = px;
  Result.m[P][Y] = py;
  Result.m[P][Z] = pz;
  Result.m[P][W] = 1.0;

  return Result;
}


/* ********************************************************************** */
/* ** Transforms a transformation-frame into the euler orientation     ** */
/* ** INPUT:  M          matrix of transformation frame                ** */
/* ** INOUT:  z1, y, z2  euler orientation                             ** */
/* ** OUTPUT: ---                                                      ** */
/* ********************************************************************** */
void GEOFrame2Euler (H_MATRIX M, double *px, double *py, double *pz,
		                 double *z1, double *y, double *z2)
{
#if 0
  double sz1, cz1;
#endif

  *px = M.m[P][X];
  *py = M.m[P][Y];
  *pz = M.m[P][Z];

#if 1
  /* Paul ? */
  *y = GRAD (atan2(sqrt(SQR(M.m[N][Z]) + SQR(M.m[O][Z])), M.m[A][Z]));
  if (*y == 0.0)
  {
    PRO0 ("[1]\n")
    *z1  = 0.0;
    *z2 = GRAD (atan2(-M.m[O][X], M.m[N][X]));
  }
  elsif (*y == 180.0)
  {
    PRO0 ("[2]\n")
    *z1  = 0.0;
    *z2 = GRAD (atan2(M.m[O][X], -M.m[N][X]));
  }
  else
  {
    PRO0 ("[3]\n")
    *z1 = GRAD (atan2(M.m[A][Y],  M.m[A][X]));
    *z2 = GRAD (atan2(M.m[O][Z], -M.m[N][Z])) + 180.0;
  }
#else
  #if 1
    /* 
       Fehlerhaft, bei Matrix von (yaw: 2, pitch: 0, roll: 0) 
       -> -90, 2, -90 statt -90, 2, 90
    */
    if (M.m[A][X] != 0.0 || M.m[A][Y] != 0.0)
      *z1 = atan2 (M.m[A][Y], M.m[A][X]);
    else 
      *z1 = 0.0;

    /*
      the other solution is: atan2 (-M.m[A][X], M.m[A][Y]);
      which is equivalent to *z1 +/- PI
      if M.m[A][X] = 0 and M.m[A][Y] = 0 then *z1 can have an
      arbitrary value, we choose *z1 = 0.0 (because of the behaviour of atan2)
    */

    sz1 = sin (*z1);
    cz1 = cos (*z1);

    *y  = GRAD (atan2( cz1 * M.m[A][X] + sz1 * M.m[A][Y], M.m[A][Z]));
    *z2 = GRAD (atan2( cz1 * M.m[O][X] + sz1 * M.m[O][Y],
	 	      -cz1 * M.m[N][X] - sz1 * M.m[N][Y]));
    *z1 = GRAD (*z1);
  #else    /* Fu, Gonzales, Li */
    *z1 = atan2 (M.m[A][X], -M.m[A][Y]);

    sz1 = sin (*z1);
    cz1 = cos (*z1);

    *z2 = GRAD (atan2(-cz1 * M.m[O][X] - sz1 * M.m[O][Y], 
	 	       cz1 * M.m[N][X] + sz1 * M.m[N][Y]));
    *y  = GRAD (atan2( sz1 * M.m[A][X] - cz1 * M.m[A][Y], M.m[A][Z]));
  #endif
#endif
}


void GEORPY2Euler (RPY Pos, EULER *Euler)
{
  EULER    Result;
  H_MATRIX M;

  M = GEORPY2Frame (Pos.px, Pos.py, Pos.pz, Pos.r, Pos.p, Pos.y);
  if (Euler != NULL)
    GEOFrame2Euler (M, &Result.px, &Result.py, &Result.pz, 
		    &Result.z1, &Result.y,  &Result.z2);
}


RPY GEOGetRPY (H_MATRIX M)
{
  RPY Result;

  GEOFrame2RPY (M, &Result.px, &Result.py, &Result.pz, 
		   &Result.y, &Result.p, &Result.r);

  return Result;
}

