#include <math.h>

#include "libgeo.h"
/* ************************************************************************* */
/* geoOps.c                                                                  */
/* stellt Funktionen zur Berechnung von Polyedereigenschaften zur Verfgung. */
/* ------------------------------------------------------------------------- */
/* Von            : Stefan Vietinghoff und Heike Ochs Vietinghoff            */
/* letzte nderung: 23.06.1997 von Heike                                     */
/* ************************************************************************* */


/* ************************************************************************* */
/* GEOTetreaderVolume                                                        */
/* berechnet das Volumen des Tetraeders mit der durch die Punkte p1, p2 und  */
/* p3 gegebenen Grundflche. Die Spitze des Tetraeders liegt im Ursprung.    */
/* Auerdem wird der Flcheninhalt der Grundflche des Tetraeders berechnet. */
/* Fr die Berechnung  wird die Formel zur Flcheninhaltsberechnung von Gau */
/* benutzt.                                                                  */
/* Dient als Hilfsfunktion fr GEOGalctetraeder.                             */
/* Literatur: c't Heft 2 1988, "Von Vektoren und Volumina" S.144 ff          */
/* ************************************************************************* */
void GEOTetraederVolume (POINT3D p1, POINT3D p2, POINT3D p3,
			 double *Area, double *Volume)
{
  double Gauss[3];               /* Parameter fr Gauss-Formel. Dienen zur   */
                                 /* Volumen und Flcheninhaltsberechnung     */

/* Berechnung des Inhalts der Grundflche. Diese ist durch die Lnge des     */
/* Flchennormalevektors "Gauss" gegeben. Die Lnge und Richtung der Kompo-  */
/* nenten von "Gauss" berechnent man, indem man die Grundflche auf die xy-, */
/* xz- und yz-Ebene projeziert. Die Flcheninhalte der projezierten berech-  */
/* net man mit                                                               */
/* Gauss[1] = 0.5 * [ Summe ber alle Punkte pi : (pi.y-pj.y)*(pi.z-pj.z) ]  */
/* Gauss[2] = 0.5 * [ Summe ber alle Punkte pi : (pi.z-pj.z)*(pi.x-pj.x) ]  */
/* Gauss[3] = 0.5 * [ Summe ber alle Punkte pi : (pi.x-pj.x)*(pi.y-pj.y) ]  */
/* fr j= (i+1) mod n mit n = Anzahl der Ecken  des Polygons; hier also n=3. */
/* Fr die Lnge von "Gauss", also den Flcheninhalt des Polygons gilt:      */
/* Area = Wurzel aus (Gauss[0]^2 + Gauss[1]^2 + Gauss[2]^2)                  */
 
  Gauss[0] = 0.5 * ((p1.y - p2.y) * (p1.z + p2.z) +
		    (p2.y - p3.y) * (p2.z + p3.z) +
		    (p3.y - p1.y) * (p3.z + p1.z));
  Gauss[1] = 0.5 * ((p1.z - p2.z) * (p1.x + p2.x) +
		    (p2.z - p3.z) * (p2.x + p3.x) +
		    (p3.z - p1.z) * (p3.x + p1.x));
  Gauss[2] = 0.5 * ((p1.x - p2.x) * (p1.y + p2.y) +
		    (p2.x - p3.x) * (p2.y + p3.y) +
		    (p3.x - p1.x) * (p3.y + p1.y));
    
  if (Area != NULL)
    *Area = sqrt (SQR(Gauss[0]) + SQR(Gauss[1]) + SQR(Gauss[2]));

/* Volumen = Grundflche * Hhe. Gauss ist der normierte Flchennormalenvek- */
/* tor der Ebene, in der das Polygon liegt, multipliziert mit dem Fchen-    */
/* inhalt des Polygons. Daraus folgt:                                        */

  if (Volume != NULL)
    *Volume = ((Gauss[0]*p1.x) + (Gauss[1]*p1.y) + (Gauss[2]*p1.z)) / 3.0;
}
  

/* ********************************************************************** */
/* ** GEOTetraederCenter                                               ** */
/* ** bestimmt den Schwerpunkt des Tetraeders mit der durch "p1", "p2" ** */
/* ** und "p3" Grundflche.                                            ** */
/* ********************************************************************** */
POINT3D GEOTetraederCenter (POINT3D p1, POINT3D p2, POINT3D p3)
{
  POINT3D Result;

  Result.x = 0.25 * (p1.x + p2.x + p3.x);
  Result.y = 0.25 * (p1.y + p2.y + p3.y);
  Result.z = 0.25 * (p1.z + p2.z + p3.z);

  return Result;
}


/* ********************************************************************** */
/* ** Computes the distance between two points.                        ** */
/* ** INPUT:  Pos1, Pos2  Both points                                  ** */
/* ** OUTPUT: (double)    length of the distance.                      ** */
/* ********************************************************************** */
double GEOPointPointDist (POINT3D Pos1, POINT3D Pos2)
{
  return sqrt(SQR(Pos2.x-Pos1.x) + SQR(Pos2.y-Pos1.y) + SQR(Pos2.z-Pos1.z));
}


/* ********************************************************************** */
/* ** Compute the Point's distance to a line which is given through    ** */
/* ** Line = Pos + lambda * Dir                                        ** */
/* ********************************************************************** */
double GEOLinePointDist (POINT3D Pos, VECTOR3D Dir, POINT3D Point)
{
  double   Result;
  VECTOR3D r, u, e;

  r = GEOVectorSub  (Point, Pos);
  u = GEOVectorNorm (Dir);

  e = GEOVectorCross (u, r);

  Result = VEC_LEN (e);

  return Result;
}


/* ********************************************************************** */
/* ** Compute the distance between a point and a polygon.              ** */
/* ** The result respects the sign of the distance !                   ** */
/* ********************************************************************** */
double GEOAreaPointDist (POLYGON *Polygon, POINT3D Point)
{
  double Result;

  if (Polygon == NULL)
    return 0.0;
 
  /* Vector product of normal and point plus distance */
  Result =  (  Polygon->Info.n.x * Point.x
	     + Polygon->Info.n.y * Point.y
	     + Polygon->Info.n.z * Point.z) + Polygon->Info.a;

  return Result;
}


/* ********************************************************************** */
/* ** Computes the minimum dstance between two lines which are given   ** */
/* ** in the form g1 = Pos1 + r * Dir1   and  g2 = Pos2 + s * Dir2     ** */
/* ** INPUT:  Pos1, Dir1  First line                                   ** */
/* **         Pos2, Dir2  Second line                                  ** */
/* ** OUTPUT: (double)    absolute distance between both lines.        ** */
/* ********************************************************************** */
double GEOLineLineDist (POINT3D Pos1, VECTOR3D Dir1, 
			POINT3D Pos2, VECTOR3D Dir2)
{
  double Det, Div;
  double Result = 0.0;

  Det =   ((Pos1.x - Pos2.x) * Dir1.y * Dir2.z)
        + ((Pos1.y - Pos2.y) * Dir1.z * Dir2.x)
        + ((Pos1.z - Pos2.z) * Dir1.x * Dir2.y)
        - (Dir2.x * Dir1.y * (Pos1.z - Pos2.z))
        - (Dir2.y * Dir1.z * (Pos1.x - Pos2.x))
        - (Dir2.z * Dir1.x * (Pos1.y - Pos2.y));

  if (Det == 0.0)
    return 0.0;                         /* lines are crossing each other */

  Div = sqrt (  SQR((Dir1.x * Dir2.y) - (Dir2.x * Dir1.y))
	      + SQR((Dir1.y * Dir2.z) - (Dir2.y * Dir1.z))
	      + SQR((Dir1.z * Dir2.x) - (Dir2.z * Dir1.x)));

  if (Div == 0.0)
    return 0.0;

  Result = fabs (Det / Div);

  return Result;
}


/* ********************************************************************** */
/* ** Computes the point on the given line, with the nearest distance  ** */
/* ** to the point p. ('Lotfusspunkt').                                ** */
/* ** INPUT:  Pos, Dir  Line given in the form g = Pos + r * Dir       ** */
/* **         p         The given point                                ** */
/* ** OUTPUT: (dual)    TRUE, if the nearest point was found.          ** */
/* **                   FALSE, if a bad line-direction was specified   ** */
/* **         Point     Line-point with minimum distance to p          ** */
/* ********************************************************************** */
dual GEOLinePointLot (POINT3D Pos, VECTOR3D Dir, POINT3D p, POINT3D *Point)
{
  double Factor, square;

  if ((square = SQR(Dir.x) + SQR(Dir.y) + SQR(Dir.z)) == 0.0)
    return FALSE;

  Factor = GEOVectorProd (GEOVectorSub(p, Pos), Dir) / square;

  *Point = GEOVectorAdd (Pos, GEOVectorMult(Factor, Dir));

  return TRUE;
} 


/* ********************************************************************** */
/* ** Computes the points on both lines, which have the minimum        ** */
/* ** distance to each other. ('Lotfusspunkte')                        ** */
/* ** INPUT:  Pos1, Dir1  First line in the form g = Pos1 + r * Dir1   ** */
/* **         Pos2, Dir2  Second line (analogy)                        ** */
/* ** OUTPUT: (dual)      success of computation (tells if Point1 and  ** */
/* **                     Point2 are valid)                            ** */
/* **         Point1                                                   ** */
/* **         Point2      Point on first/second line                   ** */
/* ********************************************************************** */
dual GEOLineLineLot (POINT3D Pos1, VECTOR3D Dir1, POINT3D Pos2, VECTOR3D Dir2,
		     POINT3D *Point1, POINT3D *Point2)
{
  VECTOR3D h, i;
  double   r = 0, s = 0;

  h.x = GEOVectorProd (GEOVectorSub(Pos1, Pos2), Dir1);
  h.y = GEOVectorProd (Dir1, Dir1);
  h.z = GEOVectorProd (Dir2, Dir1);

  i.x = GEOVectorProd (GEOVectorSub(Pos1, Pos2), Dir2);
  i.y = GEOVectorProd (Dir1, Dir2);
  i.z = GEOVectorProd (Dir2, Dir2);

  /* Compute the position of the 'Lotfusspunkt' */
  if ((h.z*i.y - h.y*i.z) != 0.0)
  {
    r = (h.x*i.z - h.z*i.x) / (h.z*i.y - h.y*i.z);
    s = (h.x*i.y - h.y*i.x) / (h.z*i.y - h.y*i.z);
  }
  else
  {
    PRO0 ("Division by zero for r and s")
    return FALSE;
  }

  if (Point1 != NULL)
    *Point1 = GEOVectorAdd (Pos1, GEOVectorMult(r, Dir1));

  if (Point2 != NULL)
    *Point2 = GEOVectorAdd (Pos2, GEOVectorMult(s, Dir2));

#if 0
  if (fabs(VEC_LEN(GEOVectorSub(v[0], v[2]))) > D_DIST)
    PRO_RETURN ("Vector not zero", FALSE)

  /* Compute first axis-point and the axis-direction of the               */
  Result = GEOVectorMult (0.5, GEOVectorAdd(v[0], v[1]));
#endif

  return TRUE;
}



/* ********************************************************************** */
/* ** Calculates the cutting point between a plane and a line.         ** */
/* ** INPUT:  n, a      Plane given in "Hesse'sche normal form"        ** */
/* **         Pos, Dir  Line in the form g = Pos + r * Dir             ** */
/* ** OUTPUT: (dual)    TRUE: intersection was computed successfully   ** */
/* **                   FALSE: the line is collinear to the plane      ** */
/* **         Point     The computed intersection-point                ** */
/* ********************************************************************** */
dual GEOFaceLinePoint (VECTOR3D n, double a, POINT3D Pos, VECTOR3D Dir, 
		       POINT3D *Point)
{
  double Factor, Div;

  if ((Div = GEOVectorProd(n, Dir)) == 0.0)
    return FALSE;

  if (Point != NULL)
  {
    Factor = (- a - GEOVectorProd(n, Pos)) / Div;
    *Point = GEOVectorAdd (Pos, GEOVectorMult(Factor, Dir));
  }

  return TRUE;
}


/* ********************************************************************** */
/* ** Computes the intersection-points between a line and a bubble.    ** */
/* ** the formula is: (bubble is centered in origin)                   ** */
/* **          PM * d      1          +------------------------------- ** */
/* ** l    = - ------ +/- --- * -    /          2     2    2    2   2  ** */
/* **  1/2        2         2    \  /   (PM * d)  - PM  * d  + r * d   ** */
/* **            d         d      \/                                   ** */
/* ** found in: Khler, Hwelmann, Krmer: page 156                    ** */
/* **   "Analytische Geometrie in vektorieller Darstellung"            ** */
/* ** INPUT:  M         Center of bubble                               ** */
/* **         r         radius of bubble                               ** */
/* **         Pos       local-vector of line                           ** */
/* **         d         direction of line                              ** */
/* ** OUTPUT  lambda1/2 cutting points: s1/s2 = Pos + lambda1/2 * d    ** */
/* **         (int)     0: no cutting-point                            ** */
/* **                   1: one cut (tangent)                           ** */
/* **                   2: two cuting-points                           ** */
/* ********************************************************************** */
int GEOComputeBubbleCut (POINT3D M, double r, POINT3D Pos, VECTOR3D d,
			 double *lambda1, double *lambda2)
{
  double  Result[2];
  double  value, mpd;
  double  root, dd;
  POINT3D MP;

  PRO3 ("M:   %f  %f  %f\n", M.x, M.y, M.z);
  PRO1 ("r:   %f\n", r);
  PRO3 ("Pos: %f  %f  %f\n", Pos.x, Pos.y, Pos.z);
  PRO3 ("Dir: %f  %f  %f\n", d.x, d.y, d.z);

  MP = GEOVectorSub (Pos, M);
  dd = GEOVectorSquare(d);

  /* compute radicant */
  mpd = GEOVectorProd (MP, d);
  root = mpd * mpd - GEOVectorSquare(MP) * dd + r * r * dd;

  if (root < 0.0)
    return 0;

  root = sqrt (root);

  value = - GEOVectorProd(MP, d) / dd;
  Result[0] = value + 1.0 / dd * root;
  Result[1] = value - 1.0 / dd * root;

  PRO2 ("=> %f  %f\n", Result[0], Result[1]);

  /* Cuts are: Pos + Result[0/1] * d */
  if (lambda1 != NULL)
    *lambda1 = Result[0];

  if (lambda2 != NULL)
  *lambda2 = Result[1];

  if (root == 0.0)
    return 1;
  else
    return 2;
}

