#include <math.h>

#include "libgeo.h"
/* ************************************************************************* */
/* POLYGON.C                                                                 */
/* enthlt Funktionen zum Erzeugen und Verndern von Polygonen.              */
/* ------------------------------------------------------------------------- */
/* Von            : Stefan Vietinghoff und Heike Ochs-Vietinghoff            */
/* Letzte nderung: 14.Jul 1997                                              */
/* ************************************************************************* */


/* ************************************************************************* */
/* GEONewPolygon                                                             */
/* erzeugt leeres Polygon.                                                   */
/* ************************************************************************* */
POLYGON *GEONewPolygon (void)
{
  POLYGON *Result;

  Result = (POLYGON *)malloc (sizeof(POLYGON));

  Result->Number = 0;
  Result->Vertex = NULL;
  Result->Points = NULL;

  Result->Normals = NULL;

  Result->Triangle = NULL;

  Result->Info.ID     = -1;
  Result->Info.Center = Vector_0;
  Result->Info.Area   = 0.0;
  Result->Info.n      = Vector_0;
  Result->Info.a      = 0.0;

  Result->Flags       = 0;

  Result->Next = NULL;

  return Result;
}


/* ************************************************************************* */
/* GEOCopyPolygon                                                            */
/* Copy polygon-data (only pointer to point-data - not data itself).         */
/* ************************************************************************* */
POLYGON *GEOCopyPolygon (POLYGON *Polygon)
{
  POLYGON *Result;

  if (Polygon == NULL)
    return NULL;

  Result = GEONewPolygon ();

  Result->Number = Polygon->Number;
  Result->Vertex = (int *)malloc (sizeof(int) * Polygon->Number);
  memcpy (Result->Vertex, Polygon->Vertex, sizeof(int) * Polygon->Number);

  /* assign the polygon points, reassign on retur, if needed */
  Result->Points = Polygon->Points;

  if (Polygon->Normals != NULL)
  {
    Result->Normals = (VECTOR3D *)malloc (sizeof(VECTOR3D) * Polygon->Number);
    memcpy (Result->Normals, Polygon->Normals, 
                             sizeof(VECTOR3D) * Polygon->Number);
  }

  if (Polygon->Triangle != NULL)
  {
    Result->Triangle = (TRIANGLE *)malloc (sizeof(TRIANGLE) * 
					   (Polygon->Number-2));
    memcpy (Result->Triangle, Polygon->Triangle, 
	    sizeof(TRIANGLE)*(Polygon->Number-2));
  }
  else
    Result->Triangle = NULL;

  Result->Info = Polygon->Info;

  return Result;
}


/* ********************************************************************** */
/* GEOPrintPolygon                                                        */
/* Lists all polygon's points                                             */
/* ********************************************************************** */
void GEOPrintPolygon (POLYGON *Polygon)
{
  int count;

  if (Polygon == NULL)
    return;

  PRO1 ("POLYGON: #%d\n", Polygon->Number)
  for (count = 0; count < Polygon->Number; ++count)
    PRO4 ("%3d. %8.2f %8.2f %8.2f\n", count,
	  Polygon->Points[Polygon->Vertex[count]].x,
	  Polygon->Points[Polygon->Vertex[count]].y,
	  Polygon->Points[Polygon->Vertex[count]].z)
}



/* ********************************************************************** */
/* GEOWritePolygon                                                        */
/* Writes the points of a polygon into the given stream.                  */
/* ********************************************************************** */
void GEOWritePolygon (FILE *fp, POLYGON *Polygon)
{
  int count;

  if (Polygon == NULL)
    return;

  fprintf (fp, "POLYGON: #%d\n", Polygon->Number);
  for (count = 0; count < Polygon->Number; ++count)
    fprintf (fp, "%3d. %8.3f %8.3f %8.3f\n", count,
	         Polygon->Points[Polygon->Vertex[count]].x,
	         Polygon->Points[Polygon->Vertex[count]].y,
	         Polygon->Points[Polygon->Vertex[count]].z);
}


/* ********************************************************************** */
/* GEOFreePolygon                                                         */
/* Gibt Speicher von Polygon frei.                                        */
/* ********************************************************************** */
void GEOFreePolygon (POLYGON *Polygon, dual all)
{
  if (Polygon == NULL)
    return;

  if (Polygon->Vertex != NULL)
    free (Polygon->Vertex);

  if (Polygon->Normals != NULL)
    free (Polygon->Normals);

  if (Polygon->Triangle != NULL)
    free (Polygon->Triangle);

  if (all && Polygon->Points != NULL)
    free (Polygon->Points);

  free (Polygon);
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/* ********************************************************************** */
/* ** Compute apolygon's normal vector using a convex vertex           ** */
/* ********************************************************************** */
VECTOR3D GEOGetPolygonNormal (POLYGON *Polygon)
{
  int       count, Index;
  VECTOR3D  Point;
  VECTOR3D  u, v;
  VECTOR3D  Result = Vector_0;
  VECTOR3D *Points = Polygon->Points;

  if (GEOTestPolygon(Polygon) == 0) 
    Index = 1;
  else
  {
    /* Find a polygon point which describes a convex vertex:
       Search the vertex with smallest x-coordinate, because the extreme
       vertices can't be concave. Otherweise there must be another extreme 
       vertex. */
    Index = 0;
    for (count = 1; count < Polygon->Number; ++count)
    {
      Point = Points[Polygon->Vertex[count]];
      if (Point.x < Points[Polygon->Vertex[Index]].x &&
	  (Point.x == Points[Polygon->Vertex[Index]].x &&
	   Point.y < Points[Polygon->Vertex[Index]].y))
	Index = count;
    }
  }

  u = GEOVectorSub (Points[Polygon->Vertex[(Index+Polygon->Number-1) % 
                                           Polygon->Number]],
		    Points[Polygon->Vertex[Index]]);
  v = GEOVectorSub (Points[Polygon->Vertex[(Index+1) % Polygon->Number]],
		    Points[Polygon->Vertex[Index]]);
  Result = GEOVectorCross (v, u);

  return Result;
}



/* ********************************************************************** */
/* ** Calculate a polygon's hesse-normal-form (HNF) using two          ** */
/* ** noncollinear edges. If the polygon does not posses such sides,   ** */
/* ** (all vectors collinear), the FALSE will be returned.             ** */
/* ********************************************************************** */
dual GEOCalcHesseForm (POLYGON *Polygon)
{
  int       count = 0;
  int       p1 = 0, p2 = 0, p3 = 0;
  int       pos = 0, neg = 0;
  VECTOR3D  u, v;
  VECTOR3D  n1 = Vector_0, n2;
  POINT3D  *Points;

  if (Polygon == NULL)
    return TRUE;

  Points = Polygon->Points;

#if 1
  while (n1.x == 0.0 && n1.y == 0.0 && n1.z == 0.0 && 
	 count < Polygon->Number-1)
  {
    p1 = Polygon->Vertex[(count-1+Polygon->Number) % Polygon->Number];
    p2 = Polygon->Vertex[count];
    p3 = Polygon->Vertex[(count+1)                 % Polygon->Number];

    u = GEOVectorSub (Points[p1], Points[p2]);
    v = GEOVectorSub (Points[p3], Points[p2]);
    n1 = GEOVectorCross (v, u);

    count++;
  }

  if (count == Polygon->Number)
    return FALSE;

  /*
    Here's a heuristic to decide whether a concave polygon is defined
    clockwise or counterclockwise. On convex polygons this calculation 
    is superfluous. It is neccessary, because all visible sides must 
    be travers in counterclockwise order. The idea: count the convex
    and concave vertices and switch the normal-vector, if there are 
    more concave than convex vertices.
  */
  p2 = Polygon->Vertex[(count-1+Polygon->Number) % Polygon->Number];
  p3 = Polygon->Vertex[count];
  while (count < Polygon->Number)
  {
    p1 = p2;
    p2 = p3;
    p3 = Polygon->Vertex[(count+1) % Polygon->Number];

    u = GEOVectorSub (Points[p1], Points[p2]);
    v = GEOVectorSub (Points[p3], Points[p2]);
    n2 = GEOVectorCross (v, u);

    if (GEOVectorAngle(n1, n2) > 90.0)
      ++neg;
    else
      ++pos;

    ++count;
  }

  if (neg > pos)
    n1 = GEOVectorMult (-1.0, n1);
#else
  n1 = GEOGetPolygonNormal (Polygon);
#endif

  /* calculate Hesse's Normal Form (HNF): */
  Polygon->Info.n = GEOVectorNorm (n1);
  Polygon->Info.a = - (Polygon->Info.n.x * Points[p2].x + 
		       Polygon->Info.n.y * Points[p2].y + 
		       Polygon->Info.n.z * Points[p2].z);

  return TRUE;
}


/* ********************************************************************** */
/* ** GEOIsConvexPolygon checks, if the committed polygon is convex    ** */
/* ** and planar.                                                      ** */
/* ** On success (convex and planar), the result will be 0. If the     ** */
/* ** polygon is concave, 1 will be returned. A nonplanar polygon      ** */
/* ** produces the result 2                                            ** */
/* ********************************************************************** */
int GEOTestPolygon (POLYGON *Polygon)
{
  int       count = -1;
  int       p1 = 0, p2 = 0, p3 = 0;
  double    Angle = 0.0;
  VECTOR3D  u, v;
  VECTOR3D  n1 = Vector_0, n2;
  POINT3D  *Points;

  if (Polygon == NULL)
    return -1;

  Points = Polygon->Points;

  count = 0;
  p1 = Polygon->Vertex[count];
  do 
  {
    ++count;
    p2 = Polygon->Vertex[count];
  } while (VEC_EQUAL(Points[p1], Points[p2]) && count < Polygon->Number);
  do
  {
    ++count;
    p3 = Polygon->Vertex[count];
  } while (VEC_EQUAL(Points[p2], Points[p3]) && count < Polygon->Number);

  u = GEOVectorSub (Points[p1], Points[p2]);
  v = GEOVectorSub (Points[p3], Points[p2]);
  n1 = GEOVectorCross (v, u);

  while (count < Polygon->Number)
  {
    p1 = p2;
    p2 = p3;
    p3 = Polygon->Vertex[(count+1) % Polygon->Number];

    u = GEOVectorSub (Points[p1], Points[p2]);
    v = GEOVectorSub (Points[p3], Points[p2]);
    n2 = GEOVectorCross (v, u);
 
    if ((Angle = GEOVectorAngle(n1, n2)) > 90.0)
    {
      PRO1 ("NORMAL-ANGLE: %f -> CONCAVE\n", Angle);
      return 1;
    }
    elsif (Angle > D_ANGLE)
    {
      PRO1 ("NORMAL-ANGLE: %f -> NOT PLANAR\n", Angle);
      return 2;
    }
    count++;
  }

  return 0;
}


/* ********************************************************************** */
/* ** Look, if there exist two adjacent points with the same           ** */
/* ** coordinates.                                                     ** */
/* ********************************************************************** */
void GEOReducePolygon (POLYGON *Polygon)
{
  int      count, index;
  POINT3D *Points;  

  if (Polygon == NULL)
    return;

  Points = Polygon->Points;
  count = 0;
  while (count < Polygon->Number)
    if (VEC_EQUAL(Points[count], Points[(count+1)%Polygon->Number])) 
    {
      for (index = count+1; index < Polygon->Number-1; ++index)
	Polygon->Vertex[index] = Polygon->Vertex[index+1];
      (Polygon->Number)--;
    }
    else
      count++;
}


/* ********************************************************************** */
/* ** Find the next concave vertex (beginning at vertex Start)         ** */
/* ** Search will be performed in ascending order.                     ** */
/* ********************************************************************** */
static int GEO_FindNextConcave (POLYGON *Polygon, int Start)
{
  int       count;
  int       p1 = 0, p2 = 0, p3 = 0;
  VECTOR3D  u, v, n1 = Vector_0, n2;
#if 0
  int       Index;
  POINT3D   Point;
#endif
  POINT3D  *Points = Polygon->Points;

  n1 = Polygon->Info.n;

  p2 = Polygon->Vertex[(Start-1+Polygon->Number) % Polygon->Number];
  p3 = Polygon->Vertex[Start];

  count = Start;
  while (count < Polygon->Number)
  {
    p1 = p2;
    p2 = p3;
    p3 = Polygon->Vertex[(count+1) % Polygon->Number];
    if (! VEC_EQUAL(Points[p2], Points[p3]))
    {
      u = GEOVectorSub (Points[p1], Points[p2]);
      v = GEOVectorSub (Points[p3], Points[p2]); 
      n2 = GEOVectorCross (v, u);

      if (GEOVectorAngle(n1, n2) > 90.0)
	return count;
#if 0
      /* Pruefen, ob ein Punkt innerhalb des bisher gefunden Polyons liegt */
      for (Index = 1; Index < Polygon->Number; ++Index)
      {
	if (! VEC_EQUAL(Points[Index], Points[p2]) && 
	    ! VEC_EQUAL(Points[Index], Points[p3]) &&
	    ! VEC_EQUAL(Points[Index], Points[Polygon->Vertex[Start]]))
	{
	  Point = Points[Polygon->Vertex[Index]];
	  if (GEOPointInTriangle(Points[Start], Points[p2], Points[p3], Point))
	    return count;
	}
      }
#endif
    }

    count++;
  }

  return -1;
}


/* ********************************************************************** */
/* ** Locate a polygon's last concave vertex. Not very effective,      ** */
/* ** because all vertexes must be touched - but it do it's job.       ** */
/* ** better: search in descending order.                              ** */
/* ********************************************************************** */
static int GEO_FindLastConcave (POLYGON *Polygon)
{
  int Result = -1, No;

  while (Result < Polygon->Number &&
	 (No = GEO_FindNextConcave(Polygon, Result+1)) > -1)
    Result = No;

  return Result;
}


/* ********************************************************************** */
/* ** Find the last vertex in ascending order, which would create an   ** */
/* ** convex polygon.                                                  ** */
/* ********************************************************************** */
static int GEO_FindValid (POLYGON *Polygon, int Start)
{
  int       count = 0;
  int       p1 = 0, p2 = 0, p3 = 0;
  double    Angle = 0.0;
  VECTOR3D  u, v, n1 = Vector_0, n2;
  POINT3D  *Points;

  Points = Polygon->Points;

  p1 = Polygon->Vertex[Start];
  do
  {
    count++;
    p2 = Polygon->Vertex[(Start+count) % Polygon->Number];
  } while (VEC_EQUAL(Points[p1], Points[p2]) && count < Polygon->Number);
  u = GEOVectorSub (Points[p2], Points[p1]);

  n1 = Polygon->Info.n;
  while (count < Polygon->Number-1 && Angle <= 90.0)
  {
    p3 = Polygon->Vertex[(Start+count+1) % Polygon->Number];
    if (! VEC_EQUAL(Points[p2], Points[p3]))
    {
      v  = GEOVectorSub (Points[p3], Points[p1]);
      n2 = GEOVectorCross (u, v);

      if (GEOVectorAngle(n1, n2) > 90.0)
	return (Start+count) % Polygon->Number;
    }
    else
      PRO2 ("%d = %d\n", p3, p2)

    count++;
  }

  return Polygon->Number;
}


/* ********************************************************************** */
/* ** Splits a concave polyon into convex polygons. Try to keep        ** */
/* ** subpolygons as big as possible.                                  ** */
/* ** First, locate two concave vertexes, which are NOT neighbours.    ** */
/* ** Create a new Polyon by connecting both vertexes. Repeat until    ** */
/* ** only one concave or only adjacent concave vertices exist. Then   ** */
/* ** split the remainig polygon into triangles.                       ** */
/* ** On errors, check, if the normal-vector of the concave polygon    ** */
/* ** was calculated correctly (-> GEOCalcHesseForm) !                 ** */
/* ********************************************************************** */
POLYGON *GEOTessellatePolygon (POLYGON *Polygon)
{
  dual     found = TRUE;
  int      count, count1 = 0, count2;
  int      Offs;
  POINT3D *Points;
  POLYGON *List = NULL, *Last = NULL, *New;
  POLYGON *Poly;

  PRO_START ("\n")

  if (Polygon == NULL)
    PRO_RETURN ("Empty Polygon", NULL)

  /* Create a copy of the polygon */
  Poly = GEOCopyPolygon (Polygon);
  Points = Poly->Points;

  while (count1 < Poly->Number)
  {
    /* Search for the next vertexes which does not produce a convex
       polygon */
    if ((count1 = GEO_FindNextConcave(Poly, count1)) > -1 && 
	count1 < Poly->Number-2 && 
	(count2 = GEO_FindNextConcave(Poly, count1+1)) > -1)
    {
      Offs = count2 - count1;
      if ((found = (Offs > 1)))
      {
	PRO3 ("A-CUT %d-%d [%d]:  ", count1, count2, Offs-1)
	for (count = count1+1; count < count2; ++count)
	  PRO1 ("\\%d  ", Poly->Vertex[count])
	PRO0 ("\\\n")

	/* Create New Polygon: */
	New = GEONewPolygon ();
	New->Points = Poly->Points;
	New->Number = Offs + 1;
	New->Vertex = (int *)malloc (sizeof(int) * New->Number);
	for (count = 0; count < New->Number; ++count)
	  New->Vertex[count] = Poly->Vertex[count1+count];
	New->Info.ID     = Poly->Info.ID;
	New->Info.Center = GEOCalcPolygonCenter (New);
	GEOCalcHesseForm      (New);
	GEOTriangulatePolygon (New);

	/* Append to list */
	if (List == NULL)
	  List = New;
	else
	  Last->Next = New;
	Last = New;

	/* Remove detached polygon's vertices */
	Poly->Number -= (Offs - 1);
	for (count = count1+1; count < Poly->Number; ++count)
	  Poly->Vertex[count] = Poly->Vertex[count+Offs-1];

	PRO1 ("-> POLYGON %d: ", Poly->Number)
	  for (count = 0; count < Poly->Number; ++count)
	PRO1 ("\\%d  ", Poly->Vertex[count])
        PRO0 ("\\\n")

	count1 = 0;
      }
      else
	count1 = count2;
    }
    else
      break;
  }

  /* Split remaining polygon */
  while ((count1 = GEO_FindLastConcave(Poly)) > -1)
  {
    if ((count2 = GEO_FindValid(Poly, count1)) > -1)
    {
      Offs = (count2-count1 + Poly->Number) % Poly->Number;
      if (Offs < 2)
	break;

      PRO3 ("B-CUT %d-%d [%d]:  ", count1, count2, Offs-1)
      for (count = 1; count < Offs; ++count)
	PRO1 ("\\%d  ", Poly->Vertex[(count1+count) % Poly->Number])
      PRO0 ("\\\n")

      New = GEONewPolygon ();
      New->Points = Poly->Points;
      New->Number = Offs + 1;
      New->Vertex = (int *)malloc (sizeof(int) * New->Number);
      for (count = 0; count < New->Number; ++count)
	New->Vertex[count] = Poly->Vertex[(count1+count) % Poly->Number];
      New->Info.ID     = Poly->Info.ID;
      New->Info.Center = GEOCalcPolygonCenter (New);
      GEOCalcHesseForm      (New);
      GEOTriangulatePolygon (New);

      /* Insert into list */
      if (List == NULL)
	List = New;
      else
	Last->Next = New;
      Last = New;

      /* Remove detached polygon's points */
      Poly->Number -= (Offs - 1);
      if (count2 < count1)
	for (count = 0; count < count1-count2+1; ++count)
	  Poly->Vertex[count] = Poly->Vertex[count2+count];
      else
	for (count = count1+1; count < Poly->Number; ++count)
	  Poly->Vertex[count] = Poly->Vertex[count+Offs-1];

      PRO1 ("-> POLYGON %d: ", Poly->Number)
	for (count = 0; count < Poly->Number; ++count)
	  PRO1 ("\\%d  ", Poly->Vertex[count])
      PRO0 ("\\\n")
    }
  }

  if (List == NULL)
    List = Poly;
  else
    Last->Next = Poly;

#if 0
  Poly = List;
  while (Poly != NULL)
  {
    printf ("POLYGON [%d]: ", Poly->Number);
    for (count = 0; count < Poly->Number; ++count)
      printf ("%d  ", Poly->Vertex[count]);
    printf ("\n");
    Poly = Poly->Next;
  }
#endif
  PRO_END

  return List;
}


/* ********************************************************************** */
/* ** Calculate a polygon's center                                     ** */
/* ********************************************************************** */
POINT3D GEOCalcPolygonCenter (POLYGON *Polygon)
{
  int     count;
  POINT3D Result = Vector_0;

  PRO_START (" ")

  if (Polygon == NULL || Polygon->Number == 0)
    PRO_RETURN ("No Polygon", Vector_0)

  Polygon->Info.Center = Vector_0;
  for (count = 0; count < Polygon->Number; ++count)
    Result = GEOVectorAdd (Result, Polygon->Points[Polygon->Vertex[count]]);

  Result = GEOVectorDiv (Result, Polygon->Number);

  PRO_END
  return Result;
}


void GEOGetPolygonBounds (POLYGON *Polygon, POINT3D *Min, POINT3D *Max)
{
  int     count;
  POINT3D Point;

  *Min = *Max = Polygon->Points[Polygon->Vertex[0]];
  for (count = 1; count < Polygon->Number; ++count)
  {
    Point = Polygon->Points[Polygon->Vertex[count]];
    Min->x = MIN (Min->x, Point.x);
    Min->y = MIN (Min->y, Point.y);
    Min->z = MIN (Min->z, Point.z);
    Max->x = MAX (Max->x, Point.x);
    Max->y = MAX (Max->y, Point.y);
    Max->z = MAX (Max->z, Point.z);
  }
}


/* ********************************************************************** */
/* ** The problem, if a triangle contains a given point may be solved  ** */
/* ** in different ways:                                               ** */
/* ** a) odd-even-rule: create a line through the point an test for    ** */
/* **    contact-points with the triangle-edges                        ** */
/* ** b) area-computing: (see alg.) advantage: maybe easier and better ** */
/* **    for possibility to estimate a inaccuracy.                     ** */
/* ** c) angle-computing: (see alg.)                                   ** */
/* **    Compute the angle to all vertices and check if the sum is 360 ** */
/* **    Seams to be the easiest and quickest approach.                ** */
/* ** Checks, if a triangle includes a point. First, the algorithm     ** */
/* ** computes the area of the triangle and then the sum of the three  ** */
/* ** triangles, which are created by combining the point p with all   ** */
/* ** vertexes. If the triangle includes the point, both areas will be ** */
/* ** the same (apart from rouding errors).                            ** */
/* ********************************************************************** */
dual GEOPointInTriangle (POINT3D p1, POINT3D p2, POINT3D p3, POINT3D p)
{
  double   Angle = 0;
  VECTOR3D pp1, pp2, pp3;

  /* 
     This version desides wheter a point lies inside or outside an triangle
     by computing the sum of all angles between two adjacent vectors from
     the point to the triangle-vertices. This sum should be 360 degree, if 
     the point lies inside the triangle and must be less, if the point lies
     outside 
  */
  pp1 = GEOVectorSub (p1, p);
  pp2 = GEOVectorSub (p2, p);  
  pp3 = GEOVectorSub (p3, p);

  Angle =   GEOVectorAngle (pp1, pp2) 
          + GEOVectorAngle (pp2, pp3)
          + GEOVectorAngle (pp3, pp1);

  return (Angle > 359.99);

#if 0      /* Old version (inefficient) */
  double a, b, c, s;
  double pp1, pp2, pp3;
  double Area, Sum = 0.0;

  if (VEC_SAME(p1, p) || VEC_SAME(p2, p) || VEC_SAME(p3, p))
    return TRUE;

  /* calculate distance of all triangle sides */
  a = VEC_LEN (GEOVectorSub(p2, p1));
  b = VEC_LEN (GEOVectorSub(p3, p2));
  c = VEC_LEN (GEOVectorSub(p1, p3));

  /* calculate center and area of triangle */
  s = 0.5 * (a + b + c);
  Area = sqrt (s * (s - a) * (s - b) * (s - c));

  /* we need the distance of the point to all vertexes */
  pp1 = VEC_LEN (GEOVectorSub(p1, p));
  pp2 = VEC_LEN (GEOVectorSub(p2, p));
  pp3 = VEC_LEN (GEOVectorSub(p3, p));

  /*
     calculate the area for all triangles built out of 2 triangle-vertexes 
     and the point. This point lies inside the triangle, if the sum of all
     areas is smaller as or equal the area of the tested triangle plus 
     delta. This test includes points, which do not lie in the same area
     but have a small distance to it (inexactness).
  */
  s = 0.5 * (pp1 + pp2 + a);
  Sum += sqrt (fabs(s * (s - pp1) * (s - pp2) * (s - a)));

  s = 0.5 * (pp2 + pp3 + b);
  Sum += sqrt (fabs(s * (s - pp2) * (s - pp3) * (s - b)));

  s = 0.5 * (pp1 + pp3 + c);
  Sum += sqrt (fabs(s * (s - pp1) * (s - pp3) * (s - c)));

  PRO3 ("Sum: %.4f  Area: %.4f   Max: %.4f\n", Sum, Area, Area*1.01)

  return (Sum <= Area * 1.01); /*  max 1 percent size-difference */
#endif
}



/*
  TODO:
  schnellere Routine fr Punkt in Dreieck implementieren. Zum Vergleich:
  Aufwand dieser Routine:
    10 Vergleiche
    10 Wurzelberechnungen
     6 Quadrierungen (entspr. Multipl.)
    16 Multiplikationen
    51 Add./Subtraktionen
     3 fabs-Berechnungen

  Berechnung ber Gauss (siehe GEOTetraederVolume) htte eine Unmenge an
  Add + Mult zur Folge.
*/
    

/* ********************************************************************** */
/* ** Compute the matrix of a polygon using one of its vetexes as      ** */
/* ** origin                                                           ** */
/* ********************************************************************** */
H_MATRIX GEOGetTransformXY (POLYGON *Polygon)
{
  int       count = 0;
  int       p1, p2, p3;
  VECTOR3D  n, o, a = Vector_0;
  POINT3D  *Points;

  Points = Polygon->Points;

  /* seach two not collinear vectors of the polygon */
  while (a.x == 0.0 && a.y == 0.0 && a.z == 0.0 && count < Polygon->Number-1)
  {
    p1 = Polygon->Vertex[(count+Polygon->Number-1) % Polygon->Number];
    p2 = Polygon->Vertex[count];
    p3 = Polygon->Vertex[(count+1)                 % Polygon->Number];

    n = GEOVectorSub (Points[p1], Points[p2]);
    o = GEOVectorSub (Points[p3], Points[p2]);
    a = GEOVectorCross (n, o);

    count++;
  }

  /* computer frame of polygon in point 0 */
  return GEOMakeRightSys (n, o, Points[Polygon->Vertex[0]]);
}


/* ********************************************************************** */
/* ** Transforms the polygon using the specified frame                 ** */
/* ********************************************************************** */
void GEOTransformPolygon (POLYGON *Polygon, H_MATRIX Frame)
{
  int count;

  PRO_START (" ")

  /* Attention: If the polygon is part of an opject, not all points might be
     transformed, because Polygon->Number don't covers the other points. */
  for (count = 0; count < Polygon->Number; ++count)
    Polygon->Points[Polygon->Vertex[count]] = 
      GEOMatrixMult (Frame, Polygon->Points[Polygon->Vertex[count]]);

  PRO_END
}


/* ********************************************************************** */
/* ** Reverse the order of the polygon's vertexes                      ** */
/* ********************************************************************** */
void GEOSwapOrientation (POLYGON *Polygon)
{
  int count;
  int Index, Swap;

  PRO_START (" ")

  if (Polygon == NULL)
    return;

  for (count = 0; count < Polygon->Number/2; ++count)
  {
    Index = Polygon->Number-count-1;
    Swap                   = Polygon->Vertex[count];
    Polygon->Vertex[count] = Polygon->Vertex[Index];
    Polygon->Vertex[Index] = Swap;
  }
  GEOCalcHesseForm (Polygon);

  PRO_END
}


/* ********************************************************************** */
/* ** Check, if two polygons have at least on common edge              ** */
/* ********************************************************************** */
dual GEOAreNeighbours (POLYGON *Polygon1, POLYGON *Polygon2)
{
  int Common = 0;
  int count1, count2;
  int Index1, Index2;

  for (count1 = 0; count1 < Polygon1->Number; ++count1)
  {
    Index1 = Polygon1->Vertex[count1];
    for (count2 = 0; count2 < Polygon2->Number; ++count2)
    {
      Index2 = Polygon2->Vertex[count2];
      if (Index1 == Index2 || GEOVectorDist(Polygon1->Points[Index1], 
			                    Polygon2->Points[Index2]) < 0.1)
      {
        Common++;
	continue;
      }
    }
    if (Common > 1)               /* There are more than one common vertices */
      return TRUE;
  }
  return FALSE;
}

/* ********************************************************************** */
/* ** Change a polygon's point from one pointindex to another.         ** */
/* ********************************************************************** */
void GEOChangePolygonPoint (POLYGON *Polygon, int From, int To)
{
  int count = 0;
  int Index;

  /* Change the point-index in the polygon */
  while (count < Polygon->Number && Polygon->Vertex[count] != From)
    count++;

  if (count < Polygon->Number)
    Polygon->Vertex[count] = To;

  /* Change the Triangle's index */
  for (count = 0; count < Polygon->Number-2; ++count)
    for (Index = 0; Index < 3; ++Index)
      if (Polygon->Triangle[count][Index] == From)
	Polygon->Triangle[count][Index] = To;  
}



void GEOFreePolygonNormals (POLYGON *Polygon)
{
  if (Polygon->Normals != NULL)
  {
    free (Polygon->Normals);
    Polygon->Normals = NULL;
  }
}


void GEOSetVertexNormal (POLYGON *Polygon, int Vertex, VECTOR3D Normal)
{
  int count;

  if (Polygon->Normals == NULL)
  {
    Polygon->Normals = (VECTOR3D *)malloc (sizeof(VECTOR3D) * Polygon->Number);
    for (count = 0; count < Polygon->Number; ++count)
      Polygon->Normals[count] = Vector_0;
  }

  for (count = 0; count < Polygon->Number; ++count)
    if (Polygon->Vertex[count] == Vertex)
      Polygon->Normals[count] = Normal;
}
