#include "libgeo.h"


typedef struct
{
  POINT2D Pos;
  int     Index;
} POLYPOINT;

typedef struct
{
  int        Number;
  POLYPOINT *Points;
  TRIANGLE  *Triangles;
} POLY;


/* ********************************************************************** */
/* ** Split a convex polyon into triangles (triangulation) using the   ** */
/* ** fan-algorithm. The first point will be used in all triangles.    ** */
/* ********************************************************************** */
static void GEO_TriangulateConvex (POLYGON *Polygon)
{
  int count;

  PRO_START (" ")

  if (Polygon == NULL)
    PRO_ABORT ("empty polyeder")

  Polygon->Triangle = (TRIANGLE*)malloc (sizeof(TRIANGLE) * 
					 (Polygon->Number-2));
  for (count = 1; count < Polygon->Number-1; ++count)
  {
    Polygon->Triangle[count-1][0] = Polygon->Vertex[0];
    Polygon->Triangle[count-1][1] = Polygon->Vertex[count];
    Polygon->Triangle[count-1][2] = Polygon->Vertex[count+1];
  }

  PRO_END
}


/* ************************************************************* */
/* **                   POLYGON-TRIANGULATION                 ** */
/* ** Found in J. O'Rourke: 'Computational Geometry in C      ** */
/* ************************************************************* */

static void GEO2D_ClipEar (POLYGON2D *Poly, int *Labels, int p)
{
  int count;

  for (count = p; count < Poly->Number-1; ++count)
  {
    Poly->Points[count] = Poly->Points[count+1];
    Labels[count] = Labels[count+1];
  }

  (Poly->Number)--;
}

/* Attention: function destroys poly. Create a copy if neccessary 
   see Josef O'Rouke: "Computational Geometry in C", page 35ff for details */
static void GEO2D_DoTriangulate (POLYGON2D *Poly, int *Labels, int *No)
{
  int count;
  int p1, p2, p3;

  if (Poly->Number > 3)
    for (p1 = 0; p1 < Poly->Number; ++p1)
    {
      p3 = (p1 + 2) % Poly->Number;
      if (GEO2DInCone(p1, p3, Poly) && GEO2DDiagonalie(p1, p3, Poly))
      {
	p2 = (p1 + 1) % Poly->Number;

	Poly->Triangle[*No][0] = Labels[p1];
	Poly->Triangle[*No][1] = Labels[p2];
	Poly->Triangle[*No][2] = Labels[p3];
        printf ("TRIANGLE: %d -> %d -> %d\n", Poly->Triangle[*No][0], 
		                              Poly->Triangle[*No][1], 
		                              Poly->Triangle[*No][2]);
	(*No)++;

        GEO2D_ClipEar (Poly, Labels, p2);	
	GEO2D_DoTriangulate (Poly, Labels, No);
        break;
      }
    }
  elsif (Poly->Number == 3)                        /* Process last triangle */
  {
    for (count = 0; count < 3; ++count)
      Poly->Triangle[*No][count] = Labels[count];
    (*No)++;
  }
}

/* Steps:
   Suppose, the polygon has correct order, that means, orientation is CCW.
   1.  Transform polygon into xy-layer
   2.  Copy values into a list containing the index values, because algorithm
       destroys data.
   3.  for all vertexes: find ears and remove them - store result in extra
       list
   The computation will be performed on a copy of the original data
*/
void GEO_TriangulateConcave (POLYGON *Polygon)
{
  int        count, No = 0;
  int        Index, VertexNo;
  int        Orientation;
  int        LSwap;
  int       *Labels;
  POINT2D    PSwap;
  POINT3D    Point;
  H_MATRIX   Trans;
  POLYGON2D  Poly;

  PRO_START ("\n")

  for (count = 0; count < Polygon->Number; ++count)
    PRO4 ("POLYGON %d:  [%5.1f %5.1f %5.1f]\n", count,
	  Polygon->Points[Polygon->Vertex[count]].x, 
	  Polygon->Points[Polygon->Vertex[count]].y, 
	  Polygon->Points[Polygon->Vertex[count]].z)

  /* remove double start- and end-vertex */
  Poly.Number = Polygon->Number;
  if (VEC_EQUAL(Polygon->Points[Polygon->Vertex[0]],
		Polygon->Points[Polygon->Vertex[Polygon->Number-1]]))
    (Poly.Number)--;

  /* Transform polygon into xy-plane */
  Poly.Points = (POINT2D *)malloc (sizeof(POINT2D) * Poly.Number);
  Labels = (int *)malloc (sizeof(int) * Poly.Number);

  Trans = GEOMatrixInv (GEOGetTransformXY(Polygon));
  for (count = 0; count < Poly.Number; ++count)
  {
    Point = GEOMatrixMult (Trans, Polygon->Points[Polygon->Vertex[count]]);
    Poly.Points[count].x = Point.x;
    Poly.Points[count].y = Point.y;
    Labels[count] = count;
    PRO4 ("POLY %d:  [%5.1f %5.1f %5.1f]\n", count, Point.x, Point.y, Point.z)
  }

  /* Check polygon-orrientation after transformation into xy-area. If 
     orientation is clockwise, reverse point's ordxer but keep index-labels, 
     because original polygon has correct direction ! 
  */
  Orientation = GEO2DPolygonOrientation (&Poly);
  PRO1 ("ORIENTATION: %s\n", (Orientation < 0)? "CW":"CCW")
  if (Orientation == GeoDirCW)
    for (count = 0; count < Poly.Number/2; ++count)
    {
      Index              = Poly.Number-1-count;
      PSwap              = Poly.Points[count];
      Poly.Points[count] = Poly.Points[Index];
      Poly.Points[Index] = PSwap;
      LSwap              = Labels[count];
      Labels[count]      = Labels[Index];
      Labels[Index]      = LSwap;
    }

  /* Allocate Memory for triangle-data */
  Poly.Triangle = (TRIANGLE *)malloc (sizeof(TRIANGLE) * (Poly.Number - 2));

  VertexNo = Poly.Number;
  GEO2D_DoTriangulate (&Poly, Labels, &No);
  if (No != VertexNo-2)
    printf ("TRIANGULATION-ERROR: %d !\n", No);

  /* Copy triangle-data into the polygon-structure */
  Polygon->Triangle = (TRIANGLE *)malloc (sizeof(TRIANGLE) * No);
  for (count = 0; count < No; ++count)
  {
    Polygon->Triangle[count][1] = Polygon->Vertex[Poly.Triangle[count][1]];
    if (Orientation < 0)
    {
      Polygon->Triangle[count][0] = Polygon->Vertex[Poly.Triangle[count][2]];
      Polygon->Triangle[count][2] = Polygon->Vertex[Poly.Triangle[count][0]];
    }
    else
    {
      Polygon->Triangle[count][0] = Polygon->Vertex[Poly.Triangle[count][0]];
      Polygon->Triangle[count][2] = Polygon->Vertex[Poly.Triangle[count][2]];
    }
    printf ("%d: %d %d %d\n", count, 
	    Polygon->Triangle[count][0], Polygon->Triangle[count][1],
	    Polygon->Triangle[count][2]);
  }

  free (Poly.Points);
  free (Poly.Triangle);
  free (Labels);

  PRO_END
}


/* This is a simple realization of ear-clipping, because the upper algorithms
   seems to have problems, if there are vertexes used twice, which happens, if
   there are holes in a polygon: one vertex of the surrounding and one vertex
   of the hole will be used twice. Because we have no degenerated polygons,
   we do not test, if there is a intersecting edge but if there is another
   vertex which lies inside the tested ear.
*/
static int GEO_GetNextValid (POLYGON *Polygon, int Start, dual *done)
{
  int count = 1;

  while (count < Polygon->Number)
  {
    if (! done[(Start+count) % Polygon->Number])
      return (Start+count);
    ++count;
  }
  return -1;
}  

static void GEO_TriangulateConcave2 (POLYGON *Polygon)
{
  dual      *done;
  int       count;
  int       Index1, Index2, Index3;
  int       No = 0;
  VECTOR3D  u, v;
  VECTOR3D  Norm1, Norm2;
  POINT3D   Point, Points[3];

  PRO_START ("\n")

  if (Polygon->Number < 3)
    PRO_ABORT ("Invalid Polygon")

  /* Allocate Memory for triangle-data */
  Polygon->Triangle = (TRIANGLE *)malloc (sizeof(TRIANGLE) * 
					  (Polygon->Number - 2));

  done = (dual *)malloc (sizeof(dual) * Polygon->Number);
  memset (done, FALSE, sizeof(dual) * Polygon->Number);

  Norm1 = GEOGetPolygonNormal (Polygon);
#if 0
  Norm1 = GEOVectorMult (-1.0, Norm1);
#endif

  Index1 = -1;
  while (No < Polygon->Number-2)
  {
    if ((Index1 = GEO_GetNextValid(Polygon, Index1, done)) >= 0 &&
	(Index2 = GEO_GetNextValid(Polygon, Index1, done)) >= 0 &&
        (Index3 = GEO_GetNextValid(Polygon, Index2, done)) >= 0)
    {
      Points[0] = Polygon->Points[Polygon->Vertex[Index1]];     
      Points[1] = Polygon->Points[Polygon->Vertex[Index2]];
      Points[2] = Polygon->Points[Polygon->Vertex[Index3]];
      u = GEOVectorSub (Points[0], Points[1]);
      v = GEOVectorSub (Points[2], Points[1]);
      Norm2 = GEOVectorCross (v, u);

      if (GEOVectorAngle(Norm1, Norm2) <= 90.0)
      {
        /* Check if there is not other point (exect the points itself) that
	   lies inside the tested ear (triangle). If so, clipp the middle 
	   vertex */
	for (count = 0; count < Polygon->Number; ++count)
	{
	  Point = Polygon->Points[Polygon->Vertex[count]];
	  if (! VEC_EQUAL(Points[0], Point) && 
	      ! VEC_EQUAL(Points[1], Point) &&
	      ! VEC_EQUAL(Points[2], Point) && 
	      GEOPointInTriangle(Points[0], Points[1], Points[2], Point))
	    break;
	}
	if (count == Polygon->Number)  /* No vertex inside the tested ear */
	{
	  done[Index2] = TRUE;
	  Polygon->Triangle[No][0] = Polygon->Vertex[Index1];
	  Polygon->Triangle[No][1] = Polygon->Vertex[Index2];
	  Polygon->Triangle[No][2] = Polygon->Vertex[Index3];
#if 0
	  printf ("Tri%d: %d %d %d\n", No, Polygon->Triangle[No][0],
		  Polygon->Triangle[No][1], Polygon->Triangle[No][2]);
#endif
	  ++No;
	  Index1 = -1;
	}
      }
    }
    else
      printf ("! No valid Point found !\n");
  }

#if 0
  printf ("READY\n");
#endif

  PRO_END
}


/* ********************************************************************** */
/* ** Checks, if the polygon is convex/concave and calls the           ** */
/* ** appropriate triangulation-routine                                ** */
/* ** The result is stored in the polygon's triangle-structure         ** */
/* ********************************************************************** */
void GEOTriangulatePolygon (POLYGON *Polygon)
{
  int State;

  PRO_START ("\n")

  if (Polygon->Triangle != NULL)
    free (Polygon->Triangle);
  Polygon->Triangle = NULL;

  switch ((State = GEOTestPolygon(Polygon)))
  {
    case  0 : GEO_TriangulateConvex (Polygon);
              break;
    case  1 : GEO_TriangulateConcave2 (Polygon);
              break;
    default : PRO1 ("Polygon-check failed: Error %d\n", State);
  }

  PRO_END
}




