#include <math.h>

#include "libgeo.h"
/* ********************************************************************** */
/* geoVector.C                                                            */
/* stellt Funktionen zur Vektormanipulation zur Verfgung                 */
/* ---------------------------------------------------------------------- */
/* Von            : Heike Ochs-Vietinghoff und Stefan Vietinghoff         */
/* ********************************************************************** */


VECTOR3D Vector_0 = {0.0, 0.0,  0.0};             /* Nullvektor           */
VECTOR3D Vector_1 = {1.0, 1.0,  1.0};             /* Einheitsvektor       */
VECTOR3D Vector_g = {0.0, 0.0, -1.0};             /* Gravitationsvektor   */
VECTOR3D Vector_X = {1.0, 0.0,  0.0};             /* X-Vector             */  
VECTOR3D Vector_Y = {0.0, 1.0,  0.0};             /* Y-Vector             */  
VECTOR3D Vector_Z = {0.0, 0.0,  1.0};             /* Z-Vector             */  


VECTOR3D GEOMakeVector (double x, double y, double z)
{
  VECTOR3D Result;

  Result.x = x;
  Result.y = y;
  Result.z = z;

  return Result;
}


/* ************************************************************************* */
/* ** Convert a 2D-point into a 3D-Vector, setting z-component to zero    ** */
/* ************************************************************************* */
VECTOR3D GEOToVector (POINT2D Point)
{
  return GEOMakeVector (Point.x, Point.y, 0.0);
}


/* ************************************************************************* */
/* ** Convert vector into 2D-point cutting the z-component                ** */
/* ************************************************************************* */
POINT2D GEOToPoint2D (VECTOR3D Vec)
{
  POINT2D Point;
 
  Point.x = Vec.x;
  Point.y = Vec.y;

  return Point;
}


/* ************************************************************************* */
/* GEOVectorAdd                                                              */
/* addiert die Vektoren "v1" und "v2".                                       */
/* ************************************************************************* */
VECTOR3D GEOVectorAdd (VECTOR3D Vec1, VECTOR3D Vec2)
{
  VECTOR3D Result;

  Result.x = Vec1.x + Vec2.x;
  Result.y = Vec1.y + Vec2.y;
  Result.z = Vec1.z + Vec2.z;

  return Result;
}


/* ************************************************************************* */
/* GEOVectorSub                                                              */
/* erzeugt Richtungsvektor der Kante mit den Eckpunkten p1 und p2.           */
/* ************************************************************************* */
VECTOR3D GEOVectorSub (VECTOR3D Vec1, VECTOR3D Vec2)
{
  VECTOR3D Result;

  Result.x = Vec1.x - Vec2.x;
  Result.y = Vec1.y - Vec2.y;
  Result.z = Vec1.z - Vec2.z;

  return Result;
}


/* ************************************************************************* */
/* GEOVectorMult                                                             */
/* mutipliziert den Vektor "v" mit dem Skalar "Faktor".                      */
/* ************************************************************************* */
VECTOR3D GEOVectorMult (double Factor, VECTOR3D Vec)
{
  VECTOR3D Result;

  Result.x = Factor * Vec.x;
  Result.y = Factor * Vec.y;
  Result.z = Factor * Vec.z;

  return Result;
}


/* ************************************************************************* */
/* GEOVectorDiv                                                              */
/* dividiert den Vektor "v" durch den Skalar "Divisor". Ist Divisor=0 wird   */
/* der Null-Vektor zurckgegeben.                                            */
/* ************************************************************************* */
VECTOR3D GEOVectorDiv (VECTOR3D Vec, double Divisor)
{
  VECTOR3D Result = Vector_0;

  if (Divisor == 0.0)
    return Result;

  Result.x = Vec.x / Divisor;
  Result.y = Vec.y / Divisor;
  Result.z = Vec.z / Divisor;

  return Result;
}


/* ************************************************************************* */
/* GEOVectorLength                                                           */
/* berechnet die Lnge des Vektors "v" (Euklidische Norm).                   */
/* ************************************************************************* */
double GEOVectorLength (VECTOR3D Vec)
{
  return (sqrt(SQR(Vec.x) + SQR(Vec.y) + SQR(Vec.z)));
}


/* ************************************************************************* */
/* GEOVectorLength                                                           */
/* normiert den gegebenen Vektor                         .                   */
/* ************************************************************************* */
VECTOR3D GEOVectorNorm (VECTOR3D Vec)
{
  double len;

  if ((len = VEC_LEN(Vec)) != 0.0)
    return GEOVectorDiv (Vec, VEC_LEN(Vec));
  else
    return Vector_0;
}


/* ********************************************************************** */
/* ** Scales the vector to the given length. This function might be    ** */
/* ** realized by a macro. But the advantage of a function is to       ** */
/* ** feed another macro (e.g. VEC_X(v) (v.x)) with our result.        ** */
/* ** INPUT:  v           Vector                                       ** */
/* **         len         desired length of the vector                 ** */
/* ** OUTPUT: (VECTOR3D)  Resulting vector.                            ** */
/* ********************************************************************** */
VECTOR3D GEOVectorScale (VECTOR3D Vec, double len)
{
  return GEOVectorMult (len, GEOVectorNorm(Vec));
}


/* ************************************************************************* */
/* GEOVektorProduct                                                          */
/* berechnet das Skalarprodukt der Vektoren "v1" und "v2".                   */
/* ************************************************************************* */
double GEOVectorProd (VECTOR3D Vec1, VECTOR3D Vec2)
{
  return (Vec1.x*Vec2.x + Vec1.y*Vec2.y + Vec1.z*Vec2.z);
}


/* ************************************************************************* */
/* GEOVektorSquare                                                           */
/* berechnet das Quadrat des Vektor v                                        */
/* ************************************************************************* */
double GEOVectorSquare (VECTOR3D Vec)
{
  return (SQR(Vec.x) + SQR(Vec.y) + SQR(Vec.z));
}


/* ************************************************************************* */
/* GEOVectorAngle                                                            */
/* berechnet den Winkel zwischen den Vektoren "v1" und "v2". Diesen liegt    */
/* immer zwischen 0 und 180.                                                */
/* ************************************************************************* */
double GEOVectorAngle (VECTOR3D Vec1, VECTOR3D Vec2)
{
  double ab, area, div;
  double Result = 0.0;

  ab = VEC_LEN (Vec1) * VEC_LEN (Vec2);
  area = (Vec1.x*Vec2.x + Vec1.y*Vec2.y + Vec1.z*Vec2.z);
  div = area / ab;

  /* Test for rounding-errors */
  if (div < -1.0)
    div = -1.0;
  elsif (div > 1.0)
    div = 1.0;

  if (ab != 0.0)
    Result = GRAD (acos(div));

  return Result;
}


/* ************************************************************************* */
/* GEOVectorDist                                                             */
/* Compares two vectors and return the distance between them.                */
/* ************************************************************************* */
double GEOVectorDist (VECTOR3D Vec1, VECTOR3D Vec2)
{
  return GEOVectorLength (GEOVectorSub(Vec1, Vec2));
}


/* ************************************************************************* */
/* GEOTriangleNormal                                                         */
/* berechnet den Normalen-Vektor zu der durch "v1","v2" und "v3" gegebenen   */
/* Ebene                                                                     */
/* ************************************************************************* */
VECTOR3D GEOTriangleNormal (POINT3D v1, POINT3D v2, POINT3D v3)
{
  VECTOR3D Result;

  PRO_START (" ")

  Result.x = - ((v2.y - v1.y) * (v2.z + v1.z) +
		(v3.y - v2.y) * (v3.z + v2.z) +
		(v1.y - v3.y) * (v1.z + v3.z));
  Result.y = - ((v2.z - v1.z) * (v2.x + v1.x) +
		(v3.z - v2.z) * (v3.x + v2.x) +
		(v1.z - v3.z) * (v1.x + v3.x));
  Result.z = - ((v2.x - v1.x) * (v2.y + v1.y) +
		(v3.x - v2.x) * (v3.y + v2.y) +
		(v1.x - v3.x) * (v1.y + v3.y));

  if (fabs(Result.x) < D_DIST)
    Result.x = 0.0;
  if (fabs(Result.y) < D_DIST)
    Result.y = 0.0;
  if (fabs(Result.z) < D_DIST)
    Result.z = 0.0;

  PRO_END
  return Result;
}


/* ************************************************************************* */
/* GEOVectorCross                                                            */
/* berechnet das Kreuzprodukt von "v1" und "v2".                             */
/* ************************************************************************* */
VECTOR3D GEOVectorCross (VECTOR3D Vec1, VECTOR3D Vec2)
{
  VECTOR3D Result;

  Result.x = Vec1.y * Vec2.z - Vec1.z * Vec2.y;
  Result.y = Vec1.z * Vec2.x - Vec1.x * Vec2.z;
  Result.z = Vec1.x * Vec2.y - Vec1.y * Vec2.x;

  return Result;
}


/* ************************************************************************* */
/* GEOCalcAngle                                                              */
/* berechnet den Winkel zwischen der Kante von "p1" nach "p2" und der Kante  */
/* von "p2" nach "p3".                                                       */
/* ************************************************************************* */
double GEOCalcAngle (POINT3D p1, POINT3D p2, POINT3D p3)
{
  VECTOR3D u, v;

  u = GEOVectorSub (p2, p1);
  v = GEOVectorSub (p3, p2);

  return GEOVectorAngle (u, v);
}


/* ********************************************************************** */
/* ** Checks, if the vectors v1 and v2 are collinear. The result is 1, ** */
/* ** if both have the same direction, -1, if they have reverse        ** */
/* ** directions and 0 otherwise.                                      ** */
/* ********************************************************************** */
int GEOVectorCollinear (VECTOR3D Vec1, VECTOR3D Vec2)
{
  double Angle;

  Angle = fabs (GEOVectorAngle(Vec1, Vec2));
  if (Angle < D_ANGLE)
    return 1;
  elsif (Angle > 180.0 - D_ANGLE)
    return -1;
  else
    return 0;
}

/* ********************************************************************** */
/* ** Compares two vectors, Result: 1=same, -1=opposite direction,     ** */
/* ** 0 else                                                           ** */
/* ********************************************************************** */
int GEOVectorDirection (VECTOR3D Vec1, VECTOR3D Vec2)
{
  VECTOR3D Norm1, Norm2;

  Norm1 = GEOVectorNorm (Vec1);
  Norm2 = GEOVectorNorm (Vec2);

  if (VEC_EQUAL(Norm1, Norm2)) 
    return 1;

  Norm2 = GEOVectorMult (-1.0, Norm2);
  if (VEC_EQUAL(Norm1, Norm2)) 
    return -1;

  return 0;
}

