#ifndef LIBGEO_H
#define LIBGEO_H

#ifdef __cplusplus
extern "C" {
#endif

#include <X11/Intrinsic.h>

#include <debug.h>

/* ********************************************************************** */
/* **                              Definitions                         ** */
/* ********************************************************************** */

#ifndef M_PI
  #define M_PI 3.1415926593924
#endif

#define X 0
#define Y 1
#define Z 2
#define W 3

#define N 0
#define O 1
#define A 2
#define P 3

#define PX 0
#define PY 1
#define PZ 2
#define RX 3
#define RY 4
#define RZ 5


#define D_ANGLE  8E-1
#define D_DIST   5E-1
#define F_DIST   1.0

#define EPSILON       0.015

#define LGS_EPSILON   2E-6              /* 1E-6 may cause problems     */
#define MAX_LGS       10                /* max. linear problem size    */

/* Creates messages for debugging CSG-ops */
/*
#define DEBUG_CSG
*/

enum GEO_DIRECTION
{
  GeoDirCW = -1,
  GeoDirNO,
  GeoDirCCW
};

enum INTERSECTION_TYPE
{
  geoCOPLANAR = 0,
  geoTRANSVERSAL
};


/* ********************************************************************** */
/* ** Some usefull macros                                              ** */
/* ********************************************************************** */
#ifndef SQR
  #define SQR(x)  ((x) * (x))
#endif

#ifndef SIGN
  #define SIGN(x) (x < 0)? (-1):(1)
#endif


#define VEC_LEN(v)       (sqrt(SQR(v.x) + SQR(v.y) + SQR(v.z)))
#define VEC_EQUAL(v1,v2) (((v1).x==(v2).x && (v1).y==(v2).y && (v1).z==(v2).z))
#define VEC_SAME(v1,v2) \
          (sqrt(SQR(v2.x-v1.x) + SQR(v2.y-v1.y) + SQR(v2.z-v1.z)) < D_DIST)
#ifndef RAD
  #define RAD(x)           ((x) * M_PI / 180.0)
#endif
#ifndef GRAD
  #define GRAD(x)          ((x) * 180.0 / M_PI)
#endif

#define PRO_VECTOR3D(V,Title) \
	  PRO4 ("%s: %.4f %.4f %.4f\n", Title, (V).x, (V).y, (V).z)
#define WRITE_VECTOR3D(fp,V,Title) \
          fprintf (fp, "%s: %.4f %.4f %.4f\n", Title, (V).x, (V).y, (V).z)

#define PRO_HMATRIX(M,Title) \
        { \
	  PRO1 ("%s:\n", Title) \
	  PRO4 ("%8.4f  %8.4f  %8.4f  %8.4f\n", \
		(M).m[N][X], (M).m[O][X], (M).m[A][X], (M).m[P][X]) \
	  PRO4 ("%8.4f  %8.4f  %8.4f  %8.4f\n", \
		(M).m[N][Y], (M).m[O][Y], (M).m[A][Y], (M).m[P][Y]) \
	  PRO4 ("%8.4f  %8.4f  %8.4f  %8.4f\n", \
		(M).m[N][Z], (M).m[O][Z], (M).m[A][Z], (M).m[P][Z]) \
	  PRO4 ("%8.4f  %8.4f  %8.4f  %8.4f\n", \
		(M).m[N][W], (M).m[O][W], (M).m[A][W], (M).m[P][W]) \
	}

#define WRITE_HMATRIX(fp,M,Title) \
        { \
	  fprintf (fp, "%s:\n", Title); \
          fprintf (fp, "  %8.4f  %8.4f  %8.4f  %8.4f\n", \
	               M.m[N][X], M.m[O][X], M.m[A][X], M.m[P][X]); \
	  fprintf (fp, "  %8.4f  %8.4f  %8.4f  %8.4f\n", \
	               M.m[N][Y], M.m[O][Y], M.m[A][Y], M.m[P][Y]); \
          fprintf (fp, "  %8.4f  %8.4f  %8.4f  %8.4f\n", \
	               M.m[N][Z], M.m[O][Z], M.m[A][Z], M.m[P][Z]); \
          fprintf (fp, "  %8.4f  %8.4f  %8.4f  %8.4f\n", \
	               M.m[N][W], M.m[O][W], M.m[A][W], M.m[P][W]); \
	}


/* ********************************************************************** */
/* **                     Type - Definitions                           ** */
/* ********************************************************************** */
typedef double ARRAY[3];
typedef int    TRIANGLE[3];


typedef struct
{
  double x, y;
} POINT2D;


typedef struct
{
  double x, y, z;                       /* allows direct return-values    */
} VECTOR3D, POINT3D;


typedef struct
{
  double   x, y, z;
  VECTOR3D Normal;
} VERTEX;


typedef struct
{
  double m[4][4];                       /* [N/O/A/P][X/Y/Z]               */
} H_MATRIX;


typedef struct
{
  double px, py, pz;                    /* 3d position                    */
  double r, p, y;                       /* orientation (roll/pitch/yaw)   */
} RPY;


typedef struct
{
  double px, py, pz;                    /* 3d position                    */
  double z1, y, z2;                     /* orientation (euler-angles)     */
} EULER;


typedef struct
{
  int start, end;                       /* Index of edge's vertexes       */
} GEO_EDGE;


typedef struct
{
  int       ID;                         /* unique identifier              */
  POINT3D   Center;                     /* center of gravity              */
  double    Area;                       /* area of polygon                */
  VECTOR3D  n;                          /* HNF: n * x + a = 0             */
  double    a;
} INFO_2D;


typedef struct POLYGON
{
         int       Number;             /* Amount of vertex-points         */
         int      *Vertex;             /* Vector of Vertex-indexes        */
         POINT3D  *Points;             /* link to vertex-points           */

         VECTOR3D *Normals;            /* Vertex Normal vectors           */ 

         TRIANGLE *Triangle;           /* Vector for triangulated data    */
         INFO_2D   Info;               /* Informations about the polygon  */

         int       Flags;              /* Flags for CSG-computation       */

  struct POLYGON  *Next;               /* Pointer onto next polygon       */
} POLYGON;


typedef struct POLYGON2D
{
  int       Number;                     /* Number of 2d-points            */
  POINT2D  *Points;                     /* List of Points                 */
  TRIANGLE *Triangle;                   /* List of triangle-indexes       */
} POLYGON2D;


typedef struct
{
  POINT3D   Center;                     /* Center of gravity              */
  POINT3D   Min, Max;                   /* min/max-values of bounding box */
  double    Area;                       /* Area of surrounding polygons   */
  double    Volume;                     /* Volume of polyeder             */
  double    Bubble;                     /* Bubble-radius                  */
} INFO_3D;


typedef struct
{
  int  Number;                          /* Number of curve-vertexes       */
  int *List;                            /* Index-vector into vertex-list  */
} LINES;


typedef struct POLYEDER
{
  int      Number;                      /* Amount of polygons             */
  POLYGON *List, *Last;                 /* List of polygons               */  
} POLYEDER;


typedef struct
{
  int    Size;                          /* Number of Elements             */
  double Field[MAX_LGS][MAX_LGS+1];     /* Elements of of LGS-Matrix      */
  int    State;                         /* State of solution              */
} LGS;


/* Data structures for calculating normal vectors for shading */
typedef struct POLYGONELEM
{
         POLYGON     *Polygon;
         dual         done;
  struct POLYGONELEM *Next;
} POLYGONELEM;


typedef struct POINTELEM
{
  int          Number;
  POLYGONELEM *PolygonList;
  POLYGONELEM *Last;
} POINTELEM;


typedef struct INTERSECT
{
         int        ID;                 /* Only for debugging purpose     */
         int        Type;               /* IS-type (coplanar/transversal) */
         POLYGON   *Poly[2];            /* Intersecting polygons          */
         int        Number[2];          /* # Point in each polygon        */
         POINT3D   *Points[2];          /* Intersection point             */
         int        Index[2][2];        /* Index of intersection edges    */
         double     Lambda[2][2];       /* Intersect-location in the edge */
         dual       done;               /* Flag marks processed intersect */
  struct INTERSECT *Next;               /* Pointer to next intersection   */
} INTERSECT;


/* ********************************************************************** */
/* **                     Exported default-values                      ** */
/* ********************************************************************** */
extern VECTOR3D  Vector_0;
extern VECTOR3D  Vector_1;
extern VECTOR3D  Vector_g;
extern VECTOR3D  Vector_X;
extern VECTOR3D  Vector_Y;
extern VECTOR3D  Vector_Z;
extern H_MATRIX  Matrix_0;
extern H_MATRIX  Matrix_1;
extern H_MATRIX  REL_M;
extern RPY       RPY_0;


/* ********************************************************************** */
/* **                        Exported functions                        ** */
/* ********************************************************************** */

/* *** 2D: *** */
POLYGON2D *GEO2DNewPolygon          ();
void       GEO2DFreePolygon         (POLYGON2D *Polygon);
int        GEO2DPolygonOrientation  (POLYGON2D *Polygon);
void       GEO2DSwapOrientation     (POLYGON2D *Polygon);
POLYGON2D *GEO2DPolygonTransformXY  (POLYGON *Polygon);
dual       GEO2DInCone              (int p1, int p3, POLYGON2D *Poly);
dual       GEO2DDiagonalie          (int p1, int p3, POLYGON2D *Poly);
POLYGON   *GEO2DConvexIntersection  (POLYGON2D *Poly1, POLYGON2D *Poly2);

/* *** VECTOR: *** */
VECTOR3D   GEOMakeVector         (double x, double y, double z);
VECTOR3D   GEOToVector           (POINT2D Point);
POINT2D    GEOToPoint2D          (VECTOR3D Vec);
VECTOR3D   GEOVectorAdd          (VECTOR3D Vec1, VECTOR3D Vec2);
VECTOR3D   GEOVectorSub          (VECTOR3D Vec1, VECTOR3D Vec2);
VECTOR3D   GEOVectorMult         (double Factor, VECTOR3D Vec);
VECTOR3D   GEOVectorDiv          (VECTOR3D Vec, double Divisor);
double     GEOVectorLength       (VECTOR3D Vec);
VECTOR3D   GEOVectorNorm         (VECTOR3D Vec);
VECTOR3D   GEOVectorScale        (VECTOR3D Vec, double len);
double     GEOVectorProd         (VECTOR3D Vec1, VECTOR3D Vec2);
double     GEOVectorSquare       (VECTOR3D Vec);
double     GEOVectorAngle        (VECTOR3D Vec1, VECTOR3D Vec2);
double     GEOVectorDist         (VECTOR3D Vec1, VECTOR3D Vec2);

VECTOR3D   GEOVectorCross        (VECTOR3D Vec1, VECTOR3D Vec2);
int        GEOVectorCollinear    (VECTOR3D Vec1, VECTOR3D Vec2);
int        GEOVectorDirection    (VECTOR3D Vec1, VECTOR3D Vec2);

/* *** POINT: *** */
VECTOR3D   GEOTriangleNormal     (POINT3D  p1,   POINT3D p2, POINT3D p3);
double     GEOCalcAngle          (POINT3D  p1,   POINT3D  p2, POINT3D p3);

/* *** MATRIX: *** */
H_MATRIX   GEOMakeMatrix         (VECTOR3D n, VECTOR3D o, 
				  VECTOR3D a, VECTOR3D p);
H_MATRIX   GEOCreateMatrix       (double nx, double ny, double nz,
				  double ox, double oy, double oz,
				  double ax, double ay, double az,
				  double px, double py, double pz);
H_MATRIX   GEOPtrMatrix          (double *Values);
H_MATRIX   GEONormMatrix         (H_MATRIX M);
VECTOR3D   GEOMatrixVector       (H_MATRIX M,   int component);
H_MATRIX   GEOSetMatrixVector    (H_MATRIX M,   VECTOR3D Vec, int Component);
H_MATRIX   GEOMatrixProd         (H_MATRIX M1,  H_MATRIX M2);
VECTOR3D   GEOMatrixMult         (H_MATRIX M,   VECTOR3D Vec);
H_MATRIX   GEOMatrixInv          (H_MATRIX M);
dual       GEOMatrixCompare      (H_MATRIX M1,  H_MATRIX M2);
H_MATRIX   GEORotMatrix          (int Axis,   double Angle);
H_MATRIX   GEOMakeRightSys       (VECTOR3D n, VECTOR3D o, VECTOR3D p);
H_MATRIX   GEOMakeRightSys2      (VECTOR3D o, VECTOR3D a, VECTOR3D p);
H_MATRIX   GEOMakeAOFrame        (VECTOR3D o, VECTOR3D a, VECTOR3D p);
H_MATRIX   GEOMakeFrame          (VECTOR3D n);
H_MATRIX   GEOMakeAFrame         (VECTOR3D a);

/* RPY->Frame */
H_MATRIX   GEORPY2Frame          (double px, double py, double pz, 
			          double y,  double p,  double r); 
H_MATRIX   GEOEuler2Frame        (double px, double py, double pz, 
		                  double z1, double y,  double z2);
void       GEOFrame2Euler        (H_MATRIX M, 
				  double *px, double *py, double *pz,
				  double *z1, double *y, double *z2);
void       GEOFrame2RPY          (H_MATRIX M, 
				  double *px, double *py, double *pz,
				  double *y,  double *p,  double *r);
void       GEORPY2Euler          (RPY Pos, EULER *Euler);
RPY        GEOGetRPY             (H_MATRIX M);

/* *** TRANSFORMATION: *** */
H_MATRIX   GEOTranslate          (H_MATRIX M,   VECTOR3D Vec);
H_MATRIX   GEOMakeTrans          (double   dx,  double   dy, double dz);
H_MATRIX   GEOMakeRot            (VECTOR3D Dir, double   Angle);
H_MATRIX   GEOMakeTransform      (H_MATRIX M1,  H_MATRIX M2);

/* *** POLYGON: *** */
POLYGON   *GEONewPolygon         (void);
POLYGON   *GEOCopyPolygon        (POLYGON *Polygon);
VECTOR3D   GEOGetPolygonNormal   (POLYGON *Polygon);
dual       GEOCalcHesseForm      (POLYGON *Polygon);
int        GEOTestPolygon        (POLYGON *Polygon);
POINT3D    GEOCalcPolygonCenter  (POLYGON *Polygon);
void       GEOGetPolygonBounds   (POLYGON *Polygon,
				  POINT3D *Min, POINT3D *Max);
POLYGON   *GEOTessellatePolygon   (POLYGON *Polygon);
void       GEOReducePolygon      (POLYGON *Polygon);
double     GEOPointPointDist     (POINT3D  p1, POINT3D p2);
dual       GEOPointInTriangle    (POINT3D  p1, POINT3D p2, POINT3D p3,
				  POINT3D  p);
dual       GEOPointInPolygon     (POLYGON *Polygon, POINT3D p);
dual       GEOPointOnPolygon     (POLYGON *Polygon, POINT3D p);
H_MATRIX   GEOGetTransformXY     (POLYGON *Polygon);
void       GEOTransformPolygon   (POLYGON *Polygon, H_MATRIX Frame);
void       GEOPrintPolygon       (POLYGON *Polygon);
void       GEOWritePolygon       (FILE *fp, POLYGON *Polygon);
void       GEOFreePolygon        (POLYGON *Polygon, dual all);
void       GEOSwapOrientation    (POLYGON *Polygon);
dual       GEOAreNeighbours      (POLYGON *Polygon1, POLYGON *Polygon2);
void       GEOChangePolygonPoint (POLYGON *Polygon, int From, int To);
void       GEOFreePolygonNormals (POLYGON *Polygon);
void       GEOSetVertexNormal    (POLYGON *Polygon, int Vertex, 
				  VECTOR3D Normal);

/* *** POLYEDER: *** */
POLYEDER  *GEONewPolyeder        (void);
POLYEDER  *GEOAddPolygon         (POLYEDER *Polyeder, POLYGON *Polygon);
void       GEODeletePolygon      (POLYEDER *Polyeder, POLYGON *Polygon);
POLYEDER  *GEOCopyPolyeder       (POLYEDER *Polyeder, VECTOR3D *VertexList); 
POLYGON   *GEOGetPolygon         (POLYEDER *Polyeder, int ID);
POLYGON   *GEOFindPolygon        (POLYEDER *Polyeder,
			 	  double x, double y, double z);
dual       GEOGetEdgePolygons    (POLYEDER *Polyeder, int Point1, int Point2,
				  POLYGON **Polygon1, POLYGON **Polygon2);
POLYGON   *GEOFindNeighbour      (POLYEDER *Polyeder, POLYGON *Polygon,
				  int       Index1,   int Index2);
dual       GEOIsEulerPolyeder    (POLYEDER *Polyeder);
INFO_3D    GEOCalcPolyederInfo   (POLYEDER *Polyeder);
void       GEOFreePolyeder       (POLYEDER *Polyeder);
void       GEOBuildPolygonList   (POLYEDER *Polyeder, POINTELEM *PointList);
void       GEOFreePolyederNormals (POLYEDER *Polyeder);
int        GEOGetTriangles       (POLYEDER *Polyeder);


/* *** OPERATIONS: *** */
void       GEOTetraederVolume    (POINT3D p1, POINT3D p2, POINT3D p3,
				  double *Area, double *Volume);
POINT3D    GEOTetraederCenter    (POINT3D   p1, POINT3D p2, POINT3D p3);
dual       GEOFaceLinePoint      (VECTOR3D  n,       double a, POINT3D Pos, 
				  VECTOR3D  Dir,     POINT3D *Point);
double     GEOLinePointDist      (POINT3D   Pos,     VECTOR3D Dir, 
				  POINT3D   Point);
double     GEOAreaPointDist      (POLYGON  *Polygon, POINT3D   Point);
double     GEOLineLineDist       (POINT3D   Pos1,    VECTOR3D  Dir1, 
				  POINT3D   Pos2,    VECTOR3D  Dir2);
dual       GEOLinePointLot       (POINT3D   Pos,     VECTOR3D  Dir, 
				  POINT3D   p,       POINT3D  *Point);
dual       GEOLineLineLot        (POINT3D   Pos1,    VECTOR3D  Dir1, 
				  POINT3D  Pos2,    VECTOR3D  Dir2,
				  POINT3D  *Point1,  POINT3D  *Point2);
int        GEOComputeBubbleCut   (POINT3D   M,       double    r, 
				  POINT3D   Pos,     VECTOR3D  d, 
				  double   *lambda1, double   *lambda2);
void       GEOTriangulatePolygon (POLYGON  *Polygon);

/* *** CONTACT: *** */
int        GEOPointOnEdge        (POINT3D   p1, POINT3D   p2, POINT3D p);
int        GEOEdgeContact        (POINT3D   p1, POINT3D   p2,
				  POINT3D   p3, POINT3D   p4, 
				  VECTOR3D *pA, VECTOR3D *pB);
POLYGON   *GEOPolygonContact     (POLYGON  *Polygon1, POLYGON *Polygon2);
int        GEOPolygonContact2    (POLYGON  *Polygon1, POLYGON *Polygon2, 
				  POLYGON  *Contact);
POLYGON   *GEOPolyederContact    (POLYEDER *Polyeder1, POLYEDER *Polyeder2);
dual       GEOLineIntersection   (POINT3D   Pos1,  VECTOR3D Dir1,
				  POINT3D   Pos2,  VECTOR3D Dir2, 
				  POINT3D  *Point, double *lambda, double *mu);
dual       GEOAreaIntersection   (VECTOR3D  n1,    double    a1, 
				  VECTOR3D  n2,    double    a2,
				  POINT3D  *Point, VECTOR3D *Dir);
int        GEOPolygonLineIntersection (POLYGON *Polygon, 
				       POINT3D  Pos,      VECTOR3D  Dir,
				       POINT3D *Point1,   POINT3D  *Point2,
				       double  *LLambda1, double   *LLambda2,
				       int     *Index1,   int      *Index2,
				       double  *ELambda1, double   *ELambda2);
int        GEOPolygonEdgeIntersection (POLYGON *Polygon, 
				       POINT3D  v1,      POINT3D   v2,
				       POINT3D *Point1,  POINT3D  *Point2);

/* *** Bounds *** */
dual       GEOIntersectBounds    (POINT3D Min1, POINT3D Max1,
				  POINT3D Min2, POINT3D Max2);
dual       GEOPointInBounds      (POINT3D Min, POINT3D Max, POINT3D Point);

/* *** LGS (Linear Program): *** */
void       GEOShowLGS            (char *Name, LGS M);
LGS        GEOSolveLGS           (LGS M);
void       GEOPrintLGS           (LGS M);
int        GEOSolve2DimLGS       (double a1, double b1, double c1,
				  double a2, double b2, double c2,
				  double *a, double *b);
/* *** CSG *** */
INTERSECT  *GEONewIntersection      (POLYGON *Poly1, POLYGON *Poly2);
void        GEOFreeIntersection     (INTERSECT *Intersect);
INTERSECT  *GEOPolygonIntersection  (POLYGON *Poly1, POLYGON *Poly2);

#ifdef __cplusplus
}
#endif

#endif





