#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <unistd.h>
#include <string.h>

#include "libnet.h"


static int  Socket[10] = {-1, -1, -1, -1, -1, -1, -1, -1, -1, -1};
static char Response[SOCKET_BUF_SIZE+1];


/* ********************************************************************** */
/* ** Initializes the client-part of a new socket-connection. This     ** */
/* ** function must be called before sending any command to the        ** */
/* ** server. It opens a socket and waits for establishing the         ** */
/* ** connection to the server.                                        ** */
/* ** Make sure to start the server before connecting to it !          ** */
/* ** INPUT:  HostName   Name of the server-station. If this value is  ** */
/* **                    NULL or "", the local system will be used     ** */
/* **         PortNo     Number of the used port                       ** */
/* ** OUTPUT: (int)      Identifier used to send commands or close the ** */
/* **                    connection.                                   ** */
/* **                    On Error -1 will be returned.                 ** */
/* ** EXAMPLE:                                                         ** */
/* **   int   SocketID;                                                ** */
/* **   ...                                                            ** */
/* **   if ((SocketID = NETInitClient("proxima.galaxy.de", 2020)) < 0) ** */
/* **     printf ("connection failed !\n");                            ** */
/* ********************************************************************** */
int NETInitClient (const char *HostName, int PortNo)
{
         int           SocketID = 0;
  long   int           ServerAddr;
         char          Name[256];
  struct sockaddr_in   Server;
  struct hostent      *Host = NULL;

  PRO_START ("\n")

  /* locate free socket-var */
  while (SocketID < 10 && Socket[SocketID] >= 0)
    SocketID++;
  if (SocketID == 10)
    PRO_RETURN ("All sockets used",-1)

  if (HostName != NULL && HostName[0] != '\0')
    strcpy (Name, HostName);
  elsif (gethostname(Name, (size_t)(sizeof(Name)/sizeof(char))) != 0)
    PRO_RETURN ("Invalid host-name", -1)

  PRO1 ("Host:   %s\n", Name)
  PRO1 ("PortNo: %d\n", PortNo)

  if ((Host = gethostbyname(Name)) == 0)
    PRO_RETURN ("Unknown host", -1)

  /* create socket */
  if ((Socket[SocketID] = socket(NET_PROTOCOL, NET_TYPE, 0)) < 0 ) 
    PRO_RETURN ("Socket creation failed", -1)

  Server.sin_family = NET_PROTOCOL;
  memcpy((char *)&Server.sin_addr, (char *)Host->h_addr, Host->h_length);
  Server.sin_port = htons (PortNo);

  PRO2 ("Connecting %s  Port: %d...\n", Name, PortNo);
  if (connect(Socket[SocketID], (struct sockaddr*)&Server, sizeof(Server)) < 0)
  {
    NETCloseClient (SocketID);
    PRO_RETURN ("Cannot connect", -1)
  }

  ServerAddr = ntohl((long int)Server.sin_addr.s_addr);
  PRO4 ("Server: %d.%d.%d.%d\n", 
	NET_IP_ADDR1(ServerAddr), NET_IP_ADDR2(ServerAddr), 
	NET_IP_ADDR3(ServerAddr), NET_IP_ADDR4(ServerAddr))

#if 0
  /* Switch to NON-Blocking-Mode, if the client should not wait for result */
  fcntl (Socket[SocketID], F_SETFL, O_NONBLOCK);
#endif

  PRO0 ("Client connected successfully !\n")

  PRO_END
  return SocketID;
}


/* ********************************************************************** */
/* ** Read the response of a sent command from server                  ** */
/* ********************************************************************** */
static dual NET_ReadResponse (int SocketID, char *Response)
{
  int len;

#if (NET_LEN_BYTES > 0)
  Response[0] = '\0' ;
  if ((len = recv(Socket[SocketID], Response, 
		  NET_LEN_BYTES, 0)) != NET_LEN_BYTES) 
  {
    PRO2 ("CLIENT: Error reading length: %d -> %s\n", len, Response)
    close (Socket[SocketID]);
    return FALSE;
  }

  Response[NET_LEN_BYTES] = '\0';
  len = atoi (Response);

  if (recv(Socket[SocketID], Response, len, 0) != len) 
  {
    PRO0 ("CLIENT: Error reading response\n")
    close (Socket[SocketID]);
    return FALSE;
  }
#else
  len = recv (Socket[SocketID], Response, SOCKET_BUF_SIZE, 0);
#endif
  Response[len] = '\0';
  
  PRO1 ("RESPONSE: '%s'\n", Response)

  return TRUE;
}


/* ********************************************************************** */
/* ** Sends a command-string to the Server and waits for response. The ** */
/* ** Connection must be opened by NETInitClient. The maximum length of ** */
/* ** the command is defined by the 'SOCKET_BUF_SIZE'-constant         ** */
/* ** INPUT:  SocketID  Identifier of the socket-connection.           ** */
/* ** INOUT:  Command   Command, which should be sent to the server    ** */
/* **                   and also space for the response-string.        ** */
/* **                   On error, an empty string will be returned     ** */
/* ** OUTPUT: (dual)    state of execution: TRUE,  on success,         ** */
/* **                                       FALSE, if an error occured ** */
/* ** EXAMPLE:                                                         ** */
/* **   int    SocketID;                                               ** */
/* **   char   Buffer[SOCKET_BUF_SIZE];                                ** */
/* **   ...                                                            ** */
/* **   SocketID = NETInitClient (NULL, 2020);                         ** */
/* **   strcpy (Buffer, "This is a sample-command");                   ** */
/* **   if (! NETSendCommand(SocketID, Buffer))                        ** */
/* **     ...                                                          ** */
/* ********************************************************************** */
dual NETSendCommand (int SocketID, char *Msg)
{
  dual Result;
  int  len;
#if (NET_LEN_BYTES > 0)
  char Format[20];
#endif
  char Buffer[SOCKET_BUF_SIZE+1];

  PRO_START ("\n")

  if (SocketID < 0 || SocketID >= 10 || Socket[SocketID] < 0)
    PRO_RETURN ("Socket not open", FALSE)

  PRO1 ("SEND: '%s'\n", Msg)

#if (NET_LEN_BYTES > 0)
    sprintf (Format, "%%%dd%%s", NET_LEN_BYTES);  
    sprintf (Buffer, Format, strlen(Msg)+1, Msg);
#else
    strcpy (Buffer, Msg);
#endif

  len = strlen (Buffer) + 1;

  if (send(Socket[SocketID], Buffer, len, 0) != len)
  {
    NETCloseClient (SocketID);
    PRO_RETURN ("Socket aborted by peer during write.", FALSE)
  }

  Result = NET_ReadResponse (SocketID, Msg);

  PRO_END
  return Result;
}


char *NETSend (int SocketID, char *Msg, ...)
{
  va_list Arguments;

  PRO_START ("\n")

  va_start (Arguments, Msg);
  vsprintf (Response, Msg, Arguments);
  va_end   (Arguments);

  if (! NETSendCommand(SocketID, Response))
    return NULL;

  PRO_END
  return Response;
}  


/* ********************************************************************** */
/* ** This function closes an open socket-connection.                  ** */
/* ** INPUT:  SocketID   Identifier of the connection                  ** */
/* **                    (return of NETInitClient)                      ** */
/* ** OUTPUT: (dual)     TRUE on success, FALSE if wrong SocketID was  ** */
/* **                    selected.                                     ** */
/* ********************************************************************** */
dual NETCloseClient (int SocketID)
{
  PRO_START ("\n")

  if (SocketID < 0 || SocketID >= 10 || Socket[SocketID] < 0)
    PRO_RETURN ("No socket open", FALSE)

  shutdown (Socket[SocketID], 0);

  close (Socket[SocketID]);

  Socket[SocketID] = -1;

  PRO_END
  return TRUE;
}
