#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <errno.h>
#include <netinet/in.h>
#include <netdb.h>
#include <unistd.h>
#include <string.h>

#include "libnet.h"


typedef struct
{
         int       Socket;
  struct sockaddr *Client;
} CLIENT_DATA;


typedef struct
{
  int              Socket;
  SERVER_FUNCTION  Function;
  CLIENT_DATA      Client;
  int              Connections;
  int              used;
} SERVER_DATA;


typedef struct
{
  int SocketID;
  int ClientID;
} CONNECT_DATA;


static SERVER_DATA Data[NET_MAX_SERVER] = 
{
  {-1, NULL, {-1, NULL}, NET_MAX_CONNECTIONS, 0},
  {-1, NULL, {-1, NULL}, NET_MAX_CONNECTIONS, 0},
  {-1, NULL, {-1, NULL}, NET_MAX_CONNECTIONS, 0},
  {-1, NULL, {-1, NULL}, NET_MAX_CONNECTIONS, 0},
  {-1, NULL, {-1, NULL}, NET_MAX_CONNECTIONS, 0},
  {-1, NULL, {-1, NULL}, NET_MAX_CONNECTIONS, 0},
  {-1, NULL, {-1, NULL}, NET_MAX_CONNECTIONS, 0},
  {-1, NULL, {-1, NULL}, NET_MAX_CONNECTIONS, 0},
  {-1, NULL, {-1, NULL}, NET_MAX_CONNECTIONS, 0},
  {-1, NULL, {-1, NULL}, NET_MAX_CONNECTIONS, 0}
};


/* ********************************************************************** */
/* ** This function initializes the server-part of a socket-connection ** */
/* ** It waits until a client connects this server via the given port- ** */
/* ** number. While the server is waiting for commands, it will be non ** */
/* ** active.                                                          ** */
/* ** INPUT: HostName  Name of the host-system. If NULL or "", the     ** */
/* **                  local host will be used.                        ** */
/* **        PortNo    Number of the port, 0 for auto-selection        ** */
/* ** OUTPUT (int)     Number of an internal socket-identifier used to ** */
/* **                  communicate with client or -1, if connection    ** */
/* **                  fails.                                          ** */
/* ** EXAMPLE:                                                         ** */
/* **   int SocketID;                                                  ** */
/* **   ...                                                            ** */
/* **   SocketID = NETInitServer ("challenger.galaxy.de", 2020);       ** */
/* **   if (SocketID < 0)                                              ** */
/* **     ...                                                          ** */
/* ********************************************************************** */
int NETInitServer (const char *HostName, int PortNo)
{
         int          SocketID = 0, AddrLen;
         char         Name[256];
  struct sockaddr_in  Server;
  struct hostent     *Host = NULL;

  PRO_START ("\n")

  /* locate free socket-var                                               */
  while (SocketID < NET_MAX_SERVER && Data[SocketID].Socket >= 0)
    SocketID++;
  if (SocketID == NET_MAX_SERVER)
    PRO_RETURN ("All sockets used",-1)

  /* decide, which host to use                                            */
  if (HostName != NULL && HostName[0] != '\0')
    strcpy (Name, HostName);
  elsif (gethostname(Name, (size_t)(sizeof(Name)/sizeof(char))) != 0) 
    PRO_RETURN ("Invalid host-name", -1)

  PRO1 ("Host:   %s\n", Name)
  PRO1 ("PortNo: %d\n", PortNo)

  if ((Host = gethostbyname(Name)) == 0)
    PRO_RETURN ("SERVER: Unkown host", -1)

  if (getdomainname(Name, (size_t)sizeof(Name)/sizeof(char)) == 0)
    PRO1 ("Domain: %s\n", Name)

  if ((Data[SocketID].Socket = socket(NET_PROTOCOL, NET_TYPE, 0)) < 0)
    PRO_RETURN ("Socket creation failed", -1)

  Server.sin_family = NET_PROTOCOL;
  memcpy ((char *)&Server.sin_addr, (char *)Host->h_addr, Host->h_length);
  Server.sin_port = htons (PortNo);

  PRO0 ("Binding...")
  AddrLen = sizeof (Server);
  if (bind(Data[SocketID].Socket, (struct sockaddr *)&Server, AddrLen) < 0)
  {
    PRO0 ("\\\n")
    switch (errno)
    {
      case EBADF :
	PRO_RETURN ("SERVER: Invalid descriptor", -1)
	break;
      case EINVAL :
	PRO_RETURN ("SERVER: Socket already bound", -1)
        break;
      case EACCES :
	PRO_RETURN ("SERVER: Address is protected", -1)
        break;
      default :
	PRO_RETURN1 ("SERVER: Binding failed: %d", errno, -1)
    }
    NETCloseServer (SocketID);
  }

  /* Print out assigned port number */
  if (PortNo == 0)
  {
    AddrLen = sizeof (Server);
    if (getsockname(Data[SocketID].Socket, 
		    (struct sockaddr *)&Server, (unsigned int *)&AddrLen) < 0)
    {
      PRO0 ("Error getting socket name\n")
      NETCloseServer (SocketID);
    }
    printf ("Socket has port: #%d\n", ntohs(Server.sin_port));
  }


  PRO0 ("\\Listening...")
  if (listen(Data[SocketID].Socket, NET_MAX_CONNECTIONS) < 0)
  {
    NETCloseServer (SocketID);
    switch (errno)
    {
      case EBADF :    
	   PRO_RETURN ("SERVER: Invalid descriptor", -1)
	   break;
      case ENOTSOCK :
	   PRO_RETURN ("SERVER: Invalid socket", -1)
	   break;
      case EOPNOTSUPP :
	   PRO_RETURN ("SERVER: listen not supported", -1)
	   break;
      default : 
	   PRO_RETURN ("SERVER: Error while listing", -1)
    }
  }

  /* 
     Set non-blocking mode:
     fcntl (Socket[SocketID], F_SETFL, O_NONBLOCK); 
  */

  PRO_END
  return SocketID;
}


/* ********************************************************************** */
/* ** Server-loop-function, which waits for incomming commands         ** */
/* ** and starts the specified user-function if one arrives.           ** */
/* ********************************************************************** */
static int NET_ServerLoop (int SocketID)
{
         int           len;
         unsigned int  FromLen;
         char          Buffer[SOCKET_BUF_SIZE+1] = "";
         char         *Response = NULL;
  struct sockaddr      From;

  /* TODO: Mutex_lock ... */
  (Data[SocketID].used)++;
  PRO2 ("# CONNECTIONS: %d (%d)\n", SocketID, Data[SocketID].used)

  do
  {
    PRO0 ("SERVER: ")

#if (NET_LEN_BYTES > 0)
    if ((len = recvfrom(Data[SocketID].Client.Socket, Buffer, 
			NET_LEN_BYTES, 0, &From, &FromLen)) == NET_LEN_BYTES)
    {
      Buffer[NET_LEN_BYTES] = '\0';
      len = atoi (Buffer);

      PRO1 ("\\#%d: ", len)
      if (recvfrom(Data[SocketID].Client.Socket, Buffer, len, 0, 
		   &From, &FromLen) != len)
	PRO_RETURN ("Error reading data", 0)

      Buffer[len] = '\0';
      PRO1 ("\\'%s'\n", Buffer)

      if (Data->Function != NULL)
      {
	Response = Data->Function (SocketID, Buffer);
	if (Response == NULL)
	  Buffer[0] = '\0';
	else
	  sprintf (Buffer, "%5d%s", strlen(Response)+1, Response);
      } 
      else
	strcpy (Buffer, "    1");
#else
    if ((len = recvfrom(Data[SocketID].Client.Socket, Buffer, 
			SOCKET_BUF_SIZE, 0, &From, &FromLen)) > 0)
    {
      Buffer[len] = '\0';
      PRO1 ("\\'%s'\n", Buffer)

      if (Data->Function != NULL)
      {
	Response = Data->Function (SocketID, Buffer);
	if (Response == NULL)
	  Buffer[0] = '\0';
	else
	  strcpy (Buffer, Response);
      }
      else
	Buffer[0] = '\0';
#endif
	
      PRO1 ("Response: %s\n", Buffer)

      len = strlen (Buffer) + 1;
      if (send(Data[SocketID].Client.Socket, Buffer, len, 0) != len)
	PRO_RETURN ("SERVER: Error while writing", 0)
    }
    else /* if there's no data: close this connection */
    {
      Data[SocketID].used--;
      PRO1 ("\\No data received: exiting #%d\n", SocketID)
    }
  } while (len != 0);

  return 1;
}


/* ********************************************************************** */
/* ** Accept-function waits for incomming connections-requests         ** */
/* ********************************************************************** */
static int NET_Accept (int SocketID)
{
         unsigned int AddrLen;
	 fd_set       receive;
  struct timeval      TimeOut; 

  PRO_START ("\n")

  if (Data[SocketID].used >= abs(Data[SocketID].Connections))
    PRO_RETURN ("max. connections reached", -1)

  PRO0 ("\\Accepting...\n")

  do
  {
    FD_ZERO (&receive);
    FD_SET (Data[SocketID].Socket, &receive);
    TimeOut.tv_sec  = 5;
    TimeOut.tv_usec = 0;
    if (select(Data[SocketID].Socket+1, &receive, 0, 0, &TimeOut) < 0)
    {
      printf ("SELECT-ERROR\n");
      continue;
    }
    if (FD_ISSET(Data[SocketID].Socket, &receive))
    {
      if ((Data[SocketID].Client.Socket = 
	   accept(Data[SocketID].Socket, (struct sockaddr*)0, &AddrLen)) < 0)
      {
	switch (errno)
	{
          case EBADF :
	    PRO1 ("%d: Invalid descriptor", SocketID)
	    break;
          case ENOTSOCK :
	    PRO1 ("%d: file-descriptor specified", SocketID)
	    break;
          case EOPNOTSUPP :
	    PRO1 ("%d: Not a socket-stream", SocketID)
	    break;
          case EFAULT :
	    PRO1 ("%d: Not in writeable address-space", SocketID)
	    break;
          case EWOULDBLOCK :
	    PRO1 ("%d: No Connections are present (non-blocking)", SocketID)
	    break;
          default :
	    PRO1 ("%d: Error accepting\n", SocketID)
        }
	PRO_RETURN ("accept-error", -1)
      }
      else
      {
	PRO0 ("connection initialized !\n")
	NET_ServerLoop (SocketID);
#ifdef NET_THREADS
	PRO0 ("THREAD STOPPED !\n")
	NETCloseServer (SocketID);    /* close this connection          */
        pthread_exit (NULL);
#endif
      }
    }
  } while (TRUE);

  PRO0 ("SERVER: Connection closed by client\n")

  PRO_END
  return 1;
}


/* ********************************************************************** */
/* ** Function which registers a function, which will be called, if    ** */
/* ** the server receives a command from client. It needs to  be       ** */
/* ** after a socket-connection has been established.                  ** */
/* ** This function returns, if the connection will be closed by an    ** */
/* ** Invalid command or bz the NETCloseServer-function                 ** */
/* ** INPUT:  SocketID      Identifier of the used socket-connection   ** */
/* **         Function      Function with two parameters (int) and     ** */
/* **                       (char*). The int-value represents the      ** */
/* **                       socketID the string contains the message.  ** */
/* **                       The ID may be used to differ between       ** */
/* **                       multiple connections.                      ** */
/* **         Connections   neg. Value: server runs in background      ** */
/* **                       pos. Value: server runs in foreground      ** */
/* **                       Amount: amount of simultaneous connections ** */
/* ** OUTPUT: (dual)    TRUE, when connection was closed               ** */
/* **                   FALSE, if the given socket-ID is invalid.      ** */
/* ** EXAMPLE:                                                         ** */
/* **   static char Response[SOCKET_BUF_SIZE];                         ** */ 
/* **   ...                                                            ** */
/* **   static char *UserFunction (int SocketID, char *Command)        ** */
/* **   {                                                              ** */
/* **     printf ("Command: %s\n", Command);                           ** */
/* **     strcpy (Response, "ok");                                     ** */
/* **     return Response;                                             ** */
/* **   }                                                              ** */
/* **                                                                  ** */
/* **   SocketID = NETInitServer (...                                  ** */
/* **   NETRunServer (SocketID, UserFunction, 2);                      ** */
/* ********************************************************************** */
int NETRunServer (int SocketID, SERVER_FUNCTION Function, int MaxConnections)
{
  int Result = 1;

  PRO_START ("\n")

  if (SocketID < 0 || SocketID >= NET_MAX_SERVER)
    PRO_RETURN ("Invalid Socket-ID", -1)

  if (MaxConnections == 0)
    PRO_RETURN ("No connections allowed", -1)

  Data[SocketID].Function    = Function;
  Data[SocketID].Connections = MaxConnections;
  Data[SocketID].used        = 0;

  Result = NET_Accept (SocketID);        /* Serve incomming calls          */
  NETCloseServer (SocketID);             /* close Server                   */

  PRO_END
  return Result;
}


/* ********************************************************************** */
/* ** Closes an open socket-connection. This function may be called    ** */
/* ** within the user-function (e.g. if an 'exit'-command was          ** */
/* ** received)                                                        ** */
/* ** INPUT:  SocketID  Identifier of the closeable connection         ** */
/* ** OUTPUT: (dual)    TRUE, on success, FALSE, if SocketID is not    ** */
/* **                   valid.                                         ** */
/* ********************************************************************** */
dual NETCloseServer (int SocketID)
{
  PRO_START (" ")

  if (SocketID < 0 || SocketID >= NET_MAX_SERVER || Data[SocketID].Socket < 0)
    PRO_RETURN ("Invalid socket-ID", FALSE)

  shutdown (Data[SocketID].Socket, 0);

  close (Data[SocketID].Client.Socket);
  Data[SocketID].Socket = -1;

  PRO_END
  return TRUE;
}


/* ********************************************************************** */
/* ** This function closes all open server-connections. This effects   ** */
/* ** to close the coresponding client-connections, too.               ** */
/* ** INPUT:  ---                                                      ** */
/* ** OUTPUT: ---                                                      ** */
/* ********************************************************************** */
void NETCloseAllServers (void)
{
  int count;

  for (count = 0; count < NET_MAX_SERVER; count++)
    if (Data[count].Socket >= 0)
      NETCloseServer (count);
}


#ifdef NET_THREADS
/* ********************************************************************** */
/* ** Same as NETRunServer, but start a threaded server. So, this      ** */
/* ** function returns emidiately, running the server as new thread.   ** */
/* ** On success a pointer to the created posix thread will be         ** */
/* ** returned.                                                        ** */
/* ********************************************************************** */
pthread_t *NETStartServer (int SocketID, SERVER_FUNCTION Function, 
			   int MaxConnections)
{
  pthread_t *NetThread = NULL;

  PRO_START ("\n")

  if (SocketID < 0 || SocketID >= NET_MAX_SERVER)
    PRO_RETURN ("Invalid Socket-ID", NULL)

  if (MaxConnections == 0)
    PRO_RETURN ("No connections allowed", NULL)

  Data[SocketID].Function    = Function;
  Data[SocketID].Connections = MaxConnections;
  Data[SocketID].used        = 0;

  NetThread = (pthread_t *)malloc (sizeof(pthread_t));
  if (pthread_create(NetThread, (pthread_attr_t *)NULL, 
		     (void *(*)(void *))NET_Accept, (void *)SocketID) != 0)
    PRO_RETURN ("error creating thread", NULL)

  PRO0 ("THREAD STARTED...\n")

  PRO_END
  return NetThread;
}


dual NETStopServer (pthread_t *NetThread)
{
  PRO_START ("");

  if (NetThread == NULL)
    PRO_RETURN ("Invalid server thread", FALSE)

  pthread_exit (NetThread);

  PRO_END
  return TRUE;
}

#endif

