#ifndef LIBOBJ_H
#define LIBOBJ_H

#include <libgeo.h>

#ifdef __cplusplus
extern "C" {
#endif


enum OBJECT_TYPES
{
  NONE = -1,
  PART_PLANE,
  PART_CURVE
};


enum CSG_OPERATION
{
  CSG_UNION = 0,
  CSG_INTERSECTION,
  CSG_DIFFERENCE
};


/* ********************************************************************** */
/* ** structure for storing object-data                                ** */
/* ********************************************************************** */
typedef struct OBJECT
{
         int           ObjNo;           /* Obj-Number (used by DXF)       */
         int           ColorIndex;      /* Color-index                    */

         unsigned int  VertexNo;        /* amount of vertexes             */
         POINT3D      *VertexList;      /* pointer to vertex-list         */

         char          Type;            /* object-type: curve/polyeder    */
         union         {
                         LINES    *Curve;
                         POLYEDER *Polyeder;
	               } Data;          /* object-data                    */


         INFO_3D       Info;            /* object-informations            */

  struct OBJECT       *Next;            /* pointer to next object         */
} OBJECT;


OBJECT   *OBJNewObject          (void);
void      OBJFreeObject         (OBJECT *Object);
OBJECT   *OBJCopyObject         (OBJECT *Object);
dual      OBJAddPolygon         (OBJECT *Object, POLYGON *Polygon);
OBJECT   *OBJAddObject          (OBJECT *Object, OBJECT *NewObj);
int       OBJAddVertex          (OBJECT *Object, POINT3D Point);

void      OBJPlace              (OBJECT *Object, H_MATRIX WTO);
void      OBJScale              (OBJECT *Object, 
				 double sx, double sy, double sz);
INFO_3D   OBJCalcInfo           (OBJECT *Object);
dual      OBJIsEuler            (OBJECT *Object);
int       OBJGetTriangles       (OBJECT *Object);

int       OBJGetIGRIPColor      (int Color);         /* From IGRIP-color */
int       OBJSetIGRIPColor      (int Color);         /* To   IGRIP-color */

OBJECT   *OBJLoadV11            (FILE *fp, int Version, 
				 int *PolygonID, dual check);
OBJECT   *OBJLoadV12            (FILE *fp, int *PolygonID, dual check);
dual      OBJWriteV11           (FILE *fp, OBJECT *Obj, dual triangle);
dual      OBJWriteV12           (FILE *fp, OBJECT *Obj);


OBJECT   *OBJCreateArrow        (H_MATRIX Matrix, 
				 double   Radius1, double Len1, 
				 double   Radius2, double Len2, 
				 double   Angle,   int    Facettes, 
				 double   Segment, int    Color);

OBJECT   *OBJCreateBolid        (H_MATRIX Matrix, 
				 double   Radius1,   double Radius2, 
				 double   Radius3,   double SizeZ, 
				 int      Facettes1, int    Facettes2, 
				 double   Segment,   int    Color);

OBJECT   *OBJCreateCone         (H_MATRIX Matrix,
				 double   Radius1, double Radius2, 
				 double   Radius3, double SizeZ, 
				 double   Segment, int    Facettes, int Color);

OBJECT   *OBJCreateCylinder     (H_MATRIX Matrix,
				 double   Radius,  double Depth, 
				 double   Segment, int    Facettes, int Color);

OBJECT   *OBJCreateGear         (H_MATRIX Matrix,
				 double Radius1, double Radius2, 
				 double Radius3, double Depth1,  double Depth2,
				 int    Teeth,   double Skew,    int    Color);

OBJECT   *OBJCreateIcosahedron  (H_MATRIX Matrix,
				 double Radius1, double Radius2, 
				 int    Density, int    Iterations, int Color);

OBJECT   *OBJCreatePipe         (H_MATRIX Matrix,
				 double   Radius1,    double Radius2, 
				 double   Thickness1, double Thickness2,
				 double   SizeZ, 
				 double   Segment,    int Facettes, int Color);

OBJECT   *OBJCreatePrism        (H_MATRIX Matrix, 
				 double Width1,  double Width2,  double OffsX,
				 double Depth1,  double Depth2,  double OffsY,
				 double Height1, double Height2, double OffsZ,
				 int Color);

OBJECT   *OBJCreateQuader       (H_MATRIX Matrix,
				 double   SizeX, double SizeY, double SizeZ, 
				 int      Color);

OBJECT   *OBJCreateRot          (H_MATRIX Matrix,
				 int      Number, POINT2D *Points, 
				 double   Segment, int Facettes,
				 double   Radius,  int Color);

OBJECT   *OBJCreateScrew        (H_MATRIX Matrix,
				 double   Radius1,   double Radius2, 
				 double   Height,    double Increase, 
				 double   Offset,    int    Facettes, 
				 dual     Direction, int    Color);

OBJECT   *OBJCreateSlice        (H_MATRIX Matrix, 
				 double   Radius1,   double Radius2,
				 double   Radius3,   double SizeZ, 
				 int      Facettes1, int    Facettes2, 
				 double   Segment,   int    Color);

OBJECT  *OBJCreateSphere        (H_MATRIX Matrix, 
				 double   Radius, 
				 double   SegmentX, double SegmentZ,
				 int      Facettes, int    Color);

OBJECT  *OBJCreateSpring        (H_MATRIX Matrix,
				 double   Radius1,   double Radius2, 
				 double   Radius3,
				 double   Height,    double Increase,
				 int      Facettes1, int    Facettes2, 
				 dual     Direction, int    Color);

OBJECT   *OBJCreateSweep        (H_MATRIX Matrix,
	  /* Extrusion */	 int      Number, POINT2D *Points, 
				 double Height,   int      Color);

OBJECT   *OBJCreateTorus        (H_MATRIX Matrix,
				 double   Radius1,   double Radius2, 
				 int      Facettes1, int    Facettes2,
				 double   Segment,   int    Color);

OBJECT   *OBJCreateWedge        (H_MATRIX Matrix,
				 double   Width1,  double Width2,
				 double   Height1, double Height2,
				 double   Depth1,  double Depth2, int Color);

/* CSG-Operations: */
OBJECT   *OBJIntersection       (OBJECT *Object1, H_MATRIX WTO1,
				 OBJECT *Object2, H_MATRIX WTO2, int Color);
OBJECT   *OBJDifference         (OBJECT *Object1, H_MATRIX WTO1, 
				 OBJECT *Object2, H_MATRIX WTO2, int Color);
OBJECT   *OBJUnion              (OBJECT *Object1, H_MATRIX WTO1, 
				 OBJECT *Object2, H_MATRIX WTO2, int Color);
INTERSECT *OBJGetIntersections  (OBJECT *Object1, H_MATRIX WTO1, 
				 OBJECT *Object2, H_MATRIX WTO2);

#ifdef __cplusplus
}
#endif

#endif
