#include <math.h>
#include <stdarg.h>

#include "libobj.h"


#define ICO_X 0.525731112119133606
#define ICO_Z 0.850650808352039932


static POINT3D IcoData[12] =
{
  {-ICO_X,    0.0,  ICO_Z},
  { ICO_X,    0.0,  ICO_Z},
  {-ICO_X,    0.0, -ICO_Z},
  { ICO_X,    0.0, -ICO_Z},
  {   0.0,  ICO_Z,  ICO_X},
  {   0.0,  ICO_Z, -ICO_X},
  {   0.0, -ICO_Z,  ICO_X}, 
  {   0.0, -ICO_Z, -ICO_X},
  { ICO_Z,  ICO_X,    0.0},
  {-ICO_Z,  ICO_X,    0.0}, 
  { ICO_Z, -ICO_X,    0.0}, 
  {-ICO_Z, -ICO_X,    0.0}
};

static int TIndices[20][3] = 
{
  {0, 4, 1}, {0,9, 4}, {9, 5,4}, { 4,5,8}, {4,8, 1},
  {8,10, 1}, {8,3,10}, {5, 3,8}, { 5,2,3}, {2,7, 3},
  {7,10, 3}, {7,6,10}, {7,11,6}, {11,0,6}, {0,1, 6},
  {6, 1,10}, {9,0,11}, {9,11,2}, { 9,2,5}, {7,2,11}
};

static int PrismData[8][4] = 
{
  {3, 2, 1, 0},
  {0, 1, 5, 4},
  {1, 2, 6, 5},
  {2, 3, 7, 6},
  {3, 0, 4, 7},
  {4, 5, 6, 7},
  {0, 0, 0, 0},  /* to be continued ? */
  {0, 0, 0, 0}
};

static int ScrewIndex[2][2][4] =
{
  { /* left-sided winding */
    { 0,  1,  3, 2},  /* lower side */
    { 1, -3, -1, 3}   /* upper side */
  },
  { /* right-sided winding*/
    { 2,  3,  1, 0},  /* lower side */
    { 3, -1, -3, 1}   /* upper side */
  }
};

static int QuaderData[6][4] =
{
  {3, 2, 1, 0},
  {1, 5, 4, 0},
  {2, 6, 5, 1},
  {3, 7, 6, 2},
  {0, 4, 7, 3},
  {5, 6, 7, 4}
};



static POLYGON *OBJ_MakeTriangle (int ID, POINT3D *VertexList, 
				  int Idx1, int Idx2, int Idx3, dual order)
{
  POLYGON *Polygon;

  Polygon = GEONewPolygon ();
  Polygon->Points = VertexList;

  Polygon->Number = 3;
  Polygon->Vertex = (int *)malloc (sizeof(int) * Polygon->Number);
  Polygon->Vertex[0] = (order)? Idx1:Idx3;
  Polygon->Vertex[1] = Idx2;
  Polygon->Vertex[2] = (order)? Idx3:Idx1;
  Polygon->Info.ID = ID;
  Polygon->Info.Center = GEOCalcPolygonCenter (Polygon);

  GEOCalcHesseForm (Polygon);
  Polygon->Triangle = (TRIANGLE *)malloc (sizeof(TRIANGLE));
  Polygon->Triangle[0][0] = Polygon->Vertex[0];
  Polygon->Triangle[0][1] = Polygon->Vertex[1];
  Polygon->Triangle[0][2] = Polygon->Vertex[2];

  return Polygon;
}


static POLYGON *OBJ_MakePolygon (int ID, int Number, 
				 POINT3D *VertexList, int Offset, dual order)
{
  int      count;
  POLYGON *Polygon;

  Polygon = GEONewPolygon ();
  Polygon->Points = VertexList;

  /* If start- and end-point are equal, kill one of them */
  if (VEC_EQUAL(Polygon->Points[Offset], Polygon->Points[Offset+Number-1]))
    Number--;

  Polygon->Number = Number;
  Polygon->Vertex = (int *)malloc (sizeof(int) * Polygon->Number);
  for (count = 0; count < Number; ++count)
    Polygon->Vertex[count] = Offset + ((order)? count:(Number-count-1));
  Polygon->Info.ID = ID;
  Polygon->Info.Center = GEOCalcPolygonCenter (Polygon);

  GEOCalcHesseForm      (Polygon);
  GEOTriangulatePolygon (Polygon);

  return Polygon;
}


static POLYGON *OBJ_VaMakePolygon (int ID, POINT3D *VertexList, 
				   int Number, ...)
{
  int      count;
  va_list  Arguments;
  POLYGON *Polygon;
  
  Polygon = GEONewPolygon ();
  Polygon->Points = VertexList;

  Polygon->Number = Number;
  Polygon->Vertex = (int *)malloc (sizeof(int) * Polygon->Number);

  va_start (Arguments, Number);
  for (count = 0; count < Number; ++count)
    Polygon->Vertex[count] = (int)va_arg (Arguments, int);
  va_end (Arguments);

  Polygon->Info.ID = ID;
  Polygon->Info.Center = GEOCalcPolygonCenter (Polygon);

  GEOCalcHesseForm (Polygon);
  GEOTriangulatePolygon (Polygon);

  return Polygon;  
}


/* When creating parts we often create convex polygons with four points.
   Because we know their orientation and convexity we can use a faster
   method to triangulate them than GEOTriangulate 
*/
static dual OBJ_TriangulateQuad (POLYGON *Polygon)
{
  if (Polygon->Number != 4)
    return FALSE;

  Polygon->Triangle = (TRIANGLE *)malloc(sizeof(TRIANGLE) * 2);
  Polygon->Triangle[0][X] = Polygon->Vertex[0];
  Polygon->Triangle[0][Y] = Polygon->Vertex[1];
  Polygon->Triangle[0][Z] = Polygon->Vertex[2];
  Polygon->Triangle[1][X] = Polygon->Vertex[0];
  Polygon->Triangle[1][Y] = Polygon->Vertex[2];
  Polygon->Triangle[1][Z] = Polygon->Vertex[3];

  return TRUE;
}


/* *********************************************************************** */


OBJECT *OBJCreateArrow (H_MATRIX Matrix, 
			double   Radius1, double Len1, 
			double   Radius2, double Len2, 
			double   Angle,   int    Facettes, 
			double   Segment, int    Color)
{
  POINT2D  Points[5];
  OBJECT  *Obj;

  PRO_START ("\n")

  if (Angle == 0.0 || Angle == 180.0)
    Angle = 90.0;

  Points[0].x = 0.0;
  Points[0].y = 0.0;
  Points[1].x = Radius1;
  Points[1].y = 0.0;
  Points[2].x = Radius1;
  Points[2].y = Len1;
  Points[3].x = Radius2;
  Points[3].y = Len1 - cos(RAD(Angle)) * (Radius2-Radius1) / sin(RAD(Angle));
  Points[4].x = 0.0;
  Points[4].y = Len1+Len2;

  Obj = OBJCreateRot (Matrix, 5, Points, Segment, Facettes, 0.0, Color);
  
  PRO_END
  return Obj;
}


OBJECT *OBJCreateBolid (H_MATRIX Matrix, 
			double   Radius1,   double Radius2, 
			double   Radius3,   double SizeZ, 
			int      Facettes1, int    Facettes2, 
			double   Segment,   int    Color)
{
  int      Index = 0, count;
  POINT2D *Points;
  OBJECT  *Obj;

  PRO_START ("\n")

  Points = (POINT2D *)malloc (sizeof(POINT2D) * (Facettes1+4));

  if (Radius1 != Radius2)
  {
    Points[Index].x = Radius1;
    Points[Index].y = 0.0;
    Index++;
  }

  for (count = 0; count < Facettes1; ++count)
  {
    Points[Index].x = Radius2 +
                      (Radius3-Radius2) * sin(RAD(count*180.0/(Facettes1-1)));
    Points[Index].y = count * SizeZ / (Facettes1-1);
    Index++;
  }
  if (Radius1 != Radius2)
  {
    Points[Index].x = Radius1;
    Points[Index].y = SizeZ;
    Index++;
  }
  Points[Index] = Points[0];

  if (Radius1 != Radius2)
    Facettes1 += 2;
    
  if (Radius1 > 0)
    Facettes1++;

  Obj = OBJCreateRot (Matrix, 
		      Facettes1, Points, Segment, Facettes2, 0.0, Color);
  free (Points);

  PRO_END
  return Obj;
}


OBJECT *OBJCreateCone (H_MATRIX Matrix,
		       double   Radius1, double Radius2, 
		       double   Radius3, double SizeZ, 
		       double   Segment, int Facettes, int Color)
{
  int      Number = 5;
  POINT2D  Points[5];
  OBJECT  *Obj;

  PRO_START ("\n")

  Points[0].x = Radius1;  Points[0].y = 0.0;
  Points[1].x = Radius2;  Points[1].y = 0.0;
  Points[2].x = Radius3;  Points[2].y = SizeZ;
  Points[3].x = Radius1;  Points[3].y = SizeZ;
  Points[4] = Points[0];

  if (Radius1 == 0.0)
    Number--;
  if (Radius3 == 0.0)
    Number--;

  Obj = OBJCreateRot (Matrix, Number, Points, Segment, Facettes, 0.0, Color);

  PRO_END
  return Obj;
}


OBJECT *OBJCreateCylinder (H_MATRIX Matrix,
			   double   Radius,  double Depth, 
			   double   Segment, int    Facettes, int Color)
{
  int      count, Number;
  double   Angle;
  POINT2D *Points;
  OBJECT  *Obj = NULL;

  PRO_START ("\n")

  Number = (Segment < 360.0)? (Facettes+2) : Facettes;
  Points = (POINT2D *)malloc (sizeof(POINT2D) * Number);
  for (count = 0; count < Number; ++count)
  {
    Angle = RAD ((double)count * (Segment / (double)Facettes));
    Points[count].x = cos (Angle) * Radius;
    Points[count].y = sin (Angle) * Radius;
  }
  if (Segment < 360.0)
  {
    Points[count-1].x = 0.0;
    Points[count-1].y = 0.0;
  }
  Obj = OBJCreateSweep (Matrix, Number, Points, Depth, Color);

  free (Points);                        /* Free points of base shape */
  
  PRO_END
  return Obj;
}


OBJECT *OBJCreateGear (H_MATRIX Matrix,
		       double   Radius1, double Radius2, 
		       double   Radius3, double Depth1,  double Depth2,
		       int      Teeth,   double Skew,    int    Color)
{
  int           count;
  int           Offset, Index;
  int           Faces, Facettes;
  int           Next, Side;
  double        Angle1, Angle2, Angle3;
  double        PosZ;
  POLYGON      *Polygon;
  POLYEDER     *Polyeder;
  OBJECT       *Obj;

  PRO_START ("\n")

  Angle2 = RAD (360.0 / (2.0*(double)Teeth));
  Angle3 = Angle2 * Skew / 100.0;

  Facettes = 2*Teeth;                   /* Amount of hole-facettes */
  Faces    = 4*Teeth;
  Offset = (Facettes + 4 * Teeth);      /* vertices on each side   */

  Obj = OBJNewObject ();
  Obj->ColorIndex = -Color;
  Obj->VertexNo = 2 * Offset;
  Obj->VertexList = (POINT3D *)malloc (sizeof(POINT3D) * Obj->VertexNo);

  /* Vertices of hole */
  for (count = 0; count < Facettes; ++count)
  {
    Angle1 = RAD ((double)count * 360.0 / (double)Facettes);
    for (Side = 0; Side < 2; ++Side)
    {
      PosZ = (double)Side * Depth1 - 0.5 * Depth1;
      Obj->VertexList[Offset*Side+count] = 
	GEOMakeVector (cos(Angle1) * Radius1, sin(Angle1) * Radius1, PosZ);
    }
  }

  /* Vertices of teeth */
  for (count = 0; count < Teeth ; ++count)
  {
    Angle1 = RAD((double)count * 360.0 / (double)Teeth);
    for (Side = 0; Side < 2; ++Side)
    {
      Index = Side*Offset + Facettes + 4*count;

      PosZ = (double)Side * Depth1 - 0.5 * Depth1;
      Obj->VertexList[Index]   = 
	GEOMakeVector (cos(Angle1+Angle3) * Radius2, 
		       sin(Angle1+Angle3) * Radius2, 
		       PosZ);
      Obj->VertexList[Index+1] = 
	GEOMakeVector (cos(Angle1+Angle2-Angle3) * Radius2, 
		       sin(Angle1+Angle2-Angle3) * Radius2, 
		       PosZ);

      PosZ = (double)Side * Depth2 - 0.5 * Depth2;
      Obj->VertexList[Index+2] = 
	GEOMakeVector (cos(Angle1+Angle2+Angle3) * Radius3, 
		       sin(Angle1+Angle2+Angle3) * Radius3, 
		       PosZ);
      Obj->VertexList[Index+3] = 
	GEOMakeVector (cos(Angle1+2*Angle2-Angle3) * Radius3,
		       sin(Angle1+2*Angle2-Angle3) * Radius3,
		       PosZ);
    }
  }

  /* Transform object coordinates */
  for (count = 0; count < (int)Obj->VertexNo; ++count)
    Obj->VertexList[count] = GEOMatrixMult (Matrix, Obj->VertexList[count]);

  Obj->Type = PART_PLANE;  

  Polyeder = GEONewPolyeder ();
  Polyeder->Number = Facettes + 8 * Teeth;

  /* Faces of hole */
  for (count = 0; count < Facettes; ++count)
  {
    Polygon = GEONewPolygon ();
    Polygon->Points = Obj->VertexList;
    Polygon->Number = 4;
    Polygon->Vertex = (int *)malloc (sizeof(int) * Polygon->Number);
    Polygon->Vertex[0] = (count==Facettes-1)? 0:(count+1);
    Polygon->Vertex[1] = count;
    Polygon->Vertex[2] = Offset + count;
    Polygon->Vertex[3] = Offset + ((count==Facettes-1)? 0:(count+1));

    Polygon->Info.ID = count;
    Polygon->Info.Center = GEOCalcPolygonCenter (Polygon);
    GEOCalcHesseForm      (Polygon);
    GEOTriangulatePolygon (Polygon);

    if (Polyeder->List == NULL)
      Polyeder->List = Polygon;
    else
      Polyeder->Last->Next = Polygon;
    Polyeder->Last = Polygon;
  }

  /* Draw upper and lower side */
  for (count = 0; count < Facettes; ++count)
    for (Side = 0; Side < 2; ++Side)
    {
      Polygon = GEONewPolygon ();
      Polygon->Points = Obj->VertexList;
      Polygon->Number = 4;
      Polygon->Vertex = (int *)malloc (sizeof(int) * Polygon->Number);

      Next = (count & 1);
      if (Side == 0) 
      {
	Polygon->Vertex[0] = count;
	Polygon->Vertex[1] = (count+1) % Facettes;
	Polygon->Vertex[2] = Facettes + ((2*count+1+Next) % Faces);
	Polygon->Vertex[3] = Facettes + 2*count-Next;
      }
      else
      {
	Polygon->Vertex[0] = Offset + Facettes + 2*count-Next;	
	Polygon->Vertex[1] = Offset + Facettes + ((2*count+1+Next) % Faces);
	Polygon->Vertex[2] = Offset + (count+1) % Facettes;
	Polygon->Vertex[3] = Offset + count;
      }

      Polygon->Info.ID = Facettes + count;
      Polygon->Info.Center = GEOCalcPolygonCenter (Polygon);
      GEOCalcHesseForm      (Polygon);
      GEOTriangulatePolygon (Polygon);

      Polyeder->Last->Next = Polygon;
      Polyeder->Last = Polygon;    
    }

  /* Faces on top and bottom of teeth */
  for (count = 0; count < Teeth; ++count)
    for (Side = 0; Side < 2; ++Side)
    {
      Polygon = GEONewPolygon ();
      Polygon->Points = Obj->VertexList;
      Polygon->Number = 4;
      Polygon->Vertex = (int *)malloc (sizeof(int) * Polygon->Number);
      
      if (Side == 0)
      {
	Polygon->Vertex[0] = Facettes + (4*count+4) % Faces;
	Polygon->Vertex[1] = Facettes + 4*count+3;
	Polygon->Vertex[2] = Facettes + 4*count+2;
	Polygon->Vertex[3] = Facettes + 4*count+1;
      }
      else
      {
	Polygon->Vertex[0] = Offset + Facettes + 4*count+1;
	Polygon->Vertex[1] = Offset + Facettes + 4*count+2;
	Polygon->Vertex[2] = Offset + Facettes + 4*count+3;
	Polygon->Vertex[3] = Offset + Facettes + (4*count+4) % Faces;
      }

      Polygon->Info.ID = Facettes + count;
      Polygon->Info.Center = GEOCalcPolygonCenter (Polygon);
      GEOCalcHesseForm      (Polygon);
      GEOTriangulatePolygon (Polygon);

      Polyeder->Last->Next = Polygon;
      Polyeder->Last = Polygon;
    }

  /* faces on side of teeth */
  for (count = 0; count < 4*Teeth; ++count)
  {
    Polygon = GEONewPolygon ();
    Polygon->Points = Obj->VertexList;
    Polygon->Number = 4;
    Polygon->Vertex = (int *)malloc (sizeof(int) * Polygon->Number);
    Polygon->Vertex[0] = Facettes + count;
    Polygon->Vertex[1] = Facettes + ((count+1) % Faces);
    Polygon->Vertex[2] = Offset + Facettes + ((count+1) % Faces);
    Polygon->Vertex[3] = Offset + Facettes + count;

    Polygon->Info.ID = Facettes + count;
    Polygon->Info.Center = GEOCalcPolygonCenter (Polygon);
    GEOCalcHesseForm      (Polygon);
    GEOTriangulatePolygon (Polygon);

    Polyeder->Last->Next = Polygon;
    Polyeder->Last = Polygon;
  }

  /* Insert polyeder into object */
  Obj->Data.Polyeder = Polyeder;
  Obj->Info = OBJCalcInfo (Obj);

  PRO_END
  return Obj;
}


static void OBJ_SubDivide (OBJECT *Obj, int *No, int i1, int i2, int i3, 
			    double Radius1, double Radius2, 
			    int Density, int Iterations)
{
  int       act;
  double    Radius;
  POINT3D   v12, v23, v31;
  POLYGON  *Polygon;
  POLYEDER *Polyeder;
 
  if (Iterations == 0)
  {
    Polyeder = Obj->Data.Polyeder;
    (Polyeder->Number)++;

    Polygon = 
      OBJ_MakeTriangle (Polyeder->Number, Obj->VertexList, i1, i2, i3, TRUE);
    if (Polyeder->List == NULL)
      Polyeder->List = Polygon;
    else
      Polyeder->Last->Next = Polygon;
    Polyeder->Last = Polygon;
    return;
  }

  Radius = (Density == 0 || ! (Iterations % Density)) ? Radius2 : Radius1;     

  v12 = GEOVectorAdd (Obj->VertexList[i1], Obj->VertexList[i2]);
  v23 = GEOVectorAdd (Obj->VertexList[i2], Obj->VertexList[i3]);
  v31 = GEOVectorAdd (Obj->VertexList[i3], Obj->VertexList[i1]);

  Obj->VertexList[*No  ] = GEOVectorMult (Radius, GEOVectorNorm(v12));
  Obj->VertexList[*No+1] = GEOVectorMult (Radius, GEOVectorNorm(v23));
  Obj->VertexList[*No+2] = GEOVectorMult (Radius, GEOVectorNorm(v31));

  act = *No;
  *No += 3;

  OBJ_SubDivide (Obj, No, i1,  act,   act+2, 
		 Radius1, Radius2, Density, Iterations-1);
  OBJ_SubDivide (Obj, No, i2,  act+1, act,   
		 Radius1, Radius2, Density, Iterations-1);
  OBJ_SubDivide (Obj, No, i3,  act+2, act+1, 
		 Radius1, Radius2, Density, Iterations-1);
  OBJ_SubDivide (Obj, No, act, act+1, act+2, 
		 Radius1, Radius2, Density, Iterations-1);
}

OBJECT *OBJCreateIcosahedron (H_MATRIX Matrix,
			      double   Radius1, double Radius2, 
			      int      Density, int    Iterations, int Color)
{
  int       count, No = 0, act;
  double    Radius;
  POLYEDER *Polyeder;
  OBJECT   *Obj;

  PRO_START ("\n")

  Obj = OBJNewObject ();
  Obj->ColorIndex = -Color;
  Obj->VertexNo = 60;
  for (count = 0; count < Iterations; ++count)
    Obj->VertexNo = Obj->VertexNo * 4 - 120;
  Obj->VertexList = (POINT3D *)malloc (sizeof(POINT3D) * Obj->VertexNo);
  Obj->Type       = PART_PLANE;

  Polyeder = GEONewPolyeder ();
  Polyeder->Number = 0;

  Radius = (Density == 0 || ! (Iterations % Density)) ? Radius2 : Radius1;  

  Obj->Data.Polyeder = Polyeder;
  for (count = 0; count < 20; ++count)
  {
    Obj->VertexList[No  ] = 
      GEOVectorMult (Radius, IcoData[TIndices[count][2]]);
    Obj->VertexList[No+1] = 
      GEOVectorMult (Radius, IcoData[TIndices[count][1]]);
    Obj->VertexList[No+2] = 
      GEOVectorMult (Radius, IcoData[TIndices[count][0]]);

    act = No;
    No += 3;
      
    OBJ_SubDivide (Obj, &No, act, act+1, act+2, 
		   Radius1, Radius2, Density, Iterations);
  }

  OBJPlace (Obj, Matrix);

  PRO_END
  return Obj;
}


OBJECT *OBJCreatePipe (H_MATRIX Matrix,
		       double   Radius1,    double Radius2, 
		       double   Thickness1, double Thickness2,
		       double   SizeZ, 
		       double   Segment,    int    Facettes, int Color)
{
  POINT2D  Points[5];
  OBJECT  *Obj;

  PRO_START ("\n")

  Radius1 = MAX (Radius1, 0.5*Thickness1);  
  Radius2 = MAX (Radius2, 0.5*Thickness2);

  Points[0].x = Radius2-0.5*Thickness2;
  Points[0].y = 0.0;
  Points[1].x = Radius2+0.5*Thickness2;
  Points[1].y = 0.0;
  Points[2].x = Radius1+0.5*Thickness1;
  Points[2].y = SizeZ;
  Points[3].x = Radius1-0.5*Thickness1;
  Points[3].y = SizeZ;
  Points[4].x = Radius2-0.5*Thickness2;
  Points[4].y = 0.0;

  Obj = OBJCreateRot (Matrix, (Radius1==0.0)? 4:5,
		      Points, Segment, Facettes, 0.0, Color);
  
  PRO_END
  return Obj;
}


OBJECT *OBJCreatePrism (H_MATRIX Matrix,
			double   Width1,  double Width2,  double OffsX,
			double   Depth1,  double Depth2,  double OffsY,
			double   Height1, double Height2, double OffsZ,
			int      Color)
{
  int       count;
  POLYGON  *Polygon;
  POLYEDER *Polyeder;
  OBJECT   *Obj;

  PRO_START ("\n")

  Obj = OBJNewObject ();
  Obj->ColorIndex = -Color;
  Obj->VertexNo   = 8;
  Obj->VertexList = (POINT3D *)malloc (sizeof(POINT3D) * Obj->VertexNo);
  Obj->VertexList[0] = GEOMakeVector (  0.0,    0.0,      0.0);
  Obj->VertexList[1] = GEOMakeVector (Width1,   0.0,  Height1);
  Obj->VertexList[2] = GEOMakeVector (Width1, Depth1, Height1);
  Obj->VertexList[3] = GEOMakeVector (  0.0,  Depth1,     0.0);
  Height1 += OffsZ;
  Obj->VertexList[4] = 
    GEOMakeVector (OffsX,        OffsY,        Height2);
  Obj->VertexList[5] = 
    GEOMakeVector (OffsX+Width2, OffsY,        Height1+Height2);  
  Obj->VertexList[6] = 
    GEOMakeVector (OffsX+Width2, OffsY+Depth2, Height1+Height2);
  Obj->VertexList[7] =
    GEOMakeVector (OffsX,        OffsY+Depth2, Height2);

  /* Transform object coordinates */
  for (count = 0; count < (int)Obj->VertexNo; ++count)
    Obj->VertexList[count] = GEOMatrixMult (Matrix, Obj->VertexList[count]);

  Obj->Type = PART_PLANE;
  Polyeder = GEONewPolyeder ();
  Polyeder->Number = 6;
  for (count = 0; count < 6; ++count)
  {
    Polygon = OBJ_VaMakePolygon (count, Obj->VertexList, 4,
				  PrismData[count][0], PrismData[count][1], 
				  PrismData[count][2], PrismData[count][3]);
    if (Polyeder->List == NULL)
      Polyeder->List = Polygon;
    else
      Polyeder->Last->Next = Polygon;
    Polyeder->Last = Polygon;
  }

  Obj->Data.Polyeder = Polyeder;
  Obj->Info = OBJCalcInfo (Obj);

  PRO_END
  return Obj;
}



OBJECT *OBJCreateQuader (H_MATRIX Matrix,
			 double   SizeX, double SizeY, double SizeZ, int Color)
{
  int       count;
  POLYGON  *Polygon;
  POLYEDER *Polyeder;
  OBJECT   *Obj;

  PRO_START ("\n")

  Obj = OBJNewObject ();
  Obj->ColorIndex = -Color;
  Obj->VertexNo   = 8;
  Obj->VertexList = (POINT3D *)malloc (sizeof(POINT3D) * Obj->VertexNo);
  Obj->VertexList[0] = GEOMakeVector (  0.0,   0.0,   0.0);
  Obj->VertexList[1] = GEOMakeVector (SizeX,   0.0,   0.0);
  Obj->VertexList[2] = GEOMakeVector (SizeX, SizeY,   0.0);
  Obj->VertexList[3] = GEOMakeVector (  0.0, SizeY,   0.0);
  Obj->VertexList[4] = GEOMakeVector (  0.0,   0.0, SizeZ);
  Obj->VertexList[5] = GEOMakeVector (SizeX,   0.0, SizeZ);
  Obj->VertexList[6] = GEOMakeVector (SizeX, SizeY, SizeZ);
  Obj->VertexList[7] = GEOMakeVector (  0.0, SizeY, SizeZ);

  /* Transform object coordinates */
  for (count = 0; count < (int)Obj->VertexNo; ++count)
    Obj->VertexList[count] = GEOMatrixMult (Matrix, Obj->VertexList[count]);

  Obj->Type = PART_PLANE;
  Polyeder = GEONewPolyeder ();
  Polyeder->Number = 6;
  for (count = 0; count < 6; ++count)
  {
    Polygon = OBJ_VaMakePolygon (count, Obj->VertexList, 4,
				 QuaderData[count][0], QuaderData[count][1], 
				 QuaderData[count][2], QuaderData[count][3]);
    if (Polyeder->List == NULL)
      Polyeder->List = Polygon;
    else
      Polyeder->Last->Next = Polygon;
    Polyeder->Last = Polygon;
  }

  Obj->Data.Polyeder = Polyeder;
  Obj->Info = OBJCalcInfo (Obj);

  PRO_END
  return Obj;
}


OBJECT *OBJCreateRoundQuader (H_MATRIX Matrix,
			      double SizeX, double SizeY, double SizeZ, 
			      int Color)
{
  int       count, Facettes = 4;
  double    rx, ry, rz;
  double    OffsX, OffsY, OffsZ;
  POINT3D   v[8];
  POLYGON  *Polygon;
  POLYEDER *Polyeder;
  OBJECT   *Obj;

  rx = 0.0;
  ry = 0.0;
  rz = 2.0;

  OffsX = MAX (ry, rz);
  OffsY = MAX (rx, rz);
  OffsZ = MAX (rx, ry);

  Obj = OBJNewObject ();
  Obj->ColorIndex = -Color;
  Obj->VertexNo   = 8 /* + ... */;

  printf ("Points: %d", 8 + (rx == 0)? 0:(8*(Facettes-1)) 
	                  + (ry == 0)? 0:(8*(Facettes-1))
	                  + (rz == 0)? 0:(8*(Facettes-1)));
  /*   printf ("Flchen: %d", 6 + (rx */

  v[0].x = OffsX;       v[0].y = OffsY;       v[0].z = OffsZ;
  v[1].x = SizeX-OffsX; v[1].y = OffsY;       v[1].z = OffsZ;
  v[2].x = SizeX-OffsX; v[2].y = SizeY-OffsY; v[2].z = OffsZ;
  v[3].x = OffsX;       v[3].y = SizeY-OffsY; v[3].z = OffsZ;
  v[4].x = OffsX;       v[0].y = OffsY;       v[0].z = SizeZ-OffsZ;
  v[1].x = SizeX-OffsX; v[1].y = OffsY;       v[1].z = SizeZ-OffsZ;
  v[2].x = SizeX-OffsX; v[2].y = SizeY-OffsY; v[2].z = SizeZ-OffsZ;
  v[3].x = OffsX;       v[3].y = SizeY-OffsY; v[3].z = SizeZ-OffsZ;
  
  /* erst mal die Unterseite */
  


  return Obj;
}



OBJECT *OBJCreateRot (H_MATRIX Matrix,  int Number,   POINT2D *Points, 
		      double   Segment, int Facettes,
		      double   Radius,  int Color)
{
  int        count, Index;
  int        Steps;
  double     Angle;
  POLYGON2D  Poly;
  POLYGON   *Polygon, *List;
  POLYEDER  *Polyeder;
  OBJECT    *Obj;

  PRO_START ("\n")

  if (fabs(Points[0].x - Points[Number-1].x) < 1e-5 && 
      fabs(Points[0].y - Points[Number-1].y) < 1e-5)
    Number--;

  Poly.Points = Points;
  Poly.Number = Number;
  if (GEO2DPolygonOrientation(&Poly) == GeoDirCW)
    GEO2DSwapOrientation (&Poly);

  if (Segment < -360.0 || Segment > 360.0)
    Segment = 360.0;

  Obj = OBJNewObject ();
  Obj->Type       = PART_PLANE;
  Obj->ColorIndex = -Color;

  /* A 360 object uses the first vertices twice. Smaller objects needs extra
     vertices */
  Steps = (fabs(Segment) < 360.0)? (Facettes+1):Facettes;

  /* Create object's vertices */
  Obj->VertexNo   = Steps * Number;
  Obj->VertexList = (POINT3D *)malloc (sizeof(POINT3D) * Obj->VertexNo);
  for (count = 0; count < Steps; ++count)
  {
    Angle = RAD ((double)count * (Segment / (double)Facettes));
    for (Index = 0; Index < Number; ++Index)
      Obj->VertexList[count*Number+Index] =
	GEOMakeVector (sin(Angle) * (Radius + Points[Index].x),
		       cos(Angle) * (Radius + Points[Index].x),
		       Points[Index].y);
    
  }
  for (count = 0; count < (int)Obj->VertexNo; ++count)
    Obj->VertexList[count] = GEOMatrixMult (Matrix, Obj->VertexList[count]);

  Polyeder = GEONewPolyeder ();

  /* Create polygons around the object */
  Polyeder->Number = Facettes * Number;
  for (count = 0; count < Facettes; ++count)
    for (Index = 0; Index < Number; ++Index)
    {
      Polygon = GEONewPolygon ();
      Polygon->Points = Obj->VertexList;
      Polygon->Number = 4;
      Polygon->Vertex = (int *)malloc (sizeof(int) * Polygon->Number);
      Polygon->Vertex[0] =   count   *Number +   Index;
      Polygon->Vertex[1] =   count   *Number + ((Index+1) % Number);
      Polygon->Vertex[2] = ((count+1)*Number + ((Index+1) % Number)) % 
                              (Obj->VertexNo);
      Polygon->Vertex[3] = ((count+1)*Number +   Index  ) % (Obj->VertexNo);
      Polygon->Info.ID = count * (Number-1) + Index;

      Polygon->Info.Center = GEOCalcPolygonCenter (Polygon);

      GEOCalcHesseForm    (Polygon);
      OBJ_TriangulateQuad (Polygon);

      if (Polyeder->List == NULL)
	Polyeder->List = Polygon;
      else
	Polyeder->Last->Next = Polygon;

      Polyeder->Last = Polygon;
    }

  /* Add both sides of segment */
  if (Segment < 360.0)
    for (count = 0; count < 2; ++count)
    {
      Polygon = OBJ_MakePolygon (Polyeder->Number, Number, Obj->VertexList, 
				 count * Facettes * Number, (count==1));
      if (GEOTestPolygon(Polygon) != 0)
      {
	/* 
	   If polygons are concave, use triangles as polygons in order to
	   get convex polygons 
	*/
	for (Index = 0; Index < Polygon->Number-2; ++Index)
	{
	  List = GEONewPolygon ();
	  List->Points = Obj->VertexList;
	  List->Number = 3;
	  List->Vertex   =      (int *)malloc (sizeof(int) * 3);
	  List->Triangle = (TRIANGLE *)malloc (sizeof(TRIANGLE));
	  List->Vertex[0] = List->Triangle[0][0] = Polygon->Triangle[Index][0];
	  List->Vertex[1] = List->Triangle[0][1] = Polygon->Triangle[Index][1];
	  List->Vertex[2] = List->Triangle[0][2] = Polygon->Triangle[Index][2];

	  List->Info.ID = Number+count;
	  List->Info.Center = GEOCalcPolygonCenter (List);
	  GEOCalcHesseForm (List);

	  Polyeder->Last->Next = List;
	  Polyeder->Last       = List;
	  (Polyeder->Number)++;
	}
      }
      else
      {
	Polyeder->Last->Next = Polygon;
	Polyeder->Last       = Polygon;
	(Polyeder->Number)++;
      }
    }

  Obj->Data.Polyeder = Polyeder;
  OBJCalcInfo (Obj);                    /* Calc object info */

  PRO_END
  return Obj;
}


OBJECT *OBJCreateScrew (H_MATRIX Matrix,
			double   Radius1,  double Radius2, 
			double   Height,   double Increase, double Offset,
			int      Facettes, dual   Direction, 
			int      Color)
{
  int       count, Steps;
  int       Side, Index;
  double    Rounds, Radius;
  double    Angle, sa, ca;
  double    Gain, PosZ, MedZ;
  POLYGON  *Polygon;
  POLYEDER *Polyeder;
  OBJECT   *Obj;

  PRO_START ("\n")

  if (Height != 0.0 && Increase != 0.0)
    Rounds = Height / Increase;
  else
    Rounds = 1.0;

  /* We need two more rounds to fill all gaps */
  Steps = (Rounds+2.0) * Facettes;
  Gain  = Increase / Facettes;
  Radius = (Radius2 - Radius1) / (Steps-1);

  PRO4 ("Rounds: %.2f   Steps: %d   Gain: %.3f   Direction: %s\n", 
	Rounds, Steps, Gain, Direction? "RIGHT":"LEFT")

  /* Probleme:
     Anzahl der Rillen berechnen, falls Abstand nicht zu Lnge pat
     (z.B. Hoehe reicht nicht fr komplette Runde.
     Evtl als Parameter...
  */

  Obj = OBJNewObject ();
  Obj->ColorIndex = -Color;
  Obj->VertexNo = Facettes + 2*Steps + Facettes;
  Obj->VertexList = (POINT3D *)malloc (sizeof(POINT3D) * Obj->VertexNo);

  /* vertexes of lower ground plate */
  for (count = 0; count < Facettes; ++count)
  {
    Angle = RAD ((double)count * 360.0 / (double)(Facettes-1));
    Obj->VertexList[count] = 
      GEOMakeVector (sin(Angle) * Radius1, cos(Angle) * Radius1, 0.0);
  }

  PosZ = -Increase; 
  Index = Facettes;
  for (count = 0; count < Steps; ++count)
  {
    Angle = RAD((double)count * 360.0 / (double)(Facettes-1));
    if (Direction)
      Angle *= -1.0;

    sa = sin (Angle);
    ca = cos (Angle);

    Obj->VertexList[Index++] = 
      GEOMakeVector (sa*Radius1, ca*Radius1, MAX(0.0, MIN(PosZ, Height)));

    if (PosZ < Height+Increase)
    {
      /* Reduce Radius of outer race if top of screw is reached */
      MedZ = PosZ + 0.5 * Increase;
      if (MedZ <= Height)
        Obj->VertexList[Index++] = 
	  GEOMakeVector (sa*(Radius1 + Offset), ca*(Radius1 + Offset), 
			 MAX(0.0, MedZ));
      else
	Obj->VertexList[Index++] = 
	  GEOMakeVector (sa*(Radius1 + MAX(0.0,
			       Offset*(Increase-(MedZ-Height))/Increase)),
		         ca*(Radius1 + MAX(0.0,
                               Offset*(Increase-(MedZ-Height))/Increase)),
		         Height);
    }

    Radius1 += Radius;
    PosZ += Gain;
  }

  /* vertexes of upper plate */
  for (count = 0; count < Facettes; ++count)
  {
    Angle = RAD ((double)count * 360.0 / (double)(Facettes-1));
    Obj->VertexList[Index+count] = 
      GEOMakeVector (sin(Angle) * (Radius2+fabs(Radius)), 
		     cos(Angle) * (Radius2+fabs(Radius)), Height);
  }
  
  /* Transform object coordinates */
  for (count = 0; count < (int)Obj->VertexNo; ++count)
    Obj->VertexList[count] = GEOMatrixMult (Matrix, Obj->VertexList[count]);

  Obj->Type = PART_PLANE;

  Polyeder = GEONewPolyeder ();
  Polyeder->Number = 1 + 2*(Steps-Facettes) + 1; 

  /* Polygon of ground plate */
  Polygon = GEONewPolygon ();
  Polygon->Points = Obj->VertexList;
  Polygon->Number = Facettes;
  Polygon->Vertex = (int *)malloc (sizeof(int) * Polygon->Number);
  for (count = 0; count < Facettes; ++count)
    Polygon->Vertex[count] = count;
  Polygon->Info.ID = 0;
  Polygon->Info.Center = GEOCalcPolygonCenter (Polygon);
  GEOCalcHesseForm      (Polygon);
  GEOTriangulatePolygon (Polygon);

  Polyeder->List = Polygon;
  Polyeder->Last = Polygon;

  /* Groove */
  for (count = 0; count < Steps-Facettes; ++count)
    for (Side = 0; Side < 2; ++Side)
    {
      if (Side == 0 || 
	  (Facettes+2*count + 2*Facettes + 1 < (int)Obj->VertexNo))
      {
	Polygon = GEONewPolygon ();
	Polygon->Points = Obj->VertexList;
	Polygon->Number = 4;
	Polygon->Vertex = (int *)malloc (sizeof(int) * Polygon->Number);

	for (Index = 0; Index < 4; ++Index)
	{
	  Polygon->Vertex[Index] = 
	    Facettes + 2*count + ScrewIndex[Direction][Side][Index];
	  if (ScrewIndex[Direction][Side][Index] < 0)
	    Polygon->Vertex[Index] += 2*Facettes+1;
	}

	Polygon->Info.ID = count+1;
	Polygon->Info.Center = GEOCalcPolygonCenter (Polygon);

	GEOCalcHesseForm      (Polygon);
	GEOTriangulatePolygon (Polygon);

	Polyeder->Last->Next = Polygon;
	Polyeder->Last = Polygon;
      }
   }

  /* Polygon of upper plate */
  Polygon = GEONewPolygon ();
  Polygon->Points = Obj->VertexList;
  Polygon->Number = Facettes;
  Polygon->Vertex = (int *)malloc (sizeof(int) * Polygon->Number);
  for (count = 0; count < Facettes; ++count)
    Polygon->Vertex[count] = Obj->VertexNo-1-count;
  Polygon->Info.ID = 1+2*Steps;
  Polygon->Info.Center = GEOCalcPolygonCenter (Polygon);
  GEOCalcHesseForm      (Polygon);
  GEOTriangulatePolygon (Polygon);

  Polyeder->Last->Next = Polygon;
  Polyeder->Last = Polygon;

  /* Insert polyeder into object */
  Obj->Data.Polyeder = Polyeder;
  Obj->Info = OBJCalcInfo (Obj);

  PRO_END
  return Obj;
}


OBJECT *OBJCreateSlice (H_MATRIX Matrix,    double Radius1, double Radius2,
			double   Radius3,   double SizeZ, 
			int      Facettes1, int    Facettes2, 
			double   Segment,   int    Color)
{
  int      count;
  double   Radius, Offset;
  POINT2D *Points;
  OBJECT  *Obj;

  PRO_START ("\n")

  Radius = sqrt (SQR(Radius3) + 
		 SQR(((SQR(Radius3)-SQR(Radius2)-SQR(SizeZ)) / (2*SizeZ))));
  Offset = Radius * cos (asin(Radius2/Radius));

  Points = (POINT2D *)malloc (sizeof(POINT2D) * (Facettes1+4));

  Points[0].x = Radius1;
  Points[0].y = Offset - SizeZ;
  for (count = 0; count <= Facettes1; ++count)
  {
    Points[count+1].y = Offset - SizeZ + count * SizeZ / Facettes1;
    Points[count+1].x = Radius * cos (asin(Points[count+1].y / Radius));
  }
  Points[count+1].x = Radius1;
  Points[count+1].y = Offset;
  Points[count+2] = Points[0];

  Obj = OBJCreateRot (Matrix, Facettes1+4, Points, 
		              Segment, Facettes2, 0.0, Color);
  free (Points);

  PRO_END
  return Obj;
}


OBJECT *OBJCreateSphere (H_MATRIX Matrix, 
			 double   Radius,   double SegmentX, double SegmentZ,
			 int      Facettes, int    Color)
{
  int      count, Rows;
  int      Number;
  double   Angle, Step;
  POINT2D *Points;
  OBJECT  *Obj;
  
  PRO_START ("\n")

  if (SegmentZ <= 0.0 || SegmentZ > 360.0)
    SegmentZ = 360.0;

  Rows = (int)(Facettes * 0.5) - 1;
  Step = SegmentX / (double)Rows;
  Number = Rows+1;
  if (SegmentX < 180.0)
    Number++;
  
  Points = (POINT2D *)malloc (sizeof(POINT2D) * Number);
  for (count = 0; count <= Rows; ++count)
  {
    Angle = RAD (count * Step);
    Points[count].x = sin(Angle) * Radius;
    Points[count].y = cos(Angle) * Radius;
  }
  if (SegmentX < 180.0)
  {
    Points[count].x = 0.0;
    Points[count].y = 0.0;/* Points[count-1].y */;
  }

  Obj = OBJCreateRot (Matrix, Number, Points, SegmentZ, Facettes, 0.0, Color);

  free (Points);

  PRO_END
  return Obj;
}


OBJECT *OBJCreateSpring (H_MATRIX Matrix,
			 double   Radius1,   double Radius2, double Radius3,
			 double   Height,    double Increase,
			 int      Facettes1, int    Facettes2, 
			 dual     Direction, int    Color)
{
  int       count, count1, count2;
  int       Steps, Swap;
  double    Rounds, Radius, Dist;
  double    Angle1, Angle2, sa, ca;
  double    Gain, PosZ;
  POLYGON  *Polygon;
  POLYEDER *Polyeder;
  OBJECT   *Obj;

  PRO_START ("\n")

  if (Height != 0.0 && Increase != 0.0)
    Rounds = Height / Increase;
  else
    Rounds = 1.0;

  Steps = Rounds * Facettes1;
  Gain  = Height / Steps;
  Radius = (Radius2 - Radius1) / (Steps-1);

  PRO5 ("Height: %.2f Rounds: %.2f   Steps: %d   Gain: %.3f   Dir: %s\n", 
	Height, Rounds, Steps, Gain, Direction? "RIGHT":"LEFT")

  /* TODO: 
     Falls Hhe = 0, und Radius1 <> Radius2: Anzahl der Drehungen berechnen 
  */

  Obj = OBJNewObject ();
  Obj->ColorIndex = -Color;
  Obj->VertexNo = Facettes2*Steps;
  Obj->VertexList = (POINT3D *)malloc (sizeof(POINT3D) * Obj->VertexNo);

  /* Build vertices of spring */
  PosZ = 0.0; /* bzw. 0.5*Radius3 */
  for (count1 = 0; count1 < Steps; ++count1)
  {
    Angle1 = RAD((double)count1 * 360.0 / (double)(Facettes1-1));
    if (Direction)
      Angle1 *= -1.0;

    sa = sin (Angle1);
    ca = cos (Angle1);

    for (count2 = 0; count2 < Facettes2; ++count2)
    {
      Angle2 = RAD ((double)count2 * 360.0 / (double)(Facettes2));
      Dist = Radius1 + cos(Angle2) * Radius3;
      Obj->VertexList[count1*Facettes2+count2] = 
        GEOMakeVector (Dist * ca, sa * Dist, PosZ + sin(Angle2)*Radius3);
    }
    Radius1 += Radius;
    PosZ += Gain;
  }

  /* Transform object coordinates */
  for (count = 0; count < (int)Obj->VertexNo; ++count)
    Obj->VertexList[count] = GEOMatrixMult (Matrix, Obj->VertexList[count]);

  Obj->Type = PART_PLANE;  

  Polyeder = GEONewPolyeder ();
  Polyeder->Number = 1 + Facettes2*(Steps-1) + 1; 

  /* Polygon of ground plate */
  Polygon = GEONewPolygon ();
  Polygon->Points = Obj->VertexList;
  Polygon->Number = Facettes2;
  Polygon->Vertex = (int *)malloc (sizeof(int) * Polygon->Number);
  if (Direction)
    for (count2 = 0; count2 < Facettes2; ++count2)
      Polygon->Vertex[count2] = Facettes2-count2-1;
  else
    for (count2 = 0; count2 < Facettes2; ++count2)
      Polygon->Vertex[count2] = count2;    
  Polygon->Info.ID = 0;
  Polygon->Info.Center = GEOCalcPolygonCenter (Polygon);
  GEOCalcHesseForm      (Polygon);
  GEOTriangulatePolygon (Polygon);

  Polyeder->List = Polygon;
  Polyeder->Last = Polygon;
 
  for (count1 = 0; count1 < Steps-1; ++count1)
    for (count2 = 0; count2 < Facettes2; ++count2)
    {
      Polygon = GEONewPolygon ();
      Polygon->Points = Obj->VertexList;
      Polygon->Number = 4;
      Polygon->Vertex = (int *)malloc (sizeof(int) * Polygon->Number);

      Polygon->Vertex[0] = count1*Facettes2 + count2;
      if (count2 == Facettes2-1)
      {
	Polygon->Vertex[1] =  count1   *Facettes2;
	Polygon->Vertex[2] = (count1+1)*Facettes2;
      }
      else
      {
	Polygon->Vertex[1] =  count1   *Facettes2 + count2+1;
	Polygon->Vertex[2] = (count1+1)*Facettes2 + count2+1;
      }
      Polygon->Vertex[3] = (count1+1)*Facettes2 + count2;

      if (! Direction)
      {
	Swap               = Polygon->Vertex[0];
	Polygon->Vertex[0] = Polygon->Vertex[3];
	Polygon->Vertex[3] = Swap;

	Swap               = Polygon->Vertex[1];
	Polygon->Vertex[1] = Polygon->Vertex[2];
	Polygon->Vertex[2] = Swap;
      }

      Polygon->Info.ID = 1 + count1*Facettes2 + count2;
      Polygon->Info.Center = GEOCalcPolygonCenter (Polygon);
      GEOCalcHesseForm      (Polygon);
      GEOTriangulatePolygon (Polygon);

      Polyeder->Last->Next = Polygon;
      Polyeder->Last = Polygon;      
    }

  /* Polygon of upper plate */
  Polygon = GEONewPolygon ();
  Polygon->Points = Obj->VertexList;
  Polygon->Number = Facettes2;
  Polygon->Vertex = (int *)malloc (sizeof(int) * Polygon->Number);
  if (Direction)
    for (count2 = 0; count2 < Facettes2; ++count2)
      Polygon->Vertex[count2] = Obj->VertexNo-Facettes2+count2;
  else
    for (count2 = 0; count2 < Facettes2; ++count2)
      Polygon->Vertex[count2] = Obj->VertexNo-1-count2;
  Polygon->Info.ID = 1+Facettes2*(Steps-1);
  Polygon->Info.Center = GEOCalcPolygonCenter (Polygon);
  GEOCalcHesseForm      (Polygon);
  GEOTriangulatePolygon (Polygon);

  Polyeder->Last->Next = Polygon;
  Polyeder->Last = Polygon;

  /* Insert polyeder into object */
  Obj->Data.Polyeder = Polyeder;
  Obj->Info = OBJCalcInfo (Obj);

  PRO_END
  return Obj;
}


OBJECT *OBJCreateSweep (H_MATRIX Matrix,
        /* Extrusion */ int Number, POINT2D *Points, double Height, int Color)
{
  int       count, Index;
  POLYGON2D Poly;
  POLYGON  *Polygon, *List;
  POLYEDER *Polyeder;
  OBJECT   *Obj;

  PRO_START ("\n")

  if (fabs(Points[0].x - Points[Number-1].x) < 1e-5 && 
      fabs(Points[0].y - Points[Number-1].y) < 1e-5)
    Number--;

  Poly.Points = Points;
  Poly.Number = Number;
  if (GEO2DPolygonOrientation(&Poly) == GeoDirCW)
    GEO2DSwapOrientation (&Poly);

  Obj = OBJNewObject ();

  Obj->ColorIndex = -Color;
  Obj->VertexNo   = 2 * Number;
  Obj->VertexList = (POINT3D *)malloc (sizeof(POINT3D) * Obj->VertexNo);
  for (count = 0; count < Number; ++count)
  {
    Obj->VertexList[count] = 
      GEOMakeVector (Points[count].x, Points[count].y, Height);
    Obj->VertexList[2*Number-1-count] =
      GEOMakeVector (Points[count].x, Points[count].y, 0.0);
  }

  /* Transform object coordinates */
  for (count = 0; count < (int)Obj->VertexNo; ++count)
    Obj->VertexList[count] = GEOMatrixMult (Matrix, Obj->VertexList[count]);

  Obj->Type = PART_PLANE;
  Polyeder = GEONewPolyeder ();
  Polyeder->Number = Number;
  /* Create side faces */
  for (count = 0; count < Number; ++count)
  {
    Polygon = GEONewPolygon ();
    Polygon->Points = Obj->VertexList;
    Polygon->Number = 4;
    Polygon->Vertex = (int *)malloc (sizeof(int) * Polygon->Number);
    for (Index = 0; Index < 2; ++Index)
    {
      Polygon->Vertex[Index]   = Number + ((2*Number-count-1-Index) % Number);
      Polygon->Vertex[2+Index] = (count+1-Index) % Number;
    }
    Polygon->Info.ID = count;
    Polygon->Info.Center = GEOCalcPolygonCenter (Polygon);

    GEOCalcHesseForm      (Polygon);
    OBJ_TriangulateQuad (Polygon);

    if (Polyeder->List == NULL)
      Polyeder->List = Polygon;
    else
      Polyeder->Last->Next = Polygon;

    Polyeder->Last = Polygon;
  }

  for (count = 0; count < 2; ++count)
  {
    PRO1 ("EXTRUSION-FACE %d...\n", count+1)

    Polygon = GEONewPolygon ();

    Polygon->Points = Obj->VertexList;
    Polygon->Number = Number;
    Polygon->Vertex = (int *)malloc (sizeof(int) * Polygon->Number);
    for (Index = 0; Index < Number; ++Index)
      Polygon->Vertex[Index] = count*Number + Index;
    
    Polygon->Info.ID     = Number+count;
    Polygon->Info.Center = GEOCalcPolygonCenter (Polygon);

    GEOCalcHesseForm      (Polygon);
    GEOTriangulatePolygon (Polygon);

    if (GEOTestPolygon(Polygon) == 0)   /* Polygon is already convex        */
    {
      Polyeder->Last->Next = Polygon;
      Polyeder->Last = Polygon;
      (Polyeder->Number)++;
    }
    else                               /* Polygon is concave, use triangles */
    {
      for (Index = 0; Index < Polygon->Number-2; ++Index)
      {
	List = GEONewPolygon ();
	List->Points   = Obj->VertexList;
	List->Number   = 3;
	List->Vertex   =      (int *)malloc (sizeof(int) * 3);
	List->Triangle = (TRIANGLE *)malloc (sizeof(TRIANGLE));
	List->Vertex[0] = List->Triangle[0][0] = Polygon->Triangle[Index][0];
	List->Vertex[1] = List->Triangle[0][1] = Polygon->Triangle[Index][1];
	List->Vertex[2] = List->Triangle[0][2] = Polygon->Triangle[Index][2];

	List->Info.ID = Number + count*(Polygon->Number-2) + Index;
	List->Info.Center = GEOCalcPolygonCenter (List);
	GEOCalcHesseForm (List);

	Polyeder->Last->Next = List;
	Polyeder->Last       = List;
	(Polyeder->Number)++;
      }
    
      GEOFreePolygon (Polygon, FALSE);
    }
  }

  Obj->Data.Polyeder = Polyeder;
  Obj->Info = OBJCalcInfo (Obj);

  PRO_END
  return Obj;
}


OBJECT *OBJCreateTorus (H_MATRIX Matrix,    double Radius1,   double Radius2, 
			int      Facettes1, int    Facettes2, double Segment, 
			int      Color)
{
  int      count;
  POINT2D *Points;
  OBJECT  *Obj;

  PRO_START ("\n")

  Points = (POINT2D *)malloc (sizeof(POINT2D) * Facettes1);
  for (count = 0; count < Facettes1; ++count)
  {
    Points[count].x = Radius1 + Radius2 * cos(RAD(count*360.0/Facettes1));
    Points[count].y =           Radius2 * sin(RAD(count*360.0/Facettes1));
  }
  
  Obj = OBJCreateRot (Matrix, Facettes1, Points, 
		              Segment, Facettes2, 0.0, Color);
  free (Points);

  PRO_END
  return Obj;
}


OBJECT *OBJCreateWedge (H_MATRIX Matrix,  double Width1,  double Width2,
			double   Height1, double Height2,
			double   Depth1,  double Depth2,  int    Color)
{
  int       count = 0, Index;
  int       Points = 12;
  POLYGON  *Polygon;
  POLYEDER *Polyeder;
  OBJECT   *Obj;

  PRO_START ("\n")

  if (Height1 == 0.0)
    Points -= 4;

  Obj = OBJNewObject ();
  Obj->ColorIndex = -Color;
  Obj->VertexNo   = Points;
  Obj->VertexList = (POINT3D *)malloc (sizeof(POINT3D) * Obj->VertexNo);

  Obj->VertexList[count++] = GEOMakeVector (0.0,    0.0, 0.0);
  Obj->VertexList[count++] = GEOMakeVector (0.0,    0.0, Depth1);
  Obj->VertexList[count++] = GEOMakeVector (Width1, 0.0, Depth1);
  Obj->VertexList[count++] = GEOMakeVector (Width1, 0.0, 0.0);
  if (Height1 > 0.0)
  {
    Obj->VertexList[count++] = GEOMakeVector (0.0, Height1,
					      0.5*(Depth2-Depth1) *
					      Height1 / Height2);
    Obj->VertexList[count++] = GEOMakeVector (0.0, Height1,
					    Depth1 - 0.5*(Depth2-Depth1) *
					    Height1 / Height2);
    Obj->VertexList[count++] = GEOMakeVector (Width1, Height1,
					      Depth1 - 0.5*(Depth2-Depth1) *
					      Height1 / Height2);
    Obj->VertexList[count++] = GEOMakeVector (Width1, Height1,
					    0.5*(Depth2-Depth1) *
					    Height1 / Height2);
  }
  Obj->VertexList[count++] = GEOMakeVector (0.5*(Width1-Width2),
					    Height2, 0.5*(Depth1-Depth2));
  Obj->VertexList[count++] = GEOMakeVector (0.5*(Width1-Width2), Height2, 
					    Depth1 - 0.5*(Depth1-Depth2));
  Obj->VertexList[count++] = GEOMakeVector (Width1-0.5*(Width1-Width2), 
					    Height2, 
					    Depth1 - 0.5*(Depth1-Depth2));
  Obj->VertexList[count++] = GEOMakeVector (Width1-0.5*(Width1-Width2), 
					  Height2, 0.5*(Depth1-Depth2));

  /* Transform object coordinates */
  for (count = 0; count < (int)Obj->VertexNo; ++count)
    Obj->VertexList[count] = GEOMatrixMult (Matrix, Obj->VertexList[count]);

  Obj->Type = PART_PLANE;
  Polyeder = GEONewPolyeder ();
  Polyeder->Number = 6 + ((Height1 > 0.0)? 2:0);
  for (count = 0; count < Polyeder->Number; ++count)
  {
    Polygon = GEONewPolygon ();
    Polygon->Points = Obj->VertexList;
    Polygon->Number = 4;
    Polygon->Vertex = (int *)malloc (sizeof(int) * Polygon->Number);
    for (Index = 0; Index < 4; ++Index)
      Polygon->Vertex[Index] = PrismData[count][Index];
    Polygon->Info.ID = count;
    Polygon->Info.Center = GEOCalcPolygonCenter (Polygon);

    GEOCalcHesseForm     (Polygon);
    OBJ_TriangulateQuad (Polygon);

    if (Polyeder->List == NULL)
      Polyeder->List = Polygon;
    else
      Polyeder->Last->Next = Polygon;

    Polyeder->Last = Polygon;
  }

  Obj->Data.Polyeder = Polyeder;
  Obj->Info = OBJCalcInfo (Obj);

  PRO_END
  return Obj;
}
