#include <math.h>

#include "libobj.h"


OBJECT *OBJNewObject (void)
{
  OBJECT *Result = NULL;

  PRO_START (" ")

  Result = (OBJECT *)malloc (sizeof(OBJECT));

  Result->ObjNo      =  0;
  Result->ColorIndex = -2;

  Result->VertexNo   = 0;
  Result->VertexList = NULL;

  Result->Type = NONE;

  Result->Info.Center = Vector_0;
  Result->Info.Min    = Vector_0;
  Result->Info.Max    = Vector_0;
  Result->Info.Area   = 0.0;
  Result->Info.Volume = 0.0;
  Result->Info.Bubble = 0.0;

  Result->Next = NULL;

  PRO_END
  return Result;
}


void OBJFreeObject (OBJECT *Object)
{
  if (Object->VertexList != NULL)
    free (Object->VertexList);

  switch (Object->Type)
  {
    case PART_CURVE :
      PRO0 ("LINES...")
      free (Object->Data.Curve->List);
      free (Object->Data.Curve);
      break;
     
    case PART_PLANE :
      PRO0 ("POLYEDER...")
      GEOFreePolyeder (Object->Data.Polyeder);
      break;
	
    case NONE :
      break;

    default :
      PRO1 ("Invalid object-type: %d\n", Object->Type)
  }
  PRO0 ("\\done\n")

  free (Object);
}



OBJECT *OBJCopyObject (OBJECT *Object)
{
  OBJECT *Result;

  if (Object == NULL)
    return NULL;

  Result = OBJNewObject ();

  Result->ColorIndex = Object->ColorIndex;
  Result->VertexNo   = Object->VertexNo;
  Result->VertexList = (POINT3D *)malloc (sizeof(POINT3D) * Object->VertexNo);
  memcpy (Result->VertexList, Object->VertexList, 
	                      sizeof(POINT3D) * Object->VertexNo);

  switch ((Result->Type = Object->Type))
  {
    case PART_CURVE :
      Result->Data.Curve = (LINES *)malloc (sizeof(LINES));
      Result->Data.Curve->Number = Object->Data.Curve->Number;
      Result->Data.Curve->List = 
	(int *)malloc (sizeof(int) * Object->Data.Curve->Number);
      memcpy (Result->Data.Curve->List, Object->Data.Curve->List,
	      sizeof(int) * Object->Data.Curve->Number);
      break;

    case PART_PLANE :
      Result->Data.Polyeder = GEOCopyPolyeder (Object->Data.Polyeder,
					       Result->VertexList);
      break;

    case NONE :
      break;

    default :
      PRO1 ("Invalid object-type: %d\n", Object->Type)
      OBJFreeObject (Result);
      return NULL;
  }
  Result->Info = Object->Info;
  Result->Next = NULL;

  return Result;
}


dual OBJAddPolygon (OBJECT *Object, POLYGON *Polygon)
{
  POLYEDER *Polyeder;

  if (Object == NULL || Polygon == NULL)
    return FALSE;

  if (Object->Type == PART_PLANE)
  {
    if (Object->Data.Polyeder == NULL)
      Object->Data.Polyeder = GEONewPolyeder ();

    Polyeder = Object->Data.Polyeder;
    if (Polyeder->List == NULL)
      Polyeder->List = Polygon;
    else
      Polyeder->Last->Next = Polygon;
    Polyeder->Last = Polygon;

    (Polyeder->Number)++;
    OBJCalcInfo (Object);

    return TRUE;
  }
  return FALSE;
}


OBJECT *OBJAddObject (OBJECT *Object, OBJECT *NewObj)
{
  OBJECT *Ptr;

  if ((Ptr = Object) == NULL)
    return NewObj;

  while (Ptr->Next != NULL)
    Ptr = Ptr->Next;
  Ptr->Next = NewObj;

  return Object;
}


int OBJAddVertex (OBJECT *Object, POINT3D Point)
{
  unsigned int  count;
  POLYGON      *Ptr = Object->Data.Polyeder->List;

  if (Object->VertexList == NULL)
    Object->VertexList = malloc (sizeof(POINT3D));
  else
  {
    /* Check, if the vertex already exists */
    for (count = 0; count < Object->VertexNo; ++count)
      if (VEC_EQUAL(Object->VertexList[count], Point))
	return count;

    /* expand the vertex list */
    Object->VertexList = realloc (Object->VertexList,
				  sizeof(POINT3D) * ((Object->VertexNo)+1));
  }
  Object->VertexList[Object->VertexNo] = Point; 
  (Object->VertexNo)++;

  /* Refresh the vertex-pointers. Not really efficient, but it preserves
     us from problems, if this will be forgotten later. */
  while (Ptr != NULL)
  {
    Ptr->Points = Object->VertexList;
    Ptr = Ptr->Next;
  }

  return (Object->VertexNo)-1;
}


void OBJPlace (OBJECT *Object, H_MATRIX WTO)
{
  unsigned int  count;
  POLYGON      *Poly;

  if (Object->VertexList == NULL)
    return;

  for (count = 0; count < Object->VertexNo; ++count)
    Object->VertexList[count] = GEOMatrixMult (WTO, Object->VertexList[count]);

#if 1
  if (Object->Type == PART_PLANE)
  {
    Poly = Object->Data.Polyeder->List;
    while (Poly != NULL)
    {
      GEOCalcHesseForm (Poly);
      Poly = Poly->Next;
    }
  }
#endif

  OBJCalcInfo (Object);
}


/* ********************************************************************** */
/* ** scales the object's coordinates relative. Advantage: The scaling ** */
/* ** needs to be computed once. By saving the actual scale-values as  ** */
/* ** absolute scale-values in the part-data, we are able to restore   ** */
/* ** the normal-values. Be aware of rounding-errors when calling this ** */
/* ** function repeated.                                               ** */
/* ********************************************************************** */
void OBJScale (OBJECT *Object, double sx, double sy, double sz)
{
  unsigned int count;

  if (Object->VertexList == NULL)
    return;

  for (count = 0; count < Object->VertexNo; ++count)
  {
    Object->VertexList[count].x *= sx;
    Object->VertexList[count].y *= sy;
    Object->VertexList[count].z *= sz;
  }

  OBJCalcInfo (Object);                 /* Update Object Info */
}


dual OBJIsEuler (OBJECT *Object)
{
  int      EdgeCount = 0;
  int      Value;
  POLYGON *Polygon;

  if (Object == NULL || Object->Type != PART_PLANE)
    return TRUE;

  Polygon = Object->Data.Polyeder->List;
  while (Polygon != NULL)
  {
    EdgeCount += Polygon->Number;
    Polygon = Polygon->Next;
  }
  
  Value = Object->VertexNo - 
          EdgeCount / 2 + 
          Object->Data.Polyeder->Number;

  PRO4 ("E - K + F = %d - %d + %d = %d\n",
        Object->VertexNo, EdgeCount, Object->Data.Polyeder->Number, Value)

  return (Value == 2);
}


static double OBJ_CalcBubble (OBJECT *Object, POINT3D Center)
{
  unsigned int count;
  double       Result = 0.0;
  POINT3D      Vector;

  PRO_START (" ")

  if (Object == NULL)
    PRO_RETURN ("Empty object", 0.0)

  for (count = 0; count < Object->VertexNo; ++count)
  {
    Vector = GEOVectorSub (Object->VertexList[count], Center);
    Result = MAX (Result, VEC_LEN(Vector));
  }

  PRO_END
  return Result;
}


static VECTOR3D OBJ_CalcRange (OBJECT *Object, 
			       POINT3D *MinSize, POINT3D *MaxSize)
{
  unsigned int count;
  POINT3D      Min, Max;
  VECTOR3D     Size = Vector_0;

  if (Object->VertexNo > 0)
  {  
    Min = Max = Object->VertexList[0];
    for (count = 1; count < Object->VertexNo; ++count)
    {
      Min.x = MIN (Min.x, Object->VertexList[count].x);
      Max.x = MAX (Max.x, Object->VertexList[count].x);
      Min.y = MIN (Min.y, Object->VertexList[count].y);
      Max.y = MAX (Max.y, Object->VertexList[count].y);
      Min.z = MIN (Min.z, Object->VertexList[count].z);
      Max.z = MAX (Max.z, Object->VertexList[count].z);
    }
  }

  if (MinSize != NULL)
    *MinSize = Min;
  if (MaxSize != NULL)
    *MaxSize = Max;

  Size = GEOVectorSub (Max, Min);

  return Size;
}


INFO_3D OBJCalcInfo (OBJECT *Object)
{
  INFO_3D Result;

  PRO_START ("\n")

  Result.Center = Vector_0;
  Result.Min    = Vector_0;
  Result.Max    = Vector_0;
  Result.Volume = 0.0;
  Result.Area   = 0.0;
  Result.Bubble = 0.0;

  if (Object == NULL)
    PRO_RETURN ("Bad Object", Result)

  PRO1 ("Type: %d\n", Object->Type)
 
  switch (Object->Type)
  {
    case PART_CURVE :
      /* TODO:
	 Result = GEOCalcCurveCenter (Object->Data.Curve);
      */
      break;

    case PART_PLANE :
      Result = GEOCalcPolyederInfo (Object->Data.Polyeder);
      break;

    case NONE :
      break;

    default :
      PRO1 ("Object-Type not supported: %d\n", Object->Type)
  }

  PRO0 ("Bubble...")
  Result.Bubble = OBJ_CalcBubble (Object, Result.Center);

  PRO0 ("\\Size...")
  OBJ_CalcRange (Object, &Result.Min, &Result.Max);

  Object->Info = Result;

  PRO_END
  return Result;
}


int OBJGetTriangles (OBJECT *Object)
{
  int Result;
 
  PRO_START ("")

  if (Object == NULL || Object->Type != PART_PLANE)
    PRO_RETURN ("empty object", 0)

  Result = GEOGetTriangles (Object->Data.Polyeder);

  PRO_END
  return Result;
}
  






