#ifndef LIBPART_H
#define LIBPART_H

#include <libobj.h>

#ifdef __cplusplus
extern "C" {
#endif


/* ********************************************************************* */
/* ** Macro-definitions                                               ** */
/* ********************************************************************* */
#define FEATURE_NAME(x) \
          ((x==PART_FACE)? "FACE": \
	    ((x==PART_SHAFT)? "SHAFT": \
	      ((x==PART_HOLE)? "HOLE": \
               ((x==PART_FIXED)? "FIXED":"UNKNOWN"))))

#define WRITE_FEATURE(fp,F) \
        { \
          if (F == NULL || F->Type == PART_FIXED) \
	    fprintf (fp, "FIXED"); \
	  else \
	  { \
	    fprintf (fp, "%s ", FEATURE_NAME(F->Type)); \
	    if (F->Type == PART_HOLE) \
	      fprintf (fp, "[%d %d %d]", F->Polygon[0]->Info.ID, \
		           F->Polygon[1]->Info.ID, F->Polygon[2]->Info.ID); \
            else \
	      fprintf (fp, " %d", F->Polygon[0]->Info.ID); \
	  } \
	}

#define PRO_FEATURE(F) \
        { \
	  if (F == NULL || F->Type == PART_FIXED) \
	    PRO0 ("\\FIXED") \
	  else \
	  { \
	    PRO1 ("\\%s ", FEATURE_NAME(F->Type)) \
	    if (F->Type == PART_HOLE) \
	      PRO3 ("\\[%d %d %d]", F->Polygon[0]->Info.ID, \
		    F->Polygon[1]->Info.ID, F->Polygon[2]->Info.ID) \
	    else \
	      PRO1 ("\\%d", F->Polygon[0]->Info.ID); \
	  } \
	}


/* *********************************************************************** */
/* ** enumerations and constants                                        ** */
/* *********************************************************************** */
enum FEATURE_TYPE
{
  PART_FACE = 1,
  PART_SHAFT,
  PART_HOLE,
  PART_FIXED
};


/* contact-types: User-flag 'U_CONTACT' overwrites a calculating value 
   (prevents from recalculation) */
#define X_CONTACT (0<<0)
#define V_CONTACT (1<<0)
#define R_CONTACT (1<<1)
#define U_CONTACT (1<<2)


/* ********************************************************************** */
/* ** feature-informations                                             ** */
/* ********************************************************************** */
typedef struct FEATURE
{
         int       Type;                /* Feature-type: FACE/SHAFT/HOLE  */
         int       ID;                  /* identifier                     */
         OBJECT   *Object;              /* Object-pointer                 */
         POLYGON  *Polygon[3];          /* up to three polygons           */
         H_MATRIX  CTF;                 /* (Part-)Center->Feature-Transf. */
         int       used;                /* x-times used                   */
  struct FEATURE  *Next;
} FEATURE;


/* ********************************************************************** */
/* ** structure, which contains the data of complex parts. A part      ** */
/* ** consists out of one or more objects.                             ** */
/* ********************************************************************** */
typedef struct PART
{
  char        *PartName;                /* Name of part                   */
  char        *FileName;                /* location of file if exists     */

  int          CSNo;                    /* Amount of Coordinate-Systems   */
  H_MATRIX    *CS;                      /* Coordinate-Systems             */
  H_MATRIX     WTP;                     /* WorldToPart transformation     */
  H_MATRIX     PTC;                     /* PartToCenter transformation    */
  double       Scale[3];                /* Scaling                        */

  int          ObjectNo;                /* Amount of objects              */
  OBJECT      *ObjectList;              /* List of Objects                */

  INFO_3D      Info;                    /* geometric-information          */

  FEATURE     *FeatureList;             /* List of features               */

  double       NormalAngle;             /* Angle for normal-calculation   */
  dual         updated;                 /* TRUE, if updated else FALSE    */ 

  struct PART *Next;                    /* Pointer to next part           */
} PART;


/* ********************************************************************** */
/* ** Exported functions                                               ** */
/* ********************************************************************** */
PART     *PARTNewPart           (void);
PART     *PARTCreate            (const char *Name, OBJECT *ObjectList);
PART     *PARTCopyPart          (PART *Part);
PART     *PARTMergeParts        (PART *Part1, PART *Part2, dual create);
PART     *PARTSmashPart         (PART *Part);
void      PARTFreePart          (PART *Part);
void      PARTAddObject         (PART *Part, OBJECT *ObjectList);

void      PARTScalePart         (PART *Part, double sx, double sy, double sz);
void      PARTSetName           (PART *Part, const char *Name, ...);
void      PARTSetOrigin         (PART *Part, H_MATRIX Matrix);

PART     *PARTLoad              (const char *FileName, 
				 char *PartName, dual check);
dual      PARTSave              (const char *FileName, 
				 PART *Part, dual triangle);
void      PARTWritePro          (FILE *fp, PART *Part);
dual      PARTSavePro           (char *FileName, PART *Part);
dual      PARTWritePLY          (PART *Part, char *BaseName);
PART     *PARTLoadDXF           (FILE *fp, char *PartName, dual check);

dual      PARTIsClosed          (PART *Part);
dual      PARTIsPlanar          (PART *Part);
dual      PARTIsConvex          (PART *Part);
dual      PARTCheckEuler        (PART *Part);
void      PARTCalcInfo          (PART *Part);
void      PARTCalcNormals       (PART *Part, double Range);

dual      PARTFindPolygon       (PART *Part, double x, double y, double z,
				 OBJECT **Object, POLYGON **Polygon);
dual      PARTGetPolygon        (PART *Part, int ID, 
				 OBJECT **Object, POLYGON **Polygon);
int       PARTPolygonAmount     (PART *Part);
int       PARTTriangleAmount    (PART *Part);
int       PARTPointAmount       (PART *Part);
void      PARTGetScale          (PART *Part, 
				 double *sx, double *sy, double *sz);

void      PARTSetFrame          (PART *Part, H_MATRIX Matrix);
void      PARTSetColor          (PART *Part, int Color);

FEATURE  *PARTNewFeature        (void);
char     *PARTMakeFeatureName   (FEATURE *Feature, char *Result);
dual      PARTCalcFaceCS        (PART *Part, POLYGON *Polygon, H_MATRIX *CS);
FEATURE  *PARTAddFeature        (PART *Part, OBJECT *Object, int Type, 
				 POLYGON *Polygon1, POLYGON *Polygon2, 
				 POLYGON *Polygon3, H_MATRIX Matrix);
int       PARTFeatureAmount     (PART *Part);
FEATURE  *PARTGetFeature        (PART *Part, char *Name);
void      PARTFreeFeature       (FEATURE *Feature);
void      PARTRemoveFeature     (PART *Part, FEATURE *Feature);

int       PARTGetFaceContact    (PART *Part1, PART *Part2, 
				 FEATURE *F1, FEATURE *F2);
int       PARTGetShaftContact   (PART *Part1, PART *Part2, 
				 FEATURE *F1, FEATURE *F2, dual fits);

PART     *PARTCreateArrow       (const char *Name, H_MATRIX Frame, 
				 double Radius1,   double   Len1, 
				 double Radius2,   double   Len2, 
				 double Angle,     int      Facettes, 
				 double Segment,   int      Color);

PART     *PARTCreateBolid       (const char *Name,  H_MATRIX Frame,
                                 double  Radius1,   double   Radius2, 
				 double  Radius3,   double   SizeZ, 
				 int     Facettes1, int      Facettes2, 
				 double  Segment,   int      Color);

PART     *PARTCreateCone        (const char *Name, H_MATRIX Frame,
				 double  Radius1,  double   Radius2, 
				 double  Radius3,  double   SizeZ, 
				 double  Segment,  int      Facettes, 
				 int     Color);

PART     *PARTCreateCylinder    (const char *Name, H_MATRIX Frame,
				 double  Radius,   double   Depth, 
				 double  Segment,  int      Facettes, 
				 int     Color);

PART     *PARTCreateGear        (const char *Name, H_MATRIX Frame,
				 double  Radius1,  double Radius2, 
				 double  Radius3, double Depth1, double Depth2,
				 int     Teeth,   double Skew,   int    Color);

PART     *PARTCreateIcosahedron (const char *Name, H_MATRIX Frame,
				 double  Radius1, double Radius2, 
				 int     Density, int    Iteration, int Color);

PART     *PARTCreatePipe        (const char *Name,   H_MATRIX Frame,
				 double  Radius1,    double   Radius2, 
				 double  Thickness1, double   Thickness2,
				 double  SizeZ,      double   Segment,
				 int     Facettes,   int      Color);

PART     *PARTCreatePrism       (const char *Name, H_MATRIX Frame,
				 double  Width1,  double Width2,  double OffsX,
				 double  Depth1,  double Depth2,  double OffsY,
				 double  Height1, double Height2, double OffsZ,
				 int     Color);

PART     *PARTCreateQuader      (const char *Name, H_MATRIX Frame,
				 double  SizeX,    double  SizeY, double SizeZ,
				 int     Color);

PART     *PARTCreateRot         (const char *Name, H_MATRIX  Frame,
				 int     Number,   POINT2D  *Points, 
				 double  Segment,  int       Facettes,
				 double  Radius,   int       Color);

PART     *PARTCreateScrew       (const char *Name,  H_MATRIX Frame,
				 double  Radius1,   double   Radius2, 
				 double  Height,    double   Increase, 
				 double  Offset,    int      Facettes, 
				 dual    Direction, int      Color);

PART     *PARTCreateSlice       (const char *Name,  H_MATRIX Frame,
				 double  Radius1,   double   Radius2, 
				 double  Radius3,   double   SizeZ, 
				 int     Facettes1, int      Facettes2, 
				 double  Segment,   int      Color);

PART     *PARTCreateSphere      (const char *Name,  H_MATRIX Frame,
				 double  Radius, 
				 double  SegmentX,  double   SegmentZ,
				 int     Facettes,  int      Color);

PART     *PARTCreateSpring      (const char *Name,  H_MATRIX Frame,
				 double  Radius1,   double   Radius2, 
				 double  Radius3,
				 double  Height,    double   Increase,
				 int     Facettes1, int      Facettes2, 
				 dual    Direction, int      Color);

PART     *PARTCreateSweep       (const char *Name,  H_MATRIX   Frame,
	  /* Extrusion */	 int    Number,     POINT2D *Points, 
				 double Height,     int   Color);

PART     *PARTCreateTorus       (const char *Name,  H_MATRIX Frame,
				 double  Radius1,   double   Radius2, 
				 int     Facettes1, int      Facettes2,
				 double  Segment,   int      Color);

PART     *PARTCreateWedge       (const char *Name, H_MATRIX Frame,
				 double  Width1,   double   Width2,
				 double  Height1,  double   Height2,
				 double  Depth1,   double   Depth2, int Color);


/* CSG-Operations: */
PART     *PARTIntersection      (const char *Name, H_MATRIX Frame,
				 const PART *Part1, H_MATRIX WTP1, 
				 const PART *Part2, H_MATRIX WTP2,
				 int  Color);
PART     *PARTDifference        (const char *Name, H_MATRIX Frame,
				 const PART *Part1, H_MATRIX WTP1, 
				 const PART *Part2, H_MATRIX WTP2,
				 int Color);
PART     *PARTUnion             (const char *Name, H_MATRIX Frame,
				 const PART *Part1, H_MATRIX WTP1, 
				 const PART *Part2, H_MATRIX WTP2,
				 int Color);
INTERSECT *PARTGetIntersections (const PART *Part1, H_MATRIX WTP1, 
				 const PART *Part2, H_MATRIX WTP2);

#ifdef __cplusplus
}
#endif

#endif



