#include <math.h>

#include "partGlobal.h"


#define NEW_VERSION


typedef struct POLYGON_LIST
{
         POLYGON      *Polygon;
         dual          used;
  struct POLYGON_LIST *Next;
} POLYGON_LIST;

typedef struct PART_DATA
{
         OBJECT    *Object;
         POINT3D   *SaveList;
         dual       updated;
         POLYGON_LIST *List;
  struct PART_DATA *Next;
} PART_DATA;


/* ********************************************************************** */
/* ** IDEA of contact-checking-algorithm:                              ** */
/* ** The related parts may be very complex. So it's useful to reduce  ** */
/* ** the data first. This will we done by storing all the possible    ** */
/* ** polygons, which are coplanar to the given polygon into a         ** */
/* ** 2dimensional-list: (1st-dim: objects, 2nd-dim: polygons).        ** */
/* ** Then we search recursively                                       ** */
/* ** a) for other polygons in the same object, which has directly or  ** */
/* **    indirect contact to the polygon and                           ** */
/* ** b) for polygons in other objects of the same part, which has a   ** */
/* **    contact to any of the found polygons in a)                    ** */
/* ** This algorithm must be done for both parts. Finally we check for ** */
/* ** physical contact between any found polygon in part one with any  ** */
/* ** polygon of part two. (fortunately - we can abort, if we find the ** */
/* ** first real contact)                                              ** */
/* ********************************************************************** */



/* ********************************************************************** */
/* ** Walk through all objects polygons and store the coplanar         ** */
/* ** polygons into the data-structure. Each object has it's own       ** */
/* ** vertex-list. For the contact-analysis we need world-coordinates. ** */
/* ** To restore the local vertex-positions after calculation we save  ** */
/* ** them into a 'SaveList'.                                          ** */
/* ********************************************************************** */
static dual PART_AddObject (PART_DATA **List, 
			    PART *Part, OBJECT *Object, POLYGON *Polygon)
{
  unsigned int  count;
  POLYGON      *PolyPtr;
  POLYGON_LIST *Elem, *PLast = NULL;
  PART_DATA    *ObjPtr, *OLast;

  /* Check, if object already stored in list */
  ObjPtr = OLast = *List;
  while (ObjPtr != NULL && ObjPtr->Object != Object)
  {
    OLast = ObjPtr;
    ObjPtr = ObjPtr->Next;
  }

  if (ObjPtr != NULL)
    return FALSE;

  /* Store all coplanar polygons: */
  PolyPtr = Object->Data.Polyeder->List;
  while (PolyPtr != NULL)
  {
    /* All objects of one part have same coord-sys, so we can use         */
    /* relative coordinates to check, if objects touches each other       */
    if (GEOVectorAngle(Polygon->Info.n, PolyPtr->Info.n) < D_ANGLE &&
	fabs(GEOAreaPointDist(Polygon, PolyPtr->Info.Center)) < D_DIST)
    {
      Elem = (POLYGON_LIST*)malloc (sizeof(POLYGON_LIST));
      Elem->Polygon = PolyPtr;
      Elem->used    = FALSE;
      Elem->Next    = NULL;

      /* If we want to check the contact between different parts, we      */
      /* needs to convert the part-points into world-coordinates.         */
      if (ObjPtr == NULL)               /* Create a new Object-List       */
      {
	ObjPtr = (PART_DATA*)malloc (sizeof(PART_DATA));
	ObjPtr->Object   = Object;
	ObjPtr->SaveList = 
	  (POINT3D *)malloc (sizeof(POINT3D) * Object->VertexNo);
	memcpy (ObjPtr->SaveList, Object->VertexList, 
		sizeof(POINT3D) * Object->VertexNo);
	for (count = 0; count < Object->VertexNo; ++count)
	  Object->VertexList[count] =
	    GEOMatrixMult (Part->WTP, Object->VertexList[count]);
	ObjPtr->updated  = FALSE;
	ObjPtr->List     = Elem;
	ObjPtr->Next     = NULL;

	if (*List == NULL)
	  *List = ObjPtr;
	else
	  OLast->Next = ObjPtr;
      }
      else
	PLast->Next = Elem;

      PLast = Elem;
    }
    PolyPtr = PolyPtr->Next;
  }

  return (ObjPtr != NULL);
}


/* ********************************************************************** */
/* ** Create a List with all valid objects                             ** */
/* ********************************************************************** */
static PART_DATA *PART_CreateList (PART *Part, 
                                   OBJECT *Object, POLYGON *Polygon)
{
  OBJECT    *ObjPtr;
  PART_DATA *List = NULL;

  PART_AddObject (&List, Part, Object, Polygon);

  ObjPtr = Part->ObjectList;
  while (ObjPtr != NULL)
  {
    if (ObjPtr != Object && ObjPtr->Type == PART_PLANE)
      PART_AddObject (&List, Part, ObjPtr, Polygon);

    ObjPtr = ObjPtr->Next;
  }

  return List;
}


/* ********************************************************************** */
/* ** locate a polygon in the list and set it's used-flag. On success  ** */
/* ** the pointer to the object's data is returned.                    ** */
/* ********************************************************************** */
static PART_DATA *PART_SetFlag (PART_DATA *List, 
				OBJECT *Object, POLYGON *Polygon)
{
  POLYGON_LIST *PolyPtr;
  PART_DATA    *ObjPtr;

  ObjPtr = List;
  while (ObjPtr != NULL && ObjPtr->Object != Object)
    ObjPtr = ObjPtr->Next;

  if (ObjPtr == NULL)
    return NULL;

  PolyPtr = ObjPtr->List;
  while (PolyPtr != NULL && PolyPtr->Polygon != Polygon)
    PolyPtr = PolyPtr->Next;

  if (PolyPtr == NULL || PolyPtr->used)
    return NULL;

  PolyPtr->used = TRUE;
  return ObjPtr;
}


/* ********************************************************************** */
/* ** Look for another object in the same part, which is contacted to  ** */
/* ** the given polygon. On successs the Object-Ptr and the found      ** */
/* ** polygon and the value TRUE will be returned. Unfortunately, it   ** */
/* ** is not sure, that IRIP parts suffice the 'Euler'-formula. So it  ** */
/* ** is inisufficient to check for common edges.                      ** */
/* ********************************************************************** */
static dual PART_SearchObj (PART_DATA *List, 
			    PART *Part, OBJECT **Object, POLYGON **Polygon)
{
  int           Points;
  POLYGON      *Contact;
  POLYGON_LIST *PolyPtr;
  PART_DATA    *ObjPtr;

  if (Part->ObjectNo == 1)
    return FALSE;

  ObjPtr = List;
  while (ObjPtr != NULL)
  {
    /* object already found or same object ? */
    if (! ObjPtr->updated && ObjPtr->Object != *Object)
    {
      PolyPtr = ObjPtr->List;
      while (PolyPtr != NULL)
      {
	if (! PolyPtr->used &&
	    (Contact = GEOPolygonContact(*Polygon, PolyPtr->Polygon)) != NULL)
	{
	  Points = Contact->Number;
	  PRO2 ("Object found (%d): %d\n", Points, PolyPtr->Polygon->Info.ID)
	  GEOFreePolygon (Contact, TRUE);
	  if (Points > 1)
	  {
	    *Object  = ObjPtr->Object;
	    *Polygon = PolyPtr->Polygon;
	    return TRUE;
	  }
	}
	PolyPtr = PolyPtr->Next;
      }
    }
    ObjPtr = ObjPtr->Next;
  }

  return FALSE;
}


/* ********************************************************************** */
/* ** This (double-)recursive function locates all polygons in the     ** */
/* ** same object which are (direct or indirect) connected to the      ** */
/* ** polygon with a common edge. Also with each found polygon, we     ** */
/* ** must search recursively for other objects in the part, which     ** */
/* ** will expand this polygon-face. To avoid cycles we mark all found ** */
/* ** objects as updated.                                              ** */
/* ********************************************************************** */
static void PART_SearchList (PART_DATA *List,
			     PART *Part, OBJECT *Object, POLYGON *Polygon)
{
  int           Points;
  POLYGON      *Poly, *Contact;
  POLYGON_LIST *PolyPtr;
  PART_DATA    *ObjPtr;

  if ((ObjPtr = PART_SetFlag(List, Object, Polygon)) == NULL)
    return;

  PolyPtr = ObjPtr->List;
  while (PolyPtr != NULL)
  {
    if (! PolyPtr->used && (Poly = PolyPtr->Polygon) != Polygon)
      if ((Contact = GEOPolygonContact(Polygon, Poly)) != NULL)
      {
	Points = Contact->Number;
	GEOFreePolygon (Contact, TRUE);
	if (Points > 1)
	  PART_SearchList (List, Part, Object, Poly);
      }

    PolyPtr = PolyPtr->Next;
  }

  ObjPtr->updated = TRUE;

  if (PART_SearchObj(List, Part, &Object, &Polygon))
    PART_SearchList (List, Part, Object, Polygon);
}


#ifdef _PROTOCOL
  /* ********************************************************************** */
  /* ** print a protocol of those polygons, wehich are identified to     ** */
  /* ** contact to the feature's polygon                                 ** */
  /* ********************************************************************** */
  static void PART_PrintList (PART_DATA *List, char *Text)
  {
    POLYGON_LIST *PolyPtr;
    PART_DATA    *ObjPtr;

    PRO1 ("Text: %s\n", Text)

    ObjPtr = List;
    while (ObjPtr != NULL)
    {
      PRO0 ("Object:")
      PolyPtr = ObjPtr->List;
      while (PolyPtr != NULL)
      {
	PRO2 ("\\ %d%c", PolyPtr->Polygon->Info.ID, PolyPtr->used? '+':'-')
	PolyPtr = PolyPtr->Next;
      }
      PRO0 ("\\\n")
      ObjPtr = ObjPtr->Next;
    }
  }
#endif


/* ********************************************************************** */
/* ** Erase the created help-list                                      ** */
/* ********************************************************************** */
static void PART_FreeList (PART_DATA *List)
{
  POLYGON_LIST *PolyPtr, *NextPoly;
  PART_DATA    *ObjPtr, *NextObj;

  ObjPtr = List;
  while (ObjPtr != NULL)
  {
    NextObj = ObjPtr->Next;
    memcpy (ObjPtr->Object->VertexList, ObjPtr->SaveList,
	    sizeof(POINT3D) * ObjPtr->Object->VertexNo);
    if (ObjPtr->SaveList != NULL)
      free (ObjPtr->SaveList);
    PolyPtr = ObjPtr->List;
    while (PolyPtr != NULL)
    {
      NextPoly = PolyPtr->Next;
      free (PolyPtr);
      PolyPtr = NextPoly;
    }
    free (ObjPtr);
    ObjPtr = NextObj;
  }
}


#if 0
/* ******************************************************************** */
/* ** Compares each entry in the List with the given polygon.        ** */
/* ******************************************************************** */
static int PART_CheckList (PART_DATA *List, 
			   PART *Part, OBJECT *Object, POLYGON *Polygon)
{
  dual          found = FALSE;
  int           count;
  VECTOR       *Points, *SaveList;
  POLYGON      *Contact;
  POLYGON_LIST *PolyPtr;
  PART_DATA    *ObjPtr;

  SaveList = Polygon->Points;

  Points = (VECTOR*)malloc (sizeof(VECTOR) * Object->VertexNo);
  for (count = 0; count < Object->VertexNo; ++count)
    Points[count] = GEOMatrixMult (Part->WTP, Object->VertexList[count]);

  Polygon->Points = Points;

  ObjPtr = List;
  while (ObjPtr != NULL && ! found)
  {
    PolyPtr = ObjPtr->List;
    while (PolyPtr != NULL && ! found)
    {
      if (PolyPtr->used && (Contact = GEOPolygonContact(PolyPtr->Polygon,
							Polygon)) != NULL)
      {
	found = (Contact->Number > 2);
	GEOFreePolygon (Contact, TRUE);
      }
      PolyPtr = PolyPtr->Next;
    }
    ObjPtr = ObjPtr->Next;
  }

  free (Points);

  Polygon->Points = SaveList;

  if (found)
    return R_CONTACT;
  else
    return V_CONTACT;
}
#endif


/* ********************************************************************** */
/* ** check each (used) polygon in List1 with each (used) polygon      ** */
/* ** in List2 for a physical contact                                  ** */
/* ********************************************************************** */
static int PART_CheckContact (PART_DATA *List1, PART_DATA *List2)
{
  int           Points;
  POLYGON      *Contact;
  POLYGON_LIST *PolyPtr1, *PolyPtr2;
  PART_DATA    *ObjPtr1, *ObjPtr2;

  ObjPtr1 = List1;
  while (ObjPtr1 != NULL)
  {
    PolyPtr1 = ObjPtr1->List;
    while (PolyPtr1 != NULL)
    {
      if (PolyPtr1->used)
      {
	ObjPtr2 = List2;
	while (ObjPtr2 != NULL)
	{
	  PolyPtr2 = ObjPtr2->List;
	  while (PolyPtr2 != NULL)
	  {
	    if (PolyPtr2->used &&
		(Contact = GEOPolygonContact(PolyPtr1->Polygon,
					     PolyPtr2->Polygon)) != NULL)
	    {
	      Points = Contact->Number;
	      GEOFreePolygon (Contact, TRUE);
	      if (Points > 2)
		return R_CONTACT;
	      else
		PRO1 ("Contact: %d Points\n", Points)
	    }
	    PolyPtr2 = PolyPtr2->Next;
	  }
	  ObjPtr2 = ObjPtr2->Next;
	}
      }
      PolyPtr1 = PolyPtr1->Next;
    }
    ObjPtr1 = ObjPtr1->Next;
  }
  PRO0 ("NO CONTACT\n")

  return V_CONTACT;
}


/* ********************************************************************** */
/* ** Check contact-type of two face-polygons [R_CONTACT | V_CONTACT]  ** */
/* ** INPUT:  Part1                                                    ** */
/* **         Part2   Parts, which shall be checked for contact-type   ** */
/* **         F1, F2  Related features (must be of type FACE)          ** */
/* ** OUTPUT: (int)   X_CONTACT: An error occured                      ** */
/* **                 R_CONTACT: if both faces have a physical contact ** */
/* **                 V_CONTACT: there's no contact polygon between    ** */
/* **                            both part's features                  ** */
/* ********************************************************************** */
int PARTGetFaceContact (PART *Part1, PART *Part2, FEATURE *F1, FEATURE *F2)
{
  int        Result = X_CONTACT;
  double     Angle = 0.0;
  POINT3D    Center1, Center2;
  VECTOR3D   WTN1, WTN2;
  PART_DATA *PList1, *PList2;

  PRO_START ("\n")

  if (F1->Type != PART_FACE || F2->Type != PART_FACE)
    PRO_RETURN ("Invalid feature-type(s)", X_CONTACT)

  if (F1->Object == NULL || F2->Object == NULL)
    PRO_RETURN ("Invalid object-Ptr", X_CONTACT)

  /* Check bubble-radius (using world-coordinates) */
  Center1 = GEOMatrixMult (Part1->WTP, F1->Object->Info.Center);
  Center2 = GEOMatrixMult (Part2->WTP, F2->Object->Info.Center);

  if (VEC_LEN(GEOVectorSub(Center2, Center1)) >
      (F1->Object->Info.Bubble + F2->Object->Info.Bubble) + D_DIST)
    PRO_RETURN ("No Bubble contact ", V_CONTACT)

  /* Do we have really coplanar polygons ? */
  WTN1 = GEOMatrixVector(GEOMatrixProd(Part1->WTP, 
				       GEOMatrixProd(Part1->PTC, F1->CTF)), N);
  WTN2 = GEOMatrixVector(GEOMatrixProd(Part2->WTP, 
				       GEOMatrixProd(Part2->PTC, F2->CTF)), N);

  Angle = GEOVectorAngle (WTN1, WTN2);
  if (Angle < D_ANGLE || (180.0-Angle) < D_ANGLE)
  {
#if 0
    /* What's about the distance between both polygon-planes ? */
    Dist = GEOPlanePointDist (F1->Polygon[0], F2->Polygon[0]->Info.Center);
    if (fabs(Dist) > D_DIST)
      PRO_RETURN ("polygons have different planes", V_CONTACT)
#endif

    /* Compute all feature-polygons and store their world-coordiante */
    PList1 = PART_CreateList (Part1, F1->Object, F1->Polygon[0]);
    PART_SearchList (PList1, Part1, F1->Object, F1->Polygon[0]);
    #ifdef _PROTOCOL
      PART_PrintList (PList1, "Part1:");
    #endif

    /* 
       Before we start the second complex search, let's check the polygon
       of the 2nd feature with the computed list of feature 1.
       It's very probable, that we find an physical contact here. So we 
       might speed up the computation.
       The needed computing time for a 21-part assembly (motor) could be
       reduced from 2 min 15 sec to 1 min 45.
    */
#if 0
    Result = PART_CheckList (PList1, Part2, F2->Object, F2->Polygon[0]);
#endif
    if (Result != R_CONTACT)
    {
      PList2 = PART_CreateList (Part2, F2->Object, F2->Polygon[0]);
      PART_SearchList (PList2, Part2, F2->Object, F2->Polygon[0]);
      #ifdef _PROTOCOL
        PART_PrintList (PList2, "Part2:");
      #endif

      /* Is there a contact-polygon between both list's polygons ? */
      Result = PART_CheckContact (PList1, PList2);

      PART_FreeList (PList1);
      PART_FreeList (PList2);
    }

    if (Result == R_CONTACT)
      PRO_RETURN ("R-CONTACT", R_CONTACT)
  }
  else
  {
    PRO1 ("Angle: %f\n", Angle);
    PRO_RETURN ("polygons are not coplanar", V_CONTACT);
  }

  PRO0 ("V-CONTACT\n")

  PRO_END
  return V_CONTACT;
}


#ifdef OLD_VERSION
static dual PART_InsertPolygon (POLYGON_LIST **PList, POLYGON *Polygon)
{
  POLYGON_LIST *Elem, *Ptr, *Last;

  /* Has the polygon been inserted before ? */
  Ptr = Last = *PList;
  while (Ptr != NULL && Ptr->Polygon != Polygon)
  {
    Last = Ptr;
    Ptr  = Ptr->Next;
  }

  if (Ptr != NULL)                      /* already stored in list         */
    return FALSE;

  Elem = (POLYGON_LIST *)malloc (sizeof(POLYGON_LIST));
  Elem->Polygon = Polygon;
  Elem->Next    = NULL;

  if (Last == NULL)                     /* insert it into the list        */
    *PList = Elem;
  else
    Last->Next = Elem;

  return TRUE;
}


/* ********************************************************************** */
/* ** Deallocate memory of a polygon-list                              ** */
/* ********************************************************************** */
static void PART_FreePolyList (POLYGON_LIST *PList)
{
  POLYGON_LIST *Elem, *Next;

  Elem = PList;
  while (Elem != NULL)
  {
    Next = Elem->Next;
    free (Elem);
    Elem = Next;
  }
}


/* ********************************************************************** */
/* ** Writes out a polygon-list (for debug-purpose only)               ** */
/* ********************************************************************** */
static void PART_PrintPolyList (POLYGON_LIST *PList, char *Text)
{
  POLYGON_LIST *Elem;

  PRO0 ("POLYGON-LIST1:")
  Elem = PList;
  while (Elem != NULL)
  {
    PRO1 ("\\  %d", Elem->Polygon->Info.ID)
    Elem = Elem->Next;
  }
  PRO0 ("\\\n")
}


/* ********************************************************************** */
/* ** Insert a new object into an object-list. If OList is NULL a new  ** */
/* ** List will be created.                                            ** */
/* ********************************************************************** */
static void PART_InsertObject (OBJECT_LIST **OList, OBJECT *Object)
{
  OBJECT_LIST *Obj, *Ptr, *Last;

  Ptr = Last = *OList;
  while (Ptr != NULL && Ptr->Object != Object)
  {
    Last = Ptr;
    Ptr  = Ptr->Next;
  }

  if (Ptr != NULL)
    return;

  Obj = (OBJECT_LIST *)malloc (sizeof(OBJECT_LIST));
  Obj->Object = Object;
  Obj->SaveList = (VECTOR *)malloc (sizeof(VECTOR) * Object->VertexNo);
  memcpy (Obj->SaveList, Object->VertexList, 
	  sizeof(VECTOR) * Object->VertexNo);
  Obj->Next = NULL;

  if (Last == NULL)
    *OList = Obj;
  else
    Last->Next = Obj;
}


/* ********************************************************************** */
/* ** Updates all object-points by multiplying with a given matrix     ** */
/* ********************************************************************** */
static void PART_UpdateObjPoints (OBJECT_LIST *OList, MATRIX WTP) 
{
  int          count;
  OBJECT_LIST *Elem;

  Elem = OList;
  while (Elem != NULL)
  {
    /* now compute the object's world-coordinates */
    for (count = 0; count < Elem->Object->VertexNo; ++count)
      Elem->Object->VertexList[count] = 
	GEOMatrixMult (WTP, Elem->Object->VertexList[count]);
    Elem = Elem->Next;
  }
}
  

/* ********************************************************************** */
/* ** For checking polygon contact we must place the parts at their    ** */
/* ** world-position. At the end we need to restore the initial        ** */
/* ** positions.                                                       ** */
/* ********************************************************************** */
static void PART_RestoreObjPoints (OBJECT_LIST *OList)
{
  OBJECT_LIST *Elem, *Next;

  Elem = OList;
  while (Elem != NULL)
  {
    Next = Elem->Next;
    memcpy (Elem->Object->VertexList, Elem->SaveList, 
	    sizeof(VECTOR) * Elem->Object->VertexNo);
    free (Elem->SaveList);
    free (Elem);
    Elem = Next;
  }
}


/* ********************************************************************** */
/* ** Searches an polygon in another object of the given part, which   ** */
/* ** is connected to the polygon and lies in the same area            ** */
/* ********************************************************************** */
static dual PART_CheckObjContact (PART *Part, 
				  OBJECT **Object, POLYGON **Polygon)
{
  OBJECT  *ObjPtr;
  POLYGON *PolyPtr, *Contact;

  if (Part->ObjectNo == 1)              /* no other object                */
    return FALSE;

  ObjPtr = Part->ObjectList;
  while (ObjPtr != NULL)
  {
    if (! ObjPtr->Info.updated && ObjPtr != *Object && ObjPtr->Type == PLANES)
    {
      PolyPtr = ObjPtr->Data.Polyeder->List;
      while (PolyPtr != NULL)
      {
	if (GEOVectorAngle((*Polygon)->Info.n, PolyPtr->Info.n) < D_ANGLE &&
	    fabs(GEOPlanePointDist(*Polygon, PolyPtr->Info.Center)) < D_DIST &&
	    (Contact = GEOPolygonContact(*Polygon, PolyPtr)) != NULL)
	{
	  PRO1 ("Object found: %d\n", PolyPtr->Info.ID)
	  GEOFreePolygon (Contact, TRUE);
	  *Object  = ObjPtr;
	  *Polygon = PolyPtr;
	  return TRUE;
	}
	PolyPtr = PolyPtr->Next;
      }
    }
    ObjPtr = ObjPtr->Next;
  }

  return FALSE;
}


/* ********************************************************************** */
/* ** Locate all polygons (also in other objects of the same part),    ** */
/* ** which are coherently with the given feature-polygon and store    ** */
/* ** them in a List                                                   ** */
/* ********************************************************************** */
static void PART_FeaturePolygons (POLYGON_LIST **PList, OBJECT_LIST **OList,
				  PART *Part, OBJECT *Object, POLYGON *Polygon)
{
  int      count1, count2;
  int      next1, next2;
  POLYGON *PolyPtr;

  if (! PART_InsertPolygon(PList, Polygon))
    return;

  /* Check for adjacent polygon(s) in the same object */
  PolyPtr = Object->Data.Polyeder->List;
  while (PolyPtr != NULL)
  {
    /* locate coplanar polygons */
    if (PolyPtr != Polygon && 
	GEOVectorAngle(Polygon->Info.n, PolyPtr->Info.n) < D_ANGLE &&
	fabs(GEOPlanePointDist(Polygon, PolyPtr->Info.Center)) < D_DIST)
    {
      PRO1 ("COPLANAR POLYGON: %d\n", PolyPtr->Info.ID)
      /* does exist a common edge ? */
      for (count1 = 0; count1 < Polygon->Number; ++count1)
	for (count2 = 0; count2 < PolyPtr->Number; ++count2)
	{
	  next1 = (count1+1) % Polygon->Number;
	  next2 = (count2+1) % PolyPtr->Number;
	  if ((Polygon->Vertex[count1] == PolyPtr->Vertex[count2] &&
	       Polygon->Vertex[next2]  == PolyPtr->Vertex[next2]) ||
	      (Polygon->Vertex[count1] == PolyPtr->Vertex[next2]  &&
	       Polygon->Vertex[next1]  == PolyPtr->Vertex[count2]))
	  {
	    PRO2 ("Common edge: %d - %d\n",
		  Polygon->Vertex[count1], Polygon->Vertex[next1])
	    PART_FeaturePolygons (PList, OList, Part, Object, PolyPtr);
	  }
	}
    }
    PolyPtr = PolyPtr->Next;
  }
  PRO0 ("Object checked\n")

  Object->Info.updated = TRUE;
  PART_InsertObject (OList, Object);

  /* search for other objects with adjacent polygon(s) */
  if (PART_CheckObjContact(Part, &Object, &Polygon))
    PART_FeaturePolygons (PList, OList, Part, Object, Polygon);
}


/* ********************************************************************** */
/* ** Test, if at least one polygon in the 1st List has a physical     ** */
/* ** contact to a polygon in the 2nd list.                            ** */
/* ********************************************************************** */ 
int PART_PolyListContact (POLYGON_LIST *List1, POLYGON_LIST *List2)
{
  int           Number;
  POLYGON      *Contact;
  POLYGON_LIST *Elem1, *Elem2;

  Elem1 = List1;
  while (Elem1 != NULL)
  {
    Elem2 = List2;
    while (Elem2 != NULL)
    {
      Contact = GEOPolygonContact (Elem1->Polygon, Elem2->Polygon);
      if (Contact != NULL)
      {
	Number = Contact->Number;
	GEOFreePolygon (Contact, TRUE);
	if (Number > 2)
	  return R_CONTACT;
      }
      Elem2 = Elem2->Next;
    }
    Elem1 = Elem1->Next;
  }
  PRO0 ("No contact between polygons\n")

  return V_CONTACT;
}

/* *********************************************************************** */
/* ** Concept: first we check if a collision is possible by testing     ** */
/* **          the bubble-radiuses. Then we test, if both Faces are     ** */
/* **          in the same area (should be because of symbolic-spatial  ** */
/* **          relations).                                              ** */
/* **          For both feature-polygon, we start at the given area and ** */
/* **          search recursively for polygons which has contact to it  ** */
/* **          and are coplanar. Also we must check other part's        ** */
/* **          objects for such polygons. All polygons of each part     ** */
/* **          are stored in a separate list (PList1, PList2).          ** */
/* **          The objects are stored in an object-list. While adding   ** */
/* **          a object to the list, we save the objects vertex-list    ** */
/* **          in oder to change the original (local) vertexes into     ** */
/* **          global positions.                                        ** */
/* **          Then a test for polygon-cut in both polygon-lists will   ** */
/* **          be performed (PART_PolyListContact). It breaks on the    ** */
/* **          first found contact-polygon (R-CONTACT).                 ** */
/* **          At last we must restore the object's vertex-lists and    ** */
/* **          free the allocated memory                                ** */
/* *********************************************************************** */
int PARTGetFaceContact (PART *Part1, PART *Part2, FEATURE *F1, FEATURE *F2)
{
  int           Result;
  double        Angle = 0.0, Dist;
  VECTOR        Center1, Center2;
  VECTOR        WTN1, WTN2;
  POLYGON_LIST *PList1 = NULL, *PList2 = NULL;
  OBJECT       *Object;
  OBJECT_LIST  *OList1 = NULL, *OList2 = NULL;

  PRO_START ("\n")

  if (F1->Type != PART_FACE || F2->Type != PART_FACE)
    PRO_RETURN ("Invalid feature-type(s)", U_CONTACT)

  if (F1->Object == NULL || F2->Object == NULL)
    PRO_RETURN ("Invalid object-Ptr", U_CONTACT)

  /* calculate world-positions for both object's center */
  Center1 = GEOMatrixMult (Part1->WTP, F1->Object->Info.Center);
  Center2 = GEOMatrixMult (Part2->WTP, F2->Object->Info.Center);

  /* check bubble-radius */
  if (VEC_LEN(GEOVectorSub(Center2, Center1)) >
      (F1->Object->Info.Bubble + F2->Object->Info.Bubble) + D_DIST)
    PRO_RETURN ("No Bubble contact ", V_CONTACT)

  /* Do we have coplanar polygons ? */
#if 0
  WTN1 = GEOMatrixMult (Part1->WTP, F1->Polygon[0]->Info.n);
  WTN2 = GEOMatrixMult (Part2->WTP, F2->Polygon[0]->Info.n);
  PRO_VECTOR (F1->Polygon[0]->Info.n, "Norm1")
  PRO_VECTOR (F2->Polygon[0]->Info.n, "Norm2")
  PRO_VECTOR (WTN1, "WTN1")
  PRO_VECTOR (WTN2, "WTN2")
#endif

  WTN1 = GEOMatrixVector(GEOMatrixProd(Part1->WTP, 
				       GEOMatrixProd(Part1->PTC, F1->CTF)), N);
  WTN2 = GEOMatrixVector(GEOMatrixProd(Part2->WTP, 
				       GEOMatrixProd(Part2->PTC, F2->CTF)), N);

  Angle = GEOVectorAngle (WTN1, WTN2);
  if (Angle < D_ANGLE || 180.0-Angle  < D_ANGLE)
  {
#if 0
    /* What's about the distance between both polygon-planes ? */
    Dist = GEOPlanePointDist (F1->Polygon[0], F2->Polygon[0]->Info.Center);
    if (fabs(Dist) > D_DIST)
      PRO_RETURN ("polygons not coplanar", V_CONTACT)
#endif

    /* Compute world-coordiantes of polygon-points */
    Object = Part1->ObjectList;
    while (Object != NULL)
    {
      Object->Info.updated = FALSE;
      Object = Object->Next;
    }
    PART_FeaturePolygons (&PList1, &OList1, Part1, F1->Object, F1->Polygon[0]);
    PART_UpdateObjPoints (OList1, Part1->WTP);
    PART_PrintPolyList (PList1, "POLYGON-LIST1:");

    Object = Part2->ObjectList;
    while (Object != NULL)
    {
      Object->Info.updated = FALSE;
      Object = Object->Next;
    }
    PART_FeaturePolygons (&PList2, &OList2, Part2, F2->Object, F2->Polygon[0]);
    PART_UpdateObjPoints (OList2, Part2->WTP);
    PART_PrintPolyList (PList2, "POLYGON-LIST2:");

    /* Is there a contact-polygon between both list's polygons ? */
    Result = PART_PolyListContact (PList1, PList2);

    PART_RestoreObjPoints (OList1);
    PART_RestoreObjPoints (OList2);

    PART_FreePolyList (PList1);
    PART_FreePolyList (PList2);

    if (Result == R_CONTACT)
      PRO_RETURN ("Contact-polygon found -> R-CONTACT", R_CONTACT)
  }
  else
  {
    PRO1 ("Angle: %f\n", Angle);
    PRO_RETURN ("polygons are not coplanar", V_CONTACT);
  }

  PRO0 ("V-CONTACT\n")

  PRO_END
  return V_CONTACT;
}
#endif


/* ********************************************************************** */
/* ** calculate a shaft's radius using it's feature's x-axis and a     ** */
/* ** point, which lies (hopely) on the shaft-mantle                   ** */
/* ********************************************************************** */
static double PART_ShaftRadius (PART *Part, FEATURE *F)
{
  double     Result = 0.0;
  POINT3D    Pos;
  VECTOR3D   Dir;
  H_MATRIX   PTF;
  POLYGON   *Polygons[3];

  PTF = GEOMatrixProd (Part->PTC, F->CTF);
  Pos = GEOMatrixVector (PTF, P);
  Dir = GEOMatrixVector (PTF, N);

  if (PART_LocateFaces(Part, F->Polygon[0], Polygons))
    Result = GEOLinePointDist (Pos, Dir,
			       Polygons[0]->Points[Polygons[0]->Vertex[0]]);

  return Result;
}


/* ********************************************************************** */
/* ** Correspondending function to PART_ShaftRadius                    ** */
/* ********************************************************************** */
static double PART_HoleRadius (PART *Part, FEATURE *F)
{
  double    Result;
  POINT3D   Pos;
  VECTOR3D  Dir;
  H_MATRIX  PTF;
  POLYGON  *Polygon;

  PTF = GEOMatrixProd (Part->PTC, F->CTF);
  Pos = GEOMatrixVector (PTF, P);
  Dir = GEOMatrixVector (PTF, N);

  Polygon = F->Polygon[0];
  Result = GEOLinePointDist (Pos, Dir, Polygon->Points[Polygon->Vertex[0]]);

  return Result;
}


/* ********************************************************************** */
/* ** Evaluates the type of contact, which is defined between a shaft  ** */
/* ** and a hole by comparing the radius and position of both features.** */
/* ** INPUT:  Part1, Part2  Both parts, which has a fits/aligned-      ** */
/* **                       relationship.                              ** */
/* **         F1, F2        Corresponding features in the first and    ** */
/* **                       second part.                               ** */
/* **         fits          Decides, whether we have a fits or aligned ** */
/* **                       relation.                                  ** */
/* ** OUTPUT: (int)         R_CONTACT: physical contact                ** */
/* **                       V_CONTACT: virtual contact                 ** */
/* ********************************************************************** */
int PARTGetShaftContact (PART *Part1, PART *Part2, FEATURE *F1, FEATURE *F2,
			 dual fits)
{
  int      Result = R_CONTACT;
  double   r1, r2, len;
  VECTOR3D Diff;
  H_MATRIX WTF1, WTF2;

  PRO_START ("\n")

  if (F1->Type == PART_HOLE)
    r1 = PART_HoleRadius (Part1, F1);
  else
    r1 = PART_ShaftRadius (Part1, F1);
  PRO1 ("F1-Radius: %f\n", r1)

  if (F2->Type == PART_HOLE)
    r2 = PART_HoleRadius (Part2, F2);
  else
    r2 = PART_ShaftRadius (Part2, F2);
  PRO1 ("F2-Radius: %f\n", r2)

  if (F1->Type == F2->Type)
    PRO_RETURN ("Must be V-CONTACT", V_CONTACT)

  if (fabs(r2 - r1) <= TOOLGetConfigDouble(CFG_DELTA_RADIUS)) 
    Result = R_CONTACT;
  else
    Result = V_CONTACT;

  /* Test, if shaft is inside the hole */
  if (Result == R_CONTACT && fits)
  {
    WTF1 = GEOMatrixProd (Part1->WTP, GEOMatrixProd(Part1->PTC, F1->CTF));
    WTF2 = GEOMatrixProd (Part2->WTP, GEOMatrixProd(Part2->PTC, F2->CTF));

    Diff = GEOVectorSub (GEOMatrixVector(WTF2, P), GEOMatrixVector(WTF1, P));
    len = VEC_LEN(Diff);
    if (len == 0.0 ||
	fabs(GEOVectorAngle(GEOMatrixVector(WTF1, N), Diff)) < 90.0)
      Result = V_CONTACT;
  }

  PRO_END
  return Result;
}


