#include <math.h>
#include <string.h>

#include "partGlobal.h"


/* ********************************************************************** */
/* ** Create a Name for a specific feature using it's type and it's ID ** */
/* ** Valid names may be: 'SHAFT2', 'HOLE3' or 'FACE5'                 ** */
/* ** INPUT:  Feature  Pointer to feature-data                         ** */
/* ** IN/OUT: Result   allocated memory for the feature-name           ** */
/* ** OUTPUT: (char*)  Copy of memory-area                             ** */
/* ********************************************************************** */
char *PARTMakeFeatureName (FEATURE *Feature, char *Result)
{
  if (Feature == NULL)
    strcpy (Result, "FIXED");
  else
    sprintf (Result, "%s%d", FEATURE_NAME(Feature->Type), Feature->ID);

  return Result;
}


/* ********************************************************************** */
/* ** Create a new and empty feature using default-values.             ** */
/* ** INPUT:  none                                                     ** */
/* ** OUTPUT: (FEATURE*)  Allocated feature-structure with default     ** */
/* **                     values.                                      ** */
/* ********************************************************************** */ 
FEATURE *PARTNewFeature (void)
{
  FEATURE *Result;

  Result = (FEATURE *)malloc (sizeof(FEATURE));
  Result->Type       = -1;
  Result->ID         = 1;
  Result->Object     = NULL;
  Result->Polygon[0] = NULL;
  Result->Polygon[1] = NULL;
  Result->Polygon[2] = NULL;
  Result->CTF        = Matrix_1;
  Result->used       = 0;
  Result->Next       = NULL;

  return Result;
}


/* ********************************************************************** */
/* ** Gives back the memory, which has been allocated for a feature.   ** */
/* ** INPUT:  Feature  Deleteable feature                              ** */
/* ** OUTPUT: none                                                     ** */
/* ********************************************************************** */
void PARTFreeFeature (FEATURE *Feature)
{
  if (Feature != NULL)
    free (Feature);
}


/* ********************************************************************** */
/* ** This function delete a feature from a part's feature-list, if    ** */
/* ** will not be used anymore (used-counter is zero). If Part is      ** */
/* ** NULL, only the feature will be deleted.                          ** */
/* ** INPUT:  Part     Part containing the feature (may be NULL)       ** */
/* **         Feature  pecific feature                                 ** */
/* ** OUTPUT: none                                                     ** */
/* ********************************************************************** */
void PARTRemoveFeature (PART *Part, FEATURE *Feature)
{
  FEATURE *F = NULL, *LastF;

  PRO_START (" ")

  if (Feature == NULL)
    PRO_ABORT ("No Feature")

  (Feature->used)--;

  /* search feature in the part */
  if (Part != NULL && Feature->used < 1)
  {
    F = LastF = Part->FeatureList;
    while (F != NULL && F != Feature)
    {
      LastF = F;
      F = F->Next;
    }

    if (F != NULL)
    {
      if (F == Part->FeatureList)
	Part->FeatureList = F->Next;
      else
	LastF->Next = F->Next;
    }
  }

  /* now remove the features data if it is unused                         */
  if (Feature != NULL && Feature->used < 1)
  {
    PRO0 ("\\delete...")
    free (Feature);
  }
  else
    PRO0 ("\\reduce")

  PRO_END
}


/* ********************************************************************** */
/* ** Search for polygons, which are connected with a specified one.   ** */
/* ********************************************************************** */
dual PART_LocateFaces (PART *Part, POLYGON *Polygon, POLYGON **Polygons)
{
  int       count, Index = 0;
  POLYGON  *Ptr, *Polygon1, *Polygon2;
  POLYEDER *Polyeder = NULL;
  OBJECT   *Object;

  PRO_START (" ")

  /* Locate part's object (polyeder) */
  Object = Part->ObjectList;
  while (Polyeder == NULL && Object != NULL)
  {
    if (Object->Type == PART_PLANE)
    {
      Ptr = Object->Data.Polyeder->List;
      while (Ptr != NULL && Ptr != Polygon)
        Ptr = Ptr->Next;
      if (Ptr != NULL)
	Polyeder = Object->Data.Polyeder;
    }
    Object = Object->Next;
  }
  if (Polyeder == NULL)
    PRO_RETURN ("Polyeder not found", FALSE)

  PRO0 ("\\faces...")
  count = 0;
  while (Index < 3 && count < Polygon->Number-1)
  {
    if (GEOGetEdgePolygons(Polyeder, 
			   Polygon->Vertex[count], Polygon->Vertex[count+1],
			   &Polygon1, &Polygon2))
    {
      if (Polygon1 == Polygon)
	Polygons[Index] = Polygon2;
      else
	Polygons[Index] = Polygon1;
      Index++;
    }
    count++;
  }

  PRO_END
  return (Index == 3);
}


/* ********************************************************************** */
/* ** Counts the number of defined features of one part.               ** */
/* ** INPUT:  Part  Interesting part                                   ** */
/* ** OUTPUT: (int) Amount of defined features in the part's feature-  ** */
/* **               list.                                              ** */
/* ********************************************************************** */
int PARTFeatureAmount (PART *Part)
{
  int      count = 0;
  FEATURE *Feature;

  if (Part == NULL)
    return 0;

  Feature = Part->FeatureList;
  while (Feature != NULL)
  {
    count++;
    Feature = Feature->Next;
  }

  return count;
}


/* ********************************************************************** */
/* ** Locate a part's feature using the feature-name used by the       ** */
/* ** 'Automatischer Montage Planer'.                                  ** */
/* ** INPUT:  Part        Part, which should be searched for the       ** */
/* **                     feature.                                     ** */
/* **         Name        Name of searched feature                     ** */
/* **                     (see PARTMakeFeatureName for detail)         ** */
/* ** OUTPUT: (FEATURE*)  Pointer to found feature or NULL             ** */
/* ********************************************************************** */
FEATURE *PARTGetFeature (PART *Part, char *Name)
{
  int      Type = 0, ID;
  FEATURE *Feature;

  if (strncmp(Name, "FACE", 4) == 0)
  {
    Type = PART_FACE;
    ID   = atoi (Name+4);
  }
  elsif (strncmp(Name, "SHAFT", 5) == 0)
  {
    Type = PART_SHAFT;
    ID   = atoi (Name+5);
  }
  elsif (strncmp(Name, "HOLE", 4) == 0)
  {
    Type = PART_HOLE;
    ID   = atoi (Name+4);
  }
  elsif (strncmp(Name, "FIXED", 5) == 0)
  {
    Type = PART_FIXED;
    ID   = atoi (Name+5);
  }
  else
  {
    PRO1 ("Invalid Feature-Name: %s\n", Name)
    return NULL;
  }

  Feature = Part->FeatureList;
  while (Feature != NULL && (Feature->Type != Type || Feature->ID != ID))
    Feature = Feature->Next;

  return Feature;
}


/* ********************************************************************** */
/* ** Find out the index of the polygon's longest edge.                ** */
/* ********************************************************************** */
static int PART_GetPolygonVertex (POLYGON *Polygon)
{
  int      count, Index;
  int      Len, MaxLen = -1;
  int      Result = 0;
  POINT3D *Points;

  if (Polygon == NULL || Polygon->Vertex == NULL)
    return -1;

  Points = Polygon->Points;
  for (count = 0; count < Polygon->Number; ++count)
  {
    Index = (count+1) % Polygon->Number;
    Len = VEC_LEN(GEOVectorSub(Points[Polygon->Vertex[Index]],
			       Points[Polygon->Vertex[count]]));
    Index = (count+2) % Polygon->Number;
    if (Len > MaxLen && ! VEC_EQUAL(Points[Polygon->Vertex[count]],
				    Points[Polygon->Vertex[Index]]))
    {
      Result = count;
      MaxLen = Len;
    }
  }
  return Result;
}


/* ********************************************************************** */
/* ** Compute the face-coodinate-system of a part's polygon            ** */
/* ********************************************************************** */
dual PARTCalcFaceCS (PART *Part, POLYGON *Polygon, H_MATRIX *CS)
{
  int       Index;
  VECTOR3D  o, a;
  POINT3D  *Points;
  H_MATRIX  Matrix, PartInv;

  Index = PART_GetPolygonVertex (Polygon);

  Points = Polygon->Points;

  o = GEOVectorSub (Points[Polygon->Vertex[(Index+1)%Polygon->Number]],
                    Points[Polygon->Vertex[Index]]);
  a = GEOVectorSub (Points[Polygon->Vertex[(Index+2)%Polygon->Number]],
                    Points[Polygon->Vertex[Index]]);

  Matrix  = GEOMakeRightSys2 (o, a, Polygon->Info.Center);
  PartInv = GEOMatrixInv (Part->PTC);
  *CS     = GEOMatrixProd (PartInv, Matrix);

  PRO_HMATRIX ((*CS), "FACE-CS (PTF)")

  return TRUE;
}


/* ********************************************************************** */
/* ** Locates the point in the polygon, which lies in the same         ** */
/* ** direction as the matrix's normal-vector and has the largest      ** */
/* ** distance to it.                                                  ** */
/* ********************************************************************** */
double PART_FindPolygonPoint (H_MATRIX M, POLYGON *Polygon, double *Angle)
{
  int      count;
  double   Len, MaxLen = -1.0;
  double   Rot;
  POINT3D  Center;
  VECTOR3D Norm, Diff;

  if (Polygon == NULL)
    return -1.0;

  Center = GEOMatrixVector (M, P);
  Norm   = GEOMatrixVector (M, N);

  for (count = 0; count < Polygon->Number; ++count)
  {
    Diff = GEOVectorSub (Polygon->Points[Polygon->Vertex[count]], Center);
    Len = VEC_LEN(Diff);
    if (Len > MaxLen && (Rot = fabs(GEOVectorAngle(Norm, Diff))) <= 90.0)
    {
      *Angle = Rot;
      MaxLen = Len;
    }
  }

  return MaxLen;
}


/* ********************************************************************** */
/* ** compute a shaft- or hole-coordinate-system using the three side  ** */
/* ** polygons (and an end-face for shafts). On success, the function  ** */
/* ** returns TRUE and the computed coordiante-system. On error,       ** */
/* ** FALSE will be returned.                                          ** */
/* ********************************************************************** */
static dual PART_CalcCylCS (PART *Part, POLYGON *Polygon1, POLYGON *Polygon2,
			    POLYGON *Polygon3, POLYGON *Face, H_MATRIX *CS)
{
  dual     changed;
  double   Len, Angle, Factor;
  double   h[3], i[3], j[3], k[3];
  double   r = 0, s = 0, t = 0, u = 0;
  POINT3D  Pos; 
  VECTOR3D Dir;
  VECTOR3D c[3], n[3];
  VECTOR3D v[4], swap;
  H_MATRIX Matrix, PartInv;

  PRO_START ("\n")

  if (Polygon1 == NULL || Polygon2 == NULL || Polygon3 == NULL)
    PRO_RETURN ("3 polygons required", FALSE)

  *CS = Matrix_0;

  c[0] = Polygon1->Info.Center;
  c[1] = Polygon2->Info.Center;
  c[2] = Polygon3->Info.Center;
  n[0] = Polygon1->Info.n;
  n[1] = Polygon2->Info.n;
  n[2] = Polygon3->Info.n;

  /* at least, two vectors (n0 and n2) must be linear independent */
  if ((changed = (GEOVectorCollinear(n[0], n[2]) != 0)))
  {
    PRO0 ("SWAPPING FACES...\n")

    swap = c[0];
    c[0] = c[1];
    c[1] = swap;

    swap = n[0];
    n[0] = n[1];
    n[1] = swap;
  }

  h[0] = GEOVectorProd (GEOVectorSub(c[0], c[1]), n[0]);
  h[1] = GEOVectorProd (n[0], n[0]);
  h[2] = GEOVectorProd (n[1], n[0]);

  i[0] = GEOVectorProd (GEOVectorSub(c[0], c[1]), n[1]);
  i[1] = GEOVectorProd (n[0], n[1]);
  i[2] = GEOVectorProd (n[1], n[1]);

  j[0] = GEOVectorProd (GEOVectorSub(c[0], c[2]), n[0]);
  j[1] = GEOVectorProd (n[0], n[0]);
  j[2] = GEOVectorProd (n[2], n[0]);

  k[0] = GEOVectorProd (GEOVectorSub(c[0], c[2]), n[2]);
  k[1] = GEOVectorProd (n[0], n[2]);
  k[2] = GEOVectorProd (n[2], n[2]);


  if (h[2]*i[1] - h[1]*i[2] != 0.0)
  {
    r = (h[0]*i[2] - h[2]*i[0]) / (h[2]*i[1] - h[1]*i[2]);
    s = (h[0]*i[1] - h[1]*i[0]) / (h[2]*i[1] - h[1]*i[2]);
  }
  else
    PRO0 ("Division by zero for r and s")

  if (j[2]*k[1] - j[1]*k[2] != 0.0)
  {
    t = (j[0]*k[2] - j[2]*k[0]) / (j[2]*k[1] - j[1]*k[2]);
    u = (j[0]*k[1] - j[1]*k[0]) / (j[2]*k[1] - j[1]*k[2]);
  }
  else
    PRO0 ("Division by zero for t and u")

  v[0] = GEOVectorAdd (c[0], GEOVectorMult(r, n[0]));
  v[1] = GEOVectorAdd (c[1], GEOVectorMult(s, n[1]));
  v[2] = GEOVectorAdd (c[0], GEOVectorMult(t, n[0]));
  v[3] = GEOVectorAdd (c[2], GEOVectorMult(u, n[2]));

  if (fabs(VEC_LEN(GEOVectorSub(v[0], v[2]))) > D_DIST)
    PRO_RETURN ("Vector not zero", FALSE)

  /* For the present, compute axis-point and the axis-direction of the CS */
  Pos = GEOVectorMult (1.0/3.0, GEOVectorAdd(v[0], GEOVectorAdd(v[1],v[3])));
  Dir = GEOVectorCross (n[0], n[2]);

  if (changed)                     /* Did we swapped vector 2 and 3 ?     */
    Dir = GEOVectorMult (-1.0, Dir);

  /* check, if 3rd normal-vector is orthogonal to calculated axis         */
  if (fabs(fabs(GEOVectorAngle(Dir, n[1]))-90.0) > D_ANGLE)
    PRO_RETURN ("Faces not on shaft", FALSE)

  Matrix = GEOMakeRightSys2 (n[0], n[2], Pos);

  /* 
     Place coordinate-system on the shaft/hole-axis:
     find polygon-point with maximum distance (used for holes)
  */
  if (Face == NULL)
  {
    Len = PART_FindPolygonPoint (Matrix, Polygon1, &Angle);
    Factor = cos (RAD(Angle)) * Len;
  }
  else                  /* calc cutting-point between axis and face-plane */
  {
    Factor = (- Face->Info.a - GEOVectorProd(Face->Info.n, Pos)) /
               GEOVectorProd(Face->Info.n, GEOMatrixVector(Matrix, N));
    PRO1 ("Factor: %f\n", Factor)
  }

  Matrix.m[P][X] += Matrix.m[N][X] * Factor;
  Matrix.m[P][Y] += Matrix.m[N][Y] * Factor;
  Matrix.m[P][Z] += Matrix.m[N][Z] * Factor;

  PartInv = GEOMatrixInv (Part->PTC);
  *CS = GEOMatrixProd (PartInv, Matrix);

  PRO_HMATRIX ((*CS), "CYLINDER-CS")

  PRO_END
  return TRUE;
}


/* ********************************************************************** */
/* ** Add a feature to the part's feature-list. If there's already     ** */
/* ** the same feature (type, polygon (face/shaft), x-axis and         ** */
/* ** position (hole) of the feature-coordinate-system) stored in      ** */
/* ** the feature-list, then only the 'used'-counter of the stored     ** */
/* ** feature will be increased.                                       ** */
/* ********************************************************************** */
FEATURE *PARTAddFeature (PART *Part, OBJECT *Object, int Type, 
			 POLYGON *Polygon1, POLYGON *Polygon2, 
			 POLYGON *Polygon3, H_MATRIX Matrix)
{
  dual      found = FALSE;
  int       ID[4] = {0, 0, 0, 0};
  H_MATRIX  CS;
  POLYGON  *Polygons[3];
  FEATURE  *Feature = NULL;
  FEATURE  *Ptr = NULL;

  PRO_START ("\n")

  if (Part == NULL)
    PRO_RETURN ("Invalid part", NULL)

  PRO1 ("%s\n", FEATURE_NAME(Type))

  switch (Type)
  {
    case PART_SHAFT :
      if (! PART_LocateFaces(Part, Polygon1, Polygons) ||
	  ! PART_CalcCylCS(Part, Polygons[0], Polygons[1], Polygons[2], 
			         Polygon1, &CS))
	ERROR_RETURN0 ("Error Calculating SHAFT-CS", NULL)
      break;

    case PART_HOLE :
      if (! PART_CalcCylCS(Part, Polygon1, Polygon2, Polygon3, NULL, &CS))
	ERROR_RETURN0 ("Error Calculating HOLE-CS", NULL)
      break;

    case PART_FACE :
      if (! PARTCalcFaceCS(Part, Polygon1, &CS))
	ERROR_RETURN0 ("Error Calculating FACE-CS", NULL)
      break;

    case PART_FIXED :
      CS = Matrix;
      PRO_HMATRIX (CS, "FIXED-CS")
      break;

    default :
      ERROR_RETURN1 ("Invalid Feature-Type: %d", Type, NULL)
  }

  /* 
     Now, search the feature in the existing feature-list. Because
     MARS does not allow to use a FIXED-feature(-name) twice, we must
     create a new feature for it. So we needn't search for it !
  */
  if (Type != PART_FIXED)
    Ptr = Part->FeatureList;

  while (Ptr != NULL && ! found)
  {
    if (Ptr->Type == Type && Ptr->Object == Object)
      switch (Type)
      {
        case PART_SHAFT :
        case PART_FACE :
	  if (Ptr->Polygon[0] == Polygon1)
	    found = TRUE;
	  break;

        case PART_HOLE :
	  {
	    VECTOR3D Norm[2];
	    POINT3D  Pos[2];

	    Norm[0] = GEOMatrixVector (Ptr->CTF, N);
	    Norm[1] = GEOMatrixVector (CS,       N);
	    Pos[0]  = GEOMatrixVector (Ptr->CTF, P);
	    Pos[1]  = GEOMatrixVector (CS,       P);
	    if (VEC_SAME(Norm[0], Norm[1]) && VEC_SAME(Pos[0], Pos[1]))
	      found = TRUE;
	  }
	  break;

        case PART_FIXED :
          #if 0
	    if (GEOMatrixCompare(Ptr->CTF, CS))
	      found = TRUE;
          #endif
	  break;

        default :
	  ERROR_RETURN1 ("Invalid Feature-Type: %d", Type, NULL)
      }

    if (! found)
      Ptr = Ptr->Next;
  }

  if (Ptr == NULL)                      /* feature not yet inserted       */
  {
    Feature = PARTNewFeature ();
    Feature->Type       = Type;
    Feature->ID         = 1;
    Feature->Object     = Object;
    Feature->Polygon[0] = Polygon1;
    Feature->Polygon[1] = Polygon2;
    Feature->Polygon[2] = Polygon3;
    Feature->CTF        = CS;

    if ((Ptr = Part->FeatureList) == NULL)
      Part->FeatureList = Feature;
    else
    {
      while (Ptr->Next != NULL)
      {
	ID[Ptr->Type-1] = Ptr->ID;
	Ptr = Ptr->Next;
      }
      ID[Ptr->Type-1] = Ptr->ID;
      Feature->ID = ID[Type-1]+1;
      Ptr->Next = Feature;
    }
  }
  else
  {                                     /* feature already in list        */
    PRO0 ("Feature already exists")
    Feature = Ptr;
  }

  (Feature->used)++;

  PRO1 ("ID: %d\n", Feature->ID)

  PRO_END
  return Feature;
}



