#ifndef _LIBSIM3D_H_
#define _LIBSIM3D_H_

#include <iostream>

#include <debug.h>
#include <libgeo.h>


#ifndef W
#define W 3
#endif

#ifndef RX
#define RX 3
#endif

#ifndef RY
#define RY 4
#endif

#ifndef RZ
#define RZ 5
#endif

#define VECTOR_EPSILON 0.000000005
#define SIM_ACCURACY   0.001


#define FZero                       (FRAME(0.0))
#define FIdentity                   (FRAME())
#define FDiagonal(Value)            (FRAME(Value));
#define FRotX(Angle)                (FRAME().RotX(Angle))
#define FRotY(Angle)                (FRAME().RotY(Angle))
#define FRotZ(Angle)                (FRAME().RotZ(Angle))
#define FTrans(dx,dy,dz)            (FRAME().Trans(HVECTOR(dx, dy,   dz)))
#define FTransX(dx)                 (FRAME().Trans(HVECTOR(dx, 0.0, 0.0)))
#define FTransY(dy)                 (FRAME().Trans(HVECTOR(0.0, dy, 0.0)))
#define FTransZ(dz)                 (FRAME().Trans(HVECTOR(0.0, 0.0, dz)))
#define FTransXYZ(dx,dy,dz)         (FRAME().Trans(HVECTOR(dx, dy, dz)))
#define FRotRPY(rx,ry,rz)           (FRAME().RotRPY(rx, ry, rz)) 
#define FRotEuler(rz1,rx,rz2)       (FRAME().RotEuler(rz1, rx, rz2))
#define FRotVectr(Vector,Angle)     (FRAME().RotVector(Vector, Angle))


class VECTOR
{
  public:
    // Constructor / Destructor:
	     VECTOR    (int Dim, ...);
	     VECTOR    (const VECTOR &Vector);
    virtual ~VECTOR    ();

    // Operators: 
             VECTOR   &operator  =   (const VECTOR &Vec);

    virtual  bool      operator  ==  (const VECTOR &Vec) const;
             bool      operator  !=  (const VECTOR &Vec) const;
	     bool      operator  <=  (const VECTOR &Vec) const;

    virtual  double   &operator  []  (      int i);
    virtual  double    operator  []  (      int i) const;

             VECTOR    operator  +   (const VECTOR &Vec) const;
             VECTOR    operator  -   (const VECTOR &Vec) const;
             VECTOR    operator  ~   () const;

             VECTOR   &operator  +=  (const VECTOR &Vec);
             VECTOR   &operator  -=  (const VECTOR &Vec);

	     VECTOR    operator  +   () const;
	     VECTOR    operator  -   () const;
             double    operator  *   (const VECTOR &Vec) const;
	     double    operator  %   (const VECTOR &Vec) const;

    // Methods:
             void      Init          (double Value = 0.0);
             int       GetDim        () const;

             double    Len           () const;
             VECTOR   &Normalize     ();

             void      SetEpsilon    (double Epsilon);
             double    GetEpsilon    () const;

             void      Write         (char *Title = NULL);

    // Friends:
    friend   VECTOR    operator  *   (      double  s,   const VECTOR &Vec);
    friend   VECTOR   &operator  *=  (      VECTOR &Vec,       double  s);
    friend   VECTOR    operator  *   (const VECTOR &Vec,       double  s);
    friend   VECTOR   &operator  /=  (      VECTOR &Vec,       double  s);
    friend   VECTOR    operator  /   (const VECTOR &Vec,       double  s);

  protected :
    // Attributes:
    int     Dim;
    double *Val;

    static double compare_epsilon;          // epsilons for comparison
};

class HVECTOR : public VECTOR
{
  public :
    // Constructor:
            HVECTOR  (double x = 0.0, double y = 0.0, 
		      double z = 0.0, double w = 0.0);
            HVECTOR  (const HVECTOR &Vector);

            HVECTOR  &operator  =   (const HVECTOR &Vector);

    // Operations:
    virtual bool      operator  ==  (const HVECTOR &Vector) const;

    virtual double   &operator  []  (      int i);
    virtual double    operator  []  (      int i) const;

            HVECTOR   operator  +   (const HVECTOR &Vec) const;
            HVECTOR   operator  -   (const HVECTOR &Vec) const;
            HVECTOR   operator  ^   (const HVECTOR &Vec) const;
            HVECTOR   operator  ~   () const;

            HVECTOR  &operator  +=  (const HVECTOR &Vec);
            HVECTOR  &operator  -=  (const HVECTOR &Vec);
            HVECTOR  &operator  ^=  (const HVECTOR &Vec);
            HVECTOR  &Normalize     ();

            HVECTOR   operator  +   () const;
            HVECTOR   operator  -   () const;

    // Friends:
    friend  HVECTOR   operator  *   (      double   s,   const HVECTOR &Vec);
    friend  HVECTOR   operator  /   (const HVECTOR &Vec,       double   s);

    friend  HVECTOR  &operator  *=  (      HVECTOR &Vec,       double   s);
    friend  HVECTOR  &operator  /=  (      HVECTOR &Vec,       double   s);

  private :
    // Attributes:
    double w;
};


class RANGE
{
  public :
    enum INDEX 
    {
      MINIMUM = 0,
      VALUE,
      MAXIMUM
    };

    // Exceptions:
    class RANGE_EXCEPTION
    {
      public :
        RANGE_EXCEPTION (double Min, double Max, double Val)
        {
	  PRO3 ("Range Error: %f..%f: %f\n", Min, Max, Val)
	}
    };

    // Enumerations:
    enum RANGE_TYPE
    {
      ROTATORIC = 0,
      TRANSLATORIC
    };

    // Constructor:
                RANGE         (double Val = 0.0, 
			       double Min = 1.0, double Max = -1.0,
	                       RANGE_TYPE JointType = ROTATORIC, 
			       double Precision = SIM_ACCURACY);

    // Operations:
    RANGE      &operator  =   (const RANGE &Range);
    double     &operator  []  (int Index);
    double      operator  []  (int Index) const;

    // Methods:
    RANGE_TYPE  GetType       () const;
    double      GetValue      () const;
    double     &GetValue      ();
    double      GetMin        () const;
    double      GetMax        () const;
    double      GetAccuracy   () const;

    void        SetType       (RANGE_TYPE Type = ROTATORIC);
    void        SetValue      (double Value = 0.0);
    void        SetRange      (double Min = 1.0, double Max = -1.0);
    void        SetAccuracy   (double Accuracy = SIM_ACCURACY);
 
    bool        Check         () const;
    bool        Check         (double Value) const;

  private :
    // Attributes:
    RANGE_TYPE  Type;
    double      Value;
    double      Min, Max;
    double      Accuracy;
};


class JOINTS
{
  public :
    enum JOINTS_MEMBER 
    {
      WAIST = 0,
      SHOULDER,
      ELBOW,
      WRISTROT,
      WRISTBEND,
      HANDROT
    };       

    // Constructor/Destructor:
           JOINTS (const JOINTS &J);
           JOINTS (int Number,       double  Value);
           JOINTS (int Number, const double *J);
           JOINTS (double J0 = 0.0, double J1 = 0.0, double J2 = 0.0, 
		   double J3 = 0.0, double J4 = 0.0, double J5 = 0.0);
          ~JOINTS ();

    // Operations:
           double    &operator []  (      int Index);
           double     operator []  (      int Index) const;

           JOINTS    &operator =   (const JOINTS &J);
           JOINTS     operator +   (const JOINTS &J) const;
           JOINTS     operator -   (const JOINTS &J) const;
           JOINTS    &operator +=  (const JOINTS &J);
           JOINTS    &operator -=  (const JOINTS &J);

           bool       operator ==  (const JOINTS &J) const;
           bool       operator !=  (const JOINTS &J) const;
           bool       operator <   (const JOINTS &J) const;
           bool       operator <=  (const JOINTS &J) const;
           bool       operator >   (const JOINTS &J) const;
           bool       operator >=  (const JOINTS &J) const;

    // Methods:
           JOINTS    &Round        (double Value = Accuracy);
           JOINTS     Round        (double Value = Accuracy) const;

           JOINTS     MakeRad      () const;
           JOINTS    &Rad          (JOINTS &J) const;
           JOINTS    &Rad          ();
           JOINTS     MakeDeg      () const;
           JOINTS    &Deg          (JOINTS &J) const;
           JOINTS    &Deg          ();

           int        GetNumber    () const;
           void       SetAccuracy  (double Accuracy);
           double     GetAccuracy  () const;
           void       Write        (FILE *fp, const char *Title = NULL) const; 

    // Friends:
    friend JOINTS    &operator *=  (      JOINTS  &J,      double   s);
    friend JOINTS     operator *   (      double  s, const JOINTS  &J);
    friend JOINTS     operator *   (const JOINTS  &J,      double   s);

  private:
    // Attributes:
    int    No;
    RANGE *Range;

    static double Accuracy;
};


#if 0
class MATRIX 
{
  public:         
    // Constructor/Destructor:
             MATRIX    (double Value = 1.0);
             MATRIX    (const HVECTOR &n, const HVECTOR &o,
		        const HVECTOR &a, const HVECTOR &p);
             MATRIX    (double n1, double n2, double n3,
		        double o1, double o2, double o3,
		        double a1, double a2, double a3,
		        double p1, double p2, double p3);
          // MATRIX    (int rows, int cols, ...);
    virtual ~MATRIX    ();

    // Methods:
    virtual  int       operator ==   (const MATRIX &Matr) const;
    virtual  MATRIX   &operator +=   (const MATRIX &Matr);
    virtual  MATRIX   &operator -=   (const MATRIX &Matr);
    virtual  MATRIX    operator *    (const MATRIX &Matr) const;

    virtual  int       operator !=   (const MATRIX &Matr) const;
    virtual  MATRIX    operator +    (const MATRIX &Matr) const;
    virtual  MATRIX    operator -    (const MATRIX &Matr) const;
    virtual  MATRIX   &operator *=   (const MATRIX &Matr);

             MATRIX    operator +    () const;
             MATRIX    operator -    () const;
    virtual  HVECTOR   operator *    (const HVECTOR &Vec) const;
    virtual  HVECTOR  &operator []   (int i);
    virtual  HVECTOR   operator []   (int i) const;

    virtual  int       operator <=   (const MATRIX &Matr) const;

    // Methods:
             void      Init          (double Value);
             void      Write         (FILE *fp,
				      const char *Title = NULL, ...) const;

    // Friends:
    friend  MATRIX     operator *    (double s, const MATRIX &Matr);

  protected:
    int       Rows, Cols;
    VECTOR    *Vec;

#if 0
    //  private :
    HVECTOR   normal;
    HVECTOR   open;
    HVECTOR   approach;
    HVECTOR   position;
#endif
};
#endif

class FRAME
{
  public :
    // Constructor/Destrurctor:
             FRAME    (double Value = 1.0);
             FRAME    (const HVECTOR &n, const HVECTOR &o,
		       const HVECTOR &a, const HVECTOR &p);
             FRAME    (double n1, double n2, double n3,
		       double o1, double o2, double o3,
		       double a1, double a2, double a3,
		       double p1, double p2, double p3);
             FRAME    (const FRAME &Frame);
    virtual ~FRAME    ();

    // Opeations:
             FRAME    operator ~     () const;		     // OrthoNorm
             FRAME    operator !     () const;                // calc inversion
             FRAME    operator +     () const;                 // unary + 
             FRAME    operator -     () const;                 // unary -

    virtual  FRAME    operator *     (const FRAME &Frame) const;
    virtual  FRAME    operator +     (const FRAME &Frame) const;
    virtual  FRAME    operator -     (const FRAME &Frame) const;
    virtual  bool     operator ==    (const FRAME &Frame) const;  
    virtual  bool     operator !=    (const FRAME &Frame) const;
    virtual  FRAME   &operator +=    (const FRAME &Frame);
    virtual  FRAME   &operator -=    (const FRAME &Frame);
    virtual  FRAME&   operator *=    (const FRAME &Frame);

    virtual  HVECTOR  operator *     (const HVECTOR &Vector) const;
    virtual  HVECTOR &operator []    (int i); 
    virtual  HVECTOR  operator []    (int i) const;
    virtual  bool     operator <=    (const FRAME &Frame) const;
  
    // Methods:
             void     Init           (double Value);
             bool     IsOrthoNorm    () const;

             FRAME   &SetOrthoNorm   ();
             FRAME   &SetInv         ();
             FRAME   &SetZero        ();
             FRAME   &SetIdentity    ();
             FRAME   &SetPos         (double x, double y, double z);
             FRAME   &SetPos         (const HVECTOR &Vector);
             FRAME   &SetRotX        (double angle);
             FRAME   &SetRotY        (double angle);
             FRAME   &SetRotZ        (double angle);
             FRAME   &SetRotVector   (const HVECTOR &Vector,    double Angle);
             FRAME   &SetRotRPY      (double phi, double theta, double psi);
             FRAME   &SetRotEuler    (double phi, double theta, double psi);   
             FRAME   &SetDHP         (double  theta, double  d, 
				      double  a,     double  alpha);

             FRAME    Zero           () const;
             FRAME    Identity       () const;
             FRAME    Trans          (double x, double y, double  z) const;
             FRAME    Trans          (const HVECTOR &Vector) const;
             FRAME    RotX           (double angle) const;
             FRAME    RotY           (double angle) const;
             FRAME    RotZ           (double angle) const;
             FRAME    RotVector      (const HVECTOR &Vector, 
				            double   Angle) const;
             HVECTOR  GetRPY         ();
             FRAME    RotRPY         (double Roll, 
				      double Pitch, 
				      double Yaw) const;
             FRAME    RotEuler       (double phi, 
				      double theta,
				      double psi) const;
             void     GetDHP         (double *Theta, double *d, 
				      double *a,     double *Alpha);
             void     CalcRotRPY     (double &Roll, double &Pitch, 
				      double &Yaw) const; 
             void     CalcRotEuler   (double &phi, double &theta, 
				      double &psi) const;
             void     CalcRotVector  (HVECTOR &Vec, double &Angle) const;

             void     Write          (FILE *fp, 
				      const char *Title = NULL, ...) const;

  private :
    // Attributes:
    HVECTOR  Vec[4];
};


class DHP
{
  public :
    // Enumerations:
    enum JOINT_TYPE
    {
      ROT = 0,
      TRANS
    };

    // Constructor:
    DHP  (double d = 0.0, double Theta = 0.0, 
          double a = 0.0, double Alpha = 0.0, JOINT_TYPE Type = ROT);
    DHP  (FRAME An);

    // Methods:
    JOINT_TYPE GetType     () const;      // Get joint tpye
    double     GetD        () const;      // Get translation along  z
    double     GetTheta    () const;      // Get rotation    around z
    double     GetA        () const;      // Get translation along  x
    double     GetAlpha    () const;      // Get rotation    around x

    FRAME      GetAMatrix  () const;      // Compute A-matrix

  private :
    // Attributes:
    JOINT_TYPE Type;                      // joint type
    double     d, Theta, a, Alpha;        // Denavit-Hartenberg-Parameter
};


class INTERFACE
{
  public :
             INTERFACE  ();
    virtual ~INTERFACE  ();

    // virtual functions:
    virtual char   *SendCmd           (const char *Cmd, ...);

    virtual double  GetSpeed          () const; 
    virtual int     GetHand           () const;

    virtual bool    SetSpeed          (double Speed);
    virtual bool    OpenHand          () const;
    virtual bool    CloseHand         () const;

    virtual bool    Place             (const JOINTS &Joint) const;

    virtual bool    Break             () const;
    virtual JOINTS  GetRobotPos       () const;
    virtual FRAME   GetPos            () const;

    virtual FRAME   Teach             (const char *Msg /* = NULL */, ...);
};


class SOCKET : public INTERFACE
{
  public :
             SOCKET  (const char *HostName = NULL, int PortNo = 2030);
    virtual ~SOCKET  ();

    // virtual functions:
    virtual  char   *SendCmd           (const char *Cmd, ...);

    virtual  double  GetSpeed          () const;
    virtual  int     GetHand           () const;

    virtual  bool    SetSpeed          (double Speed) const;
    virtual  bool    OpenHand          () const;
    virtual  bool    CloseHand         () const;

    virtual  bool    Place             (const JOINTS &Joint) const;

    virtual  bool    Break             () const;
    virtual  JOINTS  GetRobotPos       () const;
    virtual  FRAME   GetPos            () const;

    virtual  FRAME   Teach             (const char *Msg /* = NULL */, ...);

  private :
    int SocketID;
};


class MONITOR
{
  public :
    // Enumerations:
    enum MONITOR_STATUS
    {
      ABORT = -1,
      IGNORE,
      MODIFY
    };

    // Type definitions:
    // This event function is called with the *next* frame computed by 
    // the motion pipeline of the trajectory. On guided monitor, Delta 
    // returns the relative changes to this frame respecting actual 
    // circumstandes. When no changes are to be done, a zero-frame or 
    // -joint must be returned. The event routine must take care that the 
    // desired changes are allowed.  
    typedef bool (* GuardFrameMonitor)(const FRAME  &Frame);
    typedef bool (* GuardJointMonitor)(const JOINTS &Joint);

    typedef int  (* GuideFrameMonitor)(const FRAME  &Frame, FRAME  &Delta);
    typedef int  (* GuideJointMonitor)(const JOINTS &Joint, JOINTS &Delta);

    // Constructor/Destructor:
             MONITOR ();
    virtual ~MONITOR ();

    // Friends:
    friend   bool    LightBarrierGuardMonitor  (const FRAME &Frame);
    friend   bool    NearLeftGuardMonitor      (const FRAME &Frame);
    friend   bool    NearRightGuardMonitor     (const FRAME &Frame);
};


// Binary values for arm configuration
#define ARM_RIGHT     (1<<0)
#define ARM_LEFT           0
#define ELBOW_ABOVE   (1<<1) 
#define ELBOW_BELOW        0
#define WRIST_DOWN    (1<<2)
#define WRIST_UP           0
#define FLIP          (1<<3)
#define NO_FLIP            0

#define IsArmRight(Config)    (  Config & ARM_RIGHT)
#define IsArmLeft(Config)     (!(Config & ARM_RIGHT))
#define IsElbowAbove(Config)  (  Config & ELBOW_ABOVE)
#define IsElbowBelow(Config)  (!(Config & ELBOW_ABOVE))
#define IsWristDown(Config)   (  Config & WRIST_DOWN)
#define IsWristUp(Config)     (!(Config & WRIST_DOWN))


class KINEMATIC  
{
  public:
    //Exceptions
    // ErrorJoint in Ausgeloeste OutOfXXX enthaelf ein binaer 
    // Muster mit der Fehlerhaften Joints
    // zB 000101=0*32+0*16+0*8+1*4+0*2+1*1=Joint[3] & Joint[0]
    // dafr geeignete defines ErJoint0=1, ErJoint1=2,...
    // Hoeher bits zum Ausdrucken von zu weit bzw. zu nah Situation bei 
    // OutOfWorkspace
    class OutOfWorkspace                 // Frame/Joint/Inc
    {
      public : 
        OutOfWorkspace (int JointNo)
	{
	  PRO1 ("OutOfWorkSpace: %d\n", JointNo)
	  ErrorJoint = JointNo;
	};

        int ErrorJoint;
    };    

    // Die Abweichung des angegebenen letzten Winkelsatzes
    // ist bedenklich gro (grer als EpsilonJoint). Dies
    // kann auf ein Umdrehen eines Gelenkes aufgrund
    // des Erreichen des Wertebereiches zurckzufhren sein oder
    // auf einen Fehler innerhalb der Berechnung der inversen
    // Kinematik. Vorraussetzung ist natrlich, da der angegebene
    // Lastjointsatz in der Ne des zu berechnenden Punktes liegt.
    class OutOfLimit
    {
      public:
        OutOfLimit (int JointNo)
	{
	  PRO1 ("OutOfLimit: %d\n", JointNo)
	  ErrorJoint = JointNo;
	};

        int ErrorJoint;
    };		   
	
  public:
    // Constructor:
             KINEMATIC ();
    virtual ~KINEMATIC ();
   
             void        SetJointRange   (const JOINTS &Min, 
					  const JOINTS &Max);

             void        SetEpsilonJoint (double Epsilon);
             void        SetBase         (const FRAME &BaseFrame);
             void        SetOffset       (const FRAME &TCPOffs);

    virtual  JOINTS InvKin  (const FRAME  &Frame,
			           int     Configuration,
			     const JOINTS &LastJoint,
			           bool    CheckOOW = TRUE,
			           bool    ValidLastJoint = TRUE) = 0;
    virtual  JOINTS InvKin  (const FRAME  &Frame,
			           int     Configuration,  
			           bool    CheckOOW = TRUE) = 0;
    virtual  int    InvKin  (const FRAME  &Frame,
			           JOINTS &Joint,
			     const JOINTS &LastJoint,
			           bool    CheckOOW = TRUE) = 0;
    virtual  JOINTS InvKin  (const FRAME  &Frame,
			     const JOINTS &LastJoint,
			           bool    CheckOOW) = 0;
    virtual  int    InvKin  (const FRAME  &Frame, 
			           JOINTS &Joint) = 0;
    virtual  JOINTS InvKin  (const FRAME  &Frame) = 0;

    virtual  JOINTS  CheckFrame  (const FRAME &Frame) = 0;

    virtual  FRAME   FwdKin      (const JOINTS &Joint,
				        bool    Check = TRUE) = 0;

    virtual  int     CalcConfiguration (const JOINTS &Joint, 
					const FRAME  &Frame, 
					      bool    Check = TRUE) = 0;

    virtual  int     CalcConfiguration (const JOINTS &Joint,
					      bool    Check = TRUE) = 0;

    virtual  void    CheckJoint        (const JOINTS &Joint) = 0;
    virtual  void    CheckOutOfLimit   (const JOINTS &Joint,
					const JOINTS &LastJoint) = 0;
            JOINTS   GetEpsilonJoint   ();
            double   GetEpsilonJoint   (int JointNo);
            FRAME    GetBase           ();
            FRAME    GetOffset         ();
            JOINTS   GetHome           ();

  protected:
    // epsilonjoint: Wird benutzt, um zu prfen, ob bei der inversen 
    // Kinematik der berechnete Gelenkwinkelsatz zu stark von den alten 
    // Gelenkwinkeln abweicht, also ob 
    // fabs(JointLast[i]-Joint[i]) > epsilonjoint[i] ist.
    JOINTS  EpsilonJoint;
    JOINTS  JointMin, JointMax;
    JOINTS  Home;
    FRAME   Base;       // Offset for Joint0
    FRAME   Offset;     // TCP-Offset    
};


class CURVE
{
  public :
                    CURVE ();

    virtual double  GetValue (double Time);
};


class SPLINE : public CURVE
{
  public :
    // Constructor / destructor:
                     SPLINE   (int Number, POINT2D *Points);
    virtual         ~SPLINE   ();

    virtual double   GetValue (double Time);

  private :
    int       Number;
    POINT2D  *Points;
    double   *d, *u, *w, *k;
};


#if 0
class PATH
{
}


class PROFILE   // Speed-provfile
{
  public :
    PROFILE (CURVE Accel, CURVE Move, CURVE Break);


  double GetIntegral ();   // entspricht Zeit
  
  private :
    // motion-phases (normed to 1.0):
    CURVE  Accel;
    CURVE  Move;
    CURVE  Break;
  /*
    Possible profiles (examples):
          linear           Curve (Spline)      discrete without decelleration
                                                (for following motion)
         ********               ******
        *        *           ***      *          **********************
       *          *         *          *         *
      *            *      **            ***      *
     *              *  ***                 ***   * 

     delayed motion
            ********
            *
           *
           *
     ******
    
  */       
}
#endif

class MOTION
{
  public :
           MOTION (const FRAME   &Start,   const FRAME  &End,
		   /* const PROFILE &Profile, */ double  Duration);
           MOTION (const MOTION &Motion);
  //       MOTION (int Number; const FRAME *Frames, 
  //               const PROFILE &Profile, double Duration);
  //  TODO: Spezielle Konstruktoren fr Trans & Rot
  //        Interpolationsart (cartesian, joint)
          ~MOTION ();

    FRAME  GetStep     (double Time) const;
    double GetDuration () const;

  private :
    int      Number;
    FRAME   *Frames;
    //    PROFILE  Profile;
    double   Duration;

    HVECTOR  Pos;
    HVECTOR  Rot;
    double   Angle;
};


class TRAJECTORY
{
  public:
#if 0
      class IncorrectStartPos
      {
        public:
          JOINTS RobotPos, IncorrectPos;

	  IncorrectStartPos (JOINTS p_RobotPos, JOINTS p_IncorrectPos) :
	    RobotPos     (p_RobotPos), 
	    IncorrectPos (p_IncorrectPos)
	  {
	  };
      };
#endif

    enum INTERPOLATION               // Bewegungsarten
    { 
      FRAME_MOTION,
      JOINT_MOTION
    };

    enum MOTION_RESULT
    {
      FAILED = -1,
      STOPPED,
      DONE
    };

             TRAJECTORY            (KINEMATIC *pKinematic = NULL, 
				    INTERFACE *pInterface = NULL);
    virtual ~TRAJECTORY            ();

    void  SetKinematic             (KINEMATIC *pKinematic);
    void  SetInterface             (INTERFACE *pInterface);

    virtual int Move (const FRAME  &Frame, INTERPOLATION  Motion = FRAME_MOTION);
    virtual int Move (const JOINTS &Joint, INTERPOLATION  Motion = JOINT_MOTION);

    int Move (const FRAME  &Frame,   MONITOR::GuardFrameMonitor Monitor,
	      double Priority = 1.0, INTERPOLATION Motion = FRAME_MOTION);
    int Move (const JOINTS &Joint,   MONITOR::GuardJointMonitor Monitor,
	      double Priority = 1.0, INTERPOLATION Motion = JOINT_MOTION);
    int Move (const FRAME  &Frame,   MONITOR::GuideFrameMonitor Monitor, 
	      double Priority = 1.0, INTERPOLATION Motion = FRAME_MOTION);
    int Move (const JOINTS &Joint,   MONITOR::GuideJointMonitor Monitor, 
	      double Priority = 1.0, INTERPOLATION Motion = JOINT_MOTION);

    int Approach    (double Dist, double Speed = 0.0);
    int Depart      (double Dist, double Speed = 0.0);

    void Stop       ();
    void Contnue    ();

    void SetMonitor (MONITOR::GuardFrameMonitor Monitor = NULL,
		     double Priority = 1.0);
    void SetMonitor (MONITOR::GuardJointMonitor Monitor = NULL,
		     double Priority = 1.0);
    void SetMonitor (MONITOR::GuideFrameMonitor Monitor = NULL,
		     double Priority = 1.0);
    void SetMonitor (MONITOR::GuideJointMonitor Monitor = NULL,
		     double Priority = 1.0);

    // Set/Get-Joint/CartesianSpeed setzt/list die Sollgeschwindigkeiten bei 
    // gelnek-orienteritem Fahren (JointSpeed[6] in Grad/s) oder bei 
    // kartesichem Fahren (CartSpeed in mm/s). 
    void    SetClock               (double Clock);
    void    SetInterpTotal         (int    p_N);
    void    SetInterpPerMeter      (int    p_N);
    void    SetInterpPerDegree     (int    p_N);

    JOINTS  GetStartJoint          () const;
    void    FetchRobotPos          ();
    double  GetClock               () const;
    int     GetInterpTotal         () const;
    int     GetInterpPerMeter      () const;
    int     GetInterpPerDegree     () const;

    int     GetSteps               () const;	
    int     GetNCurrent            () const;
    JOINTS  GetLastJoint           () const;
    JOINTS  GetNextJoint           () const;

    JOINTS  NextJoint;
    FRAME   NextFrame;

  private :
    double CalcLambda (int n, int Steps, int NRaise, int NFall, double DeltaStep);
    bool FindAlternateConfig      (const FRAME &Frame);
    int  CalcNextJoint            ();

    KINEMATIC        *Kinematic;
    INTERFACE        *Interface;

    double Clock;    //Reglertakt in ms

    int Steps;
    int NCurrent;
    int InterpTotal;
    int InterpPerMeter;
    int InterpPerDegree;

    JOINTS         LastJoint;

    INTERPOLATION  Interp; 

    JOINTS         StartJoint, DeltaJoint; 
    HVECTOR        StartPos,   DeltaPos; 

    FRAME          GoalRotation;
    HVECTOR        DeltaRot_NVec;
    double         DeltaRot_Theta;

    int            Configuration;
	
    MONITOR::GuardFrameMonitor   GuardFrameMonitor;
    MONITOR::GuardJointMonitor   GuardJointMonitor;
    MONITOR::GuideFrameMonitor   GuideFrameMonitor;
    MONITOR::GuideJointMonitor   GuideJointMonitor;

    double GuardFramePriority;
    double GuardJointPriority;
    double GuideFramePriority;
    double GuideJointPriority;

    int MoveTo (const FRAME  &frame, INTERPOLATION Motion = FRAME_MOTION);
    int MoveTo (const JOINTS &joint, INTERPOLATION Motion = JOINT_MOTION);
};


class ROBOT : public virtual KINEMATIC, public TRAJECTORY
{
  public :
    // Constructor / Destructor:
         ROBOT  (INTERFACE *Interf = NULL);
	~ROBOT  () ;

    // TODO: Als Namespace ?
    bool    SetSpeed (double Speed);
    double  GetSpeed ();

    bool    OpenHand ();
    bool    CloseHand ();
    int     GetHand ();

    bool    Place (const JOINTS &Joint);
    bool    Break ();
    JOINTS  GetRobotPos ();
    FRAME   GetPos ();
    FRAME   Teach (char *Msg /* = NULL */, ...);

  protected :
    INTERFACE *Interface;
    double     Speed, OldSpeed;
    // JOINTS Joint;
    // FRAME  TCP;
};


class GENERIC : public ROBOT
{ 
  public:
    enum SENSOR_ID  /* -> SENSOR */
    {
      NO_SENSOR = -1,
      LIGHT_BARRIER,
      HAND_LEFT,
      HAND_RIGHT,
      SONAR,
      TRIANGULAR
    };

    // Constructor / Destructor:
             GENERIC  (INTERFACE *Interface);
            ~GENERIC  ();

    // Methods:
    JOINTS  InvKin            (const FRAME  &Frame,	      // Ursp frame
			             int     Configuration,  
			       const JOINTS &LastJoint,       // Vergl.-Joint
			             bool    CheckOOW = TRUE,
			             bool    ValidLastJoint = true);
    JOINTS  InvKin            (const FRAME  &Frame,	      // Ursp frame
			             int     Configuration,  
			             bool    CheckOOW = TRUE);
    int	    InvKin            (const FRAME  &Frame,	      // Ursp frame
			             JOINTS &Joint,	      // Ziel joint
			       const JOINTS &LastJoint,	      // Vergl joint. 
			             bool    CheckOOW = TRUE);
    JOINTS  InvKin            (const FRAME  &frame,	      // orig frame
			       const JOINTS &LastJoint,	      // Vergl joint. 
	          	             bool    CheckOOW);
    int     InvKin            (const FRAME  &Frame, 
		                     JOINTS &Joint); 
    JOINTS  InvKin            (const FRAME  &Frame);
    FRAME   FwdKin            (const JOINTS &Joint,
			             bool    Check = TRUE);

    JOINTS  CheckFrame        (const FRAME  &Frame);

    int     CalcConfiguration (const JOINTS &Joint, 
			       const FRAME  &Frame, 
			             bool    Check = TRUE);
    int     CalcConfiguration (const JOINTS &Joint,
			             bool    Check = TRUE);

    void    CheckJoint        (const JOINTS &Joint);
    void    CheckOutOfLimit   (const JOINTS &Joint,
			       const JOINTS &LastJoint);

    /*
    bool    SetSpeed          (double Speed);
    double  GetSpeed          ();
    bool    OpenHand          ();
    bool    CloseHand         ();
    int     GetHand           ();
    bool    Place             (const JOINTS &Joint);

    bool    Break             ();
    JOINTS  GetRobotPos       ();
    FRAME   GetPos            ();
    */

    double  GetSensor         (SENSOR_ID Sensor);

    double  GetEpsilon        ();
    void    SetEpsilon        (double Epsilon);
    double  GetEpsilonLength  ();
    void    SetEpsilonLength  (double Epsilon);
    void    SetDHP            (JOINTS a, JOINTS d, JOINTS Alpha, JOINTS Theta);
    void    SetDHP            (double D0, 
                               double D1, double A1, 
                               double D3, 
			       double D5);
    void    SetHome           (JOINTS Home);
    void    SetJointMin       (const JOINTS &Min);
    void    SetJointMax       (const JOINTS &Max);
    const JOINTS &GetJointMin ();
    const JOINTS &GetJointMax ();

  protected :
    // epsilon: Wird bentigt, um zu prfen, ob ein Ergebniswert 0 ist.
    // Anwendung beim Test, ob Stellung Gelenk 4/5/6 degeneriert ist.
    // epsilon ist kleiner als der minimal erreichbare Winkelabstand
    // zu whlen.
    static double  Epsilon, EpsilonLength;
           JOINTS  DHP_D, DHP_A;
           JOINTS  DHP_Alpha, DHP_Theta;     // Denavit-Hartenberg-Parameter
	   JOINTS  Home;                     // Home-Offset of Joints
};


class RX60 : public GENERIC
{
  /*
    No  Theta   d       Alpha    a             Range
     0   var    0(341)    90      0      -160.0 .. 160.0
     1   var   49          0    290      -217.5 ..  37.5 
     2   var    0        -90      0       -39.5 .. 224.5
     3   var  310          0      0      -270.0 .. 270.0
     4   var    0         90      0      -109.5 .. 120.5
     5   var   65          0      0      -270.0 .. 270.0
  */
  public :
    RX60 (INTERFACE *Interface = NULL) :
      GENERIC (Interface)
    {
      SetKinematic (this);
      SetInterface (Interface);   // Set Trajectory Interface

      SetDHP (341.0, 49.0, 290.0, 310.0, 65.0);
      SetHome (JOINTS(6, 0.0));

      SetJointRange (JOINTS(-160.0, -217.5, -39.5, -270.0, -109.5, -270.0),
                     JOINTS( 160.0,   37.5, 224.5,  270.0,  120.5,  270.0));

      //      SetBase (FTransZ(341.0));       // Robot base offset

      FetchRobotPos ();  
    };
};


class RX90 : public GENERIC
{
  /*
    No  Theta  Alpha    a     d          Range
     0   var     90      0    0 (420)  -160.0 .. 160.0
     1   var      0    450    0        -217.5 ..  37.5 
     2   var    -90      0    0         -39.5 .. 224.5
     3   var      0      0  365        -270.0 .. 270.0
     4   var     90      0    0        -109.5 .. 120.5
     5   var      0      0   85        -270.0 .. 270.0
  */
  public:
    RX90 (INTERFACE *Interface = NULL) : 
      GENERIC (Interface)
    {
      SetKinematic (this);
      SetInterface (Interface);

      SetDHP (0.0, 30.0, 471.0, 481.0, 74.0);

      SetJointRange (JOINTS(-160.0, -217.5, -39.5, -270.0, -109.5, -270.0),
                     JOINTS( 160.0,   37.5, 224.5,  270.0,  120.5,  270.0));
      SetBase (FTransZ(423));      // Robot base offset
      FetchRobotPos ();  
    };
};


class R2 : public GENERIC
{
  /*
    No  Theta  Alpha    a     d          Range
     0   var     90      0  520.5 (630)  
     1   var      0    350    0
     2   var    -90      0    0
     3   var      0      0  350
     4   var     90      0    0
     5   var      0      0   70
  */
  public:
    R2 (INTERFACE *Interface = NULL) : 
      GENERIC (Interface)
    {
      SetKinematic (this);
      SetInterface (Interface);

      // SetDHP (420.0, 0.0, 290.0, 310.0, 65.0);

      SetJointRange (JOINTS(-165.0, -125.0, -150.0, -200.0, -115.0, -265.0),
                     JOINTS( 165.0,  125.0,  150.0,  200.0,  115.0,  265.0));
      SetHome (JOINTS(0.0, 90.0, -90.0, 0.0, 0.0, 0.0));

      SetOffset (FTransZ(164.0+466.0));

      FetchRobotPos ();
    };
};


class R15 : public GENERIC
{
  /*
    No  Theta  Alpha    a     d          Range
     0   var     90      0  520.5 (630)  
     1   var      0    350    0
     2   var    -90      0    0
     3   var      0      0  350
     4   var     90      0    0
     5   var      0      0   70
  */
  public:
    R15 (INTERFACE *Interface = NULL) : 
      GENERIC (Interface)
    {
      SetKinematic (this);
      SetInterface (Interface);

      //  SetDHP (420.0, 0.0, 290.0, 310.0, 65.0);

      SetJointRange (JOINTS(-165.0, -125.0, -150.0, -200.0, -115.0, -265.0),
                     JOINTS( 165.0,  125.0,  150.0,  200.0,  115.0,  265.0));
      SetHome (JOINTS(0.0, 90.0, -90.0, 0.0, 0.0, 0.0));

      SetOffset (FTransZ(164.0+466.0));
      FetchRobotPos ();
    };
};


class MITSUBISHI : public ROBOT
{
  
  public:
    // Constructor / Destructor:
             MITSUBISHI  (INTERFACE *Interface);
            ~MITSUBISHI  ();

    // Methods:
    FRAME   FwdKin            (const JOINTS &Joint,
                                     bool    Check = TRUE);

    JOINTS  InvKin            (const FRAME  &Frame);
    JOINTS  InvKin            (const FRAME  &Frame,
                                     int     Configuration,
			       const JOINTS &LastJoint,   
			             bool    CheckOOW = TRUE,
			             bool    ValidLastJoint = TRUE);
    JOINTS  InvKin            (const FRAME  &Frame,
			             int     Configuration,  
			             bool    CheckOOW = TRUE);
    int     InvKin            (const FRAME  &Frame,
			             JOINTS &Joint,
			       const JOINTS &LastJoint,
			             bool    CheckOOW = TRUE);
    JOINTS  InvKin            (const FRAME  &Frame,
			       const JOINTS &LastJoint,
			             bool    CheckOOW);
    int     InvKin            (const FRAME  &Frame, 
			             JOINTS &Joint);

    JOINTS  CheckFrame        (const FRAME  &Frame);

    int     CalcConfiguration (const JOINTS &Joint, 
                               const FRAME  &Frame, 
                                     bool    Check = TRUE);
    int     CalcConfiguration (const JOINTS &Joint,
                                     bool    Check = TRUE);

    /*
    bool    SetSpeed          (double Speed);
    double  GetSpeed          ();
    bool    OpenHand          ();
    bool    CloseHand         ();
    int     GetHand           ();
    bool    Place              (const JOINTS &Joint);

    bool    Break             ();
    JOINTS  GetRobotPos       ();
    FRAME   GetPos            ();
    */

    double  GetEpsilon        ();
    void    SetEpsilon        (const double&);
    double  GetEpsilonLength  ();
    void    SetEpsilonLength  (const double&);
    void    SetDHP            (JOINTS a, JOINTS d, JOINTS Alpha, JOINTS Theta);
    void    SetDHP            (const double &D0, 
                               const double &D1, 
                               const double &A1,
                               const double &D3, 
                               const double &D5);
    void    SetJointMin       (JOINTS &Min);
    void    SetJointMax       (JOINTS &Max);
    const JOINTS  &GetJointMin ();
    const JOINTS  &GetJointMax ();

    void    CheckJoint        (const JOINTS &Joint);
    void    CheckOutOfLimit   (const JOINTS &Joint,
			       const JOINTS &LastJoint);

  protected :
    JOINTS  DHP_D, DHP_A;
    JOINTS  DHP_Alpha, DHP_Theta;     // Denavit-Hartenberg-Parameter
};


class RM501 : public MITSUBISHI
{
  public:
    RM501 (INTERFACE *Interface = NULL) : 
      MITSUBISHI (Interface)
    {
      SetKinematic (this);
      SetInterface (Interface);

      //   SetDHP (0.0, 49.0, 290.0, 310.0, 65.0);

      SetJointRange (JOINTS(-150.0, -30.0, -90.0,   0.0, -194.5),
                     JOINTS( 150.0, 100.0,   0.0, 180.0,  165.5));

      SetBase (FTransZ(250.0));

      FetchRobotPos ();  
    };
};


class HEXA : public ROBOT
{ 
  public:
    // Constructor / Destructor:
            HEXA  (INTERFACE *Interface = NULL);
           ~HEXA  ();

    // Methods:
    JOINTS  GetPassive         (FRAME Frame, JOINTS Joint);
    JOINTS  CheckFrame         (const FRAME  &Frame);

    void    CheckJoint         (const JOINTS &Joint);
    void    CheckOutOfLimit    (const JOINTS &Joint,
				const JOINTS &LastJoint);

    FRAME   FwdKin             (const JOINTS &Joint,
				      bool    Check = TRUE);

    JOINTS  InvKin             (const FRAME  &Frame);	
    JOINTS  InvKin             (const FRAME  &Frame,
				      int     Configuration,
				const JOINTS &LastJoint,   
				      bool    CheckOOW = TRUE,
				      bool    ValidLastJoint = TRUE);
    JOINTS  InvKin             (const FRAME  &Frame,
				      int     Configuration,  
			              bool    CheckOOW = TRUE);
    int     InvKin             (const FRAME  &Frame,
				      JOINTS &Joint,
				const JOINTS &LastJoint,
				      bool    CheckOOW = TRUE);
    JOINTS  InvKin             (const FRAME  &Frame,
				const JOINTS &LastJoint,
				      bool    CheckOOW);
    int     InvKin             (const FRAME  &Frame, 
			              JOINTS &Joint);

    int     CalcConfiguration  (const JOINTS &Joint, 
				const FRAME  &frame, 
				      bool    Check = TRUE);
    int     CalcConfiguration  (const JOINTS &Joint,
				      bool    Check = TRUE);

    bool    Place              (const JOINTS &Joint);
    /*
    bool    Break              ();
    JOINTS  GetRobotPos        ();
    */
          FRAME    GetPos            ();

#if 0
          void     SetEpsilon        (double Epsilon);
          void	   SetEpsilonLength  (double EpsilonLen);
          double   GetEpsilon        ();
          double   GetEpsilonLength  ();
#endif
          void     SetJointMin       (const JOINTS &Min);
          void     SetJointMax       (const JOINTS &Max);
    const JOINTS  &GetJointMin       ();
    const JOINTS  &GetJointMax       ();

  protected :
    // epsilon: Wird bentigt zum Prfen, ob ein Resultatwert (double) null
    // ist. Anwendung beim Test, ob Stellung Gelenk 4/5/6 degeneriert ist.
    // epsilon ist kleiner als der minimal erreichbare Winkelabstand
    // zu whlen.
    static double  epsilon, epsilonlength;
	   double  LenA, LenB;
	   double  OffsA, OffsB;
           JOINTS  Joint;       // Gelenkstellungen
	   FRAME   TCP;
	   FRAME   Motor[6];
	   FRAME   Mount[6];    // Rel. Position der Flansch-Gelenke
    // TODO: geeignete Datenstruktur fr Roboterparameter:
    //       JointLen(A/B), Offset(A/B), Range(Min,Max,Val), Activ/Passiv
    //       Freiheitsgrade (rx,ry,rz), TCP->Flange1..6, Motor-Geometrie, ...

    bool       CalcPassive  (JOINTS Joint, FRAME TCP, 
                             JOINTS &J1, JOINTS &J2, JOINTS &J3,
			     bool   check = TRUE);
};


#endif
