#include <assert.h>
#include <math.h>
#include <stdarg.h>

#include <debug.h>

#include "libsim3d.h"


#define Thres(x) (fabs(x) <= Vec[N].GetEpsilon()? 0.0:x)


#define TEST_FRAME(Frame) \
        { \
          assert (fabs((Frame).Vec[N].Len()-1.0) <= VECTOR_EPSILON); \
          assert (fabs((Frame).Vec[O].Len()-1.0) <= VECTOR_EPSILON); \
          assert (fabs((Frame).Vec[A].Len()-1.0) <= VECTOR_EPSILON); \
          assert ((Frame).Vec[N][W] == 0.0);  \
          assert ((Frame).Vec[O][W] == 0.0);  \
          assert ((Frame).Vec[A][W] == 0.0);  \
          assert ((Frame).Vec[P][W] == 1.0 || \
                  (Frame).Vec[P][W] == 0.0);  \
        }


FRAME::FRAME (double Value /* = 1.0 */)
{
  PRO_METHOD ("FRAME", "")

  Vec[N] = HVECTOR (Value, 0.0,   0.0,   0.0);
  Vec[O] = HVECTOR (0.0,   Value, 0.0,   0.0);
  Vec[A] = HVECTOR (0.0,   0.0,   Value, 0.0);
  Vec[P] = HVECTOR (0.0,   0.0,   0.0,   1.0);

  PRO_END
}


FRAME::FRAME (const HVECTOR &n, const HVECTOR &o,
	      const HVECTOR &a, const HVECTOR &p)
{
  PRO_METHOD ("FRAME", "")

  Vec[N] = n;
  Vec[O] = o;
  Vec[A] = a;
  Vec[P] = p;

  PRO_END
}


FRAME::FRAME (double nx, double ny, double nz,
	      double ox, double oy, double oz,
	      double ax, double ay, double az,
	      double px, double py, double pz)
{
  PRO_METHOD ("FRAME", "")

  Vec[N] = HVECTOR (nx, ny, nz, 0.0);
  Vec[O] = HVECTOR (ox, oy, oz, 0.0);
  Vec[A] = HVECTOR (ax, ay, az, 0.0);
  Vec[P] = HVECTOR (px, py, pz, 1.0);

  PRO_END
}


FRAME::FRAME (const FRAME &Frame)
{
  PRO_METHOD ("FRAME", "")

  Vec[N] = Frame[N];
  Vec[O] = Frame[O];
  Vec[A] = Frame[A];
  Vec[P] = Frame[P];

  PRO_END
}


FRAME::~FRAME ()
{
  PRO_METHOD ("~FRAME", "")
  PRO_END
}


FRAME FRAME::operator ~ () const
{ 
  return FRAME (*this).SetOrthoNorm ();
}


FRAME FRAME::operator ! () const
{
  return FRAME (*this).SetInv ();
}


bool FRAME::operator == (const FRAME &Frame) const
{
  TEST_FRAME (Frame);

  return ((Vec[N] == Frame[N]) &&
	  (Vec[O] == Frame[O]) &&
	  (Vec[A] == Frame[A]) &&
	  (Vec[P] == Frame[P]));
}


bool FRAME::operator != (const FRAME &Frame) const
{
  return ! (*this == Frame);
}


FRAME& FRAME::operator += (const FRAME &Frame)
{
  PRO_METHOD ("FRAME", "")

  TEST_FRAME (Frame);

  Vec[N] += Frame[N];
  Vec[O] += Frame[O];
  Vec[A] += Frame[A];
  Vec[P] += Frame[P];
  Vec[P][W] = 1.0;

  PRO_END
  return *this;
}


FRAME& FRAME::operator -= (const FRAME &Frame)
{
  PRO_METHOD ("FRAME", "")

  TEST_FRAME (Frame);

  Vec[N] -= Frame[N];
  Vec[O] -= Frame[O];
  Vec[A] -= Frame[A];
  Vec[P] -= Frame[P];
  Vec[P][W] = 1.0;

  PRO_END
  return *this;
}


FRAME FRAME::operator * (const FRAME &Frame) const
{
  int   row, col, k;
  FRAME Result (0.0);

  PRO_METHOD ("FRAME", "")

  TEST_FRAME (Frame);

  // rotation
  for (col = N; col <= A; ++col)
    for (row = X; row <= Z; ++row)
      for (k = X; k <= W; k++)
        Result[col][row] += Vec[k][row] * Frame[col][k];

  // translation
  for (row = X; row <= Z; ++row)
    for (k = N; k <= P; k++)
      Result[P][row] += Vec[k][row] * Frame[P][k];
 
  PRO_END
  return Result;
}


FRAME FRAME::operator + (const FRAME &Frame) const
{
  return FRAME (*this) += Frame;
}


FRAME FRAME::operator - (const FRAME &Frame) const
{
  return FRAME (*this) -= Frame;
}


FRAME& FRAME::operator *= (const FRAME &Frame)
{
  return *this = *this * Frame;
}


FRAME FRAME::operator + () const
{
  return FRAME (Vec[N], Vec[O], Vec[A], Vec[P]);
}


FRAME FRAME::operator - () const
{
  return FRAME (-Vec[N], -Vec[O], -Vec[A], -Vec[P]);
}


HVECTOR FRAME::operator * (const HVECTOR &Vector) const
{
  int     row, k;
  HVECTOR Result (0.0, 0.0, 0.0, Vector[W]);

  PRO_METHOD ("FRAME", "")

  for (row = X; row <= Z; ++row)  
    for (k = N; k <= P; ++k)
      Result[row] += Vec[k][row] * Vector[k];
 
  PRO_END
  return Result;
}


HVECTOR &FRAME::operator [] (int i)
{
  assert (i >= N && i <= P);

  return Vec[i];
}


HVECTOR FRAME::operator [] (int i) const
{
  assert (i >= N && i <= P);

  return Vec[i];
}


bool FRAME::operator <= (const FRAME &Frame) const
{
  TEST_FRAME (Frame);
 
  return ((Vec[N] <= Frame[N]) &&
	  (Vec[O] <= Frame[O]) &&
	  (Vec[A] <= Frame[A]) &&
	  (Vec[P] <= Frame[P]));
}


void FRAME::Init (double Value)
{
  Vec[N] = HVECTOR (Value, Value, Value, 0.0);
  Vec[O] = HVECTOR (Value, Value, Value, 0.0);
  Vec[A] = HVECTOR (Value, Value, Value, 0.0);
  Vec[P] = HVECTOR (Value, Value, Value, 1.0);
}


FRAME& FRAME::SetOrthoNorm ()  
{
  PRO_METHOD ("FRAME", "")

  if (! IsOrthoNorm())
  {
    Vec[O].Normalize ();
    Vec[A].Normalize ();
    Vec[N] = ~(Vec[O] ^ Vec[A]);          // Normal-vector is right to o and a
    Vec[O] =   Vec[A] ^ Vec[N];           // Adjust open vector
                                          // Approach vector left unchanged
  }

  PRO_END
  return *this;
}


bool FRAME::IsOrthoNorm () const
{
#if 0
  double eps = Vec[N].GetEpsilon ();
#endif

  return (/* fabs(Vec[N].Len()-1.0) <= eps &&
	  fabs(Vec[O].Len()-1.0) <= eps &&
	  fabs(Vec[A].Len()-1.0) <= eps &&
	  */
	  fabs(Vec[N] * Vec[O]) <= VECTOR_EPSILON &&
	  fabs(Vec[N] * Vec[A]) <= VECTOR_EPSILON &&
	  fabs(Vec[O] * Vec[A]) <= VECTOR_EPSILON);
}


FRAME& FRAME::SetInv ()  // caution! only for orthonomed frames
{                                
  FRAME Old (*this);

  PRO_METHOD ("FRAME", "")

  assert (IsOrthoNorm());

  Vec[P][X] =- (Old[P] * Vec[N]);
  Vec[P][Y] =- (Old[P] * Vec[O]);
  Vec[P][Z] =- (Old[P] * Vec[A]);

  // keep nx
  Vec[N][Y] = Vec[O][X];
  Vec[N][Z] = Vec[A][X];

  Vec[O][X] = Old[N][Y];
  // keep oy
  Vec[O][Z] = Vec[A][Y];

  Vec[A][X] = Old[N][Z];
  Vec[A][Y] = Old[O][Z];
  // keep az

  PRO_END
  return *this;
}


FRAME& FRAME::SetZero ()
{
  return *this = FRAME (0.0);
}


FRAME& FRAME::SetIdentity ()
{
  return *this = FRAME ();
}


FRAME& FRAME::SetPos (double x, double y, double z)
{
#if 0
  Vec[N] = HVECTOR (1.0, 0.0, 0.0, 0.0);
  Vec[O] = HVECTOR (0.0, 1.0, 0.0, 0.0);
  Vec[A] = HVECTOR (0.0, 0.0, 0.0, 0.0);
#endif
  Vec[P] = HVECTOR (  x,   y,   z, 1.0);

  return *this;
}


FRAME& FRAME::SetPos (const HVECTOR &Vector)
{
  SetPos (Vector[X], Vector[Y], Vector[Z]);

  return *this;
}


FRAME& FRAME::SetRotX (double Angle)
{
  double RadAngle = RAD (Angle);

  Vec[N] = HVECTOR (1.0,            0.0,           0.0, 0.0);
  Vec[O] = HVECTOR (0.0,  cos(RadAngle), sin(RadAngle), 0.0);
  Vec[A] = HVECTOR (0.0, -sin(RadAngle), cos(RadAngle), 0.0);
#if 0
  Vec[P] = HVECTOR (0.0,            0.0,           0.0, 1.0);
#endif

  return *this;
}


FRAME& FRAME::SetRotY (double Angle)
{
  double RadAngle = RAD (Angle);

  Vec[N] = HVECTOR (cos(RadAngle), 0.0, -sin(RadAngle), 0.0);
  Vec[O] = HVECTOR (          0.0, 1.0,            0.0, 0.0);
  Vec[A] = HVECTOR (sin(RadAngle), 0.0,  cos(RadAngle), 0.0);
#if 0
  Vec[P] = HVECTOR (          0.0, 0.0,            0.0, 1.0);
#endif

  return *this;
}


FRAME& FRAME::SetRotZ (double Angle)
{
  double RadAngle = RAD (Angle);

  Vec[N] = HVECTOR ( cos(RadAngle), sin(RadAngle), 0.0, 0.0);
  Vec[O] = HVECTOR (-sin(RadAngle), cos(RadAngle), 0.0, 0.0);
  Vec[A] = HVECTOR (           0.0,           0.0, 1.0, 0.0);
#if 0
  Vec[P] = HVECTOR (           0.0,           0.0, 0.0, 1.0);
#endif

  return *this;
}


// see Fu, Gonzalez, Lee, ROBOTICS, p. 21
FRAME& FRAME::SetRotVector (const HVECTOR &Vector, double Angle)
{
  double  sa, ca, va;
  double  vx, vy, vz;
  double  RadAngle = RAD (Angle);
  HVECTOR Normal = ~Vector;

  PRO_METHOD ("FRAME", "")

  if (fabs(Angle) <= Vec[N].GetEpsilon())
    return (*this) = FRAME ();

  sa = sin (RadAngle);
  ca = cos (RadAngle);
  va = (1.0 - ca);
  vx = (Normal[X]);
  vy = (Normal[Y]);
  vz = (Normal[Z]);

  Vec[N][X] = vx*vx*va + ca;
  Vec[N][Y] = vx*vy*va + vz*sa;
  Vec[N][Z] = vx*vz*va - vy*sa;

  Vec[O][X] = vx*vy*va - vz*sa;
  Vec[O][Y] = vy*vy*va + ca;
  Vec[O][Z] = vy*vz*va + vx*sa;

  Vec[A][X] = vx*vz*va + vy*sa;
  Vec[A][Y] = vy*vz*va - vx*sa;
  Vec[A][Z] = vz*vz*va + ca;

#if 0
  Vec[P][X] = 0.0;
  Vec[P][Y] = 0.0;
  Vec[P][Z] = 0.0;
#endif

  PRO_END
  return *this;
}


// see Fu, Gonzalez, Lee, ROBOTICS, p. 25
FRAME& FRAME::SetRotRPY (double phi, double theta, double psi)
{
  double sphi, cphi;
  double sthe, cthe;
  double spsi, cpsi;
  double RadPhi   = RAD (phi);
  double RadTheta = RAD (theta);
  double RadPsi   = RAD (psi);
  
  PRO_METHOD ("FRAME", "")

  sphi = sin (RadPhi);
  cphi = cos (RadPhi);
  sthe = sin (RadTheta);
  cthe = cos (RadTheta);
  spsi = sin (RadPsi);
  cpsi = cos (RadPsi);

  Vec[N][X] = cphi*cthe;
  Vec[N][Y] = sphi*cthe;
  Vec[N][Z] = -sthe;

  Vec[N][X] = cphi*sthe*spsi - sphi*cpsi;
  Vec[N][Y] = sphi*sthe*spsi + cphi*cpsi;
  Vec[N][Z] = cthe*spsi;

  Vec[A][X] = cphi*sthe*cpsi + sphi*spsi;
  Vec[A][Y] = sphi*sthe*cpsi - cphi*spsi;
  Vec[A][Z] = cthe*cpsi;

#if 0
  Vec[P][X] = 0.0;
  Vec[P][Y] = 0.0;
  Vec[P][Z] = 0.0;
#endif

  PRO_END
  return *this;
}


// see Fu, Gonzalez, Lee, ROBOTICS, p. 23
FRAME& FRAME::SetRotEuler (double phi, double theta, double psi)
{
  double sphi, cphi;
  double sthe, cthe;
  double spsi, cpsi;
  double RadPhi   = RAD (phi);
  double RadTheta = RAD (theta);
  double RadPsi   = RAD (psi);
  
  PRO_METHOD ("FRAME", "")

  sphi = sin (RadPhi);
  cphi = cos (RadPhi);
  sthe = sin (RadTheta);
  cthe = cos (RadTheta);
  spsi = sin (RadPsi);
  cpsi = cos (RadPsi);

  Vec[N][X] = cphi*cpsi - sphi*cthe*spsi;
  Vec[N][Y] = sphi*cpsi + cphi*cthe*spsi;
  Vec[N][Z] = sthe*spsi;

  Vec[O][X] = -cphi*spsi - sphi*cthe*cpsi;
  Vec[O][Y] = -sphi*spsi + cphi*cthe*cpsi;
  Vec[O][Z] = sthe*cpsi;

  Vec[A][X] =  sphi*sthe;
  Vec[A][Y] = -cphi*sthe;
  Vec[A][Z] =  cthe;

#if 0
  Vec[P][X] = 0.0; 
  Vec[P][Y] = 0.0;
  Vec[P][Z] = 0.0;
#endif

  PRO_END
  return *this;
}


/* Computes a frame from given DHP-parameters. While theta is the joint 
   variable, it will be zero for home-position, See Fu,Gonzales,Lee,page 40 */
FRAME &FRAME::SetDHP (double Theta, double d, double a, double Alpha) 
{
  double st, ct, sa, ca;
  double RadTheta, RadAlpha;

  PRO_METHOD ("FRAME", "")

  RadTheta = RAD (Theta);
  RadAlpha = RAD (Alpha);
  st = sin (RadTheta);
  ct = cos (RadTheta);
  sa = sin (RadAlpha);
  ca = cos (RadAlpha);

  Vec[N][X] = ct;
  Vec[N][Y] = st;
  Vec[N][Z] = 0.0;

  Vec[O][X] = -ca*st;
  Vec[O][Y] = ca*ct;
  Vec[O][Z] = sa;

  Vec[A][X] = sa*st;
  Vec[A][Y] = -sa*ct;
  Vec[A][Z] = ca;

  Vec[P][X] = a*ct;
  Vec[P][Y] = a*st;
  Vec[P][Z] = d;

  PRO_END
  return *this;
}


FRAME FRAME::Trans (double x, double y, double z) const
{
  return FRAME().SetPos (x, y, z);
}


FRAME FRAME::Trans (const HVECTOR &Vector) const
{
  return FRAME().SetPos (Vector);
}


FRAME FRAME::RotX (double Angle) const
{
  return FRAME().SetRotX (Angle);
}


FRAME FRAME::RotY (double Angle) const
{
  return FRAME().SetRotY (Angle);
}


FRAME FRAME::RotZ (double Angle) const
{
  return FRAME().SetRotZ (Angle);
}


// see Fu, Gonzalez, Lee, ROBOTICS, p. 21 
FRAME FRAME::RotVector (const HVECTOR &Vector, double Angle) const
{
  return FRAME().SetRotVector (Vector, Angle);
}


HVECTOR FRAME::GetRPY ()
{
  double sr, cr;
  double roll, pitch, yaw;
  
  PRO_METHOD ("FRAME", "")

  roll = atan2 (Vec[N][Y], Vec[N][X]);
  sr = sin (roll);
  cr = cos (roll);

  yaw   = atan2 (-1.0 * Vec[N][Z], 
		 Vec[N][X]*cr + Vec[N][Y]*sr);
  pitch = atan2 (Vec[A][X]*sr - Vec[A][Y]*cr, 
		 Vec[O][Y]*cr - Vec[O][X]*sr);

  PRO_END
  return HVECTOR (GRAD(yaw), GRAD(pitch), GRAD(roll));
}


// See Fu, Gonzalez, Lee: "Robotics", p. 25
FRAME FRAME::RotRPY (double Roll, double Pitch, double Yaw) const
{
  return FRAME().SetRotRPY (Roll, Pitch, Yaw);
}


// see Fu, Gonzalez, Lee: "Robotics", p. 23
FRAME FRAME::RotEuler (double phi, double theta, double psi) const
{
  return FRAME().SetRotEuler (phi, theta, psi);
}


// See Ho, Sriwattanathamma: "Robot Kinematics" p. 144 ff
void FRAME::CalcRotRPY (double &Roll, double &Pitch, double &Yaw) const
{
  double sr, cr;

  PRO_METHOD ("FRAME", "")

  if (!(Thres(Vec[N][Y]) == 0.0 && Thres(Vec[N][X]) == 0.0))
    Roll = atan2 (Vec[N][Y], Vec[N][X]);
  else
    Roll = 0.5 * PI;

  sr = sin (Roll);
  cr = cos (Roll);

  if (Thres(Vec[N][Z]) != 0.0 || Thres(cr*Vec[N][X] + sr*Vec[N][Y]) != 0.0)
    Pitch = GRAD (atan2(- Vec[N][Z], 
			cr*Vec[N][X] + sr*Vec[N][Y]));
  else
    Pitch = 90.0;

  if (Thres( sr*Vec[A][X] - cr*Vec[A][Y]) != 0.0 ||
      Thres(-sr*Vec[O][X] + cr*Vec[O][Y]) != 0.0)
    Yaw = GRAD (atan2(  sr*Vec[A][X] - cr*Vec[A][Y],
		      - sr*Vec[O][X] + cr*Vec[O][Y]));
  else 
    Yaw = 90.0;

  Roll = GRAD (Roll);

  PRO_END
}


// See Fu, Gonzalez, Lee: "Robotics", p. 55 ff
void FRAME::CalcRotEuler (double &phi, double &theta, double &psi) const
{
  PRO_METHOD ("FRAME", "")

  // *** Translate frame into euler-representation
  theta = GRAD (atan2(sqrt(SQR(Vec[N][Z]) + SQR(Vec[O][Z])), Vec[A][Z]));
  if (theta == 0.0)
  {
    phi = 0.0;
    psi = GRAD (atan2(-Vec[O][X], Vec[N][X]));
  }
  elsif (theta == 180.0)
  {
    phi = 0.0;
    psi = GRAD (atan2(Vec[O][X], -Vec[N][X]));
  }
  else
  {
    phi = GRAD (atan2(Vec[A][Y],  Vec[A][X]));
    psi = GRAD (atan2(Vec[O][Z], -Vec[N][Z]));
  }
  PRO_END
}


void FRAME::GetDHP (double *Theta, double *d, double *a, double *Alpha)
{
  double t, st;

  PRO_METHOD ("FRAME", "")

  if (d != NULL)
    *d = Vec[P][Z];

  if (Alpha != NULL)
    *Alpha = GRAD (atan2(Vec[O][Z], Vec[A][Z]));

  t = atan2 (Vec[N][Y], Vec[N][X]);
  st = sin (t);

  if (a != NULL)
    if (Thres(st) != 0.0)  // mte laut F.,G.,L. der cos sein !!!
      *a = Vec[P][X] / st;
    else
      *a = Vec[P][Y] / cos (t);
  
  if (Theta != NULL)
    *Theta = GRAD (t);

  PRO_END
}


// See B. K. P.Horn: p. 632 ff / J.Opt.Soc.Am. A/Vol.4,No.4/April 87
void FRAME::CalcRotVector (HVECTOR &Vector, double &Angle) const
{
  double  FourQ0Q0, Q0 = 0.0;              // Quaternion components
  double  FourQxQx, Qx = 0.0;
  double  FourQyQy, Qy = 0.0;
  double  FourQzQz, Qz = 0.0;
  HVECTOR Rot (0.0, 0.0, 0.0, 0.0);
  FRAME   Frame = ~(*this);                    // orthonormed frame

  PRO_METHOD ("FRAME", "")

  // calculate rotation HVECTOR and angle
  // see J.Opt.Soc.Am. A/Vol.4,No.4/April 87; B.K.P.Horn, page 641 ff
  // we use internal a quaternion representation

  // because the accuracy of arccos is critical near 0, we
  // do some special calculations to achieve a higher numerical 
  // accuracy

  FourQ0Q0 = 1 + Frame[N][X] + Frame[O][Y] + Frame[A][Z];
  FourQxQx = 1 + Frame[N][X] - Frame[O][Y] - Frame[A][Z];
  FourQyQy = 1 - Frame[N][X] + Frame[O][Y] - Frame[A][Z];
  FourQzQz = 1 - Frame[N][X] - Frame[O][Y] + Frame[A][Z];

  if (FourQ0Q0 >= FourQxQx && FourQ0Q0 >= FourQyQy && FourQ0Q0 >= FourQzQz) 
  {
    // FourQ0Q0 is largest value
    Q0 = sqrt (FourQ0Q0) / 2.0;
    Qx = (Frame[O][Z] - Frame[A][Y]) / (4.0 * Q0);
    Qy = (Frame[A][X] - Frame[N][Z]) / (4.0 * Q0);
    Qz = (Frame[N][Y] - Frame[O][X]) / (4.0 * Q0);
  } 
  elsif (FourQxQx >= FourQ0Q0 && FourQxQx >= FourQyQy && FourQxQx >= FourQzQz) 
  {
    // FourQxQx is largest value
    Qx = sqrt (FourQxQx) / 2.0;
    Q0 = (Frame[O][Z] - Frame[A][Y]) / (4.0 * Qx);
    Qy = (Frame[N][Y] + Frame[O][X]) / (4.0 * Qx);
    Qz = (Frame[A][X] + Frame[N][Z]) / (4.0 * Qx);
  } 
  elsif (FourQyQy >= FourQ0Q0 && FourQyQy >= FourQxQx && FourQyQy >= FourQzQz) 
  {
    // FourQyQy is largest value
    Qy = sqrt (FourQyQy) / 2.0;
    Q0 = (Frame[A][X] - Frame[N][Z]) / (4.0 * Qy);
    Qx = (Frame[N][Y] + Frame[O][X]) / (4.0 * Qy);
    Qz = (Frame[O][Z] + Frame[A][Y]) / (4.0 * Qy);
  } 
  elsif (FourQzQz >= FourQ0Q0 && FourQzQz >= FourQxQx && FourQzQz >= FourQyQy) 
  {
    // FourQzQz is largest value
    Qz = sqrt (FourQzQz) / 2.0;
    Q0 = (Frame[N][Y] - Frame[O][X]) / (4.0 * Qz);
    Qx = (Frame[A][X] + Frame[N][Z]) / (4.0 * Qz);
    Qy = (Frame[O][Z] + Frame[A][Y]) / (4.0 * Qz);
  }
  else 
  {
    PRO0 ("FRAME::GetRotHVECTOR() internal error: no max value found\n")
    assert (FALSE);
  }

  // calculate angle and rotation HVECTOR
  Angle = 2.0 * GRAD (acos(Q0));
  if (fabs(Angle) > Vec[N].GetEpsilon() && (Qx!=0.0 || Qy!=0.0 || Qz!=0.0))
  {
    Rot[X] = Qx;
    Rot[Y] = Qy;
    Rot[Z] = Qz;
    assert (Thres(sin(acos(Q0))) != 0.0);
    Vector = (1.0 / sin(acos(Q0))) * Rot; 
  }
  else
  {
    Angle = 0.0;
    Vector = Rot;
  }

  PRO_END
}


void FRAME::Write (FILE *fp /* = NULL */, 
		   const char *Title /* = NULL */, ...) const
{
  char    Header[255];
  va_list Arguments;

  PRO_METHOD ("FRAME", "")

  if (fp == NULL)
    fp = stdout;

  if (Title != NULL)
  {
    va_start (Arguments, Title);
    vsprintf (Header, Title, Arguments);
    va_end (Arguments);
    fprintf (fp, "%s:\n", Header);
  }

  fprintf (fp, "%6.3f  %6.3f  %6.3f  %9.2f\n", 
	  Vec[N][X], Vec[O][X], Vec[A][X], Vec[P][X]);
  fprintf (fp, "%6.3f  %6.3f  %6.3f  %9.2f\n", 
	  Vec[N][Y], Vec[O][Y], Vec[A][Y], Vec[P][Y]);
  fprintf (fp, "%6.3f  %6.3f  %6.3f  %9.2f\n", 
	  Vec[N][Z], Vec[O][Z], Vec[A][Z], Vec[P][Z]);

  PRO_END
}


#undef Thres
