#include <math.h>
#include <assert.h>

#include <debug.h>

#include "libsim3d.h"


//static definitions
double GENERIC::Epsilon       (0.00005);
double GENERIC::EpsilonLength (0.0001);


/* Verbesserungen:
   * Vergleich zwischen letzter Gelenkstellung prfen (Joint 4 ?)
*/

#define Thres(x) (fabs(x) <= Epsilon? 0.0:x)


GENERIC::GENERIC (INTERFACE *Interface) :
  ROBOT (Interface),
  DHP_D (0.0,  49.0, 0.0, 310.0, 0.0, 65.0),
  DHP_A (0.0, 290.0, 0.0,   0.0, 0.0,  0.0),
  Home  (0.0,   0.0, 0.0,   0.0, 0.0,  0.0) // r2: 0.0,90.0,-90.0,0.0,0.0,0.0
{
  OldSpeed = GetSpeed ();               // save speed-settings

  OpenHand ();
}


GENERIC::~GENERIC ()
{
  SetSpeed (OldSpeed);                  // Restore old speed
}


// Inverse Kinematic for Knickarm-Roboters (see Fu: 'Robotics' p. 63 ff)
// Anpassung, da a3=0 und Abstand fr d1 bentigt wird.
// Weitere Anpassungen notwendig, da andere Nullpositionen der Gelenke
// des Pumas angenommen werden als die des Stubli.
// Daher: Vorher an Puma anpassen, nachher die Winkel korrigieren.
// Achtung ist geboten bei Benutzung der atan2-Funktion fr die Parameter 0,0.
// Diese knnen sich je nach Compiler unterscheiden!
// Rechenzeit: 60us?? im optimierten Code auf Pentium 133MHz
// Ausnahmen OutOfWorkspace(falls check), OutOfLimit(falls LastJoint)
// schneller (keine Objekt Erzeugung)
JOINTS GENERIC::InvKin (const FRAME  &Frame,
	 		      int     Configuration,
		        const JOINTS &LastJoint,
		              bool    CheckOOW       /* = TRUE */,
		              bool    ValidLastJoint /* = TRUE */)
{
  int     m, k;
  double  sign;
  double  sj[6], cj[6];
  double  s0, c0;
  double  s12, c12;
  double  s3, c3;
  double  s_alpha, c_alpha;
  double  s_beta,  c_beta;
  double  s_phi,   c_phi;
  double  s_psi,   c_psi;
  double  r, R, RR;
  double  OldEpsilonLength = 0.0;
  double  Omega;
  HVECTOR Pos;
  JOINTS  Joint, TryJoint;
  FRAME   dif, maxdif;
  FRAME   TCP;

  PRO_METHOD ("GENERIC", "")

  assert (Configuration >= 0 && Configuration <= 7);
  assert (Frame.IsOrthoNorm());

  // Compute TCP in world-coordinates respecting robot-base and TCP-Offset
  TCP = (! Base) * Frame * (! Offset);

  // Subtract length of joint 4 and 5 from robot position
  Pos = TCP[P] - DHP_D[5] * TCP[A];

  // Transform relative to robot base (if z-coordinate is not zero)
  Pos[Z] -= DHP_D[0];

  // Compute Eq. 2.3-49 (b)
  if ((r = SQR(Pos[X]) + SQR(Pos[Y]) - SQR(DHP_D[1])) < 0.0)
  {
    PRO0 ("OutOfWorkspace(128): px^2 + py^2 - DHP_D[1]^2 < 0\n")
    throw KINEMATIC::OutOfWorkspace (128);
  }
  r = sqrt (r);

  // Compute Eq. 2.3-49 (a)
  if ((RR = SQR(Pos[X]) + SQR(Pos[Y]) + SQR(Pos[Z]) - SQR(DHP_D[1])) < 0.0)
  {
    PRO0 ("OutOfWorkspace(64): RR < 0\n")
    throw KINEMATIC::OutOfWorkspace (64);
  }
  R = sqrt (RR);
  assert (Thres(R) != 0.0);

  // ================== SOLUTION FOR JOINT 0 =================================
  sign = IsArmLeft(Configuration) ? 1.0:(-1.0);
  sj[0] = sign * Pos[Y] * r - Pos[X] * DHP_D[1];  // Eq. (2.3-45)
  cj[0] = sign * Pos[X] * r + Pos[Y] * DHP_D[1];  // Eq. (2.3-46)

//  assert (Thres(sj[0]) != 0.0 || Thres(cj[0]) != 0.0);
  Joint[0] = atan2 (sj[0], cj[0]);
  s0 = sin (Joint[0]);
  c0 = cos (Joint[0]);
  Joint[0] = GRAD (Joint[0]);

  // ================== SOLUTION FOR JOINT 1 =================================
  s_alpha = - Pos[Z] / R;               // Eq. (2.3-50)
  c_alpha = sign * r / R;               // Eq. (2.3-51)

//  assert (DHP_A[1] != 0.0);                              // R always != 0
  c_beta = (SQR(DHP_A[1]) + RR - (SQR(DHP_D[3]) + SQR(DHP_A[2]))) / 
	    (2 * DHP_A[1] * R);         // Eq. (2.3-52
  if (fabs(c_beta) > 1.0) 
  {
    PRO0 ("OutOfWorkspace(256): 1 - cbeta^2 < 0\n")
    throw KINEMATIC::OutOfWorkspace (256);
  }
  s_beta = sqrt (1.0 - SQR(c_beta));

  k = 1;                           // Konfigurationsbit ermitteln
  if ((IsArmLeft (Configuration) && IsElbowBelow(Configuration)) ||
      (IsArmRight(Configuration) && IsElbowAbove(Configuration)))
    k = -1;

  sj[1] = (s_alpha*c_beta + k*c_alpha*s_beta);
  cj[1] = (c_alpha*c_beta - k*s_alpha*s_beta);
  assert (Thres(sj[1]) != 0.0 || Thres(cj[1]) != 0.0);
  Joint[1] = GRAD (atan2(sj[1], cj[1]));

  if (Joint[1] > JointMax[1] && (Joint[1]-360.0) > JointMin[1])
    Joint[1] -= 360.0;

  // ================== SOLUTION FOR JOINT 2 =================================
  // values of atan2 are from -180 to 180, but the real range may be larger !
  assert (Thres(DHP_A[1] * DHP_D[3]) != 0.0);
  c_phi = (SQR(DHP_A[1]) + SQR(DHP_D[3]) - RR) / (2 * DHP_A[1] * DHP_D[3]);
                                                       // hier: DHP_A[3] = 0
  assert (fabs(c_phi) <= 1.0);                // = 1-SQR(c_phi) >= 0
  s_phi = k * sqrt(1-SQR(c_phi));           // K von Gelenk2, K = Elbow*Arm!

  s_psi = 1.0;                        // aufgrund von DHP_A[3]=0 ergibt sich 1!
  c_psi = 0.0;                        // aufgrund von DHP_A[3]=0 ergibt sich 0!

  sj[2] = (s_phi*c_psi - c_phi*s_psi);
  cj[2] = (c_phi*c_psi + s_phi*s_psi);
  assert (Thres(sj[2]) != 0.0 || Thres(cj[2]) != 0.0);
  Joint[2] = GRAD (atan2(sj[2], cj[2]));
  if (Joint[2] < JointMin[2] && (Joint[2]+360.0) < JointMax[2])
    Joint[2] += 360.0;

  s12 = sin (RAD(Joint[1]+Joint[2]));
  c12 = cos (RAD(Joint[1]+Joint[2]));

  // ================== SOLUTION FOR JOINT 3 =================================
  // Erst Konfigurationsrichtung ermitteln.

  // A-Vektor hinter Gelenk 2, nach Fu = z2
  HVECTOR z2  (c0*s12, s0*s12, c12), vza(z2 ^ TCP[A]);  // vza = z3 x TCP.A
  if (vza.Len() < EpsilonLength) // degeneriert. Fall: Achse Gelenk3/5 parallel
  {	// J3 beliebig, zur Sicherheit letzten Wert bernehmen falls vorhanden!
    if (ValidLastJoint) 
      Joint[3] = LastJoint[3];
    else
      Joint[3] = 0.0;

    c3 = cos (RAD(Joint[3]));
    s3 = sin (RAD(Joint[3]));

    // Berechnung fr Gelenk 5:
    if (EpsilonLength == 1.0)
    {
      sj[4] = (  (c0 * c12 * c3 - s0 * s3) * TCP[A][X]
	       + (s0 * c12 * c3 + c0 * s3) * TCP[A][Y]
 	       - c3 * s12                  * TCP[A][Z]);
      cj[4] = (  s0 * s12 * TCP[A][X] 
               + s0 * s12 * TCP[A][Y] 
               + c12      * TCP[A][Z]);
      // Falls Joint4 == 0, degenerierter Fall, siehe Gelenk 4.
      assert (Thres(sj[4]) != 0.0 || Thres(cj[4]) != 0.0);
      Joint[4] = GRAD (atan2(sj[4], cj[4]));
    }
  }
  else
  {
    vza.Normalize ();

    if (fabs((Omega = TCP[O] * vza)) < Epsilon)
      Omega = TCP[N] * vza;

    m = 1; // Konfigurationsbit
    if ((IsWristUp(Configuration) && (Omega <  0)) ||
	(IsWristDown(Configuration) && (Omega >= 0))) // Andere Konfiguration?
      m = -1;

    // Gelenkwinkel 4 berechnen
    sj[3] = (m * (  c0  * TCP[A][Y]
                  - s0  * TCP[A][X]));
    cj[3] = (m * (  c0  * c12 * TCP[A][X]
                  + s0  * c12 * TCP[A][Y]
                  - s12 *       TCP[A][Z]));
    assert (Thres(sj[3]) != 0.0 || Thres(cj[3]) != 0.0);
    Joint[3] = GRAD (atan2(sj[3], cj[3]));

    // Lsung liegt zwischen +-180, Gelenk 3 kann aber +-270.
    // Daher teilweise mehrdeutig! Um Umdrehen des Gelenkes zu verhindern,
    // die Lsung nehmen, die mglichst nahe bei letzter Lsung liegt.
    if (ValidLastJoint && Joint[3] < 0.0 &&
	fabs(fabs(Joint[3] - LastJoint[3]) - 360.0) < EpsilonJoint[3])
      Joint[3] += 360.0;
    elsif (ValidLastJoint && Joint[3] > 0.0 &&
           fabs(fabs(Joint[3] - LastJoint[3]) - 360.0) < EpsilonJoint[3])
      Joint[3] -= 360.0;

    c3 = cos (RAD(Joint[3]));
    s3 = sin (RAD(Joint[3]));

    // ================== SOLUTION FOR JOINT 4 ===============================
    // Im degenerierten Fall kann es - lt. Fu - zu einem Fehler in der 
    // allgemeinen Annahme zur Berechnung von Gelenk 5 kommen. Daher wird 
    // dieses im degenerierten Fall auf null gesetzt und die Berechnung so 
    // an dieser Stelle lediglich im nichtdegenerierten Fall ausgefhrt.
    sj[4] = (  (c0 * c12 * c3 - s0 * s3) * TCP[A][X]
	     + (s0 * c12 * c3 + c0 * s3) * TCP[A][Y]
	     -  c3 * s12                 * TCP[A][Z]);
    cj[4] = (   c0 * s12 * TCP[A][X]
             +  s0 * s12 * TCP[A][Y] 
             + c12       * TCP[A][Z]);
    // Falls Joint4 == 0, degenerierter Fall, siehe Gelenk 4.
    assert (Thres(sj[4]) != 0.0 || Thres(cj[4]) != 0.0);
    Joint[4] = GRAD (atan2(sj[4], cj[4]));
  } // else nichtdegenerierter Fall Gelenk 4

  // ================== SOLUTION FOR JOINT 5 =================================
  sj[5] = (  (-s0 * c3 - c0 * c12 * s3) * TCP[N][X]
	   + ( c0 * c3 - s0 * c12 * s3) * TCP[N][Y]
           + (           s3 * s12     ) * TCP[N][Z]);
  cj[5] = (  (-s0 * c3 - c0 * c12 * s3) * TCP[O][X]
           + ( c0 * c3 - s0 * c12 * s3) * TCP[O][Y]
           + (           s3 * s12     ) * TCP[O][Z]);
  assert (Thres(sj[5]) != 0.0 || Thres(cj[5]) != 0.0);
  Joint[5] = GRAD (atan2(sj[5], cj[5]));

  if (ValidLastJoint)
  {
    // Lsung liegt zwischen +-180, Gelenk 5 kann aber +-270.
    // Daher teilweise mehrdeutig! Um Umdrehen des Gelenkes zu verhindern,
    // die Lsung nehmen, die mglichst nahe bei letzter Lsung liegt.
    if (ValidLastJoint && Joint[5] < 0.0 &&
	fabs(fabs(Joint[5]-LastJoint[5]) - 360.0) < EpsilonJoint[5])
      Joint[5] += 360.0;
    elsif (ValidLastJoint && Joint[5] > 0.0 &&
	   fabs(fabs(Joint[5]-LastJoint[5]) - 360.0) < EpsilonJoint[5])
      Joint[5] -= 360.0;

    // Prfen, ob sich ein Gelenk strker als ein vorgegebenens EpsilonJoint
    // gegenber dem letzten Stand (LastJoint) gendert hat. Zur Vorsicht
    // wird dann ein OutOfLimit geliefert.
    try 
    {
      CheckOutOfLimit (Joint, LastJoint);
    }
    catch (KINEMATIC::OutOfLimit &OOL)
    {  // Versuchen LastJoint[3] zu bernehmen um den OutOfLimit zu beseitigen.
      if ((EpsilonLength < 1.0) &&               // Stetigkeit gewhrleisten
	  (OOL.ErrorJoint & (1<<3)))           //Joint[3] OutOfLimit
      {
	PRO1 ("OutOfLimit: %d (trying joint #3)\n", OOL.ErrorJoint)

        OldEpsilonLength = EpsilonLength;
	EpsilonLength = 1.0;         // bewirkt die bernahme von LastJoint[3]

	try
	{
	  TryJoint = InvKin (TCP, Configuration, LastJoint, CheckOOW);
	}
	catch (OutOfWorkspace)
	{
	  PRO0 ("OutOfWorkspace:  "
		"Versuchen Stetigkeit in Joint[3] zu gewhrleisten\n")
	  // Dies kann auf das Erreichen einer Grenze bei Gelenk 3 oder 5 
	  // zurckzufhren sein.
	  EpsilonLength = OldEpsilonLength;

     	  dif = TCP - FwdKin (TryJoint, FALSE);
	  maxdif.Init (RAD(0.1));
	  maxdif[P] = HVECTOR (0.19, 0.19, 0.19, 1.0);
	  if (dif <= maxdif)
	  {
	    Joint = TryJoint;
	    throw;                 // OOW - Ausnahme weiterreichen
	  }
	  else
	    throw OOL;
	}
	catch (KINEMATIC::OutOfLimit)
	{
	  PRO0 ("OutOfLimit: "
		"Versuchen Stetigkeit in Gelenk[3] zu gewhrleisten\n")
	  EpsilonLength = OldEpsilonLength;
	  throw OOL;
	}
	EpsilonLength = OldEpsilonLength;

	dif = TCP - FwdKin (TryJoint, FALSE);
       	maxdif.Init (0.0019);
	maxdif[P] = HVECTOR (0.19, 0.19, 0.19, 1.0);
	if (dif <= maxdif)
	  return TryJoint;
	else 
	{
	  PRO1 ("OutOfLimit: %d:  Joint[3] is discontinuous\n", OOL.ErrorJoint)
	  throw;         // Abweichung bedenklich gro !
	}
      }             //Erste Korrektionsversuch fur Joint[3] OutOfLimit
      else 
      {
	PRO1 ("OutOfLimit: %d\n", OOL.ErrorJoint)
        throw;       // Abweichung bedenklich gro !
      }
    }  // catch OutOfLimit OOL
  } //if LastJoint

  Joint += Home;

  // Prfen, ob Lsung innerhalb des Roboterbereiches liegt.
  // Falls auerhalb Arbeitsbereich OutOfWorkspace Ausnahme!!
  if (CheckOOW) 
    CheckJoint (Joint); 

  PRO_END
  return Joint;
}


JOINTS GENERIC::InvKin (const FRAME &Frame,		   //orig frame
		              int    Configuration,
		              bool   CheckOOW /* = TRUE */)
{
  JOINTS LastJoint;

  return InvKin (Frame, Configuration, LastJoint, CheckOOW, FALSE);
}


int GENERIC::InvKin (const FRAME  &Frame,			//Ursp frame
	                   JOINTS &Joint,			//Ziel joint
	             const JOINTS &LastJoint,		        //Vergl joint. 
	                   bool    CheckOOW /* = TRUE */)
{
  int Configuration;

  Configuration = CalcConfiguration (LastJoint, FALSE);
  Joint = InvKin (Frame, Configuration, LastJoint, CheckOOW, TRUE);

  return Configuration;
}


// kein default Wert fuer checkOOW, um von die nchste Funk 
// (fehlt Konf, LastJoint, checkOOW, checkOOL) zu unterscheiden
JOINTS GENERIC::InvKin (const FRAME  &Frame,			//orig frame
	           	const JOINTS &LastJoint,		//Vergl joint.
		              bool    CheckOOW)
{	                          
  int Configuration;

  Configuration = CalcConfiguration (LastJoint, FALSE);

  return InvKin (Frame, Configuration, LastJoint, CheckOOW, TRUE); 
}


int GENERIC::InvKin (const FRAME  &Frame, 
                           JOINTS &Joint)
{
  bool ValidConf;
  int  Configuration, ErrorJoint (0);

  /* Test all configurations, if one of them is valid */
  for (Configuration = 0; Configuration <= 7; Configuration++)
  {
    ValidConf = TRUE;

    try
    {
      Joint = InvKin (Frame, Configuration);
    }
    catch (const KINEMATIC::OutOfWorkspace &OOW)
    {
      PRO2 ("OutOfWorkspace: %d    Konf %d\n", OOW.ErrorJoint, Configuration)
      ErrorJoint = OOW.ErrorJoint;
      ValidConf = FALSE;
    }

    if (ValidConf) 
      return Configuration;
  }
  PRO1 ("OutOfWorkspace: %d - No volid configurations found\n", ErrorJoint)

  throw KINEMATIC::OutOfWorkspace (ErrorJoint);

  return -1;
}


JOINTS GENERIC::InvKin (const FRAME &Frame)
{
  JOINTS Joint;

  InvKin (Frame, Joint);

  return Joint;
}


// Dasselbe wie InvKin(const FRAME &p_Frame)
JOINTS GENERIC::CheckFrame (const FRAME &Frame)
{ 
  JOINTS Joint;

  InvKin (Frame, Joint);

  return Joint;
}


// Forward kinematic fr Staeubli RX 60
// Entnommen aus "Robotics", Fu, Seite 45
// Anpassung, da a3=0.0 und d1 fr Abstand Basis/Tisch hinzukommt.
// Rechenzeit: 48us?? mit optimierten Code auf Pentium 133MHz
FRAME GENERIC::FwdKin (const JOINTS  &Joint, bool Check /* = TRUE */)
{
  double c0  = cos (RAD(Joint[0]-Home[0]));
  double c1  = cos (RAD(Joint[1]-Home[1]));
  double c12 = cos (RAD(Joint[1]-Home[1]+Joint[2]-Home[2]));
  double c3  = cos (RAD(Joint[3]-Home[3]));
  double c4  = cos (RAD(Joint[4]-Home[4]));
  double c5  = cos (RAD(Joint[5]-Home[5]));
  double s0  = sin (RAD(Joint[0]-Home[0]));
  double s1  = sin (RAD(Joint[1]-Home[1]));
  double s12 = sin (RAD(Joint[1]-Home[1]+Joint[2]-Home[2]));
  double s3  = sin (RAD(Joint[3]-Home[3]));
  double s4  = sin (RAD(Joint[4]-Home[4]));
  double s5  = sin (RAD(Joint[5]-Home[5]));
  FRAME  Frame;

  PRO_METHOD ("GENERIC", "")

  if (Check)
    CheckJoint (Joint);

  // Berechnung n-Vektor
  Frame[N][X] =   c0  * (c12*(c3*c4*c5-s3*s5) - s12*s4*c5) 
                - s0  * (s3*c4*c5 + c3*s5);
  Frame[N][Y] =   s0  * (c12*(c3*c4*c5-s3*s5) - s12*s4*c5)
                + c0  * (s3*c4*c5 + c3*s5);
  Frame[N][Z] = - s12 * (c3*c4*c5 - s3*s5) 
                - c12 * s4*c5;

  // Berechnung o-Vektor
  Frame[O][X] =   c0  * (-c12*(c3*c4*s5+s3*c5) + s12*s4*s5)
                - s0  * (-s3*c4*s5 + c3*c5);
  Frame[O][Y] =   s0  * (-c12*(c3*c4*s5+s3*c5) + s12*s4*s5)
                + c0  * (-s3*c4*s5 + c3*c5);
  Frame[O][Z] =   s12 * (c3*c4*s5 + s3*c5) 
	        + c12 * s4*s5;

  // Berechnung a-Vektor
  Frame[A][X] =   c0  * (c12*c3*s4 + s12*c4)
                - s0  * s3*s4;
  Frame[A][Y] =   s0  * (c12*c3*s4 + s12*c4) 
                + c0  * s3*s4;
  Frame[A][Z] = - s12 * c3*s4 
	        + c12 * c4;

  // Berechnung p-Vektor
  Frame[P][X] =   c0  * (DHP_D[5]*(c12*c3*s4 + s12*c4)
	        + s12 * DHP_D[3] + DHP_A[1]*c1)  // A3 = 0
	        - s0  * (DHP_D[5]*s3*s4 + DHP_D[1]); 
  Frame[P][Y] =   s0  * (DHP_D[5]*(c12*c3*s4 + s12*c4) 
	        + s12 * DHP_D[3] + DHP_A[1]*c1)  // A3 = 0
                + c0  * (DHP_D[5]*s3*s4 + DHP_D[1]);
  Frame[P][Z] =   DHP_D[5] * (c12*c4-s12*c3*s4) 
	        + c12 * DHP_D[3]
	        - DHP_A[1]*s1 + DHP_D[0];  // D1 eingefgt

  Frame = Base * Frame * Offset;  // Multiply Result with Base-Offset and 
                                  // TCP-Offset to get Tool-Center-Point

  PRO_END
  return Frame;
}


// Compute a robot's arm configuration respecting the initial joint-values.
// See: "Robotics", Fu, Gonzales, Lee, page 69 (Table 2.4)
int GENERIC::CalcConfiguration (const JOINTS &Joint,
	         	        const FRAME  &Frame,
			              bool    Check /* = TRUE */)
{
  int    Configuration = 0;
  double g0, g1, g2;
  double c0  = cos (RAD(Joint[0]-Home[0]));
  double c1  = cos (RAD(Joint[1]-Home[2]));
  double c2  = cos (RAD(Joint[2]-Home[2]));
  double c12 = cos (RAD(Joint[1]-Home[1]+Joint[2]-Home[2]));
  double c3  = cos (RAD(Joint[3]-Home[3]));
  double s0  = sin (RAD(Joint[0]-Home[0]));
  double s12 = sin (RAD(Joint[1]-Home[1]+Joint[2]-Home[2]));
  double s3  = sin (RAD(Joint[3]-Home[3]));

  assert(Frame.IsOrthoNorm());

  if (Check) 
    CheckJoint (Joint);

  // Arm links/rechts?   
  g0 = Thres (-DHP_D[3] * s12 - DHP_A[1] * c1); // DHP_A3 = 0
  if (g0 < 0.0)
    Configuration += ARM_LEFT;
  else
    Configuration += ARM_RIGHT; // bei g0==0: per Definition Rechtsarm

  // Ellenbogen unten/oben?
  g1 = (Thres(DHP_D[3] * c2));   // DHP_A3 = 0
  if ((SIGN(g0)*SIGN(g1) == 1) || ((SIGN(g0) == 0) && (SIGN(g1) == 1)))
    Configuration += ELBOW_BELOW;
  else
    Configuration += ELBOW_ABOVE;

  // Handgelenk oben/unten?
  // Frame aus Joint berechnen, sofern nicht als Parameter bergeben

  // z4 Vektor berechnen
  HVECTOR z4 (-c0*c12*s3-s0*c3, -s0*c12*s3+c0*c3 , s12*s3);

  g2 = Frame[O] * z4;
  if (fabs(g2) < Epsilon)              // g2 == 0, daher n*z4 prfen!
    g2 = Frame[N] * z4;
  if (g2 < 0.0)
    Configuration += WRIST_UP; 
  else
    Configuration += WRIST_DOWN;

  return Configuration;
}


int GENERIC::CalcConfiguration (const JOINTS &Joint,
				      bool    Check /* = TRUE */)
{
  FRAME Frame;

  Frame = FwdKin (Joint, FALSE);

  return CalcConfiguration (Joint, Frame, Check);
}


// CheckJoint prft, ob die Gelenkstellung Joint durch
// den Roboter erreicht werden kann, die Grenzen also
// eingehalten werden. Ausnahme: OutOfWorkspace
void GENERIC::CheckJoint (const JOINTS &Joint)
{
  int count;
  int Error (0);                       // binary pattern for bad joints

  for (count = 0; count < 6; count++)
    if (Joint[count] < JointMin[count] || Joint[count] > JointMax[count])
    {
      Error += (1 << count);
      PRO4 ("OutOfWorkspace Joint %d: %f Range (%f - %f)\n",
	    count+1, Joint[count], JointMin[count], JointMax[count])
    }

  if (Error > 0) 
  {
    PRO1 ("OutOfWorkspace: %d\n", Error)
    throw KINEMATIC::OutOfWorkspace (Error);
  }
}


void GENERIC::CheckOutOfLimit(const JOINTS &Joint, const JOINTS &LastJoint)
{
  int count;
  int Error (0);                      // binary pattern for bad joints

  for (count = 0; count < 6; count++)
    if (fabs(LastJoint[count]-Joint[count]) > EpsilonJoint[count])
    {
      Error += (1 << count);
      PRO4 ("OutOfLimit Joint #%d: %.5f -> %.5f Epsilon:%.3f\n",
	    count+1, LastJoint[count], Joint[count], EpsilonJoint[count])
    }

  if (Error > 0) 
  {
    PRO1 ("OutOfLimit: %d\n", Error)
    throw KINEMATIC::OutOfLimit (Error); // Auerhalb des Gltigkeitsbereiches!
  }
}


double GENERIC::GetSensor (SENSOR_ID Sensor)
{
  char   *Buffer = NULL;
  char    Command[40] = "";
  double  Result = -1.0;

  if (Interface == NULL)
  {
    PRO0 ("No connection\n")
    return -1.0;
  }

  switch (Sensor)
  {
    case LIGHT_BARRIER :
      strcpy (Command, "GET SIG (1003)");
      break;

    case HAND_RIGHT : 
      strcpy (Command, "GET SIG (1001)");
      break;

    case HAND_LEFT : 
      strcpy (Command, "GET SIG (1002)");
      break;

    case SONAR :
      strcpy (Command, "GET BITS (1004,8)");
      break;

    case TRIANGULAR :
      strcpy (Command, "GET BITS (1004,8)");
      break;

    default :
      break;
  }
  Buffer = Interface->SendCmd (Command);

  if (Buffer != NULL)
    Result = atof (Buffer);

  return Result;
}


void GENERIC::SetEpsilon (double Epsilon)
{
  this->Epsilon = Epsilon;
}


void GENERIC::SetEpsilonLength (double EpsilonLen)
{
  EpsilonLength = EpsilonLen;
}


double GENERIC::GetEpsilon ()
{
  return Epsilon;
}


double GENERIC::GetEpsilonLength ()
{
  return EpsilonLength;
}


void GENERIC::SetDHP (JOINTS a, JOINTS d, JOINTS Alpha, JOINTS Theta)
{
  DHP_A     = a;
  DHP_D     = d;
  DHP_Alpha = Alpha;
  DHP_Theta = Theta;
};


void GENERIC::SetDHP (double D0, double D1, double A1, double D3, double D5)
{
  DHP_D[0] = D0;      // Position Tisch/Roboterbasis (341.0)
  DHP_D[1] = D1;
  DHP_A[1] = A1;
  DHP_D[3] = D3;
  DHP_D[5] = D5;
}


void GENERIC::SetHome (JOINTS Home)
{
  this->Home = Home;
}


void GENERIC::SetJointMin (const JOINTS &Min)
{
  JointMin = Min;
}


void GENERIC::SetJointMax (const JOINTS &Max)
{
  JointMax = Max;
}


const JOINTS &GENERIC::GetJointMin ()
{
  return JointMin;
}


const JOINTS &GENERIC::GetJointMax ()
{
  return JointMax;
}

#undef Thres
