#include <assert.h>
#include <math.h>

#include "libsim3d.h"


#define CHECK_HVECTOR(Vec) assert((Vec).w == 1.0 || (Vec).w == 0.0)


HVECTOR::HVECTOR (double x /* = 0.0 */, double y /* = 0.0 */, 
		  double z /* = 0.0 */, double w /* = 0.0 */) :
  VECTOR (3, x, y, z),
  w      (w)
{
}


HVECTOR::HVECTOR (const HVECTOR &Vector) :
  VECTOR (3, Vector[0], Vector[1], Vector[2]),
  w      (Vector.w)
{
  assert (Vector.GetDim() == 3);
}


HVECTOR &HVECTOR::operator = (const HVECTOR &Vec)
{
  int count;

  CHECK_HVECTOR (Vec);
  assert (Vec.GetDim() == 3);

  if (this != &Vec)
  {
    for (count = 0; count < 3; ++count)
      Val[count] = Vec[count];
    w = Vec.w;
  }

  return *this;
}


bool HVECTOR::operator == (const HVECTOR &Vec) const
{
  int count;

  CHECK_HVECTOR (Vec);

  for (count = 0; count < Dim; ++count)
    if (fabs(Val[count]-Vec.Val[count]) > compare_epsilon)
      return FALSE;

  return (w == Vec.w);
}


double &HVECTOR::operator [] (int i)
{
  if (i < 3)
    return Val[i];
  else
    return w;
}


double HVECTOR::operator [] (int i) const
{
  if (i < 3)
    return Val[i];
  else
    return w;
}


HVECTOR HVECTOR::operator + (const HVECTOR &Vec) const
{
  CHECK_HVECTOR (Vec);

  return HVECTOR (*this) += Vec;
}


HVECTOR HVECTOR::operator - (const HVECTOR &Vec) const
{
  CHECK_HVECTOR (Vec);

  return HVECTOR (*this) -= Vec;
}


HVECTOR HVECTOR::operator ^ (const HVECTOR &Vec) const
{
  CHECK_HVECTOR (Vec);

  return HVECTOR (*this) ^= Vec;
}


HVECTOR HVECTOR::operator ~ () const
{
  CHECK_HVECTOR (*this);

  return HVECTOR (*this).Normalize ();
}


HVECTOR &HVECTOR::operator += (const HVECTOR &Vec)
{
  int count;

  CHECK_HVECTOR (Vec);

  for (count = 0; count < 3; ++count)
    Val[count] += Vec.Val[count];

  return *this;
}


HVECTOR &HVECTOR::operator -= (const HVECTOR &Vec)
{
  int count;

  CHECK_HVECTOR (Vec);

  for (count = 0; count < 3; ++count)
    Val[count] -= Vec.Val[count];

  return *this;
}


HVECTOR &HVECTOR::operator ^= (const HVECTOR &Vec)
{
  double OldX = Val[0], OldY = Val[1];

  CHECK_HVECTOR (Vec);

  Val[0] = Val[1] * Vec[Z] - Val[2] * Vec[Y];
  Val[1] = Val[2] * Vec[X] -   OldX * Vec[Z];
  Val[2] =   OldX * Vec[Y] -   OldY * Vec[X];

  return *this;
}


HVECTOR &HVECTOR::Normalize ()
{
  return *this *= (1/Len());
}


HVECTOR HVECTOR::operator + () const
{
  return HVECTOR (+Val[0], +Val[1], +Val[2], w);
}
 

HVECTOR HVECTOR::operator - () const
{
  return HVECTOR (-Val[0], -Val[1], -Val[2], w);
}


HVECTOR operator * (double s, const HVECTOR &Vec)
{
  HVECTOR Result (Vec);

  CHECK_HVECTOR (Vec);

  return Result *= s;
}


HVECTOR operator / (const HVECTOR &Vec, double s)
{
  HVECTOR Result (Vec);

  CHECK_HVECTOR (Vec);

  return Result /= s;
}


HVECTOR &operator *= (HVECTOR &Vec, double s)
{
  int count;

  CHECK_HVECTOR (Vec);

  for (count = 0; count < 3; ++count)
    Vec[count] *= s;

  return Vec;
}


HVECTOR &operator /= (HVECTOR &Vec, double s)
{
  int count;

  CHECK_HVECTOR (Vec);

  for (count = 0; count < 3; ++count)
    Vec[count] /= s;

  return Vec;
}
