#include <math.h>
#include <assert.h>
#include <string.h>

#include <debug.h>

#include "libsim3d.h"

#define SIMPLE

#ifdef SIMPLE
#define JOINT_A_LEN    300.00
#define JOINT_A_OFFS     0.00
#define JOINT_B_LEN    500.00
#define JOINT_B_OFFS    45.00
#define JOINT_B_RADIUS  12.50
#define PLATE_ROT        0.00
#define PLATE_OFFS      -7.50
#define FLANGE_X        84.6688 
#define FLANGE_Y        48.00
#define FLANGE_Z        -7.50
#define MOTOR_R        300.00
#define MOTOR_OFFSY     80.00
#define MOTOR_Z        750.00
#else
#define JOINT_A_LEN    200.00
#define JOINT_A_OFFS     0.00
#define JOINT_B_LEN    470.00
#define JOINT_B_OFFS    22.50
#define JOINT_B_RADIUS   5.00
#define PLATE_ROT       30.00
#define PLATE_OFFS       6.0
#define FLANGE_X        33.49
#define FLANGE_Y        25.00
#define FLANGE_Z        -6.00
#define MOTOR_R        300.00
#define MOTOR_OFFSY     65.00
#define MOTOR_Z        750.00
#endif


HEXA::HEXA (INTERFACE *Interface /* = NULL */) :
  ROBOT    (Interface),
  LenA     (JOINT_A_LEN),
  LenB     (JOINT_B_LEN),
  OffsA    (JOINT_A_OFFS),
  OffsB    (JOINT_B_OFFS)
{
  int   count;
  FRAME Rot;

  PRO_START ("")

  // relative coodinates:
  FRAME Flange1 (1,0,0,   0,1,0,   0,0,1,   FLANGE_X,  FLANGE_Y, FLANGE_Z);
  FRAME Flange2 (1,0,0,   0,1,0,   0,0,1,   FLANGE_X, -FLANGE_Y, FLANGE_Z);

  // absolute coodrinates:
  FRAME Motor1  (0,0,-1,  -1,0,0,   0,1,0,   
		 MOTOR_R, -(MOTOR_OFFSY-JOINT_B_OFFS), MOTOR_Z);
  FRAME Motor2  (0,0,-1,   1,0,0,   0,-1,0,    
		 MOTOR_R,  (MOTOR_OFFSY-JOINT_B_OFFS), MOTOR_Z);

  OldSpeed = GetSpeed ();               // save speed-settings

  OpenHand ();

  for (count = 0; count < 3; count++)
  {
    Rot = FRotZ (count*120.0);
    Motor[2*count]   = Rot * Motor1;
    Motor[2*count+1] = Rot * Motor2;

    Rot = FRotZ (-count*120.0 - PLATE_ROT);
    Mount[2*count]   = Rot * Flange1;
    Mount[2*count+1] = Rot * Flange2;
  }
  SetBase (0.0);

  SetKinematic (this);
  SetInterface (Interface);
  PRO_END
}


HEXA::~HEXA ()
{
  SetSpeed (OldSpeed);                  // Restore old speed
}


// Berechnet den/die Schnittpunkt(e) zwischen einer Kugel und einem im
// Raum liegenen Kreis. Der Kreis liegt in der xy-Ebene des durch M2
// gegebenen Mittelpunkt-Frames.
// Verfahren:
// Kreis in den Ursprung der xy-Ebene transformieren. Die Kugel entsprechend
// mittransformieren. => z-Komponente der transformierten Schnittpunkte ist 0.
// Falls Abstand von Kugel zu xy-Ebene > 0, gibt es keine Lsung. Sonst den 
// Radius des Schnittkreises der Kugel und xy-Ebene berechnen (acos + sin). 
// Mittelpunkt des Schnittkreises durch Projektion des Kugelmittelpunktes
// in die xy-Ebene berechnen (z-Komponente = 0).
// Abstand zwischen beiden Kreisen gibt an ob es keine, eine oder zwei
// Lsungen gibt. Schnitt zwischen beiden Kreisen berechnen. Punkte 
// anschlieen zurcktransformieren.
int CalcSphereCircleIntersection (FRAME    M1, double   r1,      // sphere 
				  FRAME    M2, double   r2,      // circle
				  HVECTOR &P1, HVECTOR &P2)
{
  double  r, c, dist;
  double  q;
  HVECTOR n, u;
  HVECTOR l, k;
  FRAME   M;

  PRO_START ("\n")

  M = ! M2 * M1;               // Compute sphere center relative to circle

  if (M[P][Z] > r1)
    PRO_RETURN ("No sphere/circle intersection", 0)

  r = sin (acos(M[P][Z] / r1)) * r1;

  M[P][Z] = 0.0;               // project the center of circle onto xy-layer

  PRO3 ("  Circle1: M1: (%8.2f / %8.2f)  r1: %.4f\n", M[P][X], M[P][Y], r)
  PRO1 ("  Circle2: M2: (    0.00 /     0.00)  r2: %.4f\n", r2)

  dist = M[P].Len ();
  if (fabs(r-r2) > dist || dist > fabs(r+r2))  // Kreis in Kreis oder 
    PRO_RETURN ("no circles interection", 0)    // zu groer Abstand ?

  // Schnittgerade beider Kreise berechnen:
  //   g = [n] * [r] - c = 0
  //   [n] = 2 * (M[P] - M2[P]);  
  //   c = M2[P]^2 - M[P]^2 - r2^2 + r^2
  //   u: Normalenvektor zu n
  n = 2 * M[P];
  c = M[P]*M[P] + r2*r2 - r*r;

  u[X] =   n[Y];
  u[Y] = - n[X];
  u[Z] =   0.0;

  q = (r2*r2) * (n*n) - c*c;
  if (q >= 0.0)
  {
    l = c / (n * n) * n;
    k = 1 / (n*n) * sqrt(q) * u;

    P1 = M2 * (l + k);
    PRO3 ("  P1xyz:  %.3f %.3f %.3f\n", P1[X], P1[Y], P1[Z])

    if (q > 0.0)
    {
      P2 = M2 * (l - k);
      PRO3 ("  P2xyz:  %.3f %.3f %.3f\n", P2[X], P2[Y], P2[Z]);
      PRO_RETURN ("Two intersection points", 2)
    }
    else
      PRO_RETURN ("One intersection point", 1)
  }

  PRO_END
  return 0;
}


// Berechnet die Lotfusspunkte zweier Vektoren
dual CalcLineLineLot (HVECTOR  Pos1,   HVECTOR  Dir1, 
                      HVECTOR  Pos2,   HVECTOR  Dir2,
		      HVECTOR *Point1, HVECTOR *Point2)
{
  HVECTOR h, i;
  double  r = 0, s = 0;

  PRO_START ("\n")

  h[X] = (Pos1 - Pos2) * Dir1;
  h[Y] = Dir1 * Dir1;
  h[Z] = Dir2 * Dir1;

  i[X] = (Pos1 - Pos2) * Dir2;
  i[Y] = Dir1 * Dir2;
  i[Z] = Dir2 * Dir2;

  /* Compute the position of the 'Lotfusspunkt' */
  if ((h[Z]*i[Y] - h[Y]*i[Z]) != 0.0)
  {
    r = (h[X]*i[Z] - h[Z]*i[X]) / 
        (h[Z]*i[Y] - h[Y]*i[Z]);
    s = (h[X]*i[Y] - h[Y]*i[X]) / 
	(h[Z]*i[Y] - h[Y]*i[Z]);
  }
  else
    PRO_RETURN ("Division by zero for r and s", FALSE)

  if (Point1 != NULL)
    *Point1 = Pos1 + (r * Dir1);

  if (Point2 != NULL)
    *Point2 = Pos2 + (s * Dir2);

#if 0
  if (fabs(!(Point1[0] - v[2]))) > D_DIST)
    PRO_RETURN ("Vector not zero", FALSE)

  Result = GEOVectorMult (0.5, GEOVectorAdd(v[0], v[1]));
#endif

  PRO_END
  return TRUE;
}


double CalcLinesDist (HVECTOR Pos1, HVECTOR Dir1, HVECTOR Pos2, HVECTOR Dir2)
{
  double Det, Div;

  Det =   ((Pos1[X] - Pos2[X]) * Dir1[Y] * Dir2[Z])
        + ((Pos1[Y] - Pos2[Y]) * Dir1[Z] * Dir2[X])
        + ((Pos1[Z] - Pos2[Z]) * Dir1[X] * Dir2[Y])
        - (Dir2[X] * Dir1[Y] * (Pos1[Z] - Pos2[Z]))
        - (Dir2[Y] * Dir1[Z] * (Pos1[X] - Pos2[X]))
        - (Dir2[Z] * Dir1[X] * (Pos1[Y] - Pos2[Y]));

  if (Det == 0.0)
    return 0.0;                         /* lines are crossing each other */

  Div = sqrt (  SQR((Dir1[X] * Dir2[Y]) - (Dir2[X] * Dir1[Y]))
	      + SQR((Dir1[Y] * Dir2[Z]) - (Dir2[Y] * Dir1[Z]))
	      + SQR((Dir1[Z] * Dir2[X]) - (Dir2[Z] * Dir1[X])));

  if (Div == 0.0)
    return 0.0;

  return  fabs (Det / Div);
}


// Get the 6 motor angles and the corresponding TCP and calculates the 
// three angles of the passive socket-joint
bool HEXA::CalcPassive (JOINTS Joint, FRAME TCP, 
                        JOINTS &J1, JOINTS &J2, JOINTS &J3,
                        bool   check /* = TRUE */)
{
  int     count, next;
  double  Dist;
  HVECTOR Dir[6], L1, L2;
  HVECTOR Orient[6];
  double  Len1, Len2;
  FRAME   Flange[6], Mid[6];

  for (count = 0; count < 6; count++)
  {
    Flange[count] = TCP * Mount[count];

    // Compute frame of socket-joint
    Mid[count] = Motor[count] * FRotZ(Joint[count]) * FTransX (JOINT_A_LEN);

    // Diretion RELATIVE to Mid:
    Dir[count] = (! Mid[count]) * Flange[count][P];

    // Direction in world coordinates:
    Orient[count] = Flange[count][P] - Mid[count][P];

    J1[count] = - GRAD (atan2(-Dir[count][Y], Dir[count][X]));
    J2[count] = - GRAD (asin(Dir[count][Z] / JOINT_B_LEN));
    J3[count] = 0.0;    // rotation about joint-axis rounded to zero

    // falls Winkel zwischen Kugelgelenk und Gelenk2 > Max-Winkel:
    // throw (KINEMATIC::OutOfLimit);

    // Winkel zwischen Gelenk2 und Mountplatte testen
  }

  if (check)            // Check, if there`s a collision between two joints
    for (count = 0; count < 6; count++)
    {
      next = (count+1) % 6;
      Dist = CalcLinesDist ((Mid[count])[P], Orient[count], 
			    (Mid[next])[P],  Orient[next]);
      // PRO3 ("DISTANCE %d <-> %d: %.3f\n", count, next, Dist);

      if (Dist <= 2*JOINT_B_RADIUS &&
	  CalcLineLineLot((Mid[count])[P], Orient[count], 
			  (Mid[next])[P],  Orient[next], &L1, &L2))
      {
	Len1 = (L1 - (Mid[count])[P]).Len ();
	Len2 = (L2 - (Mid[next])[P]).Len ();
	if (Len1 <= JOINT_B_LEN && Len2 <= JOINT_B_LEN)
	  printf ("COLLISION !!!: %d <-> %d: %f / %f\n", count, next,
		  Len1, Len2);
      }
    }

  return TRUE;
}


JOINTS HEXA::InvKin (const FRAME &Frame)
{
  int     count, No;
  int     error(0);
  HVECTOR Pos, P1, P2;
  HVECTOR Rel;
  JOINTS  Joint (18, 0.0);
  JOINTS  J, J1, J2, J3;
  FRAME   Flange, Plate;

  Plate = Frame;

  for (count = 0; count < 6; count++)
  {
    Flange = Frame * Mount[count];
// Flange.Print ("InvKin %d", count);

    No = CalcSphereCircleIntersection (Flange, LenB, Motor[count], LenA, 
				       P1, P2);
    if (No <= 0)
      error += (1 << count);
    else
    {
      if (No == 2) // TODO: alle Gelenkstellungen testen (kleinerer zuerst ?)
	Pos = (P1[Z] < P2[Z]) ? P1 : P2;
      else
	Pos = P1;

      // TODO: Gelenkwinkel max. +/- 90.0, wegen Winkel (Vektor, Vektor)
      Rel = Pos - Motor[count][P];
      J[count] = 90.0 - GRAD(Rel % Motor[count][O]);
    }
  }

  if (error > 0)
  {
    PRO1 ("OUT OF WORKSPACE: %d\n", error);
    throw KINEMATIC::OutOfWorkspace (error);
  }

  PRO6 ("%.3f %.3f %.3f %.3f %.3f %.3f\n",
	J[0], J[1], J[2], J[3], J[4], J[5])
 
  CalcPassive (J, Frame, J1, J2, J3, FALSE);
  for (count = 0; count < 6; count++)
  {
    Joint[count*3]   = J[count];                  // active joints
    Joint[count*3+1] = J2[count];                 // passive joints
    Joint[count*3+2] = J1[count];
  }

  return Joint;
}


JOINTS HEXA::InvKin  (const FRAME  &Frame, 
		            int     Configuration,
		      const JOINTS &LastJoint, 
		            bool    checkOOW       /* = TRUE */,
		            bool    ValidLastJoint /* = TRUE */)
{
  return InvKin (Frame);
}


JOINTS HEXA::InvKin (const FRAME &Frame, 
                           int    configuration,  
		           bool   CheckOOW /* = TRUE */)
{ 
  return InvKin (Frame);
}


int HEXA::InvKin (const FRAME  &Frame, 
                        JOINTS &Joint,
		  const JOINTS &LastJoint, 
                        bool    CheckOOW /* = TRUE */)
{ 
  Joint = InvKin (Frame);

  return 0; 
}


JOINTS HEXA::InvKin (const FRAME  &Frame, 
                     const JOINTS &LastJoint,
		           bool   CheckOOW)
{
  return InvKin (Frame);
}


int HEXA::InvKin (const FRAME  &Frame, 
                        JOINTS &Joint)
{ 
  int Configuration = 0;

  Joint = InvKin (Frame);

  return Configuration;
}


FRAME HEXA::FwdKin (const JOINTS &Joint, bool Check /* = TRUE */)
{
  PRO_METHOD ("HEXA", "\n")

  if (Joint.GetNumber() == 18) 
  {
    int     count;
    HVECTOR Center[3], Dir[3];
    FRAME   Mid[6], End[6];
  
    // This is still a simple solution for computing a hexa's fwd kinematic
    if (Joint.GetNumber() < 18)
      PRO_RETURN ("Not enough angles for fwd kinematics\n", TCP)

    for (count = 0; count < 6; count++)
    {
      Mid[count] = 
	Motor[count] * FRotZ(Joint[3*count]) * FTransX (JOINT_A_LEN);

      End[count] = Mid[count] * 
	           FRotY(Joint[3*count+1]) * FRotZ(Joint[3*count+2]) * 
                   FTransX (JOINT_B_LEN);
      // End[count].Print ("FwdKin %d", count);
    }
  
    // Calc center of each side
    Center[0] = 0.5 * (End[0][P] + End[1][P]);
    Center[1] = 0.5 * (End[2][P] + End[3][P]);
    Center[2] = 0.5 * (End[4][P] + End[5][P]);

    // Compute TCP-Frame:
    Dir[0] = ~(Center[0] - Center[1]);
    Dir[1] = ~(Center[2] - Center[1]);
    Dir[2] = Dir[0] ^ Dir[1] ;
    Dir[0][W] = Dir[1][W] = Dir[2][W] = 0.0;
    FRAME TCP (Dir[0], Dir[1], Dir[2], 
	       (Center[0] + Center[1] + Center[2]) / 3.0);
    TCP.SetOrthoNorm ();

    TCP *= FTransZ (PLATE_OFFS);
    PRO_END
    return TCP;
  }
  else
  {
    FRAME   TCP;
    char   *Buffer;
    double  rpy[6];

    if (Interface == NULL || (Buffer = Interface->SendCmd("RPY")) == NULL ||
	sscanf(Buffer, "%lf %lf %lf %lf %lf %lf",
	       &rpy[0], &rpy[1], &rpy[2], &rpy[3], &rpy[4], &rpy[5]) < 6)
      PRO0 ("Error computing Forward kinematics")

    PRO6 ("RPY= %.2f %.2f %.2f   %.2f %.2f %.2f\n", 
	  rpy[0], rpy[1], rpy[2], rpy[3], rpy[4], rpy[5])

    TCP.SetRotRPY (rpy[5], rpy[4], rpy[3]);
    TCP[P][X] = rpy[0];
    TCP[P][Y] = rpy[1];
    TCP[P][Z] = rpy[2];

    TCP.Write (NULL, "HEXA.FwdKin:");

    PRO_END
    return TCP;
  }
}


int HEXA::CalcConfiguration (const JOINTS &Joint, 
			            bool  Check /* = TRUE */)
{
  // BitNo = JointNo: 
  // 0: smaller angle,  1: larger angle
  return 0; 
}


int HEXA::CalcConfiguration (const JOINTS &Joint, 
                             const FRAME  &frame, 
			           bool    Check /* = TRUE */)
{
  return 0;
}


JOINTS HEXA::GetPassive (FRAME Frame, JOINTS Joint)
{
  JOINTS Result;

  return Result;
}


JOINTS HEXA::CheckFrame (const FRAME &Frame)
{
  JOINTS Joint;

  return Joint;
}


void HEXA::CheckJoint (const JOINTS &Joint)
{
}


void HEXA::CheckOutOfLimit (const JOINTS &Joint, const JOINTS &LastJoint)
{
}


// ========================================================================
#ifdef TODO
bool HEXA::SetSpeed (double Speed)  // m/sec */
{
  if (Speed <= 0.0)
    return FALSE;

  this->Speed = Speed;
  SetInterpPerMeter ((1000 / GetClock()) / Speed);

  return TRUE;
}


double HEXA::GetSpeed ()
{
  char *Buffer = NULL;

  if (SocketID < 0 || (Buffer = NETSend(SocketID, "GET SPEED(1)")) == NULL)
  {
    PRO0 ("Error getting speed\n")
    return -1.0;
  }

  return atof (Buffer);
}


bool HEXA::OpenHand ()
{
  if (SocketID >= 0)
    NETSend (SocketID, "OPEN");

  return TRUE;
}


bool HEXA::CloseHand ()
{
  if (SocketID >= 0)
    NETSend (SocketID, "CLOSE"); 

  return TRUE;
}


int HEXA::GetHand ()
{ 
  char *Buffer = NULL;

  if (SocketID < 0 ||
      (Buffer = NETSend(SocketID, "GET HAND")) == NULL)
  {
    PRO0 ("Error getting hand-status\n")
    return -1;
  }

  return (Buffer[0] == '1');
}
#endif


bool HEXA::Place (const JOINTS &Joint)
{
  double  RPY[3];
  char   *Buffer = NULL;

printf ("HEXA::Place\n");

  if (Interface != NULL)
  {
    NextFrame.CalcRotRPY (RPY[0], RPY[1], RPY[2]);
    if ((Buffer = Interface->SendCmd("HEXA "
				     "%.3f %.3f %.3f %.3f %.3f %.3f "
				     "%.3f %.3f %.3f %.3f %.3f %.3f",
				     NextFrame[P][X], NextFrame[P][Y],
				     NextFrame[P][Z],
				     RPY[2], RPY[1], RPY[0],
				     Joint[0], Joint[1], Joint[2],
				     Joint[3], Joint[4], Joint[5])) == NULL)
    {
      PRO0 ("Error moving robot\n")
      return FALSE;
    }
  }

  return (Buffer != NULL && Buffer[0] == 'O');
}


#if 0
bool HEXA::Break ()
{
  if (SocketID < 0)
  {
    PRO0 ("No connection\n")
    return FALSE;
  }

  return (NETSend(SocketID, "WAIT") != FALSE);
}


JOINTS HEXA::GetRobotPos ()
{
  char   *Buffer;
  JOINTS  Joint (0.0, 0.0, 0.0, 0.0, 0.0, 0.0);

  if (SocketID < 0 ||
      (Buffer = NETSend(SocketID, "HERE")) == NULL ||
      sscanf(Buffer, "%lf %lf %lf %lf %lf %lf",
	             &Joint[0], &Joint[1], &Joint[2], 
	             &Joint[3], &Joint[4], &Joint[5]) < 6)
    PRO0 ("Error reading Robot-position\n")

  return Joint;
}
#endif


FRAME HEXA::GetPos ()
{
  JOINTS Joint;

  Joint = GetRobotPos ();

  return FwdKin (Joint);
}

