#include <assert.h>
#include <stdio.h>
#include <math.h>

#include "libsim3d.h"


#define CHECK_RANGE(J) (J.Min <= J.Max && J.Min <= J.Value && J.Value <= J.Max)


double JOINTS::Accuracy (0.001);


JOINTS::JOINTS (const JOINTS &J) :
  No (J.No)
{
  int count;

  Range = new RANGE[No];
  for (count = 0; count < No; count++)
    Range[count] = J.Range[count];
}


JOINTS::JOINTS (int Number, double Value /* = 0.0 */) :
  No (Number)
{
  int count;

  assert (Number > 0);

  Range = new RANGE[No];
  for (count = 0; count < Number; count++)
    Range[count].SetValue (Value);
}


JOINTS::JOINTS (int Number, const double *J) :
  No (Number)
{
  int count;

  assert (Number > 0);

  Range = new RANGE[No];
  for (count = 0; count < No; count++)    
    Range[count].SetValue (J[count]);
}


JOINTS::JOINTS (double J0 /* = 0.0 */, double J1 /* = 0.0 */,
	        double J2 /* = 0.0 */, double J3 /* = 0.0 */,
	        double J4 /* = 0.0 */, double J5 /* = 0.0 */) :
  No (6)
{
  Range = new RANGE[No];

  Range[0].SetValue (J0);
  Range[1].SetValue (J1);
  Range[2].SetValue (J2);
  Range[3].SetValue (J3);
  Range[4].SetValue (J4);
  Range[5].SetValue (J5);
}


JOINTS::~JOINTS ()
{
  delete Range;
}


bool JOINTS::operator == (const JOINTS &J) const
{
  int count = 0;

  if (No != J.No)
    return FALSE;

  while (count < No)
  {
    if (fabs(Range[count].GetValue() - 
	     J.Range[count].GetValue()) > Range[count].GetAccuracy())
      return FALSE;
    count++;
  }

  return TRUE;
}


JOINTS &JOINTS::operator = (const JOINTS &J)
{
  int count;

  if (this != &J)                          // check for self-assignment
  {
    delete Range;

    No = J.No;
    Range = new RANGE[J.No];

    for (count = 0; count < J.No; count++)
    {
      Range[count] = J.Range[count];
      Range[count].Check ();
    }
  }

  return *this;
} 

// addition and assignation of two HVECTORs (faster)   caution! w=0
JOINTS &JOINTS::operator += (const JOINTS &J)      
{
  int count;

  assert (No == J.No);

  for (count = 0; count < No; count++)
  {
    assert (Range[count].GetType() == J.Range[count].GetType());
    Range[count].SetValue (Range[count].GetValue() +
			   J.Range[count].GetValue());
    Range[count].Check ();
  }

  return *this;
}


// subtraction and assignation of two HVECTORs (faster) caution! w=0
JOINTS &JOINTS::operator -= (const JOINTS &J)      
{
  int count;

  assert (No == J.No);

  for (count = 0; count < No; count++)
  {
    assert (Range[count].GetType() == J.Range[count].GetType());
    Range[count].SetValue (Range[count].GetValue() -
			   J.Range[count].GetValue());
    Range[count].Check ();
  }

  return *this;
}


// addition of two HVECTORs (slower)  caution! w=0
JOINTS JOINTS::operator + (const JOINTS &J) const
{
  return JOINTS (*this) += J;
}


// subtraction of two HVECTORs (slower) caution! w=0
JOINTS JOINTS::operator - (const JOINTS &J) const
{
  return JOINTS (*this) -= J;
}


bool JOINTS::operator != (const JOINTS &J) const
{
  return ! (*this == J);               // use simply the equal comparison :-) 
}


double& JOINTS::operator [] (int Index)
{
  assert (Index >= 0 && Index < No);

  return Range[Index].GetValue();
}


double JOINTS::operator [] (int Index) const
{
  assert (Index >= 0 && Index < No);

  return Range[Index].GetValue();
}


bool JOINTS::operator < (const JOINTS &J) const
{
  int count = 0;

  if (No != J.No)
    return FALSE;

  while (count < No)
  {
    if (Range[count].GetValue() >= J.Range[count].GetValue())
      return FALSE;
    count++;
  }

  return TRUE;
}


bool JOINTS::operator <= (const JOINTS &J) const
{
  int count = 0;

  if (No != J.No)
    return FALSE;

  while (count < No)
  {
    if (Range[count].GetValue() > J.Range[count].GetValue())
      return FALSE;
    count++;
  }

  return TRUE;
}


bool JOINTS::operator > (const JOINTS &J) const
{
  int count = 0;

  if (No != J.No)
    return FALSE;

  while (count < No)
  {
    if (Range[count].GetValue() <= J.Range[count].GetValue())
      return FALSE;
    count++;
  }

  return TRUE;
}


bool JOINTS::operator >= (const JOINTS &J) const
{
  int count = 0;

  if (No != J.No)
    return FALSE;

  while (count <= No)
  {
    if (Range[count].GetValue() < J.Range[count].GetValue())
      return FALSE;
    count++;
  }

  return TRUE;
}


JOINTS& JOINTS::Round (double Value /* = Accuracy */)
{
  int count;

  for (count = 0; count < No; count++)
    if (fabs(Range[count].GetValue()) < Value)
      Range[count].SetValue (0.0);

  return *this;
}


JOINTS JOINTS::Round (double Value /* = Accuracy */) const
{
  return JOINTS (*this).Round (Value);
}


JOINTS JOINTS::MakeRad () const
{
  return JOINTS (*this).Rad ();
}

JOINTS &JOINTS::Rad (JOINTS &J) const
{
  int count;

  for (count = 0; count < J.No; count++)
    if (J.Range[count].GetType() == RANGE::ROTATORIC)
      J.Range[count].SetValue (J.Range[count].GetValue() * PI / 180.0);

  return J;
}

JOINTS &JOINTS::Rad ()
{
  int count;

  for (count = 0; count < No; count++)
    if (Range[count].GetType() == RANGE::ROTATORIC)
      Range[count].SetValue (Range[count].GetValue() * PI / 180.0);

  return *this;
}


JOINTS JOINTS::MakeDeg () const
{
  return JOINTS (*this).Deg ();
}

JOINTS &JOINTS::Deg (JOINTS &J) const
{
  int count;

  for (count = 0; count < J.No; count++)
    if (J.Range[count].GetType() == RANGE::ROTATORIC)
      J.Range[count].SetValue (J.Range[count].GetValue() * 180.0 / PI);

  return J;
}

JOINTS &JOINTS::Deg () 
{
  int count;

  for (count = 0; count < No; count++)
    if (Range[count].GetType() == RANGE::ROTATORIC)
      Range[count].SetValue (Range[count].GetValue() * 180.0 / PI);

  return *this;
}


int JOINTS::GetNumber () const
{
  return No;
}

void JOINTS::SetAccuracy (double Accuracy)
{
  this->Accuracy = Accuracy;
}


double JOINTS::GetAccuracy () const
{
  return Accuracy;
}


void JOINTS::Write (FILE *fp, const char *Title /* = NULL */) const
{
  int count;

  if (fp == NULL)
    fp = stdout;

  if (Title != NULL)
    fprintf (fp, "'%s' ", Title);
  fprintf (fp, "No: %d\n", No);
  for (count = 0; count < No; count++)
    fprintf (fp, "%7.2f%s ", Range[count].GetValue(), 
	    (Range[count].GetType() == RANGE::ROTATORIC)? "":"mm");
  fprintf (fp, "\n");
}


// scalar with JOINTS multiplication
JOINTS &operator *= (JOINTS &J, double s)
{
  int count;

  for (count = 0; count < J.No; count++)
    J.Range[count].SetValue (J.Range[count].GetValue() * s);

  return J;
}


JOINTS operator * (const JOINTS &J, double s)
{
  JOINTS Result (J);

  return Result *= s;
}


JOINTS operator * (double s, const JOINTS &J)
{
  JOINTS Result (J);

  return Result *= s;
}
