#include <math.h>
#include <stdlib.h>
#include <assert.h>
#include <stdarg.h>

#include <debug.h>

#include "libsim3d.h"


double VECTOR::compare_epsilon = VECTOR_EPSILON;


VECTOR::VECTOR (int Dim, ...) :
  Dim (Dim)
{
  int     count;
  va_list Arguments;

  Val = new double [Dim];

  va_start (Arguments, Dim);
  for (count = 0; count < Dim; ++count)
    Val[count] = va_arg (Arguments, double);
  va_end (Arguments);
}


VECTOR::VECTOR (const VECTOR &Vec) :
  Dim (Vec.Dim)
{
  int count;

  Val = new double [Dim];

  for (count = 0; count < Dim; ++count)
    Val[count] = Vec.Val[count];
}


VECTOR::~VECTOR ()
{
  if (Val != NULL)
    delete Val;
}


VECTOR &VECTOR::operator = (const VECTOR &Vec)
{
  int count;

  if (this != &Vec)
  {
    if (Val != NULL)
      delete Val;
    Dim = Vec.Dim;
    Val = new double [Dim];
    for (count = 0; count < Dim; ++count)
      Val[count] = Vec.Val[count];
  }

  return *this;
}


bool VECTOR::operator == (const VECTOR &Vec) const
{
  int count;

  assert (Dim == Vec.Dim);

  for (count = 0; count < Dim; ++count)
    if (fabs(Val[count]-Vec.Val[count]) > compare_epsilon)
      return FALSE;

  return TRUE;
}


bool VECTOR::operator != (const VECTOR &Vec) const
{
  return ! (*this == Vec);
}


bool VECTOR::operator <=  (const VECTOR &Vec) const
{
  int count;

  assert (Dim == Vec.Dim);

  for (count = 0; count < Dim; ++count)
    if (Val[count] > Vec[count]) 
      return FALSE;

  return TRUE;
}


VECTOR& VECTOR::operator += (const VECTOR &Vec)
{
  int count;

  assert (Dim == Vec.Dim);

  for (count = 0; count < Dim; ++count)
    Val[count] += Vec.Val[count];

  return *this;
}


VECTOR &VECTOR::operator -= (const VECTOR &Vec)
{
  int count;

  assert (Dim == Vec.Dim);

  for (count = 0; count < Dim; ++count)
    Val[count] -= Vec.Val[count];

  return *this;
}


VECTOR VECTOR::operator + (const VECTOR &Vec) const
{
  return VECTOR (*this) += Vec;
}


VECTOR VECTOR::operator - (const VECTOR &Vec) const
{
  return VECTOR (*this) -= Vec;
}


VECTOR VECTOR::operator ~ () const 
{
  return VECTOR (*this).Normalize ();
}


double VECTOR::operator * (const VECTOR &Vec) const
{ 
  int    count;
  double Result = 0.0;

  assert (Dim == Vec.Dim);

  for (count = 0; count < Dim; ++count)
    Result += Val[count] * Vec[count];

  return Result;
}


VECTOR VECTOR::operator + () const
{
  VECTOR Result (*this);

  return Result;
}


VECTOR VECTOR::operator - () const
{
  int     count;
  VECTOR Result (*this);

  for (count = 0; count < Dim; ++count)
    Result.Val[count] *= -1;

  return Result;
}


double &VECTOR::operator [] (int i)
{
  assert (i >= 0 && i < Dim);

  return Val[i];
}


double VECTOR::operator [] (int i) const
{
  assert (i >= 0 && i < Dim);

  return Val[i];
}


double VECTOR::operator % (const VECTOR &Vec) const
{
  assert (Dim == Vec.Dim);

  return acos (~(*this) * ~Vec);
}


void VECTOR::Init (double Value /* = 0.0 */)
{
  int count;

  for (count = 0; count < Dim; ++count)
    Val[count] = Value;
}


int VECTOR::GetDim () const
{
  return Dim;
}


double VECTOR::Len () const                 // Calc euclidean length
{
  int    count;
  double Result = 0.0;

  for (count = 0; count < Dim; ++count)
    Result += SQR (Val[count]);
  Result = sqrt (Result);

  return Result;
}


VECTOR &VECTOR::Normalize () 
{
  double length = Len ();

  assert (length > 0.0);

  return *this *= (1/length);  
}


void VECTOR::SetEpsilon (double Epsilon)
{
  compare_epsilon = Epsilon;
}


double VECTOR::GetEpsilon () const
{
  return compare_epsilon;
}


void VECTOR::Write (char *Title /* = NULL */)
{
  int count;

  if (Title != NULL)
    PRO1 ("%s:\n", Title)

  PRO1 ("#%d: [", Dim)
  for (count = 0; count < Dim; ++count)
  {
    PRO1 ("\\%5.3f", Val[count])
    if (count < Dim-1)
      PRO0 ("\\ / ")
  }
  PRO0 ("\\]\n")
}


VECTOR& operator *= (VECTOR &Vec, double s)
{
  int count;

  for (count = 0; count < Vec.GetDim(); ++count)
    Vec[count] *= s;

  return Vec;
}


VECTOR operator * (const VECTOR &Vec, double s)
{
  VECTOR Result (Vec);

  return Result *= s;
}


VECTOR operator * (double s, const VECTOR &Vec)
{
  VECTOR Result (Vec);

  return Result *= s;
}


VECTOR &operator /= (VECTOR &Vec, double s)
{
  int count;

  assert (s != 0.0);

  for (count = 0; count < Vec.GetDim(); ++count)
    Vec[count] /= s;

  return Vec;
}


VECTOR operator / (const VECTOR &Vec, double s)
{
  VECTOR Result (Vec);

  return Result /= s;
}
