#include <limits.h>

#include <ctype.h>

#include "libtools.h"


/* ************************************************************ */
/* ** converts a complete string into uppercase letters. The ** */
/* ** memory for the result must be allocated by calling     ** */
/* ** function                                               ** */
/* ** INPUT:  Str     string to be converted                 ** */
/* ** OUTPUT: Result  upcased string                         ** */
/* ************************************************************ */
void TOOLstrtoupper (char *Result, char *Str)
{
  int count, len;

  len = strlen (Str);
  for (count = 0; count < len; ++count)
    Result[count] = toupper(Str[count]);
  Result[count] = '\0';
}


/*----------------------------------------------------------------------
  Funktion: Durchsucht eine Zeichenkette nach Umgebungsvariablen und 
            ersetzt diese (falls sie definiert sind) durch ihren Inhalt
            Die Umgebungsvars knnen mit Klammern, z.B#. $(HOME) oder 
            ohne Klammern, d.h. $HOME enthalten sein.

  INPUT:    FileName  Dateiname (evtl. mit Umgebunsvariablen)
            Result    Zeiger auf den allokierten Speicherbereich, in den
                      das Ergebnis geschrieben werden soll
            Length    Maximale Lnge, die fr das Ergebnis verfgbar ist
  OUTPUT:   (char*)   Kopie des Adresszeigers
  
  Bsp.:     char   Result[256];
            ...
            printf (TOOLParseFileName("$(HOME)/bin/$UNIX/foo", Result, 256));
              liefert (je nach Var-Inhalt) z.B.: "/users/name/bin/linux/foo"
  ----------------------------------------------------------------------*/
char *TOOLParseFileName (const char *FileName, char *Result, int Length)
{
  int   len, start = 0;
  int   fcount = 0, ncount = 0;
  int   vcount;
  char  VarName[PATH_MAX+1];
  char *Path = NULL;

  Result[0] = '\0';

  if (FileName == NULL)
    return NULL;

  while (FileName[fcount] != '\0' && ncount < Length)
  {
    VarName[0] = '\0';
    if (FileName[fcount] == '$')
    {
      start = fcount++;
      if (FileName[fcount] == '(')
        fcount++;

      vcount = 0;
      while ((isalpha(FileName[fcount]) || FileName[fcount] == '_') &&  
             FileName[fcount] != '\0' && vcount < PATH_MAX)
        VarName[vcount++] = FileName[fcount++]; 
      VarName[vcount] = '\0';
    }
    elsif (FileName[fcount] == '~')
      strcpy (VarName, "HOME");

    if (VarName[0] != '\0')
    {
      if ((Path = getenv(VarName)) == NULL)
        for (vcount = start; vcount < fcount; ++vcount)
          Result[ncount++] = FileName[vcount];
      else
      {
	   len = MIN ((int)strlen(Path), Length-ncount-1);
	   Result[ncount] = '\0';
	   strncat (Result, Path, len);
	   ncount += len;
      }
    }

    if ((Path == NULL || (strchr(")~", FileName[fcount]) == NULL)) &&
        FileName[fcount] != '$' && ncount < Length-1)
      Result[ncount++] = FileName[fcount];

    if (FileName[fcount] != '$')
      fcount++;
  }
  while (ncount < Length)
    Result[ncount++] = '\0';

  return Result;
}


/*----------------------------------------------------------------------
  Funktion: Diese Routine ermittelt den reinen Pfad einer kompletten
            Dateiangabe. Der angegeben String kann auch Umgebungsvariablen
            enthalten.

  INPUT:    FileName  Dateiname (evtl. mit Umgebunsvariablen)
            Result    Zeiger auf den allokierten Speicherbereich, in den
                      das Ergebnis geschrieben werden soll
            Length    Maximale Lnge, die fr das Ergebnis verfgbar ist
  OUTPUT:   (char*)   Kopie des Adresszeigers
  
  Bsp.:     char   Result[256];
            ...
            printf (TOOLGetFileDir("$(HOME)/data/foo", Result, 256));
              liefert je nach Var-Inhalt z.B.: "/users/name/data"
  ----------------------------------------------------------------------*/
char *TOOLGetFileDir (const char *FileName, char *Result, int Length)
{
  char *Pos = NULL;

  Result[0] = '\0';

  if (FileName == NULL)
    return NULL;

  TOOLParseFileName (FileName, Result, Length);

  if ((Pos = strrchr(Result, '/')) != NULL)
    Pos[0] = '\0';            /* Pos[1], if last '/' is desired */
  else
    Result[0] = '\0';

  if (Result[0] == '\0')
    return NULL;

  return Result;
}


/*----------------------------------------------------------------------
  Funktion: Hierbei handelt es sich um das Gegenstck zu TOOLGetFileDir
            Sie ermittelt den reinen Name der Datei ohne Pfad

  INPUT:    FileName  Dateiname (evtl. mit Umgebunsvariablen)
            Result    Zeiger auf den allokierten Speicherbereich, in den
                      das Ergebnis geschrieben werden soll
            Length    Maximale Lnge, die fr das Ergebnis verfgbar ist
  OUTPUT:   (pchar)   Kopie des Adresszeigers
  
  Bsp.:     char   Result[256];
            ...
            printf (TOOLGetFileName("$(HOME)/data/foo", Result, 256));
              liefert "foo"
  ----------------------------------------------------------------------*/
char *TOOLGetFileName (const char *FileName, char *Result, int Length)
{
  char *Pos = NULL;
  char  File[PATH_MAX+1];

  Result[0] = '\0';

  if (FileName == NULL)
    return NULL;

  TOOLParseFileName (FileName, File, PATH_MAX);

  if ((Pos = strrchr(File, '/')) == NULL)
    Pos = File;
  else
    Pos++;

  strncpy (Result, Pos, Length);
  Result[Length-1] = '\0';

  if (Result[0] == '\0')
    return NULL;

  return Result;
}


/* ********************************************************************** */
/* ** Extracts a filename's extension and returns it                   ** */
/* ** INPUT:  FileName  Complete filename (may use sytem-variables)    ** */
/* **         Result    Pointer to result-memory (needs to be          ** */
/* **                   allocated.                                     ** */
/* **         Lenght    Size of allocated memory-area                  ** */
/* ** OUPUT:  (char *)  Copy of memory-pointer or NULL if invalid      ** */
/* **                   fileName                                       ** */
/* ** EXAMPLE: char Extension[40];                                     ** */
/* **          ...                                                     ** */
/* **          if (TOOLGetFileExt("$HOME/test.doc",                    ** */
/* **                             Extension, 40) == NULL)              ** */
/* **            ...                                                   ** */
/* ********************************************************************** */
char *TOOLGetFileExt (const char *FileName, char *Result, int Length)
{
  char File[PATH_MAX+1];
  char *Pos = NULL;

  if (FileName == NULL)
    return NULL;

  TOOLParseFileName (FileName, File, PATH_MAX);

  if ((Pos = strrchr(File, '.')) == NULL)
    strncpy (Result, File, Length);
  else
    strncpy (Result, Pos+1, Length);
  Result[Length-1] = '\0';

  if (Result[0] == '\0')
    return NULL;

  return Result;
}


/* ********************************************************************** */
/* ** Checks, if the given filename complies to the specified pattern  ** */
/* ** which might contain wildcard (* or ?). Both strings should       ** */
/* ** contain a path or not                                            ** */
/* ** INPUT:    Pattern   pattern string                               ** */
/* **           FileName  name of the file to be testet                ** */
/* ** OUTPUT:   (dual)    TRUE, if Filename matches the pattern        ** */
/* **                     FALSE, if not.                               ** */
/* ********************************************************************** */
dual TOOLTestPattern (const char *Pattern, const char *FileName)
{
  int  count1 = 0, count2 = 0;
  int  len;
  char Part[PATH_MAX+1];

  if (Pattern == NULL)
    return TRUE;
  elsif (FileName == NULL)
    return FALSE;

  while (Pattern[count1] != '\0' && FileName[count2] != '\0')
    if (Pattern[count1] == '*')                 /* '*'-wildcard found */
    {
      count1++;

      /* Get next subpattern without a wildcard */
      len = 0;
      while (Pattern[count1+len] != '\0' &&
	     Pattern[count1+len] != '*' && Pattern[count1+len] != '?' && 
	     len < PATH_MAX)
      {
	Part[len] = Pattern[count1+len];
	len++;
      };
      Part[len] = '\0';

      while (FileName[count2] != '\0' && (Part[0] =='\0' ||
	     strncmp(FileName+count2, Part, len) != 0))
	count2++;
    }
    elsif (Pattern[count1] == '?' || FileName[count2] == Pattern[count1])
    {
      count1++;
      count2++;
    }
    else
      return FALSE;

  return (Pattern[count1] == '\0' && FileName[count2] == '\0');
}


