#include <math.h>

#include <Xm/Form.h>

#include "tuxCAD.h"


#define QUADER_PAGE    2100
#define CONE_PAGE      2200
#define SPHERE_PAGE    2300
#define PIPE_PAGE      2400
#define WEDGE_PAGE     2500
#define TORUS_PAGE     2600
#define TON_PAGE       2700
#define SLICE_PAGE     2800
#define CYLINDER_PAGE  2900
#define SWEEP_PAGE     3000
#define ROT_PAGE       3100
#define ICO_PAGE       3200
#define PRISM_PAGE     3300
#define SCREW_PAGE     3400
#define SPRING_PAGE    3500
#define GEAR_PAGE      3600
#define ARROW_PAGE     3700

#define MATERIAL_PAGE  5000
#define COORD_PAGE     6000


typedef struct
{
  double SizeX;                                   // Size in x-direction
  double SizeY;                                   // Size in y-direction
  double SizeZ;                                   // Size in z-direction
} QUADER_INFO;

typedef struct
{
  double Radius;                                  // outer radius
  double SegmentX;                                // segment around x-axis
  double SegmentZ;                                // circular range (0..360)
  int    Facettes;                                // accuracy (# of facettes)
} SPHERE_INFO;

typedef struct
{ 
  double Radius1;                                 // inner radius
  double Radius2;                                 // outer base radius
  double Radius3;                                 // outer top radius
  double Thickness1;                              // upper thickness of pipe
  double Thickness2;                              // lower thickness of pipe
  double Height;                                  // height
  double Segment;                                 // circular range (0..360)
  int    Facettes;                                // number of facettes
} CONE_INFO, PIPE_INFO;

typedef struct
{
  double Radius1;                                 // radius of circle
  double Radius2;                                 // radius of ring
  int    Facettes1;                               // radius around ring
  int    Facettes2;                               // radius around circle
  double Segment;                                 // circular range (0..360)
} TORUS_INFO;

typedef struct
{
  double Radius1;                                 // inner-radius
  double Radius2;                                 // end-radius 
  double Radius3;                                 // middle radius
  double Height;                                  // height of (hyper-)bolid
  int    Facettes1;                               // shape accuracy
  int    Facettes2;                               // facettes around the obj
  double Segment;                                 // circurar range (0..360)
} BOLID_INFO;


typedef struct
{
  double Width1;                                  // base width
  double Width2;                                  // top width
  double Height1;                                 // base height
  double Height2;                                 // top height
  double Depth1;                                  // base depth
  double Depth2;                                  // top depth
} WEDGE_INFO;

typedef struct
{
  double   Length;                                // length of extrusion
  int      Num;                     
  POINT2D *Points;
} SWEEP_INFO;

typedef struct
{
  double   Radius;                                // outer radius
  double   Segment;                               // circular range (0..360)
  int      Facettes;                              // accuracy (# of facettes)
  int      Num;
  POINT2D *Points;
} ROT_INFO;

typedef struct
{
  double Radius;
  double Depth;
  double Segment;
  int    Facettes;
} CYLINDER_INFO;

typedef struct
{
  double Radius1;
  double Radius2;
  bool   Star;
  int    Density;
  int    Iterations;
} ICO_INFO;

typedef struct
{
  double Width1;
  double Width2;
  double OffsX;
  double Depth1;
  double Depth2;
  double OffsY;
  double Height1;
  double Height2;
  double OffsZ;
} PRISM_INFO;

typedef struct
{
  double Radius1;
  double Radius2;
  double Height;
  double Increase;
  double Offset;
  int    Facettes;
  int    Thread;
} SCREW_INFO;

typedef struct
{
  double Radius1;
  double Radius2;
  double Thickness;
  double Height;
  double Increase;
  int    Facettes1;
  int    Facettes2;
  int    Thread;
} SPRING_INFO;

typedef struct
{  
  double Radius1;
  double Radius2;
  double Radius3;
  double Depth1;
  double Depth2;
  int    Teeth;
  double Skew;
} GEAR_INFO;

typedef struct
{
  double Radius1;
  double Len1;
  double Radius2;
  double Len2;
  double Angle;
  double Segment;
  int    Facettes;
} ARROW_INFO;


typedef struct
{
  int            Amount;
  H_MATRIX       Offset;
  MATERIAL       Material;
  int            Page;
  QUADER_INFO    Quader;
  CONE_INFO      Cone;
  SPHERE_INFO    Sphere;
  PIPE_INFO      Pipe;
  TORUS_INFO     Torus;
  BOLID_INFO     Bolid;
  BOLID_INFO     Slice;
  WEDGE_INFO     Wedge;
  SWEEP_INFO     Sweep;
  ROT_INFO       Rot;
  CYLINDER_INFO  Cylinder;
  ICO_INFO       Ico;
  PRISM_INFO     Prism;
  SCREW_INFO     Screw;
  SPRING_INFO    Spring;
  GEAR_INFO      Gear;
  ARROW_INFO     Arrow;
  int            PosX, PosY;
  char           DevName[80];
  char           PartName[80];
} ADD_INFO;


#if 1
/* Default values: */
static ADD_INFO Add =
{
  1,                                                    // Amount to create
  Matrix_1,
  MATERIAL (COLOR::GOLD),                               // Material
  3,                                                    // Object-Page (3..15)
  {30.0,  60.0,  30.0                            },     // Quader
  { 0.0,  30.0,   0.0, 0.0,  0.0, 50.0, 360.0, 16},     // Cone
  {30.0, 180.0, 360.0, 16                        },     // Sphere
  { 0.0,  30.0,  40.0, 5.0, 25.0, 50.0, 360.0, 16},     // Pipe
  {30.0,  10.0,  16,   24,   360.0               },     // Torus
  { 0.0,  20.0,  25.0, 60.0,  12, 24, 360.0      },     // Hyperbolid/Ton
  { 0.0,  10.0,  20.0, 12.0,   8, 24, 360.0      },     // Slice
  {30.0,  10.0,   0.0, 50.0,   5.0, 15.0         },     // Wedge
  {40.0,            0,     NULL                  },     // Sweep
  { 0.0, 100.0,  16,    0, NULL                  },     // Rot
  {30.0,  10.0,  45.0, 16,                       },     // Cylinder
  {30.0,  50.0,  TRUE,  2,    2                  },     // Icosahedron/Star
  {40.0,  40.0,  30.0,
   10.0,  10.0,   0.0,
    0.0,  30.0,   0.0                            },     // Prism
  { 8.0,   8.0,  25.0,  2.0,   1.5,   24,    1   },     // Screw
  {20.0,  20.0,   2.0, 40.0,   8.0,   32,   16, 1},     // Spring
  { 5.0,  25.0,  30.0,  5.0,   3.75,  16,   15   },     // Gear
  { 1.0,   6.0,   1.5,  3.0, 135.0,  360.0, 16   },     // Arrow
  -1, -1,                                               // Dlg position
  "device", "part",                                     // Dev-/Partname
};
#else
static ADD_INFO Add;
#endif


static void SetFreeNames (WORLD *World)
{
  int       count = 0;
  DEVICE3D *Dev;

  if ((Dev = World->GetDev(Add.DevName)) == NULL)
    strcpy (Add.PartName, "part");
  else
    do
    {
      count++;
      sprintf (Add.PartName, "part%03d", count);
    } while (World->GetPart(Add.DevName, Add.PartName) != NULL);
}


int Editor2d (DIALOG *Dlg, int BoxID)
{
  CAD_EDITOR *Editor;

  DLGSetButtonActive (Dlg, 0, FALSE);
  DLGSetBoxActive    (Dlg, 0, FALSE);

  Editor = new CAD_EDITOR ("CAD-Editor", 600, 600);

  switch (BoxID)
  {
    case 3020 :
      if (Add.Sweep.Points != NULL)
      {
	free (Add.Sweep.Points);
	Add.Sweep.Points = NULL;
      }

      if ((Add.Sweep.Num = Editor->GetNum()) > 0)
      {
	Add.Sweep.Points = (POINT2D *)malloc(sizeof(POINT2D) * Add.Sweep.Num);
	memcpy (Add.Sweep.Points, Editor->GetPoints(), 
		sizeof(POINT2D) * Add.Sweep.Num);
      }
      break;

    case 3140 :
      if (Add.Rot.Points != NULL)
      {
	free (Add.Rot.Points);
	Add.Rot.Points = NULL;
      }

      if ((Add.Rot.Num = Editor->GetNum()) > 0)
      {
	Add.Rot.Points = (POINT2D *)malloc(sizeof(POINT2D) * Add.Rot.Num);
	memcpy (Add.Rot.Points, Editor->GetPoints(), 
		sizeof(POINT2D) * Add.Rot.Num);
      }
      break;
  }

  delete Editor;

  DLGSetBoxActive    (Dlg, 0, TRUE);
  DLGSetButtonActive (Dlg, 0, TRUE);

  return DlgCONFIRM;
}



static void PreviewHandle (Widget widget, CREATE_DIALOG *Dlg, XEvent *Event)
{
  // printf ("EventType: %d\n", Event->type);

  if (Event->type == MapNotify)
    Dlg->MakePreview (widget);
}


static Widget PreviewWin (DIALOG *Dlg, Widget Parent, void *Param)
{
  Widget Form = XtVaCreateManagedWidget ("preview", xmFormWidgetClass, Parent, 
					            XmNwidth,  150,
					            XmNheight, 150, 
				                    NULL);
  XtAddEventHandler (Form, StructureNotifyMask, FALSE, 
		           (XtEventHandler)PreviewHandle, Param);
  return Form;
}


void CREATE_DIALOG::MakePreview (Widget widget)
{
  if (Preview == NULL)
  {
    // printf ("MAKE PREVIEW\n");
    Preview = new WORKCELL (widget, "preview");
    Demo = new SPHERE (Preview, "sphere", GEOMakeRot(Vector_Y, -90.0), 
		       10.0, 180.0, 180.0, 32, MATERIAL(Add.Material));
    Preview->AddDevice (new DEVICE3D("preview", Demo));
    Preview->Fill      ();
    Preview->SetScale  (0.5);
  }
}


int CREATE_ChangeCB (DIALOG *Dlg, int BoxID)
{
  CREATE_DIALOG *CreateDlg = (CREATE_DIALOG *)DLGGetUserData (Dlg, 0);

  if (CreateDlg != NULL)
    CreateDlg->ChangeCB (BoxID);

  return DlgACCEPT;
}


CREATE_DIALOG::CREATE_DIALOG (WORLD *World) :
  CAD_DIALOG (),
  World      (World),
  Part       (NULL),
  Preview    (NULL),
  Demo       (NULL),
  Handle     (NULL)
{
  PRO_METHOD ("CREATE_DIALOG", "")

  assert (World != NULL);

  PRO_END
}


bool CREATE_DIALOG::Open ()
{
#if 1
  int       count;
  char      PartName[255];
  H_MATRIX  Matrix = Matrix_1;
#endif
  char     *Name;
  DEVICE3D *Dev;

  PRO_METHOD ("CREATE_DIALOG", "\n")

  if (! CAD_DIALOG::Open())
    PRO_RETURN ("Opening dialog failed", FALSE)

  DLGMakeCols (Dlg, 0, 2, DlgETCHED_IN, 10, 50);

  DLGInsertNote (Dlg, 10, "object", DlgNOBINDING);

    DLGAddPage (Dlg, 10, DlgMAJOR, "global", 1000);
      // TODO: nicht existierenden Name erzeugen
      DLGMakeRows   (Dlg, 1000, 3, DlgNO_LINE, 1010, 1020, 1030);
      SetFreeNames (World);
      DLGInsertText (Dlg, 1010, "devName",  80, Add.DevName,  25, 1);
      DLGInsertText (Dlg, 1020, "partName", 80, Add.PartName, 25, 1);
      DLGInsertUpDn (Dlg, 1030, "amount",   1, 1000, Add.Amount, 1);

      // TODO: Bezugskoordinatensystem-Eingabe / Position + Orientierung
      //       Offset fr Duplikate (Rot/Trans)

    DLGAddPage (Dlg, 10, DlgMAJOR, "object", 2000);

    DLGAddPage (Dlg, 10, DlgMINOR, "quader", QUADER_PAGE);
      DLGMakeRows (Dlg, QUADER_PAGE, 3, DlgNO_LINE, 
	                QUADER_PAGE+10, QUADER_PAGE+20, QUADER_PAGE+30);
      DLGInsertDouble (Dlg, QUADER_PAGE+10, "sizeX", Add.Quader.SizeX, 8, 3);
      DLGInsertDouble (Dlg, QUADER_PAGE+20, "sizeY", Add.Quader.SizeY, 8, 3);
      DLGInsertDouble (Dlg, QUADER_PAGE+30, "sizeZ", Add.Quader.SizeZ, 8, 3);

    DLGAddPage (Dlg, 10, DlgMINOR, "cone", CONE_PAGE);
      DLGMakeRows (Dlg, CONE_PAGE, 6, DlgNO_LINE, 
	                CONE_PAGE+10, CONE_PAGE+20, CONE_PAGE+30,
	                CONE_PAGE+40, CONE_PAGE+50, CONE_PAGE+60);
      DLGInsertDouble (Dlg, CONE_PAGE+10, "radius1",  Add.Cone.Radius1, 8, 3);
      DLGInsertDouble (Dlg, CONE_PAGE+20, "radius2",  Add.Cone.Radius2, 8, 3);
      DLGInsertDouble (Dlg, CONE_PAGE+30, "radius3",  Add.Cone.Radius3, 8, 3);
      DLGInsertDouble (Dlg, CONE_PAGE+40, "height",   Add.Cone.Height,  8, 3);
      DLGInsertScale  (Dlg, CONE_PAGE+50, "segment",  0.0, 360.0, 
		                          Add.Cone.Segment, 2, DlgHORIZONTAL);
      DLGInsertUpDn   (Dlg, CONE_PAGE+60, "facettes", 3, 256, 
		                          Add.Cone.Facettes, 1);

    DLGAddPage (Dlg, 10, DlgMINOR, "sphere", SPHERE_PAGE);
      DLGMakeRows (Dlg, SPHERE_PAGE, 4, DlgNO_LINE, 
	                SPHERE_PAGE+10, SPHERE_PAGE+20, SPHERE_PAGE+30,
	                SPHERE_PAGE+40);
      DLGInsertDouble (Dlg, SPHERE_PAGE+10, "radius", Add.Sphere.Radius, 8, 3);
      DLGInsertScale  (Dlg, SPHERE_PAGE+20, "segmentX", 0.0, 180.0, 
		                            Add.Sphere.SegmentX, 2, 
		                            DlgHORIZONTAL);
      DLGInsertScale  (Dlg, SPHERE_PAGE+30, "segmentZ", 0.0, 360.0, 
		                            Add.Sphere.SegmentZ, 2, 
		                            DlgHORIZONTAL);
      DLGInsertUpDn   (Dlg, SPHERE_PAGE+40, "facettes", 3, 256, 
		                            Add.Sphere.Facettes, 1);

    DLGAddPage (Dlg, 10, DlgMINOR, "pipe", PIPE_PAGE);
      DLGMakeRows (Dlg, PIPE_PAGE, 7, DlgNO_LINE, 
	                PIPE_PAGE+10, PIPE_PAGE+20, PIPE_PAGE+30,
	                PIPE_PAGE+40, PIPE_PAGE+50, PIPE_PAGE+60,
	                PIPE_PAGE+70);
      DLGInsertDouble (Dlg, PIPE_PAGE+10, "radius1", Add.Pipe.Radius2, 8, 3);
      DLGInsertDouble (Dlg, PIPE_PAGE+20, "thickness1", 
		                          Add.Pipe.Thickness1, 8, 3);
      DLGInsertDouble (Dlg, PIPE_PAGE+30, "radius2", Add.Pipe.Radius3, 8, 3);
      DLGInsertDouble (Dlg, PIPE_PAGE+40, "thickness2", 
		                          Add.Pipe.Thickness2, 8, 3);
      DLGInsertDouble (Dlg, PIPE_PAGE+50, "length", Add.Pipe.Height,   8, 3);
      DLGInsertScale  (Dlg, PIPE_PAGE+60, "segment", 0.0, 360.0, 
		                          Add.Pipe.Segment, 2, DlgHORIZONTAL);
      DLGInsertUpDn   (Dlg, PIPE_PAGE+70, "facettes", 3, 256, 
		                          Add.Pipe.Facettes, 1);

    DLGAddPage (Dlg, 10, DlgMINOR, "wedge", WEDGE_PAGE);
      DLGMakeRows (Dlg, WEDGE_PAGE, 6, DlgNO_LINE, 
                        WEDGE_PAGE+10, WEDGE_PAGE+20, WEDGE_PAGE+30, 
	                WEDGE_PAGE+40, WEDGE_PAGE+50, WEDGE_PAGE+60);
      DLGInsertDouble (Dlg, WEDGE_PAGE+10, "witdh1",  Add.Wedge.Width1,  8, 3);
      DLGInsertDouble (Dlg, WEDGE_PAGE+20, "width2",  Add.Wedge.Width2,  8, 3);
      DLGInsertDouble (Dlg, WEDGE_PAGE+30, "height1", Add.Wedge.Height1, 8, 3);
      DLGInsertDouble (Dlg, WEDGE_PAGE+40, "height2", Add.Wedge.Height2, 8, 3);
      DLGInsertDouble (Dlg, WEDGE_PAGE+50, "depth1",  Add.Wedge.Depth1,  8, 3);
      DLGInsertDouble (Dlg, WEDGE_PAGE+60, "depth2",  Add.Wedge.Depth2,  8, 3);

    DLGAddPage (Dlg, 10, DlgMINOR, "torus", TORUS_PAGE);
      DLGMakeRows (Dlg, TORUS_PAGE, 5, DlgNO_LINE, 
	                TORUS_PAGE+10, TORUS_PAGE+20, TORUS_PAGE+30, 
	                TORUS_PAGE+40, TORUS_PAGE+50);
      DLGInsertDouble (Dlg, TORUS_PAGE+10, "radius1", Add.Torus.Radius1, 8, 3);
      DLGInsertDouble (Dlg, TORUS_PAGE+20, "radius2", Add.Torus.Radius2, 8, 3);
      DLGInsertUpDn   (Dlg, TORUS_PAGE+30, "facettes1", 3, 256, 
		                           Add.Torus.Facettes1, 1);
      DLGInsertUpDn   (Dlg, TORUS_PAGE+40, "facettes2", 3, 256, 
		                           Add.Torus.Facettes2, 1);
      DLGInsertScale  (Dlg, TORUS_PAGE+50, "segment", 0.0, 360.0, 
		                           Add.Torus.Segment, 2, 
		                           DlgHORIZONTAL);

    DLGAddPage (Dlg, 10, DlgMINOR, "ton", TON_PAGE);
      DLGMakeRows (Dlg, TON_PAGE, 7, DlgNO_LINE, 
	                TON_PAGE+10, TON_PAGE+20, TON_PAGE+30, 
	                TON_PAGE+40, TON_PAGE+50, TON_PAGE+60, TON_PAGE+70);
      DLGInsertDouble (Dlg, TON_PAGE+10, "radius1", Add.Bolid.Radius1, 8, 3);
      DLGInsertDouble (Dlg, TON_PAGE+20, "radius2", Add.Bolid.Radius2, 8, 3);
      DLGInsertDouble (Dlg, TON_PAGE+30, "radius3", Add.Bolid.Radius3, 8, 3);
      DLGInsertDouble (Dlg, TON_PAGE+40, "height",  Add.Bolid.Height, 8, 3);
      DLGInsertUpDn   (Dlg, TON_PAGE+50, "facettes1", 3, 256, 
		                                      Add.Bolid.Facettes1, 1);
      DLGInsertUpDn   (Dlg, TON_PAGE+60, "facettes2", 3, 256, 
		                                      Add.Bolid.Facettes2, 1);
      DLGInsertScale  (Dlg, TON_PAGE+70, "segment",   0.0, 360.0, 
		                                      Add.Bolid.Segment, 2, 
		                                      DlgHORIZONTAL);

    DLGAddPage (Dlg, 10, DlgMINOR, "slice", SLICE_PAGE);
      DLGMakeRows (Dlg, SLICE_PAGE, 7, DlgNO_LINE, 
	                SLICE_PAGE+10, SLICE_PAGE+20, SLICE_PAGE+30, 
	                SLICE_PAGE+40, SLICE_PAGE+50, SLICE_PAGE+60, 
	                SLICE_PAGE+70);
      DLGInsertDouble (Dlg, SLICE_PAGE+10, "radius1", Add.Slice.Radius1, 8, 3);
      DLGInsertDouble (Dlg, SLICE_PAGE+20, "radius2", Add.Slice.Radius2, 8, 3);
      DLGInsertDouble (Dlg, SLICE_PAGE+30, "radius3", Add.Slice.Radius3, 8, 3);
      DLGInsertDouble (Dlg, SLICE_PAGE+40, "height",  Add.Slice.Height, 8, 3);
      DLGInsertUpDn   (Dlg, SLICE_PAGE+50, "facettes1", 3, 256, 
		                           Add.Slice.Facettes1, 1);
      DLGInsertUpDn   (Dlg, SLICE_PAGE+60, "facettes2", 3, 256, 
		                           Add.Slice.Facettes2, 1);
      DLGInsertScale  (Dlg, SLICE_PAGE+70, "segment",   0.0, 360.0, 
		                           Add.Slice.Segment, 2, 
		                           DlgHORIZONTAL);

    DLGAddPage (Dlg, 10, DlgMINOR, "cylinder", CYLINDER_PAGE);
      DLGMakeRows (Dlg, CYLINDER_PAGE, 4, DlgNO_LINE, 
	                CYLINDER_PAGE+10, CYLINDER_PAGE+20, 
	                CYLINDER_PAGE+30, CYLINDER_PAGE+40);
      DLGInsertDouble (Dlg, CYLINDER_PAGE+10, "radius", 
		                              Add.Cylinder.Radius, 8, 3);    
      DLGInsertDouble (Dlg, CYLINDER_PAGE+20, "depth1",
		                              Add.Cylinder.Depth,  8, 3);
      DLGInsertScale  (Dlg, CYLINDER_PAGE+30, "segment", 0.0, 360.0, 
		                              Add.Cylinder.Segment, 2, 
		                              DlgHORIZONTAL);
      DLGInsertUpDn   (Dlg, CYLINDER_PAGE+40, "facettes", 3, 256, 
		                              Add.Cylinder.Facettes, 1);

    DLGAddPage (Dlg, 10, DlgMINOR, "sweep", SWEEP_PAGE);
      DLGMakeRows (Dlg, SWEEP_PAGE, 2, DlgNO_LINE, 
		        SWEEP_PAGE+10, SWEEP_PAGE+20);
      DLGInsertDouble (Dlg, SWEEP_PAGE+10, "length", Add.Sweep.Length, 8, 3);
      DLGInsertProc   (Dlg, SWEEP_PAGE+20, "editor", Editor2d);

    DLGAddPage (Dlg, 10, DlgMINOR, "rot", ROT_PAGE);
      DLGMakeRows (Dlg, ROT_PAGE, 4, DlgNO_LINE, 
	                ROT_PAGE+10, ROT_PAGE+20, ROT_PAGE+30, ROT_PAGE+40);
      DLGInsertDouble (Dlg, ROT_PAGE+10, "radius",  Add.Rot.Radius, 8, 3);
      DLGInsertScale  (Dlg, ROT_PAGE+20, "segment", 0.0, 360.0, 
		                         Add.Rot.Segment, 2, DlgHORIZONTAL);
      DLGInsertUpDn   (Dlg, ROT_PAGE+30, "facettes", 3, 256, 
		                         Add.Rot.Facettes, 1);
      DLGInsertProc   (Dlg, ROT_PAGE+40, "editor", Editor2d);

    DLGAddPage (Dlg, 10, DlgMINOR, "icosahedron", ICO_PAGE);
      DLGMakeRows (Dlg, ICO_PAGE, 5, DlgNO_LINE, 
	                ICO_PAGE+10, ICO_PAGE+20, ICO_PAGE+30, 
	                ICO_PAGE+40, ICO_PAGE+50);
      DLGInsertDouble (Dlg, ICO_PAGE+10, "radius1",   Add.Ico.Radius1, 8, 3);
      DLGInsertUpDn   (Dlg, ICO_PAGE+20, "iterations", 0, 10, 
		                         Add.Ico.Iterations, 1);
      DLGInsertBool   (Dlg, ICO_PAGE+30, "star",    Add.Ico.Star);
      DLGInsertDouble (Dlg, ICO_PAGE+40, "radius2", Add.Ico.Radius2, 8, 3);
      DLGInsertUpDn   (Dlg, ICO_PAGE+50, "density", 1, 10, Add.Ico.Density, 1);

    DLGAddPage (Dlg, 10, DlgMINOR, "prism", PRISM_PAGE);
      DLGMakeRows (Dlg, PRISM_PAGE, 9, DlgNO_LINE, 
		        PRISM_PAGE+10, PRISM_PAGE+20, PRISM_PAGE+30, 
		        PRISM_PAGE+40, PRISM_PAGE+50, PRISM_PAGE+60, 
		        PRISM_PAGE+70, PRISM_PAGE+80, PRISM_PAGE+90);
      DLGInsertDouble (Dlg, PRISM_PAGE+10, "width1",  Add.Prism.Width1,  8, 3);
      DLGInsertDouble (Dlg, PRISM_PAGE+20, "width2",  Add.Prism.Width2,  8, 3);
      DLGInsertDouble (Dlg, PRISM_PAGE+30, "offsetX", Add.Prism.OffsX,   8, 3);
      DLGInsertDouble (Dlg, PRISM_PAGE+40, "depth1",  Add.Prism.Depth1,  8, 3);
      DLGInsertDouble (Dlg, PRISM_PAGE+50, "depth2",  Add.Prism.Depth2,  8, 3);
      DLGInsertDouble (Dlg, PRISM_PAGE+60, "offsetY", Add.Prism.OffsY,   8, 3);
      DLGInsertDouble (Dlg, PRISM_PAGE+70, "height1", Add.Prism.Height1, 8, 3);
      DLGInsertDouble (Dlg, PRISM_PAGE+80, "height2", Add.Prism.Height2, 8, 3);
      DLGInsertDouble (Dlg, PRISM_PAGE+90, "offsetZ", Add.Prism.OffsZ,   8, 3);

    DLGAddPage (Dlg, 10, DlgMINOR, "screw", SCREW_PAGE);
      DLGMakeRows (Dlg, SCREW_PAGE, 7, DlgNO_LINE, 
		        SCREW_PAGE+10, SCREW_PAGE+20, SCREW_PAGE+30, 
		        SCREW_PAGE+40, SCREW_PAGE+50, SCREW_PAGE+60, 
		        SCREW_PAGE+70);
      DLGInsertDouble (Dlg, SCREW_PAGE+10, "radius1", Add.Screw.Radius1, 8, 3);
      DLGInsertDouble (Dlg, SCREW_PAGE+20, "radius2", Add.Screw.Radius2, 8, 3);
      DLGInsertDouble (Dlg, SCREW_PAGE+30, "height",  Add.Screw.Height,  8, 3);
      DLGInsertDouble (Dlg, SCREW_PAGE+40, "increase", 
		                           Add.Screw.Increase, 8, 3);
      DLGInsertDouble (Dlg, SCREW_PAGE+50, "offset",  Add.Screw.Offset,  8, 3);
      DLGInsertUpDn   (Dlg, SCREW_PAGE+60, "facettes", 3, 256, 
		                           Add.Screw.Facettes, 1);
      DLGInsertOMenu  (Dlg, SCREW_PAGE+70, "thread", 2, "left", "right", 
                                           Add.Screw.Thread);

    DLGAddPage (Dlg, 10, DlgMINOR, "spring", SPRING_PAGE);
      DLGMakeRows (Dlg, SPRING_PAGE, 8, DlgNO_LINE, 
		        SPRING_PAGE+10, SPRING_PAGE+20, SPRING_PAGE+30, 
		        SPRING_PAGE+40, SPRING_PAGE+50, SPRING_PAGE+60, 
		        SPRING_PAGE+70, SPRING_PAGE+80);
      DLGInsertDouble (Dlg, SPRING_PAGE+10, "radius1", 
		                            Add.Spring.Radius1, 8, 3);
      DLGInsertDouble (Dlg, SPRING_PAGE+20, "radius2", 
		                            Add.Spring.Radius2, 8, 3);
      DLGInsertDouble (Dlg, SPRING_PAGE+30, "thickness", 
		                            Add.Spring.Thickness,  8, 3);
      DLGInsertDouble (Dlg, SPRING_PAGE+40, "height", Add.Spring.Height, 8, 3);
      DLGInsertDouble (Dlg, SPRING_PAGE+50, "increase", 
		                            Add.Spring.Increase, 8, 3);
      DLGInsertUpDn   (Dlg, SPRING_PAGE+60, "facettes1", 3, 256, 
		                            Add.Spring.Facettes1, 1);
      DLGInsertUpDn   (Dlg, SPRING_PAGE+70, "facettes2", 3, 256, 
		                            Add.Spring.Facettes2, 1);
      DLGInsertOMenu  (Dlg, SPRING_PAGE+80, "thread", 2, "left", "right", 
                                            Add.Spring.Thread);
     
    DLGAddPage (Dlg, 10, DlgMINOR, "gear", GEAR_PAGE);
      DLGMakeRows (Dlg, GEAR_PAGE, 7, DlgNO_LINE, 
		        GEAR_PAGE+10, GEAR_PAGE+20, GEAR_PAGE+30, 
		        GEAR_PAGE+40, GEAR_PAGE+50, GEAR_PAGE+60, 
		        GEAR_PAGE+70);
      DLGInsertDouble (Dlg, GEAR_PAGE+10, "radius1", Add.Gear.Radius1, 8, 3);
      DLGInsertDouble (Dlg, GEAR_PAGE+20, "radius2", Add.Gear.Radius2, 8, 3);
      DLGInsertDouble (Dlg, GEAR_PAGE+30, "radius3", Add.Gear.Radius3, 8, 3);
      DLGInsertDouble (Dlg, GEAR_PAGE+40, "depth1",  Add.Gear.Depth1,  8, 3);
      DLGInsertDouble (Dlg, GEAR_PAGE+50, "depth2",  Add.Gear.Depth2,  8, 3);
      DLGInsertUpDn   (Dlg, GEAR_PAGE+60, "teeth",  4, 256, Add.Gear.Teeth, 1);
      DLGInsertScale  (Dlg, GEAR_PAGE+70, "skew",   0.0, 45.0, Add.Gear.Skew,
		                          1, DlgHORIZONTAL);

    DLGAddPage (Dlg, 10, DlgMINOR, "arrow", ARROW_PAGE);
      DLGMakeRows (Dlg, ARROW_PAGE, 7, DlgNO_LINE, 
		        ARROW_PAGE+10, ARROW_PAGE+20, ARROW_PAGE+30, 
		        ARROW_PAGE+40, ARROW_PAGE+50, ARROW_PAGE+60, 
		        ARROW_PAGE+70);
      DLGInsertDouble (Dlg, ARROW_PAGE+10, "radius1", Add.Arrow.Radius1, 8, 3);
      DLGInsertDouble (Dlg, ARROW_PAGE+20, "length",  Add.Arrow.Len1,    8, 3);
      DLGInsertDouble (Dlg, ARROW_PAGE+30, "radius2", Add.Arrow.Radius2, 8, 3);
      DLGInsertDouble (Dlg, ARROW_PAGE+40, "depth",   Add.Arrow.Len2,    8, 3);
      DLGInsertScale  (Dlg, ARROW_PAGE+50, "angle", 10.0, 170.0, 
		                           Add.Arrow.Angle, 1, DlgHORIZONTAL);
      DLGInsertScale  (Dlg, ARROW_PAGE+60, "segment", 0.0, 360.0, 
		                           Add.Arrow.Segment, 2, 
		                           DlgHORIZONTAL);
      DLGInsertUpDn   (Dlg, ARROW_PAGE+70, "facettes", 3, 256, 
		                           Add.Arrow.Facettes, 1);

    /*
       weitere Objektkrper: Oktaeder,
			     Prismoid,
			     Zylinderhuf,
                             schiefgeschn. Kreiszyl, 
			     Elipsoid, 
			     Rotationsparabolid, 
			     Text, ...
    */

    DLGAddPage (Dlg, 10, DlgMAJOR, "material", MATERIAL_PAGE);
      MaterialDialog (Dlg, MATERIAL_PAGE, &Add.Material, 
		                           ALPHA_MASK | TEX_MASK);

    DLGAddPage (Dlg, 10, DlgMAJOR, "coordinates", COORD_PAGE);

    DLGSetBorder    (Dlg, 50, DlgETCHED_IN);
    DLGInsertWidget (Dlg, 50, PreviewWin, (void *)this);

  // TODO: 
  //  Icon mit Zeichnung der Objekt-Parameter bzw. Vorschau-Fenster

  DLGSetUserData     (Dlg, 0, this);
  DLGSetFunction     (Dlg, 0, CREATE_ChangeCB);
  DLGSetHelpFunction (Dlg, HelpDlgCB);

  DLGMakeButton (Dlg, DEFAULT_BUTTON);
  DLGMakeButton (Dlg, HELP_BUTTON);
  DLGMakeButton (Dlg, OK_BUTTON);

  DLGOpenDialog (Dlg, "createPart", DlgNORMAL, NULL);

  /* 
     TODO:
     Constrains prfen:
     Segment: 0..360, Bolid: r1 < r2;r3, ...
  */
  if (DLGGetExitButton(Dlg) == DlgOK)
  {
    if ((Name = DLGGetText(Dlg, 1010)) != NULL)
    {
      strncpy (Add.DevName, Name, 80);
      Add.DevName[79] = '\0';
    }
    if ((Name = DLGGetText(Dlg, 1020)) != NULL)
    {
      strncpy (Add.PartName, Name, 80);
      Add.PartName[79] = '\0';
    }

    World->GetWin()->Select ();

    Add.Amount = (int)DLGGetUpDn (Dlg, 1030);
#if 1
    Add.Offset = GEORPY2Frame (0.0, 0.0, 0.0, 0.0, 0.0, 360.0/Add.Amount);
#else
    Add.Offset = GEOMakeTrans (10.0, 10.0, 10.0);
#endif

    for (count = 0; count < Add.Amount; ++count)
    {
      if (Add.Amount > 1)
      {
	sprintf (PartName, "%s_%d", Add.PartName, count+1);
	Matrix = GEOMatrixProd (Matrix, Add.Offset);
      }
      else
        strcpy (PartName, Add.PartName);

      Part = NULL;
      switch (Add.Page-2)
      {
        case 1 :
	  PRO0 ("BLOCK\n")
	  Add.Quader.SizeX = DLGGetDouble (Dlg, QUADER_PAGE+10);
	  Add.Quader.SizeY = DLGGetDouble (Dlg, QUADER_PAGE+20);
	  Add.Quader.SizeZ = DLGGetDouble (Dlg, QUADER_PAGE+30);
	  Part = new QUADER (World, PartName, Matrix, 
			     Add.Quader.SizeX, Add.Quader.SizeY, 
			     Add.Quader.SizeZ, Add.Material);
	  break;

        case 2 :
	  PRO0 ("CONE\n")
	  Add.Cone.Radius1  =      DLGGetDouble (Dlg, CONE_PAGE+10);
	  Add.Cone.Radius2  =      DLGGetDouble (Dlg, CONE_PAGE+20);
	  Add.Cone.Radius3  =      DLGGetDouble (Dlg, CONE_PAGE+30);
	  Add.Cone.Height   =      DLGGetDouble (Dlg, CONE_PAGE+40);
	  Add.Cone.Segment  =      DLGGetScale  (Dlg, CONE_PAGE+50);
	  Add.Cone.Facettes = (int)DLGGetUpDn   (Dlg, CONE_PAGE+60);
	  Part = new CONE (World, PartName, Matrix,
			   Add.Cone.Radius1,  Add.Cone.Radius2,
			   Add.Cone.Radius3,
			   Add.Cone.Height,   Add.Cone.Segment,
			   Add.Cone.Facettes, Add.Material);
	  break;

        case 3 :
	  PRO0 ("SPHERE\n")
          Add.Sphere.Radius   =      DLGGetDouble (Dlg, SPHERE_PAGE+10);
	  Add.Sphere.SegmentX =      DLGGetScale  (Dlg, SPHERE_PAGE+20);
	  Add.Sphere.SegmentZ =      DLGGetScale  (Dlg, SPHERE_PAGE+30);
	  Add.Sphere.Facettes = (int)DLGGetUpDn   (Dlg, SPHERE_PAGE+40);
	  Part = new SPHERE (World, PartName, Matrix,
			     Add.Sphere.Radius, 
			     Add.Sphere.SegmentX, Add.Sphere.SegmentZ,
			     Add.Sphere.Facettes, Add.Material);
	  break;

        case 4 :
	  PRO0 ("PIPE\n")
	  Add.Pipe.Radius2    =      DLGGetDouble (Dlg, PIPE_PAGE+10);
	  Add.Pipe.Thickness1 =      DLGGetDouble (Dlg, PIPE_PAGE+20);
	  Add.Pipe.Radius3    =      DLGGetDouble (Dlg, PIPE_PAGE+30);
	  Add.Pipe.Thickness2 =      DLGGetDouble (Dlg, PIPE_PAGE+40);
	  Add.Pipe.Height     =      DLGGetDouble (Dlg, PIPE_PAGE+50);
	  Add.Pipe.Segment    =      DLGGetScale  (Dlg, PIPE_PAGE+60);
	  Add.Pipe.Facettes   = (int)DLGGetUpDn   (Dlg, PIPE_PAGE+70);
	  Part = new PIPE (World, PartName, Matrix,
			   Add.Pipe.Radius2,    Add.Pipe.Radius3,
			   Add.Pipe.Thickness1, Add.Pipe.Thickness2,
			   Add.Pipe.Height,     Add.Pipe.Segment, 
			   Add.Pipe.Facettes,   Add.Material);
	  break;

        case 5 :
	  PRO0 ("WEDGE\n")
	  Add.Wedge.Width1  = DLGGetDouble (Dlg, WEDGE_PAGE+10);
	  Add.Wedge.Width2  = DLGGetDouble (Dlg, WEDGE_PAGE+20);
	  Add.Wedge.Height1 = DLGGetDouble (Dlg, WEDGE_PAGE+30);
	  Add.Wedge.Height2 = DLGGetDouble (Dlg, WEDGE_PAGE+40);
	  Add.Wedge.Depth1  = DLGGetDouble (Dlg, WEDGE_PAGE+50);
	  Add.Wedge.Depth2  = DLGGetDouble (Dlg, WEDGE_PAGE+60);
	  Part = new WEDGE (World, PartName, Matrix,
			    Add.Wedge.Width1,  Add.Wedge.Width2,
			    Add.Wedge.Height1, Add.Wedge.Height2,  
			    Add.Wedge.Depth1,  Add.Wedge.Depth2, 
			    Add.Material);
	  break;

        case 6 :
	  PRO0 ("TORUS\n")
	  Add.Torus.Radius1   =      DLGGetDouble (Dlg, TORUS_PAGE+10);
	  Add.Torus.Radius2   =      DLGGetDouble (Dlg, TORUS_PAGE+20);
	  Add.Torus.Facettes1 = (int)DLGGetUpDn   (Dlg, TORUS_PAGE+30);
	  Add.Torus.Facettes2 = (int)DLGGetUpDn   (Dlg, TORUS_PAGE+40);
	  Add.Torus.Segment   =      DLGGetScale  (Dlg, TORUS_PAGE+50);
	  Part = new TORUS (World, PartName, Matrix,
			    Add.Torus.Radius1,   Add.Torus.Radius2,
			    Add.Torus.Facettes1, Add.Torus.Facettes2,
			    Add.Torus.Segment,   Add.Material);
	  break;

        case 7 : 
	  PRO0 ("TON/HYPERBOLID\n")
	  Add.Bolid.Radius1   =      DLGGetDouble (Dlg, TON_PAGE+10);
	  Add.Bolid.Radius2   =      DLGGetDouble (Dlg, TON_PAGE+20);
	  Add.Bolid.Radius3   =      DLGGetDouble (Dlg, TON_PAGE+30);
	  Add.Bolid.Height    =      DLGGetDouble (Dlg, TON_PAGE+40);
	  Add.Bolid.Facettes1 = (int)DLGGetUpDn   (Dlg, TON_PAGE+50);
	  Add.Bolid.Facettes2 = (int)DLGGetUpDn   (Dlg, TON_PAGE+60);
	  Add.Bolid.Segment   =      DLGGetScale  (Dlg, TON_PAGE+70);
	  Part = new BOLID (World, PartName, Matrix,
			    Add.Bolid.Radius1,   Add.Bolid.Radius2,
			    Add.Bolid.Radius3,
			    Add.Bolid.Height,    Add.Bolid.Facettes1,
			    Add.Bolid.Facettes2, Add.Bolid.Segment, 
			    Add.Material);
	  break;

        case 8 : 
	  PRO0 ("SLICE\n")
	  Add.Slice.Radius1   =      DLGGetDouble (Dlg, SLICE_PAGE+10);
	  Add.Slice.Radius2   =      DLGGetDouble (Dlg, SLICE_PAGE+20);
	  Add.Slice.Radius3   =      DLGGetDouble (Dlg, SLICE_PAGE+30);
	  Add.Slice.Height    =      DLGGetDouble (Dlg, SLICE_PAGE+40);
	  Add.Slice.Facettes1 = (int)DLGGetUpDn   (Dlg, SLICE_PAGE+50);
	  Add.Slice.Facettes2 = (int)DLGGetUpDn   (Dlg, SLICE_PAGE+60);
	  Add.Slice.Segment   =      DLGGetScale  (Dlg, SLICE_PAGE+70);
	  Part = new SLICE (World, PartName, Matrix,
			    Add.Slice.Radius1,   Add.Slice.Radius2,
			    Add.Slice.Radius3,   Add.Slice.Height, 
			    Add.Slice.Facettes1, Add.Slice.Facettes2,
			    Add.Slice.Segment,   Add.Material);
	  break;

        case 9 :
	  PRO0 ("CYLINDER\n")
          Add.Cylinder.Radius   =      DLGGetDouble (Dlg, CYLINDER_PAGE+10);
	  Add.Cylinder.Depth    =      DLGGetDouble (Dlg, CYLINDER_PAGE+20);
	  Add.Cylinder.Segment  =      DLGGetScale  (Dlg, CYLINDER_PAGE+30);
	  Add.Cylinder.Facettes = (int)DLGGetUpDn   (Dlg, CYLINDER_PAGE+40);
	  Part = new CYLINDER (World, PartName, Matrix,
			       Add.Cylinder.Radius,  Add.Cylinder.Depth, 
			       Add.Cylinder.Segment, Add.Cylinder.Facettes, 
			       Add.Material);
	  break;

        case 10 :
	  PRO0 ("SWEEP\n")
	  if (Add.Sweep.Num > 0 && Add.Sweep.Points != NULL)
	  {
	    Add.Sweep.Length = DLGGetDouble (Dlg, SWEEP_PAGE+10);
	    Part = new SWEEP (World, PartName, Matrix,
			      Add.Sweep.Num,    Add.Sweep.Points, 
			      Add.Sweep.Length, Add.Material);
	  }
	  else
	    PRGPrintf ("No Shape defined !");
	  break;

        case 11 :
	  PRO0 ("ROT\n")
	  if (Add.Rot.Num > 0 && Add.Rot.Points != NULL)
	  {
	    Add.Rot.Radius   =      DLGGetDouble (Dlg, ROT_PAGE+10);
	    Add.Rot.Segment  =      DLGGetScale  (Dlg, ROT_PAGE+20);
	    Add.Rot.Facettes = (int)DLGGetUpDn   (Dlg, ROT_PAGE+30);
	    Part = new ROTATIONAL (World, PartName, Matrix,
				   Add.Rot.Num,     Add.Rot.Points, 
				   Add.Rot.Segment, Add.Rot.Facettes, 
				   Add.Rot.Radius,  Add.Material);
	  }
	  else
	    PRGPrintf ("No Shape defined !");
	  break;

        case 12 :
	  PRO0 ("ISOCAHEDRON\n")
          Add.Ico.Radius1    =      DLGGetDouble   (Dlg, ICO_PAGE+10);
	  Add.Ico.Iterations = (int)DLGGetUpDn     (Dlg, ICO_PAGE+20);
	  Add.Ico.Star       =      DLGGetBool     (Dlg, ICO_PAGE+30);
	  Add.Ico.Radius2    =        DLGGetDouble (Dlg, ICO_PAGE+40);
	  Add.Ico.Density    = (int)DLGGetUpDn     (Dlg, ICO_PAGE+50);
	  Part = new ICOSAHEDRON (World, PartName, Matrix,
				  Add.Ico.Radius1,    Add.Ico.Radius2,
				  (Add.Ico.Star)? Add.Ico.Density : 0, 
				  Add.Ico.Iterations, Add.Material);
	  break;

        case 13 :
	  PRO0 ("PRISM\n")
          Add.Prism.Width1  = DLGGetDouble (Dlg, PRISM_PAGE+10);
	  Add.Prism.Width2  = DLGGetDouble (Dlg, PRISM_PAGE+20);
	  Add.Prism.OffsX   = DLGGetDouble (Dlg, PRISM_PAGE+30);
	  Add.Prism.Depth1  = DLGGetDouble (Dlg, PRISM_PAGE+40);
	  Add.Prism.Depth2  = DLGGetDouble (Dlg, PRISM_PAGE+50);
	  Add.Prism.OffsY   = DLGGetDouble (Dlg, PRISM_PAGE+60);
	  Add.Prism.Height1 = DLGGetDouble (Dlg, PRISM_PAGE+70);
	  Add.Prism.Height2 = DLGGetDouble (Dlg, PRISM_PAGE+80);
	  Add.Prism.OffsZ   = DLGGetDouble (Dlg, PRISM_PAGE+90);
	  Part = new PRISM (World, PartName, Matrix,
			    Add.Prism.Width1,  Add.Prism.Width2,
			    Add.Prism.OffsX,   Add.Prism.Depth1,
			    Add.Prism.Depth2,  Add.Prism.OffsY,
			    Add.Prism.Height1, Add.Prism.Height2,
			    Add.Prism.OffsZ,   Add.Material);
	  break;

        case 14 : 
	  PRO0 ("SCREW\n")
          Add.Screw.Radius1  =      DLGGetDouble (Dlg, SCREW_PAGE+10);
	  Add.Screw.Radius2  =      DLGGetDouble (Dlg, SCREW_PAGE+20);
	  Add.Screw.Height   =      DLGGetDouble (Dlg, SCREW_PAGE+30);
	  Add.Screw.Increase =      DLGGetDouble (Dlg, SCREW_PAGE+40);
	  Add.Screw.Offset   =      DLGGetDouble (Dlg, SCREW_PAGE+50);
	  Add.Screw.Facettes = (int)DLGGetUpDn   (Dlg, SCREW_PAGE+60);
	  Add.Screw.Thread   =      DLGGetOMenu  (Dlg, SCREW_PAGE+70);
	  Part = new SCREW (World, PartName, Matrix,
			    Add.Screw.Radius1, Add.Screw.Radius2,
			    Add.Screw.Height,  Add.Screw.Increase,
			    Add.Screw.Offset,  Add.Screw.Facettes,
			    (SCREW::THREAD)Add.Screw.Thread,
			    Add.Material);
	  break;

        case 15 :
	  PRO0 ("SPRING\n")
	  Add.Spring.Radius1   =      DLGGetDouble (Dlg, SPRING_PAGE+10);
	  Add.Spring.Radius2   =      DLGGetDouble (Dlg, SPRING_PAGE+20);
	  Add.Spring.Thickness =      DLGGetDouble (Dlg, SPRING_PAGE+30);
	  Add.Spring.Height    =      DLGGetDouble (Dlg, SPRING_PAGE+40);
	  Add.Spring.Increase  =      DLGGetDouble (Dlg, SPRING_PAGE+50);
	  Add.Spring.Facettes1 = (int)DLGGetUpDn   (Dlg, SPRING_PAGE+60);
	  Add.Spring.Facettes2 = (int)DLGGetUpDn   (Dlg, SPRING_PAGE+70);
	  Add.Spring.Thread    =      DLGGetOMenu  (Dlg, SPRING_PAGE+80);
	  Part = new SPRING (World, PartName,  Matrix,
			     Add.Spring.Radius1,   Add.Spring.Radius2,
			     Add.Spring.Thickness,
			     Add.Spring.Height,    Add.Spring.Increase,
			     Add.Spring.Facettes1, Add.Spring.Facettes2,
			     (SPRING::THREAD)Add.Spring.Thread,
			     Add.Material);
	  break;

        case 16 :
	  PRO0 ("GEAR\n")
          Add.Gear.Radius1  =      DLGGetDouble (Dlg, GEAR_PAGE+10);
	  Add.Gear.Radius2  =      DLGGetDouble (Dlg, GEAR_PAGE+20);
	  Add.Gear.Radius3  =      DLGGetDouble (Dlg, GEAR_PAGE+30);
	  Add.Gear.Depth1   =      DLGGetDouble (Dlg, GEAR_PAGE+40);
	  Add.Gear.Depth2   =      DLGGetDouble (Dlg, GEAR_PAGE+50);
	  Add.Gear.Teeth    = (int)DLGGetUpDn   (Dlg, GEAR_PAGE+60);
	  Add.Gear.Skew     =      DLGGetScale  (Dlg, GEAR_PAGE+70);
	  Part = new GEAR (World, PartName, Matrix,
			   Add.Gear.Radius1, Add.Gear.Radius2,
			   Add.Gear.Radius3, 
			   Add.Gear.Depth1,  Add.Gear.Depth2,
			   Add.Gear.Teeth,   Add.Gear.Skew,
			   Add.Material);
	  break;

        case 17 :
	  PRO0 ("ARROW\n")
	  Add.Arrow.Radius1  =      DLGGetDouble (Dlg, ARROW_PAGE+10);
	  Add.Arrow.Len1     =      DLGGetDouble (Dlg, ARROW_PAGE+20);
	  Add.Arrow.Radius2  =      DLGGetDouble (Dlg, ARROW_PAGE+30);
	  Add.Arrow.Len2     =      DLGGetDouble (Dlg, ARROW_PAGE+40);
	  Add.Arrow.Angle    =      DLGGetScale  (Dlg, ARROW_PAGE+50);
	  Add.Arrow.Segment  =      DLGGetScale  (Dlg, ARROW_PAGE+60);
	  Add.Arrow.Facettes = (int)DLGGetUpDn   (Dlg, ARROW_PAGE+70);
	  Part = new ARROW (World, PartName, Matrix,
			    Add.Arrow.Radius1,  Add.Arrow.Len1, 
			    Add.Arrow.Radius2,  Add.Arrow.Len2,
			    Add.Arrow.Angle,    Add.Arrow.Facettes, 
			    Add.Arrow.Segment,  Add.Material);
	  break;

        default :
	  PRO0 ("unsupported page\n")
      }

      if (Part != NULL)
      {
	if ((Dev = World->GetDev(Add.DevName)) != NULL)
	  Dev->AddPart (Part);            // Add part to existing device
	else                              // Create a new device for the part
	{
	  Dev = new DEVICE3D (Add.DevName, Part, World);
	  World->AddDevice (Dev);        
	}
      }

    }  // for Amount

    World->Fill    ();
    ClearSelection (World);
    SetCurrent (World);
  }

  if (Add.Sweep.Points != NULL)
  {
    free (Add.Sweep.Points);
    Add.Sweep.Points = NULL;
  }
  if (Add.Rot.Points != NULL)
  {
    free (Add.Rot.Points);
    Add.Rot.Points = NULL;
  }

  Close ();

  Preview = NULL;
  if (Demo != NULL)
  {
    delete Demo;
    Demo = NULL;
  }

  PRO_END
  return TRUE;
}


void CREATE_DIALOG::ChangeCB (int BoxID)
{
  int Page = DLGGetNote (Dlg, 10);

  PRO_METHOD ("CREATE_DIALOG", "")

  CAD_DIALOG::ChangeCB (BoxID);

  switch (Page)
  {
    case  1 : // global
              break;

    case  2 : // objects
              DLGSetNote (Dlg, BoxID, Add.Page);
              break;

    case 20 : // material
              MaterialCB (Dlg, MATERIAL_PAGE, &(Add.Material), BoxID, 
			                      ALPHA_MASK | TEX_MASK);
              break;

    default : // store selected page
              if (Page > 2 && Page < 20)
                Add.Page = Page;
  }

  if (Handle != NULL)
    Handle (Dlg, BoxID, Part);

/* TODO:
     In handle verschieben ?
*/
  if (Preview != NULL && Demo != NULL)
  {
    SetCurrent (Preview);
    Demo->SetMaterial (Add.Material);
    Preview->Redraw ();
  }

  PRO_END
}


void CREATE_DIALOG::Update ()
{
  PRO_METHOD ("CREATE_DIALOG", "\n")

  if (Dlg == NULL)
    PRO_ABORT ("Dialog is not open")

  // TODO: ...

  PRO_END
}


void CREATE_DIALOG::HelpCB (int ID)
{
  char Label[20] = "";

  //...

  HelpDlg.Open ("CreatePart", "dlg_create.hlp", Label);
}


void CREATE_DIALOG::OnChange (CHANGE_HANDLE Handle)
{
  this->Handle = Handle;
}


void CreatePart ()
{
  WORLD *World;

  if ((World = SelectWorld(TRUE)) == NULL)
    return;

  OpenCreateDialog (World);
}
