#include <stdarg.h>
#include <math.h>

#include <Xm/Form.h>
#include <Xm/ScrolledW.h>
#include <Xm/ScrollBar.h>
#include <Xm/PushB.h>
#include <Xm/Text.h>
#include <Xm/Label.h>

#include "tuxCAD.h"


static void EventHandle (Widget widget, CAD_EDITOR *Editor, XEvent *Event);


double CAD_EDITOR::SnapPos (double Pos, double Raster /* = -1.0 */)
{
  double Remain;

  if (Raster == -1.0)
    Raster = Snap;

  if (Raster > 0.0 && (Remain = fmod(Pos, Raster)) != 0)
    if (Remain < 0.5*Raster)
      Pos -= Remain;
    else
      Pos += Raster - Remain;

  return Pos;
}


void CAD_EDITOR::SetSnap (double Value)
{
  if (Snap == Value)
    return;

  Snap = Value;

  OffsX = SnapPos (0.5 * SizeX, Grid);
  OffsY = SnapPos (0.5 * SizeY, Grid);

  //  Redraw ();
}


void CAD_EDITOR::SetColor (EDITOR_COLORS Color)
{
  XSetForeground (XDisplay, XorGC, Colors[Color]);
}


void CAD_EDITOR::DrawLine (int x1, int y1, int x2, int y2, EDITOR_COLORS Color)
{
  XSetForeground (XDisplay, XorGC, Colors[Color]);
  XDrawLine (XDisplay, Win, XorGC, x1, y1, x2, y2);
}


void CAD_EDITOR::DrawGrid ()
{
  double StepX, StepY;

  XSetForeground (XDisplay, DrawGC, Colors[Raster]);

  for (StepY = 0; StepY < SizeY; StepY += Grid)
    for (StepX = 0; StepX < SizeX; StepX += Grid)
    {
      XDrawLine (XDisplay, Pic, DrawGC,
		 int(StepX-2), int(StepY), int(StepX+2), int(StepY));
      XDrawLine (XDisplay, Pic, DrawGC,
		 int(StepX), int(StepY-2), int(StepX), int(StepY+2));
    }

  XSetForeground (XDisplay, DrawGC, TOOLGetColor("Black"));
  XDrawLine (XDisplay, Pic, DrawGC,    0, int(OffsY), SizeX-1,   int(OffsY));
  XDrawLine (XDisplay, Pic, DrawGC, int(OffsX),     0,  int(OffsX), SizeY-1);
}


void CAD_EDITOR::DrawCrossHair (int x, int y)
{
  DrawLine (0, y, GetWidth()-1, y,  CrossHair);  
  DrawLine (x, 0, x, GetHeight()-1, CrossHair);
}


void CAD_EDITOR::Redraw ()
{
  int count;

  XSetForeground (XDisplay, DrawGC, Colors[Background]);
  XFillRectangle (XDisplay, Pic, DrawGC, 0, 0, SizeX, SizeY);

  if (Grid > 0.0)
    DrawGrid ();

  XSetForeground (XDisplay, DrawGC, Colors[Line] ^ Colors[Background]);
  for (count = 0; count < Num-1; ++count)
    XDrawLine (XDisplay, Pic, DrawGC,
	       int(Points[count].x),   int(Points[count].y),
	       int(Points[count+1].x), int(Points[count+1].y));

  XCopyArea (XDisplay, Pic, Win, DrawGC, 0, 0, SizeX, SizeY, 0, 0);
}


void CAD_EDITOR::AddPoint (double x, double y)
{
  if (Num == 99)
    return;

  Points[Num].x = x;
  Points[Num].y = y;

  Num++;
}


void CAD_EDITOR::ClosePath ()
{
  if (Num > 0)
  {
    Points[Num] = Points[0];
    Num++;
    EventHandle (NULL, NULL, NULL);
    Redraw ();
  }
}


void CAD_EDITOR::Clear ()
{
  Num = 0;
  Redraw ();
}


void CAD_EDITOR::DeletePoint (int No)
{
  int count;

  for (count = No; count < Num-1; ++count)
    Points[count] = Points[count+1];

  Num--;
}


static void EventHandle (Widget widget, CAD_EDITOR *Editor, XEvent *Event)
{
  static int    Btn;
  static int    px, py;
  static int    LastX, LastY;
  static double GridX, GridY;

  if (Editor == NULL)
  {
    px = py = -1;
    LastX = LastY = -1;
    GridX = GridY = 0.0;
    Btn = 0;
    return;
  }

  switch (Event->type)
  {
    case MotionNotify :
      {
	XMotionEvent *Motion = (XMotionEvent *)Event;

	Editor->DrawCrossHair (LastX, LastY);
	Editor->DrawCrossHair (Motion->x, Motion->y);

	if (Btn == 1)
	  Editor->DrawLine (px, py, int(GridX), int(GridY), CAD_EDITOR::Line);

	GridX = Editor->SnapPos (Motion->x);
	GridY = Editor->SnapPos (Motion->y);

	if (Btn == 1)
	  Editor->DrawLine (px, py, int(GridX), int(GridY), CAD_EDITOR::Line);

	LastX = Motion->x;
	LastY = Motion->y;

	Editor->PrintPos (int(GridX), int(GridY));
      }
      break;

    case EnterNotify :
      //      break;
    case LeaveNotify :
      Editor->DrawCrossHair (LastX, LastY);
      //      Editor->Redraw ();
      break;

    case ButtonPress :
      {
	XButtonEvent *Button = (XButtonEvent *)Event;

	Editor->AddPoint (Editor->SnapPos(Button->x),
			  Editor->SnapPos(Button->y));

	switch ((Btn = Button->button))
	{
  	  case 1 :   //  left: Add Point
            LastX = Button->x;
	    LastY = Button->y;
	    GridX = Editor->SnapPos (Button->x);
	    GridY = Editor->SnapPos (Button->y);
	    px = int(GridX);
	    py = int(GridY);
	    break;
  	  case 2 :   // middle: stop
	    Editor->Clear ();
	    break;
   	  default :  // right: close path
	    Editor->ClosePath ();
	}
      }
      break;

    case Expose :
      Editor->Redraw ();
      break;

    default :
      break;
  }
}


void CAD_EDITOR::SetPos (int PosX, int PosY)
{
  int Dummy;
  int Size, Inc, Page;

  XmScrollBarGetValues (ScrBarX, &Dummy, &Size, &Inc, &Page);
  XmScrollBarSetValues (ScrBarX, PosX, Size, Inc, Page, True);

  XmScrollBarGetValues (ScrBarY, &Dummy, &Size, &Inc, &Page);
  XmScrollBarSetValues (ScrBarY, PosY, Size, Inc, Page, True);  
}


void CAD_EDITOR::TransformPoints ()
{
  int count;

  // TODO:
  // Segementation fault, falls Punkt auf einer Linien liegt !

  // Transform screen pixels into part-coordinates
  for (count = 0; count < Num; ++count)
  {
    Points[count].x = SnapPos ( (Points[count].x - OffsX) / Zoom);
    Points[count].y = SnapPos (-(Points[count].y - OffsY) / Zoom);
  }
}


static void CloseCB (Widget w, CAD_EDITOR *Editor, XtPointer Data)
{
  XtPopdown (Editor->GetShell());
  XtDestroyWidget (Editor->GetShell());

  Editor->TransformPoints ();
}


void ExitEdit (CAD_EDITOR *Editor)
{
  CloseCB (NULL, Editor, NULL);
}


void SetSnap1 (CAD_EDITOR *Editor)
{
  Editor->SetSnap (1.0);  
}


void SetSnap2 (CAD_EDITOR *Editor)
{
  Editor->SetSnap (2.0);
}


void SetSnap5 (CAD_EDITOR *Editor)
{
  Editor->SetSnap (5.0);  
}


void SetSnap10 (CAD_EDITOR *Editor)
{
  Editor->SetSnap (10.0);  
}


CAD_EDITOR::CAD_EDITOR (char *Name, 
			int SizeX /* = 400 */, int SizeY /* = 400 */)
{
  Widget    Form, ScrWin;
  Widget    Menu, Tools;
  Widget    Button;
  Widget    Label;
  XmString  LabelStr;
  XGCValues GCVal;

  static MENU_ENTRY SystemMenu[] = 
  {
    ENTRY ("exit",   "",     "", ' ', TRUE, ExitEdit, this)
  };
  static MENU_ENTRY SnapMenu[] = 
  {
    ENTRY ("1x1",   "", "", ' ', TRUE, SetSnap1,  this),
    ENTRY ("2x2",   "", "", ' ', TRUE, SetSnap2,  this),
    ENTRY ("5x5",   "", "", ' ', TRUE, SetSnap5,  this),
    ENTRY ("10x10", "", "", ' ', TRUE, SetSnap10, this)
  };
  static MENU_ENTRY ConfigMenu[] = 
  {
    SUB_MENU ("snap...", "", "", ' ', TRUE,  SnapMenu),
    SUB_MENU ("grid...", "", "", ' ', FALSE, NULL)
  };
  static MENU_COLUMN MenuCols[] = 
  {
    FUNC_MENU ("system", 's', TRUE, SystemMenu),
    FUNC_MENU ("config", 'c', TRUE, ConfigMenu)
  };
  static MENU_BAR EditMenu = MAIN_MENU ("editMenu", MenuCols);

  Zoom =  1.0;
  Grid = 10.0;
  Snap =  5.0;
  this->SizeX = SizeX;
  this->SizeY = SizeY;
  OffsX = SnapPos (0.5 * SizeX, Grid);
  OffsY = SnapPos (0.5 * SizeY, Grid);
  Num = 0;

  Shell = XtVaCreatePopupShell ("cadEditor",
				topLevelShellWidgetClass, AppShell,
				XmNtitle,                 Name,
				NULL);

  Form = XtCreateManagedWidget ("form", xmFormWidgetClass, Shell, NULL, 0);

  Menu = PRGCreateMenuBar (Form, &EditMenu);

  /* 
     UpDn (Grid), Koordinatensystem, Add Point, Delete Point, Drag point
     Group select, Move group, Scale group, Append Arc,
  */

  LabelStr = XmStringLtoRCreate ("Mouse:", XmSTRING_DEFAULT_CHARSET);
  Label = XtVaCreateManagedWidget ("mouse", xmLabelWidgetClass, Form,
				            XmNleftAttachment,   XmATTACH_FORM,
                                            XmNleftOffset,       5,
				            XmNbottomAttachment, XmATTACH_FORM,
				            XmNbottomOffset,     10,
				            NULL);

  Msg1 = XtVaCreateManagedWidget ("posX", xmTextWidgetClass, Form,
				          XmNcolumns,          6,
				          XmNleftAttachment,   XmATTACH_WIDGET,
				          XmNleftWidget,       Label,
                                          XmNleftOffset,       5,
				          XmNbottomAttachment, XmATTACH_FORM,
				          XmNbottomOffset,     5,
				          XmNmarginHeight,     5,
				          NULL);

  Msg2 = XtVaCreateManagedWidget ("posY", xmTextWidgetClass, Form,
				          XmNcolumns,          6,
				          XmNleftAttachment,   XmATTACH_WIDGET,
				          XmNleftWidget,       Msg1,
                                          XmNleftOffset,       3,
				          XmNbottomAttachment, XmATTACH_FORM,
				          XmNbottomOffset,     5,
				          XmNmarginHeight,     5,
				          NULL);

  Msg = XtVaCreateManagedWidget ("msg", xmTextWidgetClass, Form,
				        XmNleftAttachment,   XmATTACH_WIDGET,
				        XmNleftWidget,       Msg2,
                                        XmNleftOffset,       3,
				        XmNrightAttachment,  XmATTACH_FORM,
				        XmNrightOffset,      5,
				        XmNbottomAttachment, XmATTACH_FORM,
				        XmNbottomOffset,     5,
				        XmNmarginHeight,     5,
				        NULL);

  Button = XtVaCreateManagedWidget ("close", xmPushButtonWidgetClass, Form,
				    XmNrightAttachment,  XmATTACH_FORM,
				    XmNrightOffset,      5,
				    XmNbottomAttachment, XmATTACH_WIDGET,
				    XmNbottomWidget,     Msg1,
				    XmNbottomOffset,     3,
				    NULL);
  XtAddCallback (Button, XmNactivateCallback, (XtCallbackProc)CloseCB, this);

  Tools = XtVaCreateManagedWidget ("tools", xmFormWidgetClass, Form, 
				   XmNtopAttachment,    XmATTACH_WIDGET,
				   XmNtopWidget,        Menu,
				   XmNleftAttachment,   XmATTACH_FORM,
				   XmNwidth,            35,
				   XmNbottomAttachment, XmATTACH_WIDGET,
				   XmNbottomWidget,     Button,
				   NULL);

  XtVaCreateManagedWidget ("test", xmPushButtonWidgetClass, Tools,
				    XmNleftAttachment,  XmATTACH_FORM,
				    XmNtopAttachment,   XmATTACH_FORM,
				    NULL);

  ScrWin = XtVaCreateManagedWidget ("scrWin", xmScrolledWindowWidgetClass, 
				    Form,
				    XmNscrollingPolicy,        XmAUTOMATIC,
				    XmNscrollBarDisplayPolicy, XmAS_NEEDED,
				    XmNtopAttachment,          XmATTACH_WIDGET,
				    XmNtopWidget,              Menu,
				    XmNleftAttachment,         XmATTACH_WIDGET,
				    XmNleftWidget,             Tools,
				    XmNrightAttachment,        XmATTACH_FORM,
				    XmNrightOffset,            5,
				    XmNbottomAttachment,       XmATTACH_WIDGET,
				    XmNbottomWidget,           Button,
				    XmNbottomOffset,           3,
				    XmNwidth,                  MIN(SizeX, 250),
				    XmNheight,                 MIN(SizeY, 250),
				    NULL);
  XtVaGetValues (ScrWin, XmNhorizontalScrollBar, &ScrBarX,
		         XmNverticalScrollBar,   &ScrBarY,
		         NULL);

  Work = XtVaCreateManagedWidget ("textArea", xmLabelWidgetClass, ScrWin,
				  XmNmarginLeft,  0,
				  XmNmarginWidth, 0,
				  XmNalignment,   XmALIGNMENT_BEGINNING,
				  XmNenableUnselectableDrag, False,
				  XmNconvertProc, NULL,
				  NULL);

  EventHandle (NULL, NULL, NULL); // Initialize Event attributes 

  XtManageChild (Shell);
  XtVaSetValues (ScrWin, XmNworkWindow, Work, NULL);

  Win = XtWindow (Work);

  Colors[Background] = TOOLGetColor ("Grey90");
  Colors[Line]       = TOOLGetColor ("Red")  ^ Colors[Background];
  Colors[Raster]     = TOOLGetColor ("Grey70");
  Colors[CrossHair]  = TOOLGetColor ("Blue") ^ Colors[Background];
  GCVal.foreground = Colors[Line];
  GCVal.background = Colors[Background];
  GCVal.graphics_exposures = False;
  DrawGC = XCreateGC (XDisplay, Win, GCGraphicsExposures, &GCVal);

  XorGC = XCreateGC (XDisplay, Win, GCForeground | GCBackground, &GCVal);
  XSetFunction (XDisplay, XorGC, GXxor);

  Pic = XCreatePixmap (XDisplay, Win, SizeX, SizeY, TOOLGetDepth());

  XtVaSetValues (Work, XmNlabelType,   XmPIXMAP,
		       XmNlabelPixmap, Pic,
		       NULL);

  XtAddEventHandler (Work, KeyPressMask    | PointerMotionMask |
		           EnterWindowMask | LeaveWindowMask   |
		           ButtonPressMask | ButtonReleaseMask |
		           ExposureMask /*| StructureNotifyMask */, 
		           FALSE, (XtEventHandler)EventHandle, this);

  SetPos (int(0.5*MAX(0,SizeX-250)), int(0.5*MAX(0,SizeY-250)));

  Redraw ();

  XtRealizeWidget (Shell);
  XtPopup         (Shell, XtGrabNonexclusive);

  TOOLSetMouseCursor (Shell, ToolCROSS, "Blue", "White");

  TOOLMainLoop (AppCon, Shell);
}


void CAD_EDITOR::PrintPos (int x, int y)
{
  char Message[20];

  sprintf (Message, "x: %4d", x-int(OffsX));
  XmTextSetString (Msg1, Message);
  
  sprintf (Message, "y: %4d", x-int(OffsY));
  XmTextSetString (Msg2, Message);
}

void CAD_EDITOR::Printf (char *Mesg, ...)
{
  char    Message[512];
  va_list Arguments;

  va_start (Arguments, Mesg);
  vsprintf (Message, Mesg, Arguments);
  va_end (Arguments);

  XmTextSetString (Msg, Message);
}


void EditPart ()
{
  CAD_EDITOR *Editor;

  Editor = new CAD_EDITOR ("CAD-Editor", 600, 600);
}  


void MergeParts ()
{
  PART3D *Part1, *Part2;
  WORLD  *World = GetCurrent();

  if ((Part1 = World->PickPart()) == NULL ||
      (Part2 = World->PickPart()) == NULL)
  {
    PRGPrintf ("Function aborted");
    return;
  }

  if (Part1 != Part2)
    World->MergeParts (Part1, Part2);

  World->Redraw ();
}


void SmashPart ()
{
  PART3D *Part;
  WORLD  *World = GetCurrent();

  if ((Part = World->PickPart()) == NULL)
  {
    PRGPrintf ("Function aborted");
    return;
  }

  World->Smash  (Part);
  World->Redraw ();
}
