#include <assert.h>
#include <limits.h>

#include <libcad3d.h>
#include "../libfilter.h"


#define DOFn(fp,n,Val) \
        { \
          for (index = 1; index < n; ++index) \
            fprintf (fp, "%d ", Val); \
          fprintf (fp, "%d\n", Val); \
        }

#define MATRIX2FRAME(M) \
        FRAME (M.m[N][X], M.m[N][Y], M.m[N][Z], \
	       M.m[O][X], M.m[O][Y], M.m[O][Z], \
	       M.m[A][X], M.m[A][Y], M.m[A][Z], \
	       M.m[P][X], M.m[P][Y], M.m[P][Z])

#define DEFAULT_TRANSPARENCY 0.5

#define FILE_TRACE TRUE


static COLOR::ID IGRIPColors[] =              // Color-mapping for igrip-files
{
  COLOR::OBJECT,                              // Use object colors
  COLOR::GREY50,                              // Indexed colors
  COLOR::GREY,
  COLOR::SILVER,
  COLOR::WHITE,
  COLOR::GREY,
  COLOR::BLACK,
  COLOR::MEDRED,
  COLOR::MEDGREEN,
  COLOR::MEDBLUE,
  COLOR::YELLOW,
  COLOR::CYAN,
  COLOR::PURPLE,
  COLOR::ORANGE,
  COLOR::MAGENTA,
  COLOR::CHERRY,
  COLOR::LIGHTBLUE,
  COLOR::BROWN,
  COLOR::LIME,
  COLOR::TAN,
  COLOR::GOLD,
  COLOR::STEELBLUE,
  COLOR::TURQUOISE,
  COLOR::SIENNA,
  COLOR::PEACHPUFF,
  COLOR::PERU,
  COLOR::LEMONCHIFFON,
  COLOR::ORCHID,
  COLOR::SILK,
  COLOR::DODGERBLUE,
  COLOR::PINK,
  COLOR::DARKBLUE,
  COLOR::SEAGREEN,
  COLOR::DARKORANGE,
  COLOR::SEAGREEN,
  COLOR::DARKRED
};

#define IGRIP_COLORS (sizeof(IGRIPColors)/sizeof(COLOR::ID));


static bool ReadLine (FILE *fp, bool show = FALSE, 
		      char *Result = NULL, int Len = 0);

class ATTRIBUTES
{
  public :
    ATTRIBUTES () :
      Type  (0),
      Min   (0.0),
      Max   (0.0),
      Range (0),
      Speed (0.0),
      Accel (0.0),
      Time  (0.0)
    {
      int count;

      for (count = 0; count < 6; ++count)
	Val[count] = 0.0;
    };
   
  public :
    int    Type;
    double Min, Max, Val[6];
    int    Range;
    double Speed, Accel, Time;
};


char *GetName ()
{
  return "IGRIP";
}


bool Check (const char *FileSpec)
{
  return TRUE;
}


static bool ReadLine (FILE *fp, bool show /* = FALSE */, 
		      char *Result /* = NULL */, int Len /* = 0 */)
{
  int  count = 0;
  char letter = ' ';

  while (! feof(fp) && letter != '\n')
  {
    letter = fgetc (fp);

    if (Result != NULL && count < Len)
      Result[count++] = letter;
      
    if (show)
      PRO1 ("\\%c", letter)
  }

  if (Result != NULL)
    Result[Len-1] = '\0';

  return (! feof(fp));
}


static DEVICE3D *ReadDev (FILE *fp, WORLD *World, double Version, 
		          const char *DataPath)
{
  int         count, index;
  int         IVal, DOF, DOF2, Axes;
  int         CSNo, CS1No, CS2No;
  int         Color, Lines;
  int         Flag = 1;
  double      Val, DVal, Value[6];
  char        letter;
  char        PosName[80];
  char        String[100];
  char        FileName[PATH_MAX+1];
  char        FileSpec[PATH_MAX+1];
  char        PartName[PATH_MAX+1];
  char        DataDir[PATH_MAX+1];
  H_MATRIX    Offset, M;
  PART       *Part;
  ATTRIBUTES *Attr;
  PART3D     *Ptr, *Prev;
  DEVICE3D   *Dev = NULL;
  ROBOT      *Robot = NULL;

  PRO_METHOD ("WORLD", "\n")

  if (DataPath == NULL)
    TOOLParseFileName ("~/tuxcad/data", DataDir, PATH_MAX);
  else
    TOOLParseFileName (DataPath,        DataDir, PATH_MAX);

  Dev = new DEVICE3D ();

  fscanf (fp, "%s\n", String);        // Device-Name
  Dev->SetName (String);
  PRO1 ("  DevName  : %s\n", String)

  fscanf (fp, "%d\n", &IVal);         // 0: no kinematic, 1: has kinematic
  PRO1 ("  KINEMATICS: %s\n", (IVal>0)? "YES":"NO")
  ReadLine (fp);                      // 16|32                           ?
  ReadLine (fp);                      // -1                              ?
  fscanf (fp, "%d\n", &DOF);          // DOF
  Dev->SetDOF (DOF);
  PRO1 ("  DOF       : %d\n", DOF)

  Attr = new ATTRIBUTES[DOF+1];

  ReadLine (fp);                      // 0                               ?
  ReadLine (fp);                      // 0                               ?
  ReadLine (fp);                      // 0|15                            ?
  if (Version < 11)
  {
    ReadLine (fp);
    ReadLine (fp);
  }
  else
  {
    fscanf (fp, "%d\n", &Flag);       // 1                               ?
    if (Flag == 1)
      ReadLine (fp);                  // 0                               ?
  }
  ReadLine (fp);                      // 0                               ?

  if (IVal > 0)
  {
    ReadLine (fp, FILE_TRACE);        // Program-FileName
    ReadLine (fp);                    // 0                            ?
    fscanf (fp, "%d\n", &IVal);       // # Home-Positions
    if (Version < 11)
      Dev->SetHomeNo (IVal*Flag);
    else
      Dev->SetHomeNo (IVal);
    PRO1 ("  #HOME-POS : %d\n", Dev->GetHomeNo())
    fscanf (fp, "%d\n", &CS1No);      // Amount of CS1s                  ?
    fscanf (fp, "%d\n", &CS2No);      // Amount of CS2s                  ?
    PRO2 ("  #CS       : %d / %d\n", CS1No, CS2No)
    fscanf (fp, "%d\n", &IVal);       // # Configs                       ?
    Dev->SetConfigNo (IVal);
    PRO1 ("  CONFIGS   : %d\n", IVal)

    ReadLine (fp, FILE_TRACE);        // VendorName
    ReadLine (fp);                    // 0                            ?
    ReadLine (fp);                    // 0                            ?
    ReadLine (fp);                    // TCP-Speed (Cart/Joint) [MM/SEC]
                                      // TCP-Accel              [MM/SEC^2]
    ReadLine (fp);                    // Angular-Speed (1..3)   [DEG/SEC]
    ReadLine (fp);                    // Angular-Accel (1..3)   [DEG/SEC^2]
    if (Version > 10)
      ReadLine (fp);                  // Angular-Time  (1..3)   [SEC]
    ReadLine (fp);                    // 0 0                          ?
    ReadLine (fp);                    // eg. 1 1 1 1 0.1 0.5 0        ?
    ReadLine (fp);                    // 0 1 0 0 0 2 0 1              ?
    fscanf (fp, "%s\n", String);      // Kin.: Move_By_Links, kin_generic, ...

    if (strcasecmp(String,    "kin_rx60") == 0)
      Robot = new RX60;
    elsif (strcasecmp(String, "kin_rx90") == 0)
      Robot = new RX90;
    elsif (strcasecmp(String, "kin_r2") == 0)
      Robot = new R2;
    elsif (strcasecmp(String, "kin_r15") == 0)
      Robot = new R15;
    elsif (strcasecmp(String, "kin_rm501") == 0)
      Robot = new RM501; 
    elsif (strcasecmp(String, "kin_hexa") == 0)
      Robot = new HEXA;
    elsif (strcasecmp(String, "kin_manutec") == 0)
      PRO0 ("Manutec Kinematic\n")
    elsif (strcasecmp(String, "kin_generic") == 0)
      PRO0 ("Generic Kinematic\n")       // TODO: Use DH-parameter
    else
      PRO1 ("Unknown Kinematic: '%s'\n", String)

    Dev->SetRobot (Robot);

    ReadLine (fp);                       // none | vellocity                ?
    ReadLine (fp);                       // none | forward vellocity        ?

    if (Version > 10)
    {
#if 0
      if (Version == 11)
	ReadLine (fp);                  // 0                               ?

      fscanf (fp, "%d\n", &IVal);       // # Connected Sub-Devices (-1)    ?
      //      ReadLine (fp);
      PRO1 ("Configs: %d\n", IVal)
      if (IVal > 0)                     // bzw. abhngig von ConfigNo
      {
	ReadLine (fp);                  // ConfigNo x Config-Type          ?
	for (count = 0; count < Dev->GetConfigNo(); ++count)
	  ReadLine (fp);                // Config_[1..n]                   ?
      }
#else
      ReadLine (fp);                    // # Connected Sub-Devices
      if (Dev->GetConfigNo() > 0)
      {
	ReadLine (fp);                  // ConfigNo x Config-Type          ?
	for (count = 0; count < Dev->GetConfigNo(); ++count)
	  ReadLine (fp);                // Config_[1..n]                   ?
      }
#endif
    }
    else
      ReadLine (fp);                    // 1 1 1 1 1 1 1 1

    ReadLine (fp, FILE_TRACE);          // 0|1 (dyn. Libs)             ?
    ReadLine (fp);                      // mot_joint_0                    ?
    ReadLine (fp);                      // mot_straight_0                 ?
    ReadLine (fp);                      // mot_circular_0/mot_straight_0  ?
    ReadLine (fp);                      // mot_joint_0 | mot_slew_0       ?
    ReadLine (fp);                      // mot_aux1_0                     ?
    ReadLine (fp, FILE_TRACE);          // mot_aux2_0                     ?

    Attr[0].Type = 0;
    for (count = 1; count <= DOF; ++count)  // Joint-types
      fscanf (fp, "%d ", &Attr[count].Type);

    if (Version > 10)
      ReadLine (fp);                    // Kinematic-Type (0: rot/1: trans)

    Attr[0].Val[5] = 0.0;
    for (count = 1; count <= DOF; ++count)
      fscanf (fp, "%lf ", &Attr[count].Val[5]); // ***** JOINT-VALUES ******

    ReadLine (fp, FILE_TRACE);                 // 1 | 2                     ?
    for (index = 0; index < Dev->GetHomeNo(); ++index)
      for (count = 0; count < DOF; ++count)  // HOME-Pos
	fscanf (fp, "%lf ", &DVal);

    if (Version > 10)
    {
      for (count = 0; count < DOF; ++count)
	ReadLine (fp, FILE_TRACE);      // = dof(n)                 ?

      Attr[0].Range = -99;
      for (count = 1; count <= DOF; ++count) // DOF x Range (-99 = no)   ?
	fscanf (fp, "%d ", &Attr[count].Range);
      for (count = 1; count <= DOF; ++count)
	fscanf (fp, "%d ", &IVal);    // DOF x 0,Trans(13,14,15), Rot(16,17,18)
      ReadLine (fp);                  // DOF x 0                         ?
      ReadLine (fp);                  // DOF x 0                         ?

      ReadLine (fp);                  // 1 1 1
    }

    for (count = 0; count < 4; ++count) // TCP/Mount - Offset
      fscanf (fp, "%lf %lf %lf %lf\n", 
	      &Offset.m[count][X], &Offset.m[count][Y], 
	      &Offset.m[count][Z], &Offset.m[count][W]);
    Dev->SetOffset (Offset);
    if (Robot != NULL)
      Robot->SetOffset (MATRIX2FRAME(Offset));

#if 1
    fscanf (fp, "%d\n", &CSNo);
    for (count = 0; count < CSNo; ++count)
#else
    ReadLine (fp);                      // 1                         ?
    for (count = 0; count < CS1No; ++count) // CS1-frames
#endif
    {
      ReadLine (fp);                    // Normal
      ReadLine (fp);                    // Open
      ReadLine (fp);                    // Approach
      ReadLine (fp);                    // Position
    }

#if 1
    fscanf (fp, "%d\n", &CSNo);
    for (count = 0; count < CSNo; ++count)
#else
    ReadLine (fp);                      // 1                         ?
    for (count = 0; count < CS2No; ++count) // CS2-frames
#endif
    {
      ReadLine (fp);                    // Normal
      ReadLine (fp);                    // Open
      ReadLine (fp);                    // Approach
      ReadLine (fp);                    // Position
    }

    ReadLine (fp, FILE_TRACE);          // DOF x Link-Length
    ReadLine (fp);                      // DOF x Offset 
    ReadLine (fp);                      // Aux1 .. Aux8

    // TODO: Min und Max auf 6 Freiheitsgrade erweitern (RPY)
    Attr[0].Min = Attr[0].Max = 0.0;
    for (count = 1; count <= DOF; ++count)
    {
      fscanf (fp, "%s\n", String);      // Min-Range [RAD|MM] | =formula
      if (String[0] == '=')
	Val = 0.0;
      else
	Val = atof (String);
      if (Attr[count].Type == 0)
	Val = GRAD (Val);
      Attr[count].Min = Val;
    }

    for (count = 1; count <= DOF; ++count)
    {
      fscanf (fp, "%s\n", String);      // Max-Range [RAD|MM]
      if (String[0] == '=')
	Val = 0.0;
      else
	Val = atof (String);
      if (Attr[count].Type == 0)
	Val = GRAD (Val);
      Attr[count].Max = Val;
      PRO3 ("  Range %d   : %.2f...%.2f\n", 
	    count, Attr[count].Min, Attr[count].Max)
    }

    Attr[0].Speed = Attr[0].Accel = Attr[0].Time = 0.0;
    for (count = 1; count <= DOF; ++count)
      fscanf (fp, "%lf ", &Attr[count].Speed); // Speed (J1..J6) [GRAD/SEC] 
    for (count = 1; count <= DOF; ++count)
      fscanf (fp, "%lf ", &Attr[count].Accel); // Accel (J1..J6) [GRAD/SEC^2]
    if (Version > 10)
      for (count = 1; count <= DOF; ++count)
	fscanf (fp, "%lf ", &Attr[count].Time);// Accel-Time (J1..J6) [SEC]

    ReadLine (fp);                      // DOF x 0                         ?
    ReadLine (fp);                      // DOF x 0                         ?
    if (Version < 11)
    {
      ReadLine (fp);                    // 8 x 0  ?
      ReadLine (fp);                    // 8 x 0  ?     

      fscanf (fp, "%d\n", &IVal);     // Mounting Plate.  neg.: PartNo in world
      Dev->SetMountNo (IVal);         //                  pos.: rel. chain-part
      PRO1 ("  Mount       : %d\n", IVal) // (own convention for hexa)

      ReadLine (fp);                    // DOF x 0                         ?
      ReadLine (fp);                    // DOF x 0                         ?
      ReadLine (fp);                    // DOF x 0                         ?

      for (count = 1; count <= DOF; ++count)
	ReadLine (fp);                  // = dof(n)                        ?

      do
	fscanf (fp, "%s\n", String);
      while (String[0] == '=');
      ReadLine (fp);                    // DOF x 1                         ?

      for (count = 0; count < 3; ++count) // 3 Frames                      ?
      {
	ReadLine (fp);                  // Normal
	ReadLine (fp);                  // Open
	ReadLine (fp);                  // Approach
	ReadLine (fp);                  // Position
      }

      ReadLine (fp);                    // DOF x 0                         ?
      ReadLine (fp);                    // DOF x 0                         ?
      ReadLine (fp);                    // 0 0 0 0 0 0 0 0                 ?
      ReadLine (fp);                    // 0 0 0 0 0 0 0 0                 ?
    }
    else
    {
      fscanf (fp, "%d\n", &Lines);    // # following lines
      PRO1 ("# Lines: %d\n", Lines)

      if (Version == 11)
	for (count = 0; count < Lines; ++count)
	  ReadLine (fp);
      else
      {
	fscanf (fp, "%d\n", &IVal);     // Mounting Plate
	Dev->SetMountNo (IVal);
	PRO1 ("  Mount       : %d\n", IVal)

	// corresponds with:
	ReadLine (fp);                  // 0 0                             ?

	for (count = 0; count < 10; ++count)
	  ReadLine (fp);                // *
	ReadLine (fp);                  // DEFAULT
	ReadLine (fp);                  // 0
	ReadLine (fp);                  // 0
	ReadLine (fp);                  // 0
	ReadLine (fp);                  // *
	ReadLine (fp);                  // *
	ReadLine (fp);                  // *
	ReadLine (fp);                  // *
 
	if (Flag)
        {
	  for (count = 0; count < Dev->GetHomeNo(); ++count)
	    fscanf (fp, "%s\n", PosName); // Position-Names
	  ReadLine (fp);                // 0 (End)

	  for (count = 0; count < Dev->GetHomeNo(); ++count)
	    ReadLine (fp);              // *                               ?
	}
      
	for (count = 1; count <= DOF; ++count)
	  ReadLine (fp);                // Joint n                         ?
      }

      ReadLine (fp);                    // 5|8                             ?
      ReadLine (fp);                    // 0.1 0.00174533                  ?
      ReadLine (fp);                    // 0 0 -1.5708 0 0 0               ?
      ReadLine (fp);                    // 0 0 0 0 0 0                     ?
      ReadLine (fp);                    // 1 2 3 4 5 6                     ?
      ReadLine (fp);                    // 0 0 0 0 0 0                     ?

      for (count = 0; count < 3; ++count) // 3 Frames  (CS1No + CS2No) ?
      {
	ReadLine (fp);                  // Normal
	ReadLine (fp);                  // Open
	ReadLine (fp);                  // Approach
	ReadLine (fp);                  // Position
      }

      ReadLine (fp);                    // 0 0 51.5498 0 0 0               ?
      ReadLine (fp);                    // 0 289.98 309.685 0.0124512 0 0  ?

      fscanf (fp, "%d\n", &Lines);      // # followin lines
      PRO1 ("# Lines: %d\n", Lines)

      if (Version == 11)
	for (count = 0; count < Lines; ++count)
	  ReadLine (fp);
      else                              // normally Lines = 7
      {
	ReadLine (fp);                  // *
	ReadLine (fp);                  // *
	ReadLine (fp);                  // 1 1 1 1 1 1                     ?
	ReadLine (fp);                  // DOF x 0|1                       ?
	ReadLine (fp);                  // 0 0 0 0                         ?
	ReadLine (fp);                  // *
	ReadLine (fp);                  // *
      }
    }
  }

  if (Version > 10)
  {
    fscanf (fp, "%d\n", &Lines);
    PRO1 ("# Lines: %d\n", Lines)
    for (count = 0; count < Lines; ++count)
      ReadLine (fp);

    ReadLine (fp);                      // 0                               ?
    ReadLine (fp, FILE_TRACE);          // App-File
    fscanf (fp, "%d\n", &Lines);        // # followin lines
    for (count = 0; count < Lines; ++count)
      ReadLine (fp);                    // Par-File, 0 1 1 1               ?

    
    ReadLine (fp);                      // # following lines: normaly  6   ?

    // 10 Alpha 1 32768(Render) 0 0    ?
    fscanf (fp, "%d ", &IVal);          // 10
    fscanf (fp, "%d ", &IVal);          // Alpha
    Dev->SetAlpha (IVal);
    fscanf (fp, "%d ", &IVal);          // 1
    fscanf (fp, "%d ", &IVal);          // 32768x, 4=normal,2=transp.,3=wire
    IVal = IVal % 10;
    if (IVal == 2)
    {
      if (World != NULL)
	World->IncAlpha ();
      Dev->SetAlpha (DEFAULT_TRANSPARENCY);
    }
#if 0
    elsif (IVal == 3)
      Dev->SetRender (WINDOW3D::WIRE_FRAME);
#endif
    ReadLine (fp);                      // Rest of line (0 0)

    ReadLine (fp);                      // -1 -1                           ?
    ReadLine (fp);                      // -1 -1 -1                        ?
    ReadLine (fp);                      // -1 -1 -1                        ?
    ReadLine (fp);                      // -1 -1 -1                        ?
  }

  fscanf (fp, "%d\n", &Lines);          // # following lines: mostly 1     ?
  for (count = 0; count < Lines; ++count)
    ReadLine (fp);                      // 0

  index = 0;
  Prev  = NULL;
  do
  {
    PRO0 ("    [PART:]\n")
    PRO1 ("    Joint: %d\n",   index)

    Ptr = new PART3D ();
    Ptr->SetParent (Dev);
    Ptr->SetPrev   (Prev);
    Ptr->SetWorld  (World);

    switch (Attr[index].Type)
    {
      case 0 :
	Ptr->SetType (PART3D::ROT);
	break;
      case 1 :
	Ptr->SetType (PART3D::TRANS);
	break;
      default :
	Ptr->SetType (PART3D::UNKNOWN);
    } 

    fscanf (fp, "%d", &Color);           // Joint-Color
    PRO1 ("    Color: %d\n", Color)
    if (Color == -1)
      Ptr->SetMaterial (MATERIAL(COLOR::OBJECT));
    else
      Ptr->SetMaterial (MATERIAL(IGRIPColors[abs(Color)]));

    fscanf (fp, "%d", &IVal);            // 1 1 262145 0                   ?
    Ptr->SetAlpha (IVal);
    fscanf (fp, "%d", &IVal);
    fscanf (fp, "%d", &IVal);
    IVal = IVal % 10;
    if (IVal == 2)
    {
      if (World != NULL)
	World->IncAlpha ();
      Ptr->SetAlpha (DEFAULT_TRANSPARENCY);
    }
#if 0
    elsif (IVal == 3)
      Dev->SetRender (WINDOW3D::WIRE_FRAME);
#endif
    ReadLine (fp);

    if (Version > 10)
    {
      ReadLine (fp);                    // -1 1                            ?
      ReadLine (fp);                    // 1 1 1                           ?
      ReadLine (fp);                    // -1 -1 -1                        ?
      ReadLine (fp);                    // -1 -1 -1                        ?
    }

    for (count = 0; count < 4; ++count)  // relative part position
      fscanf (fp, "%lf %lf %lf %lf\n", &M.m[count][X], &M.m[count][Y], 
	                               &M.m[count][Z], &M.m[count][W]);
    Ptr->SetMatrix (M);

    fscanf (fp, "%s\n", FileSpec);
    PRO1 ("      PartSpec: %s\n", FileSpec)
    sprintf (FileName, "%s/PARTS/%s", DataDir, FileSpec);
    PRO1 ("      File:     %s\n", FileName)
    TOOLGetFileName (FileSpec, PartName, PATH_MAX);
    PRO1 ("      PartName: %s\n", PartName)

    Value[PX] = Value[PY] = Value[PZ] = 0.0;
    Value[RX] = Value[RY] = Value[RZ] = 0.0;
    fscanf (fp, "%d", &DOF2);      // 0 | #DOF(Joint) 5 Trans:13-15/Rot:16-18)
    Axes = 0;
    if (DOF2 > 0)
    {
      for (count = 0; count < DOF2; ++count)
      {
	fscanf (fp, "%d", &IVal);
	IVal = abs (IVal);   // TODO: Vorzeichen gibt Dreh-/Trans-Richtung an !
	if (IVal > 12)
	{
	  if (Axes == 0)
	    Value[IVal-13] = Attr[index].Val[5];
	  Axes += (1 << (IVal - 13));
	}
      }
      PRO1 ("AXES: 0x%02x\n", Axes)
      Ptr->GetJoint().SetAxes (Axes);
    }
    ReadLine (fp, FILE_TRACE);          // rest of line

    Ptr->GetJoint().SetRange (Axes, Attr[index].Min, Attr[index].Max);
    Ptr->SetVal    (Value);
    // Ptr->SetRange  (Attr[index].Range);
    Ptr->SetSpeed  (Attr[index].Speed);
    Ptr->SetAccel  (Attr[index].Accel);
    Ptr->SetTime   (Attr[index].Time);

    ReadLine (fp);                      // # -> JointNo(s)                 ?
    fscanf (fp, "%d\n", &IVal);         // # following lines
    if (Version > 10)
    {
      if (IVal > 0)
	for (count = 0; count < IVal; ++count)
	  ReadLine (fp);                // =-DOF(2)-dof(3)                 ?
      fscanf (fp, "%d\n", &Lines);      // # following lines
      for (count = 0; count < Lines; ++count)
	ReadLine (fp);                  // 0|1 + Offset-Frame, *, *[PartName] ?
    }

    PRO0 ("    loading part...\n")
    if ((Part = PARTLoad(FileName, PartName, FALSE)) != NULL ||
	(Part = PARTLoad(FileSpec, PartName, FALSE)) != NULL)
    {
      Ptr->SetPart (Part);              // Set part and create display list it

      PRO0 ("done\n")
      PRO_HMATRIX (M, "    T")
      PRO3 ("  Min: %.2f / Val: %.2f / Max: %.2f\n", 
	    Attr[index].Min, Attr[index].Val[5], Attr[index].Max)

      PRO0 ("  DOF: ")
      for (count = 0; count < 3; ++count)      // Translation
	if (Ptr->GetJoint().IsFree(count))
	    PRO1 ("\\T%c ", 'x'+count)
      for (count = 0; count < 3; ++count)      // Rotatorical DOF
        if (Ptr->GetJoint().IsFree(count+3))
	  PRO1 ("\\R%c ", 'x'+count)

      PRO0 ("\\\n")

      Dev->AddPart (Ptr, Prev);
    }
    else
    {
      PRO0 ("\\FAILED !\n")
      printf ("Loading part '%s' failed\n", FileName);
    }

    Prev = Ptr;

    // Read hierachy-informations: 0 previous part, 1 next part (chain)
    // e.g. 1 following part is chain / 01 next part is chain of previous part
    letter = fgetc (fp);
    while (letter != '\n')
    {
      if (letter == '0' && Prev != NULL)
	Prev = Prev->GetPrev ();

      if (Prev == NULL)
      {
	fscanf (fp, "%d", &IVal);
	PRO1 ("    DEVICE-NO: %d\n", IVal)

        letter = '\n';
	ReadLine (fp);
      }

      if (letter != '\n')
	letter = fgetc (fp);
    }
    if (index < DOF)
      index++;
  } while (Prev != NULL);

  for (count = 0; count < DOF+2; ++count)
    ReadLine (fp);
  ReadLine (fp);                        // 0 (End)

  fscanf (fp, "%d\n", &IVal);
  for (count = 0; count < IVal; ++count)
    ReadLine (fp);                      // 0                             ?
  ReadLine (fp);                        // 0 (End)                       ?

  delete Attr;

  PRO_END
  return Dev; 
}


static int Read12 (FILE *fp, WORLD *World, const char *DataPath)
{
  int       count = 0, Version;
  int       Value, DevNo;
  int       CommentNo;
  double    Double;
  double    fg[3], bg[3];
  char      letter = ' ';
  char      DevFile[PATH_MAX+1];
  H_MATRIX  COI;
  DEVICE3D *Dev = NULL;
  FLOOR    *Floor;
  WINDOW3D *Win;

  PRO_METHOD ("WORLD", "\n")

  Win   = World->GetWin   ();
  Floor = World->GetFloor ();

  //  ReadLine (fp);                    // Vendor-Name
  ReadLine (fp);                        // User
  ReadLine (fp);                        // Time
  ReadLine (fp);                        // Date

  fscanf (fp, "%d\n", &CommentNo);      // # Comment lines
  for (count = 0; count < CommentNo; ++count)
    ReadLine (fp);

  for (count = 0; count < 4; ++count)
    if (fscanf(fp, "%lf %lf %lf %lf\n", 
	           &COI.m[count][X], &COI.m[count][Y], 
	           &COI.m[count][Z], &COI.m[count][W]) < 4)
      PRO_RETURN ("Missing COI", 0)
  World->SetCamera (GEOGetRPY(COI));

#if 1
  ReadLine (fp);
#else
  if (fscanf(fp, "%lf %lf %lf %lf %lf %lf\n",
	         &World.Proj.px, &World.Proj.py, &World.Proj.pz,
	         &World.Proj.y,  &World.Proj.p,  &World.Proj.r) < 6)
    PRO_RETURN ("Missing Frustum", 0)
#endif

  ReadLine (fp);                        // ?
  ReadLine (fp);                        // ?

  if (fscanf(fp, "%lf %lf %lf\n", &bg[RED], &bg[GREEN], &bg[BLUE]) < 3)
    PRO_RETURN ("Missing background color", 0)
  Win->SetBackground (COLOR(bg[RED], bg[GREEN], bg[BLUE]));

  if (fscanf(fp, "%lf %lf %lf\n", &fg[RED], &fg[GREEN], &fg[BLUE]) < 3)
    PRO_RETURN("Missing foreground color", 0)
  Floor->SetMaterial (COLOR(fg[RED], fg[GREEN], fg[BLUE]));

  // Solid/Wireframe:
  fscanf (fp, "%d ", &Value);     // [0]
  if (Value == 0)
    fscanf (fp, "%d ", &Value);   // -1
  fscanf (fp, "%d ", &Value);     // 3: Wireframe, 6: Solid
  Win->SetPolygonMode ((Value > 0)? WINDOW3D::FACES : WINDOW3D::WIRE_FRAME);
  ReadLine (fp);

  // Grid:
  fscanf (fp, "%d ", &Value);
  if (Value == 0)
    fscanf (fp, "%d ", &Value);
  Floor->SetMode ((Value == 0) ? FLOOR::OFF : FLOOR::ON);
  fscanf (fp, "%d ",  &Value);
  Floor->SetSize (Value, Value);
  fscanf (fp, "%lf ", &Double);
  Floor->SetSpacing (Double, Double);

  while (! feof(fp) && letter != '*')
    letter = fgetc (fp);
  ReadLine (fp);
  ReadLine (fp);

  if (fscanf(fp, "%d\n", &DevNo) < 1)
    PRO_RETURN ("Missing Device-Amount", 0)

  PRO1 ("DevNo: %d\n", DevNo)

  ReadLine (fp);

  for (count = 0; count < DevNo; ++count)
  {
    PRO1 ("Device %d\n", count+1)
    fscanf (fp, "%s\n", DevFile);       // Dev-FileName

    fscanf (fp, "%d\n", &Version);
    PRO1 ("DEV-VERSION: %d\n", Version)
    if (Version < 13)
      Dev = ReadDev (fp, World, Version, DataPath);

    if (Dev != NULL)
      World->AddDevice (Dev);

    ReadLine (fp);                      // 0
  }

  PRO_END
  return 1;
}


// Insert all parts of a device into a global list
static void MakePartList (PART3D *PartList, PART3D **List, int &Index)
{
  PART3D *Part, *Chain;

  Part = PartList;
  while (Part != NULL)
  {
    PRO2 ("PartList[%d] = '%s'\n", Index, Part->GetPart()->PartName)
    List[Index++] = Part;
    if ((Chain = Part->GetChain()) != NULL)
      MakePartList (Chain, List, Index);
    Part = Part->GetNext ();
  }
}


int Read14 (FILE *fp, WORLD *World, const char *DataPath)
{
  int        Result = 1;
  int        count, Version;
  int        IVal, Render;
  int        No = 0, ListNo;
  int        PartNo, DevNo;
  int        CommentNo;
  double     DVal;
  double     fg[3], bg[3];
  char       String[100];
  char       DevFile[PATH_MAX+1];
  H_MATRIX   COI;
  PART3D   **PartList = NULL;
  DEVICE3D  *Dev = NULL;
  DEVICE3D  *Last = NULL, *Next;
  FLOOR     *Floor;
  WINDOW3D  *Win;

  PRO_METHOD ("WORLD", "\n")

  Win   = World->GetWin   ();
  Floor = World->GetFloor ();

  ReadLine (fp);                        // DENEB_WORKCELL_FILE
  ReadLine (fp);                        // User
  ReadLine (fp);                        // Time
  ReadLine (fp);                        // Date

  fscanf (fp, "%d\n", &CommentNo);      // # Comment lines
  for (count = 0; count < CommentNo; ++count)
    ReadLine (fp);

  ReadLine (fp);                        // Light: hor. [RAD] Elevation [RAD]

  // Background-Color (RGB):
  if (fscanf(fp, "%lf %lf %lf\n", 
	         &bg[RED], &bg[GREEN], &bg[BLUE]) < 3)
    PRO_RETURN ("Invalid BackGround-Color", -1)
  Win->SetBackground (COLOR(bg[RED], bg[GREEN], bg[BLUE]));

  PRO3 ("BG-Color: %.2f %.2f %.2f\n", bg[RED], bg[GREEN], bg[BLUE])

  // Grid-Color (RGB):
  if (fscanf(fp, "%lf %lf %lf\n", &fg[RED], &fg[GREEN], &fg[BLUE]) < 3)
    PRO_RETURN ("Invalid Grid-Color", -2)

  Floor->SetMaterial (COLOR(fg[RED], fg[GREEN], fg[BLUE]));
  PRO3 ("FG-Color: %.2f %.2f %.2f\n", fg[RED], fg[GREEN], fg[BLUE])

  // Render-Mode
  fscanf (fp, "%d ", &IVal);            // -1
  fscanf (fp, "%d ", &Render);          // Wire-Frame/Solid
  ReadLine (fp);                        // 0 1                             ?
  Win->SetPolygonMode ((Render>0)? WINDOW3D::FACES : WINDOW3D::WIRE_FRAME);
  PRO1 ("Rendering: %s\n", (Render>0)? "Solid":"Wire-Frame")

  // Grid
  fscanf (fp, "%d ",  &IVal);           // Grid: Mode, Number, Spacing
  Floor->SetMode ((IVal==0)? FLOOR::OFF : FLOOR::ON);
  fscanf (fp, "%d ",  &IVal);
  Floor->SetSize (IVal, IVal);
  fscanf (fp, "%lf ", &DVal);
  Floor->SetSpacing (DVal, DVal);
  PRO5 ("Grid:      %d  %dx%d  [%.2f][%.2f]\n", int(Floor->GetMode()),
	Floor->GetSizeX(),  Floor->GetSizeY(), 
	Floor->GetSpaceX(), Floor->GetSpaceY())

  ReadLine (fp);                        // 0 0                             ?
  ReadLine (fp);                        // 1                               ?
  ReadLine (fp);                        // 0 0 -9810                       ?
  ReadLine (fp);                        // f 1           (f=0.1, 1)        ?

  ReadLine (fp);                        // *

  ReadLine (fp);                       // 1                               ?
  fscanf (fp, "%d\n", &PartNo);        // # Parts
  PRO1 ("PartNo:    %d\n", PartNo)

  ReadLine (fp);                       // 0                               ?
  ReadLine (fp);                       // 0                               ?
  fscanf (fp, "%d\n", &DevNo);         // # independent devices
  PRO1 ("DeviceNo:  %d\n", DevNo)

  ReadLine (fp);                       // **BEGIN_DEVICES**

  // Create list for all parts needed for attachments
  PartList = new PART3D *[PartNo];

  ListNo = World->GetPartNum ();
  for (count = 0; count < DevNo; ++count)
  {
    PRO1 ("  [DEVICE %d:]\n", count+1)
    fscanf (fp, "%s\n", DevFile);       // Device-FileName

    fscanf (fp, "%d\n", &Version);
    PRO1 ("DEV-VERSION: %d\n", Version)
    if (Version < 13)
      Dev = ReadDev (fp, World, Version, DataPath);

    if (Dev != NULL)
    {
      PRO0 ("Adding Device\n")
      World->AddDevice (Dev);

      // insert all device's parts in the global list
      MakePartList (Dev->GetPartList(), PartList, No);
    }
    else
      PRO0 ("  !!! Loading failed !!!\n")
  }

#if 0
CalcDHP ();
#endif

  // Reorder Attachment-List
  PRO0 ("Attach:\n")
  Dev = World->GetData ();
  Last = NULL;
  for (count = 0; count < DevNo; ++count)
  {
    Next = Dev->GetNext ();
    fscanf (fp, "%d\n", &IVal);        // Attached Device -1: base, >= 0 CS
    if (IVal < 0)
    {
      PRO2 ("%d: Base-Dev: '%s'\n", count, Dev->GetName())
      Last = Dev;
    }
    else
    {
      PRO3 ("%d: Sub-Dev: '%s' -> %d\n", count, Dev->GetName(), IVal)
      Dev->SetNext (NULL);
      PartList[IVal]->AttachDev (Dev);
      if (Last == NULL)
	World->SetData (Next);
      else
	Last->SetNext (Next);
    }

    Dev = Next;
  }

  // free part list
  delete PartList;

  do
    fscanf (fp, "%s\n", String);
  while (! feof(fp) && strcmp(String, "**END_DEVICES**") != 0);

  // Paths:
  ReadLine (fp);                        // ** BEGIN_PATHS**
  ReadLine (fp);                        // 0 or Paths
  do 
    fscanf (fp, "%s\n", String);
  while (! feof(fp) && strcmp(String, "**END_PATHS**") != 0);

  // IO:
  ReadLine (fp);                        // **BEGIN_IO**
  do 
    fscanf (fp, "%s\n", String);
  while (! feof(fp) && strcmp(String, "**END_IO**") != 0);

  // Collision Queue:
  ReadLine (fp);                        // **BEGIN_COLL_Q**
  do 
    fscanf (fp, "%s\n", String);
  while (! feof(fp) && strcmp(String, "**END_COLL_Q**") != 0);

  // D-Lists: ???
  ReadLine (fp);                        // **BEGIN_DLISTS**
  do 
    fscanf (fp, "%s\n", String);
  while (! feof(fp) && strcmp(String, "**END_DLISTS**") != 0);

  // CMM-Info: ???
  ReadLine (fp);                        // **BEGIN_CMM_INFO**
  do 
    fscanf (fp, "%s\n", String);
  while (! feof(fp) && strcmp(String, "**END_CMM_INFO**") != 0);

  // Chains: ???
  ReadLine (fp);                        // **BEGIN_CHAINS**
  do 
    fscanf (fp, "%s\n", String);
  while (! feof(fp) && strcmp(String, "**END_CHAINS**") != 0);

  // Light sources:
  do                                    // **BEGIN_LIGHTS**
    fscanf (fp, "%s\n", String);
  while (! feof(fp) && strcmp(String, "**BEGIN_LIGHTS**") != 0);
  fscanf (fp, "%d", &IVal);
  for (count = 0; count < IVal; ++count)
  {
    ReadLine (fp);                      // d d f f f                     ?
    ReadLine (fp);                      // f f f f f f                   ?
    ReadLine (fp);                      // f f f f f d                   ?
    ReadLine (fp);                      // 0 0 0 0 -1                    ?
  }
  ReadLine (fp);                        // 0 -d f f f f                  ?
  do 
    fscanf (fp, "%s\n", String);
  while (! feof(fp) && strcmp(String, "**END_LIGHTS**") != 0);

  // Render information:
  ReadLine (fp);                        // **BEGIN_RENDER**
  ReadLine (fp);                        // current_view                  ?
  ReadLine (fp);                        // COI(z)                        ?
  for (count = 0; count < 4; ++count)   // Camera (Frame)                ?
    fscanf (fp, "%lf %lf %lf %lf\n", &COI.m[count][X], &COI.m[count][Y], 
	                             &COI.m[count][Z], &COI.m[count][W]);
#if 0
  GEORPY (COI, &World.View[0].Camera.px,  &World.View[0].Camera.py,
	       &World.View[0].Camera.pz,  &World.View[0].Camera.r,
	       &World.View[0].Camera.p,   &World.View[0].Camera.y);
#endif
  ReadLine (fp);                        // GVL[mm] f f COI(x,y,z) 42     ?
  ReadLine (fp);                        // -1 -1                         ?
  for (count = 0; count < 2; ++count)   // 2 Frames
  {
    ReadLine (fp);                      // Normal
    ReadLine (fp);                      // Open
    ReadLine (fp);                      // Approach
    ReadLine (fp);                      // Position
  }
  ReadLine (fp);                        // 1 1.5 60                      ?
  ReadLine (fp);                        // 1                             ?
  ReadLine (fp);                        // 0                             ?
  ReadLine (fp);                        // 0                             ?
  ReadLine (fp);                        // 0                             ?
  ReadLine (fp);                        // 0 0 0 0 0                     ?
  ReadLine (fp);                        // 20 8000 3 1e+08 0.523599 32   ?
  ReadLine (fp);                        // 20 3000 20 3000 1 20000       ?
  ReadLine (fp);                        // 200 10000 50 25000 3 500      ?
  ReadLine (fp);                        // 1|2                           ?
  ReadLine (fp);                        // 1                             ?
  ReadLine (fp);                        // 0 3 50000                     ?
                                        // [0 10 1.5 1.5 100]            ?
  do 
    fscanf (fp, "%s\n", String);
  while (! feof(fp) && strcmp(String, "**END_RENDER**") != 0);

  PRO_END
  return Result;
}


static bool WriteAttached (FILE *fp, PART3D *Part, int &ListNo)
{
  PART3D   *Ptr, *Chain;
  DEVICE3D *Attach;

  Ptr = Part;
  while (Ptr != NULL)
  {
    Attach = Ptr->GetAttach ();
    while (Attach != NULL)
    {
      //      Attach->WriteIGRIP (fp, ListNo);
      WriteAttached (fp, Attach->GetPartList(), ListNo);
      Attach = Attach->GetNext ();
    }

    if ((Chain = Ptr->GetChain()) != NULL)
      WriteAttached (fp, Chain, ListNo);

    Ptr = Ptr->GetNext ();
  }

  return TRUE;
}


static bool WriteFixed (FILE *fp, PART3D *Part, int &ListNo)
{
  int       No;
  PART3D   *Ptr, *Chain;
  DEVICE3D *Attach;

  Ptr = Part;
  while (Ptr != NULL)
  {
    No = ListNo;
    ListNo++;

    Attach = Ptr->GetAttach ();
    while (Attach != NULL)
    {
      fprintf (fp, "%d\n", No);
      WriteFixed (fp, Attach->GetPartList(), ListNo);
      Attach = Attach->GetNext();
    }

    if ((Chain = Ptr->GetChain()) != NULL)
      WriteFixed (fp, Chain, ListNo);

    Ptr = Ptr->GetNext ();
  }

  return TRUE;
}


WORLD *Load (const char *FileSpec, const char *DataPath)
{
  return ImportWorld (NULL, FileSpec, DataPath);
}


bool Save (WORLD *World, const char *FileSpec)
{
  FILE *fp = NULL;

  PRO_METHOD ("WORLD", "")
  
  if ((fp = fopen(FileSpec, "w+")) == NULL)
    PRO_RETURN1 ("Error opening file: %s", FileSpec, FALSE)

  return WriteWorld (fp, World);
}


PART3D *ImportPart (WORLD *World, const char *FileSpec)
{
  
  return NULL;
}


DEVICE3D *ImportDevice (WORLD *World, 
			const char *FileSpec, const char *DataPath)
{
  int       Version;
  FILE     *fp = NULL;
  DEVICE3D *Dev = NULL;
  WINDOW3D *Win;

  PRO_METHOD ("WORLD", "\n")

  assert (World != NULL);

  if ((fp = fopen(FileSpec, "r")) == NULL)
    PRO_RETURN ("File-Error", NULL)

  Win = World->GetWin ();
  Win->Select ();
  fscanf (fp, "%d\n", &Version);
  PRO1 ("DEV-VERSION: %d\n", Version)
  if (Version < 13)
    Dev = ReadDev (fp, World, Version, DataPath);
  else
    PRO1 ("Invalid Device-Version: %d\n", Version)

  if (Dev != NULL)
    World->AddDevice (Dev);

  fclose (fp);

#ifdef FILTER
  if (AlphaNum > 0)
    Win->SetBlending (WINDOW3D::TRANSPARENT);
#endif

  World->Fill ();

  PRO_END
  return Dev;
}


WORLD *ImportWorld (WORLD *World, const char *FileSpec, const char *DataPath)
{
  int    Version;
  int    State = -1;
  char   FileName[PATH_MAX+1];
  FILE  *fp;
  WORLD *NewWorld = NULL;

  if ((fp = fopen(FileSpec, "r")) == NULL)
    PRO_RETURN1 ("Error opening file: %s", FileSpec, NULL)

  if (fscanf(fp, "%d\n", &Version) < 1)
    PRO_RETURN1 ("Not an world-file: %s", FileSpec, NULL)

  if (World == NULL)
  {
    TOOLGetFileName (FileSpec, FileName, PATH_MAX);
    NewWorld = World = new WORKCELL (FileName);
  }

  PRO1 ("Loading IGRIP-Data Version %d...\n", Version)  
  switch (Version)
  {
    case 12 :
      State = Read12 (fp, World, DataPath);
      break;
    case 14 :
      State = Read14 (fp, World, DataPath);
      break;
    default :
      PRO1 ("Unsupported File-Version : %d\n", Version)
   }

  fclose (fp);

  if (State < 0)
  {
    if (NewWorld != NULL)
      delete NewWorld;
    return NULL;
  }

  return World;
}


bool ExportPart (PART3D *Part, const char *FileSpec)
{
  return TRUE;
}


bool ExportDevice (DEVICE3D *Dev, const char *FileSpec)
{
  return TRUE;
}


bool WritePart (FILE *fp, PART3D *Part, int &ListNo, int &Depth)
{
  int       count, Axes;
  int       NewDepth;
  H_MATRIX  M;
  PART3D   *Chain;

  count = IGRIP_COLORS;
  while (count >= 0 && Part->GetMaterial()->GetColorID() != IGRIPColors[count])
    count--;
  fprintf (fp, "%2d", count);
  fprintf (fp, " 1 1 26214%d 0\n", 
	       (Part->GetMaterial()->GetAlpha() < 1.0)? 2:5);
  fprintf (fp, "-1 1\n");
  fprintf (fp, "1 1 1\n");
  fprintf (fp, "-1 -1 -1\n");
  fprintf (fp, "-1 -1 -1\n");
  M = Part->GetMatrix ();
  for (count = 0; count < 4; ++count)
    fprintf (fp, "%.4f %.4f %.4f %.4f\n",
	     M.m[count][X], M.m[count][Y], M.m[count][Z], M.m[count][W]);
  fprintf (fp, "%s\n", Part->GetPart()->FileName);

  // Write axes
  Axes = Part->GetJoint().GetAxes ();
  if (Part->GetParent() != NULL && Axes != 0)
  {
    fprintf (fp, "2 5");
    for (count = POS_X; count <= ROT_Z; ++count)
      if (Axes & (1<<count))
	fprintf (fp, " %d", 13+count);
    fprintf (fp, "\n");
    fprintf (fp, "1 %d\n", Depth);
  }
  else
  {
    fprintf (fp, "0\n");
    fprintf (fp, "0\n");
  }
  fprintf (fp, "0\n");
  fprintf (fp, "6\n");
  fprintf (fp, "0\n");
  fprintf (fp, "*\n");
  fprintf (fp, "*\n");
  fprintf (fp, "*\n");
  fprintf (fp, "*\n");
  fprintf (fp, "*%s\n", Part->GetPart()->PartName);

  ListNo++;

  // Write hierachical informations
  if ((Chain = Part->GetChain()) != NULL)
  {
    fprintf (fp, "1\n");
    NewDepth = Depth+1;
    WritePart (fp, Chain, ListNo, NewDepth);
  }

  if (Part->GetNext() != NULL)
  {
    fprintf (fp, "01\n");
    WritePart (fp, Part->GetNext(), ListNo, Depth);
  }
  else
    fprintf (fp, "0");

  return TRUE;
}


bool WriteDevice (FILE *fp, DEVICE3D *Dev, int &ListNo)
{
  int     count, index;
  int     DOF, No;
  int     Axes;
  PART3D *Ptr;
  JOINT   Joint (0);

  PRO_METHOD ("WORLD", "\n")
  
  if (fp == NULL)
    fp = stdout;

  fprintf (fp, "no_dev_fname\n");
  fprintf (fp, "12\n");
  fprintf (fp, "%s\n", Dev->GetName());

  fprintf (fp, "1\n");      // TODO: Kinematic 
  fprintf (fp, "16\n");
  fprintf (fp, "-1\n");
  
  DOF = Dev->GetDOF ();
  DOF = 6;
  fprintf (fp, "%d\n", DOF);

  fprintf (fp, "0\n");
  fprintf (fp, "0\n");
  fprintf (fp, "0\n");
  fprintf (fp, "1\n");
  fprintf (fp, "0\n");
  fprintf (fp, "0\n");

  fprintf (fp, "NOPROGRAM\n");
  fprintf (fp, "0\n");
  fprintf (fp, "%d\n", Dev->GetHomeNo());
  fprintf (fp, "1\n");                // TODO: Coord-CS1
  fprintf (fp, "1\n");                // TODO: Coord-CS2
  fprintf (fp, "%d\n", Dev->GetConfigNo());

  fprintf (fp, "UNKNOWN\n");
  fprintf (fp, "0\n");
  fprintf (fp, "0\n");
  fprintf (fp, "1500 1500 5000\n");  // TCP-Speed [mm/sec]
  fprintf (fp, "360 360 360\n");     // TCP-Accel [mm/sec^2]
  fprintf (fp, "720 720 720\n");     // Angular-Speed [deg/sec]
  fprintf (fp, "0.5 0.5 0.5\n");     // Angular-Accel [deg/sec^2]
  fprintf (fp, "0 0\n");

  fprintf (fp, "1 1 1 1 0.1 0.5 0\n");
  fprintf (fp, "0 1 0 0 0 2 0 1\n");

// TODO: spezielle Kinematik bercksichtigen 
  if (strcmp(Dev->GetName(), "hexa") == 0)
    fprintf (fp, "kin_hexa\n");
  else
    fprintf (fp, "kin_generic\n");
  fprintf (fp, "none\n");
  fprintf (fp, "none\n");
  count = 0; 
  Ptr = Dev->GetPartList ();
  while (Ptr != NULL)
  {
    count++;
    Ptr = Ptr->GetChain();
  }
  fprintf (fp, "%d\n", count-1);
  if (Dev->GetConfigNo() > 0)
  {
    for (count = 0; count < Dev->GetConfigNo(); ++count)
      fprintf (fp, "1%c", (count < Dev->GetConfigNo()-1)? ' ':'\n');
    for (count = 1; count <= Dev->GetConfigNo(); ++count)
      fprintf (fp, "Config_%d\n", count);
  }

  fprintf (fp, "%d\n", (Dev->GetPartList()->GetChain()!=NULL)?1:0); // dyn. libs

  fprintf (fp, "mot_joint_0\n");
  fprintf (fp, "mot_straight_0\n");
  fprintf (fp, "mot_circular_0\n");
  fprintf (fp, "mot_slew_0\n");
  fprintf (fp, "mot_aux1_0\n");
  fprintf (fp, "mot_aux2_0\n");

  // Joint-types
  for (count = 0; count < DOF; ++count)
    fprintf (fp, "0%c", (count < DOF-1)? ' ':'\n');

  // Kinematic-types (0: rot/ 1: trans)
  for (count = 0; count < DOF; ++count)
    fprintf (fp, "0%c", (count < DOF-1)? ' ':'\n');

  // Joint-Values
  Ptr = Dev->GetPartList ();
  if (Ptr->GetChain() != NULL)
  {
    while (Ptr != NULL)
    {
      Axes = Ptr->GetJoint().GetAxes ();
      for (count = POS_X; count <= ROT_Z; ++count)
	if (Axes & (1<<count))
	  fprintf (fp, "%.5f ", Ptr->GetVal(count));
      Ptr = Ptr->GetChain ();
    }
    fprintf (fp, "\n");
  }
  else
    for (count = 0; count < DOF; ++count)
      fprintf (fp, "0%c", (count < DOF-1)? ' ':'\n');

  fprintf (fp, "1\n");

  // Home positions
  for (count = 0; count < Dev->GetHomeNo(); ++count)
    DOFn (fp, DOF, 0)

  for (count = 1; count <= Dev->GetDOF(); ++count)
    fprintf (fp, "=dof(%d)\n", count);
    
  // Joint-range
  for (count = 0; count < DOF; ++count)
    fprintf (fp, "-99%c", (count < DOF-1)? ' ':'\n');

#if 0
  Ptr = GetPartList();
  while (Ptr->GetChain() != NULL)
  {
    Val = Ptr->GetAxes ();
    count = 0;
    while (! (Val & 1))
    {
      Val = Val >> 1;
      count++;
    }
    fprintf (fp, "%d", 13+count);
    Ptr = Ptr->GetChain ();
    fprintf (fp, "%c", (Ptr->GetChain() == NULL)? '\n':' ');
  }
#else
  for (count = 0; count < DOF; ++count)
    fprintf (fp, "0%c", (count < DOF-1)? ' ':'\n');
#endif    

  DOFn (fp, DOF, 0)
  DOFn (fp, DOF, 0)

  fprintf (fp, "1 1 1\n");

  // TODO: TCP/Mount-Offset
  fprintf (fp, "1 0 0 0\n");
  fprintf (fp, "0 1 0 0\n");
  fprintf (fp, "0 0 1 0\n");
  fprintf (fp, "0 0 0 1\n");
    
  fprintf (fp, "1\n");
  for (count = 0; count < 1; ++count)  // CS1-frames
  {
    fprintf (fp, "1 0 0 0\n");
    fprintf (fp, "0 1 0 0\n");
    fprintf (fp, "0 0 1 0\n");
    fprintf (fp, "0 0 0 1\n");
  }

  fprintf (fp, "1\n");
  for (count = 0; count < 1; ++count)  // CS2-frames
  {
    fprintf (fp, "1 0 0 0\n");
    fprintf (fp, "0 1 0 0\n");
    fprintf (fp, "0 0 1 0\n");
    fprintf (fp, "0 0 0 1\n");
  }

  DOFn (fp, DOF, 0)                      // Link-length
  DOFn (fp, DOF, 0)                      // OFfset
  fprintf (fp, "0 0 0 0 0 0 0 0\n");     // Aux1 .. Aux8

  Ptr = Dev->GetPartList ();                 // Min-Value
  if (Ptr->GetChain() != NULL)
    while (Ptr != NULL)
    {
      Joint = Ptr->GetJoint ();
      Axes = Joint.GetAxes ();
      for (count = POS_X; count <= POS_Z; ++count)
	if (Axes & (1<<count))
          fprintf (fp, "%.5f\n", Joint.GetMin(count));
      for (count = ROT_X; count <= ROT_Z; ++count)
	if (Axes & (1<<count))
	  fprintf (fp, "%.5f\n", RAD(Joint.GetMin(count)));
      Ptr = Ptr->GetChain ();
    }
  else
    for (count = 0; count < DOF; ++count)
      fprintf (fp, "-1.5708\n");

  Ptr = Dev->GetPartList ();                 // Max-Value
  if (Ptr->GetChain() != NULL)
    while (Ptr != NULL)
    {
      Joint = Ptr->GetJoint ();
      Axes = Joint.GetAxes ();
      for (count = POS_X; count <= POS_Z; ++count)
	if (Axes & (1<<count))
	  fprintf (fp, "%.5f\n", Joint.GetMax(count));
      for (count = ROT_X; count <= ROT_Z; ++count)
	if (Axes & (1<<count))
	  fprintf (fp, "%.5f\n", RAD(Joint.GetMax(count)));
      Ptr = Ptr->GetChain ();
    }
  else
    for (count = 0; count < DOF; ++count)
      fprintf (fp, "1.5708\n");

  // Speed
  for (count = 0; count < DOF; ++count)
    fprintf (fp, "90%c", (count < DOF-1)? ' ':'\n');

  // Accel
  for (count = 0; count < DOF; ++count)
    fprintf (fp, "180%c", (count < DOF-1)? ' ':'\n');

  // Time
  for (count = 0; count < DOF; ++count)
    fprintf (fp, "0.5%c", (count < DOF-1)? ' ':'\n');

  DOFn (fp, DOF, 0);
  DOFn (fp, DOF, 0);

  fprintf (fp, "31\n");                 // # following lines
  fprintf (fp, "%d\n", Dev->GetMountNo());
  fprintf (fp, "0 0\n");

  for (count = 0; count < 10; ++count)
    fprintf (fp, "*\n");

  fprintf (fp, "DEFAULT\n");
  fprintf (fp, "0\n");
  fprintf (fp, "0\n");
  fprintf (fp, "0\n");
  fprintf (fp, "*\n");
  fprintf (fp, "*\n");
  fprintf (fp, "*\n");
  fprintf (fp, "*\n");

  // Home-position names
  for (count = 0; count < Dev->GetHomeNo(); ++count)
    fprintf (fp, "Home_%d\n", count+1);
  fprintf (fp, "0\n");
  for (count = 0; count < Dev->GetHomeNo(); ++count)
    fprintf (fp, "*\n");

  for (count = 1; count <= DOF; ++count)
    fprintf (fp, "Joint %d\n", count);

  fprintf (fp, "8\n");
  fprintf (fp, "0.1 0.00174533\n");
  fprintf (fp, "0 0 -1.5708 0 0 0\n");
  fprintf (fp, "0 0 0 0 1 0\n");
  fprintf (fp, "1 2 3 4 5 6\n");
  fprintf (fp, "0 0 0 0 0 0\n");
  for (count = 0; count < 3; ++count)
  {
    fprintf (fp, "1 0 0 0\n");
    fprintf (fp, "0 1 0 0\n");
    fprintf (fp, "0 0 1 0\n");
    fprintf (fp, "0 0 0 1\n");
  }
  fprintf (fp, "0 0 51.5498 0 0 0\n");
  fprintf (fp, "0 289.98 309.685 0.0124512 0 0\n");
  fprintf (fp, "7\n");                   // Number following lines

  fprintf (fp, "*\n");
  fprintf (fp, "*\n");
  fprintf (fp, "1 1 1 1 1 1\n");
  DOFn (fp, DOF, 1);
  fprintf (fp, "0 0 0 0\n");
  fprintf (fp, "*\n");
  fprintf (fp, "*\n");

  fprintf (fp, "0\n");
  fprintf (fp, "0\n");
  fprintf (fp, "no_app_file\n");
  fprintf (fp, "2\n");
  fprintf (fp, "no_par_file\n");
  fprintf (fp, "0 1 1 1\n");

  fprintf (fp, "6\n");

  fprintf (fp, "10 1 1 327684 0 0\n");
  fprintf (fp, "-1 -1\n");
  fprintf (fp, "-1 -1 -1\n");
  fprintf (fp, "-1 -1 -1\n");
  fprintf (fp, "-1 -1 -1\n");

  fprintf (fp, "1\n");      // #following lines
  fprintf (fp, "0\n");

  No = 0;
  // Write all parts of this device
  WritePart (fp, Dev->GetPartList(), ListNo, No);
  fprintf (fp, "%d\n", ListNo-1);

  fprintf (fp, "0\n");
  if (No == 1)
    for (count = 0; count <= DOF; ++count)
      fprintf (fp, "%d\n", ListNo);
  else
    for (count = 0; count <= DOF; ++count)
      fprintf (fp, "%d\n", count);
  fprintf (fp, "0\n");                   // End

  fprintf (fp, "14\n");
  for (count = 0; count < 14; ++count)
    fprintf (fp, "0\n");
  fprintf (fp, "0\n");

  PRO_END
  return TRUE;
}


bool WriteWorld (FILE *fp, WORLD *World)
{
  int       count, DevNum = 0;
  int       ListNo = 0;
  COLOR     Color;
  MATERIAL *Material;
  DEVICE3D *Dev;
  FLOOR    *Floor;
  WINDOW3D *Win;

  PRO_METHOD ("WORLD", "")

  Floor = World->GetFloor ();
  Win   = World->GetWin   ();

  fprintf (fp, "14 DENEB_WORKCELL_FILE\n");
  fprintf (fp, "User: %s\n", getenv("USER"));
  fprintf (fp, "Time: \n");
  fprintf (fp, "Date: \n");
  fprintf (fp, "0\n");                   // Comment lines

  fprintf (fp, "0.523599 0\n");          // Light: hor. [RAD] Elevation [RAD]

  // Colors (background/floor)
  Color = Win->GetBackground ();
  fprintf (fp, "%.6f %.6f %.6f\n", Color[RED], Color[GREEN], Color[BLUE]); 
  if ((Material = Floor->GetMaterial()) != NULL)
    fprintf (fp, "%.6f %.6f %.6f\n",   
	        Material->GetRed(), Material->GetGreen(), Material->GetBlue());

  // Render-Mode
  fprintf (fp, "-1 %d 0 1\n", 
	       (Win->GetPolygonMode()==WINDOW3D::FACES)? 6:0);

  // Grid
  fprintf (fp, "%d %d %.3f\n", (Floor->GetMode()==FLOOR::OFF)? 0:1, 
	       Floor->GetSizeX(), Floor->GetSpaceX());

  fprintf (fp, "0 0\n");
  fprintf (fp, "1\n");
  fprintf (fp, "0 0 -9810\n");
  fprintf (fp, "0.1 1\n");
  fprintf (fp, "*\n");

  fprintf (fp, "1\n");
  fprintf (fp, "%d\n", World->GetPartNum());
  fprintf (fp, "0\n");
  fprintf (fp, "0\n");
  //  fprintf (fp, "%d\n", GetDevNum(NEXT_MASK));
  fprintf (fp, "%d\n", World->GetDevNum(ALL_MASK));

  fprintf (fp, "**BEGIN_DEVICES**\n");

  Dev = World->GetData ();
  while (Dev != NULL)
  {
    // Dev->WriteIGRIP (fp, ListNo);           // Write device data

    WriteAttached (fp, Dev->GetPartList(), ListNo);

    Dev = Dev->GetNext ();
    DevNum++;
  }

  // Order of attachments
  ListNo = 0;
  Dev = World->GetData ();
  while (Dev != NULL)
  {
    fprintf (fp, "-1\n");

    WriteFixed (fp, Dev->GetPartList(), ListNo);

    Dev = Dev->GetNext ();
  }

  fprintf (fp, "**END_DEVICES**\n");

  fprintf (fp, "**BEGIN_PATHS**\n");
  fprintf (fp, "0\n");
  fprintf (fp, "**END_PATHS**\n");

  fprintf (fp, "**BEGIN_IO**\n");
  fprintf (fp, "0\n");
  fprintf (fp, "**END_IO**\n");

  fprintf (fp, "**BEGIN_DLISTS**\n");
  fprintf (fp, "0\n");
  fprintf (fp, "**END_DLISTS**\n");

  fprintf (fp, "**BEGIN_CMM_INFO**\n");
  fprintf (fp, "0\n");
  fprintf (fp, "**END_CMM_INFO**\n");
  
  fprintf (fp, "**BEGIN_CHAINS**\n");
  fprintf (fp, "0\n");
  fprintf (fp, "**END_CHAINS**\n");

  fprintf (fp, "**BEGIN_LIGHTS**\n");
  fprintf (fp, "1\n");
  for (count = 0; count < 1; ++count)
  {
    fprintf (fp, "1 2 0.50000 0.500000 0.50000\n");
    fprintf (fp, "0.000000 0.000000 0.000000 0.000000 0.000000 0.000000\n");
    fprintf (fp, "0.000000 0.000000 0.500000 1.000000 90.000000 1\n");
    fprintf (fp, "0 0 0 0 -1\n");
  }
  fprintf (fp, "0 -15 1.000000 1.000000 1.000000 1.000000\n");
  fprintf (fp, "**END_IO**\n");

  fprintf (fp, "**BEGIN_RENDER**\n");
  fprintf (fp, "current_view\n");
  fprintf (fp, "215.318\n");                           // COI(z)
  fprintf (fp, "0.7222274 -0.315768 0.615314 0\n");    // Camera-frame 
  fprintf (fp, "-1.49481e-06 0.889686 0.456573 0\n"); 
  fprintf (fp, "-0.691607 -0.329771 0.642597 0\n");
  fprintf (fp, "103.154 -23.6755 294.085 1\n");
  fprintf (fp, "1460.44 411.634 4192.21 0 0.804202 177.416 42\n"); // GVL
  fprintf (fp, "-1 -1\n");

  fprintf (fp, "1 0 0 0\n");
  fprintf (fp, "0 1 0 0\n");
  fprintf (fp, "0 0 1 0\n");
  fprintf (fp, "0 0 0 1\n");
  fprintf (fp, "1 0 0 0\n");
  fprintf (fp, "0 1 0 0\n");
  fprintf (fp, "0 0 1 0\n");
  fprintf (fp, "0 0 0 1\n");

  fprintf (fp, "1 1.5 60\n");
  fprintf (fp, "1\n");
  fprintf (fp, "0\n");
  fprintf (fp, "0\n");
  fprintf (fp, "0\n");
  fprintf (fp, "0 0 0 0 0\n");
  fprintf (fp, "20 8000 3 1e+08 0.523599 32\n");
  fprintf (fp, "20 3000 20 3000 1 20000\n");
  fprintf (fp, "200 10000 50 25000 3 500\n");
  fprintf (fp, "2\n");
  fprintf (fp, "1\n");
  fprintf (fp, "0 3 50000\n");
  fprintf (fp, "1\n");
  fprintf (fp, "0 10 1.5 1.5 100\n");
  fprintf (fp, "**END_RENDER**\n");

  fclose (fp);

  PRO_END
  return TRUE;
}


PART3D *ReadPart (FILE *fp, double Version)
{

  return NULL;
}


DEVICE3D *ReadDevice (FILE *fp, double Version, const char *DataPath)
{
  return ReadDev (fp, NULL, Version, DataPath);
}


