#include <math.h>

#include "tuxCAD.h"


SCHEDULER::SCHEDULER  (WORLD *World) :
  World    (World),
  TaskList (),
  Time     (0.0),
  MaxTime  (0.0),
  Speed    (0.1),
  status   (STOPPED),
  Event    (NULL),
  Data     (NULL)
{
}
 

SCHEDULER::~SCHEDULER ()
{
  int   count;
  TASK *Task;

#ifdef USE_STL
  for (count = TaskList.size()-1; count >= 0; count--)
  {
    Task = TaskList[count];
    delete (Task);
  }
  TaskList.clear ();
#else
  for (count = TaskList.Size()-1; count >= 0; count--)
  {
    Task = TaskList.Get (count);
    TaskList.Delete (Task);
    delete (Task);
  }
#endif
}


void SCHEDULER::AddTask (TASK *Task)
{
  double End;

  if (Task == NULL)
    return;

#ifdef USE_STL
  vector<TASK *>::iterator Ptr = TaskList.begin ();
  if (! TaskList.empty())
    while (Ptr != TaskList.end())
    {
      if (Task->GetStart() > (*Ptr)->GetStart())
      {
	TaskList.insert (Ptr, Task);
	return;
      }
      ++Ptr;
    }

  TaskList.push_back (Task);

  // Test List: 
  printf ("List\n");
  Ptr = TaskList.begin ();
  while (Ptr != TaskList.end())
  {
    printf ("  Start: %f -> %f\n", (*Ptr)->GetStart(), (*Ptr)->GetEnd());
    ++Ptr;
  }
#else
  TaskList.Add (Task);  
#endif
  if ((End = Task->GetEnd()) > MaxTime)
    MaxTime = End;
}


void SCHEDULER::DelTask (TASK *Task)
{
  if (Task == NULL)
    return;

#ifdef USE_STL
  int count;

  for (count = 0; count < (int)TaskList.size(); ++count)
    if (TaskList[count] == Task)
      TaskList.erase (TaskList.begin()+count);
#else
  TaskList.Delete (Task);
#endif
  CalcMaxTime ();
}


TASK *SCHEDULER::GetTask (int No)
{
#ifdef USE_STL
  if (No >= 0 && No < (int)TaskList.size())
    return TaskList[No];
  else
    return NULL;
#else
  if (No >= 0 && No < TaskList.Size())
    return TaskList.Get (No);
  else
    return NULL;
#endif
}


int SCHEDULER::GetTaskNo () const
{
#ifdef USE_STL
  return TaskList.size ();
#else
  return TaskList.Size ();
#endif
}


void SCHEDULER::Clear ()
{
#ifdef USE_STL
  TaskList.clear ();
#else
  TaskList.Clear ();
#endif
  MaxTime = 0;
}


void SCHEDULER::MoveTask (TASK *Task, double Dist)
{
  if (Task == NULL)
    return;

  Task->SetStart (Task->GetStart() + Dist);
  // Taskliste umsortieren ?
  CalcMaxTime ();
}


void SCHEDULER::PlaceTask (TASK *Task, double Time)
{
  if (Task == NULL)
    return;

  Task->SetStart (Time);
  // Taskliste umsortieren ?
  CalcMaxTime ();
}


bool SCHEDULER::SetTime (double Time)
{
  if (Time < 0.0 || Time > MaxTime) 
    return FALSE;

  this->Time = Time;
  return TRUE;
}


void SCHEDULER::SetSpeed (double Speed)
{
  this->Speed = Speed;
} 


void SCHEDULER::SetEvent (SCHEDULER_EVENT Event, void *Data)
{
  this->Event = Event;
  this->Data  = Data;
}


static void SCHEDULER_Step (SCHEDULER *Scheduler, XtIntervalId Timer)
{
  double       Time = Scheduler->GetTime ();
  XtIntervalId TimerId;

  if (Scheduler->GetStatus() == SCHEDULER::STOPPED || 
      ! Scheduler->Execute(Time))
    return;

  if (! Scheduler->SetTime (Time + Scheduler->GetSpeed()))
    return;

  TimerId = XtAppAddTimeOut (AppCon, 10,
			     (XtTimerCallbackProc)SCHEDULER_Step, Scheduler);
}


void SCHEDULER::Run ()
{
  PRO_METHOD ("SCHEDULER", "\n")

  status = RUNNING;
  SCHEDULER_Step (this, 0);

  PRO_END
}


void SCHEDULER::Stop ()
{
  status = STOPPED;
}


void SCHEDULER::Step ()
{
  Execute (Time + Speed);
  status = STOPPED;
}


double SCHEDULER::GetTime () const
{
  return Time;
}


double SCHEDULER::GetSpeed () const
{ 
  return Speed;
}


SCHEDULER::STATUS SCHEDULER::GetStatus () const
{
  return status;
}


double SCHEDULER::GetMaxTime () const
{
  return MaxTime;
}


bool SCHEDULER::Execute (double Time)
{
  int   count;
  TASK *Task;

  PRO_METHOD ("SCHEDULER", "\n")

  if (Event != NULL && ! Event(this, Time, Data))
    return FALSE;

  if (Time < 0.0 || Time > MaxTime)
  {
#if 0
    if (Time < 0.0) 
      Time = 0.0;
    else
      Time = MaxTime;
#endif
    status = STOPPED;
    //    SetSpeed (fabs(GetSpeed()));
    return FALSE;
  }

  SetTime (Time);
#ifdef USE_STL
  for (count = 0; count < (int)TaskList.size(); ++count)
  {
    Task = TaskList[count];
    Task->Execute (Time);
  }
#else
  for (count = 0; count < TaskList.Size(); ++count)
  {
    Task = TaskList.Get (count);
    Task->Execute (Time);
  }
#endif

  World->Redraw ();

  PRO_END
  return TRUE;
}


void SCHEDULER::CalcMaxTime ()
{
  bool  first = TRUE;
  int   count;
  TASK *Task;

  // Get finishing time of all tasks
  MaxTime = 0;
#ifdef USE_STL
  for (count = 0; count < (int)TaskList.size(); ++count)
  {
    Task = TaskList[count];
    if (! first)
      MaxTime = MAX (MaxTime, Task->GetEnd());
    else
    {
      MaxTime = Task->GetEnd ();
      first = FALSE;
    }
  }
#else
  for (count = 0; count < TaskList.Size(); ++count)
  {
    Task = TaskList.Get (count);
    if (! first)
      MaxTime = MAX (MaxTime, Task->GetEnd());
    else
    {
      MaxTime = Task->GetEnd ();
      first = FALSE;
    }
  }
#endif
}
