#include "tuxCAD.h"


TASK::TASK (const MOTION &Motion, double Start, 
	    TASK_EVENT Event /* = NULL */, void *Data /* = NULL */) :
  Motion (Motion),
  Start  (Start),
  Next   (NULL),
  Event  (Event),
  Data   (Data)
{
}


TASK::~TASK ()
{
}


bool TASK::Execute (double Time)
{
  // Return FALSE, if a user event-handler is defined which forbids the 
  // current step
  if (Event != NULL && ! Event(this, Time, Data))
    return FALSE;

  // Check execution time of this task 
  return (Start <= Time && Time <= Start + Motion.GetDuration());
}


void TASK::SetStart (double Start)
{
  this->Start = Start;
}


void TASK::SetEvent (TASK_EVENT Event, void *Data /* = NULL */)
{
  this->Event = Event;
  this->Data  = Data;
}


double TASK::GetStart () const
{
  return Start;
}


double TASK::GetEnd () const
{
  return Start + Motion.GetDuration ();
}


char *TASK::GetName () const
{
  char *Name = new char[10];

  strcpy (Name, "TASK:");
	   
  return Name;
}


int Comp (const TASK *Task1, const TASK *Task2)
{
  if (Task1->Start < Task2->Start)
    return -1;
  elsif (Task1->Start > Task2->Start)
    return 1;
  else
    return 0;
}


PART_TASK::PART_TASK (PART3D *Part, const MOTION &Motion, double Start,
		      TASK_EVENT Event /* = NULL */) :
  TASK (Motion, Start, Event),
  Part (Part)
{
}


bool PART_TASK::Execute (double Time)
{
  FRAME Step;

  if (! TASK::Execute(Time) || Part == NULL)
    return FALSE;

  Step = Motion.GetStep (Time - Start);
  Part->SetMatrix (TO_MATRIX(Step));

  return TRUE;
}
  

char *PART_TASK::GetName () const
{
  char *Name = NULL;

  if (Part != NULL)
  {
    Name = (char *)malloc (sizeof(char) * 255);
    sprintf (Name, "PART: %s:%s", 
	     Part->GetParent()->GetName(), Part->GetName());
  }
	   
  return Name;
}



DEV_TASK::DEV_TASK (DEVICE3D *Dev, const MOTION &Motion, double Start,
		    TASK_EVENT Event /* = NULL */) :
  TASK (Motion, Start, Event),
  Dev (Dev)
{
}


bool DEV_TASK::Execute (double Time)
{
  FRAME Step;

  if (! TASK::Execute(Time))
    return FALSE;

  Step = Motion.GetStep (Time - Start);
  Dev->SetWTD (TO_MATRIX(Step));

  return TRUE;
}


char *DEV_TASK::GetName () const
{
  char *Name = NULL;

  if (Dev != NULL)
  {
    Name = new char[80];
    sprintf (Name, "DEV: %s", Dev->GetName());
  }
	   
  return Name;
}



TCP_TASK::TCP_TASK (DEVICE3D *Dev, const MOTION &Motion, double Start,
		    TASK_EVENT Event /* = NULL */) :
  TASK (Motion, Start, Event),
  Dev (Dev)
{
}


bool TCP_TASK::Execute (double Time)
{
  int       count;
  double    J[18];
  FRAME     Step;
  H_MATRIX  Next;
  JOINTS    Joint;
  ROBOT    *Robot;

  if (! TASK::Execute(Time))
    return FALSE;

  if ((Robot = Dev->GetRobot()) == NULL)
    return FALSE;

  Step = Motion.GetStep (Time - Start);
  Next = TO_MATRIX (Step);

  try
  {
    Joint = Robot->InvKin (TO_FRAME(Next));
    for (count = 0; count < Joint.GetNumber(); ++count)
      J[count] = Joint[count];
    Dev->SetJoint (J, Joint.GetNumber());
  }
  catch (KINEMATIC::OutOfWorkspace OOW)
  {
    PRGPrintf ("OOW: [%x] Pos: %.2f %.2f %.2f [mm]", OOW.ErrorJoint, 
	       Next.m[P][X], Next.m[P][Y], Next.m[P][Z]);
  }

  return FALSE;
}
  

char *TCP_TASK::GetName () const
{
  char *Name = NULL;

  if (Dev != NULL)
  {
    Name = new char[80];
    sprintf (Name, "TCP: %s", Dev->GetName());
  }
	   
  return Name;
}






