#include <limits.h>
#include <unistd.h>
#include <string.h>

#include "viewGlobal.h"


/* ********************************************************************** */
/* ** Read the specification of a selected part and the clicked        ** */
/* ** position. The specifictation consists of:                        ** */
/* ** Names of the device and part, the part's filename and selected   ** */
/* ** 3d-position (x,y,z)                                              ** */
/* ********************************************************************** */
static dual VIEW_MouseSelect (FILE *fp, 
			     char *DevName, char *PartName, char *FileName,
			     double *x, double *y, double *z)
{
  double   fx, fy, fz;
  char     Name[NAME_LEN+1];
  RPY      p;
  H_MATRIX M, MInv;

  PRO_START ("\n")

  if (! VIEW_ReadString(fp, Name) ||
      ! VIEW_ReadString(fp, FileName) ||
      ! VIEW_ReadPos(fp, &fx, &fy, &fz) ||
      ! VIEW_ReadRPY(fp, &p.px, &p.py, &p.pz, &p.r, &p.p, &p.y))
    PRO_RETURN ("Error reading data", FALSE)

  VIEW_ExtractDevName  (Name, DevName);
  VIEW_ExtractPartName (Name, PartName);

  M = GEORPY2Frame (p.px, p.py, p.pz, p.y, p.p, p.r);
  MInv = GEOMatrixInv (M);

  PRO_HMATRIX (M, "RPY-PART")

  if (x != NULL)
    *x = MInv.m[N][X]*fx + MInv.m[O][X]*fy + MInv.m[A][X]*fz + MInv.m[P][X];
  if (y != NULL)
    *y = MInv.m[N][Y]*fx + MInv.m[O][Y]*fy + MInv.m[A][Y]*fz + MInv.m[P][Y];
  if (z != NULL)
    *z = MInv.m[N][Z]*fx + MInv.m[O][Z]*fy + MInv.m[A][Z]*fz + MInv.m[P][Z];

  if (x != NULL && y != NULL && z != NULL)
    PRO3 ("Pos: [%.3f][%.3f][%.3f]\n", *x, *y, *z)

  PRO_END
  return TRUE;
}


/* ********************************************************************** */
/* ** start a GSL-Macro, which invites the user to click on a part's   ** */
/* ** face. The returned values depend on the function parameters.     ** */
/* ** If the exactly position of the clicked point is not required     ** */
/* ** the user might use NULL-pointers for x, y and z.                 ** */
/* ** INPUT:  Socket    Connection to IGRIP                            ** */
/* ** INOUT:  DevName   Allocated Memory for selected device-name      ** */
/* **         PartName                       selected part-name        ** */
/* **         FileName                       filename of the part      ** */
/* **         x, y, z   Pointer to result-variables                    ** */
/* ** OUTPUT: (int)     Pressed mouse-button or error-status           ** */
/* **                     0: right mouse-button was pressed (no        ** */
/* **                        part/position returned)                   ** */
/* **                     1: part(-pos.) was selected with left        ** */
/* **                        mouse-button)                             ** */
/* **                     2: User pressed middle mouse-button          ** */
/* **                        (no part/pos returned)                    ** */
/* **                    -1: an error occured                          ** */
/* ********************************************************************** */
int VIEWMouseSelect (int SocketID, 
		     char *DevName, char *PartName, char *FileName,
		     double *x, double *y, double *z)
{
  dual    Result;
  int     Button;
  double  px, py, pz;
  char    PartFile[PATH_MAX+1];
  FILE   *fp = NULL;

  PRO_START (" ")

  if (VIEWRunCLI(SocketID, "EXECUTE '%s/GSLMACROS/face.gsl'",
		           TOOLGetConfigString(CFG_HIGHLAP_PATH)) == NULL)
    PRO_RETURN ("Error executing GSL-Macro 'face.gsl'", -1)

  if (access(TOOLGetConfigString(CFG_TMP_FILE), R_OK) != 0)
    PRO_RETURN ("File not found: Function aborted", -1)

  if ((fp = fopen(TOOLGetConfigString(CFG_TMP_FILE), "r")) == NULL)
    PRO_RETURN ("Error loading file", -1)

  if (! (Result = VIEW_ReadInt(fp, &Button)))
    PRO0 ("Error reading Button-ID")
  elsif (Button == 1)
  {
    if ((Result = 
	   VIEW_MouseSelect(fp, DevName, PartName, PartFile, &px, &py, &pz)))
    {
      if (x != NULL)
	*x = px;
      if (y != NULL)
	*y = py;
      if (z != NULL)
	*z = pz;

      if (FileName != NULL)
	strncpy (FileName, PartFile, PATH_MAX);
    }
    else
      PRO0 ("Error reading Position")
  }

  fclose (fp);
  remove (TOOLGetConfigString(CFG_TMP_FILE));

  if (! Result)
    PRO_RETURN ("Error reading position", -1)

  PRO_END
  return Button;
}


/* ********************************************************************** */
/* ** Checks for a given part and returns it's world-position (if no   ** */
/* ** reference is given) or relative position (relative to Reverence  ** */
/* ** Because there's no CLI-function to get this information, we need ** */
/* ** to start a GSL-Macro, which writes the desired information into  ** */
/* ** a temporary file.                                                ** */
/* ** INPUT:  SocketID   Connection to IGRIP (ID from c/s-lib)         ** */
/* **         DevName    Name of according device                      ** */
/* **         PartName   Interesting part-name                         ** */
/* **         Reference  Do we need the position relative to a         ** */
/* **                    other object ? -> Name or NULL                ** */
/* ** INOUT:  M          Pointer to target matrix                      ** */
/* ** OUTPUT: (dual)     TRUE: fetching position was successfull, the  ** */
/* **                          values in M are valid                   ** */
/* **                    FALSE on error, M is invalid (not changed)    ** */
/* ********************************************************************** */
dual VIEWGetPos (int SocketID, 
		char *DevName, char *PartName, char *Reference, H_MATRIX *M)
{
  dual    Result;
  double  px, py, pz, r, p, y; 
  char    Name[NAME_LEN+1];
  FILE   *fp = NULL;

  PRO_START ("\n")

  if (M)
    *M = Matrix_1;

  if ((fp = fopen(TOOLGetConfigString(CFG_TMP_FILE), "w+")) == NULL)
    PRO_RETURN ("Error while creating temp-file", FALSE)

  if (PartName == NULL)
    fprintf (fp, "%s\n", DevName);
  else
    fprintf (fp, "%s:%s\n", DevName, PartName);

  if (Reference == NULL)
  {
    fclose (fp);
    if (VIEWRunCLI(SocketID, "EXECUTE '%s/GSLMACROS/WorldPos.gsl'",
		            TOOLGetConfigString(CFG_HIGHLAP_PATH)) == NULL)
      PRO_RETURN ("Error executing GSL-Macro 'WorldPos.gsl'", FALSE)
  }
  else
  {
    fprintf (fp, "%s\n", Reference);
    fclose (fp);
    if (VIEWRunCLI(SocketID, "EXECUTE '%s/GSLMACROS/GetPos.gsl'",
		             TOOLGetConfigString(CFG_HIGHLAP_PATH)) == NULL)
      PRO_RETURN ("Error executing GSL-Macro 'GetPos.gsl'", FALSE)
  }

  if ((fp = fopen(TOOLGetConfigString(CFG_TMP_FILE), "r")) == NULL)
    PRO_RETURN ("Error while reading temp-file", FALSE)

  Result = (VIEW_ReadString(fp, Name) &&
	    VIEW_ReadRPY(fp, &px, &py, &pz, &r, &p, &y));

  fclose (fp);
  remove (TOOLGetConfigString(CFG_TMP_FILE));

  if (! Result)
    PRO_RETURN ("Error reading positon", FALSE)

  *M = GEORPY2Frame (px, py, pz, y, p, r);

  PRO_END
  return TRUE;
}


/* ********************************************************************** */
/* ** Inquires a device's or part's position                           ** */
/* ** INPUT:  SocketID  Connection to IGRIP                            ** */
/* **         DevName   Name of the device                             ** */
/* **         PartName  Name of the part (NULL if a device is meant)   ** */
/* ** INOUT:  M         Frame of the device's/part's position          ** */
/* ** OUTPUT: (dual)    TRUE, on success, FALSE otherwise              ** */
/* ********************************************************************** */
dual VIEWInquirePos (int SocketID, char *DevName, char *PartName, H_MATRIX *M)
{
  char *Result;
  RPY   Pos;

  if (PartName == NULL)
    Result = VIEWRunCLI (SocketID, "INQUIRE '%s' POSITION", DevName);
  else
    Result = VIEWRunCLI (SocketID, 
			 "INQUIRE '%s:%s' POSITION", DevName, PartName);

  if (Result == NULL ||
      sscanf(Result, "0: %le, %le, %le, %le, %le, %le",
	             &Pos.px, &Pos.py, &Pos.pz, 
	             &Pos.y,  &Pos.p,  &Pos.r) < 6)
    return FALSE;

  if (M != NULL)
    *M = GEORPY2Frame (Pos.px, Pos.py, Pos.pz, Pos.y, Pos.p, Pos.r);

  return TRUE;
}


/* ********************************************************************* */
/* ** Inquires the world's filename from IGRIP                        ** */
/* ** INPUT:  SocketID  Socket-Identifier for IGRIP-connection        ** */
/* ** INOUT:  FileName  pointer to resulting filename                 ** */
/* ** OUTPUT: (dual)    TRUE:  Filename was read successfully,        ** */
/* **                   FALSE: An error occured while retrieving      ** */
/* ********************************************************************* */
dual VIEWGetWorkcell (int SocketID, char *FileName)
{
  FILE *fp = NULL;

  PRO_START ("\n")

  if (FileName == NULL)
    return FALSE;

  if (VIEWRunCLI(SocketID, "EXECUTE '%s/GSLMACROS/WorldFile.gsl'",
		           TOOLGetConfigString(CFG_HIGHLAP_PATH)) == NULL)
    PRO_RETURN ("Error executing GSL-Macro 'face.gsl'", FALSE)

  if (access(TOOLGetConfigString(CFG_TMP_FILE), R_OK) != 0)
    PRO_RETURN ("File not found: Function aborted", FALSE)

  if ((fp = fopen(TOOLGetConfigString(CFG_TMP_FILE), "r")) == NULL)
    PRO_RETURN ("Error loading file", FALSE)

  if (! VIEW_ReadString(fp, FileName))
    PRO_RETURN ("No data in temp-file", FALSE)

  PRO_END
  return TRUE;
}

